/* ------------------------------------------------------------------------------------------------------*/
/* - Implicit.h                                                                                          */
/* - **********                                                                                          */
/* -                                                                                                     */
/* - Implizite Darstellung von Paaren von flatterms durch die Stelle ihrer Ueberlappung.                 */
/* -                                                                                                     */
/* - 01.04.96 Arnim. Angelegt.                                                                           */
/* -                                                                                                     */
/* -                                                                                                     */
/* ------------------------------------------------------------------------------------------------------*/

#ifndef IMPLICIT_H
#define IMPLICIT_H

#include "general.h"
#include "InfoStrings.h"
#include "TermOperationen.h"
#include "TermpairRepresentation.h"
#include "RUndEVerwaltung.h"

#ifdef __cplusplus
extern "C" {
#endif


/* ------------------------------------------------------------------------------------------------------*/
/* ------- Initialize------------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

BOOLEAN IMP_InitAposteriori(InfoStringT DEF, InfoStringT USR);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_NewFlatsGenerated(void);


/* ------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR ------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
TermpairT IMP_Flat2TP (TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT, 
                       PCL_ObjektNummerT pclID);                
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL -------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
void IMP_TP2ParIntermed (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT,
                         PCL_ObjektNummerT *pclIDPtr);               
                                                                                                         
void IMP_TP2ParFinal (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT,
                      PCL_ObjektNummerT *pclIDPtr);               
                                                                                                         
                                                                                                         
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS ------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

void IMP_Par2Flat (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);
/* Destroys the termpair-representation irrecoverable */


/* ------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR -------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
TermpairT IMP_Par2TP (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, BOOLEAN,
                 PCL_ObjektNummerT pclID);     
                                                                                                         
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY --------------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
                                                                                                         
void      IMP_DestroyTP  (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);                        
/* Termpaar zerstoeren */                                                                                
                                                                                                         
void      IMP_DestroyPar (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);          
/* Termpaar und Flachterme zerstoeren */                                                                 
                                                                                                         
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------- TermVerfolgung -------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

void IMP_TermOperationBeiVerfolgung(void (* Operation) (TermT, TermT), TermpairT tp, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent);

/* ------------------------------------------------------------------------------------------------------*/
/* ------- Unambiguous Comparison -----------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

BOOLEAN IMP_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                   TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2);
/* Vom Entstehungszeitpunkt unabhaengiger Vergleich von Termpaaren.
   Hier: Eltern-Nummern, dann Ueberlappungsstellen.
*/

/* ------------------------------------------------------------------------------------------------------*/
/* ------- Statistics -----------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
void IMP_Statistics(InfoStringT IS);                                                                     
                                                                                                         
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------- Output ---------------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/
                                                                                                         
void IMP_PrintTermpair (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT, const char*);
                                                                                                         
                                                                                                         
/* ------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN --------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

#define IMP_CleanUp()

#ifdef __cplusplus
}
#endif

#endif
