/*=================================================================================================================
  =                                                                                                               =
  =  Interreduktion.h                                                                                             =
  =                                                                                                               =
  =  Interreduktion der Regel- und Gleichungsmenge                                                                =
  =  Subsumption                                                                                                  =
  =                                                                                                               =
  =  24.06.94  Arnim.                                                                                             =
  =  14.07.94  Arnim. Subsumption, dh. Durchlauf + Abspeichern aller Subgleichungen.                              =
  =  15.07.94  Arnim. Subgleichungsmenge implementiert.                                                           =
  =================================================================================================================*/
#ifndef INTERREDUKTION_H
#define INTERREDUKTION_H

#include "RUndEVerwaltung.h"
#include "TermOperationen.h"
#include "InfoStrings.h"

#ifdef __cplusplus
extern "C" {
#endif

void IR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);


BOOLEAN IR_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/* IR-RM rechts */

/* -------------------------------------------------------------------------------*/
/* --------------- Statistik -----------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_Statistik(InfoStringT str);


/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion initialisieren ---------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_InterreduktionStarten(void);
/* Vor Beginn der Interreduktion den Datenpuffer initialisieren.
   Meldung: Im entsprechenden Ausgabelevel -> 'Interreduktion Anfang'
*/

void IR_InterreduktionBeendet(void);
/* Meldung: Im entsprechenden Ausgabelevel -> 'Interreduktion Ende'*/

void IR_NeueGleichungIst(GleichungsT g);

void IR_NeueRegelIst(RegelT r);

/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Gleichungsmenge ----------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_GMSubsummierenMitGleichung(void);
/* Sucht alle von der angegebenen Gleichung subsummierten Gleichungen aus GM, entfernt und loescht sie.
   Seiteneffekt: Meldung, dass Gleichung xyz von neuer Gleichung subsummiert wird.
                 Entfernen der subsummierten Gleichung(en) aus der Gleichungsmenge.
   Ferner entsprechende Statistik anpassen.*/

void IR_GMInterredMitGleichung(void);
/* Entfernen aller Gleichungen aus der Gleichungsmenge (-> Puffer), die von der angegebenen Gleichung
   reduziert werden. Da Gleichung noch nicht in der GM, kein Ausschluss.
   Meldungen: Gleichung xyz reduziert. (Entfernung aus GM wird beim Auslesen des Puffers gemeldet)
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

void IR_GMInterredMitRegel(void);
/* Entfernen aller Gleichungen aus der Gleichungsmenge (-> Puffer), die von der angegebenen Regel
   links = rechts reduziert werden.
   Meldungen:  Gleichung xyz reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Regelmenge ---------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_RMLinksInterredMitGleichung(void);
/* Entfernen aller Regeln aus der Regelmenge (-> Puffer), die von der angegebenen Gleichung
   links reduziert werden.
   Meldungen: xyz links reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

void IR_RMRechtsInterredMitGleichung(void);
/* Reduktion aller rechten Seiten von Regeln, die mit der angegebenen Gleichung reduziert werden koennen, und
   Abschlus der NF-Bildung.
   Meldungen: Regel xyz rechts reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

void IR_RMLinksInterredMitRegel(void);
/* Entfernen aller Regeln aus der Regelmenge (-> Puffer), die von der angegebenen Regel
   links reduziert werden.(ACHTUNG: au\3er der Regel selbst).
   Meldungen: Regel xyz links reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

void IR_RMRechtsInterredMitRegel(void);
/* Reduktion aller rechten Seiten von Regeln, die mit der angegebenen Regel reduziert werden koennen, und
   Abschlus der NF-Bildung.
   Meldungen: Regel xyz rechts reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

void IR_FinalerRMrechtsDurchlauf(void);
/* Falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, muessen sie am Ende der Vervollstaendigung 
   nochmals durchlaufen werden. Dies geschieht hier.
   Vor.: IR-Intervall RMrechts ist nicht == 1, RM ist nicht leer.
*/

/* -------------------------------------------------------------------------------*/
/* --------------- Weiterbehandlung der entfernten Objekte -----------------------*/
/* -------------------------------------------------------------------------------*/

void IR_PufferAuslesen(void);
/* Auslesen des waehrend der Interreduktion gefuellten Regel- und Gleichungs-Puffers.
   Behandeln der darin befindlichen Objekte gemaess Modus.
   Meldungen: *Im entsprechenden Ausgabelevel -> 'Interreduktion Ende',
              *Beim Loeschen der entsprechenden Objekte automatisch durch die Loeschroutine,
              *Beim Einfuegen als KP von dieser Routine.
*/

/* -------------------------------------------------------------------------------*/
/* --------------- Test auf Killer-Eigenschaft -----------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Die beiden folgenden Prozeduren ueberpruefen, ob das gegebene Termpaar als 
   Regel (IR_TPRReduziertRE), dh. von links nach rechts, bzw. als Gleichung
   (IR_TPGReduziertRE) gesehen, die linke Seite einer Regel oder Gleichung
   aus RE reduziert. */

BOOLEAN IR_TPRReduziertRE(TermT Links, TermT Rechts);

BOOLEAN IR_TPEReduziertRE(TermT Links, TermT Rechts);

#ifdef __cplusplus
}
#endif

#endif

