/**********************************************************************
* Filename : MNF_CPs.h
*
* Kurzbeschreibung : MultipleNormalFormen-Bildung: Kritische Paare
*                    (Verwalten von Nachfolgermengen)
*
* Autoren : Andreas Jaeger
*
* 19.11.96: Erstellung
*
**********************************************************************/

#ifndef MNF_CPS_H
#define MNF_CPS_H

#include "general.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"
#include "TermOperationen.h"
#include "MNF.h"

#ifdef __cplusplus
extern "C" {
#endif


/* Parameter: Abbruch wegen Deckelfkt? */
typedef void (*MNF_ClassifyInitT) (BOOLEAN);
/* Parameter: Term, Irreduzibel?, links=TRUE/rechts=FALSE */
typedef void (*MNF_ClassifyTraverseT) (TermT, BOOLEAN, BOOLEAN);
/* Parameter: keine */
typedef void (*MNF_ClassifyExitT)(void);

typedef struct MNF_NFSetTS {
   struct MNF_NFSetTS     *next;
   TermT                  Term;
} MNF_NFSetT, *MNF_NFSetPtrT;

void MNF_CPCleanup (void);

void MNF_REMengenGeaendert (BOOLEAN RLeer, BOOLEAN ELeer);

/* KP-Generate */
/* ----------- */

BOOLEAN MNF_InitKPGAposteriori (InfoStringT DEF, InfoStringT USR,
                                unsigned short PermGrad, BOOLEAN useSubsumption);

/* Testen auf Zusammenfuehrbarkeit und setzen von Parametern fuer
   Classifikation bei der Generierung von CP.
   Wenn links und rechts zusammengefuehrt werden konnten, so
   wird TRUE zurueckgegeben und joinedTerm ist der zusammengefuehrte
   Term, joinedTerm muss vom Aufrufer mit TO_TermLoeschen freigegeben
   werden ! */
BOOLEAN MNF_KPGBearbeite (TermT links, TermT rechts, TermT *joinedTerm);

/* Classify */
/* -------- */

void MNF_InitClassifyAposteriori (MNF_ClassifyInitT CInit, MNF_ClassifyTraverseT CTrav, MNF_ClassifyExitT CExit);

/* KP-Select */
/* --------- */

BOOLEAN MNF_InitKPSAposteriori (InfoStringT DEF, InfoStringT USR);

/* Testen auf Zusammenfuehrbarkeit und bestimmen aller
   Nachfolger */
/* Die irreduziblen Nachfolger werden in NFLeft und NFRight zurueckgegeben,
   das Modul MNF legt diese Struktur an und gibt die Daten mit der Funktion
   MNF_DeleteNFSet wieder richtig frei!
   Wenn links und rechts zusammengefuehrt werden konnten, so
   wird TRUE zurueckgegeben und joinedTerm ist der zusammengefuehrte
   Term, joinedTerm muss vom Aufrufer mit TO_TermLoeschen freigegeben
   werden ! */
BOOLEAN MNF_KPSBearbeite (TermT links, TermT rechts, TermT *joinedTerm, MNF_NFSetT **NFLeft, MNF_NFSetT **NFRight);

void MNF_DeleteNFSet (MNF_NFSetT **NFLeft, MNF_NFSetT **NFRight);

#ifdef __cplusplus
}
#endif

#endif /* MNF_CPS_H */
