/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFBasis.h                                                                                                  */
/* - *********                                                                                                  */
/* -                                                                                                            */
/* - Enthaelt die Basis-Operationen zur Normalform-Bildung (alle Verfahren) :                                   */
/* -  * Pfadverwaltung (ueber Makros etc.)                                                                      */
/* -  * RegelAngewendet                                                                                         */
/* -  * GleichungAngewendet                                                                                     */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* - 20.06.94 Arnim. Ueberfluessige Makros 'rausgeworfen.                                                       */
/* - 08.12.95 Arnim. Reduzierbarkeitstests eingefuegt fuer Kapur-Kriterium.                                     */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------*/
/* - Generelle Konvention:                                                                                      */
/* - ---------------------                                                                                      */
/* - Prozeduren, die den Regel- bzw. Gleichungsbaum zum Matchen verwenden, duerfen nicht aufgerufen werden,     */
/* - wenn der betreffende Baum leer ist.                                                                        */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef NFBASIS_H
#define NFBASIS_H

#include "general.h"		   /* wg. Typ Boolean */
#include "TermOperationen.h"	   /* wg. Term-Typ */
#include "pcl.h"                   /* StellenT jetzt in pcl.h*/
#include "RUndEVerwaltung.h"
#include "MatchOperationen.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {KPsGen = 0, KPsIntermed, KPsSel, RegelnRechts, RegelnLinks, Gleichungen, Ziele} ObjektT;


/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Statistik --------------------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

#define AnzObjektTypen 7

extern unsigned long        AnzRedVersMitRegeln[AnzObjektTypen];   /* Reduktionsversuche mit der Regelmenge */
extern unsigned long        AnzRedErfRegeln[AnzObjektTypen];       /* Erfolgreiche Reduktionsversuche mit der Regelmenge */
extern unsigned long        AnzRedVersMitGleichgn[AnzObjektTypen]; /* Reduktionsversuche mit der Gleichungsmenge */
extern unsigned long        AnzRedErfGleichgn[AnzObjektTypen];     /* Erfolgreiche Reduktionsversuche mit der Gleichungsmenge */
extern unsigned long        AnzRedMisserfolgeICG;                  /* Matches mit Gleichungs-Menge wg. Indirekter Clashes gescheitert */
extern unsigned long        AnzRedMisserfolgeICR;                  /* Matches mit Regel-Menge wg. Indirekter Clashes gescheitert */

#define IncAnzRed(re,objektArt)                                                                 \
(((objektArt==KPsGen)||(objektArt==KPsIntermed)||(objektArt==KPsSel))?TP_IncRedCp(re):          \
 ((objektArt==RegelnRechts)?TP_IncRedRight(re):                                                 \
  ((objektArt==RegelnLinks)?TP_IncRedLeft(re):                                                  \
   ((objektArt==Gleichungen)?TP_IncRedEqu(re):                                                  \
    ((objektArt==Ziele)?TP_IncRedGoal(re):0)))))

#define NFB_STATISTIK normal_power_comp(1,0,0)
/* Alternativen: 0 fuer aus, 1 fuer an */

#if NFB_STATISTIK
#define IncAnzRedErfRegeln(objektArt,re) (AnzRedErfRegeln[objektArt]++,IncAnzRed(re,objektArt))           
#define IncAnzRedErfGleichgn(objektArt,re) (AnzRedErfGleichgn[objektArt]++,IncAnzRed(re,objektArt))
#define IncAnzRedVersMitRegeln(objektArt) (AnzRedVersMitRegeln[objektArt]++)                
#define IncAnzRedVersMitGleichgn(objektArt) (AnzRedVersMitGleichgn[objektArt]++)            
#define IncAnzRedMisserfolgeICR(bedingung) (bedingung?AnzRedMisserfolgeICR++:0)
#define IncAnzRedMisserfolgeICG(bedingung) (bedingung?AnzRedMisserfolgeICG++:0)

#else
#define IncAnzRedErfRegeln(objektArt,re)       IncAnzRed(re,objektArt)/* nix tun */
#define IncAnzRedErfGleichgn(objektArt,re)     IncAnzRed(re,objektArt)/* nix tun */ 
#define IncAnzRedVersMitRegeln(objektArt)   /* nix tun */ 
#define IncAnzRedVersMitGleichgn(objektArt) /* nix tun */ 
#define IncAnzRedMisserfolgeICR(bedingung)  /* nix tun */
#define IncAnzRedMisserfolgeICG(bedingung)  /* nix tun */

#endif


#define NFB_IC() (MO_IndirekterClashImBaum())

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Typen und Operationen zum Arbeiten auf einem Pfad ----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Der Reduktions-Pfad                                                                                                     */
/* -------------------                                                                                                     */
/*                                                                                                                         */
/* Bei diesem Pfad      handelt es sich im wesentlichen um einen Stack, der den Durchlauf durch einen Term nach einer      */
/* beliebigen Strategie organisiert.                                                                                       */
/*                                                                                                                         */
/* Neben den verschiedenen Verwaltungs-Informationen, die unten erklaert sind, besteht dieser Pfad aus einem Array, auf    */
/* dem ein Stack simuliert wird. Der am weitesten rechts stehende Eintrag, auf den die Komponente aktuelleStelle verweist, */
/* beinhaltet den gerade behandelten Teilterm. Wird ein neuer Eintrag auf den Stack gelegt, so wird die Stelligkeit des    */
/* entsprechenden Topsymbols sofort als AnzNochZuBehandelnderTeilterme eingetragen, um beim Backtracking die noch nicht    */
/* abgearbeiteten Stellen im Term leichter auffinden zu koennen. Ferner wird der EinfuegePunkt initial auf NULL gesetzt,   */
/* um eine Unterscheidung zwischen neu zu behandelnden und schon teilweise abgearbeiteten Stellen zu ermoeglichen.         */
/*                                                                                                                         */
/* Ein Eintrag auf dem Stack                                                                                               */
/* =========================                                                                                               */
/*                                                                                                                         */

/* Der folgende Datentyp erleichtert das Fuehren der Reduktions-Statistik.*/

typedef struct NFB_PfadZellenTS{                   /* Ein Eintrag im eigentlichen Stack besteht aus:                       */
   UTermT           Stelle;                        /* - der in diesem Eintrag gespeicherten Stelle (Teilterm)              */
   UTermT           EinfuegePunkt;                 /* - derjenigen Termzelle, die nach Reduktion des aktuell behandelten   */
                                                   /*   Nachfolgers ggf. destruktiv geaendert werden muss                  */
   StellenT         AnzNochZuBehandelnderTeilterme;/* - ... s.Name                                                         */
} NFB_PfadZellenT;                                                                                                       /**/

extern StellenT *PositionImAktuellenTeilterm;             /* - Stelle (also 1.2.1.1) des akt. Teilterms                           */

/*                                                                                                                         */
/*                                                                                                                         */
/* Der Pfad                                                                                                                */
/* ========                                                                                                                */
/*                                                                                                                         */
/* Der Pfad enthaelt folgende Daten:                                                                                       */
/*                                                                                                                         */
extern   StellenT        aktuelleStelle;                 /* - einen Verweis auf die aktuelle Stelle                              */
                                                                                                                         /**/
extern   StellenT        aktuelleGroessePfad;            /* - die # der aktuell eingerichteten PfadZellen -1                     */
                                                   /*   (-1, um den Test auf ueberlaufenden Stack zu erleichtern)          */
extern   NFB_PfadZellenT *NFPfad;                   /* - den eigentlichen Durchlauf-'Stack'                                 */

extern   ObjektT         objektArt;                      /* - Art des gerade durchlaufenen Termes (fuer Statistik)               */
extern   unsigned long int ObjektNummer;                 /* Nummer des behandelten Objekts                                       */

/*                                                                                                                         */
/*                                                                                                                         */
/*                                                                                                                         */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Zugriffe auf den Pfad                                                                                                   */
/* =====================                                                                                                   */
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
#if 0
void BL_LengthHack(int objekt);
#else
#define BL_LengthHack(o) /* nix */
#endif

#define NFB_PfadInit(/*TermT*/ Term, /*ObjektT*/ objekt)\
{\
    aktuelleStelle = 0;\
    NFB_Reduziert = FALSE;\
    BL_LengthHack(objekt);\
    NFPfad[0].Stelle = Term;\
    NFPfad[0].AnzNochZuBehandelnderTeilterme = TO_AnzahlTeilterme(Term);\
    NFPfad[0].EinfuegePunkt = NULL;\
    objektArt = objekt;\
    ObjektNummer = 0;\
}

/* Da gemaess Konvention in NFBildung die Normalformbildung mit leeren Baeumen
   nicht angestossen wird, sind negative Werte hier nicht zu befuerchten.
   Bei nichtleeren Baeumen ist die Mindesttiefe der enthaltenen Regeln bzw. Gleichungen 1. */

#define aktStelleVonVorne(){\
    NFPfad[aktuelleStelle].AnzNochZuBehandelnderTeilterme = \
			       TO_AnzahlTeilterme(NFPfad[aktuelleStelle].Stelle);\
    NFPfad[aktuelleStelle].EinfuegePunkt = NULL;\
}


#define aktStelleSei(neueStelle)    (aktuelleStelle = neueStelle)
#define AktTermIstKeineVar()        (SO_SymbIstFkt(TO_TopSymbol(NFPfad[aktuelleStelle].Stelle)))
#define PfadEinsAb()                (aktuelleStelle++)
#define PfadEinsAuf()               (aktuelleStelle--)

#define NichtWurzel()               (aktuelleStelle != 0)                 /* eigentlich > 0, dies ist aber Invariante */
#define Wurzel()                    (aktuelleStelle == 0)                                                           /**/

#define EsGibtKeineZuBehNachf()     (NFPfad[aktuelleStelle].AnzNochZuBehandelnderTeilterme == 0) /*s.o.*/
#define EsGibtNochZuBehNachf()      (NFPfad[aktuelleStelle].AnzNochZuBehandelnderTeilterme != 0)

#define PfadWurdeReduziert()        (NFB_Reduziert)  

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------------------------------ Das Reduziert-Flag -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

extern BOOLEAN NFB_Reduziert;
/* Diese Variable muss am Anfang jeder NF-Prozedur auf FALSE gesetzt werden. Sie wird in jeder Regel-Angewendet-Prozedur im
   Erfolgsfalle auf TRUE gesetzt und kann dann als Rueckgabewert der NF-Bildungs-Funktion genutzt werden.
   Bei Verwendung des 'normalen' (dh. nicht des Adhoc-Pfades) wird das Flag von PfadNeu automatisch auf FALSE gesetzt.
   
   Das Flag wird von folgenden Prozeduren im Erfolgsfalle auf TRUE gesetzt:
   - NFB_RegelAngewendet
   - NFB_RegelAngewendetAH
   - NFB_GleichungAngewendet
   - NFB_GleichungAngewendetAH
*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Initialisieren ---------------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

void NFB_Init(void);
/* Die Variable Pfad einrichten (dh. das Array NFPfad */

void NFB_AusdehnungsprozedurAnmelden(void (* fkt) (StellenT));
/* Diese Prozedur wird aufgerufen, wenn der Reduktionspfad ausgedehnt werden muss (Erspart Abfragen).
   Nur eine Prozedur anmeldbar.
*/

/* Fuer Statistik-Zwecke */
#define /*unsigned int*/ NFB_AusdehnungPfad(/*void*/) (aktuelleGroessePfad+1)
/* Liefert die maximale Ausdehnung des Reduktionspfades (=groesste eingerichtete, nicht erreichte  Red.tiefe) zurueck.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- PfadEinsWeiterSetzen -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Vor.: die aktuelle Stelle im Pfad zeigt auf einen Knoten im Term (Term als Baum interpretiert), dessen Nachfolger noch  */
/*       nicht alle abgehandelt sind. Dann wird ein neuer Eintrag im Pfad erzeugt, der den ersten noch nicht behandelten   */
/*       Teilterm des aktuellen Terms als Term hat; der aktuelle Term ist jetzt dieser Teilterm.                           */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

void NFB_PfadEinsWeiterSetzen(void);

/*-------------------------------------------------------------------------------------------------------------------------*/
/*--------------------------- Variablen und Operationen zum Arbeiten auf einem adhoc-Pfad ---------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Ein adhoc-Pfad besteht lediglich aus einem aktuellen Teilterm und einem Einfuegeindex. Dies ermoeglicht es einer        */
/* Prozedur, die diesen Pfad verwendet, einen Term Zelle fuer Zelle durchzulaufen und dabei ggf. Teilterme zu ersetzen.    */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

extern UTermT aktTeiltermAH, aktTermAH, EinfuegepunktAH, EndeAH;

#define AdHocPfadNeu(/*TermT*/ Term, /*ObjektT*/ objekt)\
{   aktTermAH = Term;\
    aktTeiltermAH = Term;\
    EinfuegepunktAH = NULL;\
    EndeAH = TO_NULLTerm(Term);\
    objektArt = objekt;\
}

BOOLEAN AdHocPfadWeitersetzen(void);
/* Setzt den AdHoc-Pfad eins weiter. Rueckgabe ist TRUE, wenn die letzte Stelle noch nicht
   ueberschritten wurde.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Regel angewendet ? -------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die Prozedur Regel angewendet versucht destruktiv, einen Reduktionschritt an der aktuellen, im Pfad gespeicherten       
   Stelle des auf Normalform zu bringenden Terms auszufuehren. Zurueckgegeben wird, ob dies moeglich war.                  
   Dabei wird unterschieden, ob es sich bei der Stelle, an der reduziert werden soll, um die Wurzel des Terms       
   handelte oder nicht.
   Nach erfolgreicher Reduktion steht in der aktuellen Stelle des Pfades der frisch reduzierte Term, wobei Stelligkeit
   und Einfuegepunkt bereits wieder neu initialisiert sind, dh. die aktuelle Stelle kann sofort wieder behandelt werden
   (ohne weitere Anpassungen).
*/                                                                                               


BOOLEAN NFB_RegelAngewendet(void);
/* Versucht, eine Regel auf den aktuellen Teilterm im Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.
   Der Pfad wird anschliessend wieder restauriert.*/

BOOLEAN NFB_RegelAngewendetAH(void);
/* Versucht, eine Regel auf den aktuellen Teilterm im Adhoc-Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.
   Nach erfolgreicher Reduktion steht in der aktuellen Stelle des Pfades der frisch reduzierte Term.*/

  

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- GleichungAngewendet ------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Diese Prozedur versucht destruktiv, einen Reduktionschritt mit einer Gleichung an der aktuell im Pfad gespeicherten     */
/* Stelle auszufuehren.                                                                                                    */
/*                                                                                                                         */
/* Dabei wird ein moeglicher Reduktionnsschritt nur dann ausgefuehrt, wenn er vertraeglich mit der Ordnung ist.            */
/* Bei den Prozeduren, die die Anwendbarkeit einer vorgegebenen Gleichung pruefen, wird die angegebene Gleichung in        */
/* beiden Richtungen geprueft (natuerlich incl. Ordnungstest).                                                             */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFB_GleichungAngewendet(void);
/* Versucht, eine Gleichung auf den aktuellen Teilterm im Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.
   Der Pfad wird anschliessend wieder restauriert.*/

BOOLEAN NFB_GleichungAngewendetAH(void);
/* Versucht, eine Gleichung auf den aktuellen Teilterm im Adhoc-Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.
   Nach erfolgreicher Reduktion steht in der aktuellen Stelle des Pfades der frisch reduzierte Term.
   Der Pfad wird anschliessend wieder restauriert.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Anwendbarkeit eines bestimmten Objektes irgendwo auf einen Term --------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Folgende Prozeduren verfolgen die Vorordnungs-Strategie, um die ihnen uebergebenen Terme auf Reduzierbarkeit mit dem    */
/* uebergebenen Objekt zu pruefen. Die Prozeduren mit 'Anwendbar' im Namen bereiten nur die Reduktionsschritt vor, welcher */
/* dann mit dem grossen Exekutor ausgefuehrt werden kann. Die Prozeduren mit 'Angewendet' im Namen fuehren diesen Schritt  */
/* unmittelbar aus. Alle Prozeduren kehren danach aber sofort mit TRUE zurueck, ohne weiterzusuchen.                       */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFB_EineRegelAnwendbar(RegelT regel, TermT Term, ObjektT objekt);

BOOLEAN NFB_EineGleichungAnwendbar(GleichungsT gleichung, TermT Term, ObjektT objekt);

BOOLEAN NFB_EineRegelAngewendet(RegelT regel, TermT Term, ObjektT objekt);
/* Versucht, die angegebene Regel auf den aktuellen Teilterm im Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.*/
   
BOOLEAN NFB_EineGleichungAngewendet(GleichungsT gleichung, TermT Term, ObjektT objekt);
/* Versucht, die angegebene Gleichung auf den aktuellen Teilterm im Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Der grosse Exekutor ----------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Zur Ausfuehrung des mit NFB_...Anwendbar() vorbereiteten Reduktionsschrittes muss nur noch der grosse Exekutor auf-     */
/* gerufen werden. Dieser fuehrt lediglich den liegengebliebenen Reduktionsschritt aus, ohne den Reduktionspfad zu         */
/* restaurieren. (Aufruf erfolgt ueber das Makro in NFBildung.                                                             */
/*-------------------------------------------------------------------------------------------------------------------------*/

extern void (*NFB_DerGrosseExekutor)(void);

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Reduzibilitaetstest ----------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFB_TermInnenReduzibelE(UTermT term);

BOOLEAN NFB_TermInnenReduzibelR(UTermT term);

BOOLEAN NFB_TermInnenReduzibelRE(UTermT term);

int NFB_AnzTermReduzibelR(UTermT Term);

int NFB_AnzTermReduzibelE(UTermT Term);

int NFB_AnzTermReduzibelRE(UTermT Term);

#ifdef __cplusplus
}
#endif

#endif /* NFBASIS_H */

