/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFBildung.h                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (normales Verfahren)                                       */
/* -                                                                                                            */
/* - 23.03.94 Arnim. Angelegt.                                                                                  */
/* - 21.04.94 Arnim. Fertiggestellt. Dabei noch die Normalformbildung zur Gleichungsmenge als eigene Prozeduren */
/* -                 eingebaut.                                                                                 */
/* - 06.07.94 Arnim. Prozedurkoepfe und -dummies fuer NFBildung mit neuer Regel/Gleichung.                      */
/* - 06.03.95 Arnim. Struktur geaendert gemaess NFBildung.doc                                                   */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef NFBILDUNG_H
#define NFBILDUNG_H


#include "TermOperationen.h"    /* Wegen TermT */
#include "RUndEVerwaltung.h"    /* Wegen GleichungsT und RegelT */
#include "NFBasis.h"            /* Wegen Exekutor und anderem */
#include "pcl.h"
#include "MNF_CPs.h"                /* Meldeproc R/E-Leer */

#ifdef __cplusplus
extern "C" {
#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Typ-Definitionen  ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

typedef BOOLEAN (*NFProzedurT)          (TermT, ObjektT);
typedef BOOLEAN (*TermpaarIrgendwoT)    (TermT, TermT, TermT, ObjektT, PCL_ObjektNummerT *);   
typedef BOOLEAN (*NFInnenReduzibelT)    (TermT);
typedef int     (*NFRedStellenT)        (TermT);

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Funktions-Pointer ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern NFProzedurT           NFBildungProz, NFZuRegelnProz, NFZuGleichungenProz, NFBildungStratProz, NFZuRegelnStratProz, NFZuGleichungenStratProz;
extern TermpaarIrgendwoT     GleichungAlsTermpaarAnwendbarProz, RegelAlsTermpaarAnwendbarProz;
extern NFInnenReduzibelT     NFInnenReduzibelProzE, NFInnenReduzibelProzRE, NFInnenReduzibelProzR;
extern NFRedStellenT         NFAnzRedStellenProzR, NFAnzRedStellenProzE, NFAnzRedStellenProzRE;

/* -------------------------------------------------------------------------------------------------------------*/
/* - Zum Irreduzibilitaetstest                                                                                  */
/* -------------------------------------------------------------------------------------------------------------*/

#define NF_TEST 0
/* Alternativen: 0,1 */

#if NF_TEST  
BOOLEAN NF_IrrtestFehler(void);
/* liefert TRUE, wenn reduzible Terme die NF-Bildung verliessen oder der Test ausgeschaltet war! */

typedef void (*IrrTestProcT) (TermT);

extern IrrTestProcT IrrNFBildungProz, IrrNFZuRegelnProz, IrrNFZuGleichungenProz;

void NF_IrrTestDummy(TermT Term);
void NF_IrrTestRE(TermT Term);
void NF_IrrTestR(TermT Term);
void NF_IrrTestE(TermT Term);
#endif /* NF_TEST */
  
/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Dummy ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFDummyProz(TermT Term, ObjektT objekt);
BOOLEAN NFInnenReduzibelDummyProz(UTermT Term);
int NFRedStellenDummyProz(UTermT Term);

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Initialisierung -----------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void NF_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void NF_InitAposteriori(InfoStringT DEF, InfoStringT USR);

/* Initialisieren der fuer die eingestellte NF-Strategie und -Variante benoetigten Module. 
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Mitteilung ueber Aenderungen in der RM bzw. GM ----------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* Baukasten  */

extern BOOLEAN RMleer, GMleer;

#define /* void */ NF_RA(/*void*/)                                                                              \
{if (RMleer){                                                                                                  \
   RMleer = FALSE;                                                                                             \
   NFZuRegelnProz = NFZuRegelnStratProz;                                                                        \
   NFInnenReduzibelProzR = NFB_TermInnenReduzibelR;                                                             \
   MNF_REMengenGeaendert (RMleer, GMleer);                                                                    \
   if (GMleer){       /* RM nicht leer, aber GM: R */                                                          \
      NFBildungProz = NFZuRegelnStratProz;                                                                      \
      NFInnenReduzibelProzRE = NFB_TermInnenReduzibelR;                                                         \
      NFAnzRedStellenProzRE = NFB_AnzTermReduzibelR;                                                            \
   }                                                                                                            \
   else{               /* RM und GM nicht leer: RG*/                                                            \
      NFBildungProz = NFBildungStratProz;                                                                       \
      NFInnenReduzibelProzRE = NFB_TermInnenReduzibelRE;                                                        \
      NFAnzRedStellenProzRE = NFB_AnzTermReduzibelRE;                                                           \
   }                                                                                                            \
}}

#define /*void*/ NF_RE(/*void*/)                                                                                \
{if (RE_RegelmengeLeer()){                                                                                      \
  RMleer = TRUE;                                                                                               \
  NFZuRegelnProz = NFDummyProz;                                                                                 \
  NFInnenReduzibelProzR = NFInnenReduzibelDummyProz;                                                            \
  MNF_REMengenGeaendert (RMleer, GMleer);                                                                     \
  if (GMleer){/* RM leer, GM leer: Dummies */                                                                  \
     NFBildungProz = NFDummyProz;                                                                               \
     NFInnenReduzibelProzRE = NFInnenReduzibelDummyProz;                                                        \
     NFAnzRedStellenProzRE = NFRedStellenDummyProz;                                                             \
  }                                                                                                             \
  else{/* RM leer, GM nicht: G */                                                                               \
     NFBildungProz = NFZuGleichungenStratProz;                                                                  \
     NFInnenReduzibelProzRE = NFB_TermInnenReduzibelE;                                                          \
     NFAnzRedStellenProzRE = NFB_AnzTermReduzibelE;                                                             \
  }                                                                                                             \
}}

#define /*void*/ NF_GA(/*void*/)                                                                                \
{if (GMleer){                                                                                                  \
  GMleer = FALSE;                                                                                              \
  NFZuGleichungenProz = NFZuGleichungenStratProz;                                                               \
  NFInnenReduzibelProzE = NFB_TermInnenReduzibelE;                                                              \
  MNF_REMengenGeaendert (RMleer, GMleer);                                                                     \
  if (RMleer){/* GM nicht leer, RM leer: G */                                                                  \
     NFBildungProz = NFZuGleichungenStratProz;                                                                  \
     NFInnenReduzibelProzRE = NFB_TermInnenReduzibelE;                                                          \
     NFAnzRedStellenProzRE = NFB_AnzTermReduzibelE;                                                             \
  }                                                                                                             \
  else{/* beide nicht leer: RG */                                                                               \
     NFBildungProz = NFBildungStratProz;                                                                        \
     NFInnenReduzibelProzRE = NFB_TermInnenReduzibelRE;                                                         \
     NFAnzRedStellenProzRE = NFB_AnzTermReduzibelRE;                                                            \
  }                                                                                                             \
}}

#define /*void*/ NF_GE(/*void*/)                                                                                \
{if (RE_GleichungsmengeLeer()){                                                                                 \
  GMleer = TRUE;                                                                                               \
  NFZuGleichungenProz = NFDummyProz;                                                                            \
  NFInnenReduzibelProzE = NFInnenReduzibelDummyProz;                                                            \
  MNF_REMengenGeaendert (RMleer, GMleer);                                                                     \
  if (RMleer){/* beide leer: Dummies */                                                                        \
     NFBildungProz = NFDummyProz;                                                                               \
     NFInnenReduzibelProzRE = NFInnenReduzibelDummyProz;                                                        \
     NFAnzRedStellenProzRE = NFRedStellenDummyProz;                                                             \
  }                                                                                                             \
  else{/* GM leer, RM nicht: R */                                                                               \
     NFBildungProz = NFZuRegelnStratProz;                                                                       \
     NFInnenReduzibelProzRE = NFB_TermInnenReduzibelR;                                                          \
     NFAnzRedStellenProzRE = NFB_AnzTermReduzibelR;                                                             \
  }                                                                                                             \
}}

#if !NF_TEST
#define /*void*/ NF_RegelAufgenommen(/*void*/) NF_RA()
#define /*void*/ NF_RegelEntfernt(/*void*/) NF_RE()
#define /*void*/ NF_GleichungAufgenommen(/*void*/) NF_GA()
#define /*void*/ NF_GleichungEntfernt(/*void*/) NF_GE()
#else
#define /*void*/ NF_RegelAufgenommen(/*void*/)\
{if (RMleer){\
  IrrNFZuRegelnProz = NF_IrrTestR;\
  if (GMleer) IrrNFBildungProz = NF_IrrTestR;/* RM nicht leer, aber GM: R */\
  else IrrNFBildungProz = NF_IrrTestRE;/* Rm und GM nicht leer: RG*/\
}\
NF_RA();\
}


#define /*void*/ NF_RegelEntfernt(/*void*/)\
{if (RE_RegelmengeLeer()){\
  IrrNFZuRegelnProz = NF_IrrTestDummy;\
  if (GMleer) IrrNFBildungProz = NF_IrrTestDummy;/* RM leer, GM leer: Dummies */\
  else IrrNFBildungProz = NF_IrrTestE;/* RM leer, GM nicht: G */\
}\
NF_RE();\
}

#define /*void*/ NF_GleichungAufgenommen(/*void*/)\
{if (GMleer){\
  IrrNFZuGleichungenProz = NF_IrrTestE;\
  if (RMleer) IrrNFBildungProz = NF_IrrTestE;/* GM nicht leer, RM leer: G */\
  else IrrNFBildungProz = NF_IrrTestRE;/* beide nicht leer: RG */\
}\
NF_GA();\
}

#define /*void*/ NF_GleichungEntfernt(/*void*/)\
{if (RE_GleichungsmengeLeer()){\
  IrrNFZuGleichungenProz = NF_IrrTestDummy;\
  if (RMleer) IrrNFBildungProz = NF_IrrTestDummy;/* beide leer: Dummies */\
  else IrrNFBildungProz = NF_IrrTestR;/* GM leer, RM nicht: R */\
}\
NF_GE();\
}
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Prozeduren ----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* Dokumentation siehe NFBildung.doc */
#if NF_TEST
/***************** Irreduzibilitaets-Test *********************************/
extern BOOLEAN NF_private_booly;

/*NFBildungZuRegeln*/
/*-----------------*/
#define /*BOOLEAN*/ NF_NFBildungZuRegeln(/*Term*/ Term, /*ObjektT*/ objekt)\
(NF_private_booly = ( (*NFZuRegelnProz) (Term, objekt)), (*IrrNFZuRegelnProz) (Term), NF_private_booly)

/*NFBildungZuGleichungen*/
/*----------------------*/
#define /*BOOLEAN*/ NF_NFBildungZuGleichungen(/*Term*/ Term, /*ObjektT*/ objekt)\
(NF_private_booly =  ( (*NFZuGleichungenProz) (Term, objekt)), (*IrrNFZuGleichungenProz) (Term), NF_private_booly)

/*NFBildung*/
/*---------*/
#define /*BOOLEAN*/ NF_NFBildung(/*Term*/ Term, /*ObjektT*/ objekt)\
(NF_private_booly =  ( (*NFBildungProz) (Term, objekt)), (*IrrNFBildungProz) (Term), NF_private_booly)

/*NFMitNeuerRegel*/
/*---------------*/
#define /*BOOLEAN*/ NF_NFMitNeuerRegelRMrechts(/* RegelT*/ Regel, /*TermT*/ Term, /*ObjektT*/ objekt)\
        (NF_private_booly =  ((NFB_EineRegelAngewendet(Regel, Term, objekt))? (((*NFBildungProz) (Term, objekt)), TRUE):FALSE), (*IrrNFBildungProz) (Term), NF_private_booly)

/*NFMitNeuerGleichung*/
/*-------------------*/
#define /*BOOLEAN*/ NF_NFMitNeuerGleichungRMrechts(/*GleichungsT*/ Gleichung, /*TermT*/ Term, /*ObjektT*/ objekt) \
        (NF_private_booly =  ((NFB_EineGleichungAngewendet (Gleichung, Term, objekt))? ((*NFBildungProz)(Term, objekt), TRUE):FALSE), (*IrrNFBildungProz) (Term), NF_private_booly)

#else  
/****************** Normale Fassung **************************************/

/*NFBildungZuRegeln*/
/*-----------------*/
#define /*BOOLEAN*/ NF_NFBildungZuRegeln(/*Term*/ Term, /*ObjektT*/ objekt) ( (*NFZuRegelnProz) (Term, objekt))

/*NFBildungZuGleichungen*/
/*----------------------*/
#define /*BOOLEAN*/ NF_NFBildungZuGleichungen(/*Term*/ Term, /*ObjektT*/ objekt) ( (*NFZuGleichungenProz) (Term, objekt))

/*NFBildung*/
/*---------*/
#define /*BOOLEAN*/ NF_NFBildung(/*Term*/ Term, /*ObjektT*/ objekt) ( (*NFBildungProz) (Term, objekt))

/*NFMitNeuerRegel*/
/*---------------*/
#define /*BOOLEAN*/ NF_NFMitNeuerRegel(/* RegelT*/ Regel, /*TermT*/ Term, /*ObjektT*/ objekt)\
        ((NFB_EineRegelAngewendet(Regel, Term, objekt))? (((*NFBildungProz) (Term, objekt)), TRUE):FALSE)

/*NFMitNeuerGleichung*/
/*-------------------*/
#define /*BOOLEAN*/ NF_NFMitNeuerGleichung(/*GleichungsT*/ Gleichung, /*TermT*/ Term, /*ObjektT*/ objekt) \
        ((NFB_EineGleichungAngewendet (Gleichung, Term, objekt))? ((*NFBildungProz)(Term, objekt), TRUE):FALSE)

#endif 

/*RegelAnwendbar*/
/*--------------*/
#define /*BOOLEAN*/ NF_RegelAnwendbar(/* RegelT*/ Regel, /*TermT*/ Term, /*ObjektT*/ Objekt)\
        (NFB_EineRegelAnwendbar(Regel, Term, Objekt))

/*RegelAngewendet*/
/*---------------*/
#define /*BOOLEAN*/ NF_RegelAngewendet(/*RegelT*/ Regel, /*TermT*/ Term, /*ObjektT*/ Objekt) \
        (NFB_EineRegelAngewendet(Regel, Term, Objekt))

/*GleichungAnwendbar*/
/*------------------*/
#define /*BOOLEAN*/ NF_GleichungAnwendbar(/*GleichungsT*/ Gleichung, /*TermT*/ Term, /*ObjektT*/ Objekt) \
        (NFB_EineGleichungAnwendbar(Gleichung, Term, Objekt))

/*GleichungAngewendet*/
/*-------------------*/
#define /*BOOLEAN*/ NF_GleichungAngewendet(/*GleichungsT*/ Gleichung, /*TermT*/ Term, /*ObjektT*/ Objekt) \
        (NFB_EineGleichungAngewendet(Gleichung, Term, Objekt))


/*Der grosse Exekutor*/
/*-------------------*/
#define /*void*/ NF_GrossenExekutorAnwenden(/*void*/) ((*NFB_DerGrosseExekutor)())
/* Fuehrt einen mit ???Anwendbar vorbereiteten Reduktionsschritt tatsaechlich aus. Diese Trennung zwischen
   Matchen und tatsaechlicher Reduktion ist an dieser Stelle noetig, weil Objekte im Baum nicht veraendert werden duerfen.
   Anwendung: waehrend der Interreduktion, wenn Objekte aus einem DSBaum reduziert wrden koennen, aber vorher aus dem
              Index entfernt werden muessen.
*/


/* Kapur-Test */
/* ---------- */

#define /*BOOLEAN*/ NF_TermInnenReduzibelR(/*UTermT*/ Term) ( (*NFInnenReduzibelProzR) (Term))
#define /*BOOLEAN*/ NF_TermInnenReduzibelE(/*UTermT*/ Term) ( (*NFInnenReduzibelProzE) (Term))
#define /*BOOLEAN*/ NF_TermInnenReduzibelRE(/*UTermT*/ Term) ( (*NFInnenReduzibelProzRE) (Term))

/* Reduzibilitaetstest */
/* ------------------- */

#define /*int*/ NF_AnzReduziblerStellenR(/*UTermT*/ Term) ( (*NFAnzRedStellenProzR) (Term))
#define /*int*/ NF_AnzReduziblerStellenE(/*UTermT*/ Term) ( (*NFAnzRedStellenProzE) (Term))
#define /*int*/ NF_AnzReduziblerStellenRE(/*UTermT*/ Term) ( (*NFAnzRedStellenProzRE) (Term))

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Statistik -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void NF_Statistik(InfoStringT IS);
void NF_AnzReduktionenGesamt(unsigned long int *AnzRedGes);


#ifdef __cplusplus
}
#endif

#endif
