/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFInnermost.h                                                                                              */
/* - *************                                                                                              */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (Leftmost-Innermost)                                       */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/
#ifndef NFINNERMOST_H
#define NFINNERMOST_H

#include "general.h"
#include "TermOperationen.h"
#include "NFBasis.h"

#ifdef __cplusplus
extern "C" {
#endif

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
void NFI_Init(void);
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dazu wird zunaechst eine 'abstrakte'      */
/* Variable Pfad initialisiert, die dann gemaess einer -aenderbaren- Strategie das Durchlaufen des Termes organisiert,     */
/* wobei das weitere Vorgehen in Abhaengigkeit vom Erfolg der letzten versuchten Reduktion sein kann.                      */
/* Dies wird solange fortgefuehrt, bis es keine reduzierbaren Stellen mehr im Anfrageterm gibt.                            */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformZuRegelnDirekt(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFI_NormalformZuRegelnNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*      >>>Falls Regelsystem leer ist<<<                                                                                   */
/* Bringt den uebergebenen Term bzgl. des aktuellen Gleichungssystems auf Normalform. Dazu wird zunaechst eine 'abstrakte' */
/* Variable Pfad initialisiert, die dann gemaess einer -aenderbaren- Strategie das Durchlaufen des Termes organisiert,     */
/* wobei das weitere Vorgehen in Abhaengigkeit vom Erfolg der letzten versuchten Reduktion sein kann.                      */
/* Dies wird solange fortgefuehrt, bis es keine reduzierbaren Stellen mehr im Anfrageterm gibt.                            */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformZuGleichungenDirekt (TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Gleichungssystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFI_NormalformZuGleichungenNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Gleichungssystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regel- und Gleichungssystems auf Normalform. Dazu wird zunaechst eine  */
/* 'abstrakte' Variable Pfad initialisiert, die dann gemaess einer -aenderbaren- Strategie das Durchlaufen des Termes      */
/* organisiert, wobei das weitere Vorgehen in Abhaengigkeit vom Erfolg der letzten versuchten Reduktion sein kann.         */
/* Dies wird solange fortgefuehrt, bis es keine reduzierbaren Stellen mehr im Anfrageterm gibt.                            */
/* Es wird erst dann versucht, Gleichungen anzuwenden, wenn es keine anwendbaren Regeln mehr gibt.                         */
/* In diesem Fall, wird versucht, eine anwendbare Gleichung zu finden. Gelingt dies, so wird die Gleichung angewendet, und */
/* anschliessend wird an dieser Stelle mit Regeln fortgefahren. Dies wiederholt sich, bis es weder anwendbare Regeln noch  */
/* Gleichungen mehr gibt.                                                                                                  */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformBildenDirekt(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFI_NormalformBildenNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
#ifdef __cplusplus
}
#endif

#endif
