#ifndef NFTOPDOWN_H
#define NFTOPDOWN_H
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFTopDown.h                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (TopDown-Strategie)                                        */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "general.h"
#include "NFBasis.h"
#include "TermOperationen.h"

#ifdef __cplusplus
extern "C" {
#endif


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

void NFT_Init(void);
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFT_NormalformZuGleichungenAbDirekt (TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */

BOOLEAN NFT_NormalformZuGleichungenAbNochmal (TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFT_NormalformZuGleichungenAufDirekt (TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFT_NormalformZuGleichungenAufNochmal (TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFT_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFT_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFT_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFT_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den  Term bzgl. des akt. Regel- und Gleichungssystems auf Normalform. Dabei wird aber  */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* Es wird erst dann versucht, Gleichungen anzuwenden, wenn es keine anwendbaren Regeln mehr gibt.                         */
/* In diesem Fall wird versucht, eine anwendbare Gleichung zu finden. Gelingt dies, so wird die Gleichung angewendet, und  */
/* anschliessend wird an dieser Stelle mit Regeln fortgefahren. Dies wiederholt sich, bis es weder anwendbare Regeln noch  */
/* Gleichungen mehr gibt.                                                                                                  */
/*-------------------------------------------------------------------------------------------------------------------------*/


BOOLEAN NFT_NormalformAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFT_NormalformAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFT_NormalformAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFT_NormalformAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


#ifdef __cplusplus
}
#endif

#endif
