/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - Packen.h                                                                                              */
/* - ********                                                                                              */
/* -                                                                                                       */
/* - Dieses Modul stellt die Funktionalitaet zum Ein- und Auspacken von Termen in Strings zur Verfuegung.  */
/* - Diese Byteketten koennen dann entweder zur platzsparenden Ablage der KP-Menge verwendet werden oder   */
/* - aber zum Versenden von Termen im Rahmen von DISCOUNT,                                                 */
/* -                                                                                                       */
/* - 27.03.95 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

/* Dokumentation:

Funktionalitaet:

1. Terme sollen in Byte-Ketten abgelegt bzw. aus solchen restauriert werden koennen.
   Dabei ist eine Normierung vorzunehmen (-> DISCOUNT).

2. Terme als Byteketten sollen als Grundlage der KP-Menge herhalten koennen. Dazu ist eine
   entsprechende Speicherverwaltung zu etablieren. Diese soll:
   - fuer alle Byteketten-Laengen in Vierer-Schritten Freispeicherlisten fuehren,
   - fuer Byteketten oberhalb einer bestimmten Laenge einzeln den Platz allokieren

*/
#ifndef PACKEN_H
#define PACKEN_H

#include "InfoStrings.h"
#include "TermOperationen.h"
#include "TermpairRepresentation.h"
#include "RUndEVerwaltung.h"


#ifdef __cplusplus
extern "C" {
#endif


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Initialize--------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN PK_InitAposteriori(InfoStringT DEF, InfoStringT USR);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_NewFlatsGenerated(void);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

TermpairT PK_Flat2TP (TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent,
                      UTermT actualPosition, PCL_ObjektNummerT pclID);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_TP2Par (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT actualParent, 
                RegelOderGleichungsT otherParent, PCL_ObjektNummerT *pclIDPtr);
/* Terme auspacken, Termpaar behalten. */



/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_Par2Flat (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                  RegelOderGleichungsT otherParent);
/* Destroys the termpair-representation irrecoverable. Gepacktes Termpaar freigeben. */


/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

TermpairT PK_Par2TP (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                     RegelOderGleichungsT otherParent, BOOLEAN changed, PCL_ObjektNummerT pclID);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void PK_DestroyTP  (TermpairT tp, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent);
/* Termpaar zerstoeren */

void  PK_DestroyPar (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                     RegelOderGleichungsT otherParent);
/* Termpaar und Flachterme zerstoeren */


/* ------------------------------------------------------------------------------------------------------*/
/* ------- Unambiguous Comparison -----------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

BOOLEAN PK_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                  TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2);
/* Vom Entstehungszeitpunkt unabhaengiger Vergleich von Termpaaren.
   Hier: Eltern-Nummern, dann Symb-Lex.
*/

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics -------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_Statistics(InfoStringT IS);


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Output -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_PrintTermpair (TermpairT tp, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent, const char*trenner);


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_CleanUp(void);

#ifdef __cplusplus
}
#endif

#endif
