/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - PriorityQueue.h                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Schnittstelle fuer das Set of unselected equations.                                                        */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef PQ_H
#define PQ_H


#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"
#include "TermpairRepresentation.h"
#include "SUEBasis.h"

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PQ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN PQ_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/*
   immortalCPsDynamic = TRUE  : CPs zu SUE-Demigod sind dynamische Objekte (extra freizugeben)
                        FALSE : ... koennen ohne Extra-Aufruf freigegeben werden

*/

/* wird bei ChangeStrategy benoetigt */
void PQ_REINITIALIZE (void);
  
  
/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
typedef void (*PQEinfuegerT)(TermpairT,WeightEntryT,RegelOderGleichungsT,RegelOderGleichungsT);
extern PQEinfuegerT PQ_InsertProc,
                    PQ_InsertImmortalProc;
extern BOOLEAN (* PQ_PostponeProc)(TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);

#define PQ_INSERT(tp, we, actualParent, otherParent) ( (* PQ_InsertProc) (tp, we, actualParent, otherParent))
/* Das aus den uebergebenen Daten bestehende KP in die PQ einfuegen. */

#define PQ_INSERTImmortal(tp, we, actualParent, otherParent) \
( (* PQ_InsertImmortalProc) (tp, we, actualParent, otherParent))
/* Das aus den uebergebenen Daten bestehende KP in die PQ einfuegen. */

#define PQ_POSTPONE(tp, we, ActualParent, OtherParent) ( (* PQ_PostponeProc) (tp, we, ActualParent, OtherParent))
/* */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern BOOLEAN (* PQ_SelectInitialProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *); 

#define PQ_SELECTINITIAL(tpPtr, ActualParentPtr, OtherParentPtr)\
( (* PQ_SelectInitialProc) (tpPtr, ActualParentPtr, OtherParentPtr) )
/* Rueckgabe der Daten des minimalen KPs. */


extern BOOLEAN (* PQ_SelectProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *);

#define PQ_SELECT(tpPtr, ActualParentPtr, OtherParentPtr)\
( (* PQ_SelectProc) (tpPtr, ActualParentPtr, OtherParentPtr) )
/* Rueckgabe der Daten des minimalen KPs. */



/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_NewParentProc)   (RegelOderGleichungsT);
extern void    (* PQ_KillParentProc)  (RegelOderGleichungsT);
extern void    (* PQ_RedirectParentProc) (void);
extern void    (* PQ_FinishParentProc) (void);

#define PQ_NEWPARENT(re) ( (* PQ_NewParentProc) (re))
/* Referenz (bei Regeln), Refernzen (bei Gleichungen) in re (bzw re und Gegenrichtung eintragen) */

#define PQ_KILLPARENT(re) ( (* PQ_KillParentProc) (re))
/* Referenzen loeschen. */

#define PQ_RedirectParent() ( (* PQ_RedirectParentProc) ())
/* Gegenrichtung verwenden. */

#define PQ_FinishParent() ( (* PQ_FinishParentProc) ())
/* KPBildung zum aktuellen Elternteil abgeschlossen. */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_WalkthroughProc) (PQ_WalkthroughProcedureT, BOOLEAN);
extern void    (* PQ_WatchingWalkthroughProc) (PQ_WatchingWalkthroughProcedureT);
extern void    (* PQ_TermVerfolgungsProc) (void (*) (TermT, TermT));
extern unsigned long int (* PQ_AnzKPsToGoProc) (void);

#define PQ_WALKTHROUGH(ApplicationProcedure, NeedsTerms) ( (* PQ_WalkthroughProc) (ApplicationProcedure, NeedsTerms))
/* ApplicationProcedure auf alle gueltigen KPs anwenden und je nach Rueckgabewert weiterbehandeln.
   Je nach NeedsTerms werden die Termpaare entpackt.
*/

#define PQ_AnzKPsToGo() ( (* PQ_AnzKPsToGoProc) ())

#define PQ_WatchingWalkthrough(WP) ( (* PQ_WatchingWalkthroughProc) (WP))

#define PQ_TermeVerfolgen(OperationAufTermen) ( (*PQ_TermVerfolgungsProc) (OperationAufTermen))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_StatisticsProc)  (InfoStringT);
extern void   (* PQ_ActSizeProc) (unsigned long int *, unsigned long int *);

#define PQ_AktuelleMengenGroesse(/*unsigned long int **/ AktKPZahlPtr,          \
                                 /*unsigned long int**/ AktGueltigeKPZahlPtr)   \
((* PQ_ActSizeProc) (AktKPZahlPtr, AktGueltigeKPZahlPtr))

#define PQ_Statistics(Degree) ( (* PQ_StatisticsProc) (Degree))
/* Mitgerechnete Statistik ausgeben auf IO_Strom. */


/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PQ_CleanupProc) (void);
extern void (* PQ_SoftCleanupProc) (void);

#define PQ_Cleanup() ( (* PQ_CleanupProc)())
#define PQ_SoftCleanup() ( (* PQ_SoftCleanupProc)())


#ifdef __cplusplus
}
#endif

#endif /* PQ_H */
