/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - SUEM.h                                                                                                     */
/* - ******                                                                                                     */
/* -                                                                                                            */
/* - Schnittstelle fuer das Set of unselected equations.                                                        */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef SUEM_H
#define SUEM_H

#include "general.h"
#include "InfoStrings.h"
#include "TermOperationen.h"
#include "RUndEVerwaltung.h"
#include "SUEBasis.h"

#ifdef __cplusplus
extern "C" {
#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ INITIALIZE -------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void SUE_InitApriori(void);

BOOLEAN SUE_InitAposteriori(InfoStringT DEFsue,  InfoStringT USRsue,
                            InfoStringT DEFclas, InfoStringT USRclas,
                            InfoStringT DEFreclas, InfoStringT USRreclas,
                            InfoStringT DEFcgclas, InfoStringT USRcgclas,
                            InfoStringT DEFpq,   InfoStringT USRpq,
                            InfoStringT DEFtpr,  InfoStringT USRtpr,
                            InfoStringT DEFcg,    InfoStringT USRcg);

void SUE_ParamInfo(ParamInfoArtT what, int whichModule,
                   InfoStringT DEFsue,  InfoStringT USRsue,
                   InfoStringT DEFclas, InfoStringT USRclas,
                   InfoStringT DEFreclas, InfoStringT USRreclas,
                   InfoStringT DEFcgclas, InfoStringT USRcgclas,
                   InfoStringT DEFpq,   InfoStringT USRpq,
                   InfoStringT DEFtpr,  InfoStringT USRtpr,
		   InfoStringT DEFcg,    InfoStringT USRcg);
/* whichModule: 0 - SUE
                1 - CLAS
                2 - RECLAS
                3 - CGCLAS
                4 - PQ
                5 - TPR
		6 - CG
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ PARENTS ----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void SUE_NewParent(RegelOderGleichungsT re, TermT lhsReplica, TermT rhsReplica);

void SUE_KillParent(RegelOderGleichungsT re);

void SUE_RedirectParent(void);

void SUE_FinishParent(void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ GENERATION -------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void SUE_CPinitial(TermT lhs, TermT rhs);

void SUE_CPkilledRE(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclID);

void SUE_CPgeneratedWithFather(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition);
void SUE_CPgeneratedWithMother(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition);

void SUE_CPgeneratedWithDatabase(RegelOderGleichungsT Konklusion);

extern CounterT SUE_AnzAktivierterFakten,
                SUE_AnzAktivierterRE,
                SUE_AnzAktivierterG,
                SUE_AnzAktivierterREFifo,
                SUE_AnzAktivierterREHeu,
                SUE_AnzAktivierterGFifo,
                SUE_AnzAktivierterGHeu;

#define SUE_recentCPlhs() SUE_recCPlhs
#define SUE_recentCPrhs() SUE_recCPrhs
#define SUE_recentCPactualParent() SUE_ActualParent
#define SUE_recentCPotherParent() SUE_recCPOtherParent

#if COMP_PCL
#define SUE_recentCPPclID() (&SUE_recCPpclID)
#else
#define SUE_recentCPPclID() 0
#endif


void SUE_recentCPjoined(void);

void SUE_recentCPMNFjoined(TermT joinedTerm);

void SUE_recentCPsubsumed(void);

void SUE_recentCPMNFsubsumed(void);

void SUE_recentCPMNFpermsubsumed(void);

void SUE_recentCPpermsubsumed(void);

void SUE_recentCPinsert(void);


/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ Smallest Weight --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_SmallestCPWeightIsZero(void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ SELECTION --------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_Select(void);

BOOLEAN SUE_SelectInitial(void);
/* Liefert so lange unsterbliche KPs mit kleinstem Gewicht. 
   Intention: initiale oder TW-verschickte */

void SUE_TouchImmortals(void (* Proc) (TermT, TermT));
/* Liefert alle unsterblichen KPs zurueck, unabhaengig vom Gewicht. 
   Intention: IR-Opfer */

#define SUE_selectedCPlhs() SUE_selCPlhs
#define SUE_selectedCPrhs() SUE_selCPrhs
#define SUE_selectedCPNr() 0

#if COMP_PCL
#define SUE_selectedCPPclID() (&SUE_selCPpclID)
#else
#define SUE_selectedCPPclID() 0
#endif



void SUE_selectedCPjoined(void);

void SUE_selectedCPMNFjoined(TermT joinedTerm);  


void SUE_selectedCPsubsumed(void);

void SUE_selectedCPMNFsubsumed(void);

void SUE_selectedCPpermsubsumed(void);

void SUE_selectedCPMNFpermsubsumed(void);
   

void SUE_selectedCPRE(void);

void SUE_selectedCPMNFRE(void);

BOOLEAN SUE_selectedCPpostponed(void);



/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ WALKTHROUGH ------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int SUE_AnzKPsToGo(void);
/* Returns number of cps to be touched during walkthrough. */

void SUE_WALKTHROUGH(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms);
/* 
   AP returns: 0 -- noop
               1 -- delete
               2 -- reweight
 */

void SUE_TermeVerfolgen(void (*OperationAufTermen) (TermT, TermT));

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ STATISTICS -------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void SUE_Statistics(InfoStringT StatisticsIS);
extern CounterT AnzKPGesamt;
void SUE_AktuelleGroesse(unsigned long int *GesamtKPZahl,
                         unsigned long int *AktKPZahl,
                         unsigned long int *AktGueltigeKPZahl);

#define SUE_NummerAktKP() 0

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void SUE_CLEANUP(void);
void SUE_SOFTCLEANUP(void);


#ifdef __cplusplus
}
#endif

#endif
