/*=================================================================================================================
  =                                                                                                               =
  =  StrahlMemorium.h                                                                                             =
  =                                                                                                               =
  =  Strahlartiger Zugriff auf Speicher                                                                           =
  =                                                                                                               =
  =  16.04.96 Thomas.                                                                                             =
  =                                                                                                               =
  =================================================================================================================*/


#ifndef STRAHLMEMORIUM_H
#define STRAHLMEMORIUM_H

#include "DomaenenMemorium.h"


#ifdef __cplusplus
extern "C" {
#endif

/* Zu jeder Strahldomaene darf es beliebig viele gefrorene, aber hoechstens einen nicht gefrorenen Strom
   geben. Jede Strahldomaene hat also genau zwei Zustaende: Zustand KALT (nur beliebig viele gefrorene 
   Strahlen) und Zustand WARM (beliebig viele gefrorene Strahlen, genau ein nicht gefrorener Strahl).
   Sei s:=Seitengroesse, a:=AnzahlSeiten, d:=sizeof(struct StrahlTS).

   Benutzungsschema:

   1. Domaene = SM_NeueStrahldomaene(Start,Obergrenze,InkrementInSeiten,Name)
      Start und Obergrenze geben den zu belegenden vurtuellen Adressraum an, InkrementInSeiten die Anzahl von 
      Speicherseiten, die beim Ausdehnen der Domaene geholt werden. Initialer Domaenenzustand ist KALT.
   
   2. Nur im Zustand KALT:
      Strahl = SM_NeuerStrahl(Domaene);
      legt einen neuen Strahl in der Domaene an. Der Strahl ist ungefroren. Er beginnt auf einer Seiten-
      adresse mit einem Verwaltungseintrag der Groesse d. Zurueckgegeben wird die erste Adresse danach.
      Initial steht zur Verfuegung der Seitenbereich [Strahl-d, Strahl+(n*a-1)*s) mit n=1.
      Der Bereich ist zu lesen als rechtsoffenes Intervall.
      Der Adressbereich [Strahl-d, Strahl) wird fuer Verwaltungszwecke benutzt.
      Nach der Operation ist der Domaenenzustand WARM.
   
                                                           |
           Nur im Zustand WARM mit Argument                |         Immer ausser im Zustand WARM mit Argument
                Strahl nicht gefroren:                     |                 Strahl nicht gefroren:
   --------------------------------------------------------+--------------------------------------------------------
                                                           |
   3. Sobald auf die Seite Strahl+(n*a-1)*s des nicht ge-  |
      frorenen Strahles zugegriffen wird, wird der Spei-   |
      cherbereich ausgedehnt auf                           |
      [Strahl-d, Strahl+((n+1)*a-1)*s).                    |
      Weiter mit n:=n+1 und Domaenenfolgezustand WARM.     |                          %
      ACHTUNG: Es muss sichergestellt werden, dass nicht   |
               auf Seiten nach Strahl+(n*a-1)*s zugegrif-  |
               fen wird! Dazu muss auf den Bereich des     |
               Strahls zugegriffen werden in Schritten     |
               hoechstens von Seitengroesse.               |
                                                           |
   4. SM_StrahlGefrieren(Strahl, void *ErsteFreieAdresse)  | 7. SM_StrahlSchrumpfen(Strahl, void *ErsteFreieAdresse)
      stellt die Ausdehnungstaetigkeit beim Zugriff ab.    |    schraenkt den Speicherbereich eines Strahles noch
      Der Strahl belegt den Speicherbereich von Strahl     |    weiter ein. Hat der Strahl die Ausdehnung
      bis ausschliesslich ErsteFreieAdresse.               |    [Strahl-d,Ende), so muss gelten
      Domaenenfolgezustand: KALT.                          |    Strahl<=ErsteFreieAdresse<=Ende.
                                                           |    Domaenenfolgezustand: unveraendert
                                                           |
   5. SM_NichtGefrorenenStrahlLoeschen(Strahl) gibt den    | 8. SM_StrahlLoeschen(Strahl) gibt den gesamten Bereich
      Bereich des Strahls an die Domaene zurueck.          |    zurueck.
      Domaenenfolgezustand: KALT.                          |    Domaenenfolgezustand: unveraendert
                                                           |
   9. SM_StrahldomaeneLoeschen(Domaene) gibt die Domaene zurueck.
      Darf aufgerufen werden, ohne dass alle Strahlen der Domaene geloescht sind.

   Sinnvollerweise sollten keine leeren, gefrorenen Strahlen vorgehalten werden, da jeder Strahl wegen der Verwal-
   tungsinformationen mindestens eine Speicherseite beansprucht. Ein leerer, gefrorener Strahl verschwendet also
   genau eine Speicherseite.

   Modifikation Februar 1998: Der Adressbereich der Strahldomaenen wird nicht mehr explizit mit Start und Ober-
   grenze angegeben, sondern durch einen Eintrag in der statischen Variable Bereiche im Modul Domaenenmemorium.c,
   der durch ein Element eines entsprechenden Aufzaehlungstyps markiert ist.
*/


/*=================================================================================================================*/
/*= I. Strahldomaenen                                                                                             =*/
/*=================================================================================================================*/

DomaenenT SM_NeueStrahldomaene(DomaenenbereichsT Bereich, unsigned int InkrementInSeiten, const char *Name);

void SM_StrahldomaeneLeeren(volatile DomaenenT Domaene);/* alle Seiten weggeben, aber Verwaltungsstruktur behalten */

void SM_StrahldomaeneLoeschen(volatile DomaenenT Domaene);                    /* auch Verwaltungsstruktur weggeben */



/*=================================================================================================================*/
/*= II. Strahlen                                                                                                  =*/
/*=================================================================================================================*/

typedef void* StrahlT;

StrahlT SM_NeuerStrahl(DomaenenT Domaene);

void SM_StrahlGefrieren(StrahlT StrahlExtern, void *ErsteFreieAdresse);

void SM_NichtGefrorenenStrahlLoeschen(StrahlT StrahlExtern);

void SM_StrahlSchrumpfen(StrahlT StrahlExtern, void *ErsteFreieAdresse);

void SM_StrahlLoeschen(StrahlT StrahlExtern);
#ifdef __cplusplus
}
#endif

#endif
