/*=================================================================================================================
  =                                                                                                               =
  =  Subsumption.h                                                                                                =
  =                                                                                                               =
  =  Subsumption mit der Gleichungsmenge bzw. einzelnen Gleichungen                                               =
  =                                                                                                               =
  =  27.03.96  Arnim. Aus der Interreduktion ausgegliedert.                                                       =
  =================================================================================================================*/
#ifndef SUBSUMPTION_H
#define SUBSUMPTION_H

#include "RUndEVerwaltung.h"
#include "TermOperationen.h"
#include "general.h"

#ifdef __cplusplus
extern "C" {
#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar mit SG-Pufferung bzgl. GM/Permsub subsummieren ------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* TRUE, wenn die angegebene Gleichung links = rechts bzw. rechts = links von einer bereits vorhandenen 
   Gleichung oder einer Permgleichung vom Grad >= angegeb. Grad subsummiert wird. 
   MIT SUBGLEICHUNGEN PUFFERN.
   Vorbedingung: - links = rechts ist noch nicht in GM aufgenommen => kein Ausschluss zu beachten.
                 - links ist nicht identisch zu rechts
   Seiteneffekt: - Meldung, wenn Gleichung subummiert wird, und ggf. von welcher Gleichung.
                 - Bei Antwort FALSE koennen alle Subgleichungen von l = r in die Subgleichungsmenge aufgenommen werden
		   (da sie bereits berechnet und in den Puffer geschrieben wurden. Dies kann nicht in dieser Prozedur
                   erledigt werden, da die entsprechenden Subgleichngen an das Objekt Gleichung angehaengt werden muessen.).
*/
BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernGM(TermT links, TermT rechts);
BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernPermsub(TermT links, TermT rechts, unsigned short Grad);
BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernGMUndPermsub(TermT links, TermT rechts, unsigned short Grad);

/* Nur besetzen der Puffer, wird beim Arbeiten mit MNF/SelectClassify benutzt,
   gibt immer FALSE zurueck */
BOOLEAN SS_OnlyInitSGPuffer(TermT links, TermT rechts);

/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar bzgl. GM subsummieren -------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* Liefert TRUE, wenn die Gleichungsmenge bzw. eine Permgleichung (Grad) das Termpaar links = rechts 
   subsummiert. Dabei werden alle Subgleichungen des Termpaares berechnet und getestet, und 
   zwar in beiden Richtungen, aber nicht gepuffert.*/

BOOLEAN SS_TermpaarSubsummiertVonGM (TermT links, TermT rechts);
BOOLEAN SS_TermpaarSubsummiertVonPermsub (TermT links, TermT rechts, unsigned short Grad);
BOOLEAN SS_TermpaarSubsummiertVonGMUndPermsub (TermT links, TermT rechts, unsigned short Grad);

/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar bzgl. einzelner G subsummieren ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* Liefert TRUE, wenn die gleichung das Termpaar links, rechts subsummiert.
   Dabei werden alle Subgleichungen des Termpaar berechnet und getestet, und zwar in beiden Richtungen.*/

BOOLEAN SS_GleichungSubsummiertTermpaar (GleichungsT gleichung, TermT links, TermT rechts);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------ Termpaar bzgl. einzelner G subsummieren (G liegt als Termpaar vor) -----------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* Liefert TRUE, wenn die gleichung Glinks = Grechts das Termpaar links, rechts subsummiert.
   Dabei werden alle Subgleichungen des Termpaar berechnet und getestet, und zwar in beiden Richtungen.*/

BOOLEAN SS_TermpaarSubsummiertTermpaar (TermT Glinks, TermT Grechts, TermT links, TermT rechts);


#ifdef __cplusplus
}
#endif

#endif
