/*=================================================================================================================
  =                                                                                                               =
  =  Termpaare.h                                                                                                  =
  =                                                                                                               =
  =  Realisierung von Termtupeln                                                                                  =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.95 Thomas.                                                                                             =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef TERMPAARE_H
#define TERMPAARE_H


#include "general.h"
#include "SpeicherVerwaltung.h"
#include "Subgleichungen.h"
#include <limits.h>  /* fuer LONG_MAX */



#ifdef __cplusplus
extern "C" {
#endif

#define TP_PETERGRAF 0                                                                       /* Alternativen: 0, 1 */


/*=================================================================================================================*/
/*= I. Typen                                                                                                      =*/
/*=================================================================================================================*/

typedef enum {weiss=0, blau=1, gelb=2, gruen=3} FarbT;                      /* Faerbung fuer Interreduktion der GM */

typedef struct GNTermpaarTS {
  TermT               linkeSeite,
                      rechteSeite;
#ifdef REFEREES_NEEDED
  unsigned long int            ref_red_cp, ref_red_right, ref_red_left, ref_red_equ, 
    ref_red_goal, ref_cp_count, ref_subsum;
  long int ref_weight;
#endif
  long int            Nummer;
  void               *SUEInfo;

#if COMP_PCL
  PCL_ObjektNummerT   pclNr;
#endif

  struct KnotenTS     *Blatt;
  BOOLEAN             RichtungAusgezeichnet, 
                      IstMonogleichung;
  FarbT               Faerbung;
  struct GNTermpaarTS *Gegenrichtung,
                      *Vorg,
                      *Nachf;
  SGM_ReferenzT       SubsumRef;
#if TP_PETERGRAF
  struct GNTermpaarTS *PGVorg,
                      *PGNachf;
#endif
} *GNTermpaarT;



/*=================================================================================================================*/
/*= II. Speicherverwaltung                                                                                        =*/
/*=================================================================================================================*/

extern SV_SpeicherManagerT TP_FreispeicherManager;

#define /*void*/ TP_GNTermpaarHolen(/*L-Ausdruck*/ Index)                                                           \
  SV_Alloc(Index,TP_FreispeicherManager,GNTermpaarT)

#define /*void*/ TP_GNTermpaarLoeschen(/*GNTermpaarT*/ Termpaar)                                                    \
  SV_Dealloc(TP_FreispeicherManager,(Termpaar))

#define /*void*/ TP_LinksRechtsLoeschen(/*GNTermpaarT*/ Termpaar)                                                   \
  TO_TermeLoeschen(TP_LinkeSeite(Termpaar),TP_RechteSeite(Termpaar))



/*=================================================================================================================*/
/*= III. Lesender Zugriff                                                                                         =*/
/*=================================================================================================================*/

#define /*BOOLEAN*/ TP_RichtungAusgezeichnet(/*GNTermpaarT*/ Termpaar)                                              \
  ((Termpaar)->RichtungAusgezeichnet)

#define /*BOOLEAN*/ TP_RichtungUnausgezeichnet(/*GNTermpaarT*/ Termpaar)                                            \
  (!(Termpaar)->RichtungAusgezeichnet)

#define /*BOOLEAN*/ TP_IstMonogleichung(/*GNTermpaarT*/ Termpaar)                                                   \
  ((Termpaar)->IstMonogleichung)

#define /*BOOLEAN*/ TP_IstKeineMonogleichung(/*GNTermpaarT*/ Termpaar)                                              \
  (!(Termpaar)->IstMonogleichung)

#define /*FarbT*/ TP_IstWeiss(/*GNTermpaarT*/ Termpaar)                                                             \
  ((Termpaar)->Faerbung == weiss)

#define /*FarbT*/ TP_IstBlau(/*GNTermpaarT*/ Termpaar)                                                              \
  ((Termpaar)->Faerbung == blau)

#define /*FarbT*/ TP_IstGelb(/*GNTermpaarT*/ Termpaar)                                                              \
  ((Termpaar)->Faerbung == gelb)

#define /*FarbT*/ TP_IstGruen(/*GNTermpaarT*/ Termpaar)                                                             \
  ((Termpaar)->Faerbung == gruen)

#define /*GNTermpaarT*/ TP_Gegenrichtung(/*GNTermpaarT*/ Termpaar)                                                  \
  ((Termpaar)->Gegenrichtung)

#define /*TermT*/ TP_LinkeSeite(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->linkeSeite)

#define /*TermT*/ TP_RechteSeite(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->rechteSeite)

#define /*void**/ TP_SUEInfo(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->SUEInfo)

#define /*unsigned*/ TP_ElternNr(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->Nummer)

#define /*SGM_ReferenzT*/ TP_SubgleichungsReferenz(/*GNTermpaarT*/ Termpaar)                                        \
  ((Termpaar)->SubsumRef)

#if COMP_PCL
#define /*PCL_ObjektNummerT**/ TP_PclID(/*GNTermpaarT*/ Termpaar)                                                   \
  (&((Termpaar)->pclNr))
#else
#define /*PCL_ObjektNummerT**/ TP_PclID(/*GNTermpaarT*/ Termpaar) 0
#endif

#if COMP_PCL
#define /*PCL_ObjektNummerT*/ TP_PclIDAs(/*GNTermpaarT*/ Termpaar)                                                   \
  ((Termpaar)->pclNr)
#else
#define /*PCL_ObjektNummerT*/ TP_PclIDAs(/*GNTermpaarT*/ Termpaar) 0
#endif

#define /*GNTermpaarT*/ TP_Vorg(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->Vorg)

#define /*GNTermpaarT*/ TP_Nachf(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->Nachf)

#if TP_PETERGRAF
#define /*GNTermpaarT*/ TP_PGVorg(/*GNTermpaarT*/ Termpaar)                                                         \
  ((Termpaar)->PGVorg)

#define /*GNTermpaarT*/ TP_PGNachf(/*GNTermpaarT*/ Termpaar)                                                        \
  ((Termpaar)->PGNachf)
#endif

#ifdef REFEREES_NEEDED
#define TP_RedCp(RE) (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_red_cp):(RE->Gegenrichtung->ref_red_cp)):\
                       (RE->ref_red_cp))
#define TP_RedRight(RE) (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_red_right):(RE->Gegenrichtung->ref_red_right)):\
                       (RE->ref_red_right))
#define TP_RedLeft(RE)  (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_red_left):(RE->Gegenrichtung->ref_red_left)):\
                       (RE->ref_red_left))
#define TP_RedEqu(RE)   (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_red_equ):(RE->Gegenrichtung->ref_red_equ)):\
                       (RE->ref_red_equ))
#define TP_RedGoal(RE)  (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_red_goal):(RE->Gegenrichtung->ref_red_goal)):\
                       (RE->ref_red_goal))
#define TP_CpCount(RE)  (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_cp_count):(RE->Gegenrichtung->ref_cp_count)):\
                       (RE->ref_cp_count))
#define TP_Subsum(RE)    (RE->Gegenrichtung?\
                       (RE->RichtungAusgezeichnet?(RE->ref_subsum):(RE->Gegenrichtung->ref_subsum)):\
                       (RE->ref_subsum))
#define TP_Weight(RE)    (RE->Gegenrichtung?\
                           (RE->RichtungAusgezeichnet?(RE->ref_weight):(RE->Gegenrichtung->ref_weight)):\
                           (RE->ref_weight))
#else
#define TP_RedCp(RE)   0
#define TP_RedRight(RE)0
#define TP_RedLeft(RE) 0
#define TP_RedEqu(RE)  0
#define TP_RedGoal(RE) 0
#define TP_CpCount(RE) 0
#define TP_Subsum(RE)  0
#define TP_Weight(RE)  0
#endif

/*=================================================================================================================*/
/*= IV. Schreibender Zugriff                                                                                      =*/
/*=================================================================================================================*/

#define /*void*/ TP_SubgleichungsReferenzSei(/*GNTermpaarT*/ Termpaar, /*SGM_ReferenzT*/ Referenz)                  \
  ((Termpaar)->SubsumRef = Referenz)

#define /*void*/ TP_GegenrichtungBesetzen(/*GNTermPaarT*/ Termpaar, /*GNTermPaarT*/ gegenrichtung)                  \
  (Termpaar)->Gegenrichtung = gegenrichtung

#define /*GNTermpaarT*/ TP_VorgSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Vorgaenger)                            \
  ((Termpaar)->Vorg = Vorgaenger)

#define /*GNTermpaarT*/ TP_NachfSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Nachfolger)                           \
  ((Termpaar)->Nachf = Nachfolger)

#define /*FarbT*/ TP_WeissMalen(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->Faerbung = weiss)

#define /*FarbT*/ TP_BlauMalen(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->Faerbung += blau)

#define /*FarbT*/ TP_GelbMalen(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->Faerbung += gelb)

#if TP_PETERGRAF
#define /*GNTermpaarT*/ TP_PGVorgSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Vorg)                                \
  ((Termpaar)->PGVorg = Vorg)

#define /*GNTermpaarT*/ TP_PGNachfSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Nachf)                              \
  ((Termpaar)->PGNachf = Nachf)
#endif

#ifdef REFEREES_NEEDED
#define TP_IncRedCp(RE) (RE->Gegenrichtung? (RE->RichtungAusgezeichnet?(RE->ref_red_cp++):\
                                                (RE->Gegenrichtung->ref_red_cp++)): (RE->ref_red_cp++))
#define TP_IncRedRight(RE) (RE->Gegenrichtung?(RE->RichtungAusgezeichnet?(RE->ref_red_right++):\
                                               (RE->Gegenrichtung->ref_red_right++)):(RE->ref_red_right++))
#define TP_IncRedLeft(RE)  (RE->Gegenrichtung?(RE->RichtungAusgezeichnet?(RE->ref_red_left++):\
                                               (RE->Gegenrichtung->ref_red_left++)):(RE->ref_red_left++))
#define TP_IncRedEqu(RE)   (RE->Gegenrichtung? (RE->RichtungAusgezeichnet?(RE->ref_red_equ++):\
                                                (RE->Gegenrichtung->ref_red_equ++)):(RE->ref_red_equ++))
#define TP_IncRedGoal(RE)  (RE->Gegenrichtung? (RE->RichtungAusgezeichnet?(RE->ref_red_goal++):\
                                                (RE->Gegenrichtung->ref_red_goal++)):(RE->ref_red_goal++))
#define TP_IncCpCount(RE)  (RE->Gegenrichtung? (RE->RichtungAusgezeichnet?(RE->ref_cp_count++):\
                                                (RE->Gegenrichtung->ref_cp_count++)):(RE->ref_cp_count++))
#define TP_IncSubsum(RE)    (RE->Gegenrichtung?(RE->RichtungAusgezeichnet?(RE->ref_subsum++):\
                                                (RE->Gegenrichtung->ref_subsum++)): (RE->ref_subsum++))
#define TP_SetWeight(RE,weight)(RE->Gegenrichtung?(RE->RichtungAusgezeichnet?(RE->ref_weight=weight):\
                                          (RE->Gegenrichtung->ref_weight=weight)): (RE->ref_weight=weight))
#define TP_SetWeightZero(RE)(RE->Gegenrichtung?(RE->RichtungAusgezeichnet?                                                                                                                                                                               \
(RE->ref_weight=RE->ref_red_cp=RE->ref_red_right=RE->ref_red_left=RE->ref_red_equ=RE->ref_red_goal=RE->ref_cp_count=RE->ref_subsum=0):                                                                                                                          \
(RE->Gegenrichtung->ref_weight=RE->Gegenrichtung->ref_red_cp=RE->Gegenrichtung->ref_red_right=RE->Gegenrichtung->ref_red_left=RE->Gegenrichtung->ref_red_equ=RE->Gegenrichtung->ref_red_goal=RE->Gegenrichtung->ref_cp_count=RE->Gegenrichtung->ref_subsum=0)): \
(RE->ref_weight=RE->ref_red_cp=RE->ref_red_right=RE->ref_red_left=RE->ref_red_equ=RE->ref_red_goal=RE->ref_cp_count=RE->ref_subsum=0))

#else
#define TP_IncRedCp(RE)    
#define TP_IncRedRight(RE) 
#define TP_IncRedLeft(RE)  
#define TP_IncRedEqu(RE)   
#define TP_IncRedGoal(RE)  
#define TP_IncCpCount(RE)  
#define TP_IncSubsum(RE)   
#define TP_SetWeight(RE,weight)   
#endif


/*=================================================================================================================*/
/*= V. Praedikate                                                                                                 =*/
/*=================================================================================================================*/

#define /*BOOLEAN*/ TP_TermpaarIstRegel(/*GNTermpaarT*/ Termpaar)                                                   \
  !((Termpaar)->Gegenrichtung)

#define /*BOOLEAN*/ TP_TermpaareGleich(/*GNTermpaarT*/ Termpaar1, /*GNTermpaarT*/ Termpaar2)                        \
  (TO_TermeGleich(TP_LinkeSeite(Termpaar1),TP_LinkeSeite(Termpaar2))                                                \
  && TO_TermeGleich(TP_RechteSeite(Termpaar1),TP_RechteSeite(Termpaar2)))



/*=================================================================================================================*/
/*= VI. Initialisierung                                                                                           =*/
/*=================================================================================================================*/

void TP_InitApriori(void);


#ifdef __cplusplus
}
#endif

#endif /* Termpaare_H */
