/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - TermpairRepresentation.h                                                                                   */
/* - ************************                                                                                   */
/* -                                                                                                            */
/* - Schnittstelle fuer das Set of unselected equations.                                                        */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/


#ifndef TERMPAIRREPRESENTATION_H
#define TERMPAIRREPRESENTATION_H

#include "InfoStrings.h"
#include "TermOperationen.h"
#include "RUndEVerwaltung.h"

#include "SUEBasis.h"

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize-------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void TPR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN TPR_InitAposteriori (InfoStringT DEF, InfoStringT USR);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* TPR_NewFlatsGeneratedProc) (void);

#if COMP_PCL
#define TPR_NewFlatsGenerated() ((* TPR_NewFlatsGeneratedProc) ())
#else
#define TPR_NewFlatsGenerated()
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern TermpairT (* TPR_Flat2TPProc) (TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT, PCL_ObjektNummerT);

#define TPR_FLATTERMStoTERMPAIR(lhs, rhs, actualParent, otherParent, actualPosition, pclID)          \
( (* TPR_Flat2TPProc) (lhs, rhs, actualParent, otherParent, actualPosition, pclID))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* TPR_TP2ParIntermedProc) (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT, PCL_ObjektNummerT *);
extern void (* TPR_TP2ParFinalProc) (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT, PCL_ObjektNummerT *);

#define TPR_TERMPAIRtoPARALLEL_INTERMED(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr)\
( (* TPR_TP2ParIntermedProc) (tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr))

#define TPR_TERMPAIRtoPARALLEL_FINAL(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr)\
( (* TPR_TP2ParFinalProc) (tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr))


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* TPR_Par2FlatProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

#define TPR_PARALLELtoFLATTERMS(tp, lhs, rhs, actualParent, otherParent)\
( (* TPR_Par2FlatProc) (tp, lhs, rhs, actualParent, otherParent))
/* Destroys the termpair-representation irrecoverable */


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern TermpairT (* TPR_Par2TPProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, BOOLEAN, PCL_ObjektNummerT);

/* Achtung: tp wird evtl. veraendert, deswegen unbedingt mit
     tp = TPR_PARALLELtoTERMPAIR (tp,...) aufrufen */
#define TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualParent, otherParent, changed, pclID)\
( (* TPR_Par2TPProc) (tp, lhs, rhs, actualParent, otherParent, changed, pclID))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


extern void      (* TPR_DestroyTPProc)  (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);
extern void      (* TPR_DestroyParProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

#define /*void*/ TPR_DESTROYTERMPAIR(tp, actualParent, otherParent)\
( (* TPR_DestroyTPProc)  (tp, actualParent, otherParent))
/* Termpaar zerstoeren */

#define /*void*/ TPR_DESTROYPARALLEL(tp, lhs, rhs, actualParent, otherParent)\
( (* TPR_DestroyParProc)  (tp, lhs, rhs, actualParent, otherParent))
/* Termpaar und Flachterme zerstoeren */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- TermVerfolgung --------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* TPR_TermOperationBeiVerfolgungProc) (void (*) (TermT, TermT), TermpairT, 
                                                    RegelOderGleichungsT, RegelOderGleichungsT);

#define TPR_OperationWaehrendTermverfolgung(Operation, tp, father, mother)\
( (* TPR_TermOperationBeiVerfolgungProc) (Operation, tp, father, mother))

/* ------------------------------------------------------------------------------------------------------*/
/* ------- Unambiguous Comparison -----------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

extern BOOLEAN (* TPR_UnambiguousGreaterThanProc) (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT, 
                                                   TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);

#define TPR_UnambiguousGreaterThan(tp1, father1, mother1, tp2, father2, mother2)                                \
( (* TPR_UnambiguousGreaterThanProc) (tp1, father1, mother1, tp2, father2, mother2))


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void TPR_Statistics(InfoStringT IS);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Output ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* TPR_PrintTermpairProc) (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT, const char *);

#define TPR_PrintTermpair(tp, actualParent, otherParent, trenner)\
((* TPR_PrintTermpairProc) (tp, actualParent, otherParent, trenner))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define TPR_CleanUp()


#ifdef __cplusplus
}
#endif

#endif /* TERMPAIRREPRESENTATION_H */
