/*=================================================================================================================
  =                                                                                                               =
  =  Unifikation1.h                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =  07.10.94  Thomas.                                                                                            =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef UNIFIKATION1_H
#define UNIFIKATION1_H

#include "general.h"
#include "TermOperationen.h"
#include "RUndEVerwaltung.h"


#ifdef __cplusplus
extern "C" {
#endif



/*=================================================================================================================*/
/*= I. Variablenbank fuer Substitutionen; Stack angelegter Bindungen; sonstige globale Groessen                   =*/
/*=================================================================================================================*/

void U1_VariablenfelderAnpassen(SymbolT NeuesteBaumvariable);
/* Variablenbank und Stapel angelegter Bindungen so anpassen,
   dass Variablen x(1), x(2), ... , x(2*BO_NeuesteBaumvariable()) verdaut werden */


/*=================================================================================================================*/
/*= II. Operationen auf oder mit Substitutionen                                                                   =*/
/*=================================================================================================================*/

void U1_DruckeMGU(void);


/*=================================================================================================================*/
/*= III. Operationen zum flachen Unifizieren                                                                      =*/
/*=================================================================================================================*/

BOOLEAN U1_Unifiziert(UTermT Term1, UTermT Term2);
/* Sind Term1 und Term2 unifizierbar? Aufzurufen nur dann, wenn xU trivial */


/*=================================================================================================================*/
/*= XVI. KP-Bildung zu neuer Regel                                                                                =*/
/*=================================================================================================================*/

void U1_KPsBildenZuRegelFailing(RegelT Regel);

void U1_KPsBildenZuRegelUnfailing(RegelT Regel);


/*=================================================================================================================*/
/*= XVII. KP-Bildung zu neuer Gleichung                                                                           =*/
/*=================================================================================================================*/

void U1_KPsBildenZuGleichung(GleichungsT Gleichung);


/*=================================================================================================================*/
/*= XVIII. Rekonstruktion von kritischen Paaren                                                                   =*/
/*=================================================================================================================*/

void U1_KPRekonstruieren(RegelOderGleichungsT Vater, unsigned int Stellennummer, RegelOderGleichungsT Mutter,
/* Die Stellennummer wird ab Null gezaehlt. */                                      TermT *KPLinks, TermT *KPRechts);

void U1_KPAusTermenRekonstruieren(TermT VaterLinks, TermT VaterRechts, TermT MutterLinks, TermT MutterRechts, 
   unsigned int Stellennummer, TermT *KPLinks, TermT *KPRechts);                           /* dito  ohne Termpaare */


/*=================================================================================================================*/
/*= XIX. Mass fuer relative Unifizierbarkeit                                                                      =*/
/*=================================================================================================================*/

unsigned long int U1_Unifikationsmass(TermT links, TermT rechts);
/* links und rechts muessen vom Typ TermT sein. Rueckgabe 0 <=> Terme sind unifizierbar. */


/*=================================================================================================================*/
/*= XX. Statistik                                                                                                 =*/
/*=================================================================================================================*/

void U1_Statistik(InfoStringT InfoString);


/*=================================================================================================================*/
/*= XXI. Initialisierung                                                                                          =*/
/*=================================================================================================================*/

void U1_InitApriori(void);

void U1_ParamInfo(ParamInfoArtT what, char* KapurString);

void U1_InitAposteriori(char *KapurString);



#ifdef __cplusplus
}
#endif

#endif
