/**********************************************************************
* Filename : SpeicherVerwaltung.c
* Funktionen fuer Freispeicherverwaltung
*
* Autoren : Max Wolf, erstellt am : 20.01.94
*
* Am 03. und 04.03.94 kopiert und angepasst. Arnim
* Am 09.03. Speicher-Statistik eingefuehrt.
*
* 17.02.95  Wesentliche Teile fertig entmaxt.
*           Debug-Betrieb verbessert. Thomas.
* 14.03.95  Nested-Betrieb fertiggestellt. Thomas.
*
**********************************************************************/

#include <string.h>

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "DomaenenMemorium.h"
#include "FehlerBehandlung.h"
#include "ManagerMemorium.h"
#include "SpeicherVerwaltung.h"
#include "WaldmeisterII.h"


/*=================================================================================================================*/
/*= II. Typdeklarationen;  Globalitaeten                                                                          =*/
/*=================================================================================================================*/

volatile DomaenenT SV_Domaene;

const size_t ALIGN_SIZE = sizeof(struct{char c; int i;}) - sizeof(int);

#define StichAdrPlus(/*void**/ Zeiger,/*long int*/ Delta)                                                           \
 ((StichAdressT)((ADDRESS) (Zeiger) + Delta))

static SV_SpeicherManagerT *ManagerListe;



/*=================================================================================================================*/
/*= III. SV_Speichertest                                                                                          =*/
/*=================================================================================================================*/

#if SV_TEST

#define /*unsigned long int*/ KleinsteAdresseImBlock(/*SpeicherblockT*/ Speicherblock, /*SV_SpeicherManagerT*/ mbm) \
  ((uintptr_t) &(Speicherblock->mem_block))

#define /*unsigned long int*/ GroessteAdresseImBlock(/*SpeicherblockT*/ Speicherblock, /*SV_SpeicherManagerT*/ mbm) \
  (((uintptr_t) &(Speicherblock->next)) + mbm->mbl_size - 1)

#define /*BOOLEAN*/ AdresseImBlock(/*void**/ Adresse, /*SpeicherblockT*/ Speicherblock, /*SV_SpeicherManagerT*/ mbm)\
  (KleinsteAdresseImBlock(Speicherblock,mbm)<=(uintptr_t) Adresse                                           \
    && (uintptr_t) Adresse<=GroessteAdresseImBlock(Speicherblock,mbm))

static void CheckManagers(void* Adresse)     /* Prueft bei allen bisher fuer die Speichermanager angelegten Bloek- */
{                    /* ken, ob adresse innerhalb eines solchen Blockes liegt. Wenn ja: Fehlermeldung und Abbruch. */
 SV_SpeicherManagerT *runny = ManagerListe;
  while (runny) {
    SpeicherblockT *Index = runny->block_list;
    while (Index) {
      if (AdresseImBlock(Adresse,Index,runny))
        our_error("Fehlerhafte Adresse in CheckManagers erkannt!");
      Index = Index->next;
    } 
    runny = runny->Nachf;
  }
}

void check_address(SV_SpeicherManagerT *mbm, void *Adresse)
{
  BOOLEAN AdresseGefunden = FALSE;
  SV_SpeicherManagerT *runny = ManagerListe;
  while (runny) {
    SpeicherblockT *Index = runny->block_list;
    while (Index) {
      if (AdresseImBlock(Adresse,Index,runny))
        if (runny!=mbm)
          our_error("Fehlerhafte Adresse in check_address erkannt! (1)");
        else
          AdresseGefunden = TRUE;
      Index = Index->next;
    }
    if (runny==mbm && !AdresseGefunden)
      our_error("Fehlerhafte Adresse in check_address erkannt! (2)");
    runny = runny->Nachf;
  }
  if (!AdresseGefunden)
    our_error("Fehlerhafte Adresse in check_address erkannt! (3)");
}


#else

#define /*void*/ CheckManagers(/*void**/ adresse)

#endif



/*=================================================================================================================*/
/*= IV. Zugriffe auf die Systemspeicherverwaltung                                                                 =*/
/*=================================================================================================================*/

void *our_alloc (size_t size)
{
  void *rescue = (void *) malloc(size);
  if (!rescue)
    our_out_of_mem_error("\nSorry, no more memory!\n\nBye...\n\n");
  CheckManagers(rescue);
  CheckManagers((ADDRESS)rescue+size-1);
  return rescue;
}



void our_dealloc (void *memory_ptr)
{
  CheckManagers(memory_ptr);
  free(memory_ptr);
}


void *our_realloc (void *memory_ptr,                 /* Zeiger auf den "alten" Speicherbereich (als VAR-Parameter) */
                   size_t size,                                                                    /* neue Groesse */
                   const char *whatfor)                                    /* Speicher wofuer (fuer Fehlermeldung) */
{
  void *rescue;
  CheckManagers(memory_ptr);
  rescue = (void *) realloc (memory_ptr, size);
  if (!rescue) {
    fprintf(IO_Strom, "Out of memory during call to realloc.");
    our_out_of_mem_error(whatfor);
  }
  CheckManagers(rescue);
  CheckManagers((ADDRESS)rescue+size-1);
  return rescue;
}



/*=================================================================================================================*/
/*= V., VI. Funktionen fuer gewoehnlichen und fuer Nested-Speichermanager                                         =*/
/*=================================================================================================================*/

#define /*unsigned long int*/ CeilQuot(/*unsigned long int*/ a, /*unsigned long int*/ b)                            \
  ( (b*(a/b)==a) ? a/b : a/b+1)                                                              /* Wert: ceiling(a/b) */


void SV_BeliebigenManagerInitialisieren(SV_SpeicherManagerT *Manager, size_t element_size,
  unsigned long int block_size, BOOLEAN Nested, const char Name[] MAYBE_UNUSEDPARAM)
/* Manager ist Zeiger auf den zu initialisierenden Manager.
   element_size gibt die Groesse der zu speichernden Elemente in Bytes an. 
   block_size gibt die Anzahl der Elemente fuer einen Speicherblock an.
   Name ist der fuer die Statistik eingerichtete Name des Managers.
   Die Funktion ist mehrfachaufruf-clean in dem Sinn, dass weitere Aufrufe auf denselben Manager ignoriert werden.
   Ein Manager gilt genau dann als initialisiert, wenn die Komponente block_size einen von 0 verschiedenen Wert 
   hat. Folglich sind als statische Variablen deklarierte Manager initial nicht initialisiert. Bei dynamisch 
   allokierten Managern muss die Durchfuehrung der Initialisierung dagegen durch eine explizite Zuweisung des 
   Wertes 0 an block_size erzwungen werden.
*/
{
  if (!Manager->block_size) {
    size_t align_rest;
    Manager->Nested = Nested;                                                 /* Flag fuer Kategorie Nested setzen */
    if ((align_rest = element_size % ALIGN_SIZE))                                                     /* Alignment */
      element_size += ALIGN_SIZE - align_rest;
    if ((Manager->block_size = block_size) < 2)                     /* sonst keine Stichpunkte-Liste einschreibbar */
      our_error("Fehler: block_size muss mindestens zwei betragen.");

    Manager->element_size = element_size;
#ifdef SV_BASIS_MMAP
    {                                                        /* Mindestens erforderliche Groesse fuer einen Block. */
      size_t Sollgroesse = sizeof(struct mem_block_list *)+element_size*block_size,         /* Aufrunden auf Soft- */
             Zielgroesse = DM_Seitengroesse()*CeilQuot(Sollgroesse,DM_Seitengroesse());             /* ware-Seiten */
      Manager->mbl_size = Zielgroesse;
      Manager->block_size = (Zielgroesse-sizeof(struct mem_block_list *))/element_size;
      MM_InitManager(Manager,SV_Domaene);                                     /* Dort wird auch next_free besetzt. */
    }
#else
    Manager->mbl_size = sizeof(struct mem_block_list *)+element_size*block_size;
    Manager->next_free = NULL;
#endif
    Manager->block_list = NULL;
    Manager->Nachf = ManagerListe;
    ManagerListe = Manager;
    if_sst({
      Manager->Name = Name;
      Manager->Anforderungen =
      Manager->Freigaben =
      Manager->InsgesamtAngelegteBloecke =
      Manager->MaximalGleichzeitigBelegteElemente = 0;
    });
  }
}


void *SV_AllocMitNeuemBlock(SV_SpeicherManagerT *mbm)
{                              /* Anlegen eines neuen Blocks von Elementen, Rueckgabe des ersten freien Elementes. */
  void *first;
  first = SV_Blockdurchzeigerung(mbm,our_alloc(mbm->mbl_size),NULL);
  return first;
}


void *SV_Blockdurchzeigerung(SV_SpeicherManagerT *mbm, void *StartNeuerBlock, void *Waechter)
{
  StichAdressT first = (StichAdressT) &((SpeicherblockT *)StartNeuerBlock)->mem_block,           /* erstes Element */
               pastlast = StichAdrPlus(first,mbm->block_size*mbm->element_size),         /* hinter letztem Element */
               IndexVorg = StichAdrPlus(first,mbm->element_size),                               /* zweites Element */
               Index = StichAdrPlus(IndexVorg,mbm->element_size);                               /* drittes Element */
  if_sst(mbm->InsgesamtAngelegteBloecke++);
  ((SpeicherblockT *)StartNeuerBlock)->next = mbm->block_list;
  mbm->block_list = StartNeuerBlock;
  mbm->next_free = IndexVorg;   /* globales next_free = 2. Element. Fuer Nesting unnoetig; sonst wichtig, wichtig! */
  check_address(mbm,first);
  first->next_free = IndexVorg;                                  /* fuer Nesting wichtig, wichtig! Sonst unnoetig. */
  while (Index<pastlast)                                         /* next_free Zeiger im neuen Block initialisieren */
    Index = StichAdrPlus(IndexVorg = IndexVorg->next_free = Index,mbm->element_size);
  IndexVorg->next_free = Waechter;                   /* Nachfolger des letzten Einstichpunktes je nach Betriebsart */
  return first;
}



/*=================================================================================================================*/
/*= VII. Leeren der Freispeicherverwaltung                                                                        =*/
/*=================================================================================================================*/

void SV_AlleManagerLeeren(void)              /* Inhalt aller Manager weggeben, aber Verwaltungsstrukturen behalten */
{
#ifdef SV_BASIS_MMAP
  MM_ManagerdomaeneLeeren(SV_Domaene,ManagerListe);
#else
  SV_SpeicherManagerT *Index = ManagerListe;
  while (Index) {
    SpeicherblockT *Index2 = Index->block_list;                   /* alle Speicherbloecke dieses Managers weggeben */
    while (Index2) {
      SpeicherblockT *Index2Nachf = Index2->next;
      our_dealloc(Index2);
      Index2 = Index2Nachf;
    }
    Index->block_list = NULL;                                                             /* keine Bloecke mehr da */
    Index->next_free = NULL;                                                     /* und keine Speicherstueckelchen */
    Index = Index->Nachf;                                                           /* naechsten Manager vornehmen */
  }
#endif
  if_sst({
    SV_SpeicherManagerT *Manager = ManagerListe;
    while (Manager) {
      Manager->Anforderungen =
      Manager->Freigaben =
      Manager->InsgesamtAngelegteBloecke =
      Manager->MaximalGleichzeitigBelegteElemente = 0;
      Manager = Manager->Nachf;
    }
  });
}


/*=================================================================================================================*/
/*= VIII. Initialisierung und Statistik                                                                           =*/
/*=================================================================================================================*/

void SV_InitApriori(void) {
  if (WM_ErsterAnlauf()) {
#ifdef SV_BASIS_MMAP
    SV_Domaene = MM_NeueManagerdomaene(DM_SV,"Domain for memory managers");
#endif
#if SV_TEST
    malloc_debug(2);
#endif
  }
}



unsigned long int SV_AnzNochBelegterElemente(SV_NestedSpeicherManagerT mbm MAYBE_UNUSEDPARAM)
{
#if DM_STATISTIK
  if (mbm.Nested) {
    StichAdressT StichpunktIndex = mbm.next_free;
    unsigned long int AnzFreierElemente = 0;
#ifdef SV_BASIS_MMAP
    while (StichpunktIndex!=mbm.Sperre_next_free)                          /* Laenge der next_free-Liste ermitteln */
#else
    while (StichpunktIndex)                                                /* Laenge der next_free-Liste ermitteln */
#endif
    {
      AnzFreierElemente++;
      StichpunktIndex = StichpunktIndex->next_free;
    }                                                  /* und von der Laenge der maximal moeglichen Liste abziehen */
  return mbm.InsgesamtAngelegteBloecke*mbm.block_size - AnzFreierElemente;
  }
  else
    return mbm.Anforderungen-mbm.Freigaben;
#else
  return 0;
#endif
}


static void DruckeSpeicherManagerStatistik(int, const char*, unsigned long, unsigned long, unsigned long,
  unsigned long, unsigned long, unsigned long) MAYBE_UNUSEDFUNC;
static void DruckeSpeicherManagerStatistik(int Grad, const char *Name, unsigned long Blockgroesse, 
  unsigned long ElementGroesse, unsigned long Anforderungen, unsigned long Freigaben,
  unsigned long InsgesamtAngelegteBloecke, unsigned long MaximalGleichzeitigBelegteElemente )
/* Grad: 0 - Ausgabe nur, wenn Manager benutzt wurde und Diskrepanzen
         1 - Ausgabe, wenn Manager benutzt wurde, unabhaengig von Diskrepanzen
*/
{
  BOOLEAN Ausgeben = FALSE;
  
  switch (Grad){
  case 0:
    Ausgeben = ((Anforderungen+Freigaben>0) && (Anforderungen != Freigaben));
    break;
  case 1:
    Ausgeben = (Anforderungen+Freigaben>0);
    break;
  default:
    our_error("Falscher Ausgabe-Grad in IO_DruckeSpeicherManagerStatistik\n");
    break;
  }
  if (Ausgeben){
    IO_DruckeKurzUeberschrift(Name, '+');
    IO_DruckeStatistikZeile("Block size", Blockgroesse);
    IO_DruckeStatistikZeile("Element size", ElementGroesse);
    IO_DruckeStatistikZeile("Total number of allocations",Anforderungen );
    IO_DruckeStatistikZeile("Total number of deallocations",Freigaben );
    IO_DruckeStatistikZeile("Total number of allocated blocks",InsgesamtAngelegteBloecke );
    IO_DruckeStatistikZeile("Max. number of elements used at the same time",MaximalGleichzeitigBelegteElemente );
  }
}


static void DruckeNestedSpeicherManagerStatistik(const char*, unsigned long, unsigned long, unsigned long,
  BOOLEAN, unsigned long, unsigned long) MAYBE_UNUSEDFUNC;
static void DruckeNestedSpeicherManagerStatistik(const char *Name, unsigned long Blockgroesse,
  unsigned long ElementGroesse, unsigned long Anforderungen, BOOLEAN Fehlerfall, 
  unsigned long AnzNochBelegterElemente, unsigned long InsgesamtAngelegteBloecke )
{
  char VerlaengerterName[SV_maxLaengeNameNestedManager];
  strcpy(VerlaengerterName,Name);
  strcat(VerlaengerterName," (Nested-Manager)");
  IO_DruckeKurzUeberschrift(VerlaengerterName, '+');
  IO_DruckeStatistikZeile("Block size", Blockgroesse);
  IO_DruckeStatistikZeile("Element size", ElementGroesse);
  IO_DruckeStatistikZeile("Total number of allocations",Anforderungen );
  if (Fehlerfall)
    IO_DruckeStatistikZeileString("Remaining number of allocated elements","nicht ermittelbar");
  else
    IO_DruckeStatistikZeile("Remaining number of allocated elements",AnzNochBelegterElemente );
  IO_DruckeStatistikZeile("Total number of allocated blocks",InsgesamtAngelegteBloecke );
  IO_DruckeStatistikZeile("Max. number of elements usable at the same time",Blockgroesse*InsgesamtAngelegteBloecke);
}


void SV_Statistik(BOOLEAN Fehlerfall MAYBE_UNUSEDPARAM, BOOLEAN NurDiskrepanzen MAYBE_UNUSEDPARAM, InfoStringT IS)   
{                                                                /* Gibt die Statistik der Speicherverwaltung aus. */
  
  if (IS_IsIn("complete", IS) || IS_IsIn("SV", IS)){
#if DM_STATISTIK
    BOOLEAN UeberschriftGedruckt = FALSE;
    unsigned long GesamtAnforderungen = 0,
                  GesamtFreigaben = 0,
                  GesamtAngelegteBloecke = 0,
                  GesamtBelegterSpeicher = 0;
                  SV_SpeicherManagerT *runny = ManagerListe;
      
    if (!NurDiskrepanzen) {
      UeberschriftGedruckt = TRUE;
      IO_DruckeUeberschrift("Memory statistics");
    }
      
    while (runny) {
      SV_SpeicherManagerT mbm = *runny;
      if (mbm.Nested) {
        if (Fehlerfall) {
	  if (!UeberschriftGedruckt) {
	    IO_DruckeKurzUeberschrift("Discrepancies in memory allocation", '*');
	    UeberschriftGedruckt = TRUE;
	  }
          DruckeNestedSpeicherManagerStatistik(mbm.Name, mbm.block_size, mbm.element_size, 
                                                   mbm.Anforderungen, TRUE, 0, mbm.InsgesamtAngelegteBloecke);
	} else {
          unsigned long int AnzNochBelegterElemente = SV_AnzNochBelegterElemente(mbm);
          GesamtFreigaben += mbm.Anforderungen - AnzNochBelegterElemente;
          /* Hier kann nur festgestellt werden, ob noch nicht freigegebene Elemente in den Bloecken stehen */
          if (AnzNochBelegterElemente || FALSE)
            DruckeNestedSpeicherManagerStatistik(mbm.Name, mbm.block_size, mbm.element_size, mbm.Anforderungen,
                                                 FALSE, AnzNochBelegterElemente, mbm.InsgesamtAngelegteBloecke);
        }
      }
      else {
        int AusgabeGrad = NurDiskrepanzen?0:1;
        GesamtFreigaben += mbm.Freigaben;
	if (!UeberschriftGedruckt && mbm.Anforderungen != mbm.Freigaben) {
	  IO_DruckeKurzUeberschrift("Discrepancies in memory allocation", '*');
	  UeberschriftGedruckt = TRUE;
	}
        DruckeSpeicherManagerStatistik(AusgabeGrad,mbm.Name, mbm.block_size, mbm.element_size, mbm.Anforderungen,
                                       mbm.Freigaben, mbm.InsgesamtAngelegteBloecke, mbm.MaximalGleichzeitigBelegteElemente);
      }
      GesamtAnforderungen += mbm.Anforderungen;
      GesamtAngelegteBloecke += mbm.InsgesamtAngelegteBloecke;
      GesamtBelegterSpeicher += mbm.InsgesamtAngelegteBloecke * mbm.element_size * mbm.block_size;
      runny = runny->Nachf;
    }
    if (UeberschriftGedruckt) {
      IO_DruckeZeile('+');
      IO_DruckeStatistikZeile("Total number of allocations",GesamtAnforderungen );
      IO_DruckeStatistikZeile("Total number of deallocations",GesamtFreigaben );
      IO_DruckeStatistikZeile("Total number of allocated blocks",GesamtAngelegteBloecke );
      IO_DruckeStatistikZeile("Thereby allocated memory in bytes", GesamtBelegterSpeicher);
    }
#elif !COMP_POWER
    fprintf(IO_Strom, "\nMemory statistics not compiled into object module.\n");
    fprintf(IO_Strom, "If desired, define constant DM_STATISTIK as 1\n");
    fprintf(IO_Strom, "in header file DomaenenMemorium.h.\n\n");
#endif
  }
  else if (IS_IsIn("config", IS)){
#if DM_STATISTIK
    fprintf(IO_Strom, "SV: Statistics ON, ");
#else
    fprintf(IO_Strom, "SV: Statistics OFF, ");
#endif
#if SV_TEST
    fprintf(IO_Strom, "Memorytest ON, ");
#else
    fprintf(IO_Strom, "Memorytest OFF, ");
#endif
    fprintf(IO_Strom, "free-list-based memory management using ");
#ifdef SV_BASIS_MMAP
    fprintf(IO_Strom, "mmap, ");
#else
    fprintf(IO_Strom, "malloc, ");
#endif
    fprintf(IO_Strom, "nesting.\n");
  }
}
