/*=================================================================================================================
  =                                                                                                               =
  =  StrahlMemorium.c                                                                                             =
  =                                                                                                               =
  =  Strahlartiger Zugriff auf Speicher                                                                           =
  =                                                                                                               =
  =  16.04.96 Thomas.                                                                                             =
  =                                                                                                               =
  =================================================================================================================*/

#include "FehlerBehandlung.h"
#include "StrahlMemorium.h"
#include "compiler-extra.h"


/*=================================================================================================================*/
/*= I. Strahldomaenen                                                                                             =*/
/*=================================================================================================================*/

#define /*size_t*/ Inkrement(/*DomaenenT*/ Domaene)                                                                 \
  (*(uintptr_t*)(void*)&Domaene->Sonstiges)

#define /*size_t*/ InkrementSei(/*DomaenenT*/ Domaene,/*size_t*/ Inkrement)                                         \
  Domaene->Sonstiges = (void *)(Inkrement)

static void StatistikInkrement(DomaenenT Domaene MAYBE_UNUSEDPARAM)
{
  if_sst(
    Domaene->AnzahlDomaenenausdehnungen++;  
    Domaene->AnzahlGemapptePages += Inkrement(Domaene)/DM_Seitengroesse();
    MaximierenLong(&Domaene->AnzahlMaxGemapptePages,Domaene->AnzahlGemapptePages-Domaene->AnzahlGeunmapptePages));
}


static void Strahlhaken(void *Kollision)
{
  DomaenenT Domaene = *((DomaenenT*)Kollision-1);
  DM_AmEndeGesperrtAllokieren(Domaene,Inkrement(Domaene),NULL);
  StatistikInkrement(Domaene);
}


DomaenenT SM_NeueStrahldomaene(DomaenenbereichsT Bereich, unsigned int InkrementInSeiten, const char *Name)
{
  DomaenenT Domaene = DM_NeueDomaene(Strahldomaene,Bereich,Strahlhaken,Name);
  if (!InkrementInSeiten)
    our_error("Ungueltiges Inkrement in DM_NeueDomaene");
  InkrementSei(Domaene,InkrementInSeiten*DM_Seitengroesse());

  DM_AmEndeGesperrtAllokieren(Domaene,Inkrement(Domaene),NULL);
  StatistikInkrement(Domaene);
  return Domaene;
}


void SM_StrahldomaeneLeeren(volatile DomaenenT Domaene)
{
  DM_DomaeneLeeren(Domaene);
  DM_AmEndeGesperrtAllokieren(Domaene,Inkrement(Domaene),NULL);
  StatistikInkrement(Domaene);
}


void SM_StrahldomaeneLoeschen(volatile DomaenenT Domaene)
{
  DM_DomaeneLoeschen(Domaene);
}


/*=================================================================================================================*/
/*= II. Strahlen                                                                                                  =*/
/*=================================================================================================================*/

typedef struct StrahlTS {
  DomaenenT Domaene;
  void *Ende;
} *InternerStrahlT;                    /* Ist der Strahl gefroren, so benutzt er genau die Seiten [Strahl-d,Ende). */

#define /*InternerStrahlT*/ e2i(/*StrahlT*/ StrahlExtern)                                                           \
  (InternerStrahlT)((char*)StrahlExtern-sizeof(struct StrahlTS))

#define /*StrahlT*/ i2e(/*InternerStrahlT*/ StrahlIntern)                                                           \
  (StrahlT)((char*)StrahlIntern+sizeof(struct StrahlTS))


StrahlT SM_NeuerStrahl(DomaenenT Domaene)
{
  InternerStrahlT Strahl = Domaene->Frei;               /* Hier kann's auch schon krachen! Dann neue Seiten holen. */
  Strahl->Domaene = Domaene;
  return i2e(Strahl);
}


static void *Seitenadresse(void *ErsteFreieAdresse)        /* gibt Seitenadresse der naechste freien Seite zurueck */
{
  unsigned long int Adresse = ptr2ul(ErsteFreieAdresse);
  return (void*)((Adresse & (DM_Seitengroesse()-1) ? Adresse+DM_Seitengroesse() : Adresse)&~(DM_Seitengroesse()-1));
}


void SM_StrahlGefrieren(StrahlT StrahlExtern, void *ErsteFreieAdresse)
{
  InternerStrahlT Strahl = e2i(StrahlExtern);
  Strahl->Domaene->Frei =
    Strahl->Ende = Seitenadresse(ErsteFreieAdresse);
  if_sst(Strahl->Domaene->AnzahlGefroreneStrahlen++);
}


void SM_NichtGefrorenenStrahlLoeschen(StrahlT StrahlExtern MAYBE_UNUSEDPARAM)
{
  ;
}


void SM_StrahlSchrumpfen(StrahlT StrahlExtern, void *ErsteFreieAdresse)
{
  InternerStrahlT Strahl = e2i(StrahlExtern);
  void *Seite = Seitenadresse(ErsteFreieAdresse);
  DM_munmap(Seite,Strahl->Ende);
  Strahl->Ende = Seite;
  if_sst(
    Strahl->Domaene->AnzahlGeunmapptePages += ((char*)Seite-(char*)Strahl)/DM_Seitengroesse());
}


void SM_StrahlLoeschen(StrahlT StrahlExtern)
{
  InternerStrahlT Strahl = e2i(StrahlExtern);
  if_sst(
    Strahl->Domaene->AnzahlGeloeschteStrahlen++;
    Strahl->Domaene->AnzahlGeunmapptePages += ((char*)Strahl->Ende-(char*)Strahl)/DM_Seitengroesse());
  DM_munmap(Strahl,Strahl->Ende);
}
