#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "compiler-extra.h"
#include "general.h"
#include "SUEBasis.h"
#include "InfoStrings.h"
#include "Ausgaben.h"
#include "CStatistics.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe CStatistics.h) -----------------------*/
/* ------------------------------------------------------------------------*/
void (* CS_BerechnetesGewichtIstProc)(long int);
PQ_WatchingWalkthroughProcedureT CS_WeightStatProc;

/* ---------------------------------------------------------------------------------------- */
/* ---------------------- Definitionen ---------------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */

const int realWidth = 80;
const int realHeight = 25;

#define MaxGewBreite 10
#define MaxAnzBreite 7

static unsigned long int AnzahlVerschiedenerGewichte, EintraegeUnterLowerBound,
                         EintraegeUeberUpperBound, MaxAnzEintraege;

static long int MaxGewicht, MinGewicht, lowerBound, upperBound;

/* ---------------------------------------------------------------------------------------- */
/* ----------------------1. Gewichts-Baeume ----------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */

typedef struct GKTS{
   long int Gewicht;
   unsigned long int AnzahlEintraege;
   struct GKTS *leftSon, *rightSon;
} GewichtsKnotenT;

static GewichtsKnotenT *NeuenGewichtsKnotenAnlegen(long int Gewicht)
{
   GewichtsKnotenT *kn = our_alloc(sizeof(GewichtsKnotenT));
   kn->leftSon = kn->rightSon = NULL;
   kn->AnzahlEintraege = 0;
   kn->Gewicht = Gewicht;
   
   AnzahlVerschiedenerGewichte++;
   if (Gewicht > MaxGewicht) MaxGewicht = Gewicht;
   else if (Gewicht < MinGewicht) MinGewicht = Gewicht;
   
   return kn;
}

static void GewichtsBaumLoeschen(GewichtsKnotenT *Baum)
{
   if (Baum->leftSon)
      GewichtsBaumLoeschen(Baum->leftSon);
   if (Baum->rightSon)
      GewichtsBaumLoeschen(Baum->rightSon);
   our_dealloc(Baum);
}



/* ---------------------------------------------------------------------------------------- */
/* ----------------------2. Ausgabe-Funktionen -------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */


/* ------------- Liste --------------------------*/

static int GewichteProZeile, aktPosInZeile;

static void DruckeGewichtsBaumZahlen(GewichtsKnotenT *Baum)
{
   if (Baum->leftSon)
      DruckeGewichtsBaumZahlen(Baum->leftSon);
   
   fprintf(IO_Strom, "%10ld:%7ld  ", Baum->Gewicht, Baum->AnzahlEintraege);
   if (aktPosInZeile == GewichteProZeile){fprintf(IO_Strom, "\n");aktPosInZeile=1;}
   else aktPosInZeile++;
   
   if (Baum->AnzahlEintraege > MaxAnzEintraege)
      MaxAnzEintraege = Baum->AnzahlEintraege;

   if (Baum->rightSon)
      DruckeGewichtsBaumZahlen(Baum->rightSon);
}


static void GewichteAuflisten(GewichtsKnotenT *Baum)
{
   GewichteProZeile = realWidth/(MaxGewBreite + MaxAnzBreite + 2); /* MaxGewBreite ':' MaxAnzBreite '  ' */

   aktPosInZeile = 1;
   
   if (GewichteProZeile<1) GewichteProZeile = 1;
   
   MaxAnzEintraege = 
      (EintraegeUnterLowerBound>EintraegeUeberUpperBound)?
      EintraegeUnterLowerBound:EintraegeUeberUpperBound;


   fprintf(IO_Strom, "< lowerBd.:%7ld\n", EintraegeUnterLowerBound);
   
   DruckeGewichtsBaumZahlen(Baum);

   fprintf(IO_Strom, "> upperBd.:%7ld\n", EintraegeUeberUpperBound);
}


/* ------------- Balken -------------------------*/

static unsigned long int Skalierung;
static unsigned long int GesamtKps;

static void EinenBalkenDrucken(unsigned long int Wert)
{
   unsigned long int i = 0;
   while (i<Wert){
      fprintf(IO_Strom, "*");
      i += Skalierung;
   }
   fprintf(IO_Strom, "\n");
   GesamtKps += Wert;
}

static void DruckeGewichtsBaumBalken(GewichtsKnotenT *Baum)
{
   if (Baum->leftSon)
      DruckeGewichtsBaumBalken(Baum->leftSon);
   
   fprintf(IO_Strom, "%7ld", Baum->Gewicht);
   EinenBalkenDrucken(Baum->AnzahlEintraege);
   
   if (Baum->AnzahlEintraege > MaxAnzEintraege)
      MaxAnzEintraege = Baum->AnzahlEintraege;

   if (Baum->rightSon)
      DruckeGewichtsBaumBalken(Baum->rightSon);
}


static void HorizontaleBalkenDrucken(GewichtsKnotenT *Baum)
{
   Skalierung = MaxAnzEintraege / (realWidth-MaxGewBreite) + 1; /* KPs pro Stern */
   GesamtKps = 0;

   if (Skalierung < 1) Skalierung = 1;
   
   fprintf(IO_Strom, "\n\n<lowerB");
   EinenBalkenDrucken(EintraegeUnterLowerBound);
   
   DruckeGewichtsBaumBalken(Baum);

   fprintf(IO_Strom, ">upperB");
   EinenBalkenDrucken(EintraegeUeberUpperBound);

   fprintf(IO_Strom, "\n\nGesamt-KPs = %ld, Skalierung = %ld, Max # KPs = %ld\n", GesamtKps, 
           Skalierung, MaxAnzEintraege);
   
}

/* ---------------------------------------------------------------------------------------- */
/* ----------------------3. Time-Distribution --------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */

/* Es werden die Gewichte der neu gebildeten KPs aufgesammelt und dabei in Intervallen
   die Durchschnitte berechnet. Diese Durchschnitte werden in eine Liste von Arrays
   gesteckt und auf Wunsch (graphisch aufbereitet) ausgegeben.

   Eine Bereichs-Kontrolle findet aus Geschwindigkeitsgruenden nicht statt.
*/

#define EinzelFeldLaenge 1000   /* Laenge eines einzelnen Teil-Arrays */

typedef struct StatistikTS{
   long int *Array;
   struct StatistikTS *next;
} StatistikZellenT;

static StatistikZellenT *StatistikListe = NULL;              /* Liste der StatistikFelder */
static StatistikZellenT *AktuelleStatistikZelle = NULL;      /* Aktuelle Zelle */
static int AktuellerIndex = 0;

static void WertEintragen(long int w1)
{
   if (AktuellerIndex == EinzelFeldLaenge){
      AktuelleStatistikZelle->next = our_alloc(sizeof(StatistikZellenT));
      AktuelleStatistikZelle = AktuelleStatistikZelle->next;
      AktuelleStatistikZelle->Array = our_alloc(EinzelFeldLaenge * sizeof(long int));
      AktuelleStatistikZelle->next = NULL;
      AktuellerIndex = 0;
   }
   AktuelleStatistikZelle -> Array[AktuellerIndex] = w1;
   AktuellerIndex++;
}


static long int Summe = 0;                 /* Summe der letzten Gewichte */
static int Intervall = 0, nn = 0;              /* # aufzusummierender Gewichte, # bisher aufsummierter Gewichte */

static void BerechnetesGewichtIst_NoStat(long int w1 MAYBE_UNUSEDPARAM){;}

static void BerechnetesGewichtIst_Stat(long int w1)
{
   Summe += w1;
   if (++nn == Intervall){
      WertEintragen(Summe/nn);
      nn = Summe = 0;
   }
}

void CS_TimeDistribution(InfoStringT IS MAYBE_UNUSEDPARAM)
{
   if (StatistikListe){
      long int max = LONG_MIN, min = LONG_MAX;
      StatistikZellenT *next, *akt;
      int i;
      long int AnzWerte = 0;
      unsigned long int Achsenwert = 0; /* KP-Nummer bzw. Periodenwert fuer vertikale Nummerierung */

      GesamtKps = 0;

      next = StatistikListe->next;
      akt = StatistikListe;

      while (next){
         for (i=0; i<EinzelFeldLaenge; i++){ /* Maximum und Minimum bestimmen */
            long int Wert = akt->Array[i];
            if (Wert<min) min = Wert;
            else if (Wert>max) max = Wert;
         }
         akt = next;
         next = next->next;
         AnzWerte += EinzelFeldLaenge;
      }
      
      /* akt ist das letzte Einzel-Feld */
      for (i=0; i<AktuellerIndex; i++){ /* Maximum und Minimum bestimmen */
         long int Wert = akt->Array[i];
         if (Wert<min) min = Wert;
         else if (Wert>max) max = Wert;
      }
      AnzWerte += AktuellerIndex;

      if (min>0) min = 0;
      Skalierung = (max-min) / (realWidth - 12) + 1; /* Intervall pro Stern */

      next = StatistikListe->next;
      akt = StatistikListe;

      while (next){             /* Naechstes Feld existiert => dieses komplett ausgeben */
         for (i=0; i<EinzelFeldLaenge; i++){
            long int Wert = akt->Array[i];
            Achsenwert += Intervall;
            fprintf(IO_Strom, "%10ld ", Achsenwert);
            EinenBalkenDrucken((unsigned long int) (Wert - min));
         }
         akt = next;
         next = next->next;
      }

      /* akt ist das letzte Einzel-Feld */
      for (i=0; i<AktuellerIndex; i++){
         long int Wert = akt->Array[i];
         fprintf(IO_Strom, "%10ld ", Achsenwert);
         Achsenwert += Intervall;
         EinenBalkenDrucken((unsigned long int) (Wert - min));
         if (Wert<min) min = Wert;
         else if (Wert>max) max = Wert;
      }

      fprintf(IO_Strom, "Skalierung = %ld, min = %ld, max = %ld, intervall = %d\n"
              "Gesamtzahl Eintraege %ld, Schnitt %ld\n", 
              Skalierung, min, max, Intervall,AnzWerte, GesamtKps/AnzWerte);
   }
}


/* ---------------------------------------------------------------------------------------- */
/* ----------------------4. Weight-Distribution ------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */

static struct GKTS *GewichtsBaum;   

#define WeightStat_BODY(/*{1,2}*/ PrimSekGewicht)                                                               \
{                                                                                                               \
   int Gew = wt.w ## PrimSekGewicht;                                                                            \
   GewichtsKnotenT *Knoten = GewichtsBaum;                                                                      \
                                                                                                                \
   if (Gew<lowerBound)                                                                                          \
      EintraegeUnterLowerBound++;                                                                               \
   else if (Gew>upperBound)                                                                                     \
      EintraegeUeberUpperBound++;                                                                               \
   else{                                                                                                        \
      if (!GewichtsBaum){                                                                                       \
         Knoten = GewichtsBaum = NeuenGewichtsKnotenAnlegen(Gew);                                               \
      }                                                                                                         \
      else{                     /* Knoten mit passendem Gewicht suchen bzw. erzeugen */                         \
         while((Knoten->Gewicht) != Gew){                                                                       \
            if (Gew < (Knoten->Gewicht)){                                                                       \
               if (!(Knoten->leftSon))                                                                          \
                  Knoten->leftSon = NeuenGewichtsKnotenAnlegen(Gew);                                            \
               Knoten = Knoten->leftSon;                                                                        \
            }                                                                                                   \
            else if (Gew > (Knoten->Gewicht)){                                                                  \
               if (!(Knoten->rightSon))                                                                         \
                  Knoten->rightSon = NeuenGewichtsKnotenAnlegen(Gew);                                           \
               Knoten = Knoten->rightSon;                                                                       \
            }                                                                                                   \
         }                                                                                                      \
      }                                                                                                         \
      Knoten->AnzahlEintraege++;                                                                                \
      if (Knoten->AnzahlEintraege > MaxAnzEintraege)                                                            \
         MaxAnzEintraege = Knoten->AnzahlEintraege;                                                             \
   }                                                                                                            \
}


static BOOLEAN PrimaryStatistics = TRUE;


static void CS_WeightStat (WeightEntryT wt, TermpairT tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT aP MAYBE_UNUSEDPARAM, RegelOderGleichungsT oP MAYBE_UNUSEDPARAM)
{
   if (PrimaryStatistics){
      WeightStat_BODY(1)
   }
   else{
      WeightStat_BODY(2)
   }
}


/* Verteilung in aktueller Menge. Dazu Durchlauf durch KP-Menge.
   */

void CS_WeightStatisticsStart(InfoStringT StatisticsIS)
{
   AnzahlVerschiedenerGewichte = EintraegeUnterLowerBound = 
      EintraegeUeberUpperBound = MaxAnzEintraege = 0;

   MaxGewicht = LONG_MIN;
   MinGewicht = LONG_MAX;
   GewichtsBaum = NULL;
   
   if (IS_ValueOf("lowerBound", StatisticsIS))
      lowerBound = atoi(IS_ValueOf("lowerBound", StatisticsIS));
   else
      lowerBound = LONG_MIN;

   if (IS_ValueOf("upperBound", StatisticsIS))
      upperBound = atoi(IS_ValueOf("upperBound", StatisticsIS));
   else
      upperBound = LONG_MAX;

   if (IS_IsIn("secondary", StatisticsIS))
      PrimaryStatistics = FALSE;
   else
      PrimaryStatistics = TRUE;

}


void CS_WeightStatisticsEnd(void)
{
   if (GewichtsBaum){

      GewichteAuflisten(GewichtsBaum);

      HorizontaleBalkenDrucken(GewichtsBaum);
      
      GewichtsBaumLoeschen(GewichtsBaum);
   }
}



/* ---------------------------------------------------------------------------------------- */
/* ----------------------5. Initialize ---------------------------------------------------- */
/* ---------------------------------------------------------------------------------------- */


BOOLEAN CS_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   char *val;

   if ((val = IS_ValueOfOption(DEF, USR, "statistics"))){
      if (strcmp(val, "on") == 0) {
         /* Variablen initialisieren */
        /* muss immer aufgerufen werden, da Listen beim Strategiewechsel geloescht werden */
        StatistikListe = our_alloc(sizeof(StatistikZellenT));
        AktuelleStatistikZelle = StatistikListe;
        AktuelleStatistikZelle->Array = our_alloc(EinzelFeldLaenge * sizeof(long int));
        AktuelleStatistikZelle->next = NULL;
        AktuellerIndex = 0;
        /* Intervall fuer Zeit-Verteilung */
        if (IS_ValueOfOption(DEF, USR, "statint")){
          Intervall = atoi(IS_ValueOfOption(DEF, USR, "statint"));
          if (Intervall < 1)
            Intervall = 1;
        }
        else
          Intervall = 1;
        CS_BerechnetesGewichtIstProc = BerechnetesGewichtIst_Stat;
      }
      else
         CS_BerechnetesGewichtIstProc = BerechnetesGewichtIst_NoStat;
      CS_WeightStatProc = CS_WeightStat;
   }
   else{
      CS_BerechnetesGewichtIstProc = BerechnetesGewichtIst_NoStat;
   }
   return TRUE;
}

void CS_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
      char *val;
      BOOLEAN stat = FALSE;
      int intervall=0;
      if ((val = IS_ValueOfOption(DEF, USR, "statistics"))){
         if ( (stat = (strcmp(val, "on") == 0))){
            if (IS_ValueOfOption(DEF, USR, "statint")){
               intervall = atoi(IS_ValueOfOption(DEF, USR, "statint"));
               if (intervall < 1)
                  intervall = 1;
            }
            else
               intervall = 1;
         }
      }
      fprintf(IO_Strom, "  Weight-Statistics: time-distribution %s (with intervall %d)\n\n",
              stat?"on":"off", intervall);
   }
   else{
      fprintf(IO_Strom, "Weight-Statistics\n*****************\n\n"
              "Compute Weight-distribution over time at run-time.\n"
              "statistics=[on|off]:intervall=<value>\n\n");
   }
}
           

void CS_CleanUp(void)
{
   while(StatistikListe){
      StatistikZellenT *next = StatistikListe->next;
      our_dealloc(StatistikListe->Array);
      our_dealloc(StatistikListe);
      StatistikListe = next;
   }
}
