/**********************************************************************
* Filename : MNF_Classify.c
*
* Kurzbeschreibung : Classifizieren von Nachfolgermengen bei Generate
*
* Autoren : Andreas Jaeger
*
* 02.10.96: Erstellung
*
*
*
**********************************************************************/

#include <math.h>
#include <limits.h>

#include "compiler-extra.h"
#include "general.h"
#include "MNF_Classify.h"
#include "SymbolGewichte.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"

/* exportierte Variablen */
unsigned long MNFC_WeightLeft  = 0,
   MNFC_WeightRight = 0,
   MNFC_IrredLeft   = 0,
   MNFC_IrredRight  = 0,
   MNFC_WeightHarmLeft  = 0,
   MNFC_WeightHarmRight = 0,
   MNFC_WeightGeoLeft = 0,
   MNFC_WeightGeoRight= 0,
   MNFC_MinWeightLeft = 0,
   MNFC_MinWeightRight= 0,
   MNFC_MaxWeightLeft = 0,
   MNFC_MaxWeightRight= 0;

/* interne Variablen */
static unsigned long mnfc_weightleft,
   mnfc_weightright,
   mnfc_anzirredl,
   mnfc_anzirredr,
   mnfc_anztotal;
static double avgHarmL, avgHarmR,
   avgGeoL, avgGeoR;
static unsigned long minL, minR, maxL, maxR;
/* Speichern, ob Abbruch wg. Deckelfkt. */
static BOOLEAN  mnfc_abbruch; /* Variable wird momentan nur gesetzt, aber nicht benutzt */

void MNF_ClassifyInitFunc (BOOLEAN AbbruchDeckel) {

   mnfc_weightleft = 0;
   mnfc_weightright= 0;
   mnfc_anzirredl  = 0;
   mnfc_anzirredr  = 0;

   avgHarmL = 0.0;
   avgHarmR = 0.0;
   avgGeoL  = 1.0;
   avgGeoR  = 1.0;

   minL = minR = ULONG_MAX;
   maxL = maxR = 0;
   
   mnfc_abbruch = AbbruchDeckel;
   mnfc_anztotal = 0;
}


static unsigned long Weight (TermT term) 
{
   SymbolT       symb;
   unsigned long weight = 0;
   
   TO_doSymbole(term, symb,{
      weight  += SG_SymbolGewichtCP(symb);
   });
   
   return weight;
}


void MNF_ClassifyTraverseFunc (TermT term, BOOLEAN irreduzibel,
                               BOOLEAN links)
{
   unsigned long weight;
   
   mnfc_anztotal++;
   if (irreduzibel) {
      weight = Weight(term);
      if (links) {
         mnfc_weightleft += weight;
         if (weight < minL)
            minL = weight;
         if (weight > maxL)
            maxL = weight;
         mnfc_anzirredl++;
         avgHarmL += 1.0/(double) weight;
         avgGeoL  *= (double) weight;
      }
      else {
         mnfc_weightright += weight;
         if (weight < minR)
            minR = weight;
         if (weight > maxR)
            maxR = weight;
         mnfc_anzirredr++;
         avgHarmR += 1.0/(double) weight;
         avgGeoR  *= (double) weight;
      }
   }
}

INLINE static void
calc_results (unsigned long anzirred,
                                 unsigned long *ArithWeight, unsigned long *HarmWeight, unsigned long *GeoWeight,
                                 unsigned long weightTotal, double avgHarm, double avgGeo) 
{
   if (anzirred == 0) {
      *ArithWeight = 100;
      *HarmWeight  = 100;
      *GeoWeight   = 100;
   }
   else {double t;
   
      *ArithWeight = weightTotal / anzirred;
      *HarmWeight  = (unsigned long)((double)anzirred / avgHarm);
      t = pow(avgGeo, 1.0/(double)anzirred);
      *GeoWeight   = (unsigned long) t;
   }
}

      
void MNF_ClassifyExitFunc (void) 
{
   calc_results (mnfc_anzirredl,  &MNFC_WeightLeft, &MNFC_WeightHarmLeft, &MNFC_WeightGeoLeft,
                 mnfc_weightleft, avgHarmL, avgGeoL);
   calc_results (mnfc_anzirredr, &MNFC_WeightRight, &MNFC_WeightHarmRight, &MNFC_WeightGeoRight,
                 mnfc_weightright, avgHarmR, avgGeoR);
   MNFC_IrredLeft   = mnfc_anzirredl;
   MNFC_IrredRight  = mnfc_anzirredr;
   MNFC_MinWeightLeft = minL;
   MNFC_MinWeightRight= minR;
   MNFC_MaxWeightLeft = maxL;
   MNFC_MaxWeightRight= maxR;
}
