#include "compiler-extra.h"
#include "OccNest.h"
#include "TermOperationen.h"
#include "general.h"
#include "Ausgaben.h"
#include "SpeicherParameter.h"
#include "SymbolOperationen.h"
#include "Zielverwaltung.h"



/*=================================================================================================================*/
/*= I. Strukturdefinitionen, globale Variablen, Makros                                                            =*/
/*=================================================================================================================*/

#define ON_TEST 0                                                     /* interne Konsistenztests, Alternativen 0,1 */

struct SymbolTabelleS{                                      /* fuer jedes Symbol wird ein solcher Eintrag angelegt */
   long int cur,                                                    /* Laenge der bisher angetroffenen Symbolkette */
            nest,                               /* Laenge der insgesamt laengsten bisher angetroffenen Symbolkette */
            occ;                                                                /* # Vorkommen des Symbols im Term */
} *SymbolTabelle;                                                                   /* fuer jedes Symbol anzulegen */

struct SeitenInfoS {
   long int *Occ,
            *Nest;
} OccInfoLinks,                                                                           /* zum Sichern von Daten */
  OccInfoZiel;

static unsigned int OccTiefe;

struct GlobalerPfadS {
   int unbehNachf;
   long int *cur;                                                               /* Ruecksetzinfo fuer alle Symbole */
} *GlobalerPfad;                                                                          /* Array ueber die Tiefe */
static unsigned int GlobalerPfadDimension;

static void GlobalerPfadInitialisieren(unsigned int Start);

static void GgfGlobalerPfadErweitern(void)
{
  if (OccTiefe==GlobalerPfadDimension) {
    unsigned int Alt = GlobalerPfadDimension;
    array_realloc(GlobalerPfad,GlobalerPfadDimension += ON_ErsteStapelhoehe,struct GlobalerPfadS);
    GlobalerPfadInitialisieren(Alt);
  }
}


/*=================================================================================================================*/
/*= II. Ausgabefunktionen                                                                                         =*/
/*=================================================================================================================*/

static void DruckeGlobaleInfo(char *Text MAYBE_UNUSEDPARAM, TermT T MAYBE_UNUSEDPARAM)
{
#if ON_TEST
  SymbolT symb;
  IO_DruckeFlex("Global. %s, %t\n\n", Text,T);
  fprintf(IO_Strom, "Symbol  Occ   Nest\n-----------------\n");
  SO_forFktSymbole(symb)
    fprintf(IO_Strom, "%6s%5ld%5ld\n", IO_FktSymbName(symb), SymbolTabelle[symb].occ, SymbolTabelle[symb].nest);
#else
  ;
#endif
}
   
static void DruckeSeitenInfo(char *Text MAYBE_UNUSEDPARAM, struct SeitenInfoS SI MAYBE_UNUSEDPARAM)
{
#if ON_TEST
  SymbolT symb;
  fprintf(IO_Strom, "SeitenInfo. %s\n", Text);
  fprintf(IO_Strom, "Symbol  Occ   Nest\n-----------------\n");
  SO_forFktSymbole(symb)
    fprintf(IO_Strom, "%6s%5ld%5ld\n", IO_FktSymbName(symb), SI.Occ[symb], SI.Nest[symb]);
  fprintf(IO_Strom, "\n\n");
#else
  ;
#endif
}
   
#if ON_TEST
static void Mom(void)
{
  SymbolT symb;
  unsigned long i;
  DruckeSeitenInfo("Ziel", OccInfoZiel);
  DruckeSeitenInfo("Links", OccInfoLinks);
  IO_DruckeFlex("Global.\n");
  fprintf(IO_Strom, "Symbol  Occ Nest  Cur\n---------------------\n");
  SO_forFktSymbole(symb)
    fprintf(IO_Strom, "%6s%5ld%5ld%5ld\n", IO_FktSymbName(symb), 
            SymbolTabelle[symb].occ, SymbolTabelle[symb].nest,
            SymbolTabelle[symb].cur);
  fprintf(IO_Strom, "Globaler Pfad\n");
  fprintf(IO_Strom, "Tiefe | UnBHN Cur:");
  SO_forFktSymbole(symb)
    fprintf(IO_Strom, "%6s", IO_FktSymbName(symb));
  fprintf(IO_Strom, "\n");
  fprintf(IO_Strom, "------------------");
  SO_forFktSymbole(symb)
    fprintf(IO_Strom, "------");
  fprintf(IO_Strom, "\n");
  for (i=0;i<15;i++){
    fprintf(IO_Strom, "%5ld | %5d     ", i, GlobalerPfad[i].unbehNachf);
    SO_forFktSymbole(symb)
      fprintf(IO_Strom, " %5ld", GlobalerPfad[i].cur[symb]);
    fprintf(IO_Strom, "\n");
  }
  fprintf(IO_Strom, "\nOccTiefe = %d\n\n", OccTiefe);
}
#endif

static void DruckeOccTiefeUngleich1(void)
{
#if ON_TEST
  if (OccTiefe!=1) 
    fprintf(IO_Strom, "OccTiefe = %d != 1\n", OccTiefe);
#else
  ;
#endif
}



/*=================================================================================================================*/
/*= III. Zielanmeldung                                                                                            =*/
/*=================================================================================================================*/

static void OccNestSchleife(TermT term);

static void ZielIst(TermT term1, TermT term2)
{
   SymbolT Index;

   OccNestSchleife(term1);

   SO_forUnkonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                              /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
      OccInfoLinks.Nest[Index] = SymbolTabelle[Index].nest;                           /* Nest sichern und loeschen */
      SymbolTabelle[Index].nest=0;
   }
   SO_forKonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                              /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
   }

   DruckeOccTiefeUngleich1();

   OccNestSchleife(term2);

   SO_forFktSymbole(Index){
      OccInfoZiel.Occ[Index] = max(SymbolTabelle[Index].occ, OccInfoLinks.Occ[Index]);
      SymbolTabelle[Index].occ = 0;
      OccInfoZiel.Nest[Index] = max(SymbolTabelle[Index].nest, OccInfoLinks.Nest[Index]);
      SymbolTabelle[Index].nest= 0;
      SymbolTabelle[Index].cur = 0;
   }
#if ON_TEST
   IO_DruckeFlex("Occ-Nest-ZIEL ist %t = %t.\n", term1, term2);
#endif
   DruckeOccTiefeUngleich1();
   OccTiefe = 1;
}



/*=================================================================================================================*/
/*= IV. Occ- und Nest-Berechnung auf Termen                                                                       =*/
/*=================================================================================================================*/

/* Iterative Funktion zur Berechnung von nest und occ(auf allen Symbolen).
   Nest bestimmt die groesste Zahl von hintereinanderfolgenden Vorkommen von f in Aesten von T, wobei
   T als Baum gesehen wird. 
*/

static void OccNestSchleife(TermT term)
{
   SymbolT symb, Index;

   TO_doSymbole(term, symb,{                                                         /* aktuelles Symbol behandeln */

     if (SO_SymbIstVar(symb)) {
         SO_forUnkonstSymbole(Index)
            SymbolTabelle[Index].cur = 0;
      } 
     else {
         SymbolTabelle[symb].occ++;

         SO_forUnkonstSymbole(Index)
            if (SO_SymbUngleich(Index,symb))
              SymbolTabelle[Index].cur = 0;

         if (SO_Stelligkeit(symb)){                                             /*d. h. Symbol ist keine Konstante */
            SymbolTabelle[symb].cur++;
            SymbolTabelle[symb].nest = max(SymbolTabelle[symb].cur,SymbolTabelle[symb].nest);
         }
      }
                                                                 /* Pfad-Position des folgenden Symbols bestimmen: */
      GgfGlobalerPfadErweitern();
      GlobalerPfad[OccTiefe].unbehNachf = SO_Stelligkeit(symb);

      if (GlobalerPfad[OccTiefe].unbehNachf > 0){             /* es gibt noch unbeh. Nachf => naechstes Symbol ist */
         GlobalerPfad[OccTiefe].unbehNachf--;                /* Nachfolger von symb; ein Nachf wird dann behandelt */
         SO_forUnkonstSymbole(Index)
            GlobalerPfad[OccTiefe].cur[Index] = SymbolTabelle[Index].cur; /* aktuelle Werte zum Ruecksetzen merken */
         OccTiefe++;                                                                                    /* Abstieg */
      }
      else{      /* symb war nullstellig oder alle Teilterme abgearbeitet => naechstes Symbol ist Vorgaenger davon */
         do{                                         /* Also Aufstieg, bis unbeh. Geschwister oder Wurzel erreicht */
            OccTiefe--;
         }while((OccTiefe>0) && (GlobalerPfad[OccTiefe].unbehNachf==0));
         GlobalerPfad[OccTiefe].unbehNachf--;                    /* einer der bisher unbeh. Nachf. wird jetzt beh. */
         SO_forUnkonstSymbole(Index)
            SymbolTabelle[Index].cur = GlobalerPfad[OccTiefe].cur[Index];              /* akt. Werte zuruecksetzen */
         OccTiefe++;                                                                              /* und absteigen */
      }
   });
}

#define PSI(/*long int*/ x) ((x<0)? 1:(x+1))

#define PSI2(/*long int*/ x, /*long int*/ y) ((x>y)?(x-y+1):1)                    /* Es gilt: PSI(x-y) = PSI2(x,y) */

static WeightT OccNest(TermT term1, TermT term2, WeightT Phi)
{
   SymbolT Index;

   OccNestSchleife(term1);

   DruckeGlobaleInfo("Linke Seite", term1);

   SO_forUnkonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                              /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
      OccInfoLinks.Nest[Index] = SymbolTabelle[Index].nest;                           /* Nest sichern und loeschen */
      SymbolTabelle[Index].nest=0;
   }
   SO_forKonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                              /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
   }

   DruckeOccTiefeUngleich1();
   DruckeSeitenInfo("Gerettet als", OccInfoLinks);

   OccNestSchleife(term2);

   SO_forUnkonstSymbole(Index){
         OccInfoLinks.Occ[Index] = max(OccInfoLinks.Occ[Index], SymbolTabelle[Index].occ);
         OccInfoLinks.Nest[Index] = max(OccInfoLinks.Nest[Index], SymbolTabelle[Index].nest);
            Phi *= PSI2(OccInfoLinks.Occ[Index], OccInfoZiel.Occ[Index])
            * PSI2(OccInfoLinks.Nest[Index], OccInfoZiel.Nest[Index]);
      SymbolTabelle[Index].occ = 0;
      SymbolTabelle[Index].nest= 0;
   }
   SO_forKonstSymbole(Index){
         OccInfoLinks.Occ[Index] = max(OccInfoLinks.Occ[Index], SymbolTabelle[Index].occ);
         Phi *= PSI2(OccInfoLinks.Occ[Index], OccInfoZiel.Occ[Index]);
      SymbolTabelle[Index].occ = 0;
   }

   DruckeOccTiefeUngleich1();
   return Phi;
}

static TermT aktuellesKPlinks,                                                     /* Terme des zu bewertenden KPs */
             aktuellesKPrechts;
static WeightT aktuellesPhi,                                                     /* kleinstes bisher erzieltes Phi */
               aktuellesBasisPhi;                                                            /* Basis-Wert des KPs */

static void OccNestEinZiel(TermT term1, TermT term2, unsigned long number MAYBE_UNUSEDPARAM)
{
  WeightT localPhi;
  ZielIst(term1, term2);
  localPhi = OccNest(aktuellesKPlinks, aktuellesKPrechts, aktuellesBasisPhi);
  if (localPhi < aktuellesPhi) 
    aktuellesPhi = localPhi;
}

WeightT ON_OccNest(TermT term1, TermT term2, WeightT Phi)
{                                           /* Berechnet das Minimum aller OccNest-Werte des KPs mit allen Zielen. */
  aktuellesKPlinks = term1, aktuellesKPrechts = term2;
  aktuellesPhi = LONG_MAX;
  aktuellesBasisPhi = Phi;

  Z_ZielmengenBeobachtungsDurchlauf(OccNestEinZiel);
  
  return aktuellesPhi;
}



/*=================================================================================================================*/
/*= V. Initialisierung                                                                                            =*/
/*=================================================================================================================*/

static void GlobalerPfadInitialisieren(unsigned int Start)
{
  for (; Start<GlobalerPfadDimension; Start++) {
    SymbolT Index;
    GlobalerPfad[Start].cur = array_alloc(SO_AnzahlFunktionen,long int);
    SO_forFktSymbole(Index)
      GlobalerPfad[Start].cur[Index] = 0;
    GlobalerPfad[Start].unbehNachf = 0;
  }
}


static BOOLEAN Initialised = FALSE;
/* Achtung: Initialised darf hier nicht durch WM_AnlaufNr ersetzt werden,
   da sonst Strategiewechsel nicht klappt (ON_Init wird nur aufgerufen, wenn
   OccNest benoetigt wird.
 */

void ON_Init(void)
{
   SymbolT Index;

   if (Initialised)
     return;

   Initialised = TRUE;
   
   GlobalerPfad = array_alloc(GlobalerPfadDimension = ON_ErsteStapelhoehe,struct GlobalerPfadS);
   GlobalerPfadInitialisieren(0);

   OccInfoLinks.Occ =  array_alloc(SO_AnzahlFunktionen,long int);
   OccInfoLinks.Nest = array_alloc(SO_AnzahlFunktionen,long int);
   OccInfoZiel.Occ =   array_alloc(SO_AnzahlFunktionen,long int);
   OccInfoZiel.Nest =  array_alloc(SO_AnzahlFunktionen,long int);

   SymbolTabelle = array_alloc(SO_AnzahlFunktionen,struct SymbolTabelleS);
   SO_forFktSymbole(Index)
      SymbolTabelle[Index].cur = 
        SymbolTabelle[Index].nest = 
        SymbolTabelle[Index].occ = 
        OccInfoLinks.Occ[Index] =
        OccInfoLinks.Nest[Index] =
        OccInfoZiel.Nest[Index] =
     OccInfoZiel.Occ[Index] = 0;

   OccTiefe = 1;
}
