#include "Ausgaben.h"
#include <ctype.h>
#include "SpeicherVerwaltung.h"
#include "SymbolGewichte.h"
#include "SymbolOperationen.h"


long int *SG_SymbolGewichtCG, 
         *SG_SymbolGewichtCP;

static unsigned int AnzVariablen = SO_AnzErsterVariablen;

static BOOLEAN AuswertungZuVeraendert(InfoStringT DEF, InfoStringT USR, BOOLEAN fuerCPs, long int **ActArray)
/* Eintraege sind durch ':' getrennt. Jeder Eintrag enthaelt den Bezeichner, das Zuweisungssymbol '=' und das 
   Gewicht. Der Bezeichner ist entweder ein Funktionssymbol der Spezifikation oder eines der beiden Schluessel-
   worte VAR fuer Variablengewichte bzw. DEF fuer den Defaultwert nicht-angegebener Funktionssymbole. */
{
  /* 1. def und var raussammeln */
  long int def = IS_ValueOfOption(DEF, USR, "DEF") ? atoi(IS_ValueOfOption(DEF, USR, "DEF")) : 1,
           var = IS_ValueOfOption(DEF, USR, "VAR") ? atoi(IS_ValueOfOption(DEF, USR, "VAR")) : 1;
  BOOLEAN ValueChanged = def!=1 || var!=1;
  SymbolT i;

  /* 2. Identifizieren der Eintraege */
  *ActArray = negarray_alloc(SO_AnzahlFunktionen+AnzVariablen,long int,AnzVariablen);
  SO_forFktSymbole(i) {
     char *wert = IS_ValueOfOption(DEF, USR, IO_FktSymbName(i));
     if (wert && (isdigit((int)*wert) || *wert=='-')) {
       (*ActArray)[i] = atoi(wert);
       ValueChanged = TRUE;
     }
     else
       (*ActArray)[i] = def;
  }
  SO_forSymboleAbwaerts(i,SO_ErstesVarSymb,SO_NummerVar(AnzVariablen))
    (*ActArray)[i] = var;

  /* 3. Test-Ausgabe */
  if (!COMP_POWER) {
    fprintf(IO_Strom, "  Symbol-Weights (%s): ", fuerCPs? "CP" : "CG");
    SO_forFktSymbole(i){
      IO_DruckeSymbol(i); 
      fprintf(IO_Strom, "=%ld, ", (*ActArray)[i]);
    }
    fprintf(IO_Strom, "Vars=%ld\n", var);
  }

  return ValueChanged;
}


BOOLEAN SG_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR, BOOLEAN fuerCPs)
{
  if (what==ParamEvaluate) {
    long int *ActArray;
    BOOLEAN ValueChanged = AuswertungZuVeraendert(DEF,USR,fuerCPs,&ActArray);
    negarray_dealloc(ActArray,AnzVariablen);
    return ValueChanged;
  }
  else {
    fprintf(IO_Strom, "Symbol-Weights\n**************\n\n"
            "Assign each symbol a different individual weight (default is 1). This works for\n"
            "both critical pairs and critical goals. Variables cannot be assigned\n"
            "individual weights, but can be assigned a value different from 1 by VAR=<value>.\n"
            "The default value for non-variable symbols can be changed by DEF=<value>.\n"
            "Other symbols can be assigned values by <symbol>=<value>.\n\n"
            "Example: DEF=2:VAR=5:f=5:g=0\n\n");
    return TRUE;
  }
}


BOOLEAN SG_SymbGewichteEintragen(InfoStringT DEF, InfoStringT USR, BOOLEAN fuerCPs)
{              /* Aufdroeseln des Aufruf-Parameters nach Default-, Variablen- und den einzelnen Funktionsgewichten */
   return AuswertungZuVeraendert(DEF,USR,fuerCPs,fuerCPs ? &SG_SymbolGewichtCP : &SG_SymbolGewichtCG);
}


static void VariablenfeldAnpassen(unsigned int AnzVariablenNeu, long int **ActArray)
{
  long int *ActArrayNeu = negarray_alloc(SO_AnzahlFunktionen+AnzVariablenNeu,long int,AnzVariablenNeu);
  SymbolT i;
  SO_forFktSymbole(i)
    ActArrayNeu[i] = (*ActArray)[i];
  SO_forSymboleAbwaerts(i,SO_ErstesVarSymb,SO_NummerVar(AnzVariablenNeu))
    ActArrayNeu[i] = (*ActArray)[SO_ErstesVarSymb];
  negarray_dealloc(*ActArray,AnzVariablen);
  *ActArray = ActArrayNeu;
}

void SG_VariablenfeldAnpassen(SymbolT NeuesteBaumvariable)
{
  if (SO_VariableIstNeuer(NeuesteBaumvariable,SO_NummerVar(AnzVariablen))) {
    unsigned int AnzVariablenNeu = minKXgrY(SO_AnzErsterVariablen,-NeuesteBaumvariable);
    VariablenfeldAnpassen(AnzVariablenNeu,&SG_SymbolGewichtCP);
    VariablenfeldAnpassen(AnzVariablenNeu,&SG_SymbolGewichtCG);
    AnzVariablen = AnzVariablenNeu;
  }
}
