/**********************************************************************
* Filename : CritGoals.c
*
* Kurzbeschreibung : Verwalten der Kritischen Ziele
*
* Autoren : WEC
*
* 08.12.96: Erstellung aus SUEM.c heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
* 02.07.97: Umstellung auf Heaps. AB.
*
*
* Benoetigt folgende Module:
*
* Benutzte Globale Variablen aus anderen Modulen:
*
**********************************************************************/



/* Die Datei wurde mal erst aus SUEM.c heraus ausgelagert. Die Schnittstelle
   ist aber noch nicht ganz sauber.
   */


#include "Achtzigstel.h"
#include "Ausgaben.h"
#include "compiler-extra.h"
#include "CritGoals.h"
#include "general.h"
#include "InfoStrings.h"
#include <limits.h>  /* fuer LONG_MAX */
#include "NFBildung.h"
#include "pcl.h"
#include "NewClassification.h"
#include "ParseInterleave.h"
#include "RUndEVerwaltung.h"
#include "Stichpunkte.h"
#include "SUEBasis.h"
#include "SUEM.h"
#include "TermOperationen.h"
#include "Zielverwaltung.h"

#define CG_TEST 0

/* Globale Variablen */
static CounterT CG_AnzCritGoalsGesamt, CG_AnzMaxCritGoals, CG_AnzCritGoalsActual;
static CounterT CG_AnzIRGleich, CG_AnzIRReduziert, CG_AnzIRWMord;

typedef enum {CG_lebt, CG_waise, CG_selektiert} CG_LebtNochT;

/* Heap fuer CG-Aufbewahrung */
typedef struct CGHeapTS {
  TermpairT            tp;
  WeightT              w1;
  WeightT              w2;
  CG_LebtNochT         LebtNoch;
  RegelOderGleichungsT aP;
  RegelOderGleichungsT oP;
} CGHeapT;

static CGHeapT *CGHeap;
static unsigned long int NaechstesFreies;
static unsigned long int CGHeapAktuelleGroesse;

static StichpunktT *StichpunktListe;
static unsigned long int AnzahlStichpunkte;
static unsigned long int Periode;
static int PeriodenStaerke;
static const char *strengthArray[4]={"o", "r", "e", "re"};

/* fuer Interleaving der Queues */
static unsigned int modulo;
static unsigned int threshold;


BOOLEAN (*CG_SelectProc)(TermpairT*,RegelOderGleichungsT*,RegelOderGleichungsT*);          /* analog PQ_SelectProc */


#if CG_TEST
static void HeapTest(char *str)
{
   unsigned long int i;
   fprintf(IO_Strom, "************Heap-Test %s*************\n",str);
   for(i=0; i < NaechstesFreies; i++){
      if ((2*i+1< NaechstesFreies) &&
          ((CGHeap[2*i+1].w1<CGHeap[i].w1)||
           ((CGHeap[2*i+1].w1==CGHeap[i].w1)&&(CGHeap[2*i+1].w2<CGHeap[i].w2))))
         fprintf (IO_Strom, "CGHeap-Sortierung falsch (Vater (%ld,%ld), Sohn (%ld,%ld))\n",
                CGHeap[i].w1, CGHeap[i].w2, CGHeap[2*i+1].w1, CGHeap[2*i+1].w2);

      if ((2*i+2< NaechstesFreies) &&
          ((CGHeap[2*i+2].w1<CGHeap[i].w1)||
           ((CGHeap[2*i+2].w1==CGHeap[i].w1)&&(CGHeap[2*i+2].w2<CGHeap[i].w2))))
         fprintf (IO_Strom, "CGHeap-Sortierung falsch (Vater (%ld,%ld), Sohn (%ld,%ld))\n",
                CGHeap[i].w1, CGHeap[i].w2, CGHeap[2*i+2].w1, CGHeap[2*i+2].w2);
      /*      fprintf(IO_Strom, "%ld --- KP zu Eltern (%ld, %ld) \n", i, 
             CGHeap[i].LebtNoch?CGHeap[i].aP->Nummer:0,
             CGHeap[i].LebtNoch?CGHeap[i].oP->Nummer:0);*/
   }
}
#else
#define HeapTest(bla)
#endif


static INLINE BOOLEAN LasseEinsinken(unsigned long int position)
{
   BOOLEAN changed = FALSE, isHeap = FALSE;
   unsigned long j = position * 2 + 1;
   CGHeapT merker = CGHeap[position];

   while ((!isHeap) && (j < NaechstesFreies)) {
      if ((j+1 < NaechstesFreies) &&                                      /* Position j+1 noch gueltig? */
          ((CGHeap[j].w1 > CGHeap[j+1].w1) ||
           ((CGHeap[j].w1 == CGHeap[j+1].w1) && (CGHeap[j].w2>CGHeap[j+1].w2))))
        j++;                                                           /* => j hat jetzt den Wert des leichteren Nachfolgers */
      if ((merker.w1 > CGHeap[j].w1) ||
          ((merker.w1 == CGHeap[j].w1)&&(merker.w2 > CGHeap[j].w2))){
         CGHeap[position]=CGHeap[j];
         changed = TRUE;
         position = j;
         j *=2;
         j++;
      }
      else
         isHeap = TRUE;
   }
   if (changed)
      CGHeap[position] = merker;
   return changed;
}

static INLINE void CG_DeleteEntry (int index)
{
  if (CGHeap[index].LebtNoch == CG_waise) {
    TPR_DESTROYTERMPAIR(CGHeap[index].tp, CGHeap[index].aP, CGHeap[index].oP);
  }
  CG_AnzCritGoalsActual--;  
}

#if 0
/* Das Benutzen dieser Fkt fuehrt zu Fehlern, wenn kein PCL benutzt wird.
   Da dies ganze eine Optimierung ist, habe ich es mal erst rausgeschmissen
   AJ. 10.5.98 */
/* Heap so bearbeiten, dass letztes Element lebt */
static INLINE void CG_ReorgHeapLast (void)
{  
  while ((NaechstesFreies > 0) && (CGHeap[NaechstesFreies-1].LebtNoch != CG_lebt)) {
    --NaechstesFreies;
    CG_DeleteEntry (NaechstesFreies);
  }
}
#endif

/* Heap so bearbeiten, dass erstes Element lebt */
static INLINE void CG_ReorgHeapFirst (void) 
{

  while ((NaechstesFreies > 0) && (CGHeap[0].LebtNoch != CG_lebt)) {
    CG_DeleteEntry (0);
    if (--NaechstesFreies > 0) {
      CGHeap[0] = CGHeap[NaechstesFreies];
      LasseEinsinken(0);
    }
  }
}

static BOOLEAN Funktionen__(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, 
                            PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{
  return FALSE;
}


#define REDUKTIONS_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit)                      \
{                                                                                                               \
  BOOLEAN reduziert;                                                                                            \
  PCL_Reduktionsbeginn(KP, pclID, linkeSeite, rhs);                                                             \
  reduziert = NF_ ## Womit  (lhs, KPsIntermed);                                                                 \
  PCL_Reduktionsbeginn(KP, pclID, rechteSeite, lhs);                                                            \
  reduziert = NF_ ## Womit (rhs, KPsIntermed) || reduziert;                                                     \
                                                                                                                \
  return reduziert;                                                                                             \
}


static BOOLEAN Funktionen_R(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, 
                     PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
REDUKTIONS_BODY(NFBildungZuRegeln)

static BOOLEAN Funktionen_E(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, 
                     PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
REDUKTIONS_BODY(NFBildungZuGleichungen)

static BOOLEAN Funktionen_RE(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, 
                     PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
REDUKTIONS_BODY(NFBildung)

static BOOLEAN (*Funktionen[])(TermT,TermT,PCL_ObjektNummerT*) = {
  Funktionen__,
  Funktionen_R,
  Funktionen_E,
  Funktionen_RE
};


/*
  Interreduktion                                                                      
*/

static void Durchlauf(unsigned int strength)
{
  unsigned long int i = 0;
  BOOLEAN changed = FALSE;
  BOOLEAN (*Anwendungsproc)(TermT,TermT,PCL_ObjektNummerT*) = Funktionen[strength];
  AZ_DruckeMeldungStart("critical goals", strengthArray[strength], NaechstesFreies);
  
  while (i < NaechstesFreies) {
    AZ_DruckeMeldungItem();
    if (CGHeap[i].LebtNoch == CG_lebt) {                          /* Auspacken des Termpaares als eq (s,t) = false */
      TermT lhs, rhs;
      PCL_ObjektNummerT pclID;
      TPR_TERMPAIRtoPARALLEL_INTERMED (CGHeap[i].tp, &lhs, &rhs, CGHeap[i].aP, CGHeap[i].oP, &pclID);
      if ((*Anwendungsproc) (lhs, rhs, &pclID)) {
        UTermT s,t;
        if (Z_UngleichungEntmantelt(&s,&t, lhs,rhs)) {                 /* gdw weder SUCCESS noch wegzuwerfendes CG */
          WeightEntryT wt;                /* Wir haben jetzt: eq (s,t) = false und brauchen s,t zur Classifikation */
          C_CG_Classify (&wt, s, t, CGHeap[i].aP, CGHeap[i].oP);
          if (CGHeap[i].w1 != wt.w1) {
            changed = TRUE;
            CGHeap[i].w1 = wt.w1;                                                      /* w2 wird nicht veraendert */
          }
          CGHeap[i].tp = TPR_PARALLELtoTERMPAIR (CGHeap[i].tp, lhs, rhs,CGHeap[i].aP, CGHeap[i].oP,TRUE, pclID);
          CG_AnzIRReduziert++;
          i++;                                                                     /* naechstes Element betrachten */
        }
        else if (TO_TermeGleich (lhs, rhs)) {                               /* false = false entstanden; wegwerfen */
          TPR_DESTROYPARALLEL (CGHeap[i].tp, lhs, rhs, CGHeap[i].aP, CGHeap[i].oP);
          CGHeap[i] = CGHeap[--NaechstesFreies];
          CG_AnzIRGleich++;
          CG_AnzCritGoalsActual--;
          changed = TRUE;             /* naechstes Element betrachten - aber i nicht hochzaehlen, da kopiert wurde */
        }
        else {                                                                 /* true = false entstanden; SUCCESS */
          changed = TRUE;
          CGHeap[i].w1 = 0;                                                    /* entsprechendes CG sofort bringen */
          CGHeap[i].tp = TPR_PARALLELtoTERMPAIR (CGHeap[i].tp, lhs, rhs,CGHeap[i].aP, CGHeap[i].oP,TRUE, pclID);
          CG_AnzIRReduziert++;
          i = NaechstesFreies;                      /* Schleife direkt verlassen, weitere Arbeit ist ueberfluessig */
        }
      }
      else {                                                       /* fokussiertes kritisches Ziel war irreduzibel */
        CGHeap[i].tp = TPR_PARALLELtoTERMPAIR(CGHeap[i].tp, lhs, rhs,CGHeap[i].aP, CGHeap[i].oP,FALSE, pclID);
        i++;
      }
    }
    else {
      /* Waisenmord/FIFO selektiert */
      CG_DeleteEntry (i);
      CGHeap[i] = CGHeap[--NaechstesFreies];
      changed = TRUE;
      CG_AnzIRWMord++;
      /* naechstes Element betrachten - aber i nicht hochzaehlen, da kopiert wurde */
    }
  }
  
  /* Reorganisieren: */
  if (changed && NaechstesFreies) {
    i = NaechstesFreies / 2;
    while (i > 0) {
      i--;
      LasseEinsinken(i);
    }
  }
  AZ_DruckeMeldungEnde();
}


void
CG_CGMengeInterreduzieren(BOOLEAN Erzwingen, char *Staerke)
{
  unsigned int strength;
  if (Erzwingen) {
    strength = 3;
    while (strcmp(Staerke,strengthArray[strength]) && strength)
      strength--;
    Durchlauf(strength);
  }
  else if (SP_IstStichpunkt(StichpunktListe, AnzahlStichpunkte,
                            RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt(), 
                            Periode, PeriodenStaerke,
                            &strength))
    Durchlauf(strength);
}

     
void CG_Insert(TermpairT tp, WeightEntryT wt,
               RegelOderGleichungsT actPar, RegelOderGleichungsT otherPar)
{
  unsigned long int i = NaechstesFreies;
  
  /* ggf. dynHeap ausdehnen */
  if (NaechstesFreies == CGHeapAktuelleGroesse) {
    CGHeapAktuelleGroesse *= 2;
    CGHeap = our_realloc(CGHeap,
                         CGHeapAktuelleGroesse * sizeof (CGHeapT),
                         "CG: Expansion of Heap");
  }

  CGHeap[NaechstesFreies].tp = tp;
  CGHeap[NaechstesFreies].oP = otherPar;
  CGHeap[NaechstesFreies].aP = actPar;
  CGHeap[NaechstesFreies].w1 = wt.w1;
  CGHeap[NaechstesFreies].w2 = wt.w2;
  CGHeap[NaechstesFreies].LebtNoch = CG_lebt;

  NaechstesFreies++;

  do {
      i = (i - 1) / 2;
  } while(LasseEinsinken (i) && (i > 0));

  CG_AnzCritGoalsGesamt ++;
  CG_AnzCritGoalsActual ++;
  MaximierenLong(&CG_AnzMaxCritGoals, CG_AnzCritGoalsActual);
}


void CG_KillParent(RegelOderGleichungsT re)
/* Ueberschreibt die Eltern-Eintraege der CGs im Heap mit NULL.*/
{
  unsigned long int i;

  for (i = 0; i < NaechstesFreies; i++) {
    if ((CGHeap[i].aP->Nummer == re->Nummer) || (CGHeap[i].oP->Nummer == re->Nummer))
      CGHeap[i].LebtNoch = CG_waise;
  }
}


/* Programmende */
void CG_CleanUp (void)
{
  unsigned long int i;

  /* Heap leerraeumen */
  for (i = 0; i < NaechstesFreies; i++) {
    if (CGHeap[i].LebtNoch != CG_selektiert) {
      TPR_DESTROYTERMPAIR(CGHeap[i].tp, CGHeap[i].aP, CGHeap[i].oP);
    }
  }
}

/* fuer Wiederaufsetzen */
void CG_SoftCleanUp (void)
{
  our_dealloc (CGHeap);
  CGHeap = NULL;
}

BOOLEAN
CG_InitAposteriori (InfoStringT DEF, InfoStringT USR)
{
  int fifo, heuristic;

  if (PI_ParseInterleave (USR, &fifo, &heuristic)
      || PI_ParseInterleave (DEF, &fifo, &heuristic)) {
    modulo = fifo + heuristic;
    threshold = fifo;

#if !COMP_POWER
    fprintf (IO_Strom, "\n* Interleaving of Critical Goal Queues\n");
    fprintf (IO_Strom, "  Fifo: %d, Heuristic: %d\n", fifo, heuristic);
#endif
  } else {
    modulo = 1;
    threshold = 0;
  }
  

  SP_StichpunktlisteGenerieren(DEF,USR,&Periode,&PeriodenStaerke,
                               &StichpunktListe, &AnzahlStichpunkte,
                               4, strengthArray);
#if !COMP_POWER
  fprintf(IO_Strom, "\n* INTERMEDIATE CG-SET-REDUCTION\n\n");
  SP_Ausgabe (Periode, PeriodenStaerke,
              StichpunktListe, AnzahlStichpunkte,
              strengthArray, TRUE);
#endif

  CG_AnzCritGoalsActual = 0; /* Tatsaechliche Groesse */
  
  NaechstesFreies = 0;
  CGHeapAktuelleGroesse = 2000;
  
  CGHeap = our_alloc (CGHeapAktuelleGroesse * sizeof(CGHeapT));
  
  CG_SelectProc = CG_Select;
  return TRUE;
}


void CG_Statistics (InfoStringT IS)
{

  if (!(IS_IsIn("all", IS) || IS_IsIn("CG", IS)))
      return;
  
  if (IS_IsIn("standard", IS)) {
    IO_DruckeKurzUeberschrift("Critical Goals", '+');
    IO_DruckeStatistikZeile("total number of critical goals", CG_AnzCritGoalsGesamt);
    IO_DruckeStatistikZeile("current number of critical goals", CG_AnzCritGoalsActual);
    IO_DruckeStatistikZeile("max number of critical goals", CG_AnzMaxCritGoals);
    IO_DruckeStatistikZeile("joined during interreduction", CG_AnzIRGleich);
    IO_DruckeStatistikZeile("reduced during interreduction", CG_AnzIRReduziert);
    IO_DruckeStatistikZeile("killed orphans during interreduction", CG_AnzIRWMord);
  }
  else if (IS_IsIn("latex-head", IS)) { /* korrespondiert zu standard */
    fprintf(IO_Strom, "CG: total CGs & current CGs & Max CGs&CGIR: joined&red.&killed&");
  }
}


#if PUNKT_IRP_HACK
extern WeightEntryT AltesGewicht; /* Punkt-IRP-Hack */
#endif

static BOOLEAN CG_SelectFifo (TermpairT *tpPtr,
          RegelOderGleichungsT *aPPtr,
          RegelOderGleichungsT *oPPtr)
{
  /* Fifo */
  WeightT min = LONG_MAX;
  unsigned int PosMin = 0;
  unsigned int Pos;

  for (Pos = 0; Pos < NaechstesFreies; Pos++) {
    if ((CGHeap[Pos].LebtNoch == CG_lebt) && CGHeap[Pos].w2 < min) {
      min = CGHeap[Pos].w2;
      PosMin = Pos;
    }
  }
  /* keine lebenden CGs */
  if (min == LONG_MAX)
    return FALSE;
  
  /* Sonst: Daten holen */
  *tpPtr = CGHeap[PosMin].tp;
  *aPPtr = CGHeap[PosMin].aP;
  *oPPtr = CGHeap[PosMin].oP;
#if PUNKT_IRP_HACK
  AltesGewicht.w1 = CGHeap[PosMin].w1; /* Punkt-IRP-Hack */
  AltesGewicht.w2 = CGHeap[PosMin].w2; /* Punkt-IRP-Hack */
#endif
  CGHeap[PosMin].LebtNoch = CG_selektiert;

  /* Selektierte+Waisen am Ende loeschen */
#if 0
  /* Aufruf mal erst rausgenommen, da irgendetwas hier schief laeuft, was ich nicht verstehe */
  CG_ReorgHeapLast ();
#endif
  
  SUE_LastSelectedWasFifo = TRUE;
  return TRUE;
}


BOOLEAN
CG_Select(TermpairT *tpPtr,
          RegelOderGleichungsT *aPPtr,
          RegelOderGleichungsT *oPPtr)
{
  /* Tote an Top loeschen */
  HeapTest("Anfang Select");

  if (SUE_AnzAktivierterG % modulo < threshold) {
    return CG_SelectFifo (tpPtr, aPPtr, oPPtr);
  }

  CG_ReorgHeapFirst ();
  /* jetzt steht an Position 0 ein lebendes CP */

  /* CGHeap jetzt leer? */
  if (NaechstesFreies == 0)
    return FALSE;

  /* Sonst: Daten holen */
  CG_AnzCritGoalsActual--;
  *tpPtr = CGHeap[0].tp;
  *aPPtr = CGHeap[0].aP;
  *oPPtr = CGHeap[0].oP;
#if PUNKT_IRP_HACK
  AltesGewicht.w1 = CGHeap[0].w1; /* Punkt-IRP-Hack */
  AltesGewicht.w2 = CGHeap[0].w2; /* Punkt-IRP-Hack */
#endif
  if (--NaechstesFreies > 0) {
    CGHeap[0] = CGHeap[NaechstesFreies];
    LasseEinsinken (0);
  }
  HeapTest("Ende Select");

  SUE_LastSelectedWasFifo = FALSE;
  return TRUE;
}
