/**********************************************************************
* Filename : MNF_Ziele.h
*
* Kurzbeschreibung : Zielbehandlung mit MultipleNormalFormen
*
* Autoren : Andreas Jaeger
*
* 09.10.96: Erstellung, Teilweise aus Zielverwaltung "geklaut"
* 22.02.97: Mehrfachaufrufclean. AJ.
*
*
* Benoetigt folgende Module:
*  - MNF: Verwalten der Ziele
*
**********************************************************************/

#include "compiler-extra.h"
#include "general.h"
#include "Ausgaben.h"
#include "InfoStrings.h"
#include "Konsulat.h"
#include "MatchOperationen.h"
#include "MNF.h"
#include "MNF_Print.h"
#include "MNF_Ziele.h"
#include "pcl.h"
#include "SpeicherVerwaltung.h"
#include "Subsumption.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "Zielverwaltung.h"
#include "Zielverwaltung_private.h"
#include "ZielAusgaben.h"

/* Typen     */
/*===========*/

typedef struct privateMNFZieleTS 
{
   struct MNF_MengeTS     *nfMenge;
} privateMNFZieleT, *privateMNFZielePtrT;



/* Variablen */
/*===========*/

/* Statistikparameter */
static CounterT       MNFZ_MaxCountMNFSet;

/* Konfiguration */
static BOOLEAN        MNFZ_SubsumeGoals = TRUE;


static void MNFZ_PrintJoined (ZielePtrT goal) 
{
   TermT  joinedTerm;
   
   if (goal->status == Z_Joined) {
      if (IO_AusgabeLevel() >= 2) {
         fprintf(IO_Strom, "joined goal:      %8u  ", goal->nummer);
         joinedTerm = MNF_GetJoinedTerm (goal->privat.MNF->nfMenge);
         IO_DruckeFlex("%t ?= %t to %t\n", goal->orginalLinks, goal->orginalRechts,
                       joinedTerm);
      }
      else
         fprintf(IO_Strom, "joined goal:      %8u\n", goal->nummer);
      ZA_ZielZusammengefuehrtDrucken();
      if (IO_AusgabeLevel() > 2 && !COMP_POWER)
         MNF_PrintJoinedComplete (goal->privat.MNF->nfMenge);
      
      MNF_PrintJoinedPCL (goal->privat.MNF->nfMenge, &goal->pclID);
   }
   else { /* goal->status == Z_Subsumed */
      MNF_PrintSubsumedPCL (goal->privat.MNF->nfMenge, &goal->pclID,
                            TP_PclID(MO_SubsummierendeGleichung));
      ZA_ZielSubsummiertDrucken (goal->nummer, TP_ElternNr(MO_SubsummierendeGleichung),
				 goal->orginalLinks, goal->orginalRechts,
				 TP_LinkeSeite(MO_SubsummierendeGleichung),
				 TP_RechteSeite(MO_SubsummierendeGleichung));
   }
}


/* ------------------------------------------------------------------------- */
/*                                                                           */
/* ------------------------------------------------------------------------- */

static void MNFZ_MNFStat (ZieleT *ziel) 
{
   unsigned int size;
   
   size = MNF_GetSize (ziel->privat.MNF->nfMenge);
   if (size >  MNFZ_MaxCountMNFSet)
      MNFZ_MaxCountMNFSet = size;
   if (ziel->status != Z_Open)
         MNFZ_PrintJoined (ziel);
}

void MNFZ_InitZielBeobachtungsProz (ZielePtrT zielPtr) {

  zielPtr->beobachtProz (zielPtr->orginalLinks, zielPtr->orginalRechts);
}

void MNFZ_ZielEinfuegen (ZieleT * ziel) 
{
   ziel->privat.MNF = (privateMNFZieleT *)our_alloc (sizeof(privateMNFZieleT));
   ziel->privat.MNF->nfMenge = MNF_InitNMenge (ziel->orginalLinks,
                                               ziel->orginalRechts);
}


void MNFZ_LoeschePrivateData (ZieleT *ziel) 
{
   MNF_DeleteNMenge ( ziel->privat.MNF->nfMenge);
   our_dealloc (ziel->privat.MNF);
}


void MNFZ_AusgabeZiel (ZieleT * ziel,Z_AusgabeArtT Z_AusgabeArt) 
{
   TermT left, right;
   
   switch (Z_AusgabeArt) {
   case Z_AusgabeAlles:
      if (MNF_JoinedNMenge (ziel->privat.MNF->nfMenge)) {
         MNF_GetJoinedTermpair (ziel->privat.MNF->nfMenge, &left, &right);
         IO_DruckeFlex(" current  %t = %t\n", left, right);
         MNF_PrintJoinedComplete (ziel->privat.MNF->nfMenge);
      }
#if 0
        MNF_PrintMengeComplete (ziel->privat.MNF->nfMenge, "lhs: ", "rhs: ");
#endif
      break;
   case Z_AusgabeInitial:
      if (MNF_JoinedNMenge (ziel->privat.MNF->nfMenge)) {
         MNF_GetJoinedTermpair (ziel->privat.MNF->nfMenge, &left, &right);
         IO_DruckeFlex(" current  %t = %t\n", left, right);
      }
      break;
   case Z_AusgabeKurz:
      if (MNF_JoinedNMenge (ziel->privat.MNF->nfMenge)) {
         MNF_GetJoinedTermpair (ziel->privat.MNF->nfMenge, &left, &right);
         IO_DruckeFlex("%t = %t", left, right);
      }
      break;
   case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
   }
}

   
void MNFZ_ZielBeobachtung (ZieleT *ziel, Z_BeobachtungsDurchlaufProcT AP) 
{
   AP (ziel->orginalLinks, ziel->orginalRechts, ziel->nummer);
}


INLINE static BOOLEAN MNF_ZielErledigt (ZieleT *ziel, BOOLEAN expand)
{
   BOOLEAN               joined;
   
   joined = MNF_JoinedNMenge (ziel->privat.MNF->nfMenge);
   if (!joined && expand) {
      MNF_ExpandNMenge (ziel->privat.MNF->nfMenge);
      joined = MNF_JoinedNMenge (ziel->privat.MNF->nfMenge);
   }
   if (joined)
      ziel->status = Z_Joined;
   else if (MNFZ_SubsumeGoals && expand) {
      if (MNF_DurchlaufPaarweise (ziel->privat.MNF->nfMenge, SS_TermpaarSubsummiertVonGM)) {
         ziel->subSumNr = TP_ElternNr(MO_SubsummierendeGleichung);
         ziel->status = Z_Subsumed;
      }
   }
   MNFZ_MNFStat (ziel);
   
   return ziel->status != Z_Open;
}

BOOLEAN MNFZ_TesteZielErledigt (ZieleT *ziel) 
{
   return (MNF_ZielErledigt (ziel, FALSE));
}

BOOLEAN MNFZ_ZielErledigtMitNeuerRegel (ZieleT *ziel, RegelT Regel,
                                        BOOLEAN expand) 
{
   BOOLEAN Resultat;
   MNF_AddRegelNM (ziel->privat.MNF->nfMenge, Regel);
   Resultat = MNF_ZielErledigt (ziel, expand);
   if (Resultat && PA_KernAn)
     KO_KonsultationErforderlich = TRUE;
   return Resultat;
}



BOOLEAN MNFZ_ZielErledigtMitNeuerGleichung (ZieleT *ziel, GleichungsT Gleichung,
                                            BOOLEAN expand) 
{
   MNF_AddGleichungNM (ziel->privat.MNF->nfMenge, Gleichung);
   return MNF_ZielErledigt (ziel, expand);
}


BOOLEAN MNFZ_ZielFinalerDurchlauf (ZieleT * ziel) 
{
   return MNF_ZielErledigt (ziel, TRUE);
}


void MNFZ_Statistik(InfoStringT str) 
{
   if (IS_IsIn("standard", str)){
      IO_DruckeStatistikZeile("Maximal size of elems in goal set",  MNFZ_MaxCountMNFSet);
   }
   else if (IS_IsIn("latex-head", str))
      fprintf(IO_Strom, "Max Size &");
   else if (IS_IsIn("short", str)) {
      IO_DruckeKurzUeberschrift("Treatment of Goals", '.');
      IO_DruckeStatistikZeile("Maximal size of elems in goal set",  MNFZ_MaxCountMNFSet);
   }
   else if (IS_IsIn("slatex-head", str)){
      fprintf(IO_Strom, "G:MaxSize &");
   }
}


static void MNFZ_ParseInfo (InfoStringT DEF, InfoStringT USR, BOOLEAN *subsume) 
{
   *subsume = IS_FindOption (DEF, USR, 1, "s") == 0;

#if !COMP_POWER
   fprintf (IO_Strom, "    MNF Goals: %s use Subsumption.\n", *subsume? "":"don't");
#endif
}



BOOLEAN MNFZ_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  
  MNFZ_ParseInfo (DEF, USR, &MNFZ_SubsumeGoals);

  return MNF_InitNMengeAposteriori (DEF, USR);
}


void MNFZ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   BOOLEAN  subsume;

   switch (what) {
   case ParamEvaluate:
      MNFZ_ParseInfo (DEF, USR, &subsume);
      break;
   case ParamExplain:
      break;
   }
   MNF_ParamInfo (what, DEF, USR);
}
