/**********************************************************************
* Filename : NormaleZiele.c
*
* Kurzbeschreibung : Normale Zielverwaltung
*
* Autoren : Andreas Jaeger
*
* 10.10.96: Erstellung aus Zielverwaltung.h heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
*
* Die Funktionen in diesem Modul werden nie direkt aufgerufen, sondern
* nur von der Zielverwaltung!
*
**********************************************************************/

#include <stdlib.h>
#include <string.h>

#include "compiler-extra.h"
#include "antiwarnings.h"
#include "general.h"
#include "pcl.h"

#include "Ausgaben.h"
#include "InfoStrings.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "NormaleZiele.h"
#include "RUndEVerwaltung.h"
#include "Subsumption.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "Zielverwaltung.h"
#include "Zielverwaltung_private.h"
#include "ZielAusgaben.h"

/*********************/
/* Typdeklarationen  */
/*********************/

/*
  reduziert: wird gesetzt, wenn links oder rechts reduziert wurde.  Falls im
  Wiederholenmodus gearbeitet wird, kann somit festgestellt werden, ob links/rechts
  die originalen Ziele sind, oder reduzierte Elemente.
*/
typedef struct privateNormaleZieleTS {
   TermT               links;
   TermT               rechts;
   BOOLEAN             reduziert;
#if COMP_PCL
   PCL_ObjektNummerT   orig_pclID;
#endif   
} privateNormaleZieleT, *privateNormaleZielePtrT;


void NZ_InitZielBeobachtungsProz (ZieleT  *ziel) {

  ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
}


void NZ_ZielEinfuegen (ZieleT * ziel, TermT links, TermT rechts) 
{
   ziel->privat.NZ = (privateNormaleZieleT *) our_alloc (sizeof(privateNormaleZieleT));
   ziel->privat.NZ->links = TO_Termkopie (links);
   ziel->privat.NZ->rechts = TO_Termkopie (rechts);
   ziel->privat.NZ->reduziert = FALSE;
#if COMP_PCL
   ziel->privat.NZ->orig_pclID = ziel->pclID;
#endif   
}


void NZ_LoeschePrivateData (ZieleT *ziel) 
{
   TO_TermLoeschen (ziel->privat.NZ->links);
   TO_TermLoeschen (ziel->privat.NZ->rechts);
   our_dealloc (ziel->privat.NZ);
}

INLINE static void InitElemente (ZieleT *ziel)
{
   if (ziel->privat.NZ->reduziert && ziel->modul == Z_Repeat) {
      TO_TermLoeschen (ziel->privat.NZ->links);
      TO_TermLoeschen (ziel->privat.NZ->rechts);
      ziel->privat.NZ->links = TO_Termkopie (ziel->orginalLinks);
      ziel->privat.NZ->rechts = TO_Termkopie (ziel->orginalRechts);
      ziel->privat.NZ->reduziert = FALSE;
#if COMP_PCL
      ziel->pclID = ziel->privat.NZ->orig_pclID;
#endif
   }
}

void NZ_AusgabeZiel (ZieleT * ziel, Z_AusgabeArtT Z_AusgabeArt) 
{
  if (ziel->modul == Z_Paramodulation) {
    switch (Z_AusgabeArt) {
    case Z_AusgabeAlles:
    case Z_AusgabeInitial:
      IO_DruckeFlex ("%t ?= %t\n",
                     ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeFlex ("        using narrowing to prove %t ?= %t\n", 
                     ziel->orginalLinks, ziel->orginalRechts);
      break;
    case Z_AusgabeKurz:
      IO_DruckeFlex ("true ?= false, current %t ?= %t\n", 
                     ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeFlex ("        using narrowing to prove %t ?= %t\n", 
                     ziel->orginalLinks, ziel->orginalRechts);
      break;
    case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
    }
  } else {
    switch (Z_AusgabeArt) {
    case Z_AusgabeAlles:
    case Z_AusgabeInitial:
      IO_DruckeFlex(" current  %t ?= %t\n", ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      break;
    case Z_AusgabeKurz:
      IO_DruckeFlex("%t = %t", ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      break;
    case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
    }
  }
}

   
void NZ_ZielBeobachtung (ZieleT *ziel, Z_BeobachtungsDurchlaufProcT AP)  
{
   AP (ziel->privat.NZ->links, ziel->privat.NZ->rechts, ziel->nummer);
}


#define Test_SubsummiertVonGM(ZIEL)                                                                     \
       if (SS_TermpaarSubsummiertVonGM (ZIEL->privat.NZ->links, ZIEL->privat.NZ->rechts))               \
         {                                                                                              \
            /* falls Ziel subsummiert */                                                                \
            PCL_ZielSubsummiert (&ZIEL->pclID,                                                          \
                                 ZIEL->privat.NZ->links,                                                \
                                 ZIEL->privat.NZ->rechts,                                               \
                                 TP_PclID(MO_SubsummierendeGleichung));                                 \
            ZA_ZielSubsummiertDrucken (ZIEL->nummer, TP_ElternNr(MO_SubsummierendeGleichung),           \
                                            ZIEL->privat.NZ->links, ZIEL->privat.NZ->rechts,            \
                                            TP_LinkeSeite(MO_SubsummierendeGleichung),                  \
                                            TP_RechteSeite(MO_SubsummierendeGleichung));                \
            ZIEL->subSumNr = TP_ElternNr(MO_SubsummierendeGleichung);                                   \
            ZIEL->status = Z_Subsumed;                                                                  \
            return TRUE;                                                                                \
         }


BOOLEAN NZ_TesteZielErledigt (ZieleT *ziel) 
{
   if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
      PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (IO_AusgabeLevel() >= 2) {
         fprintf(IO_Strom, "joined goal:    %8u  ", ziel->nummer);
         IO_DruckeFlex("%t ?= %t to %t\n", ziel->orginalLinks, ziel->orginalRechts,
                       ziel->privat.NZ->rechts);
      }
      else
         fprintf(IO_Strom, "joined goal:    %8u\n", ziel->nummer);
      ZA_ZielZusammengefuehrtDrucken(); /* ... und somit zusammengefuehrt.*/
      ziel->status = Z_Joined;
      return TRUE;
   } else
      Test_SubsummiertVonGM (ziel);
   return FALSE;
}


BOOLEAN NZ_ZielErledigtMitNeuerRegel (ZieleT *ziel, RegelT Regel)
{
   BOOLEAN reduziert;

   InitElemente (ziel);
   if (ziel->modul == Z_Repeat) {
     PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
     reduziert = NF_NFBildung (ziel->privat.NZ->links, Ziele);
     PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
     reduziert |= NF_NFBildung (ziel->privat.NZ->rechts, Ziele);
   } else {
     PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
     reduziert = NF_NFMitNeuerRegel(Regel, ziel->privat.NZ->links, Ziele);
     PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
     reduziert |= NF_NFMitNeuerRegel(Regel, ziel->privat.NZ->rechts, Ziele);
   }
   if (reduziert){
      ziel->privat.NZ->reduziert = TRUE;
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (ziel->modul != Z_Repeat)
	ZA_ZielReduziertDrucken (ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (NZ_TesteZielErledigt (ziel))
         return TRUE;
   }
   return FALSE;
}


BOOLEAN NZ_ZielErledigtMitNeuerGleichung (ZieleT *ziel, GleichungsT Gleichung)
{
   BOOLEAN reduziert;
   
   InitElemente (ziel);
   if (ziel->modul == Z_Repeat) {
     PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
     reduziert = NF_NFBildung (ziel->privat.NZ->links, Ziele);
     PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
     reduziert |= NF_NFBildung (ziel->privat.NZ->rechts, Ziele);
   } else {
     PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
     reduziert = NF_NFMitNeuerGleichung(Gleichung, ziel->privat.NZ->links, Ziele);
     PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
     reduziert |= NF_NFMitNeuerGleichung(Gleichung, ziel->privat.NZ->rechts, Ziele);
   }
   if (reduziert) {
      ziel->privat.NZ->reduziert = TRUE;
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (ziel->modul != Z_Repeat)
	 ZA_ZielReduziertDrucken (ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (NZ_TesteZielErledigt (ziel))
         return TRUE;
   } else
      /* wenn nicht reduziert wurde muss nur auf Subsumption mit neuer Gleichung getestet
         werden */
      if (SS_GleichungSubsummiertTermpaar (Gleichung, ziel->privat.NZ->links,
                                           ziel->privat.NZ->rechts))
         {
            /* falls Ziel subsummiert */
            PCL_ZielSubsummiert (&ziel->pclID,
                                 ziel->privat.NZ->links,
                                 ziel->privat.NZ->rechts,
                                 TP_PclID(Gleichung));
            ZA_ZielSubsummiertDrucken (ziel->nummer, TP_ElternNr(Gleichung),
				       ziel->privat.NZ->links,
				       ziel->privat.NZ->rechts,
				       TP_LinkeSeite(Gleichung),
				       TP_RechteSeite(Gleichung));
            ziel->subSumNr = TP_ElternNr(Gleichung);
            ziel->status = Z_Subsumed;
            return TRUE;
         }
   return FALSE;
}


BOOLEAN NZ_ZielFinalerDurchlauf (ZieleT * ziel) 
{
   BOOLEAN reduziert;

   InitElemente (ziel);
   PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
   reduziert = NF_NFBildung (ziel->privat.NZ->links, Ziele);
   PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
   reduziert |= NF_NFBildung (ziel->privat.NZ->rechts, Ziele);
   if (reduziert){
      ziel->privat.NZ->reduziert = TRUE;
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (ziel->modul != Z_Repeat)
	 ZA_ZielReduziertDrucken (ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
         PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
         ZA_ZielZusammengefuehrtDrucken(); /* ... und somit zusammengefuehrt.*/
         ziel->status = Z_Joined;
         return TRUE;
      }
   }
   Test_SubsummiertVonGM (ziel);

   return FALSE;
}


/*=================================================================================================================
  =  Zielbehandlung                                                                                               =
  =================================================================================================================*/


void NZ_Statistik(InfoStringT str)
{
   if (!(IS_IsIn("all", str) || IS_IsIn("Z", str)))
      return;
}


static void NZ_ParseInfoStrings (InfoStringT DEF MAYBE_UNUSEDPARAM,
                                 InfoStringT USR MAYBE_UNUSEDPARAM) 
{
   ;
}


void NZ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   switch (what) {
   case ParamEvaluate:
      NZ_ParseInfoStrings (DEF, USR);
      break;
   case ParamExplain:
      fprintf (IO_Strom, "Classical Goal Treatment (without MNF)\n");
      fprintf (IO_Strom, "**************************************\n");
      fprintf (IO_Strom, "To select: snf (keeps goals always reduced)\n");
      fprintf (IO_Strom, "To select: rnf (tries to reduce initial goal)\n");
      break;
   }
}


BOOLEAN NZ_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   NZ_ParseInfoStrings (DEF, USR);
   return TRUE;
}

