/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Hauptkomponenten.c                                                                                         */
/* - *******************                                                                                        */
/* -                                                                                                            */
/* - Enthaelt die Operationen, die zum Programm-Start mit den in der KP-Menge enthaltenen Gleichungen der Spezi-*/
/* - fikation ausgefuehrt werden (I) sowie eine Prozedur, die alle KP's der Reihe nach behandelt und sie zu-    */
/* - sammenzufuehren sucht (II).                                                                                */
/* - 29.06.94 Arnim. AlsRegel/GleichungAufnehmen fertiggemacht.                                                 */
/* - 10.94/1.11.94 Arnim. Prozeduren fuer die failing completion eingebaut.                                     */
/* -------------------------------------------------------------------------------------------------------------*/
#include "antiwarnings.h"
#include "compiler-extra.h"
#include "general.h"

#include <string.h>
#include <math.h>

#include "Ausgaben.h"
#include "CritGoals.h"
#include "DSBaumOperationen.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "KBO.h"
#include "Konsulat.h"
#include "KPBildung.h"
#include "KPHeap.h"
#include "KPMengenInterreduktion.h"
#include "LPO.h"
#include "LispLists.h"
#include "MNF.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "Ordnungen.h"
#include "Parameter.h"
#include "PhilMarlow.h"
#include "PriorityQueue.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "SelectClassify.h"
#include "Signale.h"
#include "SpezNormierung.h"
#include "Statistiken.h"
#include "Subgleichungen.h"
#include "Subsumption.h"
#include "Termpaare.h"
#include "Unifikation1.h"
#include "WaldmeisterII.h"
#include "general.h"
#include "pcl.h"
#include "Zielverwaltung.h"



/* -------------------------------------------------------------------------------*/
/* --------------- KPMenge Behandeln (I und II) ----------------------------------*/
/* -------------------------------------------------------------------------------*/

BOOLEAN HK_fair;

static void AlsGleichungAufnehmen(TermT links, TermT rechts, BOOLEAN ProofModus);
static void AlsRegelAufnehmenOhneGM(BOOLEAN Kreuzen);
static void AlsRegelAufnehmen_FAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus);

/* Ausgaben  */
/*-----------*/
#if IO_AUSGABEN
static void NeueRegelDrucken(RegelT regel)
{
   if (IO_level2){
#if 0
       signed long int a = min(TP_ElternNr(SUE_selCPactualParent),TP_ElternNr(SUE_selCPotherParent));
       signed long int b = max(TP_ElternNr(SUE_selCPactualParent),TP_ElternNr(SUE_selCPotherParent));
       fprintf(IO_Strom, "new rule:            %5ld  (%5ld,%5ld)  ", regel->Nummer, a, b);
#else       
       fprintf(IO_Strom, "new rule:            %5ld  ", regel->Nummer);
#endif
      IO_DruckeFaktum(TP_LinkeSeite(regel), TRUE, TP_RechteSeite(regel));
   }
   else if (IO_level3)
      fprintf(IO_Strom, "... and added as new rule %ld.\n\n", regel->Nummer);
   else if (IO_level4){
      fprintf(IO_Strom, "... and added as new rule %ld:\n", regel->Nummer);
      IO_DruckeFaktum(TP_LinkeSeite(regel), TRUE, TP_RechteSeite(regel));
      fprintf(IO_Strom, "\n");
   }
}

static void NeueGleichungDrucken(GleichungsT gl)
{
   if (IO_level2){
#if 0
      signed long int a = min(TP_ElternNr(SUE_selCPactualParent),TP_ElternNr(SUE_selCPotherParent));
      signed long int b = max(TP_ElternNr(SUE_selCPactualParent),TP_ElternNr(SUE_selCPotherParent));
      fprintf(IO_Strom, "new equation:        %5ld  (%5ld,%5ld)  ", -gl->Nummer, a, b);
#else      
      fprintf(IO_Strom, "new equation:        %5ld  ", -gl->Nummer);
#endif      
      IO_DruckeFaktum(TP_LinkeSeite(gl), FALSE, TP_RechteSeite(gl));
   }
   else if (IO_level3)
      fprintf(IO_Strom, "...and added to E as %s with number %ld.\n\n",TP_IstMonogleichung(gl)?"mono equation":"such", -gl->Nummer);
   else if (IO_level4){
      fprintf(IO_Strom, "...and added to E as %s with number %ld:\n       ",TP_IstMonogleichung(gl)?"mono equation":"such", -gl->Nummer);
      IO_DruckeFaktum(TP_LinkeSeite(gl), FALSE, TP_RechteSeite(gl));
   }
}


static void GebildeteGleichungDrucken (void)
{
   if (IO_level3 || IO_level4) {
      fprintf(IO_Strom, "not orientable, added to E-selected\n");
   }
}
#else /* Dummies benutzen */
#define NeueRegelDrucken(regel)
#define NeueGleichungDrucken(gl)
#define GebildeteGleichungDrucken()
#endif

/* (I) */
/*-----*/



BOOLEAN HK_GewichtsloseKPsBehandeln(void)
/* Behandelt der Reihe nach die kritischen Paare mit Gewicht Null (also die Ausgangsgleichungen) in
   Abhaengigkeit von den verschiedenen (einfachen!) Modi wie folgt:

   reduction:        Richten links->rechts,  -> R.
   termination:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer.
   confluence :      Richten links->rechts, -> R. KP-Bildung.
   convergence:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer. KPBildung.

   Rueckgabe                  == TRUE   : Alle behandelten KP's konnten gerichtet werden.
                                 FALSE  : nicht alle behandelten KP's konnten gerichtet werden.

   Seiteneffekte                        : Nicht richtbare Ausgangsgleichungen werden ausgegeben und in TP-Puffer aufgenommen.

   Vor.: Alle KPs mit Gewicht 0 sind noch gueltig.
   In allen Modi bleibt die Gleichungsmenge leer.
*/
{
   BOOLEAN AlleRichtbar = TRUE;
   TermT links, rechts;
   RegelOderGleichungsT rog;

#ifndef _ACIDWM_AN
   IO_DruckeUeberschrift("Gleichungen aus der Spezifikation behandeln");
   IO_DruckeFlex("... Richten %s und \n    %s Interreduktion und KP-Bildung in Regel- bzw. Gleichungsmenge aufnehmen.\n\n",
                 ((PA_ExtModus == termination)||(PA_ExtModus==convergence))? "nach Ordnung":"von links nach rechts",
                 ((PA_ExtModus == confluence)||(PA_ExtModus==convergence))? "mit":"ohne");
#endif


   if ((PA_ExtModus == confluence) || (PA_ExtModus == convergence) || (PA_ExtModus == termination))
      PU_TPPufferNeu();

   while(SUE_SelectInitial()){
      links = SUE_selectedCPlhs();
      rechts = SUE_selectedCPrhs();

      switch (PA_ExtModus){

      case confluence:                                                      /* Konfluenz */
        SUE_selectedCPRE();
        AlsRegelAufnehmenOhneGM(FALSE);
        break;

      case convergence:                                                     /* Konvergenz */
        switch(ORD_VergleicheTerme(links, rechts)){
        case Groesser:
           SUE_selectedCPRE();
           AlsRegelAufnehmenOhneGM(FALSE);
           break;
        case Kleiner:
           SUE_selectedCPRE();
           AlsRegelAufnehmenOhneGM(TRUE);
           break;
        case Unvergleichbar:
           /* In diesem Falle werden die KPs zu den nachfolgenden Regeln mit der Regelmenge weiterhin
              gebildet. Da sie aber nicht behandelt werden, wird das Zeitverhalten davon nicht merklich
              beeinflu\3t .*/
           SUE_selectedCPRE();
           PU_TPEinfuegen(links, rechts, SUE_selectedCPNr());
           AlleRichtbar = FALSE;
           break;
        default:
           SUE_selectedCPjoined();
           break;
        }
        break;

      case reduction:                                                       /* Reduktion */
        RE_IncAnzRegelnGesamt();
        rog = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
#ifndef _ACIDWM_AN
        NeueRegelDrucken(rog);
#endif
        RE_ZuletztErzeugteRegelEinfuegen();
        SUE_selectedCPRE();
        break;

      case termination:                                                      /* Termination */
        switch(ORD_VergleicheTerme(links, rechts)){
        case Groesser:
          RE_IncAnzRegelnGesamt();
          rog = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
          NeueRegelDrucken(rog);
          RE_ZuletztErzeugteRegelEinfuegen();
          SUE_selectedCPRE();
          break;
        case Kleiner:
          RE_IncAnzRegelnGesamt();
          rog = RE_ErzeugteRegel(rechts, links, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
          NeueRegelDrucken(rog);
          RE_ZuletztErzeugteRegelEinfuegen();
          SUE_selectedCPRE();
          break;
        case Unvergleichbar:
          RE_IncAnzGleichungenGesamt();
          /* PU_TPEinfuegen(links, rechts, SUE_NummerAktKP());
             SUE_selectedCPRE(); */
          rog = RE_ErzeugteGleichung(links, rechts, RE_AnzGleichungenGesamt(), SUE_selectedCPPclID());
          GebildeteGleichungDrucken ();
          NeueGleichungDrucken(rog);
          RE_ZuletztErzeugteHinrichtungEinfuegen(NULL);
          RE_ZuletztErzeugteHerrichtungEinfuegen();
          AlleRichtbar = FALSE;
          break;
        default:
          SUE_selectedCPjoined();
          break;
        }
        break;
    default: break;                    /* in den Modi proof und completion wird diese Routine sowieso nicht aufgerufen */
    }
  }
#ifndef _ACIDWM_AN
  IO_LF();
#endif
  return AlleRichtbar;
}

/* (II) */
/*------*/

BOOLEAN HK_KPsZusammenfuehrbar(void)
/* Behandelt alle kritischen Paare der Reihe nach (wie sie in der KP-Menge zu finden sind) wie folgt:
   Es wird versucht, linke und rechte Seite zusammenzufuehren. War dies moeglich,
   so wird fortgefahren, andernfalls wird das kritische Paar (nach der Reduktion!) ausgegeben
   Der Rueckgabewert gibt an, ob alle KP's zusammenfuehrbar waren.
   Seiteneffekt : nicht richtbare KP's werden ausgegeben und im TP-Puffer gespeichert.
*/
{
   BOOLEAN AlleZusammenfuehrbar = TRUE;
   TermT links, rechts;

   IO_DruckeUeberschrift("Kritische Paare auf Zusammenfuehrbarkeit testen");
   PU_TPPufferNeu();
   while(SUE_Select()){
      links = SUE_selectedCPlhs();
      rechts = SUE_selectedCPrhs();
      PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), linkeSeite, rechts);
      NF_NFBildungZuRegeln(links, KPsSel);
      PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), rechteSeite, links);
      NF_NFBildungZuRegeln(rechts, KPsSel);
      if (!TO_TermeGleich(links, rechts)){
         RE_IncAnzGleichungenGesamt();
#if IO_AUSGABEN
         if ((IO_level3) || (!IO_level4))
            IO_DruckeFlex("couldn't be joined!\n %t # %t\n",links,rechts);
#endif
         PU_TPEinfuegen(links, rechts, SUE_NummerAktKP());
         SUE_selectedCPRE();
         AlleZusammenfuehrbar = FALSE;
      }
      else
         SUE_selectedCPjoined();
   }
   IO_DruckeFlex("\n\n\nEs waren %salle kritischen Paare zusammenfuehrbar. :-%s\n\n",
		 (AlleZusammenfuehrbar)? "":"nicht ",
		 (AlleZusammenfuehrbar)? ")":"(");
   return AlleZusammenfuehrbar;
}


/* -------------------------------------------------------------------------------*/
/* --------------- Ziel-Behandlung (III) -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

void HK_GleichungenUndZieleHolen(void)
/* Die Gleichungen aus der Spezifikation plattklopfen und mit Gewicht 0 in die KP-Menge stecken.
   Fuer jedes Symbol testen, ob es AC-Eigenschaft hat.
   Die Ziele aus der Spezifikation plattklopfen, dabei skolemisieren und in die Ziele-Liste stecken.*/
{
   ListenT Index;
#if !COMP_ACIDWM && !COMP_POWER
   if (PA_InteraktionAn)
     IO_DruckeUeberschrift("Fetching goals and equations");
#endif
   if (PA_ExtModus!=completion && PA_ExtModus!=proof) IO_AusgabenAus();

   for (Index=PM_Gleichungen; Index; Cdr(&Index)) {
     ListenT Eintrag = car(Index);
     TermT Flunder1 = TO_Termkopie(car(Eintrag)),
           Flunder2 = TO_Termkopie(cdr(Eintrag));
     SUE_CPinitial(Flunder1,Flunder2);
     SUE_recentCPinsert();
   }
   Z_ZieleAusSpezifikationHolen();

   if (PA_ExtModus!=completion && PA_ExtModus!=proof) IO_AusgabenEin();
}


/* (III) */
/*-------*/


BOOLEAN HK_ZieleAufZusammenfuehrbarkeitTesten(void)
/* Alle Ziele aus der Zielmenge bzgl. des aktuellen Regel- und Gleichungs-Systems abschliessend auf NF bringen. */
{
   if (Z_ZielmengeLeer())
     return TRUE;
   IO_DruckeUeberschrift("Treatment of Goals");
   return Z_FinalerZielDurchlauf();
}


/* -------------------------------------------------------------------------------*/
/* --------------- Aufnahme neuer Regeln/Gleichungen in RM/GM---------------------*/
/* -------------------------------------------------------------------------------*/

static RegelT zuletztEingefuegteRegel;
static GleichungsT zuletztEingefuegteGleichung;

static void AlsRegelAufnehmenOhneGM(BOOLEAN Kreuzen)
/* Nimmt kp.links=kp.rechts als Regel auf, und zwar von links nach rechts, falls Kreuzen == FALSE,
   ist Kreuzen == TRUE, dann von rechts nach links, fuehrt die Interreduktion der rechten Seiten der Regelmenge
   sowie die KP-Bildung durch.
   Gleichungen werden nicht beachtet!!!!!!!
   Vor.: linke und rechte Seite des KPs sind auf NF bzgl. R!
   Diese Prozedur wird in HK_GewichtsloseKPsbehandlen in den Modi confluence und convergence verwendet.
   Es erfolgt keine Interreduktion auf den linken Regel-Seiten!
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;
   if (Kreuzen){                                                     /* Richtige Richtung herausfinden: */
     links = SUE_selectedCPrhs();
     rechts = SUE_selectedCPlhs();
   }
   else{
     links = SUE_selectedCPlhs();
     rechts = SUE_selectedCPrhs();
   }
   RE_IncAnzRegelnGesamt();
   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
   NeueRegelDrucken(zuletztEingefuegteRegel);
   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts,
                 TP_PclID(zuletztEingefuegteRegel), zuletztEingefuegteRegel->Nummer);
   RE_ZuletztErzeugteRegelEinfuegen();
   U1_KPsBildenZuRegelFailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.
      Dabei dient das KP als Eltern-Verweis fuer das PCL-Protokoll.*/
   /* =====> KPMenge ist jetzt komplett; */
}

static void AlsGleichungAufnehmen(TermT links, TermT rechts, BOOLEAN ProofModus)
/* Nimmt selectedCP als Gleichung auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R und R(E)!
   Rueckgabe == TRUE, wenn die neue Gleichung nicht wegsubsummiert wurde!*/
{
   RE_IncAnzGleichungenGesamt();
   zuletztEingefuegteGleichung = RE_ErzeugteGleichung(links, rechts, RE_AnzGleichungenGesamt(), SUE_selectedCPPclID());
   NeueGleichungDrucken(zuletztEingefuegteGleichung);
   PCL_KPZuGleichung (links, rechts, TP_PclID(zuletztEingefuegteGleichung),-zuletztEingefuegteGleichung->Nummer);
   IR_NeueGleichungIst(zuletztEingefuegteGleichung);
   IR_InterreduktionStarten();						    /* Meldung; Puffer vorbereiten */
   IR_GMSubsummierenMitGleichung();
   /* Suche alle mit l=r oder r=l subsummierbaren Gleichungen
      und entferne sie aus der GM (---> wegwerfen, nicht puffern!!!)*/
   IR_GMInterredMitGleichung();
   /* Suche alle mit l=r oder r=l reduzierbaren Gleichungen, fuehre den Reduktionsschritt
      aus und entferne sie aus der GM (--->PU_REPuffer)*/
   IR_RMLinksInterredMitGleichung();
   /* Suche alle links mit l=r oder r=l reduzierbaren Regeln, fuehre den Reduktionsschritt aus und entferne sie aus RM
      (---> PU_REPuffer) */
   /* ====> GM und RM sind links interreduziert */
   RE_ZuletztErzeugteHinrichtungEinfuegen(SGM_SubgleichungenZuObjektAufnehmen((SGM_BetroffenesObjektT) zuletztEingefuegteGleichung));
   /* Fuege l=r als Gleichung in GM ein. Spaete Aufnahme erspart Ausschluss bei
      Subsumption und Interred. der GM. */
   RE_ZuletztErzeugteHerrichtungEinfuegen();
   /* ====> NF-Bildung wieder moeglich (dh. minimal) */
   IR_RMRechtsInterredMitGleichung();
   /* Suche alle rechts mit l=r oder r=l reduzierbaren Regeln, fuehre den Reduktionsschritt aus
      und fahre mit der NF-Bildung fort*/
   IR_InterreduktionBeendet();
   /* ====> RM ist auch rechts interreduziert */
   /* ====> System ist interreduziert */
   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerGleichung(zuletztEingefuegteGleichung);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();     /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteGleichung, NULL, NULL);
         SUE_RedirectParent();
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
      /* ====> Ziele sind bearbeitet */
   }
   KPIR_KPMengeInterreduzieren();
   /* Interreduktion der KP-Menge erst hier moeglich, da vorher NF-Bildung nicht vollst.*/
   /* insb. auch vor KP-Bildung (sonst doppelte Arbeit auf den neuen KPs) */
   CG_CGMengeInterreduzieren(FALSE, "");
   U1_KPsBildenZuGleichung(zuletztEingefuegteGleichung);
   IR_PufferAuslesen();
   /* =====> KPMenge ist jetzt komplett; */
   KPB_KonklusionenAbarbeiten(zuletztEingefuegteGleichung);
}

static void AlsRegelAufnehmen_UNFAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus)
/* Nimmt links=rechts als Regel auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R und R(E)!
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;
   if (Kreuzen){
     links = rhs;
     rechts = lhs;
   }
   else{
     links = lhs;
     rechts = rhs;
   }
   RE_IncAnzRegelnGesamt();
   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
   NeueRegelDrucken(zuletztEingefuegteRegel);
   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts,
                 TP_PclID(zuletztEingefuegteRegel),zuletztEingefuegteRegel->Nummer);
   IR_NeueRegelIst(zuletztEingefuegteRegel);
   IR_InterreduktionStarten();
   IR_RMLinksInterredMitRegel();
   /* Suche alle Regeln, deren linke Seite von l gematcht werden, fuehre diesen einen Reduktionsschritt aus und
      entferne sie aus der Regelmenge; (wohin? ---> Puffer).
      Vorgezogen, um Ausschluss ueberfluessig zu machen */
   IR_GMInterredMitRegel();
   /* Suche alle mit l->r reduzierbaren Gleichungen, fuehre diesen einen Reduktionsschritt aus und entferne sie aus GM;
      (wohin? ---> Puffer) */
   /* =====> GM ist wieder interreduziert */
   /* =====> NF-Bildung ist wieder konsistent.*/
   RE_ZuletztErzeugteRegelEinfuegen();                 /* Fuege l->r als neue Regel in RM ein; */
   IR_RMRechtsInterredMitRegel();
   /* Suche alle rechts mit l->r reduzierbaren Regeln, fuehre den Reduktionsschritt aus und
      rufe die Normalformbildung dazu auf */
   /* =====> RM ist wieder interreduziert */
   IR_InterreduktionBeendet();
   /* =====> System ist interreduziert */
   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerRegel(zuletztEingefuegteRegel);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();      /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteRegel, NULL, NULL);
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
         /* ====> Ziele sind bearbeitet */
   }
   KPIR_KPMengeInterreduzieren(); /* Interreduktion der KP-Menge vor KP-Bildung (sonst doppelte Arbeit auf den neuen KPs) */
   CG_CGMengeInterreduzieren(FALSE, "");
   U1_KPsBildenZuRegelUnfailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.*/
   IR_PufferAuslesen();
   /* =====> KPMenge ist jetzt komplett; */
   KPB_KonklusionenAbarbeiten(zuletztEingefuegteRegel);
}


static void AlsRegelAufnehmen_FAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus)
/* Nimmt links=rechts als Regel auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R !
   Failing-Version ==> keine Beruehrung der Gleichungsmenge.
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;

   if (Kreuzen){
     links = rhs;
     rechts = lhs;
   }
   else{
     links = lhs;
     rechts = rhs;
   }
   RE_IncAnzRegelnGesamt();
   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
   NeueRegelDrucken(zuletztEingefuegteRegel);
   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts,
                 TP_PclID(zuletztEingefuegteRegel),zuletztEingefuegteRegel->Nummer);
   IR_NeueRegelIst(zuletztEingefuegteRegel);
   IR_InterreduktionStarten();  /* RE-Puffer initialisieren */
   IR_RMLinksInterredMitRegel();
   /* Suche alle Regeln, deren linke Seite von l gematcht werden, fuehre diesen einen Reduktionsschritt aus und
      entferne sie aus der Regelmenge; (wohin? ---> Puffer).
      Vorgezogen, um Ausschluss ueberfluessig zu machen */
   /* =====> NF-Bildung ist wieder konsistent.*/
   RE_ZuletztErzeugteRegelEinfuegen();
   IR_RMRechtsInterredMitRegel();
   /* Suche alle rechts mit l->r reduzierbaren Regeln, fuehre den Reduktionsschritt aus und
      rufe die Normalformbildung dazu auf */
   /* =====> RM ist wieder interreduziert */
   /* =====> System ist interreduziert, entfernte Regeln in KPMenge, ggf. NF-Bildung beenden */
   IR_InterreduktionBeendet();
   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerRegel(zuletztEingefuegteRegel);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();      /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteRegel, NULL, NULL); /* So tun als ob! */
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
         /* ====> Ziele sind bearbeitet */
   }
   KPIR_KPMengeInterreduzieren(); /* KP-Mengen-IR vor der Aufnahme der aktuellen KPs */
   CG_CGMengeInterreduzieren(FALSE, "");
   U1_KPsBildenZuRegelFailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.*/
   IR_PufferAuslesen();
   /* =====> KPMenge ist jetzt komplett; */
   KPB_KonklusionenAbarbeiten(zuletztEingefuegteRegel);
}

/* -------------------------------------------------------------------------------*/
/* --------------- Vervollstaendigung (IV) ---------------------------------------*/
/* -------------------------------------------------------------------------------*/
/* Hauptschleife(n) der Vervollstaendigung.

   Die Routine Vervollstaendigen enthaelt die Hauptschleife des Beweisers.

   Die verschiedenen parametrisierten Auspraegungen sind durch Funktionsvariablen realisiert.
   Diese Variablen werden im failing-Betrieb entsprechend restriktiv besetzt.

   Auffaecherung der Parametrisierung:
   +-----------+------------+------+------+-----+-----+-----+
   |  PrCo     |   FunF     | NF-R | NF-E | Sub | psR | psE |
   +-----------+------------+------+------+-----+-----+-----+
   |Proof      |  Unfailing |  _   |  _   | _   |  _  |  _  |
   |           |            |      |      |     |     |     |
   |Completion |  Failing   |  R   |  E   | S   |  P  |  E  |
   +-----------+------------+------+------+-----+-----+-----+
   Es gibt also 2^7 = 128 verschiedene Moeglichkeiten.

   Im Proofmodus ergibt sich ein geringfuegig veraendertes Verhalten gegenueber dem
   Vervollstaendigungsmodus.

   Bei der Verwendung multipler Normalformen (MNF) nach der Auswahl von KPs verwenden wir
   die Routine MNF_Unfailing(); der failing-Betrieb wird bei MNF nicht unterstuetzt.

*/

static BOOLEAN Proofmodus;

static unsigned short PermGrad_E, PermGrad_R;
static BOOLEAN VervollstaendigungAbgebrochen, Normalform_R=FALSE, Normalform_E=FALSE, Subsumption=FALSE;



#define MeldungBerechnungBeendet()                                                              \
  if (!COMP_POWER)                                                                              \
    fprintf(IO_Strom, "\n%s.\n", HK_fair() ? "system completed" : "no more critical pairs")


/* Normalisierung */
/******************/

typedef void (* NormalisierungsFkt) (TermT, TermT, PCL_ObjektNummerT*);

NormalisierungsFkt Normalisiere;

#define KPRed(NFBWomit)                                                                           \
   PCL_Reduktionsbeginn(Gleichung, pclID, linkeSeite, rhs); /* letztes Arg. ist GEGENRICHTUNG! */ \
   NF_## NFBWomit(lhs, KPsSel);                                                                   \
   PCL_Reduktionsbeginn(Gleichung, pclID, rechteSeite, lhs);                                      \
   NF_## NFBWomit(rhs, KPsSel);                                                                   \

static void KPNormalisierung_RE(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{KPRed(NFBildung)}

static void KPNormalisierung_R_(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{KPRed(NFBildungZuRegeln)}

static void KPNormalisierung__E(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{KPRed(NFBildungZuGleichungen)}

static void KPNormalisierung___(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{;}

/* Permsubsummieren */
/********************/

typedef BOOLEAN (* PermsubFkt) (unsigned int, TermT, TermT);

PermsubFkt RegelPermsubsummiert, GleichungPermsubsummiert;

static BOOLEAN Permsubsummiert_P (unsigned int Grad, TermT lhs, TermT rhs)
{
  BOOLEAN Resultat = TO_ACnGleich(Grad,lhs,rhs);
  HK_fair = HK_fair && !Resultat;
  return Resultat;
}

static BOOLEAN Permsubsummiert__ (unsigned int Grad MAYBE_UNUSEDPARAM, TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM)
{
  return FALSE;
}

/* Subsummieren */
/****************/

typedef BOOLEAN (* SubFkt) (TermT, TermT);

SubFkt Subsummiert;

static BOOLEAN Subsummiert_S(TermT lhs, TermT rhs)
{
  return SS_GleichungWirdSubsummiertMitSGPuffernGM(lhs, rhs);
}

static BOOLEAN Subsummiert__(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM)
{
  return FALSE;
}

/* Regel/Gleichung aufnehmen */
/*****************************/

RegelAufnFkt HK_RegelAufnahme;
GleichungAufnFkt HK_GleichungsAufnahme;

static BOOLEAN ActionMitGleichung_Error(BOOLEAN proofmodus MAYBE_UNUSEDPARAM)
{
  IO_DruckeFlex("Completion aborted since critical pair can't be oriented:\n %t <<##>> %t\n",
                SUE_selCPlhs, SUE_selCPrhs);
  VervollstaendigungAbgebrochen = TRUE;
  SUE_selectedCPsubsumed();
  return FALSE;
}

static BOOLEAN ActionMitGleichung_Action(BOOLEAN proofmodus)
{
  SUE_selectedCPRE();
  AlsGleichungAufnehmen(SUE_selCPlhs, SUE_selCPrhs, proofmodus);
  return TRUE;
}


static VergleichsT VariantentestUndVergleich(TermT links, TermT rechts)
{
  BOOLEAN Wegwerfen = (PA_GolemAn && (Z_IstSkolemvarianteZuUngleichung(links,rechts) || Z_IstGrundungleichung(links,rechts)));
  VergleichsT Resultat = Wegwerfen ? Gleich : ORD_VergleicheTerme(links,rechts);
  HK_fair = HK_fair && (!Wegwerfen || TO_TermeGleich(links,rechts));
  return Resultat;
}


static unsigned int SelectZaehler;

static BOOLEAN TestStrategieWechsel (StrategieT *Strategie)
{
  return Strategie && ++SelectZaehler==Strategie->Schrittzahl;
}



/* Vervollstaendigungs-Hauptschleife */
/*************************************/

static HK_ErgebnisT Vervollstaendigen(StrategieT *Strategie)
{
  while ((!Proofmodus || !Z_ZieleBewiesen()) && SUE_Select()){
    VergleichsT Result;
    (*Normalisiere) (SUE_selectedCPlhs(), SUE_selectedCPrhs(), SUE_selectedCPPclID());

    switch ((Result = VariantentestUndVergleich(SUE_selectedCPlhs(), SUE_selectedCPrhs()))){
    case Gleich:
      SUE_selectedCPjoined();
      break;
    case Unvergleichbar:
      GebildeteGleichungDrucken (); 
            /* Meldung vorgezogen wg. Interreduktion */
      if ((*GleichungPermsubsummiert)(PermGrad_E, SUE_selectedCPlhs(), SUE_selectedCPrhs()))
        SUE_selectedCPpermsubsumed();           /*Gleichung wurde permsubsummiert => wegwerfen. */
      else if ((*Subsummiert) (SUE_selectedCPlhs(), SUE_selectedCPrhs()))
        SUE_selectedCPsubsumed();  /*Gleichung wurde subsummiert => wegwerfen. */
      else{
        (*HK_GleichungsAufnahme)(Proofmodus);
      }
      break;
    case Kleiner :
    case Groesser:
      /* kp als neue Regel rechts->links oder links->rechts aufnehmen */
      if ((*RegelPermsubsummiert)(PermGrad_R, SUE_selectedCPlhs(), SUE_selectedCPrhs()))
        SUE_selectedCPpermsubsumed();  /*Gleichung wurde permsubsummiert => wegwerfen. */
      else {
        SUE_selectedCPRE();
        (*HK_RegelAufnahme)(SUE_selectedCPlhs(), SUE_selectedCPrhs(), Result==Kleiner, Proofmodus);
      }
    default:
      break;
    }
    if (VervollstaendigungAbgebrochen)   return HK_VervollAbgebrochen;
    if (SI_InterruptflagGesetzt())       ST_IRMenue("main loop");
    if (TestStrategieWechsel(Strategie)) return HK_StrategieWechseln;
    KO_Konsultieren();
  }

  if (Proofmodus){
    if (Z_ZieleBewiesen()) {
      return HK_ZielBewiesen;
    }
    else {
      /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */
      /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */
      IR_FinalerRMrechtsDurchlauf();
      MeldungBerechnungBeendet();
    }
  }
  else
    MeldungBerechnungBeendet();
  return HK_Vervollstaendigt;
}


static HK_ErgebnisT MNF_Unfailing (StrategieT *Strategie)
{
  while ((!Proofmodus || !Z_ZieleBewiesen()) && SUE_Select()){
    SC_REPairsT *REPairs=NULL;
    MNF_NFSetT  *NFLeft;
    MNF_NFSetT  *NFRight;
    TermT       joinedTerm;
    TermT       lhs = SUE_selectedCPlhs();
    TermT       rhs = SUE_selectedCPrhs();

    if (MNF_KPSBearbeite(lhs, rhs, &joinedTerm, &NFLeft, &NFRight))     /* MNF-joined */
      SUE_selectedCPMNFjoined(joinedTerm);
    else {
      SC_REPairsT *runny;

      /*In SC_Classify wird subsummmiert und somit kann es passieren,
        dass der Rueckgabeparameter REPairs NULL ist.
        SC_Classify ruft SUE_selectedCPMNF* Funktionen auf
        */
      SC_Classify(NFLeft, NFRight, &REPairs);
      MNF_DeleteNFSet (&NFLeft, &NFRight);
      runny = REPairs;
      while (runny){
        switch (runny->vergleich){
        case Kleiner :
          /* kp als neue Regel rechts->links aufnehmen */
          AlsRegelAufnehmen_UNFAILING(runny->left, runny->right, TRUE, Proofmodus);
          break;
        case Groesser :
          /* kp als neue Regel links->rechts aufnehmen */
          AlsRegelAufnehmen_UNFAILING(runny->left, runny->right, FALSE, Proofmodus);
          break;
        case Unvergleichbar:
          /* vorgezogen wg. Interreduktion: */
          GebildeteGleichungDrucken ();
          /* Provisorium: SG-Info auch in SelectClassify berechenbar, aber etwas kompliziert */
          SS_OnlyInitSGPuffer (runny->left, runny->right);
          AlsGleichungAufnehmen(runny->left, runny->right, Proofmodus);
          break;
        default:
          our_error ("Fehler: SelectClassify liefert Termpaar mit gleich/undef zurueck!");
        }
        runny = runny->next;
      }
      SC_DeletePairs(&REPairs);
    }
    if (SI_InterruptflagGesetzt())
      ST_IRMenue("main loop");
    if (TestStrategieWechsel(Strategie))
      return HK_StrategieWechseln;
  }

  if (Proofmodus) {
    if (Z_ZieleBewiesen()) {
      return HK_ZielBewiesen;
    }
    else {
      /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */
      /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */
      IR_FinalerRMrechtsDurchlauf();
      MeldungBerechnungBeendet();
    }
  }
  else
    MeldungBerechnungBeendet();
  return HK_Vervollstaendigt;
}


static HK_ErgebnisT MNF_Failing (StrategieT *Strategie MAYBE_UNUSEDPARAM)
{
   our_error ("using MNF for CP-Selection is only implemented for Unfailing Completion");
   return HK_Beendet;
}

/* (IV)  */
/*-------*/
/* Verteiler: Je nach Variante (failing/unfailing) und Modus (Normal/MultipleNF/KPSharing) wird der
   entsprechende Alg. aufgerufen. Dies geschieht auf dieser hohen Ebene, um Abfragen innerhalb der
   Algorithmen einzusparen.*/

static BOOLEAN MNF_VV = FALSE;

HK_ErgebnisT HK_Vervollstaendigung (StrategieT *Strategie)
/* Vervollstaendigung, Ziele werden erst am Ende behandelt.
   Rueckgabe ist FALSE, wenn bei failing-Variante Vervollstaendigung abgebrochen wurde. */
{
  HK_ErgebnisT Ergebnis;

  VervollstaendigungAbgebrochen = FALSE;
  HK_fair = TRUE;

  if (WM_ErsterAnlauf()) {
     if ((Proofmodus = (PA_ExtModus == proof)))
        IO_DruckeUeberschrift("COMPLETION - PROOF");
     else
        IO_DruckeUeberschrift("COMPLETION");
  }

  if (MNF_VV) {
    if (PA_FailingCompletion) Ergebnis = MNF_Failing(Strategie);
    else                      Ergebnis = MNF_Unfailing(Strategie);
  }
  else{
    if (PA_FailingCompletion){
      if (Normalform_R)
        Normalisiere = KPNormalisierung_R_;
      else
        Normalisiere = KPNormalisierung___;
      RegelPermsubsummiert = GleichungPermsubsummiert = Permsubsummiert__;
      Subsummiert = Subsummiert__;
      HK_RegelAufnahme = AlsRegelAufnehmen_FAILING;
      HK_GleichungsAufnahme = ActionMitGleichung_Error;
    }
    else{                         /* unfailing */
      if (Normalform_R && Normalform_E)
        Normalisiere = KPNormalisierung_RE;
      else if (Normalform_R)
        Normalisiere = KPNormalisierung_R_;
      else if (Normalform_E)
        Normalisiere = KPNormalisierung__E;
      else
        Normalisiere = KPNormalisierung___;
      if (PermGrad_R == 0)
        RegelPermsubsummiert = Permsubsummiert__;
      else
        RegelPermsubsummiert = Permsubsummiert_P;
      if (PermGrad_E == 0)
        GleichungPermsubsummiert = Permsubsummiert__;
      else
        GleichungPermsubsummiert = Permsubsummiert_P;
      if (Subsumption)
        Subsummiert = Subsummiert_S;
      else
        Subsummiert = Subsummiert__;
      HK_RegelAufnahme = AlsRegelAufnehmen_UNFAILING;
      HK_GleichungsAufnahme = ActionMitGleichung_Action;
    }

    Ergebnis = Vervollstaendigen(Strategie);        /* Hauptschleife im Nicht-MNF-Falle */
  }

  if (Proofmodus){
    if (!Z_ZieleBewiesen())
      /* falls nicht nach jeder Einfuegung die Ziele behandelt wurden, muessen sie am */
      Z_FinalerZielDurchlauf(); /* Ende der Vervollstaendigung nochmals durchlaufen werden. */
  }
  else{
    /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */
    /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */

    IR_FinalerRMrechtsDurchlauf();
  }

  return Ergebnis;
}



/* -------------------------------------------------------------------------------*/
/* --------------- Intialisierung ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Dieses Makro soll die Auswertung der Parameter uebernehmen. Die auszufuehrenden
   Operationen werden als Bloecke uebergeben.*/

#define Auswertung(R_GRAD, E_GRAD, NF_R, NF_E, SUB, MNF_ACTION)                                 \
{                                                                                               \
  R_GRAD=0;E_GRAD=0;NF_R=NF_E=SUB=FALSE;                                                        \
                                                                                                \
  if (!COMP_POWER)                                                                              \
    fprintf(IO_Strom, "\n* TREATMENT OF CPS AFTER SELECTION:  ");                               \
                                                                                                \
  if (IS_FindValuedOption(DEF, USR, 1, "permsubE") == 0){                                       \
    char *str = IS_ValueOfOption(DEF, USR, "permsubE");                                         \
    if (!(str && ((E_GRAD = atoi(str)) > 0))) E_GRAD = 0;                                       \
  }                                                                                             \
  if (IS_FindValuedOption(DEF, USR, 1, "permsubR") == 0){                                       \
    char *str = IS_ValueOfOption(DEF, USR, "permsubR");                                         \
    if (!(str && ((R_GRAD = atoi(str)) > 0))) R_GRAD = 0;                                       \
  }                                                                                             \
  if (IS_FindOption(DEF, USR, 1, "s") == 0) SUB  = TRUE;                                        \
  if (IS_FindValuedOption(DEF, USR, 1, "mnf") == 0){                                            \
    if (!COMP_POWER) {                                                                          \
      fprintf(IO_Strom, " %s", SUB?"S":"-");                                                    \
      fprintf(IO_Strom, " (E-Perm = %u; R-Perm = %u)", E_GRAD, R_GRAD);                         \
    }                                                                                           \
    MNF_ACTION;                                                                                 \
  }                                                                                             \
  else {                                                                                        \
    if (IS_FindOption(DEF, USR, 1, "r") == 0) NF_R = TRUE;                                      \
    if (IS_FindOption(DEF, USR, 1, "e") == 0) NF_E = TRUE;                                      \
    if (!COMP_POWER) {                                                                          \
      fprintf(IO_Strom, " %s%s", NF_R?"R":"-", NF_E?"E":"-");                                   \
      fprintf(IO_Strom, "%s", SUB?"S":"-");                                                     \
      fprintf(IO_Strom, "%s", (E_GRAD+R_GRAD>0)?"P":"-");                                       \
      if (E_GRAD+R_GRAD>0) fprintf(IO_Strom, " (E-Perm = %u; R-Perm = %u)",E_GRAD, R_GRAD);     \
      fprintf(IO_Strom,"\n");                                                                   \
     }                                                                                          \
  }                                                                                             \
}

void HK_ParamInfo (ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    int RGRADInfo, EGRADInfo;
    BOOLEAN Rinfo, Einfo, Sinfo;
    Auswertung(RGRADInfo,EGRADInfo, Rinfo, Einfo, Sinfo,
               {fprintf(IO_Strom, "\n");MNF_ParamInfo(what, DEF, USR);SC_ParamInfo(what,DEF,USR);});
  }
  else{
    IO_DruckeKurzUeberschrift("Treatment of CPs after Selection", '-');
    fprintf(IO_Strom, "1. [r]\n");
    fprintf(IO_Strom, "2. [e]\n");
    fprintf(IO_Strom, "3. [s]\n");
    fprintf(IO_Strom, "4. [permsubE=<n>]\n");
    fprintf(IO_Strom, "5. [permsubR=<n>]\n");
    fprintf(IO_Strom, "6. [mnf=[r][e]...]\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "Explanation:\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "1-3.  These letters cause Waldmeister to normalize newly generated\n");
    fprintf(IO_Strom, "      critical pairs with _r_ules and/or _e_quations, with or without\n");
    fprintf(IO_Strom, "      additional _s_ubsumption\n");
    fprintf(IO_Strom, "4.,5. These options tell Waldmeister to suppress AC-equations from the\n");
    fprintf(IO_Strom, "      degree <n> on.\n");
    fprintf(IO_Strom, "6.    (see below)\n\n");
    MNF_ParamInfo(what, DEF, USR);
    SC_ParamInfo(what, DEF, USR);
  }
}


void HK_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* Einrichtung der Funktionspointer fuer die Ordnungsaufrufe.*/
{
   Auswertung(PermGrad_R, PermGrad_E,Normalform_R,Normalform_E,Subsumption,
              {
                fprintf(IO_Strom, "\n");
                if ((MNF_InitKPSAposteriori(DEF, USR) &&
                     SC_InitAposteriori (DEF, USR, PermGrad_R, PermGrad_E, Subsumption)))
                  MNF_VV = TRUE;
              });
   SelectZaehler = 0;
}
