/*=================================================================================================================
  =                                                                                                               =
  =  Interreduktion.c                                                                                             =
  =                                                                                                               =
  =  Interreduktion der Regel- und Gleichungsmenge                                                                =
  =  Subsumption                                                                                                  =
  =                                                                                                               =
  =  24.06.94  Arnim.                                                                                             =
  =  14.07.94  Arnim. Subsumption, dh. Durchlauf + Abspeichern aller Subgleichungen.                              =
  =  15.07.94  Arnim. Subgleichungsmenge implementiert.                                                           =
  =                                                                                                               =
  =                                                                                                               =
  =  Zur Beachtung: Bei der Interreduktion der rechten Regelseiten kann evtl. die Abfrage auf Aufrufoption -kr    =
  =                 entfernt werden, falls dies als zu zeitaufwendig erkannt wurde.                               =
  =================================================================================================================*/
#include <stdlib.h>
#include <string.h>

#include "Ausgaben.h"
#include "DSBaumKnoten.h"
#include "DSBaumOperationen.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "KPBildung.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "Puffer.h"
#include "pcl.h"
#include "RUndEVerwaltung.h"
#include "Subgleichungen.h" 
#include "Subsumption.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "SUEM.h"


/*=================================================================================================================
  =  Globale Variablen                                                                                            =
  =================================================================================================================*/

GleichungsT IR_NeueGleichung;
RegelT IR_NeueRegel;

/*=================================================================================================================
  =                                                                                                               =
  =  Ein Wort zu den waehrend der Interreduktion ausgeloesten Meldungen:                                          =
  =                                                                                                               =
  =  - Wird eine Regel links oder eine Gleichung reduziert, so wird dies sofort (nur) unter                       =
  =    Angabe der Nummer gemeldet. Wird eine Regel rechts reduziert, so wird zusaetzlich (falls Ausgabelevel      =
  =    entsprechend) die auf NF gebrachte rechte Seite mit ausgegeben (dies ist ja bei den linken Seiten der      =
  =    Regeln und den Gleichungen nicht sofort moeglich, da NF-Bildung noch nicht konsistent).                    =
  =  - Wird nun der Puffer ausgelesen, in dem die reduzierten Regeln oder Gleichungen zwischengespeichert         =
  =    sind, so wird die Entfernung der entsprechenden Regel/Gleichung aus der Regel/Gleichungsmenge              =
  =    ohne nochmalige Angabe der Ursache gemeldet. Die Terme werden dann ausgegeben, wenn die Bildung            =
  =    eines kritischen Paares aus der ehemaligen Regel/Gleichung gemeldet wird.                                  =
  =  - Wird eine Gleichung subsummiert, so wird sie ohne Federlesen aus der GM entfernt und incl. Termen ver-     =
  =    nichtet (natuerlich nach einer entsprechenden Meldung).                                                    =
  =================================================================================================================*/


/* -------------------------------------------------------------------------------*/
/* --------------- Statistik -----------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

static unsigned long        AnzGleichungenInterRed;            /* Gleichungen, die der Interreduktion zum Opfer fielen. */
static unsigned long        AnzRegRechtsRed;                   /* Reduktionen an rechten Seiten von Regeln.*/
static unsigned long        AnzRegWgLinksRedWeg;               /* Regeln, die wegen reduzierter linker Seite rausgeworfen wurden.*/
static unsigned long        AnzErfSubsumptionenAlt;            /* Alte Gleichungen, die von neuen subsummiert wurden. */

#define IncAnzErfSubsumptionenAlt()   (AnzGleichungenInterRed++)
#define IncAnzGleichungenInterRed()   (AnzErfSubsumptionenAlt++)  
#define IncAnzRegWgLinksRedWeg()      (AnzRegWgLinksRedWeg++)  
#define IncAnzRegRechtsRed()          (AnzRegRechtsRed++) 

#if IO_AUSGABEN

/* ------------- Regeln ----------------------------------*/

#define DruckeMeldungRegelRechtsReduziert(/* long */ Nummer, /* TermT */ links, /* TermT */ rechts)  \
{                                                                                                       \
    if (IO_level2)                                                                                      \
       fprintf(IO_Strom, "simplify rhs of rule:%5ld\n", Nummer);                                        \
    else if (IO_level3)                                                                                 \
       fprintf(IO_Strom, "simplify rhs of rule:%5ld.\n", Nummer);                                       \
    else if (IO_level4){                                                                                \
       fprintf(IO_Strom, "simplify rhs of rule:%5ld:\n       ", Nummer);                                \
       IO_DruckeFaktum(links, TRUE, rechts);                                                            \
    }                                                                                                   \
 }

#define DruckeMeldungRegelLinksReduziert(/* long */ Nummer)                                          \
{                                                                                                       \
    if (IO_level2)                                                                                      \
       fprintf(IO_Strom, "remove rule:         %5ld\n", Nummer);  /* Eigentlich "Reduziere Regel" */     \
    else if ((IO_level3) || (IO_level4))                                                                \
       fprintf(IO_Strom, "simplify lhs of rule %ld, removed from R.\n", Nummer);                        \
 } 

#define DruckeMeldungRegelGeloescht(/* long */ Nummer)       \
{                                                               \
    if ((IO_level4) || (IO_level3))                             \
       fprintf(IO_Strom, "rule %ld removed from R.\n", Nummer); \
 } 

/* ------------- Gleichungen -----------------------------*/


#define DruckeMeldungGleichungReduziert(/* long */ Nummer)           \
{                                                                       \
    if (IO_level2)                                                      \
       fprintf(IO_Strom, "remove equation:     %5ld\n", -Nummer);        \
    else if ((IO_level4) || (IO_level3))                                \
       fprintf(IO_Strom, "equation no. %ld removed.\n", Nummer);        \
} 
 
#define DruckeMeldungGleichungGeloescht(/* long */ Nummer)           \
{                                                                       \
    if ((IO_level4) || (IO_level3))                                     \
       fprintf(IO_Strom, "equation no. %ld removed.\n", Nummer);        \
 } 


#define DruckeMeldungNeueGleichungSubsummiertAlte(/* long */ Nummer, /* TermT */ links, /* TermT */ rechts)  \
{                                                                                                               \
    if (IO_level2)                                                                                              \
       fprintf(IO_Strom, "remove equation:     %5ld\n", -Nummer);                                                \
    else if (IO_level3)                                                                                         \
       fprintf(IO_Strom, "... the new equation subsumes equation no. %ld.\n", Nummer);                          \
    else if (IO_level4){                                                                                        \
       fprintf(IO_Strom, "... the new equation subsumes equation no. %ld:\n       ", Nummer);                   \
       IO_DruckeFaktum(links, FALSE, rechts);                                                                   \
    }                                                                                                           \
 } 


#else
/*------------- Regeln ----------------------------------*/

#define DruckeMeldungRegelRechtsReduziert(/*long*/ Nummer, /*TermT*/ links, /*TermT*/ rechts) /* nix tun */
#define DruckeMeldungRegelLinksReduziert(/*long*/ Nummer) /* nix tun */
#define DruckeMeldungRegelGeloescht(/*long*/ Nummer) /* nix tun */

/*------------- Gleichungen -----------------------------*/

#define DruckeMeldungGleichungReduziert(/*long*/ Nummer) /* nix tun */
#define DruckeMeldungGleichungGeloescht(/*long*/ Nummer) /* nix tun */
#define DruckeMeldungNeueGleichungSubsummiertAlte(/*long*/ Nummer, /*TermT*/ links, /*TermT*/ rechts) /* nix tun */

#endif /* IO_AUSGABEN */

void IR_Statistik(InfoStringT str)
{
   if (! (IS_IsIn("all", str) || IS_IsIn("IR", str)))
      return;
   if (IS_IsIn("standard", str)){
      IO_DruckeKurzUeberschrift("Interreduktion", '+');
      IO_DruckeStatistikZeile("Regeln fielen der Interreduktion zum Opfer", AnzRegWgLinksRedWeg);
      IO_DruckeStatistikZeile("Regeln rechts reduziert", AnzRegRechtsRed);
      IO_DruckeStatistikZeile("Gleichungen fielen der Interreduktion zum Opfer", AnzGleichungenInterRed);
      IO_DruckeStatistikZeile("Gleichungen fielen der Backward-Subsumption zum Opfer", AnzErfSubsumptionenAlt);
   }
   else if (IS_IsIn("latex-head", str))
      fprintf(IO_Strom, "IR: R links & R rechts & E & E: back-Sub & ");
}


/* -------------------------------------------------------------------------------*/
/* --------------- Subsumption auf der Gleichungsmenge ---------------------------*/
/* -------------------------------------------------------------------------------*/


static void FinaleKillprozSubsumption(RegelOderGleichungsT objekt)
/* Entfernen der Gegenrichtungen der gepufferten Gleichungen aus der Gleichungsmenge und
   physikalisches Loeschen der Objekte. Anwendung nach Subsumption.*/
{
  RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
  TP_LinksRechtsLoeschen(TP_Gegenrichtung(objekt));
  TP_LinksRechtsLoeschen(objekt);
  RE_GleichungLoeschen(TP_Gegenrichtung(objekt));
  RE_GleichungLoeschen(objekt);
}

void IR_GMSubsummierenMitGleichung(void)
/* Sucht alle von der angegebenen Gleichung subsummierten Gleichungen aus GM, entfernt und loescht sie.
   Seiteneffekt: Meldung, dass Gleichung xyz von neuer Gleichung subsummiert wird.
                 Entfernen der subsummierten Gleichung(en) aus der Gleichungsmenge.
   Ferner entsprechende Statistik anpassen.*/
{
   TermT l, r;
   SGM_ReferenzT ref;
   GleichungsT BehGleichung;

   if (RE_GleichungsmengeNichtLeer()){
      /* Subsumptionsdurchlauf auf Toplevel in GM */
      /* ---------------------------------------- */
      RE_forGleichungenRobust(BehGleichung,{ /* Im Gleichungsbaum gibt es evtl. > 1 Eintrag je Blatt.*/ 
         /* Getestet werden jeweils Richtung und Gegenrichtung jeder Gleichung (falls nicht ohnehin schon
            entfernt). 
            Monogleichungen: 
            - ist die behGleichung eine Monogleichung, so ist ihre Gegenrichtung nicht im Baum. Ok.
            - Da die Gleichung ohnehin nur in einer Richtung angewendet wird, ist es gleichgueltig, ob sie 
              eine Monogleichung ist oder nicht. 
            Gegenrichtung muss nicht getestet werden, da beim Durchlauf ueber alle Gleichungen auch die Gegenrichtung
            irgendwann angetroffen wird (Ausnahme: Monogleichungen, s. o.).
          */
         if (TP_IstWeiss(BehGleichung)&& /* Gegenrichtung der beh. Gleichung nicht bereits erfolgreich subsummiert */
             (MO_TermpaarSubsummiertZweites(TP_LinkeSeite(IR_NeueGleichung), TP_RechteSeite(IR_NeueGleichung), 
                                            TP_LinkeSeite(BehGleichung), TP_RechteSeite(BehGleichung)))){
            TP_BlauMalen(BehGleichung);
            TP_GelbMalen(TP_Gegenrichtung(BehGleichung));
            DruckeMeldungNeueGleichungSubsummiertAlte(TP_ElternNr(BehGleichung),
                                                         TP_LinkeSeite(BehGleichung), TP_RechteSeite(BehGleichung));
            DruckeMeldungGleichungGeloescht(TP_ElternNr(BehGleichung));
            IncAnzErfSubsumptionenAlt();
            RE_GleichungEntfernen(BehGleichung); /* aber nur diese eine Seite entfernen! */
            PU_REObjektEinfuegen(BehGleichung);
            /* In REPuffer stecken.*/
         }
      });
      PU_REPufferDurchlauf(FinaleKillprozSubsumption); /* die bis jetzt subsummierten Gleichungen physikalisch loeschen */

      /* Subsumptionsdurchlauf ueber alle Subgleichungen der GM */
      /* ------------------------------------------------------ */
      PU_REPufferNeu();
      SGM_StarteDurchlauf();
      while (SGM_NochSubgleichungen()){
         SGM_NaechsteSubgleichung(&l, &r, &ref); /* Im der SGM befinden sich nur Subgleichungen einer Richtung der Gleichungen */
         /* Ist die behandelnde Gleichung eine Monogleichung, so ist der Test in ihrer Gegenrichtung ueberfluessig! */
         if (MO_TermpaarSubsummiertZweites(TP_LinkeSeite(IR_NeueGleichung), TP_RechteSeite(IR_NeueGleichung), l ,r) ||
             (TP_IstKeineMonogleichung(IR_NeueGleichung) && 
              MO_TermpaarSubsummiertZweites(TP_LinkeSeite(IR_NeueGleichung), TP_RechteSeite(IR_NeueGleichung), r, l))){
            BehGleichung = (GleichungsT) ref;
            IncAnzErfSubsumptionenAlt();
            DruckeMeldungNeueGleichungSubsummiertAlte(TP_ElternNr(BehGleichung),
                                                         TP_LinkeSeite(BehGleichung), TP_RechteSeite(BehGleichung));
            DruckeMeldungGleichungGeloescht(TP_ElternNr(BehGleichung));
            SGM_NaechsteGleichung();
            RE_GleichungEntfernen(BehGleichung);
            PU_REObjektEinfuegen(BehGleichung);
         }
      }
   }
   PU_REPufferDurchlauf(FinaleKillprozSubsumption);
   PU_REPufferNeu();
}

/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Gleichungsmenge ----------------------------*/
/* -------------------------------------------------------------------------------*/

/* Am Ende der Interreduktion sind alle Gegenrichtungen
   aus der GM entfernt worden. Ferner wurde die Gegenrichtung mitsamt Termen bereits geloescht.
*/

static void FinaleEntfernungsprozGMIRGleichung(RegelOderGleichungsT objekt)
/* Alle evtl. noch in GM verbliebenen Gegenrichtungen der reduzierten Gleichungen werden entfernt, ferner wird die Reduktion
   auf den entsprechenden Termen ausgefuehrt. Das physische Loeschen der Gegenrichtung erfolgt hier noch nicht! */
{
   if (TP_RichtungAusgezeichnet(objekt)){
      DruckeMeldungGleichungReduziert(TP_ElternNr(objekt));
      IncAnzGleichungenInterRed();
      if (TP_IstBlau(objekt)){
         if (TP_IstKeineMonogleichung(objekt))
            RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
      }
      else if (TP_IstGelb(objekt) || TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses/gelbes Objekt im Puffer!");
   }
   else{
      if (TP_IstBlau(objekt)){
         RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
         DruckeMeldungGleichungReduziert(TP_ElternNr(TP_Gegenrichtung(objekt)));
         IncAnzGleichungenInterRed();
      }
      if (TP_IstGelb(objekt))
         our_error("\nGM-IR: nicht ausgezeichnetes gelbes Objekt im Puffer!");
      if (TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses Objekt im Puffer!");
   }
}

static void FinaleEntfernungsprozGMIRRegel(RegelOderGleichungsT objekt)
/* Beim Puffer-Durchlauf zum Entfernen der jeweiligen Gegenrichtung der bereits einseitig entfernten Gleichungen.*/
{
   if (TP_RichtungAusgezeichnet(objekt)){
      DruckeMeldungGleichungReduziert(TP_ElternNr(objekt));
      IncAnzGleichungenInterRed();
      if (TP_IstBlau(objekt)){
         if (TP_IstKeineMonogleichung(objekt))
            RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
      }
      else if (TP_IstGelb(objekt) || TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses/gelbes Objekt im Puffer!");
   }
   else{
      if (TP_IstBlau(objekt)){
         RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
         DruckeMeldungGleichungReduziert(TP_ElternNr(TP_Gegenrichtung(objekt)));
         IncAnzGleichungenInterRed();
      }
      if (TP_IstGelb(objekt))
         our_error("\nGM-IR: gelbes Objekt im Puffer!");
      if (TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses Objekt im Puffer!");
   }
}

static BOOLEAN GMIRGleichungTestProc(GleichungsT g)
/* Zur sukzessiven Anwendungen auf die Referenz-Gleichungen aller Blaetter im Gleichungsbaum. */
{
   return NF_GleichungAnwendbar(IR_NeueGleichung, TP_LinkeSeite(g), Gleichungen);
}

static BOOLEAN GMIRRegelTestProc(GleichungsT g)
/* Zur sukzessiven Anwendungen auf die Referenz-Gleichungen aller Blaetter im Gleichungsbaum. */
{
   return NF_RegelAnwendbar(IR_NeueRegel, TP_LinkeSeite(g), Gleichungen);
}


static void GMIRLoeschProc(GleichungsT g)
/* Nachdem die TestProc ein TRUE geliefert hat, werden die Gleichungen des Blattes sukzessive 
   an diese Prozedur (zwecks Entfernung) uebergeben.*/
{
   RE_GleichungEntfernen(g);
   TP_BlauMalen(g);
   TP_GelbMalen(TP_Gegenrichtung(g));
   PU_REObjektEinfuegen(g);
}


void IR_GMInterredMitGleichung(void)
/* Entfernen aller Gleichungen aus der Gleichungsmenge (-> Puffer), die von der angegebenen Gleichung
   reduziert werden. Da Gleichung noch nicht in der GM, kein Ausschluss.
   Meldungen: Gleichung xyz reduziert. (Entfernung aus GM wird beim Auslesen des Puffers gemeldet)
   Ferner entsprechende Reduktions-Statistiken anpassen.*/
{
   PU_REMerkerSetzen();
   RE_forGMReferenzen(GMIRGleichungTestProc, GMIRLoeschProc);
   PU_REPufferDurchlaufAbMerker(FinaleEntfernungsprozGMIRGleichung); /* Gegenseite ggf. aus E entfernen, Red. ausfuehren. */
}


void IR_GMInterredMitRegel(void)
/* Entfernen aller Gleichungen aus der Gleichungsmenge (-> Puffer), die von der angegebenen Regel
   links -> rechts reduziert werden.
   Meldungen: Gleichung xyz reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/
{
   PU_REMerkerSetzen();
   RE_forGMReferenzen(GMIRRegelTestProc, GMIRLoeschProc);
   PU_REPufferDurchlaufAbMerker(FinaleEntfernungsprozGMIRRegel); /* Gegenseite ggr. aus E entfernen, Red. ausfuehren. */
}


/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Regelmenge ---------------------------------*/
/* -------------------------------------------------------------------------------*/

unsigned long int IntervallRMrechts;

void IR_RMLinksInterredMitGleichung(void)
/* Entfernen aller Regeln aus der Regelmenge (-> Puffer), die von der angegebenen Gleichung
   links reduziert werden.
   Meldungen: Regel xyz links reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/
{
   RegelT BehRegel;
   
   RE_forRegelnRobust(BehRegel,{            /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/ 
      PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), linkeSeite, TP_RechteSeite(BehRegel));
      if (NF_GleichungAnwendbar(IR_NeueGleichung, TP_LinkeSeite(BehRegel), RegelnLinks)){
         DruckeMeldungRegelLinksReduziert(TP_ElternNr(BehRegel));
         IncAnzRegWgLinksRedWeg();
         RE_RegelEntfernen(BehRegel);
         /* Entfernen (nicht Loeschen) aus RM) */
         NF_GrossenExekutorAnwenden();
         PU_REObjektEinfuegen(BehRegel);
         /* In REPuffer stecken.*/
      }
   });
}


void IR_RMRechtsInterredMitGleichung(void)
/* Reduktion aller rechten Seiten von Regeln, die mit der angegebenen Gleichung reduziert werden koennen, und
   Abschlus der NF-Bildung.
   Meldungen: Regel xyz rechts reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/
{
   RegelT BehRegel;


   if (IntervallRMrechts == 1){
      /* Alle Regeln rechts durchlaufen und ggf. mit gleichung reduzieren und NF-Bildung beenden.*/
      RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
        PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
        if (NF_NFMitNeuerGleichung(IR_NeueGleichung, TP_RechteSeite(BehRegel), RegelnRechts)){
          DruckeMeldungRegelRechtsReduziert(TP_ElternNr(BehRegel), TP_LinkeSeite(BehRegel), TP_RechteSeite(BehRegel));
          IncAnzRegRechtsRed();
        }
      });
   }
   else if (IntervallRMrechts>1) {
      if ((RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt())%IntervallRMrechts == 0){
         RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
            PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
            if (NF_NFBildung(TP_RechteSeite(BehRegel), RegelnRechts)){
               DruckeMeldungRegelRechtsReduziert(TP_ElternNr(BehRegel), TP_LinkeSeite(BehRegel), TP_RechteSeite(BehRegel));
               IncAnzRegRechtsRed();
            }
         });
      }
   }
}


void IR_RMLinksInterredMitRegel(void)
/* Entfernen aller Regeln aus der Regelmenge (-> Puffer), die von der angegebenen Regel
   links reduziert werden.(links->rechts ist noch nicht in RM aufgenommen)
   Meldungen: Regel xyz links reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/
{
   RegelT BehRegel; /* diejenige, welche zu entfernen ist */
   TermT ll, rr;
   
   RE_forRegelnRobust(BehRegel,{            /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/ 
      ll = TP_LinkeSeite(BehRegel); rr = TP_RechteSeite(BehRegel);
      PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), linkeSeite, rr);
      if (NF_RegelAnwendbar(IR_NeueRegel, ll, RegelnLinks)){
         DruckeMeldungRegelLinksReduziert(TP_ElternNr(BehRegel));
         IncAnzRegWgLinksRedWeg();
         RE_RegelEntfernen(BehRegel);                                        /* Entfernen (nicht Loeschen) aus RM */
         NF_GrossenExekutorAnwenden();
         PU_REObjektEinfuegen(BehRegel);                                                    /* In REPuffer stecken.*/
      }
   });
}

void IR_RMRechtsInterredMitRegel(void)
/* Reduktion aller rechten Seiten von Regeln, die mit der angegebenen Regel reduziert werden koennen, und
   Abschlus der NF-Bildung. Die rechte Seite der Regel selbst muss nicht betrachtet werden (dort ist wegen der
   Ordnung keine Reduktion moeglich!!).
   Meldungen: Regel xyz rechts reduziert.
   Ferner entsprechende Reduktions-Statistiken anpassen.*/

{

   if (IntervallRMrechts == 1){
      /* Alle Regeln rechts durchlaufen und ggf. mit Regel reduzieren und NF-Bildung beenden.*/
     RegelT BehRegel;

     RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
       if(IR_NeueRegel != BehRegel) {  /*Regel mit sich selbst? Niemals...*/
         PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
         if(NF_NFMitNeuerRegel(IR_NeueRegel, TP_RechteSeite(BehRegel), RegelnRechts)){
            DruckeMeldungRegelRechtsReduziert(TP_ElternNr(BehRegel), TP_LinkeSeite(BehRegel), TP_RechteSeite(BehRegel));
            IncAnzRegRechtsRed();
         }
       }
     });
   }
   else if (IntervallRMrechts>1){
     if ((RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt())%IntervallRMrechts == 0){
       RegelT BehRegel;

       RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
         if(IR_NeueRegel != BehRegel){/*Regel mit sich selbst? Niemals...*/
           PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
           if(NF_NFBildung(TP_RechteSeite(BehRegel), RegelnRechts)){
              DruckeMeldungRegelRechtsReduziert(TP_ElternNr(BehRegel), TP_LinkeSeite(BehRegel), TP_RechteSeite(BehRegel));
              IncAnzRegRechtsRed();
           }
         }
       });
     }
   }
}

void IR_FinalerRMrechtsDurchlauf(void)
/* Falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, muessen sie am Ende der Vervollstaendigung 
   nochmals durchlaufen werden. Dies geschieht hier.
   Vor.: IR-Intervall RMrechts ist nicht == 1, RM ist nicht leer.
*/
{
   if (IntervallRMrechts > 1){
      RegelT BehRegel;
      
      RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
         PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
         if(NF_NFBildung(TP_RechteSeite(BehRegel), RegelnRechts)){
            DruckeMeldungRegelRechtsReduziert(TP_ElternNr(BehRegel), TP_LinkeSeite(BehRegel), TP_RechteSeite(BehRegel));
            IncAnzRegRechtsRed();
         }
      });
   }
}

/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion initialisieren ---------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Die folgenden beiden Melde-Prozeduren versorgen saemtliche Interreduktionsprozeduren 
   dieses Moduls mit der aktuellen Regel bzw. Gleichung. Damit entfaellt die Parameteruebergabe
   in jeder einzelnen Prozedur. (Ausnahme: Zielmengen-Behandlung)
*/

void IR_NeueGleichungIst(GleichungsT g)
{
   IR_NeueGleichung = g;
}

void IR_NeueRegelIst(RegelT r)
{
   IR_NeueRegel = r;
}

void IR_InterreduktionStarten(void)
/* Vor Beginn der Interreduktion den Datenpuffer initialisieren.
   Meldung: Im entsprechenden Ausgabelevel -> 'Interreduktion Anfang'
*/
{
#if IO_AUSGABEN
   if ((IO_level3) || (IO_level4))
      fprintf(IO_Strom, ">>interreduction<<\n\n");
#endif
    PU_REPufferNeu();
}

void IR_InterreduktionBeendet(void)
/* Meldung: Im entsprechenden Ausgabelevel -> 'Interreduktion Ende'
*/
{
#if IO_AUSGABEN
   if ((IO_level4) || (IO_level3))
      fprintf(IO_Strom, "\n>>interreduction finished<<\n\n");
#else
   ;
#endif
}

/* -------------------------------------------------------------------------------*/
/* --------------- Weiterbehandlung der entfernten Objekte -----------------------*/
/* -------------------------------------------------------------------------------*/

static void Anwendungsprozedur(RegelOderGleichungsT objekt)
{
   TermT ll, rr;
   PCL_ObjektNummerT KpID;
   
   ll = TP_LinkeSeite(objekt);
   rr = TP_RechteSeite(objekt);
   PCL_IDZuweisung(&KpID, *(TP_PclID(objekt)));
   
   if(RE_IstRegel(objekt)){
      DruckeMeldungRegelGeloescht(TP_ElternNr(objekt));
      RE_RegelLoeschen(objekt);        /* Loeschen der Regel als Objekt, -> Meldung an KPVerwaltung u. Schirm*/
   }
   else
      if (TP_RichtungAusgezeichnet(objekt)){
         if (TP_IstBlau(objekt) || TP_IstMonogleichung(objekt)){       /* Gegenrichtung ist nicht im Puffer */
            TP_LinksRechtsLoeschen(TP_Gegenrichtung(objekt));   /* sonst: Gegenrichtung wird noch/wurde schon */
            RE_GleichungLoeschen(TP_Gegenrichtung(objekt));            /* behandelt */
         }
         DruckeMeldungGleichungGeloescht(TP_ElternNr(objekt));
         RE_GleichungLoeschen(objekt);    /* Loeschen der Gleichung als Objekt, -> Meldung an KPVerwaltung und Schirm */
      }
      else{                     /* Richtung ist nicht ausgezeichnet */
         if (TP_IstBlau(objekt)){ /* ausgezeichnete Richtung nicht im Puffer! */
            RegelOderGleichungsT merker = TP_Gegenrichtung(objekt); /* ausgezeichnete Gegenrichtung merken */
            TP_LinksRechtsLoeschen(objekt);
            RE_GleichungLoeschen(objekt);
            DruckeMeldungGleichungGeloescht(TP_ElternNr(merker));
            ll = TP_LinkeSeite(merker);
            rr = TP_RechteSeite(merker);
            RE_GleichungLoeschen(merker);    /* Loeschen der Gleichung als Objekt, -> Meldung an KPVerwaltung und Schirm */
         }
         else{                  /* ausgezeichnete Richtung selbst ist auch im Puffer */
            TP_LinksRechtsLoeschen(objekt);
            RE_GleichungLoeschen(objekt);
            return;
         }
      }

   SUE_CPkilledRE(ll, rr, &KpID);

   KPB_IROpferBehandeln(ll, rr); /* diese Routine erledigt auch das letztendliche Aufnehmen in SUE */
}


void IR_PufferAuslesen(void)
/* Auslesen des waehrend der Interreduktion gefuellten Regel- und Gleichungs-Puffers.
   Behandeln der darin befindlichen Objekte gemaess Modus.
   Meldungen: Im entsprechenden Ausgabelevel -> 'Interreduktion Ende',
              beim Loeschen der entsprechenden Objekte automatisch durch die Loeschroutine,
              beim Einfuegen als KP von dieser Routine.
*/
{
   PU_REPufferDurchlauf(Anwendungsprozedur);
}

/* -------------------------------------------------------------------------------*/
/* --------------- Test auf Killer-Eigenschaft -----------------------------------*/
/* -------------------------------------------------------------------------------*/

BOOLEAN IR_TPRReduziertRE(TermT Links, TermT Rechts)
{
  RegelT Regel;
  RegelT RegelIndex;
  unsigned int bla;
  TP_GNTermpaarHolen(Regel);
  BO_TermpaarNormieren(Links,Rechts,&bla);
  Regel->rechteSeite = Rechts;
  Regel->linkeSeite = Links;
  Regel->SUEInfo = NULL;
  TP_GegenrichtungBesetzen(Regel,NULL);
  RE_forRegelnRobust(RegelIndex, {
    TermT TermIndex = TP_LinkeSeite(RegelIndex);
    do
      if (TO_TermIstNichtVar(TermIndex) && MO_RegelPasst(Regel,TermIndex)) {
        TP_GNTermpaarLoeschen(Regel);
        return TRUE;
      }
    while ((TermIndex = TO_Schwanz(TermIndex)));
  });
  RE_forGleichungenRobust(RegelIndex, {
    TermT TermIndex = TP_LinkeSeite(RegelIndex);                            /* reicht, da jede Seite indiziert ist */
    do
      if (TO_TermIstNichtVar(TermIndex) && MO_RegelPasst(Regel,TermIndex)) {
        TP_GNTermpaarLoeschen(Regel);
        return TRUE;
      }
    while ((TermIndex = TO_Schwanz(TermIndex)));
  });
  TP_GNTermpaarLoeschen(Regel);
  return FALSE;
}

BOOLEAN IR_TPEReduziertRE(TermT Links, TermT Rechts)
{
  RegelT RegelIndex;
  GleichungsT Gleichung,
              Antigleichung;
  unsigned int gna;
  TP_GNTermpaarHolen(Gleichung);
  TP_GNTermpaarHolen(Antigleichung);
  BO_TermpaarNormieren(Links,Rechts,&gna);
  Gleichung->rechteSeite = Rechts;
  Gleichung->linkeSeite = Links;
  Gleichung->RichtungAusgezeichnet = TRUE;
  Gleichung->SUEInfo = NULL;
  BO_TermpaarNormieren(
    Antigleichung->linkeSeite = TO_Termkopie(Rechts),
    Antigleichung->rechteSeite = TO_Termkopie(Links),
    &gna);
  Gleichung->IstMonogleichung = 
    Antigleichung->IstMonogleichung =
      TP_TermpaareGleich(Gleichung,Antigleichung);
  Antigleichung->RichtungAusgezeichnet = FALSE;
  Antigleichung->SUEInfo = Gleichung->SUEInfo;
  (Gleichung->Gegenrichtung = Antigleichung)->Gegenrichtung = Gleichung;

  RE_forRegelnRobust(RegelIndex, {
    TermT TermIndex = TP_LinkeSeite(RegelIndex);
    do
      if (TO_TermIstNichtVar(TermIndex) && MO_GleichungPasst(Gleichung,TermIndex)) {
        TP_GNTermpaarLoeschen(Gleichung);
        TP_GNTermpaarLoeschen(Antigleichung);
        return TRUE;
      }
    while ((TermIndex = TO_Schwanz(TermIndex)));
  });
  RE_forGleichungenRobust(RegelIndex, {
    TermT TermIndex = TP_LinkeSeite(RegelIndex);                            /* reicht, da jede Seite indiziert ist */
    do
      if (TO_TermIstNichtVar(TermIndex) && MO_GleichungPasst(Gleichung,TermIndex)) {
        TP_GNTermpaarLoeschen(Gleichung);
        TP_GNTermpaarLoeschen(Antigleichung);
        return TRUE;
      }
    while ((TermIndex = TO_Schwanz(TermIndex)));
  });
  TP_GNTermpaarLoeschen(Gleichung);
  TP_GNTermpaarLoeschen(Antigleichung);
  return FALSE;
}

/* -------------------------------------------------------------------------------*/
/* --------------- Initialisierung -----------------------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   long int intervall;
   
   if (what == ParamEvaluate){
      if (IS_ValueOfOption(DEF, USR, "period"))
         intervall = atoi(IS_ValueOfOption(DEF, USR, "period"));
      else
         intervall = 1;
      fprintf(IO_Strom, "\n* INTERREDUCTION:   Period for right hand sides of rules: %ld\n", intervall);
   }
   else{
      fprintf(IO_Strom, "Module: Interreduction\n");
      fprintf(IO_Strom, "**********************\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. period=<n>\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Interval between simplifying the right-hand-sides of rules.\n");
      fprintf(IO_Strom, "<n> is a natural number, where 0 means that right-hand-sides of rules will\n");
      fprintf(IO_Strom, "not be simplified.\n");
      fprintf(IO_Strom, "As usual, <n> concerns the number of rules & equations inserted into R and E.\n");
      fprintf(IO_Strom, "\n");      
   }
}

   

BOOLEAN IR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* IR-RM rechts */
{
   if (IS_ValueOfOption(DEF, USR, "period"))
      IntervallRMrechts = atoi(IS_ValueOfOption(DEF, USR, "period"));
   else
      IntervallRMrechts = 1;

#if !COMP_POWER
   fprintf(IO_Strom, "\n* INTERREDUCTION:   Period for right hand sides of rules: %ld\n", IntervallRMrechts);
#endif

   return TRUE;
}
