/*=================================================================================================================
  =                                                                                                               =
  =  KPBehandlung.c                                                                                               =
  =                                                                                                               =
  =  Behandlung von kritischen Paaren nach Entstehung (auch von IR-Opfern)                                        =
  =                                                                                                               =
  =  24.06.94  Arnim.                                                                                             =
  =     05.96  Makrofiziert, PermSub eingebaut. Arnim.                                                            =
  =  04.09.96  MNF eingebaut. Arnim+Andreas                                                                       =
  =                                                                                                               =
  =                                                                                                               =
  =================================================================================================================*/
#include <string.h>
#include "compiler-extra.h"

#include "general.h"
#include "Ausgaben.h"
#include "ChaseManhattan.h"
#include "Interreduktion.h"
#include "KPBildung.h"
#include "MNF_CPs.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "Subsumption.h"
#include "TermOperationen.h"
#include "Zielverwaltung.h"
#include "NFInnermost.h"
#include "NFOutermost.h"
#include "NFTopDown.h"


/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe KPBildung.h) -------------------------*/
/* ------------------------------------------------------------------------*/
void (* KPBehandlungsProcFATHER) (RegelOderGleichungsT, TermT, TermT, UTermT);
void (* KPBehandlungsProcMOTHER) (RegelOderGleichungsT, TermT, TermT, UTermT);
void (* KPBehandlungsProcIROpfer) (TermT, TermT);

/* -----------------------------------------------------------------------------------------------*/
/* --------------- KPs nach Entstehung behandeln und ggf. in KP-Menge aufnehmen ------------------*/
/* -----------------------------------------------------------------------------------------------*/

static unsigned short GRAD;

static BOOLEAN (* VerwendeteMNFProzedur) (TermT, TermT, TermT *);
/* entweder der MNFDummy oder MNF_KPGBearbeite */


static BOOLEAN MNFDummy(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM,
			TermT *joinedTerm MAYBE_UNUSEDPARAM) {
   return FALSE;
}

#define /*BOOLEAN*/MNFBearbeiten(lhs, rhs, joinedTerm) ((* VerwendeteMNFProzedur) (lhs, rhs, joinedTerm))

#define USE_MNF_CG 0
/* Dieser Hack bringt nichts :-( AJ */
#if USE_MNF_CG
extern BOOLEAN MNF_CGBearbeite (TermT links, TermT rechts, TermT *joinedTerm, TermT *neuL, TermT *neuR);

static BOOLEAN MNFCGBearbeiten (TermT lhs, TermT rhs, TermT *joinedTerm) 
{
  TermT   s,t;
  if (Z_UngleichungEntmantelt(&s,&t,SUE_recCPlhs,SUE_recCPrhs)) {
    TermT joinedTerm,
          neuL = NULL, neuR = NULL;
    s = TO_Termkopie(s), t = TO_Termkopie(t);
    if (MNF_CGBearbeite (s, t, &joinedTerm, &neuL, &neuR)) {
       UTermT rechts = TO_Termkopie (joinedTerm);
       Z_UngleichungEinmanteln(&joinedTerm,&rechts);
       IO_DruckeFlex("%t\n", joinedTerm);
       fprintf (stderr, "\a\a\a Terme wurden zusammengefuehrt!\a\n");
    }
    TO_TermeLoeschen(s,t);
    if (neuL)
      Z_UngleichungEinmanteln(&neuL,&neuR);
   }

  return FALSE;
  
  /*return MNF_KPGBearbeite (lhs, rhs, joinedTerm);*/
}
#endif

#define NFPERMSUB_KERN(NFPROZ)                                                                          \
{TermT joinedTerm;                                                                                      \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.   \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                        \
                                                                                                        \
   if (MNFBearbeiten(lhs, rhs, &joinedTerm)) SUE_recentCPMNFjoined(joinedTerm);                         \
   else {                                                                                               \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), linkeSeite, SUE_recentCPrhs());                     \
      NFPROZ(SUE_recentCPlhs(), KPsGen);                                                                \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), rechteSeite, SUE_recentCPlhs());                    \
      NFPROZ(SUE_recentCPrhs(), KPsGen);                                                                \
                                                                                                        \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                         \
         SUE_recentCPjoined();                                                                          \
      else if(SS_TermpaarSubsummiertVonPermsub(SUE_recentCPlhs(), SUE_recentCPrhs(), GRAD))             \
         SUE_recentCPpermsubsumed();                                                                    \
      else if(SS_TermpaarSubsummiertVonGM(SUE_recentCPlhs(), SUE_recentCPrhs()))                        \
         SUE_recentCPsubsumed();                                                                        \
      else                                                                                              \
         SUE_recentCPinsert();                                                                          \
   }                                                                                                    \
}

#define PERMSUB_KERN()                                                                          \
{TermT joinedTerm;                                                                              \
   if (MNFBearbeiten(lhs, rhs, &joinedTerm)) SUE_recentCPMNFjoined(joinedTerm);                 \
   else {                                                                                       \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                 \
         SUE_recentCPjoined();                                                                  \
      else if(SS_TermpaarSubsummiertVonPermsub(SUE_recentCPlhs(), SUE_recentCPrhs(), GRAD))     \
         SUE_recentCPpermsubsumed();                                                            \
      else if(SS_TermpaarSubsummiertVonGM(SUE_recentCPlhs(), SUE_recentCPrhs()))                \
         SUE_recentCPsubsumed();                                                                \
      else                                                                                      \
         SUE_recentCPinsert();                                                                  \
   }                                                                                            \
}

#define NFPERM_KERN(NFPROZ)                                                                                     \
{TermT joinedTerm;                                                                                                               \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.           \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                                \
                                                                                                                \
   if (MNFBearbeiten(lhs, rhs,&joinedTerm)) SUE_recentCPMNFjoined(joinedTerm); 				      \
   else {												      \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), linkeSeite, SUE_recentCPrhs());                                \
      NFPROZ(SUE_recentCPlhs(), KPsGen);                                                                              \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), rechteSeite, SUE_recentCPlhs());                               \
      NFPROZ(SUE_recentCPrhs(), KPsGen);                                                                              \
                                                                                                                   \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else if (SS_TermpaarSubsummiertVonPermsub(SUE_recentCPlhs(), SUE_recentCPrhs(), GRAD))                       \
         SUE_recentCPpermsubsumed();                                                                                   \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }													      \
}

#define NFSUB_KERN(NFPROZ)                                                                                      \
{TermT joinedTerm;                                                                                                               \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.           \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                                \
                                                                                                                \
   if (MNFBearbeiten(lhs, rhs,&joinedTerm)) SUE_recentCPMNFjoined(joinedTerm); 				      \
   else {												      \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), linkeSeite, SUE_recentCPrhs());                                \
      NFPROZ(SUE_recentCPlhs(), KPsGen);                                                                              \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), rechteSeite, SUE_recentCPlhs());                               \
      NFPROZ(SUE_recentCPrhs(), KPsGen);                                                                              \
                                                                                                                   \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else if(SS_TermpaarSubsummiertVonGM(SUE_recentCPlhs(), SUE_recentCPrhs()))                                   \
         SUE_recentCPsubsumed();                                                                                   \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }													      \
}

#define SUB_KERN()                                                                                                \
{TermT joinedTerm;                                                                                                               \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.           \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                                \
                                                                                                                \
   if (MNFBearbeiten(lhs, rhs, &joinedTerm)) SUE_recentCPMNFjoined(joinedTerm); 					\
   else {												      \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else if(SS_TermpaarSubsummiertVonGM(SUE_recentCPlhs(), SUE_recentCPrhs()))                                   \
         SUE_recentCPsubsumed();                                                                                   \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }                                                                                                               \
}

#define PERM_KERN()                                                                                               \
{TermT joinedTerm;                                                                                                               \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.           \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                                \
                                                                                                                \
   if (MNFBearbeiten(lhs, rhs,&joinedTerm)) SUE_recentCPMNFjoined(joinedTerm); 					\
   else {												      \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else if(SS_TermpaarSubsummiertVonPermsub(SUE_recentCPlhs(), SUE_recentCPrhs(), GRAD))                        \
         SUE_recentCPpermsubsumed();                                                                                   \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }													      \
}

#define NF_KERN(NFPROZ)                                                                                         \
{TermT joinedTerm;                                                                                                               \
  /* Nach unserer Erfahrung kommt es nur selten vor, dass KPs schon bei ihrer Bildung identisch sind.           \
     Deshalb erfolgt an dieser Stelle auch kein Termvergleich.*/                                                \
                                                                                                                \
   if (MNFBearbeiten(lhs, rhs, &joinedTerm)) SUE_recentCPMNFjoined(joinedTerm);                                 \
   else {												      \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), linkeSeite, SUE_recentCPrhs());                                \
      NFPROZ(SUE_recentCPlhs(), KPsGen);                                                                              \
      PCL_Reduktionsbeginn(KP, SUE_recentCPPclID(), rechteSeite, SUE_recentCPlhs());                               \
      NFPROZ(SUE_recentCPrhs(), KPsGen);                                                                              \
                                                                                                                   \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }													      \
}


#define NOTHING_KERN()                                                                                            \
{TermT joinedTerm;                                                                                                               \
   if (MNFBearbeiten(lhs, rhs, &joinedTerm)) SUE_recentCPMNFjoined(joinedTerm); 			      \
   else {												      \
      if (TO_TermeGleich(SUE_recentCPlhs(), SUE_recentCPrhs()))                                                    \
         SUE_recentCPjoined();                                                                                     \
      else                                                                                                         \
         SUE_recentCPinsert();                                                                                     \
   }													      \
}


#define NFPERMSUB_BODY(NFPROZ, otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)  \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   NFPERMSUB_KERN(NFPROZ)                                                                                       \
}

#define NFSUB_BODY(NFPROZ, otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)      \
{                                                                                                               \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   NFSUB_KERN(NFPROZ)                                                                                           \
                                                                                                                \
}


#define NFPERM_BODY(NFPROZ, otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)     \
{                                                                                                               \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   NFPERM_KERN(NFPROZ)                                                                                          \
                                                                                                                \
}


#define PERMSUB_BODY(otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)            \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   PERMSUB_KERN()                                                                                                 \
                                                                                                                \
}


#define SUB_BODY(otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)                \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   SUB_KERN()                                                                                                     \
                                                                                                                \
}


#define PERM_BODY(otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)               \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   PERM_KERN()                                                                                                    \
                                                                                                                \
}


#define NF_BODY(NFPROZ, otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)         \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   NF_KERN(NFPROZ)                                                                                              \
                                                                                                                \
}


#define NOTHING_BODY(otherParent, lhs, rhs, actualPosition, /* {WithFather, WithMother} */ WithWhat)            \
{                                                                                                               \
                                                                                                                \
   SUE_CPgenerated ## WithWhat(otherParent, lhs, rhs, actualPosition);                                          \
                                                                                                                \
   NOTHING_KERN()                                                                                                 \
                                                                                                                \
}


static void KPB_GebildetesKPBehandelnSVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
SUB_BODY(father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnSMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
SUB_BODY(mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnS(TermT lhs, TermT rhs)
SUB_KERN()


static void KPB_GebildetesKPBehandelnPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
PERM_BODY(father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
PERM_BODY(mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnP(TermT lhs, TermT rhs)
PERM_KERN()


static void KPB_GebildetesKPBehandelnRVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildungZuRegeln, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildungZuRegeln, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnR(TermT lhs, TermT rhs)
NF_KERN(NF_NFBildungZuRegeln)


static void KPB_GebildetesKPBehandelnEVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildungZuGleichungen, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnEMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildungZuGleichungen, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnE(TermT lhs, TermT rhs)
NF_KERN(NF_NFBildungZuGleichungen)


static void KPB_GebildetesKPBehandelnRPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildungZuRegeln, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildungZuRegeln, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRP(TermT lhs, TermT rhs)
NFPERM_KERN(NF_NFBildungZuRegeln)


static void KPB_GebildetesKPBehandelnEPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildungZuGleichungen, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnEPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildungZuGleichungen, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnEP(TermT lhs, TermT rhs)
NFPERM_KERN(NF_NFBildungZuGleichungen)


static void KPB_GebildetesKPBehandelnSPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
PERMSUB_BODY(father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnSPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
PERMSUB_BODY(mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnSP(TermT lhs, TermT rhs)
PERMSUB_KERN()


static void KPB_GebildetesKPBehandelnREVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildung, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnREMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NF_BODY(NF_NFBildung, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRE(TermT lhs, TermT rhs)
NF_KERN(NF_NFBildung)



static void KPB_GebildetesKPBehandelnRSVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildungZuRegeln, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRSMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildungZuRegeln, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRS(TermT lhs, TermT rhs)
NFSUB_KERN(NF_NFBildungZuRegeln)


static void KPB_GebildetesKPBehandelnESVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildungZuGleichungen, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnESMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildungZuGleichungen, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnES(TermT lhs, TermT rhs)
NFSUB_KERN(NF_NFBildungZuGleichungen)


static void KPB_GebildetesKPBehandelnREPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildung, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnREPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERM_BODY(NF_NFBildung, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnREP(TermT lhs, TermT rhs)
NFPERM_KERN(NF_NFBildung)



static void KPB_GebildetesKPBehandelnRSPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildungZuRegeln, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRSPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildungZuRegeln, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRSP(TermT lhs, TermT rhs)
NFPERMSUB_KERN(NF_NFBildungZuRegeln)


static void KPB_GebildetesKPBehandelnESPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildungZuGleichungen, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnESPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildungZuGleichungen, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnESP(TermT lhs, TermT rhs)
NFPERMSUB_KERN(NF_NFBildungZuGleichungen)


static void KPB_GebildetesKPBehandelnRESVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildung, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRESMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFSUB_BODY(NF_NFBildung, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRES(TermT lhs, TermT rhs)
NFSUB_KERN(NF_NFBildung)


static void KPB_GebildetesKPBehandelnRESPVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildung, father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnRESPMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NFPERMSUB_BODY(NF_NFBildung, mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnRESP(TermT lhs, TermT rhs)
NFPERMSUB_KERN(NF_NFBildung)



static void KPB_GebildetesKPBehandelnOhneAllesVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
NOTHING_BODY(father, lhs, rhs, actualPosition, WithFather)

static void KPB_GebildetesKPBehandelnOhneAllesMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
NOTHING_BODY(mother, lhs, rhs, actualPosition, WithMother)

static void KPB_IROpferBehandelnOhneAlles(TermT lhs, TermT rhs)
NOTHING_KERN()


void KPB_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
      int levelInfo=0, GRADInfo=0;
      
      if (IS_FindOption(DEF, USR, 1, "r") == 0) levelInfo += 1;
      if (IS_FindOption(DEF, USR, 1, "e") == 0) levelInfo += 2;
      if (IS_FindOption(DEF, USR, 1, "s") == 0) levelInfo += 4;
      if (IS_FindValuedOption(DEF, USR, 1, "permsub") == 0){
         char *str = IS_ValueOfOption(DEF, USR, "permsub");
         if (str && ((GRADInfo = atoi(str)) > 0))
            levelInfo += 8;
      }
      fprintf(IO_Strom, "\n* TREATMENT OF CPS ON GENERATION:    ");

      if (IS_FindValuedOption(DEF, USR, 1, "mnf") != -1) 
        MNF_ParamInfo(what, DEF, USR);
      
      if (!((levelInfo>=0)&&(levelInfo<16))) levelInfo = 7;
      
      switch(levelInfo){
      case 0: fprintf(IO_Strom, "-\n");  break;
      case 1: fprintf(IO_Strom, "R\n");  break;
      case 2: fprintf(IO_Strom, "E\n");  break;
      case 3: fprintf(IO_Strom, "RE\n"); break;
      case 4: fprintf(IO_Strom, "S\n");  break;
      case 5: fprintf(IO_Strom, "RS\n");  break;
      case 6: fprintf(IO_Strom, "ES\n");  break;
      case 7: fprintf(IO_Strom, "RES\n");  break;
      case 8: fprintf(IO_Strom, "P (Permgrad = %u)\n", GRADInfo); break;
      case 9: fprintf(IO_Strom, "RP (Permgrad = %u)\n", GRADInfo); break;
      case 10: fprintf(IO_Strom, "EP (Permgrad = %u)\n", GRADInfo); break;
      case 11: fprintf(IO_Strom, "REP (Permgrad = %u)\n", GRADInfo); break;
      case 12: fprintf(IO_Strom, "SP (Permgrad = %u)\n", GRADInfo); break;
      case 13: fprintf(IO_Strom, "RSP (Permgrad = %u)\n", GRADInfo); break;
      case 14: fprintf(IO_Strom, "ESP (Permgrad = %u)\n", GRADInfo); break;
      case 15: fprintf(IO_Strom, "RESP (Permgrad = %u)\n", GRADInfo); break;
      default: break;
      }
   }
   else{
      fprintf(IO_Strom, "Treatment of CPs after Generation\n");
      fprintf(IO_Strom, "*********************************\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. [r]\n");
      fprintf(IO_Strom, "2. [e]\n");
      fprintf(IO_Strom, "3. [s]\n");
      fprintf(IO_Strom, "4. [permsub=<n>]\n");
      fprintf(IO_Strom, "5. [mnf=[r][e]]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Explanation:\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1-3. These letters cause Waldmeister to normalize newly generated\n");
      fprintf(IO_Strom, "     critical pairs with _r_ules and/or _e_quations, with or without\n");
      fprintf(IO_Strom, "     additional _s_ubsumption\n");
      fprintf(IO_Strom, "4.   This option tells Waldmeister to assume the AC-equations from the given\n");
      fprintf(IO_Strom, "     degree <n> as given (not used for proving).\n");
      fprintf(IO_Strom, "5.   Multiple Normal Forms. Required for some CP-Heuristics.\n");
      MNF_ParamInfo(what, DEF, USR);
   }
}
static char KPB_BehandlungsString[100];

BOOLEAN KPB_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   int level = 0;
   BOOLEAN useSubsumption;
   
   GRAD = 0;

   if (IS_FindOption(DEF, USR, 1, "r") == 0) level += 1;
   if (IS_FindOption(DEF, USR, 1, "e") == 0) level += 2;
   
   useSubsumption =  (IS_FindOption(DEF, USR, 1, "s") == 0);
   if (useSubsumption) level += 4;
   if (IS_FindValuedOption(DEF, USR, 1, "permsub") == 0){
      char *str = IS_ValueOfOption(DEF, USR, "permsub");
      if (str && ((GRAD = atoi(str)) > 0))
         level += 8;
   }
   
#if !COMP_POWER
   fprintf(IO_Strom, "\n* TREATMENT OF CPS ON GENERATION:    ");
#endif

   VerwendeteMNFProzedur = MNFDummy;
   if (IS_FindValuedOption(DEF, USR, 1, "mnf") != -1) 
      if (MNF_InitKPGAposteriori(DEF, USR, GRAD, useSubsumption)) /* nur, falls gueltige Initialisierung */
#if USE_MNF_CG	
	VerwendeteMNFProzedur = MNFCGBearbeiten;
#else
	VerwendeteMNFProzedur = MNF_KPGBearbeite;
#endif   

   if (!((level>=0)&&(level<16))) level = 7;

   switch(level){
   case 0:                      /* nix */
#if !COMP_POWER
      fprintf(IO_Strom, "-\n");
      sprintf(KPB_BehandlungsString,"-");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnOhneAllesVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnOhneAllesMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnOhneAlles;
      break;
   case 1:                      /* R */
#if !COMP_POWER
      fprintf(IO_Strom, "R\n");
      sprintf(KPB_BehandlungsString, "r");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnR;
      break;
   case 2:                      /* E */
#if !COMP_POWER
      fprintf(IO_Strom, "E\n");
      sprintf(KPB_BehandlungsString, "e");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnEVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnEMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnE;
      break;
   case 3:                      /* RE */
#if !COMP_POWER
      fprintf(IO_Strom, "RE\n");
      sprintf(KPB_BehandlungsString, "r:e");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnREVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnREMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRE;
      break;
   case 4:                      /* S */
#if !COMP_POWER
      fprintf(IO_Strom, "S\n");
      sprintf(KPB_BehandlungsString, "s");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnSVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnSMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnS;
      break;
   case 5:                      /* RS */
#if !COMP_POWER
      fprintf(IO_Strom, "RS\n");
      sprintf(KPB_BehandlungsString, "r:s");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRSVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRSMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRS;
      break;
   case 6:                      /* ES */
#if !COMP_POWER
      fprintf(IO_Strom, "ES\n");
      sprintf(KPB_BehandlungsString, "e:s");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnESVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnESMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnES;
      break;
   case 7:                      /* RES */
#if !COMP_POWER
      fprintf(IO_Strom, "RES\n");
      sprintf(KPB_BehandlungsString, "r:e:s");
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRESVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRESMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRES;
      break;
   case 8:                      /* PERM */
#if !COMP_POWER
      fprintf(IO_Strom, "P (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnP;
      break;
   case 9:                      /* RP */
#if !COMP_POWER
      fprintf(IO_Strom, "RP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "r:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRP;
      break;
   case 10:                      /* EP */
#if !COMP_POWER
      fprintf(IO_Strom, "EP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "e:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnEPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnEPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnEP;
      break;
   case 11:                      /* REP */
#if !COMP_POWER
      fprintf(IO_Strom, "REP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "r:e:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnREPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnREPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnREP;
      break;
   case 12:                      /* SP */
#if !COMP_POWER
      fprintf(IO_Strom, "SP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "s:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnSPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnSPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnSP;
      break;
   case 13:                      /* RSP */
#if !COMP_POWER
      fprintf(IO_Strom, "RSP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "r:s:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRSPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRSPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRSP;
      break;
   case 14:                      /* ESP */
#if !COMP_POWER
      fprintf(IO_Strom, "ESP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "e:s:p=%u",GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnESPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnESPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnESP;
      break;
   case 15:                      /* RESP */
#if !COMP_POWER
      fprintf(IO_Strom, "RESP (Permgrad = %u)\n", GRAD);
      sprintf(KPB_BehandlungsString, "r:e:s:p=%u", GRAD);
#endif
      KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnRESPVater;
      KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnRESPMutter;
      KPBehandlungsProcIROpfer = KPB_IROpferBehandelnRESP;
      break;
   default:
      return FALSE;
      break;
   }
   return TRUE;
}


static void (* KPFATHER_ALT) (RegelOderGleichungsT, TermT, TermT, UTermT);
static void (* KPMOTHER_ALT) (RegelOderGleichungsT, TermT, TermT, UTermT);
static void (* KPIROpfer_ALT) (TermT, TermT);
static BOOLEAN abgeschaltet = FALSE;

void KPB_KeineBehandlung(void)
{
  if (abgeschaltet)
    fprintf (IO_Strom, "\nFehler: KPB_KeineBehandlung aufgerufen, aber bereits abgeschaltet!\n");
  abgeschaltet = TRUE;

  KPFATHER_ALT=KPBehandlungsProcFATHER;
  KPMOTHER_ALT=KPBehandlungsProcMOTHER;
  KPIROpfer_ALT=KPBehandlungsProcIROpfer;

  KPBehandlungsProcFATHER = KPB_GebildetesKPBehandelnOhneAllesVater;
  KPBehandlungsProcMOTHER = KPB_GebildetesKPBehandelnOhneAllesMutter;
  KPBehandlungsProcIROpfer = KPB_IROpferBehandelnOhneAlles;
}

void KPB_WiederBehandeln(void)
{
  if (!abgeschaltet)
    fprintf(IO_Strom, "\nFehler: KPB_WiederBehandeln aufgerufen, aber nicht abgeschaltet!\n");
  abgeschaltet = FALSE;

  KPBehandlungsProcFATHER  = KPFATHER_ALT;
  KPBehandlungsProcMOTHER  = KPMOTHER_ALT;
  KPBehandlungsProcIROpfer = KPIROpfer_ALT;
}

char *KPB_ArtBehandlung(void)
{
  return KPB_BehandlungsString;
}

void KPB_KonklusionenAbarbeiten(RegelOderGleichungsT r)
/* Meldet die entsprechende Regel oder Gleichung dem Datenbankexperten an und
   verarbeitet die gezogenen Schlussfolgerungen.
*/
{
  RegelOderGleichungsT Konklusion;
  
  CM_IntegriertesFaktumMelden(r);
  for (Konklusion=CM_ErsteKonklusionsinstanz(); Konklusion; Konklusion = CM_NaechsteKonklusionsinstanz()){
    SUE_CPgeneratedWithDatabase(Konklusion);
    /* keinerlei weitere Verarbeitung! */
    SUE_recentCPinsert();
  }
}
