/*=================================================================================================================
  =                                                                                                               =
  =  KPMengenInterreduktion.c                                                                                     =
  =                                                                                                               =
  =  Interreduktion/Subsumption der KP-Menge                                                                      =
  =                                                                                                               =
  =  27.03.96  Arnim. Aus der Interreduktion ausgegliedert.                                                       =
  =================================================================================================================*/

#include "Achtzigstel.h"
#include "Ausgaben.h"
#include "compiler-extra.h"
#include "general.h"
#include "InfoStrings.h"
#include "KPMengenInterreduktion.h"
#include "NFBildung.h"
#include "Subsumption.h"
#include "Stichpunkte.h"
#include <string.h>
#include "SUEM.h"
#include "TermOperationen.h"


/* -------------------------------------------------------------------------------*/
/* --------------- Statistik -----------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

static BOOLEAN StatistikAn;

static unsigned long        AnzKPIRBehandelt;                  /* Bei der KPM-IR insgesamt betrachtete KPs */
static unsigned long        AnzKPIRReduziert;                  /* Bei der KPM-IR reduzierte KPs */
static unsigned long        AnzKPIRZusammengefuehrt;           /* Bei der KPM-IR von der Anwendungsprozedur zusammengefuehrte KPs */
static unsigned long        AnzKPIRSubsummiert;                /* Bei der KPM-IR von der Anwendungsprozedur subsummierte KPs */
static unsigned long        AnzKPnurW;                         /* Beim Neugewichten angefasste KPs */

#define IncAnzKPIRBehandelt_Stat()        (AnzKPIRBehandelt++)
#define IncAnzKPIRReduziert_Stat()        (AnzKPIRReduziert++)
#define IncAnzKPIRZusammengefuehrt_Stat() (AnzKPIRZusammengefuehrt++)
#define IncAnzKPIRSubsummiert_Stat()      (AnzKPIRSubsummiert++)
#define IncAnzKPIR_nurW_Stat()            (AnzKPnurW++)

#define IncAnzKPIRBehandelt_NoStat()
#define IncAnzKPIRReduziert_NoStat()
#define IncAnzKPIRZusammengefuehrt_NoStat()
#define IncAnzKPIRSubsummiert_NoStat()
#define IncAnzKPIR_nurW_NoStat()

void KPIR_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("KPIR", IS) || IS_IsIn("all", IS)))
      return;
   if (IS_IsIn("standard", IS)){
      if (StatistikAn){
         int tiefe1 = IO_Platzverbrauch(Maximum(2, AnzKPIRReduziert, AnzKPIRSubsummiert));
         
         IO_DruckeKurzUeberschrift("Intermediate Reprocessing (IRP)", '+');
         IO_DruckeStatistikZeile("CPs touched (total)", AnzKPIRBehandelt);
         IO_DruckeStatistikZeileEingerueckt("of them subsumed", AnzKPIRSubsummiert, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them simplified", AnzKPIRReduziert, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them joined", AnzKPIRZusammengefuehrt, tiefe1+IO_Platzverbrauch(AnzKPIRZusammengefuehrt));
         IO_DruckeStatistikZeile("CPs touched only for reweighting", AnzKPnurW);
      }
   }
   else if (IS_IsIn("latex-head", IS)){
      if (StatistikAn)
         fprintf(IO_Strom, "KPIR: touched & subs & red & join & rew & ");
   }
   else if (IS_IsIn("config", IS)){
      fprintf(IO_Strom, "KPIR: Statistik %s\n", StatistikAn?" AN":"AUS");
   }
   else if (IS_IsIn("intern", IS)){
      ;
   }
}
      

/* -------------------------------------------------------------------------------*/
/* --------------- Strings in Strength umwandeln ---------------------------------*/
/* -------------------------------------------------------------------------------*/

static int String2Strength(char *str)
{
   int val = 0;

   if (strchr(str, 'r')) val |= 1;
   if (strchr(str, 'e')) val |= 2;
   if (strchr(str, 's')) val |= 4;
   if (strchr(str, 'w')) val |= 8;

   return val;
}
      
/* -------------------------------------------------------------------------------*/
/* --------------- Anwendungsprozeduren ------------------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Anwendungsprozeduren, die dem KP-Mengendurchlauf uebergeben werden, um die KP-Mengen-Interreduktion
   durchzufuehren.
   Vor Aufruf jeder Anwendungsprozedur kann davon ausgegangen werden, dass das entsprechende KP als aktuelles (bzgl. IR) angemeldet
   und physikalisch noch intakt ist. Es sind also sowohl Zugriffe auf die Eintrage des Objekts kp als auch ueber die
   KPO_...AktKP()-Funktionen statthaft.*/


#define REDUKTIONS_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik)                \
{                                                                                                                                       \
   PCL_Reduktionsbeginn(KP, pclID, linkeSeite, rhs);                                                                                    \
   reduziert = NF_ ## Womit (lhs, KPsIntermed);                                                                                                   \
   PCL_Reduktionsbeginn(KP, pclID, rechteSeite, lhs);                                                                                   \
   reduziert |= NF_ ## Womit (rhs, KPsIntermed);                                                                                                  \
                                                                                                                                        \
   if (reduziert){                                                                                                                      \
      IncAnzKPIRReduziert_ ## statistik();                                                                                              \
      if (TO_TermeGleich(lhs, rhs)){                                                                                                    \
         IncAnzKPIRZusammengefuehrt_ ## statistik();                                                                                    \
         return 1;                                                                                                                      \
      }                                                                                                                                 \
   }                                                                                                                                    \
}   

#define SUBSUMPTIONS_BODY(/*{Stat, NoStat}*/ statistik)                                                                                 \
{                                                                                                                                       \
   if (SS_TermpaarSubsummiertVonGM(lhs, rhs)){                                                                                             \
      IncAnzKPIRSubsummiert_## statistik();                                                                                             \
      return 1;                                                                                                                         \
   }                                                                                                                                    \
}

#define COMBINE_RS_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik, ReturnExpr) \
{                                                                                                                                    \
   BOOLEAN reduziert;                                                                                                                \
                                                                                                                                     \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                              \
   AZ_DruckeMeldungItem();                                                                                                           \
   REDUKTIONS_BODY(Womit, statistik);                                                                                                \
   SUBSUMPTIONS_BODY(statistik);                                                                                                     \
   return (ReturnExpr);                                                                                                              \
}

#define S_BODY(/*{Stat, NoStat}*/ statistik, ReturnExpr)                                                                                \
{                                                                                                                                       \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                                 \
   AZ_DruckeMeldungItem();                                                                                                              \
   SUBSUMPTIONS_BODY(statistik);                                                                                                        \
   return (ReturnExpr);                                                                                                                 \
}

#define R_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik, ReturnExpr)             \
{                                                                                                                                       \
   BOOLEAN reduziert;                                                                                                                   \
                                                                                                                                        \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                                 \
   AZ_DruckeMeldungItem();                                                                                                              \
   REDUKTIONS_BODY(Womit, statistik);                                                                                                   \
   return (ReturnExpr);                                                                                                                 \
}

#define W_BODY(/*{Stat, NoStat}*/ statistik)                                                                                            \
{                                                                                                                                       \
   AZ_DruckeMeldungItem();                                                                                                              \
   IncAnzKPIR_nurW_ ## statistik();                                                                                                     \
   return 2;                    /* immer neugewichten */                                                                                \
}


static int AP_15_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildung, Stat, 2)
static int AP_15_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildung, NoStat, 2)

static int AP_14_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuGleichungen, Stat, 2)
static int AP_14_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuGleichungen, NoStat, 2)

static int AP_13_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuRegeln, Stat, 2)
static int AP_13_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuRegeln, NoStat, 2)

static int AP_12_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
S_BODY(Stat,2)
static int AP_12_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
S_BODY(NoStat,2)

static int AP_11_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildung, Stat, 2)
static int AP_11_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildung, NoStat, 2)

static int AP_10_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuGleichungen, Stat, 2)
static int AP_10_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuGleichungen, NoStat, 2)

static int AP_9_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuRegeln, Stat, 2)
static int AP_9_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuRegeln, NoStat, 2)


static int AP_8_Stat(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
W_BODY(Stat)
static int AP_8_NoStat(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
W_BODY(NoStat)

static int AP_7_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildung, Stat, reduziert?2:0)
static int AP_7_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildung, NoStat, reduziert?2:0)

static int AP_6_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuGleichungen, Stat, reduziert?2:0)
static int AP_6_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuGleichungen, NoStat, reduziert?2:0)

static int AP_5_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuRegeln, Stat, reduziert?2:0)
static int AP_5_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
COMBINE_RS_BODY(NFBildungZuRegeln, NoStat, reduziert?2:0)

static int AP_4_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
S_BODY(Stat,0)
static int AP_4_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
S_BODY(NoStat,0)

static int AP_3_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildung, Stat, reduziert?2:0)
static int AP_3_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildung, NoStat, reduziert?2:0)

static int AP_2_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuGleichungen, Stat, reduziert?2:0)
static int AP_2_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuGleichungen, NoStat, reduziert?2:0)

static int AP_1_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuRegeln, Stat, reduziert?2:0)
static int AP_1_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
R_BODY(NFBildungZuRegeln, NoStat, reduziert?2:0)



static int AP_0(TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclID MAYBE_UNUSEDPARAM)
{
   return 0;
}



static const char *strengthArray[16]={"o", "r", "e", "re", "s", "rs", "es", "res", "w",
			       "rw", "ew", "rew", "sw", "rsw", "esw", "resw"};

typedef int (* WTProcT) (TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID);

static WTProcT AP_Stat[16] = {AP_0, AP_1_Stat, AP_2_Stat, AP_3_Stat, AP_4_Stat, AP_5_Stat, AP_6_Stat, AP_7_Stat, AP_8_Stat,
                              AP_9_Stat, AP_10_Stat, AP_11_Stat, AP_12_Stat, AP_13_Stat, AP_14_Stat, AP_15_Stat};

static WTProcT AP_NoStat[16] = {AP_0, AP_1_NoStat, AP_2_NoStat, AP_3_NoStat, AP_4_NoStat, AP_5_NoStat, AP_6_NoStat, AP_7_NoStat, AP_8_NoStat,
                                AP_9_NoStat, AP_10_NoStat, AP_11_NoStat, AP_12_NoStat, AP_13_NoStat, AP_14_NoStat, AP_15_NoStat};

static WTProcT *AP;


/* Staerke:
   0 -- nix (dh. nur Waisenmord)
   1 -- r
   2 -- e
   3 -- re
   4 -- s
   5 -- rs
   6 -- es
   7 -- res
   8 -- w        _W_eight
   9 -- rw
  10 -- ew
  11 -- rew
  12 -- sw
  13 -- rsw
  14 -- esw
  15 -- resw
*/

/* -------------------------------------------------------------------------------*/
/* --------------- KP-Mengen-Durchlaeufe -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

static StichpunktT *StichpunktListe;
static unsigned long int AnzahlStichpunkte;
static unsigned long int Period;
static int PeriodStrength;



#define KPIR_BODY(strength)                                                                                     \
{                                                                                                               \
   AZ_DruckeMeldungStart("critical pairs", strengthArray[strength], SUE_AnzKPsToGo());                          \
   SUE_WALKTHROUGH(AP[strength], TRUE);                                                                         \
   AZ_DruckeMeldungEnde();                                                                                      \
}
   


void KPIR_KPMengeInterreduzieren(void)
/* Interreduktion der KP-Menge gemaess Aufruf-Einstellung.*/
{
   int strength;
   /* Bedingungsreihenfolge relevant: Stichpunkte werden immer geprueft. Damit
      wird kein Stichpukt durch paralleles periodisches Interreduzieren uebersprungen!
      */
   if (SP_IstStichpunkt(StichpunktListe, AnzahlStichpunkte,
			RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt(),
			Period, PeriodStrength, &strength))
      KPIR_BODY(strength)
}


void KPIR_Wahlfrei(char *howStrong)
/* Fuehrt die KP-Mengen-Interreduktion zu einem beliebigen Zeitpunkt aus.
   Funktionalitaet (Maechtigkeit und dyn. Neubewerten)  wie beim Programm-Start
   vorgesehen. */
{
   int strength = String2Strength(howStrong);

   KPIR_BODY(strength);
}


/* -------------------------------------------------------------------------------*/
/* --------------- Initialisieren ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

 
  
void KPIR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
     StichpunktT *localStichpunktListe = StichpunktListe; /* retten */
     long unsigned int Period=0;
     int PeriodStrength=0;
     unsigned long int AnzahlStichpunkte;
     BOOLEAN StatistikAn;

     StatistikAn = (IS_ValueOfOption(DEF, USR, "statistics") 
		    && (strcmp(IS_ValueOfOption(DEF, USR, "statistics"), "on") == 0));
     
     SP_StichpunktlisteGenerieren(DEF,USR, &Period, &PeriodStrength,
				  &StichpunktListe, &AnzahlStichpunkte,16,
				  strengthArray);
      
     fprintf(IO_Strom, "\n* INTERMEDIATE CP-SET-REDUCTION\n\n");
     SP_Ausgabe (Period, PeriodStrength, StichpunktListe, AnzahlStichpunkte,
		 strengthArray, StatistikAn);

     our_dealloc(StichpunktListe);
     StichpunktListe = localStichpunktListe;
   }
   else{
      IO_DruckeKurzUeberschrift("INTERMEDIATE CP-SET-REDUCTION", '-');
      fprintf(IO_Strom, "1. [period=<n><strength>]\n");
      fprintf(IO_Strom, "2. [<strength>=<n1>.<n2>.<n3>...]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Explanation:\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. Tells Waldmeister to periodically normalize the members of SUE every <n> new rules/equations\n");
      fprintf(IO_Strom, "    with <strength>. One may declare only one period.\n");
      fprintf(IO_Strom, "2. Tells Waldmeister to normalize all members of SUE after a certain number of new rules/equations\n");
      fprintf(IO_Strom, "    (<n1>, <n2>, <n3>) with <strength>.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "<strength> is a string within which only the following letters will be recognized:\n");
      fprintf(IO_Strom, "r - rules           : use rules for normalization\n");
      fprintf(IO_Strom, "e - equations       : use equations for normalization\n");
      fprintf(IO_Strom, "s - subsumption     : subsume\n");
      fprintf(IO_Strom, "w - weight          : reweight every touched ue\n");
      fprintf(IO_Strom, "o - orphan-murder   : remove all the orphans.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "If normalizing the cps intermediately, reweighting of the simplified ones\n");
      fprintf(IO_Strom, "will be done automatically. Those joined or subsumed will be removed from\n");
      fprintf(IO_Strom, "the set.\n");
      fprintf(IO_Strom, "It is not necessary to state an 'o' when intermediate reductions are invoked:\n");
      fprintf(IO_Strom, "full orphan murder will be executed on these ocasions automatically.\n");
   }
}
   

BOOLEAN KPIR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{

  if ((StatistikAn = (IS_ValueOfOption(DEF, USR, "statistics") 
		      && (strcmp(IS_ValueOfOption(DEF, USR, "statistics"), "on") == 0))))
    AP = AP_Stat;
  else 
    AP = AP_NoStat;
  
  SP_StichpunktlisteGenerieren(DEF,USR, &Period, &PeriodStrength,
			       &StichpunktListe, &AnzahlStichpunkte,16,
			       strengthArray);
#if !COMP_POWER  
  fprintf(IO_Strom, "\n* INTERMEDIATE CP-SET-REDUCTION\n\n");
  SP_Ausgabe (Period, PeriodStrength, StichpunktListe, AnzahlStichpunkte,
	      strengthArray, StatistikAn);
#endif
  
  return TRUE;
}
