/*=================================================================================================================
  =                                                                                                               =
  =  Konsulat.c                                                                                                   =
  =                                                                                                               =
  =                                                                                                               =
  =  25.10.98  Thomas.                                                                                            =
  =                                                                                                               =
  =================================================================================================================*/


#include "Ausgaben.h"
#include "Konsulat.h"
#include "Hauptkomponenten.h"
#include "LispLists.h"
#include "MNF.h"
#include "parser.h"
#include "PhilMarlow.h"
#include "Ordnungen.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "Termpaare.h"
#include "WaldmeisterII.h"
#include "XFilesFuerPMh.h"
#include "Zielverwaltung.h"


/*=================================================================================================================*/
/*= I. Kern-Strategie                                                                                             =*/
/*=================================================================================================================*/

/* ------- 0. Apparat -------------------------------------------------------------------------------------------- */

BOOLEAN KO_KonsultationErforderlich;
unsigned int KO_Stufe;
TermT KO_Kern,
      KO_Fixpunktkombinator;
static struct ListenTS InitialesZiel;
static SymbolT Skolemfunktion;

#define /*SymbolT*/ _     TO_ApplyOperator
#define /*SymbolT*/ combi SO_const1
#define /*SymbolT*/ sfp   SO_const2
#define /*SymbolT*/ x1    SO_ErstesVarSymb
#define /*SymbolT*/ eq    SO_EQ
#define /*SymbolT*/ true  SO_TRUE
#define /*SymbolT*/ false SO_FALSE


static void PassivesFaktumAufnehmen(TermT l, TermT r)
{
  SUE_CPinitial(l,r);
  SUE_recentCPinsert();
}

static void AktivesFaktumWegwerfen(RegelOderGleichungsT Faktum)
{
  void (*Wegwerfen)(RegelOderGleichungsT Faktum) =
    TP_TermpaarIstRegel(Faktum) ? RE_RegelVerfluechtigen : RE_GleichungVerfluechtigen;
  Wegwerfen(Faktum);
}

static void PassiveFaktenWegwerfen(void)
{ 
  while (SUE_Select())    
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
}


/* ------- 1. Stufe 1 -------------------------------------------------------------------------------------------- */

static void SchwacheFixpunkteigenschaftAufnehmen(void)
{
  TermT l = TO_(x1),
        r = TO_(_,combi,x1);
  ListenT Ziel = car(PM_Ziele);
  InitialesZiel = *Ziel;
  Ziel->car = l;
  Ziel->cdr = r;
}


/* ------- 2. Stufe 2 -------------------------------------------------------------------------------------------- */

static void NegativeAktiveFaktenWegwerfen(void)
{
  RegelOderGleichungsT Faktum;
  for (Faktum = RE_ErstesFaktum(); Faktum; Faktum = RE_NaechstesFaktum())
    if (RE_IstUngleichung(Faktum))
      AktivesFaktumWegwerfen(Faktum);
}


static void FixedPositivAufnehmen(void)
{
  const SymbolT f = Skolemfunktion;
  TermT l = TO_(eq, f,x1, _,x1,f,x1),
        r = TO_(true);
  PassivesFaktumAufnehmen(l,r);
}


static void KernAbstrahieren(TermT *Term1, TermT *Term2)
{ 
  SymbolT x = SO_VorErstemVarSymb,
          y;
  UTermT t;
  TO_doSymbole(KO_Kern,y, {
    if (SO_VariableIstNeuer(y,x))
      x = y;
  });
  x = SO_NaechstesVarSymb(x);
  TO_doStellen(KO_Kern,t, {
    if (SO_SymbGleich(TO_TopSymbol(t),combi))
      TO_SymbolBesetzen(t,x);
  });
  *Term1 = TO_(x);
  *Term2 = KO_Kern;
}


static void FixedNegativAufnehmen(TermT s, TermT t)
{
  Z_UngleichungEinmanteln(&s,&t);
  PassivesFaktumAufnehmen(s,t);
}


/* ------- 3. Stufe 3 -------------------------------------------------------------------------------------------- */

static void AktiveFaktenWegwerfen(void)
{
  RegelOderGleichungsT Faktum;
  for (Faktum = RE_ErstesFaktum(); Faktum; Faktum = RE_NaechstesFaktum())
    AktivesFaktumWegwerfen(Faktum);
}

static void KombinatorenAufnehmen(void)
{
  ListenT Index;
  for (Index=PM_Gleichungen; Index; Cdr(&Index)) {
    ListenT Eintrag = car(Index);
    TermT Flunder1 = TO_Termkopie(car(Eintrag)),
          Flunder2 = TO_Termkopie(cdr(Eintrag));
    PassivesFaktumAufnehmen(Flunder1,Flunder2);
  }
}

static void FixpunktkombinatorAufnehmen(void)
{
  TermT l = TO_(sfp),
        r = KO_Fixpunktkombinator;
  PassivesFaktumAufnehmen(l,r);
}


static void ZielStarkeFixpunkteigenschaftAufnehmen(void)
{
  TermT l = TO_(_,sfp,combi),
        r = TO_(_,combi,_,sfp,combi);
  MNF_InitNMengeAposteriori(AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,"mnf=fifo");
  Z_EinZielEinfuegen(l,r,3,Z_MNF_Ziele);
}


/* ------- 4. Stufe 4 -------------------------------------------------------------------------------------------- */

static void StarkeFixpunkteigenschaftAufnehmen(void)
{
  TermT l = TO_(_,sfp,x1),
        r = TO_(_,x1,_,sfp,x1);
  PassivesFaktumAufnehmen(l,r);
}


static void InitialesZielAufnehmen(void)
{
  TermT l = TO_Termkopie(car(&InitialesZiel)),
        r = TO_Termkopie(cdr(&InitialesZiel));
  ListenT Ziel = car(PM_Ziele);
  MNF_InitNMengeAposteriori(AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,"");
  Z_EinZielEinfuegen(l,r,4,Z_Paramodulation);
#if 0
  TO_TermeLoeschen((TermT)car(Ziel),(TermT)cdr(Ziel));
#endif
  *Ziel = InitialesZiel;
}


static void EqXXIsTrueAufnehmen(void)
{
  TermT l = TO_(eq,x1,x1),
        r = TO_(true);
  PassivesFaktumAufnehmen(l,r);
}


/* ------- 5. Aggregationen --------------------------------------------------------------------------------- */

extern unsigned short int AnzErfSubsumptionenZiele,
                          AnzErfZusfuehrungenZiele;

void KO_Konsultieren(void)
{
  if (KO_KonsultationErforderlich) {
    TermT s, t;
    KO_KonsultationErforderlich = FALSE;
    switch (++KO_Stufe) {

    case 1:
      ORD_OrdnungUmstellen(ordering_empty);
      SchwacheFixpunkteigenschaftAufnehmen();
      break;

    case 2:
      PassiveFaktenWegwerfen();
      NegativeAktiveFaktenWegwerfen();
      FixedPositivAufnehmen();
      KernAbstrahieren(&s,&t);
      FixedNegativAufnehmen(s,t);
      break;

    case 3:
      PassiveFaktenWegwerfen();
      AktiveFaktenWegwerfen();
      ORD_OrdnungUmstellen(ordering_col3);
      KombinatorenAufnehmen();
      FixpunktkombinatorAufnehmen();
      ZielStarkeFixpunkteigenschaftAufnehmen();
      break;

    case 4:
      PassiveFaktenWegwerfen();
      AktiveFaktenWegwerfen();
      ORD_OrdnungUmstellen(ordering_col4);
      StarkeFixpunkteigenschaftAufnehmen();
      InitialesZielAufnehmen();
      EqXXIsTrueAufnehmen();
      AnzErfSubsumptionenZiele = 0;
      AnzErfZusfuehrungenZiele = -1;                               /* Hack: Statistik in Zielverwaltung bereinigen */
      break;

    default:
      break;
    }
  }
}



/*=================================================================================================================*/
/*= II. Initialisierung                                                                                           =*/
/*=================================================================================================================*/


static BOOLEAN Passabel(TermT l, TermT r, SymbolT f)
{
  return TO_TermEntspricht(l,4, _,x1,f,x1)
    && TO_TermEntspricht(r,7, _,f,x1,_,x1,f,x1);
}


extern StrategieT **GewaehlteStrategien;

static BOOLEAN ZielIstStarkerFixpunktsatz(void)
{
  eq_list Ziele = get_conclusions();
  if (!Ziele || eq_rest(Ziele))
    our_error("Option -kern requires exactly one hypothesis.");
  else {
    term L = eq_first_lhs(Ziele),
         R = eq_first_rhs(Ziele);
    TermT l,r;
    SymbolT f;
    TO_PlattklopfenLRMitMalloc(L,R,&l,&r);
    SO_forConclusionSymbole(f)
      if (SO_Stelligkeit(f)==1)
        if (Passabel(l,r,f) || Passabel(r,l,f)) {
          Skolemfunktion = f;
          break;
        }
    if (Skolemfunktion==SO_ErstesVarSymb) {
      if (PA_Automodus) {
        GewaehlteStrategien[WM_AnlaufNr-1]->Schrittzahl = 1;    /* Hack: Strategiewechsel nach erstem KP erzwingen */
        return FALSE;
      }
      else
        our_error("unsuitable hypothesis for -kern option");
    }
    TO_TermLoeschenM(l);
    TO_TermLoeschenM(r);
  }
  return TRUE;
}


void KO_InitAposteriori(void)
{
  KO_Kern = KO_Fixpunktkombinator = NULL;
  KO_Stufe = 0;
  Skolemfunktion = SO_ErstesVarSymb;
  if ((KO_KonsultationErforderlich = PA_KernAn)) {
    TO_KombinatorapparatAktivieren();
    if (ZielIstStarkerFixpunktsatz())
      KO_Konsultieren();
  }
}
