/*=================================================================================================================
  =                                                                                                               =
  =  MatchOperationen.c                                                                                           =
  =                                                                                                               =
  =  Matchen von Termen mit Regel- oder Gleichungsbaum                                                            =
  =                                                                                                               =
  =                                                                                                               =
  =  26.03.94  Thomas.                                                                                            =
  =                                                                                                               =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "DSBaumKnoten.h"
#include "general.h"
#include "Hauptkomponenten.h"
#include "MatchOperationen.h"
#include "NFBasis.h"
#include "RUndEVerwaltung.h"
#include "SpeicherVerwaltung.h"
#include "SymbolOperationen.h"
#include "TermIteration.h"
#include "Ordnungen.h"
#include "WaldmeisterII.h"
#if       COMP_ACIDWM
#include "memory.h"
#endif


TI_IterationDeklarieren(MO)
TI2_IterationDeklarieren(MO)

#define MO_STATISTIK normal_power_comp(1,0,0)     /* Misserfolge wg. Ordnung zaehlen. Alternativen: 0 fuer aus, 1 fuer an */
#define MO_TAUSEND_TERME 0                                /* ggf. mehrere tausend Anfragen in Datei protokollieren */
#define TAUSEND 5000


/* Bei konkurrenter Freispeicherbenutzung sind Kommentare "$nested-feature" zu beruecksichtigen. */



/*=================================================================================================================*/
/*= I. Variablenbank fuer Substitutionen; sonstige globale Groessen; Statistik-Funktionen                         =*/
/*=================================================================================================================*/

GleichungsT MO_SubsummierendeGleichung;

BOOLEAN MO_IndirekterClashImBaum;

if_MO_Objekt(RegelOderGleichungsT MO_AngewendetesObjekt;)        /* fuer PCL-Meldung, falls Regel angewendet wurde */

#if COMP_ACIDWM
dt_QUERY MO_RegelbaumDurchwandernAcidWM(UTermT Term);/* Durchlaeuft den ganzen Baum, um nach Generalisierungen zu suchen. */
#endif

static int xMDimension;           /* xO realisiert Abbildung Variablensymbol->UTerm. Das Feld wird mit den Variab- */
typedef struct {    /* lensymbolen negativ indiziert und hat die Laenge xMDimension. Fuer xM[0] wird kein Speicher */
  UTermT Substitut,    /* angelegt. Vorg ist die letzte Zelle im Anfrageterm, Nachf entsprechend am Substitutende. */
         Vorg,
         Nachf;                       /* Das Flag SchonEingebaut wird beim Anwenden der Substitution benoetigt und */
  BOOLEAN SchonEingebaut;   /* gibt an, ob die das Substitut repraesentierende Termzellenkette aus dem Anfrageterm */
} MO_EintragsT;                                        /* schon in den neuen Term eingebaut worden ist oder nicht. */
static MO_EintragsT *xM;                                                             /* vor Beginn des Substituts. */
static SymbolT xMletztes;

unsigned int MO_AusdehnungVariablenbank(void)
{
  return -xMDimension;
}

void MO_VariablenfeldAnpassen(SymbolT NeuesteBaumvariable)              /* xM wird initial als Feld mit Ausdehnung */
{                            /* MO_AnzErsterVariablen angelegt und muss gegebenfalls nach links ausgedehnt werden. */
  if (SO_VariableIstNeuer(NeuesteBaumvariable,xMDimension)) {
    our_dealloc(xM+xMDimension);                            /* Addition von xMDimension fuer negative Adressierung */
    xMDimension = SO_VarDelta(NeuesteBaumvariable,MO_AnzErsterVariablen);
    xM = (MO_EintragsT *) our_alloc(-xMDimension*sizeof(MO_EintragsT)) - xMDimension;
  }
}


static KantenT BaumIndex;                                         /* Zeiger fuer Durchlauf des Baumes beim Matchen */
#define aktTermrest BK_aktTIndexM(BaumIndex)
#define SymbolAusTerm TO_TopSymbol(aktTermrest)                 /* naechstes zu matchendes Zeichen aus Anfrageterm */

static TermT gefundeneRechteSeite,    /* haelt gefundene rechte Seite zwischen Matchen und Erzeugen von SigmaR vor */
             gefundeneLinkeSeite;
#define TopsymbolRechteSeite TO_TopSymbol(gefundeneRechteSeite)  

static UTermT Anfrageterm,                                         /* globale Variablen; Zugriff von fast ueberall */
              letzteZelleAnfrageterm,                                                   /* dessen letzte Termzelle */
              letzteNichtVernutzteZelle;                                     /* letzte nicht rezyklierte Termzelle */

static BOOLEAN NichtGanzerTermGebunden;                           /* Information fuer Beseitigung des Anfrageterms */


static unsigned long AnzRedMisserfolgeOrdnung=0;                /* An >-Test gescheiterte Reduktionsversuche mit E */
#if MO_STATISTIK
#define IncAnzRedMisserfolgeOrdnung() (AnzRedMisserfolgeOrdnung++)
#else
#define IncAnzRedMisserfolgeOrdnung()  /* nix tun */
#endif

#if MO_TAUSEND_TERME
static FILE *TausendTermeDatei;
static unsigned long int TausendTermeZaehler;
#endif



/*=================================================================================================================*/
/*= II. Elementare Operationen auf oder mit Substitutionen                                                        =*/
/*=================================================================================================================*/

#define SubstAufIdentitaetSetzen()                                                                                  \
  xMletztes = SO_VorigesVarSymb(SO_ErstesVarSymb);


#define MatcherInitialisieren(/*TermT*/ Term)                                                                       \
  letzteZelleAnfrageterm = TO_TermEnde(Anfrageterm = Term);                                                         \
  NichtGanzerTermGebunden = TRUE;                                                                                   \
  SubstAufIdentitaetSetzen();


static void SubstErgaenzen(UTermT Substitut, UTermT Vorg)      /* fuegt zur Substitution x(n)<-t(n) ... x(1)<-t(1) */
{                                                                           /* die Bindung x(n+1)<-Substitut hinzu */
  SO_NaechstesVarSymbD(xMletztes);
  xM[xMletztes].Substitut = Substitut;
  xM[xMletztes].Nachf = TO_NaechsterTeilterm(Substitut);
  xM[xMletztes].SchonEingebaut = FALSE;
  if (!(xM[xMletztes].Vorg = Vorg))
    NichtGanzerTermGebunden = FALSE;
}


#define /*BOOLEAN*/ VarUngebunden(/*SymbolT*/ VarSymbol)                                                            \
  SO_VariableIstNeuer(VarSymbol,xMletztes)


#define /*BOOLEAN*/ VarGebunden(/*SymbolT*/ VarSymbol)                                                              \
  SO_VariableIstNichtNeuer(VarSymbol,xMletztes)


#define /*BOOLEAN*/ SymbKeineGebundeneVar(/*SymbolT*/ Symbol)                                                       \
  (SO_SymbIstFkt(Symbol) || VarUngebunden(Symbol))


#define SubstitutVon(/*SymbolT*/ VarSymbol)                                                                         \
  xM[VarSymbol].Substitut


static BOOLEAN EsFolgtSubstitutVon(SymbolT VarSymbol, UTermT TermIndex) 
/* ermittelt, ob der bezeichnete Term als Praefix Sigma(VarSymbol) hat */
#define /*Block*/ RumpfEsFolgtSubstitutVon(/*Anweisung*/ Abbruchaktion, /*SymbolT*/ VarSymbol, /*UTermT*/ TermIndex)\
{                                                                                                                   \
  UTermT SubstitutIndex = SubstitutVon(VarSymbol);                                                                  \
  UTermT NULLSubstitut = TO_NaechsterTeilterm(SubstitutIndex);                                                      \
  do {                                                                                                              \
    if (SO_SymbUngleich(TO_TopSymbol(TermIndex),TO_TopSymbol(SubstitutIndex))) {                                    \
      Abbruchaktion;                                                                                                \
      return FALSE;                                                                                                 \
    }                                                                                                               \
    TermIndex = TO_Schwanz(TermIndex);                                                                              \
    SubstitutIndex = TO_Schwanz(SubstitutIndex);                                                                    \
  } while (TO_PhysischUngleich(SubstitutIndex,NULLSubstitut));                                                      \
  return TRUE;                                                                                                      \
}
RumpfEsFolgtSubstitutVon(/*nichts tun*/,VarSymbol,TermIndex)


static BOOLEAN EsFolgtSubstitutOderIC(SymbolT VarSymbol, UTermT TermIndex)
/* wie EsFolgtSubstitutVon, jedoch wird im Fehlerfall das Flag MO_IndirekterClashImBaum gesetzt */
RumpfEsFolgtSubstitutVon(MO_IndirekterClashImBaum = TRUE,VarSymbol,TermIndex)


void MO_DruckeMatch(void)
{
  SymbolT VarIndex;

  SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb,xMletztes) {
    fprintf(IO_Strom, "x");
    IO_DruckeGanzzahl(SO_VarNummer(VarIndex));
    fprintf(IO_Strom, "<-");
    IO_DruckeTerm(SubstitutVon(VarIndex));
    IO_LF();
  }
}



/*=================================================================================================================*/
/*= III. Operationen zum Matchen mit Regeln auf Blattebene                                                        =*/
/*=================================================================================================================*/

static SymbolT xMletztesAlt;

static BOOLEAN RegelInBlattGefunden(void)       /* Im Baum begonnenen Match-Versuch fortsetzen und falls moeglich beenden */
{
  UTermT Termrest = aktTermrest,
         Regelrest = BK_Rest(BaumIndex),
         Vorg = BK_MVorgZelle(BaumIndex);        /* letzte Termzelle vor gerade betrachtetem Rest des Anfrageterms */

  xMletztesAlt = xMletztes;                   /* letzte im Baum vorgenommene Bindung vermerken (fuer Ruecksetzung) */
  while (Regelrest) {                              /* anstatt Abfrage auf Ende des (unbeschnittenen!) Anfrageterms */
    SymbolT SymbAusRegel = TO_TopSymbol(Regelrest);
    if (SO_SymbIstVar(SymbAusRegel)) {
      if (VarUngebunden(SymbAusRegel)) {                                     /* ungebundene Variable aus Regelterm */
        SubstErgaenzen(Termrest,Vorg);                                                    /* in Anfrageterm binden */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));
        Regelrest = TO_Schwanz(Regelrest);
      }
      else if (EsFolgtSubstitutOderIC(SymbAusRegel,Termrest)) {                /* fuer gebundene Variable pruefen, */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));                             /* ob sich der Anfragterm */
        Regelrest = TO_Schwanz(Regelrest);                                               /* entsprechend fortsetzt */
      }
      else {                                                           /* Falls nicht, den Matchversuch abbrechen. */
        xMletztes = xMletztesAlt;                                          /* Dabei Bindungstabelle zuruecksetzen. */
        return FALSE;
      }
    }
    else if (SO_SymbUngleich(SymbAusRegel,TO_TopSymbol(Termrest))) {              /* Abbruch ebenso bei ungleichen */
      xMletztes = xMletztesAlt;                                                              /* Funktionssymbolen. */
      return FALSE;
    }
    else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
      Termrest = TO_Schwanz(Vorg = Termrest);
      Regelrest = TO_Schwanz(Regelrest);
    }
  }
  return TRUE;
}


#define /*void*/ MatcherNachBlattZuruecksetzen(/*void*/)                                                            \
  xMletztes = xMletztesAlt
  


/*=================================================================================================================*/
/*= IV. Operationen zum Matchen mit Regel- sowie Gleichungsbaum                                                   =*/
/*=================================================================================================================*/

#define /*void*/ BaumwurzelInitialisieren(/*DSBaumT*/ DSBaum, /*UTermT*/ Term)                                      \
  MO_IndirekterClashImBaum = FALSE;                                                                                 \
  BK_MSymbolSei(BaumIndex = DSBaum.Wurzel,BK_NULL);           /* erstmalig besuchten Knoten als solchen markieren */\
  BK_aktTIndexMSei(BaumIndex,Term)                                        /* Anfrageterm in Wurzelknoten schreiben */


#define AbstiegUmSymbol aktTermrest
#define AbstiegUmTeilterm TO_TermEnde(aktTermrest)
#define AbstiegVerwalten(/*SymbolT*/ Symbol, /*UTermT*/ VorNeuemTerm)                                               \
{                                                                                                                   \
  KantenT Nachf = BK_Nachf(BaumIndex,Symbol);                                                                       \
  BK_MSymbolSei(BaumIndex,Symbol);                                                                                  \
  BK_aktTIndexMSei(Nachf,TO_Schwanz(VorNeuemTerm));                   /* noch zu bearbeitenden Termrest eintragen */\
  BK_MVorgZelleSei(Nachf,VorNeuemTerm);                                                                             \
  BK_MSymbolSei(Nachf,BK_NULL);                                 /* Knoten mit erster Uebergangsfunktion markieren */\
  BaumIndex = Nachf;                                /* absteigen. Achtung: aktTermrest ist Makro ueber BaumIndex! */\
  return TRUE;                                                                                                      \
}


#define AbMitFunktionssymbol()                                                                                      \
  if (SO_SymbIstFkt(SymbolAusTerm) && BK_Nachf(BaumIndex,SymbolAusTerm)) {         /* Abstieg mit Funktionssymbol */\
    AbstiegVerwalten(SymbolAusTerm,AbstiegUmSymbol);                                                                \
  }


#define AbMitNeuerVariable()                                                                                        \
  if (BK_letzteVarBelegt(BaumIndex)) {                                                                              \
    SubstErgaenzen(aktTermrest,BK_MVorgZelle(BaumIndex));                                                           \
    AbstiegVerwalten(BK_letzteVar(BaumIndex),AbstiegUmTeilterm);                                                    \
  }


#define BindungAufheben()                                                                                           \
  SO_VorigesVarSymbD(xMletztes)


#define KeineBindungAufheben()


#define AbMitAlterVariable(/*SymbolT*/ ersteKandidatin, /*Statement*/ BindungAufhebenOderAuchNicht)                 \
{                                                                                                                   \
  SymbolT letzteKandidatin = min(SO_ErstesVarSymb,BK_maxSymb(BaumIndex)),                                           \
          VarIndex;                                                                                                 \
  BindungAufhebenOderAuchNicht;                                                                                     \
  SO_forSymboleAufwaerts(VarIndex,ersteKandidatin,letzteKandidatin)                                                 \
    if (BK_Nachf(BaumIndex,VarIndex) && EsFolgtSubstitutOderIC(VarIndex,aktTermrest)) {                             \
      AbstiegVerwalten(VarIndex,AbstiegUmTeilterm);                                                                 \
    }                                                                                                               \
}


static BOOLEAN EinenKnotenAbOderVieleKnotenAufgestiegen(void)
{ /* Entweder einen Knoten absteigen oder so lange hoch, bis es moeglich ist, in einen noch nicht besuchten Knoten */
  do {                                                    /* abzusteigen, dies dann tun, oder bis ueber die Wurzel */
    SymbolT MSymbol = BK_MSymbol(BaumIndex);
    if (SO_SymbGleich(MSymbol,BK_NULL)) {                                         /* Knoten wird erstmalig besucht */
      AbMitFunktionssymbol();
      AbMitNeuerVariable();
      AbMitAlterVariable(BK_minSymb(BaumIndex),KeineBindungAufheben());
    }
    else if (SO_SymbIstFkt(MSymbol)) {                            /* Wiederbesuch nach Abstieg mit Funktionssymbol */
      AbMitNeuerVariable();
      AbMitAlterVariable(BK_minSymb(BaumIndex),KeineBindungAufheben());
    }
    else if (SO_SymbGleich(MSymbol,BK_letzteVar(BaumIndex))) {  /* ... nach Abstieg mit gerade gebundener Variable */
      AbMitAlterVariable(SO_VorigesVarSymb(MSymbol),BindungAufheben());
    }
    else {                                               /* ... nach Abstieg mit schon frueher gebundener Variable */
      AbMitAlterVariable(SO_VorigesVarSymb(MSymbol),KeineBindungAufheben());
    }
  } while ((BaumIndex = BK_Vorg(BaumIndex)));
  return FALSE;
}         

#if TP_PETERGRAF
BOOLEAN MO_RegelGefunden(UTermT Term)     /* Sucht im Regelbaum nach einer Generalisierung von Term. SubstAnwenden
  liefert dann diezugehoerige substituierte rechte Seite der ersten Regel in der zum Blatt gehoerenden Regelliste.
  Suchstrategie: Tiefensuche. Besuchsreihenfolge fuer Unterbaeume: zu passendem Funktionssymbol, zu noch nicht ge- */
{             /* bundener Variable, zu schon gebundenen Variablen, beginnend mit den davon neueren, endend bei x1. */
  SymbolT Topsymbol = TO_TopSymbol(Term);

  if (SO_SymbIstFkt(Topsymbol)) {
    RegelT RegelIndex = (GNTermpaarT) BK_Nachf(RE_Regelbaum.Wurzel,Topsymbol);
    while (RegelIndex) {
      if (MO_RegelPasst(RegelIndex,Term)){
        return TRUE;
      }
      RegelIndex = TP_Nachf(RegelIndex);
    }
  }
  return BK_Nachf(RE_Regelbaum.Wurzel,SO_ErstesVarSymb)                          /* exotisch, exotisch: Regel x->t */
    && MO_RegelPasst((GNTermpaarT) BK_Nachf(RE_Regelbaum.Wurzel,SO_ErstesVarSymb),Term);
}
#else


#if MO_TAUSEND_TERME
#define /*void*/ GgfProtokollieren(/*UTermT*/ Term)                                                                 \
  fprintf(TausendTermeDatei,"\t\t");                                                                                \
  IO_DruckeTermStrom(TausendTermeDatei,Term);                                                                       \
  fprintf(TausendTermeDatei," = x1\n");                                                                             \
  if (!--TausendTermeZaehler) {                                                                                     \
    if (fclose(TausendTermeDatei))                                                                                  \
      our_error("TausendTermeDatei kann nicht geschlossen werden.");                                                \
    fprintf(IO_Strom, "\n\nDie Protokollierung ist abgeschlossen.\n\n");                                            \
    RE_DruckeRegelbaum();                                                                                           \
    exit(0);                                                                                                        \
  }
#else
#define /*void*/ GgfProtokollieren(/*UTermT*/ Term)
#endif

/* ******************************************************************************************************************* */
/* ******************************************************************************************************************* */
/* ******************************************************************************************************************* */

#define DONT_MONSTER 1
#define MONSTER(TRUE_BODY)

#define SubstErgaenzenM(/*BOOLEAN*/DO_EQNS,/*UTermT*/ SubstitutT, /*UTermT*/ VorgT)	\
{									\
  SO_NaechstesVarSymbD(lxMletztes);					\
  xM[lxMletztes].Substitut = SubstitutT;/*Opt bringt nichts */		\
  /*xM[lxMletztes].Nachf = TO_NaechsterTeilterm(SubstitutT);*/		\
  /*xM[lxMletztes].SchonEingebaut = FALSE;*/				\
  xM[lxMletztes].Vorg = VorgT;						\
  /* nicht noetig bei Regeln: */					\
  if (DO_EQNS && !(VorgT))							\
      NichtGanzerTermGebunden = FALSE;					\
}

#if 1
/* Bactracking im Baum */
#define DeclareBacktrack()			\
  KantenT BackTrIndex;  

#define InitBacktrack()				\
  BackTrIndex = NULL;

#define BacktrackPush(MSymbol)			\
{						\
  BK_backtrKnotenSei(BaumIndex,BackTrIndex);	\
  BK_backtrSubSei(BaumIndex,lxMletztes);	\
  BK_MVorgZelleSei(BaumIndex,Vorg);		\
  BK_MSymbolSei(BaumIndex,MSymbol);		\
  BackTrIndex = BaumIndex;			\
}

#define BacktrackPop()											\
{													\
  BaumIndex   = BackTrIndex;										\
  BackTrIndex = BK_backtrKnoten(BaumIndex);								\
  lxMletztes  = BK_backtrSub(BaumIndex);								\
  Symbol = BK_MSymbol(BaumIndex);									\
  Vorg = BK_MVorgZelle(BaumIndex);									\
  Termrest = (Vorg) ? TO_Schwanz(Vorg) : Term /* Problem: Backtrack to root-node -> Vorg == NULL!!*/;	\
}

#define canBacktrack()				\
  (BackTrIndex != NULL)

#else
/* Backtracking in extra Stack */
typedef struct {
  SymbolT MSymbol;
  SymbolT xMletztes;
  KantenT BaumIndex;
  UTermT  Vorg;
} BacktrackT;

static BacktrackT bt[1000];

#define DeclareBacktrack()			\
  BacktrackT *BackTrIndex;  

#define InitBacktrack()				\
  BackTrIndex = bt - 1;

#define BacktrackPush(Symbol)				\
{						\
  BackTrIndex++;				\
  BackTrIndex->MSymbol   = Symbol;		\
  BackTrIndex->xMletztes = lxMletztes;		\
  BackTrIndex->BaumIndex = BaumIndex;		\
  BackTrIndex->Vorg      = Vorg;		\
}

#define BacktrackPop()											\
{													\
  Symbol      = BackTrIndex->MSymbol;									\
  lxMletztes  = BackTrIndex->xMletztes;									\
  BaumIndex   = BackTrIndex->BaumIndex;									\
  Vorg        = BackTrIndex->Vorg;									\
  Termrest = (Vorg) ? TO_Schwanz(Vorg) : Term /* Problem: Backtrack to root-node -> Vorg == NULL!!*/;	\
  BackTrIndex--;											\
}

#define canBacktrack()				\
  (BackTrIndex >= bt)

#endif



#define AbstiegUmSymbolM Termrest
#define AbstiegUmTeiltermM TO_TermEnde(Termrest)

#define AbstiegVerwaltenM(/*UTermT*/ VorNeuemTerm)		\
{								\
  Vorg = VorNeuemTerm;						\
  Termrest = TO_Schwanz(Vorg);					\
  BaumIndex = BK_Nachf(BaumIndex,Symbol);			\
  if (BK_IstBlatt(BaumIndex)) goto leave; else goto tryFct;	\
}

#define SubstInOrdnungBringen()										   \
{													   \
    letzteZelleAnfrageterm = ((Anfrageterm = Term)->Ende);						   \
    NichtGanzerTermGebunden = TRUE;									   \
    xMletztes = lxMletztes;										   \
    for (Symbol = SO_ErstesVarSymb; Symbol >= lxMletztes; Symbol--){					   \
/*	xM[Symbol].Substitut = TO_Schwanz(xM[Symbol].Vorg); / *schlecht wg. Test fuer bereits geb. Vars */ \
	xM[Symbol].Nachf = TO_NaechsterTeilterm(xM[Symbol].Substitut);					   \
	xM[Symbol].SchonEingebaut = FALSE;								   \
    }													   \
}

#if 0

/* Hier muss Symbol++ auskommentiert werden!! */

static SymbolT calcLetzteKandidatin(KantenT BaumIndex)
{
  SymbolT lk;
  for (lk = -1; lk > BK_minSymb(BaumIndex); lk --){
    if (BK_Nachf(BaumIndex,lk)){
      return lk;
    }
  }
  return lk;
  /*  return min(SO_ErstesVarSymb,BK_maxSymb(BaumIndex));*/
}
static BOOLEAN esGibtNochAlteVarMitBindung(KantenT BaumIndex, SymbolT Symbol, UTermT Termrest)
{
  SymbolT lk;
  for (lk = -1; lk > Symbol; lk --){
    if (BK_Nachf(BaumIndex,lk) && EsFolgtSubstitutOderIC/*Von*/(lk,Termrest)){
      return TRUE;
    }
  }
  return FALSE;
}
static SymbolT naechsteAlteVarMitBindung(KantenT BaumIndex, SymbolT Symbol, UTermT Termrest)
{
  SymbolT lk;
  for (lk = Symbol+1; lk < 0; lk++){
    if (BK_Nachf(BaumIndex,lk) && EsFolgtSubstitutOderIC(lk,Termrest)){
      return lk;
    }
  }
  return 0;
}

#define Backtrack1()								\
{										\
    if (BK_Nachf(BaumIndex,BK_letzteVar(BaumIndex))){				\
	BacktrackPush(Symbol);							\
    }										\
    else {									\
	SymbolT x = naechsteAlteVarMitBindung(BaumIndex,Symbol,Termrest);	\
	if (x) BacktrackPush(x);						\
    }										\
}

#define Backtrack2()							\
{									\
    SymbolT x = naechsteAlteVarMitBindung(BaumIndex,Symbol,Termrest);	\
    if (x) BacktrackPush(x);						\
}

#define Backtrack3()							\
{									\
    SymbolT x = naechsteAlteVarMitBindung(BaumIndex,Symbol,Termrest);	\
    if (x) BacktrackPush(x);						\
}

#else


#define BacktrackP1() /* SO_SymbIstVar(BK_minSymb(BaumIndex))*/			\
 (BK_Nachf(BaumIndex,BK_letzteVar(BaumIndex)) ||				\
  esGibtNochAlteVarMitBindung(BaumIndex,BK_letzteVar(BaumIndex),Termrest))

#define BacktrackP2() /* BK_maxSymb(BaumIndex)*/ /*calcLetzteKandidatin(BaumIndex) > Symbol)*/	\
  (Symbol != SO_ErstesVarSymb) && 								\
    esGibtNochAlteVarMitBindung(BaumIndex,Symbol,Termrest)

#define BacktrackP3() /*letzteKandidatin > Symbol*/		\
  esGibtNochAlteVarMitBindung(BaumIndex,Symbol,Termrest)

#define Backtrack1()							\
{									\
    if (SO_SymbIstVar(BK_minSymb(BaumIndex))) BacktrackPush(Symbol);	\
}

#define Backtrack2()											\
{													\
    if ((Symbol != SO_ErstesVarSymb) && (BK_maxSymb(BaumIndex) > Symbol)) BacktrackPush(Symbol);	\
}

#define Backtrack3()						\
{								\
    if (letzteKandidatin > Symbol) BacktrackPush(Symbol);	\
}

static SymbolT calcLetzteKandidatin(KantenT BaumIndex)
{
    /**/
  SymbolT lk;
  for (lk = -1; lk > BK_minSymb(BaumIndex); lk --){
    if (BK_Nachf(BaumIndex,lk)){
      return lk;
    }
  }
  return lk;
  /*  

      return min(SO_ErstesVarSymb,BK_maxSymb(BaumIndex));*/
}
#endif
BOOLEAN MO_RegelGefunden(UTermT Term) 
{
  SymbolT Symbol,lxMletztes,letzteKandidatin;
  KantenT BaumIndex;
  UTermT Termrest, Regelrest, Vorg;
  DeclareBacktrack();

  /* Initialisierung der globalen Variablen */
  MO_IndirekterClashImBaum = FALSE;

  /* Initialisierung der lokalen Variablen */
  BaumIndex = RE_Regelbaum.Wurzel;
  Termrest = Term;
  Vorg = NULL;
  lxMletztes = SO_VorigesVarSymb(SO_ErstesVarSymb);
  InitBacktrack();

  tryFct:
      Symbol = TO_TopSymbol(Termrest);
      if (SO_SymbIstFkt(Symbol) && 
	  BK_Nachf(BaumIndex,Symbol)){
	  Backtrack1();
	  AbstiegVerwaltenM(AbstiegUmSymbolM);
      }
  tryNVar:
      Symbol = BK_letzteVar(BaumIndex);
      if (BK_Nachf(BaumIndex,Symbol)) {
	  Backtrack2();
	  SubstErgaenzenM(FALSE,Termrest,Vorg);
	  AbstiegVerwaltenM(AbstiegUmTeiltermM);
      }
/*tryOVar:*/
      Symbol = BK_minSymb(BaumIndex); /* Optimierung: Wir wissen, dass es keine neue Var gibt */
  tryOVar1:
      letzteKandidatin = calcLetzteKandidatin(BaumIndex);
      while (Symbol <= letzteKandidatin){
	  if (BK_Nachf(BaumIndex,Symbol) && 
	      EsFolgtSubstitutOderIC(Symbol,Termrest)) {
	      Backtrack3();
	      AbstiegVerwaltenM(AbstiegUmTeiltermM);
	  }
	  Symbol++;
      }
  backtrack:
      if (!canBacktrack()) 
	  return FALSE;
      BacktrackPop();
      if (SO_SymbIstFkt(Symbol))
        goto tryNVar;
      Symbol++;
      goto tryOVar1;
  leave:
      Regelrest = BK_Rest(BaumIndex);

      while (Regelrest) {                              /* anstatt Abfrage auf Ende des (unbeschnittenen!) Anfrageterms */
	  SymbolT SymbAusRegel = TO_TopSymbol(Regelrest);
	  if (SO_SymbIstVar(SymbAusRegel)) {
	      if (SO_VariableIstNeuer(SymbAusRegel,lxMletztes)) {                     /* ungebundene Variable aus Regelterm */
		  SubstErgaenzenM(FALSE,Termrest,Vorg);                                                    /* in Anfrageterm binden */
	      }
	      else if (!EsFolgtSubstitutOderIC(SymbAusRegel,Termrest)) {                /* fuer gebundene Variable pruefen, */
		  goto backtrack;                              /* ob sich der Anfragterm entsprechend fortsetzt oder eben nicht*/
	      }
	      Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));                           /* einen Teilterm weitergehen */
	      Regelrest = TO_Schwanz(Regelrest);
	  }
	  else if (SO_SymbUngleich(SymbAusRegel,TO_TopSymbol(Termrest))) {              /* Abbruch ebenso bei ungleichen */
	      goto backtrack;
	  }
	  else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
	      Termrest = TO_Schwanz(Vorg = Termrest);                                                 /* ein Symbol weiter */
	      Regelrest = TO_Schwanz(Regelrest);
	  }
      }
/*success:*/
      SubstInOrdnungBringen();

      gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
      gefundeneRechteSeite = (if_MO_Objekt(MO_AngewendetesObjekt = )BK_Regeln(BaumIndex))->rechteSeite;
      return TRUE;
}

#endif


#if COMP_ACIDWM
dt_QUERY MO_RegelbaumDurchwandernAcidWM(UTermT Term)
{
  dt_RESULT Ergebnisse = NULL;
  dt_QUERY Ergebnis = NULL;
  MatcherInitialisieren(Term);  
  BaumwurzelInitialisieren(RE_Regelbaum,Term);
  
  while (EinenKnotenAbOderVieleKnotenAufgestiegen())
    if (BK_IstBlatt(BaumIndex)) {
      if (RegelInBlattGefunden()) {
        RegelT RegelIndex = BK_Regeln(BaumIndex);
        do {
          dt_RESULT NeuesErgebnis = (dt_RESULT) memory_Malloc(sizeof(dt_RESULT_NODE));
          NeuesErgebnis->pointer = (POINTER) TP_LinkeSeite(RegelIndex);
          NeuesErgebnis->position = NULL;
          NeuesErgebnis->next = Ergebnisse;
          Ergebnisse = NeuesErgebnis;
        } while ((RegelIndex = TP_Nachf(RegelIndex)));
        MatcherNachBlattZuruecksetzen();        
      }
      BaumIndex = BK_Vorg(BaumIndex);
    }

  if (Ergebnisse)
    (Ergebnis = memory_Malloc(sizeof(dt_QUERY_NODE)))->result = Ergebnisse;
  return Ergebnis;
}
#endif


static TermT SigmaRGanzNeu(void);

#if TP_PETERGRAF

static BOOLEAN GleichungPasstVonLinksNachRechts(GleichungsT Gleichung, UTermT Term);

BOOLEAN MO_GleichungGefunden(UTermT Term) /* Sucht im Regelbaum nach einer Generalisierung von Term. SubstAnwenden
  liefert dann diezugehoerige substituierte rechte Seite der ersten Regel in der zum Blatt gehoerenden Regelliste.
  Suchstrategie: Tiefensuche. Besuchsreihenfolge fuer Unterbaeume: zu passendem Funktionssymbol, zu noch nicht ge- */
{             /* bundener Variable, zu schon gebundenen Variablen, beginnend mit den davon neueren, endend bei x1. */
  SymbolT Topsymbol = TO_TopSymbol(Term);
  if (SO_SymbIstFkt(Topsymbol)) {
    GleichungsT GleichungsIndex = (GNTermpaarT) BK_Nachf(RE_Gleichungsbaum.Wurzel,Topsymbol);
    while (GleichungsIndex) {
      if (GleichungPasstVonLinksNachRechts(GleichungsIndex,Term))
        return TRUE;                                   /* Ende der Zeitmessung in GleichungPasstVonLinksNachRechts */
      GleichungsIndex = TP_Nachf(GleichungsIndex);
    }
  }
  if (BK_Nachf(RE_Gleichungsbaum.Wurzel,SO_ErstesVarSymb)) {                     /* exotisch, exotisch: Regel x->t */
    GleichungsT GleichungsIndex = (GNTermpaarT) BK_Nachf(RE_Gleichungsbaum.Wurzel,SO_ErstesVarSymb);
    while (GleichungsIndex) {
      if (GleichungPasstVonLinksNachRechts(GleichungsIndex,Term))
        return TRUE;                                   /* Ende der Zeitmessung in GleichungPasstVonLinksNachRechts */
      GleichungsIndex = TP_Nachf(GleichungsIndex);
    }
  }
  return FALSE;
}
#else


BOOLEAN MO_GleichungGefunden(UTermT Term) /* Sucht im Gleichungsbaum nach einer Generalisierung von Term. SubstAn-
  wenden liefert dann diezugehoerige substituierte rechte Seite der ersten Regel in der zum Blatt gehoerenden Re-
  gelliste. Suchstrategie: Tiefensuche. Besuchsreihenfolge fuer Unterbaeume: zu passendem Funktionssymbol, zu noch */
{    /*  nicht gebundener Variable, zu schon gebundenen Variablen, beginnend mit den davon neueren, endend bei x1. */
#if 0
  MatcherInitialisieren(Term);
  BaumwurzelInitialisieren(RE_Gleichungsbaum,Term);
  while (EinenKnotenAbOderVieleKnotenAufgestiegen())
    if (BK_IstBlatt(BaumIndex)) {
      if (RegelInBlattGefunden()) {
        GleichungsT Index = BK_Regeln(BaumIndex);
        gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
        do {
          TermT Kopie;
          BOOLEAN Resultat;

          Kopie = (gefundeneRechteSeite = TP_RechteSeite(Index),                         /* umsetzen, da Gleichung */
                   SigmaRGanzNeu());                                                  /* mit anderer rechter Seite */
         Resultat = ORD_TermGroesserRed(Term,Kopie);                /* Umschaltung auf Ordnungszeitmessung in HK_> */
          TO_TermLoeschen(Kopie);
          if (Resultat) {
            if_MO_Objekt(MO_AngewendetesObjekt = Index);
            return TRUE;
          }
          IncAnzRedMisserfolgeOrdnung();            /* Red.versuch wg. >-Test gescheitert; nicht im else-Zweig, da */
          MO_IndirekterClashImBaum = TRUE;
        } while ((Index = Index->Nachf));
        MatcherNachBlattZuruecksetzen();
      }
      BaumIndex = BK_Vorg(BaumIndex);
    }
  return FALSE;
#else
  SymbolT Symbol,lxMletztes,letzteKandidatin;
  KantenT BaumIndex;
  UTermT Termrest, Regelrest, Vorg;
  DeclareBacktrack();

  /* Initialisierung der globalen Variablen */
  MO_IndirekterClashImBaum = FALSE;

  /* Initialisierung der lokalen Variablen */
  BaumIndex = RE_Gleichungsbaum.Wurzel;
  Termrest = Term;
  Vorg = NULL;
  lxMletztes = SO_VorigesVarSymb(SO_ErstesVarSymb);
  InitBacktrack();

  tryFct:
      Symbol = TO_TopSymbol(Termrest);
      if (SO_SymbIstFkt(Symbol) && 
	  BK_Nachf(BaumIndex,Symbol)){
	  Backtrack1();
	  AbstiegVerwaltenM(AbstiegUmSymbolM);
      }
  tryNVar:
      Symbol = BK_letzteVar(BaumIndex);
      if (BK_Nachf(BaumIndex,Symbol)) {
	  Backtrack2();
	  SubstErgaenzenM(TRUE,Termrest,Vorg);
	  AbstiegVerwaltenM(AbstiegUmTeiltermM);
      }
/*tryOVar:*/
      Symbol = BK_minSymb(BaumIndex); /* Optimierung: Wir wissen, dass es keine neue Var gibt */
  tryOVar1:
      letzteKandidatin = calcLetzteKandidatin(BaumIndex);
      while (Symbol <= letzteKandidatin){
	  if (BK_Nachf(BaumIndex,Symbol) && 
	      EsFolgtSubstitutOderIC(Symbol,Termrest)) {
	      Backtrack3();
	      AbstiegVerwaltenM(AbstiegUmTeiltermM);
	  }
	  Symbol++;
      }
  backtrack:
      if (!canBacktrack()) 
	  return FALSE;
      BacktrackPop();
      if (SO_SymbIstFkt(Symbol))
        goto tryNVar;
      Symbol++;
      goto tryOVar1;
  leave:
      Regelrest = BK_Rest(BaumIndex);

      while (Regelrest) {                              /* anstatt Abfrage auf Ende des (unbeschnittenen!) Anfrageterms */
	  SymbolT SymbAusRegel = TO_TopSymbol(Regelrest);
	  if (SO_SymbIstVar(SymbAusRegel)) {
	      if (SO_VariableIstNeuer(SymbAusRegel,lxMletztes)) {                     /* ungebundene Variable aus Regelterm */
		  SubstErgaenzenM(TRUE,Termrest,Vorg);                                                    /* in Anfrageterm binden */
	      }
	      else if (!EsFolgtSubstitutOderIC(SymbAusRegel,Termrest)) {                /* fuer gebundene Variable pruefen, */
		  goto backtrack;                              /* ob sich der Anfragterm entsprechend fortsetzt oder eben nicht*/
	      }
	      Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));                           /* einen Teilterm weitergehen */
	      Regelrest = TO_Schwanz(Regelrest);
	  }
	  else if (SO_SymbUngleich(SymbAusRegel,TO_TopSymbol(Termrest))) {              /* Abbruch ebenso bei ungleichen */
	      goto backtrack;
	  }
	  else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
	      Termrest = TO_Schwanz(Vorg = Termrest);                                                 /* ein Symbol weiter */
	      Regelrest = TO_Schwanz(Regelrest);
	  }
      }
/*success:*/
      SubstInOrdnungBringen();
      {
	  GleichungsT Index = BK_Regeln(BaumIndex);
	  gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
	  do {
	      TermT Kopie;
	      BOOLEAN Resultat;
	      
	      Kopie = (gefundeneRechteSeite = TP_RechteSeite(Index),                         /* umsetzen, da Gleichung */
		       SigmaRGanzNeu());                                                  /* mit anderer rechter Seite */
	      Resultat = ORD_TermGroesserRed(Term,Kopie);                /* Umschaltung auf Ordnungszeitmessung in HK_> */
	      TO_TermLoeschen(Kopie);
	      if (Resultat) {
		  if_MO_Objekt(MO_AngewendetesObjekt = Index);
		  return TRUE;
	      }
	      IncAnzRedMisserfolgeOrdnung();            /* Red.versuch wg. >-Test gescheitert; nicht im else-Zweig, da */
	      MO_IndirekterClashImBaum = TRUE;
	  } while ((Index = Index->Nachf));
      }
      goto backtrack;
#endif
}
#endif


/*=================================================================================================================*/
/*= V. Matching mit nur einer Regel oder Gleichung                                                                =*/
/*=================================================================================================================*/

#if TP_PETERGRAF                                          /* wg. Verwendung von MO_RegelPasst als Baumersatz */
#define EsFolgtSubstitutVon(VarSymbol,TermIndex) EsFolgtSubstitutOderIC(VarSymbol,TermIndex)
#endif

BOOLEAN MO_TermpaarPasst(TermT links, TermT rechts, UTermT Termrest)
/* Liefert TRUE, wenn das Termpaar links->rechts als Regel auf Term anwendbar ist. In diesem Falle ist die 
   entsprechende Substitution wie gehabt zum Einpassen bereit und kann mit den unten aufgefuehrten Routinen in
   den Gesamtterm eingebaut werden. Das Termpaar muss bereits normiert sein. */
{
  UTermT Vorg = NULL;

  gefundeneLinkeSeite = links;                        /* Nach erfolgreichem Schleifendurchlauf gilt links == NULL. */
  MatcherInitialisieren(Termrest);
  do {
    if (SO_SymbIstVar(TO_TopSymbol(links))) {
      if (VarUngebunden(TO_TopSymbol(links))) {                              /* ungebundene Variable aus Regelterm */
        SubstErgaenzen(Termrest,Vorg);                                                    /* in Anfrageterm binden */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));
        links = TO_Schwanz(links);
      }
      else if (EsFolgtSubstitutVon(TO_TopSymbol(links),Termrest)) {            /* fuer gebundene Variable pruefen, */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));                             /* ob sich der Anfragterm */
        links = TO_Schwanz(links);                                                       /* entsprechend fortsetzt */
      }
      else                                                             /* Falls nicht, den Matchversuch abbrechen. */
        return FALSE;
    }
    else if (SO_SymbUngleich(TO_TopSymbol(links),TO_TopSymbol(Termrest)))         /* Abbruch ebenso bei ungleichen */
      return FALSE;                                                                           /* Funktionssymbolen */
    else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
      Termrest = TO_Schwanz(Vorg = Termrest);
      links = TO_Schwanz(links);
    }
  } while (links);
  gefundeneRechteSeite = rechts;
  return TRUE;
}


BOOLEAN MO_SigmaTermIstGeneralisierung(UTermT links, UTermT Termrest)                                /* Fuer Joerg */
/* Liefert TRUE, wenn sigma(links) Verallgemeinerung von Termrest ist.
   Es sind dann aber keine neuen Terme unter dieser Substitution erzeugbar.
   Der Term links muss in Fortsetzung der bisherigen Anfragen normiert sein (insbesondere keine Variablen-Luecken!). 
   Die Substitution wird stets wieder zurueckgesetzt.
   */
{
  SymbolT xMletztesAlt = xMletztes;
  UTermT Vorg = NULL,
         NULLTerm = TO_NULLTerm(links);
  do {
    if (SO_SymbIstVar(TO_TopSymbol(links))) {
      if (VarUngebunden(TO_TopSymbol(links))) {                              /* ungebundene Variable aus Regelterm */
        SubstErgaenzen(Termrest,Vorg);                                                    /* in Anfrageterm binden */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));
        links = TO_Schwanz(links);
      }
      else if (EsFolgtSubstitutVon(TO_TopSymbol(links),Termrest)) {            /* fuer gebundene Variable pruefen, */
        Termrest = TO_Schwanz(Vorg = TO_TermEnde(Termrest));                             /* ob sich der Anfragterm */
        links = TO_Schwanz(links);                                                       /* entsprechend fortsetzt */
      }
      else {                                                           /* Falls nicht, den Matchversuch abbrechen. */
        xMletztes = xMletztesAlt;                                                    /* Substitution zuruecksetzen */
        return FALSE;
      }
    } else if (SO_SymbUngleich(TO_TopSymbol(links),TO_TopSymbol(Termrest))) {       /* Abbruch ebenso bei ungleichen */
      xMletztes = xMletztesAlt;                                                               /* Funktionssymbolen */
      return FALSE;                                                                  /* Substitution zuruecksetzen */
    }
    else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
      Termrest = TO_Schwanz(Vorg = Termrest);
      links = TO_Schwanz(links);
    }
  } while (TO_PhysischUngleich(links,NULLTerm));
  xMletztes = xMletztesAlt;                                                          /* Substitution zuruecksetzen */
  return TRUE;
}


BOOLEAN MO_TermIstGeneralisierung(UTermT links, UTermT Termrest)                                     /* Fuer Joerg */
/* Liefert TRUE, wenn links Verallgemeinerung von Termrest ist.
   Es sind dann aber keine neuen Terme unter der Substitution erzeugbar.
   Der Term links muss bereits normiert sein. 
   Die Substitution bleibt erhalten und kann fuer MO_SigmaTermIstGeneralisierung verwendet werden.
   */
{ 
  SubstAufIdentitaetSetzen();
  return MO_SigmaTermIstGeneralisierung(links,Termrest);
}


BOOLEAN MO_TermpaarPasstMitResultat(TermT links, TermT rechts, TermT TermOld, TermT *TermNew)
/* Liefert TRUE, wenn das Termpaar links=rechts auf Stelle aktTeiltermAH in
   TermOld auf Toplevel anwendbar ist und
   liefert in diesem Fall das Ergebnis den geaenderten Term als TermNew mit neuen
   Termzellen zurueck. Es wird nicht geprueft, ob TermOld > TermNew ist.
   Die Funktion ist also nicht destruktiv, TermNew muss freigegeben werden.
   links, rechts muss normiert sein!
   */
{
   if (MO_TermpaarPasst(links, rechts, aktTeiltermAH)) {
      *TermNew = MO_AllTermErsetztNeu(TermOld, aktTeiltermAH);
      return TRUE;
   }
   return FALSE;
}


#if TP_PETERGRAF
#undef EsFolgtSubstitutVon

static BOOLEAN GleichungPasstVonLinksNachRechts(GleichungsT Gleichung, UTermT Term)
/* Testet, ob die gegebene Gleichung den Term t reduzieren kann. Dieser Test geschieht 
   nur von links nach rechts und unter Beruecksichtigung der eingestellten Ordnung.*/
{
  if (MO_RegelPasst(Gleichung,Term)) {
    UTermT Kopie;
    BOOLEAN Resultat;

    Kopie = SigmaRGanzNeu();
    Resultat = ORD_TermGroesserRed(Term,Kopie);
    TO_TermLoeschen(Kopie);
    if (Resultat) {
      if_MO_Objekt(MO_AngewendetesObjekt = TP_Gegenrichtung(Gleichung));
      return TRUE;
    }
    IncAnzRedMisserfolgeOrdnung();                                        /* Red.versuch wg. >-Test gescheitert */
    MO_IndirekterClashImBaum = TRUE;
  }
  return FALSE;
}
#endif


BOOLEAN MO_GleichungPasst(GleichungsT Gleichung, UTermT Term)
/* Testet, ob die gegebene Gleichung den Term t reduzieren kann. Dieser Test geschieht 
   in beiden Richtungen und unter Beruecksichtigung der eingestellten Ordnung.*/
{
  if (MO_RegelPasst(Gleichung,Term)) {
    UTermT Kopie;
    BOOLEAN Resultat;

    Kopie = SigmaRGanzNeu();
    Resultat = ORD_TermGroesserRed(Term,Kopie);
    TO_TermLoeschen(Kopie);
    if (Resultat) {
      if_MO_Objekt(MO_AngewendetesObjekt = Gleichung);
      return TRUE;
    }
  }
  if (TP_IstKeineMonogleichung(Gleichung) && MO_RegelPasst(TP_Gegenrichtung(Gleichung),Term)) {
    UTermT Kopie;
    BOOLEAN Resultat;
    Kopie = SigmaRGanzNeu();
    Resultat = ORD_TermGroesserRed(Term,Kopie);
    TO_TermLoeschen(Kopie);
    if_MO_Objekt(MO_AngewendetesObjekt = TP_Gegenrichtung(Gleichung));
    return Resultat;
  }
  else
    return FALSE;
}



/*=================================================================================================================*/
/*= VI. Anwenden von Substitutionen auf Terme                                                                     =*/
/*=================================================================================================================*/

#define /*BOOLEAN*/ SchonEingebaut(/*SymbolT*/ VarSymbol)                                                           \
  xM[VarSymbol].SchonEingebaut


#define /*BOOLEAN*/ NochNichtEingebaut(/*SymbolT*/ VarSymbol)                                                       \
  !xM[VarSymbol].SchonEingebaut


#define /*BOOLEAN*/ SubstitutNichtAmAnfragetermEnde(/*SymbolT*/ VarSymbol)                                          \
  TO_PhysischUngleich(TO_TermEnde(SubstitutVon(VarSymbol)),letzteZelleAnfrageterm)


#define /*BOOLEAN*/ NichtGanzerTermRezykliert(/*void*/)                                                             \
  (NichtGanzerTermGebunden || NochNichtEingebaut(SO_ErstesVarSymb))


static void SubstitutAlsEingebautVermerken(SymbolT VarSymbol) /* setzt entsprechendes Flag und erhaelt die Verket- */
{                                               /* tung der nicht benoetigten Termzellen des Anfrageterms aufrecht */
/* Konzept:
   Komponenten Vorg und Nachf geben bei noch nicht eingebauten Variablen die jeweils erste Zelle im Anfrageterm
   vor bzw. nach dem Substitut an. Nachf zeigt schlimmstenfalls auf TO_Schwanz(letzteZelleAnfrageterm),
   Vorg auf Anfrageterm, da Aufruf nur bei mehrstelligen Anfragetermen erfolgt. */

  UTermT SubstitutVorg = xM[VarSymbol].Vorg;

  if (SubstitutVorg) {
    UTermT Substitut = SubstitutVon(VarSymbol),
           SubstitutNachf = xM[VarSymbol].Nachf;

    /* Aktualisierung der Zeiger auf erste vorhergehende nicht eingebaute Termzelle bei den noch nicht
       eingebauten Variablen. Betrifft max. eine Variable, deren Substitut das von VarIndex als Vorg hat. */
    SymbolT VarIndex;
    SO_forSymboleAbwaerts(VarIndex,SO_NaechstesVarSymb(VarSymbol),xMletztes)
      if (NochNichtEingebaut(VarIndex)) {
        if (TO_PhysischGleich(TO_TermEnde(Substitut),xM[VarIndex].Vorg))
          xM[VarIndex].Vorg = SubstitutVorg;
        break;
      }

    /* analog fuer Nachf-Zeiger */
    SO_forSymboleAufwaerts(VarIndex,SO_VorigesVarSymb(VarSymbol),SO_ErstesVarSymb)
      if (NochNichtEingebaut(VarIndex)) {
        if (TO_PhysischGleich(Substitut,xM[VarIndex].Nachf))
          xM[VarIndex].Nachf = SubstitutNachf;
        break;
      }

    /* Nun Verkettung der noch nicht rezyklierten Termzellen aktualisieren. In vorhergehender noch nicht
       eingebauter Zelle, die immer existiert (Topzelle!), den Nachfolger auf Kandidat setzen und ggf.
       letzteNichtVernutzteZelle aktualisieren. */
    if (TO_PhysischGleich(TO_TermEnde(Substitut),letzteNichtVernutzteZelle))
      letzteNichtVernutzteZelle = SubstitutVorg;
    TO_NachfBesetzen(SubstitutVorg,SubstitutNachf);
  }
  xM[VarSymbol].SchonEingebaut = TRUE;
}


#define /*UTermT*/ EVSubstitutVon(/*SymbolT*/ Symbol)                                                               \
  (SubstitutAlsEingebautVermerken(Symbol),SubstitutVon(Symbol))


#define /*UTermT*/ AufAngehaengteZelle(/*UTermT*/ Index, /*SymbolT*/ Symbol)                                        \
{                                                                                                                   \
  UTermT NeueZelle;                                                                                                 \
  TO_TermzelleHolen(NeueZelle);                                                                                     \
  TO_NachfBesetzen(Index,NeueZelle);                                                                                \
  TO_SymbolBesetzen(Index = NeueZelle,Symbol);                                                                      \
}


#define /*UTermT*/ AufAngehaengtesNullstelliges(/*UTermT*/ Index, /*SymbolT*/ Symbol)                               \
{                                                                                                                   \
  AufAngehaengteZelle(Index,Symbol);                                                                                \
  TO_EndeBesetzen(Index,Index);                                                                                     \
}


#define /*UTermT*/ AufAngehaengtesSubstitut(/*UTermT*/ Index, /*UTermT*/ Substitut)                                 \
{                                                                                                                   \
  UTermT NeuerTeilterm = Substitut;                                                                                 \
  TO_NachfBesetzen(Index,NeuerTeilterm);                                                                            \
  Index = TO_TermEnde(NeuerTeilterm);                                                                               \
}



/*=================================================================================================================*/
/*= VII. Erzeugung von SigmaR beginnend mit der ersten Zelle des Anfrageterms                                     =*/
/*=================================================================================================================*/

void MO_SigmaRInEZ(void)       /* SigmaR beliebiger Stelligkeit beginnend mit erster Zelle des Anfrageterms erzeu- */
{                                                         /* gen. Fallunterscheidung nach Laenge des Anfrageterms. */
  if (TO_Nullstellig(Anfrageterm)) { /* Ist der Anfrageterm nullstellig, muessen keine Termzellen geloscht werden. */
    if (TO_Nullstellig(gefundeneRechteSeite)) { /* Ist die rechte Seite eine Konstante oder eine ungebundene Vari- */
      if (SymbKeineGebundeneVar(TopsymbolRechteSeite))                                    /* able, so muss nur das */
        TO_SymbolBesetzen(Anfrageterm,TopsymbolRechteSeite);   /* entsprechende Symbol uebertragen werden. Ist die
           rechte Seite eine gebundene Variable, also x1, so geht die Bindung in den Anfrageterm. Dieser ist null-
           stellig; also geht die Bindung an den ganzen Term. Beim Substituieren wuerde der Anfrageterm durch sich
                selbst ersetzt werden; es muss also nichts unternommen werden. Ein exotischer Fall ("Regel" x->x). */
    }  /* Ansonsten geht mit einer ebenfalls exotischen Regel eine Variable oder eine Konstante an einen Grundterm
        oder einen Term, in dem eine gebundene Variable, also x1, wieder vorkommt, oder eine ungebundene Variable.
       Wie oben erfolgt  Bindung hoechstens an nullstelligen Term. Unterscheiden, ob eine Variable gebunden wurde. */
    else if (VarGebunden(SO_ErstesVarSymb))        /* Dann Substitut merken, da es beim Kopieren der rechten Seite */
      TO_TermkopieEZx1Subst(gefundeneRechteSeite,Anfrageterm);
    else                                                                            /* Ansonsten sorglos kopieren. */
      TO_TermkopieEZ(gefundeneRechteSeite,Anfrageterm);
  } else                                                                 /* Falls der Anfrageterm mehrstellig ist: */
    if (TO_Nullstellig(gefundeneRechteSeite)) {      /* Vorgehen fuer nullstellige rechte Seite unterscheiden nach */
      if (SymbKeineGebundeneVar(TopsymbolRechteSeite)) {                                  /* Konstanten und freien */
        TO_SymbolBesetzen(Anfrageterm,TopsymbolRechteSeite);            /* Variablen (dann ein Symbol uebertragen, */
        TO_TermzellenlisteLoeschen(TO_Schwanz(Anfrageterm),TO_TermEnde(Anfrageterm));
        TO_NachfBesetzen(Anfrageterm,NULL);                                      /* Rest des Anfrageterms loeschen */
        TO_EndeBesetzen(Anfrageterm,Anfrageterm);                             /* und Zeiger in Termzelle anpassen) */
      }
      else { /* sowie gebundenen Variablen; ausfuehrliche Erlaeuterung siehe oben. Erste Termzelle vom Anfrageterm
          mit TO_Schwanz(Substitut) versehen, dabei restliche Zellen des Anfrageterms verkettet lassen. Exotischer 
                Fall, wenn eine Variable an ganzen Term gebunden ist, also Regel x->gebundene Variable, also x->x. */
        if (NichtGanzerTermGebunden) {                                                 /* Dann nichts unternehmen. */
          UTermT Substitut = SubstitutVon(TopsymbolRechteSeite);
          TO_SymbolBesetzen(Anfrageterm,TO_TopSymbol(Substitut));     /* Topsymbol in erste Zelle des Anfrageterms */
          if (TO_Nullstellig(Substitut)) {  /* Falls Substitut nullstellig ist, den gesamten Rest des Anfrageterms */
            TO_TermzellenlisteLoeschen(TO_Schwanz(Anfrageterm), TO_TermEnde(Anfrageterm));                  /* weg */
            TO_NachfBesetzen(Anfrageterm,NULL);                                         /* und Termzeiger anpassen */
            TO_EndeBesetzen(Anfrageterm,Anfrageterm);
          }
          else {   /* Ansonsten TO_Schwanz(Substitut) bis TO_TermEnde(Substitut) aus der Verkettung der Termzellen */
            UTermT SchwanzAnfrageterm = TO_Schwanz(Anfrageterm);                    /* im Anfrageterm herausnehmen */
            TO_NachfBesetzen(Anfrageterm,TO_Schwanz(Substitut));
            /* Fallunterscheidung, ob Substitut am Ende des Anfrageterms steht */
            if (SubstitutNichtAmAnfragetermEnde(TopsymbolRechteSeite)) {                            /* Falls nein: */
              TO_NachfBesetzen(Substitut,TO_NaechsterTeilterm(Substitut));
              TO_TermzellenlisteLoeschen(SchwanzAnfrageterm,letzteZelleAnfrageterm);
              TO_EndeBesetzen(Anfrageterm,TO_TermEnde(Substitut));                                /*$nested-feature*/
              TO_NachfBesetzen(TO_TermEnde(Anfrageterm),NULL);       /* Da Substitut mitten im Term stand, gibt es */
            }                                                            /* eine nicht mehr gueltige Nachf-Angabe. */
            else
              TO_TermzellenlisteLoeschen(SchwanzAnfrageterm,Substitut);
          }
        }
      }
    }
  else {
    TermzellenT DummyTopzelle;
    UTermT IndexNeu = &DummyTopzelle,
           NULLAnfrageterm = TO_Schwanz(letzteZelleAnfrageterm);
    letzteNichtVernutzteZelle = letzteZelleAnfrageterm;
    TO_SymbolBesetzen(Anfrageterm,TopsymbolRechteSeite);
    TI_PushMitAnz(IndexNeu,TO_AnzahlTeilterme(gefundeneRechteSeite));
    do {
      SymbolT TopSymbol = TO_TopSymbol(gefundeneRechteSeite = TO_Schwanz(gefundeneRechteSeite));
      if (TO_Nullstellig(gefundeneRechteSeite)) {
        if (SymbKeineGebundeneVar(TopSymbol))
          AufAngehaengtesNullstelliges(IndexNeu,TopSymbol)
        else
          AufAngehaengtesSubstitut(IndexNeu,SchonEingebaut(TopSymbol) ? 
            TO_Termkopie0(SubstitutVon(TopSymbol)) : EVSubstitutVon(TopSymbol));
        while (!(--TI_TopZaehler)) {
          TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
          if (TI_PopUndLeer()) {
            TO_EndeBesetzen(Anfrageterm,IndexNeu);
            TO_NachfBesetzen(IndexNeu,NULL);
            if (NichtGanzerTermRezykliert() && TO_PhysischUngleich(TO_Schwanz(Anfrageterm),NULLAnfrageterm))
              TO_TermzellenlisteLoeschen(TO_Schwanz(Anfrageterm),letzteNichtVernutzteZelle);
            TO_NachfBesetzen(Anfrageterm,DummyTopzelle.Nachf);
            return;
          }
        }
      }
      else {
        AufAngehaengteZelle(IndexNeu,TopSymbol);
        TI_Push(IndexNeu);
      }
    } while (TRUE);
  }
}



/*=================================================================================================================*/
/*= VIII. Erzeugung von SigmaR endend mit der letzten Zelle des Anfrageterms                                      =*/
/*=================================================================================================================*/

static UTermT VonEndeNachMitteEVSubstitutVon(SymbolT VarSymbol)
{
  UTermT NeuesEnde,
         Substitut = SubstitutVon(VarSymbol);
  TO_TermzelleHolen(NeuesEnde);
  SubstitutAlsEingebautVermerken(VarSymbol);                /* ausreichend, da am Ende des Anfrageterms befindlich */
  TO_NachfBesetzen(NeuesEnde,NULL);
  TO_EndeBesetzen(NeuesEnde,NeuesEnde);
  if (TO_Nullstellig(Substitut)) {
    TO_SymbolBesetzen(NeuesEnde,TO_TopSymbol(Substitut));
    return NeuesEnde;
  }
  else {
    UTermT AltesEnde = TO_TermEnde(Substitut),
           SubstitutVorg;
    TO_EndeBesetzen(Substitut,NeuesEnde);
    while (TO_PhysischUngleich(Substitut = TO_Schwanz(SubstitutVorg = Substitut),AltesEnde))
      if (TO_PhysischGleich(TO_TermEnde(Substitut),AltesEnde))
        TO_EndeBesetzen(Substitut,NeuesEnde);
    TO_NachfBesetzen(SubstitutVorg,NeuesEnde);
    TO_SymbolBesetzen(NeuesEnde,TO_TopSymbol(AltesEnde));
    return SubstitutVon(VarSymbol);
  }
}


static UTermT VonMitteNachEndeEVSubstitutVon(SymbolT VarSymbol)
{
  UTermT Substitut = SubstitutVon(VarSymbol);

/* Verkettung der nicht benoetigten Termzellen des Anfrageterms aufrechterhalten. Die letzte Zelle des Substituts 
   bleibt in der Verkettung stehen, daher bei folgenden Substituten nichts zu unternehmen, sondern nur bei voran-
   gehenden. LetzteNichtVernutzteZelle muss nicht aktualisiert werden. */
  xM[VarSymbol].SchonEingebaut = TRUE;
  if (TO_Nullstellig(Substitut)) {
    TO_SymbolBesetzen(letzteZelleAnfrageterm,TO_TopSymbol(Substitut));
    return letzteZelleAnfrageterm;
  }
  else {
    SymbolT VarIndex;  
    UTermT AltesEnde = TO_TermEnde(Substitut),
           SubstitutVorg;
    SO_forSymboleAufwaerts(VarIndex,SO_VorigesVarSymb(VarSymbol),SO_ErstesVarSymb)
      if (NochNichtEingebaut(VarIndex)) {
        if (TO_PhysischGleich(Substitut,xM[VarIndex].Nachf))
          xM[VarIndex].Nachf = AltesEnde;
        break;
      }
    TO_NachfBesetzen(xM[VarSymbol].Vorg/*==SubstitutVorg*/,AltesEnde);    
    TO_EndeBesetzen(Substitut,letzteZelleAnfrageterm);
    while (TO_PhysischUngleich(Substitut = TO_Schwanz(SubstitutVorg = Substitut),AltesEnde))
      if (TO_PhysischGleich(TO_TermEnde(Substitut),AltesEnde))
        TO_EndeBesetzen(Substitut,letzteZelleAnfrageterm);
    TO_NachfBesetzen(SubstitutVorg,letzteZelleAnfrageterm);
    TO_SymbolBesetzen(letzteZelleAnfrageterm,TO_TopSymbol(AltesEnde));
    return SubstitutVon(VarSymbol);
  }
}


#define /*UTermT*/ SubstAnwendenLZ(/*UTermT*/ InitialwertLetzteNichtVernutzteZelle,                                 \
                                   /*UTermT*/ RezykliertesSubstitut,                                                \
                                   /*Block*/ AktionMitLetzterZelle,                                                 \
                                   /*Block*/ KorrekturLetzteNichtVernutzteZelle)                                    \
{                                                                                                                   \
  TermT TermNeu;                                                                                                    \
  UTermT IndexNeu;                                                                                                  \
  TO_TermzelleHolen(TermNeu);                                                                                       \
  IndexNeu = TermNeu;                                                                                               \
  letzteNichtVernutzteZelle = InitialwertLetzteNichtVernutzteZelle;                                                 \
  TO_SymbolBesetzen(TermNeu,TopsymbolRechteSeite);                                                                  \
  TI_Push(TermNeu);                                                                                                 \
  while (TO_PhysischUngleich(gefundeneRechteSeite = TO_Schwanz(gefundeneRechteSeite),EndeRechteSeite)) {            \
    SymbolT TopSymbol = TO_TopSymbol(gefundeneRechteSeite);                                                         \
    if (TO_Nullstellig(gefundeneRechteSeite)) {                                                                     \
      if (SymbKeineGebundeneVar(TopSymbol))                                                                         \
        AufAngehaengtesNullstelliges(IndexNeu,TopSymbol)                                                            \
      else                                                                                                          \
        AufAngehaengtesSubstitut(IndexNeu,                                                                          \
          SchonEingebaut(TopSymbol) ? TO_Termkopie0(SubstitutVon(TopSymbol)) : (RezykliertesSubstitut));            \
      while (!(--TI_TopZaehler)) {                                                                                  \
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);                                                                   \
        TI_Pop();                                                                                                   \
      }                                                                                                             \
    }                                                                                                               \
    else {                                                                                                          \
      AufAngehaengteZelle(IndexNeu,TopSymbol);                                                                      \
      TI_Push(IndexNeu);                                                                                            \
    }                                                                                                               \
  }                                                                                                                 \
  AktionMitLetzterZelle;                                                                                            \
  do                                                                                                                \
    TO_EndeBesetzen(TI_TopTeilterm,letzteZelleAnfrageterm);                                                         \
  while (TI_PopUndNichtLeer());                                                                                     \
  if (NichtGanzerTermRezykliert()) {                                                                                \
    KorrekturLetzteNichtVernutzteZelle;                                                                             \
    TO_TermzellenlisteLoeschen(Anfrageterm,letzteNichtVernutzteZelle);                                              \
  }                                                                                                                 \
  return TermNeu;                                                                                                   \
}


UTermT MO_SigmaRInLZ(void)
{
  if (TO_Nullstellig(Anfrageterm)) {                                  /* dann Anfrageterm == letzteZelleAnfrageterm */
    if (TO_Nullstellig(gefundeneRechteSeite)) {
      if (SymbKeineGebundeneVar(TopsymbolRechteSeite))
        TO_SymbolBesetzen(Anfrageterm,TopsymbolRechteSeite);
      return Anfrageterm;
    }
    else if (VarGebunden(SO_ErstesVarSymb))
      return TO_TermkopieLZx1Subst(gefundeneRechteSeite,Anfrageterm);
    else                                                                            /* Ansonsten sorglos kopieren. */
      return TO_TermkopieLZm(gefundeneRechteSeite,Anfrageterm);
  }
  else {
    if (TO_Nullstellig(gefundeneRechteSeite)) {
      if (SymbKeineGebundeneVar(TopsymbolRechteSeite)) {                                  /* Konstanten und freien */
        TO_SymbolBesetzen(letzteZelleAnfrageterm,TopsymbolRechteSeite); /* Variablen (dann ein Symbol uebertragen, */
        TO_TermzellenlisteLoeschen(Anfrageterm,TO_VorigeTermzelle(letzteZelleAnfrageterm,Anfrageterm));
        return letzteZelleAnfrageterm;
      }
      else {                                                                        /* nun Regel f(...,x,...) -> x */
        UTermT Substitut = SubstitutVon(TopsymbolRechteSeite);
        if (SubstitutNichtAmAnfragetermEnde(TopsymbolRechteSeite)) {
          if (TO_Nullstellig(Substitut)) {        /* nullstelliges Substitut in letzteZelleAnfrageterm uebertragen */
            TO_SymbolBesetzen(letzteZelleAnfrageterm,TO_TopSymbol(Substitut));
            TO_TermzellenlisteLoeschen(Anfrageterm,TO_VorigeTermzelle(letzteZelleAnfrageterm,Anfrageterm));
            return letzteZelleAnfrageterm;
          }
          else {                                /* mehrstelliges Substitut mit letzteZelleAnfrageterm enden lassen */
            UTermT SubstitutVorg,
                   SubstitutEnde = TO_TermEnde(Substitut);
            TO_SymbolBesetzen(letzteZelleAnfrageterm,TO_TopSymbol(SubstitutEnde));
            TO_EndeBesetzen(Substitut,letzteZelleAnfrageterm);
            while (TO_PhysischUngleich(Substitut = TO_Schwanz(SubstitutVorg = Substitut),SubstitutEnde))
              if (TO_PhysischGleich(TO_TermEnde(Substitut),SubstitutEnde))
                TO_EndeBesetzen(Substitut,letzteZelleAnfrageterm);
            TO_NachfBesetzen(xM[TopsymbolRechteSeite].Vorg,SubstitutEnde);
            TO_TermzellenlisteLoeschen(Anfrageterm,TO_VorigeTermzelle(letzteZelleAnfrageterm,SubstitutEnde));
            TO_NachfBesetzen(SubstitutVorg,letzteZelleAnfrageterm);
            return SubstitutVon(TopsymbolRechteSeite);
          }
	}
        else {                                                   /* falls Substitut am Ende des Anfrageterms steht */
          if (NichtGanzerTermGebunden)
            TO_TermzellenlisteLoeschen(Anfrageterm,xM[TopsymbolRechteSeite].Vorg);
          return Substitut;
        }
      }
    }
    else {                                      /* Kasus Knackus: Anfrageterm und gefundeneRechteSeite mehrstellig */
      UTermT EndeRechteSeite = TO_TermEnde(gefundeneRechteSeite);
      SymbolT letztesSymbRechteSeite = TO_TopSymbol(EndeRechteSeite),
              letztesSymbLinkeSeite = TO_TopSymbol(TO_TermEnde(gefundeneLinkeSeite));

      #define /*Block*/ AnderenSchlussterm(/*void*/)                                                                \
        if (SymbKeineGebundeneVar(letztesSymbRechteSeite)) {                                                        \
          TO_SymbolBesetzen(letzteZelleAnfrageterm,letztesSymbRechteSeite);                                         \
          TO_NachfBesetzen(IndexNeu,letzteZelleAnfrageterm);                                                        \
        }                               /* 1 Am Ende der gefundenen rechten Seite steht keine gebundene Variable. */\
        else {                          /* 2 Am Ende der gefundenen rechten Seite steht eine gebundene Variable.  */\
          TO_NachfBesetzen(IndexNeu,                                                                                \
            SchonEingebaut(letztesSymbRechteSeite) ?                                                                \
              TO_TermkopieLZ(SubstitutVon(letztesSymbRechteSeite),letzteZelleAnfrageterm)                           \
            :                                                                                                       \
              VonMitteNachEndeEVSubstitutVon(letztesSymbRechteSeite));                                              \
        }
      
      #define /*void*/ DenselbenSchlussterm(/*void*/)                                                               \
       TO_NachfBesetzen(IndexNeu,EVSubstitutVon(letztesSymbRechteSeite))

      #define /*void*/ KeineKorrekturLNVZ(/*void*/)


      /* 1 Der Anfrageterm wird durch kein Substitut beendet. */

      if (SymbKeineGebundeneVar(letztesSymbLinkeSeite) || SubstitutNichtAmAnfragetermEnde(letztesSymbLinkeSeite))
        SubstAnwendenLZ(TO_VorletzteZelle(Anfrageterm), 
          EVSubstitutVon(TopSymbol), AnderenSchlussterm(), KeineKorrekturLNVZ())

      /* 2 Der Anfrageterm wird durch ein Substitut beendet. */
      else 

        /* 2.1 Der Anfrageterm wird durch dasselbe Substitut beendet wie der neue Term. */
        if (SO_SymbGleich(letztesSymbLinkeSeite,letztesSymbRechteSeite))
          SubstAnwendenLZ(letzteZelleAnfrageterm,
            SO_SymbGleich(TopSymbol,letztesSymbRechteSeite) ?
              TO_Termkopie0(SubstitutVon(TopSymbol)) : EVSubstitutVon(TopSymbol),
            DenselbenSchlussterm(), KeineKorrekturLNVZ())

        /* 2.2 Der Anfrageterm wird nicht durch dasselbe Substitut beendet wie der neue Term. */
        else
          SubstAnwendenLZ(xM[letztesSymbLinkeSeite].Vorg, 
            SO_SymbGleich(TopSymbol,letztesSymbLinkeSeite) ?
              VonEndeNachMitteEVSubstitutVon(TopSymbol) : EVSubstitutVon(TopSymbol),
            AnderenSchlussterm(),
            if (!SchonEingebaut(letztesSymbLinkeSeite) && TO_NichtNullstellig(SubstitutVon(letztesSymbLinkeSeite)))
              letzteNichtVernutzteZelle = TO_VorletzteZelle(SubstitutVon(letztesSymbLinkeSeite)));
    }
  }
}



/*=================================================================================================================*/
/*= IX. Erzeugung von SigmaR nur mit neuen Termzellen                                                              */
/*=================================================================================================================*/

#define /*UTermT*/ AufAngehaengtesSubstitutGNVon(/*UTermT*/ Index, /*SymbolT*/ VarSymbol)                           \
{                                                                                                                   \
  UTermT NeuerTeilterm = TO_Termkopie0(SubstitutVon(VarSymbol));                                                    \
  TO_NachfBesetzen(Index,NeuerTeilterm);                                                                            \
  Index = TO_TermEnde(NeuerTeilterm);                                                                               \
}


static TermT SigmaRGanzNeu(void)
{
  if (TO_Nullstellig(gefundeneRechteSeite)) {         /* Vorgehen fuer nullstellige rechte Seite unterscheiden nach */
    if (SymbKeineGebundeneVar(TopsymbolRechteSeite)) {                                    /* Konstanten und freien */
      TermT TermNeu;
      TO_TermzelleHolen(TermNeu);
      TO_SymbolBesetzen(TermNeu,TopsymbolRechteSeite);                  /* Variablen (dann ein Symbol uebertragen, */
      TO_NachfBesetzen(TermNeu,NULL);                                            /* Rest des Anfrageterms loeschen */
      TO_EndeBesetzen(TermNeu,TermNeu);                                       /* und Zeiger in Termzelle anpassen) */
      return TermNeu;
    }
    else
      return TO_Termkopie(SubstitutVon(TopsymbolRechteSeite));
  }
  else {
    TermT TermNeu,
          TermAlt = gefundeneRechteSeite;
    UTermT IndexNeu;
    TO_TermzelleHolen(TermNeu);
    IndexNeu = TermNeu;
    TO_SymbolBesetzen(TermNeu,TopsymbolRechteSeite);
    TI_Push(TermNeu);
    do {
      SymbolT TopSymbol = TO_TopSymbol(TermAlt = TO_Schwanz(TermAlt));
      if (TO_Nullstellig(TermAlt)) {
        if (SymbKeineGebundeneVar(TopSymbol))
          AufAngehaengtesNullstelliges(IndexNeu,TopSymbol)
        else
          AufAngehaengtesSubstitutGNVon(IndexNeu,TopSymbol);
        while (!(--TI_TopZaehler)) {
          TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
          if (TI_PopUndLeer()) {
            TO_NachfBesetzen(IndexNeu,NULL);
            return TermNeu;
          }
        }
      }
      else {
        AufAngehaengteZelle(IndexNeu,TopSymbol);
        TI_Push(IndexNeu);
      }
    } while (TRUE);
  }
}



/*=================================================================================================================*/
/*= X. Subsumptionstests                                                                                          =*/
/*=================================================================================================================*/

static void SubstErgaenzenSubsumption(UTermT Substitut)        /* fuegt zur Substitution x(n)<-t(n) ... x(1)<-t(1) */
{                                 /* die Bindung x(n+1)<-Substitut hinzu. Dabei wird aber kein Einbau vorbereitet! */
  SO_NaechstesVarSymbD(xMletztes);
  xM[xMletztes].Substitut = Substitut;
}


BOOLEAN MO_TermpaarSubsummiertZweites(TermT links1, TermT rechts1, UTermT links2, UTermT rechts2)
/* Liefert TRUE, wenn das Termpaar links1==rechts1 allgemeiner ist als das Termpaar links2==rechts2, dh. wenn
   eine Substitution @ existiert, so dass links1 == @(links2) und rechts1 == @(rechts2).
   Getestet wird dabei nur in dieser einen Richtung.
*/
{
  SubstAufIdentitaetSetzen();
  do {
    SymbolT Symbol1 = TO_TopSymbol(links1);
    if (SO_SymbIstVar(Symbol1)) {
      if (VarUngebunden(Symbol1)) {                                          /* ungebundene Variable aus Regelterm */
        SubstErgaenzenSubsumption(links2);                                                /* in Anfrageterm binden */
        links2 = TO_NaechsterTeilterm(links2);
        links1 = TO_Schwanz(links1);
      }
      else if (EsFolgtSubstitutVon(Symbol1,links2)) {                          /* fuer gebundene Variable pruefen, */
        links2 = TO_NaechsterTeilterm(links2);                                           /* ob sich der Anfragterm */
        links1 = TO_Schwanz(links1);                                                     /* entsprechend fortsetzt */
      }
      else                                                             /* Falls nicht, den Matchversuch abbrechen. */
        return FALSE;
    }
    else if (SO_SymbUngleich(Symbol1,TO_TopSymbol(links2)))                       /* Abbruch ebenso bei ungleichen */
      return FALSE;                                                                           /* Funktionssymbolen */
    else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
      links2 = TO_Schwanz(links2);
      links1 = TO_Schwanz(links1);
    }
  } while (links1);
  do {
    SymbolT Symbol1 = TO_TopSymbol(rechts1);
    if (SO_SymbIstVar(Symbol1)) {
      if (VarUngebunden(Symbol1)) {                                          /* ungebundene Variable aus Regelterm */
        SubstErgaenzenSubsumption(rechts2);                                               /* in Anfrageterm binden */
        rechts2 = TO_NaechsterTeilterm(rechts2);
        rechts1 = TO_Schwanz(rechts1);
      }
      else if (EsFolgtSubstitutVon(Symbol1,rechts2)) {                         /* fuer gebundene Variable pruefen, */
        rechts2 = TO_NaechsterTeilterm(rechts2);                                         /* ob sich der Anfragterm */
        rechts1 = TO_Schwanz(rechts1);                                                   /* entsprechend fortsetzt */
      }
      else                                                             /* Falls nicht, den Matchversuch abbrechen. */
        return FALSE;
    }
    else if (SO_SymbUngleich(Symbol1,TO_TopSymbol(rechts2)))                      /* Abbruch ebenso bei ungleichen */
      return FALSE;                                                                           /* Funktionssymbolen */
    else {                                               /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
      rechts2 = TO_Schwanz(rechts2);
      rechts1 = TO_Schwanz(rechts1);
    }
  } while (rechts1);

  return TRUE;
}


static BOOLEAN SubsummierendeGleichungInBlattGefunden(UTermT rechteSeite)
{
  SymbolT xMletztesAlt = xMletztes;
  UTermT Termrest = aktTermrest,
         Regelrest = BK_Rest(BaumIndex);
  GNTermpaarT RegelIndex;

  /* 1. Ueberpruefen, ob im Blatt befindlicher Rest der linken Seite passt */

  while (Regelrest) {
    SymbolT SymbAusRegel = TO_TopSymbol(Regelrest);
    if (SO_SymbIstVar(SymbAusRegel)) {
      if (VarUngebunden(SymbAusRegel)) {
        SubstErgaenzenSubsumption(Termrest);
        Termrest = TO_NaechsterTeilterm(Termrest);
        Regelrest = TO_Schwanz(Regelrest);
      }
      else if (EsFolgtSubstitutVon(SymbAusRegel,Termrest)) {
        Termrest = TO_NaechsterTeilterm(Termrest);
        Regelrest = TO_Schwanz(Regelrest);
      }
      else {
        xMletztes = xMletztesAlt;
        return FALSE;
      }
    }
    else if (SO_SymbUngleich(SymbAusRegel,TO_TopSymbol(Termrest))) {
      xMletztes = xMletztesAlt;
      return FALSE;
    }
    else {
      Termrest = TO_Schwanz(Termrest);
      Regelrest = TO_Schwanz(Regelrest);
    }
  }


  /* 2. Nun alle im Blatt vorhandenen rechten Seiten durchnudeln */

  RegelIndex = BK_Regeln(BaumIndex);

  do {
    UTermT rechts1 = TP_RechteSeite(RegelIndex),                         /* von potentiell allgemeinerer Gleichung */
           rechts2 = rechteSeite;                                                          /* von Anfragegleichung */
    SymbolT xMletztesAlt2 = xMletztes;
    do {
      SymbolT Symbol1 = TO_TopSymbol(rechts1);
      if (SO_SymbIstVar(Symbol1)) {
        if (VarUngebunden(Symbol1)) {                                        /* ungebundene Variable aus Regelterm */
          SubstErgaenzenSubsumption(rechts2);                                             /* in Anfrageterm binden */
          rechts2 = TO_NaechsterTeilterm(rechts2);
          rechts1 = TO_Schwanz(rechts1);
        }
        else if (EsFolgtSubstitutVon(Symbol1,rechts2)) {                       /* fuer gebundene Variable pruefen, */
          rechts2 = TO_NaechsterTeilterm(rechts2);                                       /* ob sich der Anfragterm */
          rechts1 = TO_Schwanz(rechts1);                                                 /* entsprechend fortsetzt */
        }
        else {                                                         /* Falls nicht, den Matchversuch abbrechen. */
          xMletztes = xMletztesAlt2;
          break;
        }
      }
      else if (SO_SymbUngleich(Symbol1,TO_TopSymbol(rechts2))) {                  /* Abbruch ebenso bei ungleichen */
        xMletztes = xMletztesAlt2;
        break;
      }                                                                                       /* Funktionssymbolen */
      else {                                             /* Bei gleichen Funktionssymbolen naechste Symbole lesen. */
        rechts2 = TO_Schwanz(rechts2);
        rechts1 = TO_Schwanz(rechts1);
      }
    } while (rechts1);

    if (!rechts1)  {                                                               /* Falls Subsumption geglueckt: */
      MO_SubsummierendeGleichung = RegelIndex;
      return TRUE;
    }
  } while ((RegelIndex = RegelIndex->Nachf));

  xMletztes = xMletztesAlt;
  return FALSE;
}


#if TP_PETERGRAF
#define /*BOOLEAN*/ TPSubsummiertLR(/*GNTermpaarT*/ Termpaar, /*UTermT*/ links2, /*UTermT*/ rechts2)                \
  (MO_TermpaarSubsummiertZweites(TP_LinkeSeite(Termpaar),TP_RechteSeite(Termpaar),links2,rechts))
BOOLEAN MO_SubsummierendeGleichungGefunden(UTermT links, UTermT rechts)
{
  SymbolT Topsymbol = TO_TopSymbol(links);
  if (!RE_Gleichungsbaum.Wurzel)
    return FALSE;
  if (SO_SymbIstFkt(Topsymbol)) {
    GleichungsT GleichungsIndex = (GNTermpaarT) BK_Nachf(RE_Gleichungsbaum.Wurzel,Topsymbol);
    while (GleichungsIndex) {
      if (TPSubsummiertLR(GleichungsIndex,links,rechts)) {
        MO_SubsummierendeGleichung = GleichungsIndex;
        return TRUE;
      }
      GleichungsIndex = TP_Nachf(GleichungsIndex);
    }
  }
  if (BK_Nachf(RE_Gleichungsbaum.Wurzel,SO_ErstesVarSymb)) {                     /* exotisch, exotisch: Regel x->t */
    GleichungsT GleichungsIndex = (GNTermpaarT) BK_Nachf(RE_Gleichungsbaum.Wurzel,SO_ErstesVarSymb);
    while (GleichungsIndex) {
      if (TPSubsummiertLR(GleichungsIndex,links,rechts)) {
        MO_SubsummierendeGleichung = GleichungsIndex;
        return TRUE;
      }
      GleichungsIndex = TP_Nachf(GleichungsIndex);
    }
  }
  return FALSE;
}
#else


BOOLEAN MO_SubsummierendeGleichungGefunden(UTermT links, UTermT rechts)
/* Sucht im Gleichungsbaum nach einer Generalisierungen von links==rechts.
   Suchstrategie: Tiefensuche. Besuchsreihenfolge fuer Unterbaeume: zu passendem Funktionssymbol, zu noch nicht    */
{           /* gebundener Variable, zu schon gebundenen Variablen, beginnend mit den davon neueren, endend bei x1. */
#if DONT_MONSTER
  SubstAufIdentitaetSetzen();
  BaumwurzelInitialisieren(RE_Gleichungsbaum,links);
  while (EinenKnotenAbOderVieleKnotenAufgestiegen()) {
    if (BK_IstBlatt(BaumIndex)) {
      if (SubsummierendeGleichungInBlattGefunden(rechts))
        return TRUE;
      else
        BaumIndex = BK_Vorg(BaumIndex);
    }
  }
  return FALSE;
#else
  return FALSE;
#endif
}
#endif



/*=================================================================================================================*/
/*= XI. Initialisierung                                                                                           =*/
/*=================================================================================================================*/

void MO_InitApriori(void)
{
  if (WM_ErsterAnlauf()) {
    xM = (MO_EintragsT *) our_alloc(MO_AnzErsterVariablen*sizeof(MO_EintragsT)) + MO_AnzErsterVariablen;
    xMDimension = -MO_AnzErsterVariablen;
    TI_IterationInitialisieren();
    TI2_IterationInitialisieren();
#if MO_TAUSEND_TERME
    TausendTermeDatei = fopen("TausendTerme","w");
    if (!TausendTermeDatei)
      our_error("TausendTermeDatei konnte nicht geoeffnet werden.");
    TausendTermeZaehler = TAUSEND;
#endif
  }
}



/*=================================================================================================================*/
/*= XII. Statistik                                                                                                =*/
/*=================================================================================================================*/

void MO_Statistik(InfoStringT IS)
{
  if (!(IS_IsIn("all", IS) || IS_IsIn("MO", IS)))
    return;
  if (IS_IsIn("standard", IS)) {
      if(WerteDa(1,AnzRedMisserfolgeOrdnung)) {
        IO_DruckeKurzUeberschrift("MatchOperationen", '+');
        IO_DruckeStatistikZeile("An >-Test gescheiterte Reduktionsversuche mit GM", AnzRedMisserfolgeOrdnung);
      }
  }
  else if (IS_IsIn("latex-head", IS))
    fprintf(IO_Strom, "MO: failed >-Test & ");
  else if (IS_IsIn("intern", IS)) {
    IO_DruckeStatistikZeile("Maximale Ausdehnung der Reduktionsvariablenbank",MO_AusdehnungVariablenbank());
    IO_DruckeStatistikZeile("Maximale Ausdehnung des MO_Kopierstapels", MO_AusdehnungKopierstapel());
  }
  else if (IS_IsIn("config", IS)) {
#if MO_STATISTIK
    fprintf(IO_Strom, "MO: Statistik AN.\n");
#else
    fprintf(IO_Strom, "MO: Statistik AUS.\n");
#endif
  }
}



/*=================================================================================================================*/
/*= XIII. Alle Matches                                                                                            =*/
/*=================================================================================================================*/

static BOOLEAN MO_AllMatchFirst;
/* Initialisierung fuer MO_All*MatchNext */

/* MO_AllRegelMatchFirst/Next sind geklaut von MO_RegelGefunden */
void MO_AllRegelMatchInit (UTermT Term) 
{
   
  MatcherInitialisieren(Term);  
  BaumwurzelInitialisieren(RE_Regelbaum,Term);
  MO_AllMatchFirst = TRUE;
}

BOOLEAN MO_AllRegelMatchNext (long int *RENummer,  PCL_ObjektNummerT *REpclID) {

#if 1
   /* FEHLER BEI MONSTER !! */
   if (MO_AllMatchFirst) {
     MO_AllMatchFirst = FALSE;
   }
   else {
     MatcherNachBlattZuruecksetzen();
     BaumIndex = BK_Vorg(BaumIndex);
  }
  while (EinenKnotenAbOderVieleKnotenAufgestiegen())
    if (BK_IstBlatt(BaumIndex)) {
         if (RegelInBlattGefunden()) {
            gefundeneLinkeSeite = (BK_Regeln(BaumIndex))->linkeSeite;
            gefundeneRechteSeite = (BK_Regeln(BaumIndex))->rechteSeite;
            *RENummer = (BK_Regeln(BaumIndex))->Nummer;
            *REpclID  = TP_PclIDAs(BK_Regeln(BaumIndex));
            return TRUE; 
      }
         else
            BaumIndex = BK_Vorg(BaumIndex);
    }
   return FALSE;
#else
   if (MO_AllMatchFirst) {
     MO_AllMatchFirst = FALSE;
   }
   else {
     MatcherNachBlattZuruecksetzen();
     BaumIndex = BK_Vorg(BaumIndex);
  }
   MONSTER({
      if (RegelInBlattGefunden()) {
            gefundeneLinkeSeite = (BK_Regeln(BaumIndex))->linkeSeite;
            gefundeneRechteSeite = (BK_Regeln(BaumIndex))->rechteSeite;
            *RENummer = (BK_Regeln(BaumIndex))->Nummer;
            *REpclID  = TP_PclIDAs(BK_Regeln(BaumIndex));
            return TRUE; 
      }
      goto backtrack;
   })
#endif
}


/* MO_AllGleichungMatchInit/Next sind geklaut aus MO_GleichungGefunden */
void MO_AllGleichungMatchInit (UTermT Term)
{
   MatcherInitialisieren(Term);  
   BaumwurzelInitialisieren(RE_Gleichungsbaum,Term);
   MO_AllMatchFirst = TRUE;
}


/* Ich will alle Gleichungen haben, die Matchen. Es ist egal, ob die Ordnungsrelation stimmt oder nicht
   - das wird in der Ebene hoeher bestimmt. */
static GleichungsT AllGleichungMatchNextIndex;

BOOLEAN MO_AllGleichungMatchNext (long int *RENummer,  PCL_ObjektNummerT *REpclID, PCL_SeitenT *Seite)
{
#if 1
   /* FEHLER BEI MONSTER !! */
   if (MO_AllMatchFirst) {
      MO_AllMatchFirst = FALSE;
      AllGleichungMatchNextIndex = NULL;
   }
   else {
      if (AllGleichungMatchNextIndex->Nachf) {
         MO_IndirekterClashImBaum = TRUE;
         AllGleichungMatchNextIndex = AllGleichungMatchNextIndex->Nachf;
         gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
         *RENummer = AllGleichungMatchNextIndex->Nummer;
         *REpclID  = TP_PclIDAs(AllGleichungMatchNextIndex);
         *Seite = TP_RichtungAusgezeichnet(AllGleichungMatchNextIndex)?linkeSeite:rechteSeite;
         gefundeneRechteSeite = TP_RechteSeite(AllGleichungMatchNextIndex);
         return TRUE;
      }
      MatcherNachBlattZuruecksetzen();
      BaumIndex = BK_Vorg(BaumIndex);
   }
   while (EinenKnotenAbOderVieleKnotenAufgestiegen()) {
      if (BK_IstBlatt(BaumIndex)) {
         if (RegelInBlattGefunden()) {
            GleichungsT Index = BK_Regeln(BaumIndex);
            gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
            *RENummer = Index->Nummer;
            *REpclID  = TP_PclIDAs(Index);
            *Seite = TP_RichtungAusgezeichnet(Index)?linkeSeite:rechteSeite;
            gefundeneRechteSeite = TP_RechteSeite(Index);
            AllGleichungMatchNextIndex = Index;
            return TRUE;
         }
         BaumIndex = BK_Vorg(BaumIndex);
      }
   }
   return FALSE;
#else
   if (MO_AllMatchFirst) {
      MO_AllMatchFirst = FALSE;
      AllGleichungMatchNextIndex = NULL;
   }
   else {
      if (AllGleichungMatchNextIndex->Nachf) {
         MO_IndirekterClashImBaum = TRUE;
         AllGleichungMatchNextIndex = AllGleichungMatchNextIndex->Nachf;
         gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
         *RENummer = AllGleichungMatchNextIndex->Nummer;
         *REpclID  = TP_PclIDAs(AllGleichungMatchNextIndex);
         *Seite = TP_RichtungAusgezeichnet(AllGleichungMatchNextIndex)?linkeSeite:rechteSeite;
         gefundeneRechteSeite = TP_RechteSeite(AllGleichungMatchNextIndex);
         return TRUE;
      }
      MatcherNachBlattZuruecksetzen();
      BaumIndex = BK_Vorg(BaumIndex);
   }
   MONSTER({
         if (RegelInBlattGefunden()) {
            GleichungsT Index = BK_Regeln(BaumIndex);
            gefundeneLinkeSeite = BK_Regeln(BaumIndex)->linkeSeite;
            *RENummer = Index->Nummer;
            *REpclID  = TP_PclIDAs(Index);
            *Seite = TP_RichtungAusgezeichnet(Index)?linkeSeite:rechteSeite;
            gefundeneRechteSeite = TP_RechteSeite(Index);
            AllGleichungMatchNextIndex = Index;
            return TRUE;
         }
         goto backtrack;
   });
#endif
}


#define /*Block*/ EinSymbolWeiter(/*UTermT*/ TermAlt, /*UTermT*/ TermNeu, /*Block*/ NullAction)                     \
{                                                                                                                   \
  TO_TermzelleHolen(TO_Schwanz(TermNeu));       /* neue Termzelle bereitstellen und an bisherige Zellen anhaengen */\
  TO_SymbolBesetzen(TermNeu = TO_Schwanz(TermNeu),TO_TopSymbol(TermAlt));         /* ausserdem Symbol uebertragen */\
  if (TO_Nullstellig(TermAlt)) {                       /* mehrsymbolige Terme wegen offenen Ende-Zeigers stapeln, */\
    TO_EndeBesetzen(TermNeu,TermNeu);                  /* einsymboliger gleich mit richtigem Ende-Zeiger versehen */\
    while (!(--TI2_TopZaehler)) {                          /* und offene Ende-Zeiger aus Stapel einmuenden lassen */\
      TO_EndeBesetzen(TI2_TopTeilterm,TermNeu);                                                                     \
      if (TI2_PopUndLeer())                                                           /* fertig, wenn Stapel leer */\
        NullAction;                                                                                                 \
    }                                                                                                               \
  }                                                                                                                 \
  else                                                                                                              \
    TI2_Push(TermNeu);                                                                                              \
}


TermT MO_AllTermErsetztNeu(UTermT TermAlt, UTermT Stelle)
{
  if (TO_PhysischGleich(TermAlt,Stelle))
    return SigmaRGanzNeu();
  else {
    TermT TermNeu,
          StartTermNeu;
    TO_TermzelleHolen(TermNeu);
    StartTermNeu = TermNeu;
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TI2_Push(TermNeu);
    while (TO_PhysischUngleich(TermAlt = TO_Schwanz(TermAlt),Stelle))
      EinSymbolWeiter(TermAlt,TermNeu,{;});
    TO_NachfBesetzen(TermNeu,SigmaRGanzNeu());
    TermNeu = TO_TermEnde(TO_Schwanz(TermNeu));
    while (!(--TI2_TopZaehler)) {
      TO_EndeBesetzen(TI2_TopTeilterm,TermNeu);
      if (TI2_PopUndLeer())
        return StartTermNeu;
    }
    TermAlt = TO_NaechsterTeilterm(TermAlt);
    do
      EinSymbolWeiter(TermAlt,TermNeu, {
        TO_NachfBesetzen(TermNeu,NULL);
        return StartTermNeu;
      })
    while ((TermAlt = TO_Schwanz(TermAlt), TRUE));
  }
  /* hier kommen wir nie hin, aber um einige Compiler gluecklich zu machen :-) */
  our_error ("MO_AllTermErsetztNeu reached unreachable code?!");
  return NULL;
}



/*=================================================================================================================*/
/*= XIV. Matching modulo Skolemfunktionen                                                                         =*/
/*=================================================================================================================*/

BOOLEAN MO_SkolemGleich(TermT s, TermT t)
{
  UTermT NULLs = TO_NULLTerm(s);
  do
    if (TO_TopSymboleUngleich(s,t))
      return FALSE;
    else if (TO_IstSkolemTerm(s))
      s = TO_NaechsterTeilterm(s), t = TO_NaechsterTeilterm(t);
    else
      s = TO_Schwanz(s), t = TO_Schwanz(t);
  while (TO_PhysischUngleich(s,NULLs));
  return TRUE;
}

static UTermT SkolemSubstitut1(SymbolT x, TermT si, TermT s, TermT t)
{
  while (TO_PhysischUngleich(s,si))
    if (SO_SymbGleich(x,TO_TopSymbol(s)))
      return t;
    else if (TO_TermIstVar(s) || TO_IstSkolemTerm(s))
      s = TO_NaechsterTeilterm(s), t = TO_NaechsterTeilterm(t);
    else
      s = TO_Schwanz(s), t = TO_Schwanz(t);
  return NULL;
}

static UTermT SkolemSubstitut(TermT si, TermT s, TermT t, TermT s0, TermT t0)
{
  SymbolT x = TO_TopSymbol(si);
  UTermT Resultat = s0 ? SkolemSubstitut1(x, TO_NULLTerm(s0), s0,t0) : NULL;
  return Resultat ? Resultat : SkolemSubstitut1(x, si, s,t);
}

BOOLEAN MO_SkolemAllgemeiner(TermT s, TermT t,  TermT s0, TermT t0)
{
  TermT si = s,
        ti = t;
  UTermT NULLs = TO_NULLTerm(s),
         Si;
  do
    if (TO_TermIstVar(si))
      if ((Si = SkolemSubstitut(si, s,t,s0,t0)) && !MO_SkolemGleich(Si,ti))
        return FALSE;
      else
        si = TO_Schwanz(si), ti = TO_NaechsterTeilterm(ti);
    else if (TO_TopSymboleUngleich(si,ti))
      return FALSE;
    else if (TO_IstSkolemTerm(si))
      si = TO_NaechsterTeilterm(si), ti = TO_NaechsterTeilterm(ti);
    else
      si = TO_Schwanz(si), ti = TO_Schwanz(ti);
  while (TO_PhysischUngleich(si,NULLs));
  return TRUE;
}


BOOLEAN MO_TupelSkolemAllgemeiner(TermT s, TermT t,  TermT u, TermT v)
{
  return (MO_SkolemAllgemeiner(s,u,NULL,NULL) && MO_SkolemAllgemeiner(t,v,s,u)) || 
         (MO_SkolemAllgemeiner(s,v,NULL,NULL) && MO_SkolemAllgemeiner(t,u,s,v));
}
