/*=================================================================================================================
  =                                                                                                               =
  =  RUndEVerwaltung.c                                                                                            =
  =                                                                                                               =
  =  Verwaltung der Regel- und der Gleichungsmenge                                                                =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.94  Thomas.                                                                                            =
  =  25.03.94  Ziel-Verwaltung eingebaut. Arnim.                                                                  =
  =  27.03.94  Ausgaben eingebaut.                                                                                =
  =  11.06.96  Statistik eingebaut, Ziel-Verwaltung ausgelagert. Arnim.                                           =
  =================================================================================================================*/


#include "Ausgaben.h"
#include "compiler-extra.h"
#include "DSBaumAusgaben.h"
#include "DSBaumOperationen.h"
#include "DSBaumTest.h"
#include "InfoStrings.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "pcl.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "Subgleichungen.h"
#include "SymbolOperationen.h"
#include "WaldmeisterII.h"
#include "Zielverwaltung.h"



/*=================================================================================================================*/
/*= 0. Globale Variablen                                                                                          =*/
/*=================================================================================================================*/

DSBaumT RE_Regelbaum,
        RE_Gleichungsbaum;
unsigned long int RE_AnzRegelnGesamt, 
                  RE_AnzGleichungenGesamt;



/*=================================================================================================================*/
/*= I. Bedingter Testbetrieb und Testfunktion                                                                     =*/
/*=================================================================================================================*/

#define TestMitSprungverweisen TRUE            /* Alternativen: TRUE und FALSE; sollte moeglichst auf TRUE stehen! */
#define Testbeginn (unsigned long int) 1                                        /* legt fest, ab wievieltem Aufruf */
                                                              /* wirklich getestet bzw. getestet und gedruckt wird */


#if RE_TEST
static BOOLEAN RegelmengenTest(void);
static BOOLEAN GleichungsmengenTest(BOOLEAN GegenrichtungenSuchen);
#endif

void RE_RUndETest(void)
{
#if RE_TEST
  fprintf(IO_Strom, "\nChecking set of rules ... ");
  if (RegelmengenTest())
    fprintf(IO_Strom, "\n                      ... ok\n");
  else
    our_error("set of rules inconsistent");
  fprintf(IO_Strom, "\nChecking set of equations ... ");
  if (GleichungsmengenTest(TRUE))
    fprintf(IO_Strom, "\n                          ... ok\n");
  else
    our_error("set of equations inconsistent");
#endif
}


#if RE_TEST==2
static long unsigned int TestaufrufNr = 0;
#define RTest(/*char[]*/ Ankuendigungstext)                                                                         \
  TestaufrufNr++;                                                                                                   \
  if (TestaufrufNr>=Testbeginn) {                                                                                   \
    fprintf(IO_Strom, "\nRETest %ld: %s\n",TestaufrufNr,Ankuendigungstext);                                         \
    if (!RegelmengenTest())                                                                                         \
      our_fatal_error("Regelmenge inkonsistent.");                                                                  \
  }
#define ETest(/*BOOLEAN*/ GegenrichtungenSuchen, /*char[]*/ Ankuendigungstext)                                      \
  TestaufrufNr++;                                                                                                   \
  if (TestaufrufNr>=Testbeginn) {                                                                                   \
    fprintf(IO_Strom, "\nRETest %ld: %s\n",TestaufrufNr,Ankuendigungstext);                                         \
    if (!GleichungsmengenTest(GegenrichtungenSuchen))                                                               \
      our_fatal_error("Gleichungsmenge inkonsistent.");                                                             \
  }
#else
#define RTest(/*char[]*/ Ankuendigungstext)
#define ETest(/*BOOLEAN*/ GegenrichtungenSuchen, /*char[]*/ Ankuendigungstext)
#endif



/*=================================================================================================================*/
/*= II. Regelmenge                                                                                                =*/
/*=================================================================================================================*/

static unsigned long int AnzRegelnAktuell;    /* Lokaler Zaehler. Enthaelt immer die aktuelle Regelzahl im System. */

/* 1. Erzeugen von Regeln aus Termpaaren ------------------------------------------------------------------------- */

static RegelT ZuletztErzeugteRegel;

static unsigned int TiefeLinks;

RegelT RE_ErzeugteRegel(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID MAYBE_UNUSEDPARAM)
{
  TP_GNTermpaarHolen(ZuletztErzeugteRegel);
  BO_TermpaarNormieren(links,rechts,&TiefeLinks);
  ZuletztErzeugteRegel->rechteSeite = rechts;
  ZuletztErzeugteRegel->linkeSeite = links;
  ZuletztErzeugteRegel->Nummer = Nummer;
  ZuletztErzeugteRegel->SUEInfo =
    ZuletztErzeugteRegel->Gegenrichtung = NULL;
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteRegel),*PclID);
  return ZuletztErzeugteRegel;
}


/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteRegelEinfuegen(void)
{
  AnzRegelnAktuell++;
  RTest("Regelbaum vor dem Einfuegen");
  BO_ObjektEinfuegen(&RE_Regelbaum,ZuletztErzeugteRegel,TiefeLinks);
  RTest("Regelbaum nach dem Einfuegen");
  NF_RegelAufgenommen();
}


/* 4. Loeschoperation -------------------------------------------------------------------------------------------- */

void RE_RegelEntfernen(RegelT Regel)
{
  AnzRegelnAktuell--;
  RTest("Regelbaum vor dem Entfernen");
  BO_ObjektEntfernen(&RE_Regelbaum,Regel);
  RTest("Regelbaum nach dem Entfernen");
  SUE_KillParent(Regel);
  NF_RegelEntfernt();
}


void RE_RegelLoeschen(RegelT Regel)                    /* Regel als Objekt loeschen sowie Meldung an KP-Verwaltung */
{
   TP_GNTermpaarLoeschen(Regel);
}


void RE_RegelVerfluechtigen(RegelT Regel)
{
  RE_RegelEntfernen(Regel);
  TP_LinksRechtsLoeschen(Regel);
  TP_GNTermpaarLoeschen(Regel);
}

void RE_RegelmengeLoeschen(void)
{
  IO_AusgabenAus();
  while (RE_Regelbaum.Wurzel) {
#if TP_PETERGRAF
    RegelT Regel = (RegelT) RE_Regelbaum.ErstesBlatt;
#else
    RegelT Regel = BK_Regeln(RE_Regelbaum.ErstesBlatt);
#endif
    RE_RegelVerfluechtigen(Regel);
  }
  NF_RegelEntfernt();
  IO_AusgabenEin();
}


/* 5. Ausgabeoperation ------------------------------------------------------------------------------------------- */

void RE_AusgabeRegelmenge(int Grad)
{
  RegelT RegelIndex;
  if (Grad) {
     if (Grad == 7)
        IO_DruckeUeberschrift("Completed system: Set of rules");
     else 
        IO_DruckeUeberschrift("Set of rules");
  }
  
  switch (Grad) {
  case 0:                                                           /* wird fuer Ausgabe als Spezifikation benutzt */
    BK_forRegeln(RegelIndex,RE_Regelbaum, {
      IO_DruckeFlex("\t%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
    });
    break;
  case 1:
  case 7:
    BK_forRegeln(RegelIndex,RE_Regelbaum, {
      fprintf(IO_Strom, "No. %4ld     ",TP_ElternNr(RegelIndex));
      IO_DruckeFlex("%t -> %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
    });
    break;
  case 2:
  case 3:
    (Grad==2 ? BA_BaumMalenDefaultOhneG : BA_BaumMalenDefaultMitG)(RE_Regelbaum);
    break;
  case 4:
  case 5:
    BA_DumpBaum(RE_Regelbaum, Grad==5);
  default:
    break;
  }
   if (Grad)
     fprintf(IO_Strom, "\n\n");
}


/* 7. Testfunktion ----------------------------------------------------------------------------------------------- */

#if RE_TEST
static BOOLEAN RegelPraedikat(RegelT Regel)
{
  if (TP_Nachf(Regel)) {
    fprintf(IO_Strom, "\nFehler: Regel mit eingetragener Nachfolgerregel zur selben linken Seite.\n");
    return FALSE;
  }
  if (TP_Gegenrichtung(Regel)) {
    fprintf(IO_Strom, "\nFehler: Regel mit eingetragener Gegenrichtung.\n");
    return FALSE;
  }
  if (TP_ElternNr(Regel)<=0) {
    fprintf(IO_Strom, "\nFehler: Regel mit nichtpositiver Nummer.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(TP_LinkeSeite(Regel))) {
    fprintf(IO_Strom, "\nFehler: Regel hat fehlerhafte linke Seite.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(TP_RechteSeite(Regel))) {
    fprintf(IO_Strom, "\nFehler: Regel hat fehlerhafte rechte Seite.\n");
    return FALSE;
  }
  return TRUE;
}


static BOOLEAN RegelmengenTest(void)
{
  if (BT_IndexTesten(RE_Regelbaum,RegelPraedikat,TestMitSprungverweisen))
    return TRUE;
  RE_AusgabeRegelmenge(1);
  return FALSE;
}
#endif


/* 8. Umbenennungen ---------------------------------------------------------------------------------------------- */

void RE_RegelUmbenennen(RegelT Regel, signed int Delta)
{
  TO_VariablenUmbenennen(TP_LinkeSeite(Regel),Delta);
  TO_VariablenUmbenennen(TP_RechteSeite(Regel),Delta);
}

SymbolT RE_neuesteVariable(RegelT Regel)
{
  SymbolT neuestesXlinks = TO_neuesteVariable(TP_LinkeSeite(Regel)),
          neuestesXrechts = TO_neuesteVariable(TP_RechteSeite(Regel));
  return min(neuestesXlinks,neuestesXrechts);
}


  
/*=================================================================================================================*/
/*= III. Gleichungsmenge                                                                                          =*/
/*=================================================================================================================*/

static unsigned long int AnzGleichungenAktuell,                      /* lokale Zaehler, die die Ausdehnung */
                         AnzAktMonogleichungen,                      /* der Gleichungsmenge beschreiben */
                         AnzMonogleichungen;                         /* Insgesamt aufgetretene Monogleichungen */

/* 1. Erzeugen von Gleichungen aus Termpaaren -------------------------------------------------------------------- */

static GleichungsT ZuletztErzeugteGleichung;

static unsigned int TiefeRechts;

GleichungsT RE_ErzeugteGleichung(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID MAYBE_UNUSEDPARAM)
{
  GNTermpaarT ZuletztErzeugteAntigleichung;
  TP_GNTermpaarHolen(ZuletztErzeugteAntigleichung);
  TP_GNTermpaarHolen(ZuletztErzeugteGleichung);
  BO_TermpaarNormieren(links,rechts,&TiefeLinks);
  ZuletztErzeugteGleichung->rechteSeite = rechts;
  ZuletztErzeugteGleichung->linkeSeite = links;
  ZuletztErzeugteGleichung->SUEInfo = NULL;
  ZuletztErzeugteGleichung->Nummer = -Nummer;
  ZuletztErzeugteGleichung->RichtungAusgezeichnet = TRUE;
  TP_WeissMalen(ZuletztErzeugteGleichung);
  /* Dies ist die einzige Stelle im System, an der die Abbildung der Gleichungs-Nummern auf negative Zahlen stattfindet!*/
  TP_SubgleichungsReferenzSei(ZuletztErzeugteGleichung,SGM_InitRef());
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteGleichung),*PclID);

  BO_TermpaarNormieren(
    ZuletztErzeugteAntigleichung->linkeSeite = TO_Termkopie(rechts),
    ZuletztErzeugteAntigleichung->rechteSeite = TO_Termkopie(links),
    &TiefeRechts);
  if ((ZuletztErzeugteGleichung->IstMonogleichung = 
    ZuletztErzeugteAntigleichung->IstMonogleichung =
    TP_TermpaareGleich(ZuletztErzeugteGleichung,ZuletztErzeugteAntigleichung)))
    AnzMonogleichungen++;
  ZuletztErzeugteAntigleichung->RichtungAusgezeichnet = FALSE;
  ZuletztErzeugteAntigleichung->Nummer = -Nummer;
  ZuletztErzeugteAntigleichung->SUEInfo = NULL;
  TP_WeissMalen(ZuletztErzeugteAntigleichung);
  TP_SubgleichungsReferenzSei(ZuletztErzeugteAntigleichung,SGM_InitRef());
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteAntigleichung),*PclID);
  
  (ZuletztErzeugteGleichung->Gegenrichtung = ZuletztErzeugteAntigleichung)->Gegenrichtung = ZuletztErzeugteGleichung;

  return ZuletztErzeugteGleichung;
}


/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteHinrichtungEinfuegen(SGM_ReferenzT Referenz)
{
  AnzGleichungenAktuell ++;
  ETest(TRUE,"Gleichungsbaum vor dem Einfuegen der letzten Hinrichtung");
  BO_ObjektEinfuegen(&RE_Gleichungsbaum,ZuletztErzeugteGleichung,TiefeLinks);
  ETest(FALSE,"Gleichungsbaum nach dem Einfuegen der letzten Hinrichtung");
  TP_SubgleichungsReferenzSei(ZuletztErzeugteGleichung,Referenz);
}


void RE_ZuletztErzeugteHerrichtungEinfuegen(void)                    /* Nur wenn Gleichung keine Monogleichung ist.*/
{
  if (TP_IstKeineMonogleichung(ZuletztErzeugteGleichung)) {
    ETest(FALSE,"Gleichungsbaum vor dem Einfuegen der letzten Herrichtung");
    BO_ObjektEinfuegen(&RE_Gleichungsbaum,TP_Gegenrichtung(ZuletztErzeugteGleichung),TiefeRechts);
    ETest(TRUE,"Gleichungsbaum nach dem Einfuegen der letzten Herrichtung");
  }
  else
    AnzAktMonogleichungen++;
  NF_GleichungAufgenommen();
}


/* 4. Loeschoperationen ------------------------------------------------------------------------------------------ */

void RE_GleichungEntfernen(GleichungsT Gleichung)            /* Nur, wenn nicht Gegenrichtung einer Monogleichung. */
{
  if (TP_RichtungAusgezeichnet(Gleichung) || TP_IstKeineMonogleichung(Gleichung)) {
    ETest(FALSE,"Gleichungsbaum vor dem Entfernen");
    BO_ObjektEntfernen(&RE_Gleichungsbaum,Gleichung);
    ETest(FALSE,"Gleichungsbaum nach dem Entfernen");
  }
  if (TP_RichtungAusgezeichnet(Gleichung))
    SUE_KillParent(Gleichung);

  NF_GleichungEntfernt();
}


void RE_GleichungLoeschen(GleichungsT Gleichung)     /* Gleichung als Objekt loeschen, dabei Meldung an SUE und NF */
{                                                             /* sowie SGM loeschen, falls ausgezeichnete Richtung */
  if (TP_RichtungAusgezeichnet(Gleichung)){
    AnzGleichungenAktuell--;
    if (TP_IstMonogleichung(Gleichung))
      AnzAktMonogleichungen--;
    SGM_LoescheSubgleichungen(TP_SubgleichungsReferenz(Gleichung));
  }
  TP_GNTermpaarLoeschen(Gleichung);
}


void RE_GleichungVerfluechtigen(GleichungsT Gleichung)
{
  RE_GleichungEntfernen(TP_Gegenrichtung(Gleichung));                 /* Test auf Monogl. erfolgt in 'Entfernen' */
  RE_GleichungEntfernen(Gleichung);
  TP_LinksRechtsLoeschen(TP_Gegenrichtung(Gleichung));
  RE_GleichungLoeschen(TP_Gegenrichtung(Gleichung));
  TP_LinksRechtsLoeschen(Gleichung);
  RE_GleichungLoeschen(Gleichung);
}

void RE_GleichungsmengeLoeschen(void)
{
  IO_AusgabenAus();
  while (RE_Gleichungsbaum.Wurzel) {
#if TP_PETERGRAF
    GleichungsT Gleichung = (GleichungsT) RE_Gleichungsbaum.ErstesBlatt;
#else
    GleichungsT Gleichung = BK_Regeln(RE_Gleichungsbaum.ErstesBlatt);
#endif
    RE_GleichungVerfluechtigen(Gleichung);
  }
  IO_AusgabenEin();
}


/* 5. Ausgabeoperationen ----------------------------------------------------------------------------------------- */

void RE_AusgabeGleichungsmenge(int Grad)
{
  RegelT RegelIndex;
  if (Grad) {
     if (Grad == 7)
        IO_DruckeUeberschrift("Completed system: Set of equations");
     else 
        IO_DruckeUeberschrift("Set of equations");
  }
  switch (Grad) {
  case 0: {                                                                      /* fuer Ausgabe als Spezifikation */
    BK_forRegeln(RegelIndex,RE_Gleichungsbaum, {
      if (TP_RichtungAusgezeichnet(RegelIndex)) {
        IO_DruckeFlex("\t%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
      }
    });
    break;
  }
  case 1:
  case 7:
    BK_forRegeln(RegelIndex,RE_Gleichungsbaum, {
      if (TP_RichtungAusgezeichnet(RegelIndex)) {
        fprintf(IO_Strom, "Nr. %4ld     ",TP_ElternNr(RegelIndex));
        IO_DruckeFlex("%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
      }
    });
    break;
  case 2: 
  case 3:
    (Grad==2 ? BA_BaumMalenDefaultOhneG : BA_BaumMalenDefaultMitG)(RE_Gleichungsbaum);
    break;
  case 4:
  case 5:
    BA_DumpBaum(RE_Gleichungsbaum, Grad==5);
  default:
    break;
  }
  if (Grad)
    fprintf(IO_Strom, "\n\n");
}


/* 6. Schleifenkonstrukte ---------------------------------------------------------------------------------------- */

GleichungsT RE_ErsteGleichung(void)
{
  GleichungsT ErsteGleichung = BK_ErstesTermpaar(RE_Gleichungsbaum);
  while (ErsteGleichung && TP_RichtungUnausgezeichnet(ErsteGleichung))
    ErsteGleichung = BK_NaechstesTermpaar();
  return ErsteGleichung;
}
GleichungsT RE_NaechsteGleichung(void)
{
  GleichungsT NaechsteGleichung = BK_NaechstesTermpaar();
  while (NaechsteGleichung && TP_RichtungUnausgezeichnet(NaechsteGleichung))
    NaechsteGleichung = BK_NaechstesTermpaar();
  return NaechsteGleichung;
}

GleichungsT RE_ErsteGleichungM(KantenT *iBlatt, GNTermpaarT *iTermpaar)
{
  GleichungsT ErsteGleichung = BK_ErstesTermpaarM(RE_Gleichungsbaum,iBlatt,iTermpaar);
  while (ErsteGleichung && TP_RichtungUnausgezeichnet(ErsteGleichung))
    ErsteGleichung = BK_NaechstesTermpaarM(iBlatt,iTermpaar);
  return ErsteGleichung;
}
GleichungsT RE_NaechsteGleichungM(KantenT *iBlatt, GNTermpaarT *iTermpaar)
{
  GleichungsT NaechsteGleichung = BK_NaechstesTermpaarM(iBlatt,iTermpaar);
  while (NaechsteGleichung && TP_RichtungUnausgezeichnet(NaechsteGleichung))
    NaechsteGleichung = BK_NaechstesTermpaarM(iBlatt,iTermpaar);
  return NaechsteGleichung;
}


static BOOLEAN SchonInGleichungen;
RegelOderGleichungsT RE_ErstesFaktum(void)
{
  RegelT ErsteRegel = RE_ErsteRegel();
  return ((SchonInGleichungen = !ErsteRegel)) ? RE_ErsteGleichung() : ErsteRegel;
}
RegelOderGleichungsT RE_NaechstesFaktum(void)
{
  if (SchonInGleichungen)
    return RE_NaechsteGleichung();
  else {
    RegelT NaechsteRegel = RE_NaechsteRegel();
    return (NaechsteRegel ? NaechsteRegel : (SchonInGleichungen = TRUE, RE_ErsteGleichung()));
  }
}

RegelOderGleichungsT RE_ErstesFaktumM(KantenT *iBlatt, GNTermpaarT *iTermpaar, BOOLEAN *SchonInGleichungen)
{
  RegelT ErsteRegel = RE_ErsteRegelM(iBlatt,iTermpaar);
  return ((*SchonInGleichungen = !ErsteRegel)) ? RE_ErsteGleichungM(iBlatt,iTermpaar) : ErsteRegel;
}
RegelOderGleichungsT RE_NaechstesFaktumM(KantenT *iBlatt, GNTermpaarT *iTermpaar, BOOLEAN *SchonInGleichungen)
{
  if (*SchonInGleichungen)
    return RE_NaechsteGleichungM(iBlatt,iTermpaar);
  else {
    RegelT NaechsteRegel = RE_NaechsteRegelM(iBlatt,iTermpaar);
    return (NaechsteRegel ? NaechsteRegel : (*SchonInGleichungen = TRUE, RE_ErsteGleichungM(iBlatt,iTermpaar)));
  }
}


/* 7. Testfunktion ----------------------------------------------------------------------------------------------- */

#if RE_TEST
static BOOLEAN ImGleichungsbaumVorhanden(GleichungsT Gleichung)
{
  KantenT Knoten = RE_Gleichungsbaum.Wurzel;
  UTermT TermIndex = TP_LinkeSeite(Gleichung),
         SuffixIndex;
  GleichungsT GleichungsIndex;
  if (!Knoten)
    return FALSE;
  do {
    Knoten = BK_Nachf(Knoten,TO_TopSymbol(TermIndex));
    if (!Knoten)
      return FALSE;
    TermIndex = TO_Schwanz(TermIndex);
  } while (!BK_IstBlatt(Knoten));
  SuffixIndex = BK_Rest(Knoten);
  while (SuffixIndex) {
    if (!TO_TopSymboleGleich(SuffixIndex,TermIndex))
      return FALSE;
    SuffixIndex = TO_Schwanz(SuffixIndex);
    TermIndex = TO_Schwanz(TermIndex);
  }
  GleichungsIndex = BK_Regeln(Knoten);
  while (GleichungsIndex) {
    if (GleichungsIndex==Gleichung)
      return TRUE;
    GleichungsIndex = TP_Nachf(GleichungsIndex);
  }
  return FALSE;
}


static BOOLEAN GleichungsPraedikatSchwach(GleichungsT Gleichung)
{
  GleichungsT Antigleichung = TP_Gegenrichtung(Gleichung);
  TermT Links = TP_LinkeSeite(Gleichung),
        Rechts = TP_RechteSeite(Gleichung),
        AntiLinks = TP_LinkeSeite(Antigleichung),
        AntiRechts = TP_RechteSeite(Antigleichung);
  unsigned int TiefeTerm1;

  if (!Antigleichung) {
    fprintf(IO_Strom, "\nFehler: Gleichung mit Antigleichung NULL.\n");
    return FALSE;
  }
  if (TP_RichtungAusgezeichnet(Gleichung)==TP_RichtungAusgezeichnet(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung beide ausgezeichnet oder beide unausgezeichnet.\n");
    return FALSE;
  }
  if (TP_ElternNr(Gleichung)>=0) {
    fprintf(IO_Strom, "\nFehler: Gleichung mit nichtnegativer Nummer.\n");
    return FALSE;
  }
  if (TP_ElternNr(Gleichung)!=TP_ElternNr(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung haben verschiedene Nummern.\n");
    return FALSE;
  }
  if (TP_Gegenrichtung(Antigleichung)!=Gleichung) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung sind nicht korrekt miteinander verzeigert.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(Links)) {
    fprintf(IO_Strom, "\nFehler: Gleichung hat fehlerhafte linke Seite.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(Rechts)) {
    fprintf(IO_Strom, "\nFehler: Gleichung hat fehlerhafte rechte Seite.\n");
    return FALSE;
  }
  BO_TermpaarNormieren(AntiRechts,AntiLinks,&TiefeTerm1);
  if (!TO_TermeGleich(Links,AntiRechts) || !TO_TermeGleich(Rechts,AntiLinks)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung sind nicht Varianten voneinander.\n");
    return FALSE;
  }
  BO_TermpaarNormieren(AntiLinks,AntiRechts,&TiefeTerm1);
  if (TP_IstMonogleichung(Gleichung)!=TP_IstMonogleichung(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung haben unterschiedliche"
      " Monogleichungskennzeichnung.\n");
    return FALSE;
  }
  if (TP_IstMonogleichung(Gleichung)) {
    GleichungsT BaumExterne = TP_RichtungUnausgezeichnet(Gleichung) ? Gleichung : Antigleichung;
    if (!TO_TermKorrekt(TP_LinkeSeite(BaumExterne))) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung hat fehlerhafte linke Seite.\n");
      return FALSE;
    }
    if (!TO_TermKorrekt(TP_RechteSeite(BaumExterne))) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung hat fehlerhafte rechte Seite.\n");
      return FALSE;
    }
    if (ImGleichungsbaumVorhanden(BaumExterne)) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung steht im Gleichungsbaum.\n");
      return FALSE;
    }
  }
  return TRUE;
}


static BOOLEAN GleichungsPraedikatNormal(GleichungsT Gleichung)
{
  if (!GleichungsPraedikatSchwach(Gleichung))
    return FALSE;
  if (TP_IstKeineMonogleichung(Gleichung) && !ImGleichungsbaumVorhanden(TP_Gegenrichtung(Gleichung))) {
    fprintf(IO_Strom, "\nFehler: Gegenrichtung von Stereogleichung steht nicht im Gleichungsbaum.\n");
    return FALSE;
  }
  return TRUE;
}


static BOOLEAN GleichungsmengenTest(BOOLEAN GegenrichtungenSuchen)
{
  if (BT_IndexTesten(RE_Gleichungsbaum,GegenrichtungenSuchen ? GleichungsPraedikatNormal : 
    GleichungsPraedikatSchwach,TestMitSprungverweisen))
    return TRUE;
  RE_AusgabeGleichungsmenge(4);
  return FALSE;
}
#endif



/*=================================================================================================================*/
/*= IV. Ergebnisausgabe R-, E- und KP-Menge                                                                       =*/
/*=================================================================================================================*/

void RE_SonstigeErgebnisse(BOOLEAN flag) /* flag dient zur Detailsteuerung in den verschiedenen Modi:
                                   convergence: TRUE  = im TP-Puffer befinden sich divergente KPs */
{                               /*              FALSE = im TP-Puffer befinden sich unrichtbare Ausgangsgleichungen */
  if ((PA_ExtModus == convergence) || (PA_ExtModus == termination) || (PA_ExtModus == confluence)){
    TermT links, rechts;
    unsigned long int nr;
    if (PA_ExtModus == termination)
      IO_DruckeKurzUeberschrift("Unrichtbare Ausgangsgleichungen", '+');
    else if (PA_ExtModus == confluence)
      IO_DruckeKurzUeberschrift("Divergente kritische Paare", '+');
    else                        /* convergence */
      IO_DruckeKurzUeberschrift(flag ? "Divergente kritische Paare" : "Unrichtbare Ausgangsgleichungen", '+');
    while(PU_TPPufferNichtLeer()){
      PU_TPAuslesen(&links, &rechts, &nr);
      fprintf(IO_Strom, "%5lu  ", nr);
      IO_DruckeFlex("%t = %t\n", links, rechts);
    }
    fprintf(IO_Strom, "\n\n");
  }
}


#if COMP_PCL
void RE_DruckeEndsystemFuerPCL(void)
{
  if (PA_ExtModus == completion){
    GNTermpaarT Index;
    RE_forRegelnRobust(Index, {
      PCL_VVAbschlussFaktDrucken(TP_PclID(Index),TP_LinkeSeite(Index),TP_RechteSeite(Index),Regel);
    });
    RE_forGleichungenRobust(Index, {
      if (TP_RichtungAusgezeichnet(Index))
        PCL_VVAbschlussFaktDrucken(TP_PclID(Index),TP_LinkeSeite(Index),TP_RechteSeite(Index),Gleichung);
    });
  }
}
#endif



/*=================================================================================================================*/
/*= V. Regeln und Gleichungen als Eltern von KPs                                                                  =*/
/*=================================================================================================================*/


BOOLEAN RE_IstUngleichung(RegelOderGleichungsT Faktum)
{ return Z_IstGemantelteUngleichung(TP_LinkeSeite(Faktum),TP_RechteSeite(Faktum)); }

void RE_UngleichungEntmanteln(UTermT* s, UTermT* t, RegelOderGleichungsT Faktum)
{ Z_UngleichungEntmanteln(s,t, TP_LinkeSeite(Faktum),TP_RechteSeite(Faktum)); }

BOOLEAN RE_UngleichungEntmantelt(UTermT* s, UTermT* t, RegelOderGleichungsT Faktum)
{ return Z_UngleichungEntmantelt(s,t, TP_LinkeSeite(Faktum),TP_RechteSeite(Faktum)); }



/*=================================================================================================================*/
/*= VI. Initialisierungen                                                                                         =*/
/*=================================================================================================================*/

void RE_InitApriori(void)
{
  BO_DSBaumInitialisieren(&RE_Regelbaum);                      /* muss vor jedem neuen Strategiedurchlauf erfolgen */
  BO_DSBaumInitialisieren(&RE_Gleichungsbaum);
}  


/*=================================================================================================================*/
/*= VII. Statistik                                                                                                =*/
/*=================================================================================================================*/


void RE_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("all", IS) || IS_IsIn("RE", IS)))
      return;
   if (IS_IsIn("standard", IS)){
      int tiefe1;
      /******** Regelmenge ****************************** */
      if (RE_AnzRegelnGesamt > 0){
         unsigned int KnotenImBaum, GesparteKnoten;
         BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Regelbaum);
         IO_DruckeKurzUeberschrift("Set of Rules",'+');
         tiefe1 = IO_Einruecktiefe(0, AnzRegelnAktuell);
         IO_DruckeStatistikZeile("Total Number of Rules",RE_AnzRegelnGesamt);
         IO_DruckeStatistikZeileEingerueckt("of them still in R", AnzRegelnAktuell, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Nodes in R-tree", KnotenImBaum);
         IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
         IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                       GesparteKnoten, GesparteKnoten+KnotenImBaum);
      }
      /******** Gleichungsmenge ************************* */
      if (RE_AnzGleichungenGesamt > 0){
         unsigned int KnotenImBaum, GesparteKnoten;
         BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Gleichungsbaum);
         IO_DruckeKurzUeberschrift("Set of Equations",'+');
         IO_DruckeStatistikZeile("Total Number of Equations",RE_AnzGleichungenGesamt);
         tiefe1 = IO_Einruecktiefe(0, Maximum(2, AnzGleichungenAktuell, AnzMonogleichungen));
         IO_DruckeStatistikZeileEingerueckt("of them still in E", AnzGleichungenAktuell, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them Mono-Equations", AnzMonogleichungen, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them still in E", AnzAktMonogleichungen, tiefe1+IO_Platzverbrauch(AnzAktMonogleichungen)+2);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Nodes in E-tree", KnotenImBaum);
         IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
         IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                       GesparteKnoten, GesparteKnoten+KnotenImBaum);
      }
   }
   else if (IS_IsIn("short", IS)){
      IO_DruckeKurzUeberschrift("Sets of Rules and Equations",'+');
      IO_DruckeStatistikZeile("Total Number of Rules",RE_AnzRegelnGesamt);
      IO_DruckeStatistikZeile("Total Number of Equations",RE_AnzGleichungenGesamt);
   }
   else if (IS_IsIn("latex-head", IS)){
      if (RE_AnzRegelnGesamt > 0)
         fprintf(IO_Strom, "R: total & current & ");
      if (RE_AnzGleichungenGesamt > 0)
         fprintf(IO_Strom, "E: total & current & mono & curr. mono & ");
   }
   else if (IS_IsIn("slatex-head", IS)){
      fprintf(IO_Strom, "R & E & ");
   }
   else if (IS_IsIn("intern", IS)){
      IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in R-tree", RE_Regelbaum.TiefeBlaetterDimension);
      IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in E-tree", RE_Gleichungsbaum.TiefeBlaetterDimension);
   }
   else if (IS_IsIn("config", IS)){
#if RE_TEST
      fprintf(IO_Strom, "RE: Test  ON, Teststart %lud, %s Jumplist-Test.\n", Testbeginn, TestMitSprungverweisen?"with":"without");
#else
      fprintf(IO_Strom, "RE: Test OFF.\n");
#endif
   }
}

void RE_AktuelleGroessen(unsigned long int *Regelmenge, unsigned long int *Gleichungsmenge)
{
  *Regelmenge = AnzRegelnAktuell;
  *Gleichungsmenge = AnzGleichungenAktuell;
}

