/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - SUEM.c                                                                                                */
/* - ******                                                                                                */
/* -                                                                                                       */
/* - Schnittstelle fuer das Set of unselected equations.                                                   */
/* -                                                                                                       */
/* - 14.03.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "CStatistics.h"
#include "CritGoals.h"
#include "general.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "NewClassification.h"
#include "pcl.h"
#include "PhilMarlow.h"
#include "PriorityQueue.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "TermOperationen.h"
#include "TermpairRepresentation.h"
#include "Unifikation1.h"
#include "Zielverwaltung.h"


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ STATISTICS --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

       CounterT        SUE_AnzAktivierterFakten,          /* beinhaltet auch die entfernten Fakten */
                       SUE_AnzAktivierterRE,              /* aktivierte Elemente aus KP-Menge */
                       SUE_AnzAktivierterG,               /* dito aus CG-Menge */
                       SUE_AnzAktivierterREFifo,          /* aus KP-Menge mit Fifo */
                       SUE_AnzAktivierterREHeu,           /* aus KP-Menge mit Heuristik */
                       SUE_AnzAktivierterGFifo,           /* dito aus CG-Menge */
                       SUE_AnzAktivierterGHeu;            /* dito aus CG-Menge */


       CounterT        AnzKPGesamt;                       /* Insgesamt berechnete kritischer Paare */

static CounterT        AnzKPAusSpezifikation;             /* Kritische Paare aus der Spezifikation */
static CounterT        AnzKPIROpfer;                      /* KPs aus geloeschten Regeln und/oder Gleichungen */
static CounterT        AnzKPDatenbank;                    /* Mit Hilfe der Datenbank generierte KPs */

static CounterT        AnzKPdirektZusammengefuehrt;       /* Sofort nach der Bildung zusammenfuehrbare kritische Paare */
static CounterT        AnzKPdirektSubsummiert;            /* Sofort nach der Bildung subsummierte kritische Paare */
static CounterT        AnzKPdirektMNFSubsummiert;         /* Sofort nach der Bildung MNF-subsummierte kritische Paare */
static CounterT        AnzKPdirektPermSubsummiert;        /* Sofort nach der Bildung PERM-subsummierte kritische Paare */
static CounterT        AnzKPdirektMNFPermSubsummiert;     /* Sofort nach der Bildung MNF-PERM-subsummierte kritische Paare */
static CounterT        AnzKPInserted;                     /* In die Menge eingefuegte KPs */

static CounterT        AnzKPBehandeltInVervollst;         /* Zum Vervollstaendigungsalgorithmus hochgereichte KPs */
static CounterT        AnzKPZurueckgestellt;              /* Kritische Paare, die wegen Unrichtbarkeit zurueckgestellt wurden. */
static CounterT        AnzKPspaeterZusammengefuehrt;      /* Kritische Paare, die bei ihrer Behandlung zusammengefuehrt werden konnten. */
static CounterT        AnzKPspaeterSubsummiert;           /* Kritische Paare, die bei ihrer Behandlung subsummiert werden konnten. */
static CounterT        AnzKPspaeterMNFSubsummiert;        /* Kritische Paare, die bei ihrer Behandlung MNF-subsummiert werden konnten. */
static CounterT        AnzKPspaeterPermSubsummiert;       /* Kritische Paare, die bei ihrer Behandlung PERM-subsummiert werden konnten. */
static CounterT        AnzKPspaeterMNFPermSubsummiert;    /* Kritische Paare, die bei ihrer Behandlung MNF-PERM-subsummiert werden konnten. */
static CounterT        AnzKPZuRE;                         /* Zu Regeln/Gleichungen gemachte KPs */
static CounterT        AnzKPWeggeworfen;                  /* Weggeworfen, weil Gewicht zu gross */

#if PUNKT_IRP_HACK
static CounterT        AnzKPPunktIRP;                     /* Punkt-IRP-Hack */
#endif

#define IncAnzKPGesamt() (AnzKPGesamt++)

#define IncAnzBehandelterKPsVervollst() (AnzKPBehandeltInVervollst++)

#define IncAnzKPAusSpezifikation() (AnzKPAusSpezifikation++)

#define IncAnzKPDatenbank() (AnzKPDatenbank++)

#define SUE_STATISTIK normal_power_comp(1,1,1)
/* Alternativen: 0 fuer aus, 1 fuer an */

#if SUE_STATISTIK
#define NrNaechstesKP() (AnzKPGesamt + 1)

#define IncAnzKPIROpfer() (AnzKPIROpfer++)

#define IncAnzKPdirektZusammengefuehrt() (AnzKPdirektZusammengefuehrt++) 
#define IncAnzKPdirektSubsummiert() (AnzKPdirektSubsummiert++) 
#define IncAnzKPdirektMNFSubsummiert() (AnzKPdirektMNFSubsummiert++) 
#define IncAnzKPdirektPermSubsummiert() (AnzKPdirektPermSubsummiert++) 
#define IncAnzKPdirektMNFPermSubsummiert() (AnzKPdirektMNFPermSubsummiert++) 
#define IncAnzKPInserted() (AnzKPInserted++)

#define IncAnzKPZurueckgestellt() (AnzKPZurueckgestellt++)
#define IncAnzKPspaeterZusammengefuehrt() (AnzKPspaeterZusammengefuehrt++)
#define IncAnzKPspaeterSubsummiert() (AnzKPspaeterSubsummiert++)
#define IncAnzKPspaeterMNFSubsummiert() (AnzKPspaeterMNFSubsummiert++)
#define IncAnzKPspaeterPermSubsummiert() (AnzKPspaeterPermSubsummiert++)
#define IncAnzKPspaeterMNFPermSubsummiert() (AnzKPspaeterMNFPermSubsummiert++)
#define IncAnzKPZuRE() (AnzKPZuRE++)
#define IncAnzKPWeggeworfen() (AnzKPWeggeworfen++)

#define MaximiereAnzKPBlaeh(neuwert){if (neuwert>AnzKPBlaeh)AnzKPBlaeh = neuwert;}

#else

#define NrNaechstesKP()                                  /* Nix Tun */

#define IncAnzKPIROpfer()                                /* Nix Tun */

#define IncAnzKPdirektZusammengefuehrt()                 /* Nix Tun */
#define IncAnzKPdirektSubsummiert()                      /* Nix Tun */
#define IncAnzKPdirektMNFSubsummiert()                   /* Nix Tun */
#define IncAnzKPdirektPermSubsummiert()                  /* Nix Tun */
#define IncAnzKPdirektMNFPermSubsummiert()               /* Nix Tun */
#define IncAnzKPInserted()                               /* Nix Tun */

#define IncAnzKPZurueckgestellt()                        /* Nix Tun */
#define IncAnzKPspaeterZusammengefuehrt()                /* Nix Tun */
#define IncAnzKPspaeterSubsummiert()                     /* Nix Tun */
#define IncAnzKPspaeterMNFSubsummiert()                  /* Nix Tun */
#define IncAnzKPspaeterMNFPermSubsummiert()              /* gar nichts */
#define IncAnzKPspaeterPermSubsummiert()                 /* Nix Tun */
#define IncAnzKPZuRE()                                   /* Nix Tun */
#define IncAnzKPWeggeworfen()                            /* dito    */

#endif

static long int lowerBound, upperBound;

/* fuer Interleaving von CP und CG */
static unsigned long int modulo;
static unsigned long int threshold;

/***********************************************************************
  Ausgaben 
***********************************************************************/
#if IO_AUSGABEN
static void KPZusammengefuehrtDrucken (TermT ergebnis)
{
    if (IO_level3) fprintf(IO_Strom, "... joined.\n\n");
    else if(IO_level4) IO_DruckeFlex("... joined into %t.\n\n", ergebnis);
}

static void KPSubsummiertDrucken(void)
{
   if ((IO_level3) || (IO_level4))
      fprintf(IO_Strom, "... subsumed.\n");
}

static void KPPermSubsummiertDrucken (void)
{
   if ((IO_level3) || (IO_level4))
      fprintf(IO_Strom, "... perm-subsumed.\n");
}
static void KPwirdBehandeltDrucken(long int VaterNr, long int MutterNr, TermT links, TermT rechts)
{
   if (IO_level1)
      fprintf(IO_Strom, ".");
   else if (IO_level3)
      fprintf(IO_Strom, "\n************************************\nue (%ld, %ld) touched.\n", VaterNr, MutterNr);
   else if (IO_level4){
      fprintf(IO_Strom, "\n************************************\nue (%ld, %ld) touched:\n", VaterNr, MutterNr);
      IO_DruckeFlex("       %t # %t\n",links, rechts);
   }
}

#define DruckeMeldungKPgebildet(/* unsigned long */ Nummer, /* long */ VaterNr, /* long */ MutterNr, /* TermT */ links, /* TermT */ rechts)  \
{                                                                                                                                               \
    if (IO_level3)                                                                                                                                \
       fprintf(IO_Strom, "critical pair %lu built with parents %ld and %ld.\n", Nummer, VaterNr, MutterNr);                            \
    else if (IO_level4){                                                                                                                          \
       fprintf(IO_Strom, "critical pair %lu built with parents %ld and %ld:\n", Nummer, VaterNr, MutterNr);                            \
       IO_DruckeFlex("       %t # %t\n",links, rechts);                                                                                         \
    }                                                                                                                                           \
 } 

#define DruckeMeldungKPmitDBgebildet(/* unsigned long */ Nummer, /* TermT */ links, /* TermT */ rechts)      \
{                                                                                                               \
    if (IO_level3)                                                                                              \
       fprintf(IO_Strom, "critical pair %lu generated from database.\n", Nummer);                               \
    else if (IO_level4){                                                                                        \
       fprintf(IO_Strom, "critical pair %lu generated from database:\n", Nummer);                               \
       IO_DruckeFlex("       %t # %t\n",links, rechts);                                                         \
    }                                                                                                           \
 } 



#define DruckeMeldungZuKPgemacht(/* unsigned long */ Nummer, /* TermT */ links, /* TermT */ rechts)          \
{                                                                                                               \
    if (IO_level3)                                                                                              \
       fprintf(IO_Strom, "... made ues %lu.\n",Nummer);                                                         \
    else if (IO_level4)                                                                                         \
       IO_DruckeFlex("... made ues %lu: %t # %t\n",Nummer, links, rechts);                                      \
 } 

#define DruckeMeldungAlsKPaufgenommen(lhs,rhs)					\
{											\
    if (IO_level4)									\
       IO_DruckeFlex("... normalized to:\n       %t # %t\n... added to SUE.\n\n", lhs, rhs);	\
    else if (IO_level3)									\
       fprintf(IO_Strom, "... added to SUE.\n\n");					\
} 

#define DruckeMeldungKPWeggeworfen(/*void*/)                                                                 \
{                                                                                                               \
    if ((IO_level4) || (IO_level3))                                                                             \
       fprintf(IO_Strom, "... too big - thrown away.\n\n");                                                     \
} 


#else /* !IO_AUSGABEN */

#define KPZusammengefuehrtDrucken(/*TermT*/ ergebnis)
#define KPSubsummiertDrucken()
#define KPPermSubsummiertDrucken()
#define KPwirdBehandeltDrucken(/*long*/ VaterNr, /*long*/ MutterNr, /*TermT*/ links, /* TermT*/ rechts) /* nix tun */

#define DruckeMeldungKPgebildet(/*unsigned long*/ Nummer, /*long*/ VaterNr, /*long*/ MutterNr, /*TermT*/ links, /*TermT*/ rechts) /* nix tun */
#define DruckeMeldungKPmitDBgebildet(/*unsigned long*/ Nummer, /*TermT*/ links, /*TermT*/ rechts) /* nix tun */
#define DruckeMeldungZuKPgemacht(/* unsigned long */ Nummer, /* TermT */ links, /* TermT */ rechts) /* nix tun */
#define DruckeMeldungAlsKPaufgenommen(lhs,rhs) /* nix tun */
#define DruckeMeldungKPWeggeworfen(/*void*/)
#endif /* IO_AUSGABEN */

static void UEOutput_TW(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld), Weight: ", TP_ElternNr(actualP), TP_ElternNr(otherP));
      fprintf(IO_Strom, "%11ld.%11ld.", wt.w1, wt.w2);
      fprintf(IO_Strom, " : ");
      TPR_PrintTermpair(tp, actualP, otherP, " ?= ");
      fprintf(IO_Strom, "\n");
   }
}

static void UEOutput_W(WeightEntryT wt, TermpairT tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld), Weight: ", TP_ElternNr(actualP), TP_ElternNr(otherP));
      fprintf(IO_Strom, "%11ld.%11ld.", wt.w1, wt.w2);
      fprintf(IO_Strom, "\n");
   }
}

static void UEOutput_T(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld) :", TP_ElternNr(actualP), TP_ElternNr(otherP));
      TPR_PrintTermpair(tp, actualP, otherP, " ?= ");
      fprintf(IO_Strom, "\n");
   }
}


static void UEOutput__(WeightEntryT wt, TermpairT tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld) \n", TP_ElternNr(actualP), TP_ElternNr(otherP));
   }
}

void SUE_Statistics(InfoStringT StatisticsIS)
/* InfoString:
   output -- Mengen-Ausgabe
   weight -- Gewichtsverteilung
   lowerBound=..., upperBound=...,
   noweights,
   noterms,
   time   -- Zeitliche Verteilung der Gewichte
*/
{
   int tiefe1, tiefe2;

   /* output:limit=value:noterms:noweights. Dabei gilt limit fuer das 1. int-Gewicht
      Dabei ist der Default: kein limit, Terme und Gewichte ausgeben */

   
   if (!(IS_IsIn("all", StatisticsIS) || (IS_IsIn("SUEM", StatisticsIS)))){
      if (IS_IsIn("PQ", StatisticsIS))
         PQ_Statistics(StatisticsIS);
      if (IS_IsIn("C", StatisticsIS))
         C_Statistics(StatisticsIS);
      if (IS_IsIn("TPR", StatisticsIS))
         TPR_Statistics(StatisticsIS);
      if (IS_IsIn("CG", StatisticsIS))
	CG_Statistics (StatisticsIS);
      return;
   }
   if (IS_IsIn("output", StatisticsIS)){

      if (IS_ValueOf("lowerBound", StatisticsIS))
         lowerBound = atoi(IS_ValueOf("lowerBound", StatisticsIS));
      else
         lowerBound = LONG_MIN;

      if (IS_ValueOf("upperBound", StatisticsIS))
         upperBound = atoi(IS_ValueOf("upperBound", StatisticsIS));
      else
         upperBound = LONG_MAX;
   
      
      fprintf(IO_Strom, "SUE_Statistics: output (%ld - %ld): %s\n", lowerBound, upperBound, StatisticsIS);

      if (IS_IsIn("noterms", StatisticsIS)) {
         if (IS_IsIn("noweights", StatisticsIS))
            PQ_WatchingWalkthrough(UEOutput__);
         else
            PQ_WatchingWalkthrough(UEOutput_W);
      } else {
         if (IS_IsIn("noweights", StatisticsIS))
            PQ_WatchingWalkthrough(UEOutput_T);
         else
            PQ_WatchingWalkthrough(UEOutput_TW);
      }
      return;
   }
   else if (IS_IsIn("weight", StatisticsIS)){
      CS_WeightStatisticsStart(StatisticsIS);
      PQ_WatchingWalkthrough(CS_WeightStatProc);
      /* weight:upperBond=..:lowerBound=.. Es werden nur lebende KPs angemeldet */
      CS_WeightStatisticsEnd();
      return;
   }
   else if (IS_IsIn("time", StatisticsIS)){
      CS_TimeDistribution(StatisticsIS);
      return;
   }
   else if (IS_IsIn("standard",StatisticsIS)){
      IO_DruckeKurzUeberschrift("Unselected Equations",'+');
      
      IO_DruckeStatistikZeile("Total number of ues",AnzKPGesamt);
      tiefe1 = IO_Einruecktiefe(0, Maximum(9, AnzKPAusSpezifikation, AnzKPIROpfer,
                                           AnzKPdirektZusammengefuehrt, AnzKPdirektSubsummiert, AnzKPInserted,
					   AnzKPWeggeworfen, AnzKPdirektMNFSubsummiert, AnzKPDatenbank,
                                           AnzKPdirektPermSubsummiert));
      IO_DruckeStatistikZeileEingerueckt("of them from Specification",AnzKPAusSpezifikation, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them from Killed RE", AnzKPIROpfer, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them by   ChaseManhattan", AnzKPDatenbank, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them from Generation", 
                                         AnzKPGesamt-AnzKPAusSpezifikation-AnzKPIROpfer-AnzKPDatenbank, tiefe1);
      IO_DruckeStatistikLeerZeile();
      
      IO_DruckeStatistikZeileEingerueckt("of them joined immediately",AnzKPdirektZusammengefuehrt, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them subsumed immediately", AnzKPdirektSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-subsumed immediately", AnzKPdirektMNFSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed immediately", AnzKPdirektPermSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them thrown away", AnzKPWeggeworfen, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them inserted into PQ", AnzKPInserted, tiefe1);
#if PUNKT_IRP_HACK
      IO_DruckeStatistikZeileEingerueckt("of them reinserted into PQ (P-IRP)", AnzKPPunktIRP, tiefe1); /* Punkt-IRP-Hack */
#endif
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeileEingerueckt("ues selected",AnzKPBehandeltInVervollst, tiefe1);
      
      tiefe2 = IO_Einruecktiefe(tiefe1, Maximum(7, AnzKPspaeterZusammengefuehrt, AnzKPZuRE,
                                                AnzKPZurueckgestellt, AnzKPspaeterSubsummiert, 
                                                AnzKPspaeterMNFSubsummiert, AnzKPspaeterPermSubsummiert,
                                                AnzKPspaeterMNFPermSubsummiert));
      
      IO_DruckeStatistikZeileEingerueckt("of them not orientable, postponed", AnzKPZurueckgestellt, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them joined", AnzKPspaeterZusammengefuehrt, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them subsumed (forward subsumption)", AnzKPspaeterSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-subsumed (forward subsumption)", AnzKPspaeterMNFSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed (forward subsumption)", AnzKPspaeterPermSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-perm-subsumed (forward subsumption)", AnzKPspaeterMNFPermSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them made RE", AnzKPZuRE, tiefe2);
      
      PQ_Statistics (StatisticsIS); /* complete, damit bei 'SUEM' die Submodule auch ausgeben */
      C_Statistics  (StatisticsIS);
      TPR_Statistics(StatisticsIS);
      CG_Statistics (StatisticsIS);
   }
   else if (IS_IsIn("short", StatisticsIS)){
      IO_DruckeKurzUeberschrift("Unselected Equations",'+');
      IO_DruckeStatistikZeile("Generated", AnzKPGesamt);
      IO_DruckeStatistikZeile("Selected", AnzKPBehandeltInVervollst);
   }
   else if (IS_IsIn("config", StatisticsIS)){
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
   }
   else if (IS_IsIn("intern", StatisticsIS)){
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
   }
   else if (IS_IsIn("latex-head", StatisticsIS)){
      fprintf(IO_Strom, "SUE: CPs & spec & IR-vict & GEN: join & subs & cut & inserted & SEL: & postp & join & subs & MNF subs & perm-subs & MNF perm-subs & RE & ");
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
      CG_Statistics (StatisticsIS);
   }
   else if (IS_IsIn("slatex-head", StatisticsIS)){
      fprintf(IO_Strom, "CPs gen & sel & ");
   }
}

void SUE_AktuelleGroesse(unsigned long int *GesamtKPZahl, unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  *GesamtKPZahl = AnzKPGesamt;
  PQ_AktuelleMengenGroesse(AktKPZahl,AktGueltigeKPZahl);
}

static BOOLEAN ParseCGCP (const char *str, int *cg, int *cp)
{
   char *kopie, *kopieDummy, *token;

   kopie = (char *) our_alloc (sizeof (char) * strlen (str));
   strcpy (kopie, str);
   kopieDummy = kopie;
   
  token = strtok (kopie, ":");
  while (token != NULL) {
    if (strncasecmp (token, "cgcp=", strlen ("cgcp=")) == 0) {
      token = strtok (token+strlen("cgcp="), ".");
      *cg = atoi (token);
      token = strtok (NULL, ".:");
      if (token)
	*cp = atoi (token);
      else
	*cp = 1;
      our_dealloc (kopieDummy);
      if (*cg == 0 && *cp == 0)
        our_error ("At least one of the cg and cp parameters for cgcp must be non-zero!");
      return TRUE;
    }
    token = strtok (NULL, ":");
  }

  our_dealloc (kopieDummy);

  return FALSE;
}

static void Auswertung (unsigned long int *cut,
			unsigned long int *modulo, unsigned long int *threshold,
			InfoStringT DEFclas, InfoStringT USRclas,
			InfoStringT DEFsue, InfoStringT USRsue)

{
  int cg, cp;
  *cut = 0;
  *modulo = 2;
  *threshold = 1;
  if (IS_ValueOfOption(DEFclas, USRclas, "cut1"))
    *cut = atol (IS_ValueOfOption (DEFclas, USRclas, "cut1"));
  if (ParseCGCP (USRsue, &cg, &cp) || ParseCGCP (DEFsue, &cp, &cg)) {
    *modulo = cg + cp;
    *threshold = cg;
  } else {
    cp = cg = 1;
  }
  if (!COMP_POWER) {
    fprintf(IO_Strom, "\n* SET OF UNSELECTED EQUATIONS:  Narrowing interval CG= %d, CP= %d\n", cg, cp);
    if (*cut != 0)
      fprintf (IO_Strom,"                                throw away CPs with weight w1 > %ld\n", *cut);
  }
}

void SUE_ParamInfo(ParamInfoArtT what, int whichModule,
                   InfoStringT DEFsue,    InfoStringT USRsue,
                   InfoStringT DEFclas,   InfoStringT USRclas,
                   InfoStringT DEFreclas, InfoStringT USRreclas,
                   InfoStringT DEFcgclas, InfoStringT USRcgclas,
                   InfoStringT DEFpq,     InfoStringT USRpq,
                   InfoStringT DEFtpr,    InfoStringT USRtpr,
		   InfoStringT DEFcg,    InfoStringT USRcg)
{
   if ((whichModule==0) || (whichModule==99)){                /* SUE */
     if (what == ParamEvaluate){
       unsigned long int infoCUT, infoMod, infoTh;
       Auswertung(&infoCUT, &infoMod, &infoTh, DEFclas, USRclas, DEFsue, USRsue);
     }
     else{
       IO_DruckeKurzUeberschrift("Set of unselected equations", '-');
       fprintf(IO_Strom, "1. [cut1=<n>]\n");
       fprintf(IO_Strom, "2. [cgcp=<n1>.<n2>]\n");
       fprintf(IO_Strom, "\n");
       fprintf(IO_Strom, "1. Throw away CPs with primary weight > n\n");
       fprintf(IO_Strom, "2. Intervall for paramodulation with existentially quantified goals.\n");
     }
   }

   if ((whichModule==1) || (whichModule==2) ||(whichModule==3))
     C_ParamInfo(what, whichModule, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
   if (whichModule == 99) {   
      C_ParamInfo(what, 1, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
      C_ParamInfo(what, 2, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
      C_ParamInfo(what, 3, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
   }
   
   if ((whichModule==4) || (whichModule==99)) PQ_ParamInfo(what, DEFpq, USRpq);

   if ((whichModule==5) || (whichModule==99)) TPR_ParamInfo(what, DEFtpr, USRtpr);

   if ((whichModule==6) || (whichModule==99)) TPR_ParamInfo(what, DEFcg, USRcg);
}
      

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ VARIABLES ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static long int cut1 = 0;

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ INITIALIZE --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_InitAposteriori(InfoStringT DEFsue,    InfoStringT USRsue,
                            InfoStringT DEFclas,   InfoStringT USRclas,
                            InfoStringT DEFreclas, InfoStringT USRreclas,
                            InfoStringT DEFcgclas, InfoStringT USRcgclas,
                            InfoStringT DEFpq,     InfoStringT USRpq,
                            InfoStringT DEFtpr,    InfoStringT USRtpr,
                            InfoStringT DEFcg,     InfoStringT USRcg)
{
   if (!TPR_InitAposteriori(DEFtpr, USRtpr))
      return FALSE;

   if (!C_InitAposteriori(DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas))
      return FALSE;

   if (!PQ_InitAposteriori(DEFpq, USRpq))
      return FALSE;

   if (!CG_InitAposteriori(DEFcg, USRcg))
      return FALSE;

   Auswertung(&cut1, &modulo, &threshold, DEFclas, USRclas, DEFsue, USRsue);

   SUE_Demigod.linkeSeite = NULL;
   SUE_Demigod.rechteSeite = NULL;
   SUE_Demigod.Nummer = 0;
   SUE_Demigod.SUEInfo = NULL;
   PQ_NEWPARENT(&SUE_Demigod);

   return TRUE;
}
/* --------------------------------------------------------------------------------------------------------*/
/* ------------ PARENTS -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void SUE_NewParent(RegelOderGleichungsT re, TermT lhsReplica, TermT rhsReplica)
{
   SUE_ActualParent = re;
   PQ_NEWPARENT(re);   /* In PQ: SUE_ActualParent = re;*/
   SUE_ReplicationOfLHS = lhsReplica;
   SUE_ReplicationOfRHS = rhsReplica;
}

void SUE_KillParent(RegelOderGleichungsT re)
/* Vorausgesetzt wird, dass alle Regeln/Gleichungen einen gueltigen Info-Block erhalten haben. */
{
  CG_KillParent(re);
  PQ_KILLPARENT(re);
}

void SUE_RedirectParent(void)
{
   TermT switty = SUE_ReplicationOfLHS;

   SUE_ReplicationOfLHS = SUE_ReplicationOfRHS;
   SUE_ReplicationOfRHS = switty;
   PQ_RedirectParent();         /* SUE_ActualParent ist jetzt noch auf der ausgezeichneten Richtung.*/
   SUE_ActualParent = TP_Gegenrichtung(SUE_ActualParent);
}

void SUE_FinishParent(void)
{
   PQ_FinishParent();
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ GENERATION --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* In den folgenden Funktionen, die jeweils bei Entstehung eines KPs aufgerufen werden,
   wird bei bereits existierenden Termpaaren direkt der StartChain-Befehl erteilt. Bei neu erzeugten
   muss zun"achst ihre Bildung mittels TPR_NewFlatsGenerated() angezeigt werden, wo dann ggf.
   die entsprechende Chain angelegt wird.
*/


void SUE_CPinitial(TermT lhs, TermT rhs)
{
   IncAnzKPGesamt(); IncAnzKPAusSpezifikation();
#if IO_AUSGABEN
   if (IO_level3)
      fprintf(IO_Strom, "initial equation %lu turned into ues %lu.\n", AnzKPAusSpezifikation, AnzKPGesamt);
   else if (IO_level4){
      fprintf(IO_Strom, "initial equation %lu turned into ues %lu.\n", AnzKPAusSpezifikation, AnzKPGesamt);
      IO_DruckeFlex("       %t # %t\n", lhs, rhs);
   }
   else if (IO_level2){
      if (!IO_GUeberschrift){
	 fprintf(IO_Strom, "\nInitial equations:\n" );
	 fprintf(IO_Strom, "------------------\n\n" );
	 IO_GUeberschrift = TRUE;
      }
      fprintf(IO_Strom, "(%4ld)  ", AnzKPAusSpezifikation);
      IO_DruckeFlex("%t = %t\n", lhs, rhs);
   }
#endif
   PCL_GleichungAusDerSpezifikation(lhs, rhs, &SUE_recCPpclID);
   PCL_StartChain();
   SUE_recCPrhs = rhs;
   SUE_recCPlhs = lhs;
   SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
   SUE_recCPactualPosition = NULL;
   SUE_recCPGenerationHistory = Initial;
}

void SUE_CPkilledRE(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclIDPtr MAYBE_UNUSEDPARAM)
{
   IncAnzKPGesamt(); IncAnzKPIROpfer();
   DruckeMeldungZuKPgemacht(AnzKPGesamt, lhs, rhs);
   PCL_StartChain();
   SUE_recCPrhs = rhs;
   SUE_recCPlhs = lhs;
   SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
   SUE_recCPactualPosition = NULL;
   SUE_recCPGenerationHistory = KilledRE;
   PCL_IDZuweisung(&SUE_recCPpclID, *pclIDPtr);
}

void SUE_CPgeneratedWithFather(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
{
   IncAnzKPGesamt(); 
   DruckeMeldungKPgebildet(AnzKPGesamt, TP_ElternNr(SUE_ActualParent), TP_ElternNr(father), lhs, rhs);
   TPR_NewFlatsGenerated();
   PCL_Unifikation(&SUE_recCPpclID, lhs, rhs,
                   father, TP_LinkeSeite(father), actualPosition, SUE_ActualParent);
   SUE_recCPlhs = lhs;
   SUE_recCPrhs = rhs;
   SUE_recCPOtherParent = father;
   SUE_recCPactualPosition = actualPosition;
   SUE_recCPGenerationHistory = CPWithFather;
}


void SUE_CPgeneratedWithMother(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
{
   IncAnzKPGesamt(); 
   DruckeMeldungKPgebildet(AnzKPGesamt, TP_ElternNr(SUE_ActualParent), TP_ElternNr(mother), lhs, rhs);
   TPR_NewFlatsGenerated();
   PCL_Unifikation(&SUE_recCPpclID, lhs, rhs,
                   SUE_ActualParent, SUE_ReplicationOfLHS, actualPosition, mother);
   SUE_recCPlhs = lhs;
   SUE_recCPrhs = rhs;
   SUE_recCPOtherParent = mother;
   SUE_recCPactualPosition = actualPosition;
   SUE_recCPGenerationHistory = CPWithMother;
}

void SUE_CPgeneratedWithDatabase(RegelOderGleichungsT Konklusion)
{
  SUE_recCPlhs = TP_LinkeSeite(Konklusion);
  SUE_recCPrhs = TP_RechteSeite(Konklusion);
  IncAnzKPGesamt(); 
  IncAnzKPDatenbank();
  DruckeMeldungKPmitDBgebildet(AnzKPGesamt, SUE_recCPlhs, SUE_recCPrhs);
  PCL_StartChain();
  PCL_DatenbankKonklusion(Konklusion);
  SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
  SUE_recCPactualPosition = NULL;
  SUE_recCPGenerationHistory = ChaseManhattanCP;
  PCL_IDZuweisung(&SUE_recCPpclID, COMP_PCL ? *TP_PclID(Konklusion) : 0);
}

void SUE_recentCPjoined(void)
{
  KPZusammengefuehrtDrucken (SUE_recCPlhs);
  PCL_DeleteChain();
  IncAnzKPdirektZusammengefuehrt();
  TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}

void SUE_recentCPMNFjoined(TermT joinedTerm)
{
   KPZusammengefuehrtDrucken (joinedTerm);
   TO_TermLoeschen(joinedTerm);
   PCL_DeleteChain();
   IncAnzKPdirektZusammengefuehrt();/* neuer Zaehler? */
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs); 
}

void SUE_recentCPsubsumed(void)
{
   KPSubsummiertDrucken ();
   PCL_DeleteChain();
   IncAnzKPdirektSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPMNFsubsumed(void)
{
   KPSubsummiertDrucken();
   PCL_DeleteChain();
   IncAnzKPdirektMNFSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}

void SUE_recentCPpermsubsumed(void)
{
   KPPermSubsummiertDrucken ();
   PCL_DeleteChain();
   IncAnzKPdirektPermSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPMNFpermsubsumed(void)
{
   KPPermSubsummiertDrucken();
   PCL_DeleteChain();
   IncAnzKPdirektMNFPermSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPinsert(void)
{
   TermpairT tp;
   WeightEntryT wt;
   TermT lhs,rhs;
   BOOLEAN IsCritGoal = Z_UngleichungEntmantelt(&lhs,&rhs,SUE_recCPlhs,SUE_recCPrhs);
   if (IsCritGoal){
     C_CG_Classify (&wt, lhs, rhs, SUE_ActualParent, SUE_recCPOtherParent);
   }
   else {
     C_Classify (&wt, SUE_recCPlhs, SUE_recCPrhs, SUE_ActualParent, SUE_recCPOtherParent, 
		 SUE_recCPactualPosition, SUE_recCPGenerationHistory);
     
     if (cut1 && wt.w1 > cut1) {      /* throw CP away */
        HK_fair = FALSE;
        DruckeMeldungKPWeggeworfen ();
        PCL_DeleteChain();
        IncAnzKPWeggeworfen();
        TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
        return;
     }
   }

   DruckeMeldungAlsKPaufgenommen(SUE_recCPlhs, SUE_recCPrhs);
   tp = TPR_FLATTERMStoTERMPAIR(SUE_recCPlhs, SUE_recCPrhs, SUE_ActualParent, SUE_recCPOtherParent, SUE_recCPactualPosition, SUE_recCPpclID);
   
   if (IsCritGoal)
     CG_Insert(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   else if (IsDemigod(SUE_recCPOtherParent))
     PQ_INSERTImmortal(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   else
     PQ_INSERT(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   
   IncAnzKPInserted();
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ SELECTION ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_SelectInitial(void)
/* Liefert alle unsterblichen KPs zurueck, die zusaetzlich ein (absolut) minimales Gewicht haben. */
{
   if (PQ_SELECTINITIAL(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent)){

     TPR_TERMPAIRtoPARALLEL_FINAL(SUE_selCPTermpair, &SUE_selCPlhs, &SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, &SUE_selCPpclID);
      
     KPwirdBehandeltDrucken(TP_ElternNr(SUE_selCPactualParent), TP_ElternNr(SUE_selCPotherParent), SUE_selCPlhs, SUE_selCPrhs);
     
     IncAnzBehandelterKPsVervollst();
     
     return TRUE;
   }
   else
     return FALSE;
}

static void (* IRProc) (TermT, TermT);

static void TouchOnlyImmortals(WeightEntryT wt MAYBE_UNUSEDPARAM, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
  if (IsDemigod(otherP)){
    TermT lhs, rhs;
    unsigned long int dummyPCL_ID;
    TPR_TERMPAIRtoPARALLEL_INTERMED(tp, &lhs, &rhs, actualP, otherP, &dummyPCL_ID);
    IRProc(lhs, rhs);
    /*
      XXXXXXXXXX
      TPR_PARALLELtoTERMPAIR kann tp aendern - und gibt tp dann zurueck.
       Ich bin mir deswegen nicht sicher ob der folgende Aufruf so korrekt ist.  Da diese
       Funktion aber zur Zeit nicht benutzt wird, werde ich dies jetzt nicht aendern.
       AJ. 17.3.98. XXXXXXXXX */
    TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualP, otherP, FALSE, dummyPCL_ID); /* nicht veraendert */
  }
}

void SUE_TouchImmortals(void (* Proc) (TermT, TermT))
/* Liefert alle unsterblichen KPs zurueck, unabhaengig vom Gewicht. */
{
  IRProc = Proc;
  PQ_WatchingWalkthrough(TouchOnlyImmortals);
}


#if PUNKT_IRP_HACK
WeightEntryT AltesGewicht; /* Punkt-IRP-Hack */
#endif

typedef enum {GarNix, EinCP, EinCG} WasDennT;

static WasDennT WasSelektiert(void)
{
  static BOOLEAN (**Selektoren[3])(TermpairT*,RegelOderGleichungsT*,RegelOderGleichungsT*) = {NULL,&PQ_SelectProc,&CG_SelectProc};
# define /*BOOLEAN*/ Selected(/*{EinCP;EinCG}*/ Gattung)                                                            \
    (**Selektoren[Gattung])(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent)

  if (PM_Existenzziele()) {
    WasDennT Erstgattung = (SUE_AnzAktivierterFakten % modulo < threshold) ? EinCG : EinCP,
             Zweitgattung = 3-Erstgattung;
    return Selected(Erstgattung) ? Erstgattung : Selected(Zweitgattung) ? Zweitgattung : GarNix;
  }
  else
    return Selected(EinCP) ? EinCP : GarNix;
} 
#if !PUNKT_IRP_HACK

BOOLEAN SUE_Select(void)
{
  if (WasSelektiert()==GarNix) 
    return FALSE;
  else {
    TPR_TERMPAIRtoPARALLEL_FINAL(SUE_selCPTermpair, &SUE_selCPlhs, &SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, &SUE_selCPpclID);
    KPwirdBehandeltDrucken(TP_ElternNr(SUE_selCPactualParent), TP_ElternNr(SUE_selCPotherParent), SUE_selCPlhs, SUE_selCPrhs);
    IncAnzBehandelterKPsVervollst();
    return TRUE;
  }
}

#else



#define Auspacken()   TPR_TERMPAIRtoPARALLEL_INTERMED(SUE_selCPTermpair, &SUE_selCPlhs, &SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, &SUE_selCPpclID)
#define Einpacken()   TPR_PARALLELtoTERMPAIR(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, TRUE, SUE_selCPpclID)
#define Wegwerfen()   TPR_DESTROYTERMPAIR (SUE_selCPTermpair, SUE_selCPactualParent, SUE_selCPotherParent)   /*Lesbarkeit++*/

extern void KPNormalisierung_RE(TermT, TermT, PCL_ObjektNummerT*);
/* XXX unsauber - muesste sich aendern mit dem Parameter von -kg  */
/* XXX Funktion ist jetzt in Hauptkomponenten.c nicht static! */
/* XXX Der Select-Zaehler muesste ggf. auch hochgesetzt werden! */

static void Nachwiegen(WasDennT WasDenn, WeightEntryT *wt_neu)
{
  TermT lhs, rhs;
  Auspacken();
  KPNormalisierung_RE(SUE_selectedCPlhs(), SUE_selectedCPrhs(), SUE_selectedCPPclID());
  switch (WasDenn) {
  case EinCP:
     C_ReClassify (wt_neu, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
     break;
  case EinCG:
     if (!Z_UngleichungEntmantelt(&lhs,&rhs,SUE_selCPlhs,SUE_selCPrhs))
       lhs = SUE_selCPlhs, rhs = SUE_selCPrhs;
     C_CG_Classify (wt_neu, lhs, rhs, SUE_selCPactualParent, SUE_selCPotherParent);
  default:
     break;
   }
}

BOOLEAN SUE_Select(void)
{
  WasDennT EtwasSelektiert;
  WeightEntryT wt_neu;

  /* XXX AltesGewicht wird aus PQ_Select/CG_Select rausgereicht werden - HACK ALERT */
  /* XXX Achtung HeapOfHeaps setzt AltesGewicht nicht */
  while (GarNix != (EtwasSelektiert = WasSelektiert())) {  /* evtl. Abfrage, ob dieser Kode ausgefuehrt werden soll */
     Nachwiegen(EtwasSelektiert,&wt_neu);
#if 0
     if (AltesGewicht.w1!=LONG_MIN && wt_neu.w1>AltesGewicht.w1) {                             /* wieder aufnehmen */
#else
     if (AltesGewicht.w1!=LONG_MIN && wt_neu.w1>AltesGewicht.w1
         && (double)wt_neu.w1/(double)AltesGewicht.w1 > 3) {                             /* wieder aufnehmen */
#endif
       TermpairT tp = Einpacken();
       PQEinfuegerT Einfueger = EtwasSelektiert==EinCG ? CG_Insert : 
         IsDemigod(SUE_selCPotherParent) ? PQ_InsertProc : PQ_InsertImmortalProc;
       AnzKPPunktIRP++;
       printf ("Punkt-IRP: alt:%d neu:%d Quotient: %.2f\n", AltesGewicht.w1, wt_neu.w1, (double)wt_neu.w1/(double)AltesGewicht.w1);
       (*Einfueger)(tp, wt_neu, SUE_selCPactualParent, SUE_selCPotherParent);
     } 
     else {
       /* Wegwerfen(); */
       KPwirdBehandeltDrucken(TP_ElternNr(SUE_selCPactualParent), TP_ElternNr(SUE_selCPotherParent), SUE_selCPlhs, SUE_selCPrhs);
       IncAnzBehandelterKPsVervollst();
       return TRUE;
     }
  }
  return FALSE;
}
#endif


void SUE_selectedCPjoined(void)
{
   KPZusammengefuehrtDrucken (SUE_selCPlhs);
   TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
   IncAnzKPspaeterZusammengefuehrt();
}

void SUE_selectedCPMNFjoined(TermT joinedTerm)
{
   KPZusammengefuehrtDrucken (joinedTerm);
   TO_TermLoeschen(joinedTerm);
   TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
   IncAnzKPspaeterZusammengefuehrt();
}


void SUE_selectedCPsubsumed(void)
{
    KPSubsummiertDrucken();
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterSubsummiert();
}


void SUE_selectedCPMNFsubsumed(void)
{
    KPSubsummiertDrucken();
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterMNFSubsummiert();
}

void SUE_selectedCPpermsubsumed(void)
{
    KPPermSubsummiertDrucken();
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterPermSubsummiert();
}


void SUE_selectedCPMNFpermsubsumed(void)
{
    KPPermSubsummiertDrucken();
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterMNFPermSubsummiert();
}


void SUE_selectedCPRE(void)
{
    TPR_PARALLELtoFLATTERMS(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    SUE_AnzAktivierterFakten++;
    if (Z_IstGemantelteUngleichung(SUE_selCPlhs, SUE_selCPrhs)) {
      SUE_AnzAktivierterG++;
      if (SUE_LastSelectedWasFifo)
        SUE_AnzAktivierterGFifo++;
      else 
        SUE_AnzAktivierterGHeu++;
    } else {
      SUE_AnzAktivierterRE++;
      if (SUE_LastSelectedWasFifo)
        SUE_AnzAktivierterREFifo++;
      else
        SUE_AnzAktivierterREHeu++;
    }
    IncAnzKPZuRE();
}

void SUE_selectedCPMNFRE(void)
{
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    SUE_AnzAktivierterFakten++;
    if (Z_IstGemantelteUngleichung(SUE_selCPlhs, SUE_selCPrhs)) {
      SUE_AnzAktivierterG++;
      if (SUE_LastSelectedWasFifo++)
        SUE_AnzAktivierterGFifo++;
      else
        SUE_AnzAktivierterGHeu++;
    } else {
      SUE_AnzAktivierterRE++;
      if (SUE_LastSelectedWasFifo)
        SUE_AnzAktivierterREFifo++;
      else
        SUE_AnzAktivierterREHeu++;
    }
    IncAnzKPZuRE();
}


BOOLEAN SUE_selectedCPpostponed(void)
{
    return FALSE;
}




/* --------------------------------------------------------------------------------------------------------*/
/* ------------ WALKTHROUGH -------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

unsigned long int SUE_AnzKPsToGo(void)
{
   return PQ_AnzKPsToGo();
}

void SUE_WALKTHROUGH(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
{
   PQ_WALKTHROUGH(AP, NeedsTerms);
}

void SUE_TermeVerfolgen(void (*OperationAufTermen) (TermT, TermT))
/* Alle Terme. Achtung: Geht nicht mit CGs!!!  */
{
   PQ_TermeVerfolgen(OperationAufTermen);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void SUE_CLEANUP(void)
{
   PQ_Cleanup();
   PQ_KILLPARENT(&SUE_Demigod);
   TPR_CleanUp();
   CG_CleanUp();
   C_CleanUp();
}


void SUE_SOFTCLEANUP(void)
{
  PQ_SoftCleanup();             /* Strahlen bei Heap und HeapOfHeaps loeschen */
  /* KillParent leert nur den Info-Eintrag in SUE_Demigod; dieser wird aber ohnehin geklaert. */
  /* TPR_Cleanup tut eh' nix */
  CG_SoftCleanUp();
  C_CleanUp();                  /* bleibt noetig, weil CStatistik ueber our_alloc laeuft! */
}
