/*=================================================================================================================
  =                                                                                                               =
  =  Unifikation1.c                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =  07.10.94  Thomas.                                                                                            =
  =  07.02.95  Thomas. Rekursionen eliminiert.                                                                    =
  =                                                                                                               =
  =================================================================================================================*/


#include "compiler-extra.h"
#include "general.h"
#include "Ordnungen.h"
#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "FehlerBehandlung.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Konsulat.h"
#include "KPBildung.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "Packen.h"
#include "Parameter.h"
#include "PhilMarlow.h"
#include "SpeicherParameter.h"
#include <string.h>
#include "SUEM.h"
#include "TermIteration.h"
#include "Termpaare.h"
#include "Unifikation1.h"
#include "Unifikation2.h"
#include "WaldmeisterII.h"
#include "Zielverwaltung.h"



TI_IterationDeklarieren(U1)
TI2_IterationDeklarieren(U1)



/*=================================================================================================================*/
/*= 0. Statistik-Variablen und Makros                                                                             =*/
/*=================================================================================================================*/

#define U1_STATISTIK normal_power_comp(0,0,0)
/* Alternativen: 0 fuer aus, 1 fuer an */

static unsigned long AnzKPOpferKapur = 0;       /* KPs, die dem Kapur-Kriterium bei ihrer Bildung zum Opfer fielen */
static unsigned long AnzKPGG = 0;                                                    /* KPs zwischen 2 Gleichungen */
static unsigned long AnzKPRR = 0;                                                         /* KPs zwischen 2 Regeln */
static unsigned long AnzKPRG = 0;                                           /* KPs zwischen Regeln und Gleichungen */
static unsigned long AnzUnifVersRegelnToplevel = 0;       /* Unifikationsversuche mit der Regelmenge auf Toplevel. */
static unsigned long AnzUnifVersRegelnTeilterme = 0;    /* Unifikationsversuche mit den Teiltermen der Regelmenge. */
static unsigned long AnzUnifVersRegelnEigeneTterme = 0; /* Unifikationsversuche mit den eigenen Teiltermen (Regel) */
static unsigned long AnzUnifErfRegelnToplevel = 0;    /* Erfolgreiche Unifikationsversuche mit der R auf Toplevel. */
static unsigned long AnzUnifErfRegelnTeilterme = 0; /* Erfolgreiche Unifikationsversuche mit den Teiltermen der R. */
static unsigned long AnzUnifErfRegelnEigeneTterme = 0;      /* Erfolgreiche Unifikationsversuche den eigenen Teiltermen (Regel). */
static unsigned long AnzUnifVersGleichgnToplevel = 0;/* Unifikationsversuche mit der Gleichungsmenge auf Toplevel. */
static unsigned long AnzUnifVersGleichgnTeilterme = 0;         /* Versuche mit den Teiltermen der Gleichungsmenge. */
static unsigned long AnzUnifVersGleichgnEigeneTterme = 0;        /* Versuche mit den eigenen Teiltermen (Gleichg.) */
static unsigned long AnzUnifErfGleichgnToplevel = 0;              /* Erfolge mit der Gleichungsmenge auf Toplevel. */
static unsigned long AnzUnifErfGleichgnTeilterme = 0;           /* Erfolge mit den Teiltermen der Gleichungsmenge. */
static unsigned long AnzUnifErfGleichgnEigeneTterme = 0;         /* Erfolge mit den eigenen Teiltermen (Gleichg.). */
static unsigned long AnzUnifMisserfolgeOrdnungRG = 0;       /* Wegen >-Test nicht gebildete KPs zwischen Regel und Gleichung */
static unsigned long AnzUnifMisserfolgeOrdnungGG = 0;       /* Wegen >-Test nicht gebildete KPs zwischen zwei Gleichungen */
static unsigned long AnzUnifUeberlappungenGebildet = 0;                    /* Fuer >-Test gebildete Ueberlappungen */
static unsigned long AnzUnifUeberflKritTerme = 0;                /* Ueberfluessigerweise gebildete kritische Terme */
static unsigned long AnzNonEinsSternKPs = 0;              /* Anzahl der Paare, die dem 1*-Kriterium nicht genuegen */
static unsigned long AnzNonNusfuKPs = 0;               /* Anzahl der Ueberlappungen unterhalb von Skolemfunktionen */

#if U1_STATISTIK
  #define IncAnzKapurOpfer() (AnzKPOpferKapur++)
  #define IncAnzKPRR() (AnzKPRR++)
  #define IncAnzKPGG() (AnzKPGG++)
  #define IncAnzKPRG() (AnzKPRG++)
  #define IncAnzUnifVersRegelnToplevel() (AnzUnifVersRegelnToplevel++)  
  #define IncAnzUnifVersRegelnTeilterme() (AnzUnifVersRegelnTeilterme++)  
  #define IncAnzUnifVersRegelnEigeneTterme() (AnzUnifVersRegelnEigeneTterme++)  
  #define IncAnzUnifErfRegelnToplevel() (AnzUnifErfRegelnToplevel++)
  #define IncAnzUnifErfRegelnTeilterme() (AnzUnifErfRegelnTeilterme++)
  #define IncAnzUnifErfRegelnEigeneTterme() (AnzUnifErfRegelnEigeneTterme++)
  #define IncAnzUnifVersGleichgnToplevel() (AnzUnifVersGleichgnToplevel++)                
  #define IncAnzUnifVersGleichgnTeilterme() (AnzUnifVersGleichgnTeilterme++)
  #define IncAnzUnifVersGleichgnEigeneTterme() (AnzUnifVersGleichgnEigeneTterme++)  
  #define IncAnzUnifErfGleichgnToplevel() (AnzUnifErfGleichgnToplevel++)
  #define IncAnzUnifErfGleichgnTeilterme() (AnzUnifErfGleichgnTeilterme++)
  #define IncAnzUnifErfGleichgnEigeneTterme() (AnzUnifErfGleichgnEigeneTterme++)
  #define IncAnzUnifMisserfolgeOrdnungRG() (AnzUnifMisserfolgeOrdnungRG++)
  #define IncAnzUnifMisserfolgeOrdnungGG() (AnzUnifMisserfolgeOrdnungGG++)
  #define IncAnzUnifUeberlappungenGebildet() (AnzUnifUeberlappungenGebildet++)
  #define IncAnzUnifUeberflKritTerme() (AnzUnifUeberflKritTerme++)
  #define IncAnzNonEinsSternKPs() (AnzNonEinsSternKPs++)
  #define IncAnzNonNusfuKPs() (AnzNonNusfuKPs++)
#else
  #define IncAnzKapurOpfer()                                                                            /* Nix Tun */
  #define IncAnzKPRR()                                                                                  /* Nix Tun */
  #define IncAnzKPGG()                                                                                  /* Nix Tun */
  #define IncAnzKPRG()                                                                                  /* Nix Tun */
  #define IncAnzUnifVersRegelnToplevel()                                                                /* nix tun */
  #define IncAnzUnifVersRegelnTeilterme()                                                               /* nix tun */
  #define IncAnzUnifVersRegelnEigeneTterme()                                                            /* nix tun */
  #define IncAnzUnifErfRegelnToplevel()                                                                 /* nix tun */
  #define IncAnzUnifErfRegelnTeilterme()                                                                /* nix tun */
  #define IncAnzUnifErfRegelnEigeneTterme()                                                             /* nix tun */
  #define IncAnzUnifVersGleichgnToplevel()                                                              /* nix tun */
  #define IncAnzUnifVersGleichgnTeilterme()                                                             /* nix tun */
  #define IncAnzUnifVersGleichgnEigeneTterme()                                                          /* nix tun */
  #define IncAnzUnifErfGleichgnToplevel()                                                               /* nix tun */
  #define IncAnzUnifErfGleichgnTeilterme()                                                              /* nix tun */
  #define IncAnzUnifErfGleichgnEigeneTterme()                                                           /* nix tun */
  #define IncAnzUnifMisserfolgeOrdnungRG() 0                                                            /* nix tun */
  #define IncAnzUnifMisserfolgeOrdnungGG() 0                                                            /* nix tun */
  #define IncAnzUnifUeberlappungenGebildet()                                                            /* nix tun */
  #define IncAnzUnifUeberflKritTerme()                                                                  /* nix tun */
  #define IncAnzNonEinsSternKPs()                                                                       /* nix tun */
  #define IncAnzNonNusfuKPs()                                                                           /* nix tun */
#endif



/*=================================================================================================================*/
/*= I. Variablenbank fuer Substitutionen; Stack angelegter Bindungen; sonstige globale Groessen                   =*/
/*=================================================================================================================*/

static int xUDimension;                              /* neueste bei gegenwaertiger Ausdehnung zu bindende Variable */
typedef struct {
  UTermT Substitut;
} U1_EintragsT;
static U1_EintragsT *xU;


static unsigned int StackZeiger;
typedef SymbolT StackElementT;
static StackElementT *BindungsStack;
/* StackZeiger ist stets der Index des letzten beschriebenen Elements.
   Das nullte Element wird nicht beschrieben.
   -xUDimension ist der Index des letzten bei gegenwaertiger Ausdehnung noch zu schreibenden Elements.
*/

#define /*void*/ PushVariable(/*SymbolT*/ Variable)                                                                 \
  BindungsStack[++StackZeiger] = Variable;

#define /*unsigned int*/ PopVariable(/*void*/)                                                                      \
  (BindungsStack[StackZeiger--])

#define /*BOOLEAN*/ SchonVariablenGebunden(/*void*/)                                                                \
  (StackZeiger)


static unsigned int Endestapel1Zeiger,
                    Endestapel2Zeiger,
                    Endestapel3Zeiger;
typedef struct {
  UTermT Weiter,
         Ende;
} EndeEintragsT;
static EndeEintragsT *Endestapel1,
                     *Endestapel2,
                     *Endestapel3;                                     /* Endestapel laufen ebenfalls ab Element 1 */


void U1_VariablenfelderAnpassen(SymbolT DoppelteNeuesteBaumvariable)   /* Variablenbank und Stapel angelegter Bin- */
{    /* dungen so anpassen, dass Variablen x1, x2, ... , x(DoppelteNeuesteBaumvariable) verdaut werden. Darf nicht */
  if (SO_VariableIstNeuer(DoppelteNeuesteBaumvariable,xUDimension)) { /* waehrend des Unifizierens aufgerufen wer- */
    SymbolT Index;                                                                      /* den, sondern nur davor. */
    unsigned int n;
    negarray_dealloc(xU,-xUDimension);
    xUDimension = SO_VarDelta(DoppelteNeuesteBaumvariable,U1_AnzErsterVariablen);
    n = SO_VarNummer(xUDimension);
    xU = negarray_alloc(n,U1_EintragsT,n);
    SO_forSymboleAbwaerts(Index,SO_ErstesVarSymb,xUDimension)
      xU[Index].Substitut = NULL;
    posarray_realloc(BindungsStack,n,StackElementT,1);
    posarray_realloc(Endestapel1,n,EndeEintragsT,1);
    posarray_realloc(Endestapel2,n,EndeEintragsT,1);
    posarray_realloc(Endestapel3,n,EndeEintragsT,1);
  }
}


static KantenT BaumIndex;                                     /* Zeiger fuer Durchlauf des Baumes beim Unifizieren */
#define aktTermrest BK_aktTIndexU(BaumIndex)
#define SymbolAusTerm TO_TopSymbol(aktTermrest)                 /* naechstes zu matchendes Zeichen aus Anfrageterm */

static unsigned int AnzImKnotenAngelegterBindungen;                                     /* wie der Name schon sagt */
static BOOLEAN AbstiegMitZielErfolgt;
static UTermT TermNachZiel;

/* Zur Verwaltung der Groesse AnzImKnotenAngelegterBindungen

   Initialwert wird in U1_InitApriori gesetzt und ist 0.
   Dieser Wert wird nach dem Verlassen von U1_KPsBilden wieder angenommen.

   Unifikation erfolgt in den drei Kategorien
   -MitEigenenTeiltermen,
   -MitDSBaumTeiltermen,
   -MitDSBaum.

   Jede dieser drei Kategorien muss nach dem Verlassen die Variable wieder
   auf 0 setzen. Fuer die ersten beiden Kategorien ist dies trivial und
   geschieht durch Aufruf von KnotenbindungenAufheben() nach jedem Aufruf
   von Unifiziert, also auch nach dem letzten.

   Unifizieren mit dem Baum: Wird ein Knoten betrachtet, so gibt es mehre-
   re Moeglichkeiten, aus diesem Knoten weiterzukommen. Diese sind unter-
   schieden nach der Beschaffenheit des aktuellen Symbols aus dem Anfrage-
   term. Der Knoten kann nach unten oder oben verlassen werden. In diesem
   Fall wird AnzIKAB auf 0 gesetzt; ggf. werden Bindungen aufgehoben.
   Damit diese Entscheidung zustande kommen kann, werden eine Reihe von
   Abstiegsvarianten ueberprueft. Nur fuer die erfolglose Unifikation sind
   dabei Bindungen aufzuheben und AnzIKAB rueckzusetzen. Die Funktion
   Unifiziert eine Variante MGUMitBlattGefunden. Deren Veraenderungen von
   MGU und AnzIKAB werden aber durch AusBlattAufsteigen() explizit rueck-
   gaengig gemacht.

   Da bei jedem Aufstieg zu einem Knoten AnzIKAB den Wert 0 hat, gilt dies
   auch beim letzten Aufstieg zur Wurzel. Die dort unternommenen Versuche,
   nochmals abzusteigen, schlagen fehl und belassen AnzIKAB den Wert 0.
*/



/*=================================================================================================================*/
/*= II. Operationen auf oder mit Substitutionen                                                                   =*/
/*=================================================================================================================*/

#define /*UTermT*/ SubstitutVon(/*SymbolT*/ VarSymbol)                                                              \
  (xU[VarSymbol].Substitut)


#define /*BOOLEAN*/ VarGebunden(/*SymbolT*/ VarSymbol)                                                              \
  (SubstitutVon(VarSymbol))


#define /*BOOLEAN*/ SymbGebundeneVar(/*SymbolT*/ Symbol)                                                            \
  (SO_SymbIstVar(Symbol) && (VarGebunden(Symbol)))


static void BindungenAufheben(unsigned int Anzahl)
{
  while (Anzahl--)
    xU[PopVariable()].Substitut = NULL;
}


#define /*void*/ Entbinden(/*SymbolT*/ Topsymbol, /*UTermT*/ Term)                                                  \
  while (SymbGebundeneVar(Topsymbol))                                                                               \
    Topsymbol = TO_TopSymbol(Term = SubstitutVon(Topsymbol))


#define /*void*/ FertigEntbindenMitEnde(/*SymbolT*/ Topsymbol, /*UTermT*/ TermAlt, /*UTermT*/ TermAltEnde)          \
{                                                                                                                   \
  do                                                                                                                \
    Topsymbol = TO_TopSymbol(TermAlt = SubstitutVon(Topsymbol));                                                    \
  while (SymbGebundeneVar(Topsymbol));                                                                              \
  TermAltEnde = TO_TermEnde(TermAlt);                                                                               \
}


#define /*void*/ KnotenbindungenAufheben(/*void*/)                                                                  \
 BindungenAufheben(AnzImKnotenAngelegterBindungen);                                                                 \
 AnzImKnotenAngelegterBindungen = 0;



/*=================================================================================================================*/
/*= III. Operationen zum flachen Unifizieren                                                                      =*/
/*=================================================================================================================*/

/* Verwaltung der Endestapel */

#define /*void*/ Endestapel12Initialisieren(/*void*/)                                                               \
  Endestapel1Zeiger = Endestapel2Zeiger = 0;


#define /*void*/ Endestapel1Initialisieren(/*void*/)                                                                \
  Endestapel1Zeiger = 0;


#define /*void*/ Endestapel3Initialisieren(/*void*/)                                                                \
  Endestapel3Zeiger = 0;


#define /*void*/ PushEndestapel(/*{1;2;3}*/ Wohin, /*UTermT*/ Rest, /*UTermT*/ Enderest)                            \
{                                                                                                                   \
  Endestapel ## Wohin [++Endestapel ## Wohin ## Zeiger].Weiter = Rest;                                              \
  Endestapel ## Wohin [Endestapel ## Wohin ## Zeiger].Ende = Enderest;                                              \
}


#define /*void*/ PopEndestapelIn(/*{1;2;3}*/ Woher, /*UTermT*/ Rest, /*UTermT*/ Enderest)                           \
{                                                                                                                   \
  Rest = Endestapel ## Woher [Endestapel ## Woher ## Zeiger].Weiter;                                                \
  Enderest = Endestapel ## Woher [Endestapel ## Woher ## Zeiger--].Ende;                                            \
}


#define /*BOOLEAN*/ EndestapelLeer(/*{1;2;3}*/ Welcher)                                                             \
  (!Endestapel ## Welcher ## Zeiger)


static BOOLEAN DoppeltWeitergeschaltet12(UTermT *Term1, UTermT *Term1Ende, UTermT *Term2, UTermT *Term2Ende)
{
  while (TO_PhysischGleich(*Term1,*Term1Ende))
    if (EndestapelLeer(1))
      return FALSE;
    else
      PopEndestapelIn(1,*Term1,*Term1Ende);
  while (TO_PhysischGleich(*Term2,*Term2Ende))
    PopEndestapelIn(2,*Term2,*Term2Ende);
  *Term1 = TO_Schwanz(*Term1);
  *Term2 = TO_Schwanz(*Term2);
  return TRUE;
}


static BOOLEAN Weitergeschaltet3(UTermT *Term, UTermT *EndeTerm)
{
  while (TO_PhysischGleich(*Term,*EndeTerm))
    if (EndestapelLeer(3))
      return FALSE;
    else
      PopEndestapelIn(3,*Term,*EndeTerm);
  *Term = TO_Schwanz(*Term);
  return TRUE;
}


#define /*Statements*/ RumpfWeiterschalten(/*{1;2;3}*/ StapelNr, /*UTermT*/ Term, /*UTermT*/ Schluss)               \
  while (TO_PhysischGleich(Term,Schluss) && !EndestapelLeer(StapelNr))                                              \
    PopEndestapelIn(StapelNr,Term,Schluss);                                                                         \
  Term = TO_Schwanz(Term);                                                                                          \


#define /*void*/ Weiterschalten3(/*UTermT*/ Term, /*UTermT*/ Schluss)                                               \
RumpfWeiterschalten(3,Term,Schluss)


#define /*void*/ Weiterschalten1(/*UTermT*/ Term, /*UTermT*/ Schluss)                                               \
RumpfWeiterschalten(1,Term,Schluss)


static BOOLEAN SubstErgaenzt(SymbolT Variable, UTermT Term)
{                 /* Vorbedingung: Variable darf noch nicht gebunden sein. Macht dann occur check und passt xU an. */
  if (SchonVariablenGebunden()) {
    SymbolT Topsymbol = TO_TopSymbol(Term);                                        /* Bindung x(i)<-x(i) abfangen; */
    Entbinden(Topsymbol,Term);                                        /* klappt naemlich immer und bewirkt nichts. */
    if (SO_SymbGleich(Topsymbol,Variable))
      return TRUE;
    /* Ausserdem wurde, falls Term=x(i), x(i)<-x(j), x(j)<-x(k), ... gleich das am weitesten aussubstituierte Sub-
       stitut gewaehlt, das ohne Anlegen neuer Termzellen gebildet werden kann.
       Term ist nun sicher keine gebundene Variable mehr.  */
    if (SO_SymbIstFkt(Topsymbol)) {                         /* wenn Topsymbol Variable, dann frei => gleich binden */
      UTermT TermIndex = Term,                            /* sonst Term durchlaufen, aussubstituieren, occur check */
             TermIndexEnde = TO_TermEnde(Term);
      SymbolT Symbol;
      Endestapel3Initialisieren();
      while (Weitergeschaltet3(&TermIndex,&TermIndexEnde)) {
      Start:
        Symbol = TO_TopSymbol(TermIndex);
        if (SO_SymbIstVar(Symbol)) {
          if (SO_SymbGleich(Symbol,Variable))
            return FALSE;
          else if (VarGebunden(Symbol)) {
            PushEndestapel(3,TermIndex,TermIndexEnde);
            TermIndexEnde = TO_TermEnde(TermIndex = SubstitutVon(Symbol));
            goto Start;
          }
	}
      }
    }
  } /* Da die miteinander zu unifizierenden Terme variablendisjunkt sind,
       kann man sich bei der ersten Bindung den Occur check sparen. */

  xU[Variable].Substitut = Term;
  PushVariable(Variable);
  AnzImKnotenAngelegterBindungen++;
  return TRUE;
}


#define /*void*/ AufTeiltermende(/*{1;2}*/ Wo)                                                                      \
  Term ## Wo = TO_TermEnde(Term ## Wo)


#define /*void*/ AufSubstitut(/*{1;2}*/ Wo)                                                                         \
  PushEndestapel(Wo,Term ## Wo,Term ## Wo ## Ende);                                                                 \
  Term ## Wo ## Ende = TO_TermEnde(Term ## Wo = SubstitutVon(Symbol ## Wo));                                        \
  Symbol ## Wo = TO_TopSymbol(Term ## Wo);


#define /*BOOLEAN*/ RumpfUnifiziertBis(/*UTermT*/ Term1, /*UTermT*/ Term1Ende,/*UTermT*/ Term2,/*UTermT*/ Term2Ende)\
{                                                                                                                   \
  Endestapel12Initialisieren();                                                                                     \
  do {                                                                                                              \
    SymbolT Symbol1 = TO_TopSymbol(Term1),                                                                          \
            Symbol2 = TO_TopSymbol(Term2);                                                                          \
  Start:                                                                                                            \
    if (SO_SymbUngleich(Symbol1,Symbol2)) {                                                                         \
      if (SO_SymbIstVar(Symbol1)) {                                                                                 \
        if (VarGebunden(Symbol1)) {                                                                                 \
          AufSubstitut(1);                                                                                          \
          goto Start;                                                                                               \
        }                                                                                                           \
        else if (SubstErgaenzt(Symbol1,Term2))                                                                      \
          AufTeiltermende(2);                                                                                       \
        else                                                                                                        \
          return FALSE;                                                                                             \
      }                                                                                                             \
      else if (SO_SymbIstVar(Symbol2)) {                                                                            \
        if (VarGebunden(Symbol2)) {                                                                                 \
          AufSubstitut(2);                                                                                          \
          goto Start;                                                                                               \
        }                                                                                                           \
        else if (SubstErgaenzt(Symbol2,Term1))                                                                      \
          AufTeiltermende(1);                                                                                       \
        else                                                                                                        \
          return FALSE;                                                                                             \
      }                                                                                                             \
      else                                                                                                          \
        return FALSE;                                                                                               \
    }                                                                                                               \
  } while (DoppeltWeitergeschaltet12(&Term1,&Term1Ende,&Term2,&Term2Ende));                                         \
  return TRUE;                                                                                                      \
}


static BOOLEAN UnifiziertBis(UTermT Term1, UTermT Term1Ende, UTermT Term2, UTermT Term2Ende)
  RumpfUnifiziertBis(Term1,Term1Ende,Term2,Term2Ende)


static BOOLEAN Unifiziert(UTermT Term1, UTermT Term2)
{
  UTermT Term1Ende = TO_TermEnde(Term1),
         Term2Ende = TO_TermEnde(Term2);
  RumpfUnifiziertBis(Term1,Term1Ende,Term2,Term2Ende)
}

BOOLEAN U1_Unifiziert(UTermT Term1, UTermT Term2)  /* Dies ist eine kleine Routine zum Unifizieren zweier beliebi- */
{  /* ger Terme. Der verwendete Unifikationsapparat geht normalerweise davon aus, dass die Terme variablendisjunkt */
  BOOLEAN Unifizierbar;     /* sind; daher ist beim Anlegen der ersten Bindung kein Occur check erforderlich. Dies */
  PushVariable(xUDimension);   /* wird hier mit einem Trick umschifft. Fuer zeitkritischen Einsatz ist diese Funk- */
  AnzImKnotenAngelegterBindungen++;                                                      /* tion weniger geeignet. */
  Unifizierbar = Unifiziert(Term1,Term2);
  KnotenbindungenAufheben();
  return Unifizierbar;
}



/*=================================================================================================================*/
/*= IV. Erzeugung von Termkopien unter Substitutionen und Termersetzungen                                         =*/
/*=================================================================================================================*/

/* ------- 0. Praeliminarien ------------------------------------------------------------------------------------- */

#define /*void*/ AufAngehaengteZelle(/*UTermT*/ Index, /*SymbolT*/ Symbol)                                          \
{                                                                                                                   \
  TO_AufNaechsteNeueZelle(Index);                                                                                   \
  TO_SymbolBesetzen(Index,Symbol);                                                                                  \
}


#define /*void*/ AufNaechstesLesesymbol(/*UTermT*/ TermAlt, /*{1;2;3}*/ Wohin)                                      \
{                                                                                                                   \
  Topsymbol = TO_TopSymbol(TermAlt);                                                                                \
  if (SymbGebundeneVar(Topsymbol)) {                       /* sichern, dass TermAlt keine gebundene Variable ist, */\
    PushEndestapel(Wohin,TermAlt,TermAlt ## Ende);                        /* sondern freie Variable oder Funktion */\
    FertigEntbindenMitEnde(Topsymbol,TermAlt,TermAlt ## Ende);                                                      \
  }                                                                                                                 \
}                                                           /* vorausgesetzte Bezeichner: 'Topsymbol, TermAlt'Ende */


#define /*Block*/ EinSymbolWeiter(/*UTermT*/ TermAlt, /*{1;2;3}*/ EndestapelNr,                                     \
                                                                  /*{TI;TI2}*/ KopierstapelNr, /*Block*/ NullAction)\
{                                                                                                                   \
  AufNaechstesLesesymbol(TermAlt,EndestapelNr);                                                                     \
  AufAngehaengteZelle(IndexNeu,Topsymbol);     /* neue Termzelle bereitstellen und an bisherige Zellen anhaengen, */\
                                                                                  /* ausserdem Symbol uebertragen */\
  if (TO_Nullstellig(TermAlt)) {                       /* mehrsymbolige Terme wegen offenen Ende-Zeigers stapeln, */\
    TO_EndeBesetzen(IndexNeu,IndexNeu);                /* einsymboliger gleich mit richtigem Ende-Zeiger versehen */\
    while (!(--KopierstapelNr ## _TopZaehler)) {           /* und offene Ende-Zeiger aus Stapel einmuenden lassen */\
      TO_EndeBesetzen(KopierstapelNr ## _TopTeilterm,IndexNeu);                                                     \
      if (KopierstapelNr ## _PopUndLeer()) {                                          /* fertig, wenn Stapel leer */\
        NullAction;                                                                                                 \
      }                                                                                                             \
    }                                                                                                               \
    Weiterschalten ## EndestapelNr(TermAlt,TermAlt ## Ende);           /* weiter auf naechstes Symbol aus TermAlt */\
  }                                              /* dabei ggf. aus gerade zu Ende gelesenem Substitut herausgehen */\
  else {                                                                                                            \
    KopierstapelNr ## _Push(IndexNeu);                                                                              \
    TermAlt = TO_Schwanz(TermAlt);                                                                                  \
  }                                                                                                                 \
}                                                /* vorausgesetzte Bezeichner: 'Topsymbol, TermAlt'Ende, 'IndexNeu */


static TermT MGU(UTermT TermAlt);

void U1_DruckeMGU(void)
{
  SymbolT VarIndex;

  SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb,xUDimension)
    if (VarGebunden(VarIndex)) {
      TermT Substitut = MGU(SubstitutVon(VarIndex));
      IO_DruckeFlex("x%d <- %t\n",SO_VarNummer(VarIndex),Substitut);
      TO_TermLoeschen(Substitut);
    }
}


/* ------- 1. Funktionen fuer gewoehnliche Termdarstellung ------------------------------------------------------- */

static TermT MGU(UTermT TermAlt)                                                /* liefert Term mit EndeNachf NULL */
{
  SymbolT Topsymbol =  TO_TopSymbol(TermAlt);
  Entbinden(Topsymbol,TermAlt);

  if (TO_Nullstellig(TermAlt))    /* wenn nullstellig (Konstante oder freie Variable), dann eine Termzelle zurueck */
    return TO_NullstelligerTerm(Topsymbol);
  else {                                                                         /* ansonsten TermAlt durchwandern */
    UTermT TermNeu = TO_HoleErsteNeueZelle(),
           IndexNeu = TermNeu,
           TermAltEnde = TO_TermEnde(TermAlt);
    TO_SymbolBesetzen(TermNeu,Topsymbol);                       /* Termzelle fuer Topsymbol besetzen mit Topsymbol */
    TI_Push(TermNeu);
    TermAlt = TO_Schwanz(TermAlt);                                      /* weiter auf naechstes Symbol aus TermAlt */
    Endestapel3Initialisieren();
    do
      EinSymbolWeiter(TermAlt,3,TI, {                                                  /* fertig, wenn Stapel leer */
        TO_LetzteNeueZelleMelden(IndexNeu);
        TO_NachfBesetzen(IndexNeu,NULL);                                                      /* Ende-Nachf setzen */
        return TermNeu;                                                                                 /* und weg */
      })
    while (TRUE);
  }
}


static TermT MGUMitSt(TermT TermAlt, UTermT Stelle, TermT *SubstStelle)
{
  if (TO_PhysischGleich(TermAlt,Stelle))           /* Wenn Stelle = TermAlt, dann sofort MGU(Stelle) zurueckgeben. */
    return (*SubstStelle = MGU(Stelle));          /* Im anderen Fall den Term durchwandern. Dieser ist dann sicher */
  else {                                /* nicht nullstellig, da nullstellige Terme nur eine einzige Stelle haben. */
    UTermT TermNeu = TO_HoleErsteNeueZelle(),
           IndexNeu = TermNeu,
           TermAltEnde = TO_TermEnde(TermAlt);
    SymbolT Topsymbol = TO_TopSymbol(TermAlt);

    /*** 1. Schritt: Erzeugung von sigma(TermAlt) bis vor Stelle ***/
    TO_SymbolBesetzen(TermNeu,Topsymbol);                               /* Erste Termzelle besetzen mit Topsymbol. */
    TI_Push(TermNeu);         /* Nun TermAlt durchwandern, bis Stelle erreicht. Stelle liegt im Inneren des Terms. */
    TermAlt = TO_Schwanz(TermAlt);
    Endestapel3Initialisieren();
    while (TO_PhysischUngleich(TermAlt,Stelle))
      EinSymbolWeiter(TermAlt,3,TI,{;});              /* keine Aktion bei leerem Stapel, da Stapel nicht leer wird */

    /*** 2. Schritt: Erzeugung von sigma(Stelle) ***/
    EinSymbolWeiter(TermAlt,3,TI, {                                                    /* fertig, wenn Stapel leer */
      TO_LetzteNeueZelleMelden(IndexNeu);                                /* ggf. dann diese innere Stelle zuweisen */
      TO_NachfBesetzen(*SubstStelle = IndexNeu,NULL);                                         /* Ende-Nachf setzen */
      return TermNeu;                                                                                   /* und weg */
    });
    *SubstStelle = IndexNeu;       /* andernfalls folgen noch weitere Teilterme; jetzt aber innere Stelle zuweisen */

    /*** 3. Schritt: Erzeugung des Endes von sigma(TermAlt) ***/
    do
      EinSymbolWeiter(TermAlt,3,TI, {                                                  /* fertig, wenn Stapel leer */
        TO_LetzteNeueZelleMelden(IndexNeu);
        TO_NachfBesetzen(IndexNeu,NULL);                                                      /* Ende-Nachf setzen */
        return TermNeu;                                                                                 /* und weg */
      })
    while (TRUE);

  }
}


static TermT MGUNachTE(TermT TermAlt, UTermT Stelle, TermT Einbau)
#define RumpfMGUNachTE(/*LAusdruck=*/ Ergaenzung1, /*Zuweisung*/ Ergaenzung2)                                       \
{                                                                                                                   \
  if (TO_PhysischGleich(TermAlt,Stelle))          /* Wenn Stelle = TermAlt, dann sofort MGU(Einbau) zurueckgeben. */\
    return (Ergaenzung1 MGU(Einbau));                           /* Bei MGUNachTEMitSt hier Eingesetztes zuweisen. */\
  else {                      /* Im anderen Fall den Term durchwandern. Dieser ist dann sicher nicht nullstellig, */\
    UTermT TermNeu = TO_HoleErsteNeueZelle(),             /* da nullstellige Terme nur eine einzige Stelle haben. */\
           IndexNeu = TermNeu,                                                                                      \
           TermAltEnde = TO_TermEnde(TermAlt);                                                                      \
    SymbolT Topsymbol = TO_TopSymbol(TermAlt);                                                                      \
                                                                                                                    \
    /*** 1. Schritt: Erzeugung von sigma(TermAlt) bis vor der Einbaustelle ***/                                     \
    TO_SymbolBesetzen(TermNeu,Topsymbol);                              /* Erste Termzelle besetzen mit Topsymbol. */\
    TI2_Push(TermNeu);       /* Nun TermAlt durchwandern, bis Stelle erreicht. Stelle liegt im Inneren des Terms. */\
    TermAlt = TO_Schwanz(TermAlt);                                                                                  \
    Endestapel3Initialisieren();                                                                                    \
    while (TO_PhysischUngleich(TermAlt,Stelle))                                                                     \
      EinSymbolWeiter(TermAlt,3,TI2,{;});            /* keine Aktion bei leerem Stapel, da Stapel nicht leer wird */\
                                                                                                                    \
    /*** 2. Schritt: Erzeugung von sigma(Einbau) ***/                                                               \
    {                                                                                                               \
      Topsymbol =  TO_TopSymbol(Einbau);                             /* IndexNeu ist die bisher letzte neue Zelle */\
      Entbinden(Topsymbol,Einbau);                           /* sichern, dass Einbau keine gebundene Variable ist */\
      AufAngehaengteZelle(IndexNeu,Topsymbol);                 /* Termzelle fuer Topsymbol besetzen mit Topsymbol */\
      Ergaenzung2;                                               /* bei MGUNachTEMitSt hier Eingesetztes zuweisen */\
                                                                                                                    \
      if (TO_Nullstellig(Einbau))   /* wenn nullstellig (Konst. oder freie Variable), dann eine Termzelle zurueck */\
        TO_EndeBesetzen(IndexNeu,IndexNeu);                                                                         \
      else {                                                                     /* ansonsten Einbau durchwandern */\
        UTermT EinbauEnde = TO_TermEnde(Einbau);                                                                    \
        TI_Push(IndexNeu);                                                                                          \
        Einbau = TO_Schwanz(Einbau);                                                                                \
        Endestapel1Initialisieren();                                                                                \
        do                                                                                                          \
          EinSymbolWeiter(Einbau,1,TI,{goto Fertig;})                                                               \
        while (TRUE);                                                                                               \
        Fertig: ;                                                                                                   \
      }                                                                                                             \
    }                                                                                                               \
    while (!(--TI2_TopZaehler)) {                              /* offene Ende-Zeiger aus Stapel einmuenden lassen */\
      TO_EndeBesetzen(TI2_TopTeilterm,IndexNeu);                                                                    \
      if (TI2_PopUndLeer()) {                                                         /* fertig, wenn Stapel leer */\
        TO_LetzteNeueZelleMelden(IndexNeu);                                                                         \
        TO_NachfBesetzen(IndexNeu,NULL);                                                     /* Ende-Nachf setzen */\
        return TermNeu;                                                                               /* und weg. */\
      }                                                                                                             \
    }                                                                                                               \
                                                                                                                    \
    /*** 3. Schritt: Erzeugung des Endes von sigma(TermAlt) ***/                                                    \
    TermAlt = TO_NaechsterTeilterm(TermAlt);                                        /* Einbaustelle ueberspringen */\
    do                                                                                                              \
      EinSymbolWeiter(TermAlt,3,TI2, {                                                /* fertig, wenn Stapel leer */\
        TO_LetzteNeueZelleMelden(IndexNeu);                                                                         \
        TO_NachfBesetzen(IndexNeu,NULL);                                                     /* Ende-Nachf setzen */\
        return TermNeu;                                                                                /* und weg */\
      })                                                                                                            \
    while (TRUE);                                                                                                   \
  }                                                                                                                 \
}
  RumpfMGUNachTE(,)

static TermT MGUNachTEMitSt(TermT TermAlt, UTermT Stelle, TermT Einbau, TermT *Eingesetztes)
  RumpfMGUNachTE(*Eingesetztes =,*Eingesetztes = IndexNeu)



/*=================================================================================================================*/
/*= V. Operationen zum Unifizieren auf Blattebene                                                                 =*/
/*=================================================================================================================*/

#define /*UTermT*/ RegelrestBeimBlatterreichen(/*void*/)                                                            \
  (AbstiegMitZielErfolgt ? TermNachZiel : BK_Rest(BaumIndex))


#define /*BOOLEAN*/ MGUMitBlattGefunden(/*DSBaumT*/ Baum)                                                           \
  (!RegelrestBeimBlatterreichen() ||                                                                                \
   UnifiziertBis(RegelrestBeimBlatterreichen(),TO_TermEnde(BK_linkeSeite(BaumIndex)),                               \
                 aktTermrest,TO_TermEnde(BK_aktTIndexU(Baum.Wurzel))))


#define /*BOOLEAN*/ MGUMitBlattGefundenAusschluss(/*DSBaumT*/ Baum, /*GNTermpaarT*/ Ausschluss)                     \
  (BK_Regeln(BaumIndex)!=Ausschluss && MGUMitBlattGefunden(Baum))
/* Prueft, ob das Ausschlussobjekt als erstes im Blatt steht. */



/*=================================================================================================================*/
/*= VI. Im Baum absteigen fuer SymbolAusTerm Funktionssymbol                                                      =*/
/*=================================================================================================================*/

static BOOLEAN Delta0(void);
static BOOLEAN DeltaFktSymbol1(void);

#define /*void*/ AbstiegVerwalten(/*KantenT*/ Nachfolger,                                                           \
                                  /*{Ziel,Symbol}*/ SorteDatum,                                                     \
                                  /*SprungeintragsT U SymbolT*/ naechstesDatum,                                     \
                                  /*UDeltaT*/ naechstesDelta,                                                       \
                                  /*UTermT*/ neuerTermrest,                                                         \
                                  /*unsigned short int*/ AnzahlBindungen)                                           \
{                                                                                                                   \
  KantenT Nachf = Nachfolger;                                                                                       \
  BK_UVorgSei(Nachf,BaumIndex);                                                                                     \
  BK_UAnzBindungenSei(BaumIndex,AnzahlBindungen);                                                                   \
  BK_UDatumSei(BaumIndex,SorteDatum,naechstesDatum);                                                                \
  BK_UDeltaSei(BaumIndex,naechstesDelta);                                                                           \
  BK_aktTIndexUSei(Nachf,neuerTermrest);                                                                            \
  BK_UDeltaSei(Nachf,Delta0);                                                                                       \
  BaumIndex = Nachf;                                                                                                \
  AnzImKnotenAngelegterBindungen = 0;                                                                               \
  return TRUE;                                                                                                      \
}


#define MitSelbemSymbolAb(/*SymbolT*/ naechstesSymbol, /*UDeltaT*/ naechstesDelta)                                  \
{                                                                                                                   \
  if (BK_Nachf(BaumIndex,SymbolAusTerm))                                                                            \
    AbstiegVerwalten(BK_Nachf(BaumIndex,SymbolAusTerm),                                                             \
      Symbol,naechstesSymbol,naechstesDelta,TO_Schwanz(aktTermrest),0);                                             \
}


#define TermAnBVarGebundenUndAb(/*SymbolT*/ BaumVar, /*UDeltaT*/ naechstesDelta)                                    \
/* versucht Bindung Baumvariable<-aktueller Teilterm des Anfrageterms vorzunehmen. FALSE, falls Fehler beim Occur   \
   check; ansonsten wird Baum absteigen. */                                                                         \
/* Aufruf nur in TermAnEineBVarGebundenOderUnifiziertUndAb(erste,letzte) */                                         \
{                                                                                                                   \
  if (SubstErgaenzt(BaumVar,aktTermrest))                                                                           \
    AbstiegVerwalten(BK_Nachf(BaumIndex,BaumVar),Symbol,                                                            \
      SO_VorigesVarSymb(BaumVar),naechstesDelta,TO_NaechsterTeilterm(aktTermrest),AnzImKnotenAngelegterBindungen);  \
}


#define TermMitBVarUnifiziertUndAb(/*SymbolT*/ BaumVar, /*UDeltaT*/ naechstesDelta)                                 \
/* versucht aktuellen Teilterm des Anfrageterms mit gebundener Variable aus dem Baum zu unifizieren. */             \
/* Aufruf nur in TermAnEineBVarGebundenOderUnifiziertUndAb(erste,letzte) */                                         \
{                                                                                                                   \
  if (Unifiziert(SubstitutVon(BaumVar),aktTermrest))                                                                \
    AbstiegVerwalten(BK_Nachf(BaumIndex,BaumVar),Symbol,SO_VorigesVarSymb(BaumVar),naechstesDelta,                  \
      TO_NaechsterTeilterm(aktTermrest),AnzImKnotenAngelegterBindungen);                                            \
  KnotenbindungenAufheben();                                                                                        \
}


#define TermAnEineBVarGebundenOderUnifiziertUndAb(/*SymbolT*/ ersteKandidatin,/*UDeltaT*/ naechstesDelta)           \
/* wird auch fuer SymbolAusTerm gebundene Variable verwendet */                                                     \
{                                                                                                                   \
  SymbolT letzteKandidatin = min(SO_ErstesVarSymb,BK_maxSymb(BaumIndex)),                                           \
          BaumVar;                                                                                                  \
  SO_forSymboleAufwaerts(BaumVar,ersteKandidatin,letzteKandidatin)                                                  \
    if (BK_Nachf(BaumIndex,BaumVar)) {                                                                              \
      if (VarGebunden(BaumVar))                                                                                     \
        TermMitBVarUnifiziertUndAb(BaumVar,naechstesDelta)                                                          \
      else                                                                                                          \
        TermAnBVarGebundenUndAb(BaumVar,naechstesDelta);                                                            \
    }                                                                                                               \
}


static BOOLEAN DeltaFktSymbol0(void)
{
  MitSelbemSymbolAb(BK_minSymb(BaumIndex),DeltaFktSymbol1);
  TermAnEineBVarGebundenOderUnifiziertUndAb(BK_minSymb(BaumIndex),DeltaFktSymbol1);
  return FALSE;
}

static BOOLEAN DeltaFktSymbol1(void)
{
  TermAnEineBVarGebundenOderUnifiziertUndAb(BK_USymbol(BaumIndex),DeltaFktSymbol1);
  return FALSE;
}



/*=================================================================================================================*/
/*= VII. Im Baum absteigen fuer SymbolAusTerm gebundene Variable                                                  =*/
/*=================================================================================================================*/

static BOOLEAN DeltaGebundeneVar1(void);
static BOOLEAN DeltaGebundeneVar2(void);

#define EinZielMitTVarUnifiziertUndAb(/*SprungeintragsT*/ erstesZiel, /*UDeltaT*/ naechstesDelta)                   \
{                                                                                                                   \
  SprungeintragsT ZielIndex;                                                                                        \
  BK_forSprungziele(ZielIndex,erstesZiel) {                                                                         \
    if (Unifiziert(ZielIndex->Teilterm,aktTermrest)) {                                                              \
      TermNachZiel = TO_NaechsterTeilterm(ZielIndex->Teilterm);                                                     \
      AbstiegMitZielErfolgt = TRUE;                                                                                 \
      AbstiegVerwalten(ZielIndex->Zielknoten,Ziel,ZielIndex->NaechsterZieleintrag,naechstesDelta,                   \
        TO_NaechsterTeilterm(aktTermrest),AnzImKnotenAngelegterBindungen);                                          \
    }                                                                                                               \
    KnotenbindungenAufheben();                                                                                      \
  }                                                                                                                 \
}

static BOOLEAN DeltaGebundeneVar0(void)
{
  TermAnEineBVarGebundenOderUnifiziertUndAb(BK_minSymb(BaumIndex),DeltaGebundeneVar1);
  EinZielMitTVarUnifiziertUndAb(BK_Sprungausgaenge(BaumIndex),DeltaGebundeneVar2);
  return FALSE;
}

static BOOLEAN DeltaGebundeneVar1(void)
{
  TermAnEineBVarGebundenOderUnifiziertUndAb(BK_USymbol(BaumIndex),DeltaGebundeneVar1);
  EinZielMitTVarUnifiziertUndAb(BK_Sprungausgaenge(BaumIndex),DeltaGebundeneVar2);
  return FALSE;
}

static BOOLEAN DeltaGebundeneVar2(void)
{
  EinZielMitTVarUnifiziertUndAb(BK_UZiel(BaumIndex),DeltaGebundeneVar2);
  return FALSE;
}



/*=================================================================================================================*/
/*= VIII. Im Baum absteigen fuer SymbolAusTerm freie Variable                                                     =*/
/*=================================================================================================================*/

static BOOLEAN DeltaFreieVar1(void);
static BOOLEAN DeltaFreieVar2(void);

#define EineBVarAnTVarGebundenUndAb(/*SymbolT*/ ersteKandidatin, /*UDeltaT*/ naechstesDelta)                        \
{                                                                                                                   \
  SymbolT letzteKandidatin = min(SO_ErstesVarSymb,BK_maxSymb(BaumIndex)),                                           \
          BaumVar;                                                                                                  \
  SO_forSymboleAufwaerts(BaumVar,ersteKandidatin,letzteKandidatin) {                                                \
    if (BK_Nachf(BaumIndex,BaumVar) && SubstErgaenzt(SymbolAusTerm,TO_Variablenterm(BaumVar)))                      \
      AbstiegVerwalten(BK_Nachf(BaumIndex,BaumVar),Symbol,                                                          \
        SO_VorigesVarSymb(BaumVar),naechstesDelta,TO_Schwanz(aktTermrest),AnzImKnotenAngelegterBindungen);          \
  }                                                                                                                 \
}


#define EinZielAnTVarGebundenUndAb(/*SprungeintragsT*/ erstesZiel, /*UDeltaT*/ naechstesDelta)                      \
{                                                                                                                   \
  SprungeintragsT ZielIndex;                                                                                        \
  BK_forSprungziele(ZielIndex,erstesZiel)                                                                           \
    if (SubstErgaenzt(SymbolAusTerm,ZielIndex->Teilterm)) {                                                         \
      TermNachZiel = TO_NaechsterTeilterm(ZielIndex->Teilterm);                                                     \
      AbstiegMitZielErfolgt = TRUE;                                                                                 \
      AbstiegVerwalten(ZielIndex->Zielknoten,Ziel,                                                                  \
        ZielIndex->NaechsterZieleintrag,naechstesDelta,TO_Schwanz(aktTermrest),AnzImKnotenAngelegterBindungen);     \
    }                                                                                                               \
}


static BOOLEAN DeltaFreieVar0(void)
{
  EineBVarAnTVarGebundenUndAb(BK_minSymb(BaumIndex),DeltaFreieVar1);
  EinZielAnTVarGebundenUndAb(BK_Sprungausgaenge(BaumIndex),DeltaFreieVar2);
  return FALSE;
}

static BOOLEAN DeltaFreieVar1(void)
{
  EineBVarAnTVarGebundenUndAb(BK_USymbol(BaumIndex),DeltaFreieVar1);
  EinZielAnTVarGebundenUndAb(BK_Sprungausgaenge(BaumIndex),DeltaFreieVar2);
  return FALSE;
}

static BOOLEAN DeltaFreieVar2(void)
{
  EinZielAnTVarGebundenUndAb(BK_UZiel(BaumIndex),DeltaFreieVar2);
  return FALSE;
}



/*=================================================================================================================*/
/*= IX. Unifizieren eines Terms mit eigenen Teiltermen, mit DSBaum auf Toplevel, mit DSBaum-Teiltermen            =*/
/*=================================================================================================================*/

static BOOLEAN Delta0(void)                                                                           /* Verteiler */
{
  return
    SO_SymbIstFkt(SymbolAusTerm) ?
      DeltaFktSymbol0()
    :
      VarGebunden(SymbolAusTerm) ?
        DeltaGebundeneVar0()
      :
        DeltaFreieVar0();
}


#define /*void*/ EinenKnotenAufsteigen(/*void*/)                                                                    \
  if ((BaumIndex = BK_UVorg(BaumIndex))) {                                                                          \
    BindungenAufheben(BK_UAnzBindungen(BaumIndex));                                                                 \
    AnzImKnotenAngelegterBindungen = 0;                                                                             \
  }


#define /*void*/ AusBlattAufsteigen(/*void*/)                                                                       \
  BindungenAufheben(AnzImKnotenAngelegterBindungen);                                                                \
  EinenKnotenAufsteigen()


#define /*void*/ EinenKnotenAbOderVieleKnotenAufsteigen(/*void*/)                                                   \
{                                                                                                                   \
  AbstiegMitZielErfolgt = FALSE;                                                                                    \
  do                                                                                                                \
    if (BK_UDelta(BaumIndex))                                                                                       \
      break;                                                                                                        \
    else                                                                                                            \
      EinenKnotenAufsteigen()                                                                                       \
  while (BaumIndex);                                                                                                \
}


#if TP_PETERGRAF
#undef BK_Regeln
#define BK_Regeln(BaumIndex) GleichungsIndex


#define /*Block*/ RumpfTermMitDSBaumUnifizieren(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*Block*/ Ausschluss,            \
                                                                                                   /*Block*/ Action)\
{                                                                                                                   \
  if (Baum.Wurzel) {                                                                                                \
    GleichungsT GleichungsIndex = (GleichungsT) BK_Nachf(Baum.Wurzel,TO_TopSymbol(Term));                           \
    while (GleichungsIndex) {                                                                                       \
      Ausschluss;                                                                                                   \
      if (Unifiziert(Term,TP_LinkeSeite(GleichungsIndex))) {                                                        \
        Action;                                                                                                     \
      }                                                                                                             \
      BindungenAufheben(AnzImKnotenAngelegterBindungen);                                                            \
      AnzImKnotenAngelegterBindungen = 0;                                                                           \
      GleichungsIndex = TP_Nachf(GleichungsIndex);                                                                  \
    }                                                                                                               \
  }                                                                                                                 \
}      /* Beachte: Variable GleichungsIndex ist Entsprechung zur namensgleichen Groesse in MitAllen nach Standard. */


#define /*void*/ TermMitDSBaumUnifizieren(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*Block*/ Action)                      \
  RumpfTermMitDSBaumUnifizieren(Term,Baum,,Action)


#define /*void*/ TermMitDSBaumUnifizierenAusschluss(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*GNTermpaarT*/ Selbst,      \
                                                                                                   /*Block*/ Action)\
  RumpfTermMitDSBaumUnifizieren(Term,Baum,                                                                          \
    if (GleichungsIndex==Selbst) {GleichungsIndex = TP_Nachf(GleichungsIndex); continue;},Action)
#else


#define /*Block*/ RumpfTermMitDSBaumUnifizieren(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*Block*/ Action,                \
                                                                                            /*BOOLEAN*/ MGUMitBlatt)\
{                                                                                                                   \
  if (Baum.Wurzel) {                                                                                                \
    BK_UDeltaSei(BaumIndex = Baum.Wurzel,Delta0);                                                                   \
    BK_aktTIndexUSei(BaumIndex,Term);                                                                               \
                                                                                                                    \
    EinenKnotenAbOderVieleKnotenAufsteigen();                                                                       \
    while (BaumIndex) {                                                                                             \
      if (BK_IstBlatt(BaumIndex)) {                                                                                 \
        if (MGUMitBlatt) {                                                                                          \
          Action;                                                                                                   \
        }                                                                                                           \
        AusBlattAufsteigen();                                                                                       \
      }                                                                                                             \
      EinenKnotenAbOderVieleKnotenAufsteigen();                                                                     \
    }                                                                                                               \
  }                                                                                                                 \
}


#define /*void*/ TermMitDSBaumUnifizieren(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*Block*/ Action)                      \
  RumpfTermMitDSBaumUnifizieren(Term,Baum,Action,MGUMitBlattGefunden(Baum))


#define /*void*/ TermMitDSBaumUnifizierenAusschluss(/*UTermT*/ Term, /*DSBaumT*/ Baum, /*GNTermpaarT*/ Ausschluss,  \
                                                                                                   /*Block*/ Action)\
  RumpfTermMitDSBaumUnifizieren(Term,Baum,Action,MGUMitBlattGefundenAusschluss(Baum,Ausschluss))
#endif


#define /*void*/ TermMitDSBaumTeiltermenUnifizieren(/*TermT*/ Term, /*DSBaumT*/ Baum, /*Block*/ Action)             \
{                                                                                                /* furchtbar ... */\
  GNTermpaarT RegelIndex;                                                                                           \
  BK_forRegeln(RegelIndex,Baum, {                   /* Baum leer => Regelliste leer; daher for-Schleife angebracht*/\
    UTermT LinkeSeiteIndex = TO_Schwanz(RegelIndex->linkeSeite);                                                    \
    while (LinkeSeiteIndex) {                                                  /* alle linken Seiten durchwandern */\
      if (SO_SymbIstFkt(TO_TopSymbol(LinkeSeiteIndex))) {                                                           \
        if (Unifiziert(Term,LinkeSeiteIndex)) {                              /* Wenn Unifikation erfolgreich war: */\
          Action;                                                                                                   \
        }                                                                                                           \
        KnotenbindungenAufheben();                                                                                  \
      }                                                                                                             \
      LinkeSeiteIndex = TO_Schwanz(LinkeSeiteIndex);                                                                \
    }                                                                                                               \
  })                                                                                                                \
}


#define /*void*/ Term1MitTerm2Unifizieren(/*TermT*/ Term1, /*TermT*/ Term2, /*Block*/ Action)                       \
{                                                                                                                   \
  if (Unifiziert(Term1,Term2))                                                                                      \
    Action;                                                                                                         \
  KnotenbindungenAufheben();                                                                                        \
}



/*=================================================================================================================*/
/*= X. Apparat fuer KP-Filter                                                                                     =*/
/*=================================================================================================================*/

static TermT KPLinks,
             KPRechts,
             KPRechtsInnen,
             Ueberlappung,
             UeberlappungInnen;

/* nun Makros, da sonst Vater, VaterStelle, Mutter als Parameter durchzureichen */

#define KPLinksProduzieren()      (KPLinks = MGU(TP_RechteSeite(Vater)))
#define KPRechtsProduzieren()     (KPRechts = MGUNachTEMitSt(TP_LinkeSeite(Vater),VaterStelle,TP_RechteSeite(Mutter),&KPRechtsInnen))
#define UeberlappungProduzieren() (Ueberlappung = MGUMitSt(TP_LinkeSeite(Vater),VaterStelle,&UeberlappungInnen))

#define /*void*/ Beginnen(/*BOOLEAN*/ KPLinksAn, /*BOOLEAN*/ KPRechtsAn)                                            \
{ if (KPLinksAn)               { KPLinksProduzieren(); }     else KPLinks = NULL;                                   \
  if (KPRechtsAn)              { KPRechtsProduzieren(); }    else KPRechts = KPRechtsInnen = NULL;                  \
  if (KPLinksAn || KPRechtsAn) { UeberlappungProduzieren();  IncAnzUnifUeberlappungenGebildet(); }                  \
                                                             else Ueberlappung = UeberlappungInnen = NULL;          \
}

#define /*void*/ Verwerfen(/*void*/)                                                                                \
{ if (KPLinks)      { TO_TermLoeschen(KPLinks);        IncAnzUnifUeberflKritTerme(); }                              \
  if (KPRechts)     { TO_TermLoeschen(KPRechts);       IncAnzUnifUeberflKritTerme(); }                              \
  if (Ueberlappung) { TO_TermLoeschen(Ueberlappung); }                                                              \
  return FALSE;                                                                                                     \
}

#define /*void*/ KPTermeProduzieren(/*void*/)                                                                       \
  ((void)(KPLinks ? NULL : KPLinksProduzieren(), KPRechts ? NULL : KPRechtsProduzieren()))

#define /*void*/ Vollenden(/*{"RR","RG","GG"}*/ Wo)                                                                 \
{ KPTermeProduzieren();                                                                                             \
  if (Ueberlappung) { TO_TermLoeschen(Ueberlappung); }                                                              \
  IncAnzKP##Wo();                                                                                                   \
  return TRUE;                                                                                                      \
}

#define an  TRUE
#define aus FALSE

typedef BOOLEAN (*FilterT)(RegelOderGleichungsT Vater, UTermT si, RegelOderGleichungsT Mutter);

static FilterT *KPFilter;
static unsigned int AnzahlFilter;

static void KPFilterInitialisieren(void)
{
  AnzahlFilter = 0;
  KPFilter = array_alloc(AnzahlFilter+1,FilterT);
  KPFilter[AnzahlFilter] = NULL;
}

static void KPFilterErgaenzen(FilterT Filter)
{
  KPFilter[AnzahlFilter++] = Filter;
  array_realloc(KPFilter,AnzahlFilter+1,FilterT);
  KPFilter[AnzahlFilter] = NULL;
}

static BOOLEAN Weggefiltert(RegelOderGleichungsT Vater, UTermT VaterStelle, RegelOderGleichungsT Mutter)
{
  FilterT *Filter = KPFilter;
  while (*Filter)
    if (!(*(Filter++))(Vater,VaterStelle,Mutter))
      return TRUE;
  return FALSE;
}



/*=================================================================================================================*/
/*= XI. KP-Filter Verbunden-unterhalb                                                                             =*/
/*=================================================================================================================*/

NFInnenReduzibelT *Kapurreduzibel;                          /* nur mit Regeln, nur mit Gleichungen oder mit beidem */
/* Arnim sei Dank Zeiger auf Zeiger auf Funktion */

static BOOLEAN Primueberlappung(RegelOderGleichungsT Vater MAYBE_UNUSEDPARAM,
  UTermT VaterStelle, RegelOderGleichungsT Mutter MAYBE_UNUSEDPARAM)
{
  if (!UeberlappungInnen)
    UeberlappungProduzieren();
  if ((*Kapurreduzibel)(UeberlappungInnen)) {
    IncAnzKapurOpfer();
    return FALSE;
  }
  else {
    return TRUE;
  }
}



/*=================================================================================================================*/
/*= XII. KP-Filter 1*                                                                                             =*/
/*=================================================================================================================*/

static BOOLEAN AnEinsSternIn(UTermT si, UTermT s)
{
  while (TO_PhysischUngleich(s,si))
    if (TO_Nullstellig(s))
      return FALSE;
    else
      s = TO_ErsterTeilterm(s);
  return TRUE;
}


static BOOLEAN EinsSternUeberlappung(RegelOderGleichungsT Vater, UTermT si, RegelOderGleichungsT Mutter)
{
  TermT s = TP_LinkeSeite(Vater);
  BOOLEAN Resultat =
    PM_Existenzziele() && TO_TopSymbol(s)==SO_EQ ?
      TO_TopSymbol(TP_LinkeSeite(Mutter))==SO_EQ ?
        TO_TopSymbol(TP_RechteSeite(Vater))==SO_TRUE || TO_TopSymbol(TP_RechteSeite(Mutter))==SO_TRUE
      :
        AnEinsSternIn(si,s=TO_ErsterTeilterm(s)) || AnEinsSternIn(si,s=TO_NaechsterTeilterm(s))
    :
      AnEinsSternIn(si,s); 
  if (!Resultat) {
    HK_fair = FALSE;
    IncAnzNonEinsSternKPs();
  }
  return Resultat;
}






/*=================================================================================================================*/
/*= XIIa. KP-Filter Nusfu                                                                                          =*/
/*=================================================================================================================*/

static BOOLEAN Nusfu(UTermT sj, UTermT s)
{
  UTermT NULLs = TO_NULLTerm(s);
  do {
    SymbolT f = TO_TopSymbol(s);
    if (SO_IstSkolemSymbol(f) && (!PM_Existenzziele || f<SO_EQ)) {
      UTermT NULLsi = TO_NULLTerm(s);
      while (TO_PhysischUngleich(s = TO_Schwanz(s),NULLsi))
        if (TO_PhysischGleich(s,sj))
          return FALSE;
    }
    else
      s = TO_Schwanz(s);
  }
  while (TO_PhysischUngleich(s,NULLs));
  return TRUE;
}


static BOOLEAN NusfUeberlappung(RegelOderGleichungsT Vater, UTermT si, RegelOderGleichungsT Mutter MAYBE_UNUSEDPARAM)
{
  TermT s = TP_LinkeSeite(Vater);
  BOOLEAN Resultat = Nusfu(si,s);
  /*  IO_DruckeFlex("Nusfu(%t,%t)=%s\n",si,s,Resultat?"TRUE":"FALSE");*/
  if (!Resultat) {
    HK_fair = FALSE;                                                       /*XXX ggf. doch als fair interpretieren */
    IncAnzNonNusfuKPs();
  }
  return Resultat;
}



/*=================================================================================================================*/
/*= XIII. KP-Filter Kern                                                                                          =*/
/*=================================================================================================================*/

/* ------- 0. Rueckwaerts beweisen ------------------------------------------------------------------------------- */

static BOOLEAN IstVerdrehterKombinator(RegelOderGleichungsT Faktum)
{
  TermT s = TP_RechteSeite(Faktum),
        t = TP_LinkeSeite(Faktum);
  return TO_IstKombinatorTupel(s,t);
}


/* ------- 1. Stufe 1 -------------------------------------------------------------------------------------------- */

static BOOLEAN IstRechtsInZiel(UTermT ti, RegelOderGleichungsT Faktum)
{                                    /* keine Seitenrelaxierung erforderlich, da Bezug auf intern kreiertes Faktum */
  UTermT s, t;
  return RE_UngleichungEntmantelt(&s,&t, Faktum) && TO_PhysischTeilterm(ti,t);
}


static BOOLEAN IstKombinatorSymbol(SymbolT f)
{                                                  /* Aufwendig, aber schick. Macht nichts, da nicht zeitkritisch. */
  if (SO_SymbIstKonst(f)) {
    ListenT Fakten;
    for (Fakten = PM_Gleichungen; Fakten; Fakten = cdr(Fakten)) {
      RegelOderGleichungsT Faktum = car(Fakten);
      TermT u = TP_LinkeSeite(Faktum),
            v = TP_RechteSeite(Faktum);
      UTermT s = TO_IstKombinatorTupel(u,v) ? u : TO_IstKombinatorTupel(v,u) ? v : NULL;
      if (s && TO_SymbolEnthalten(f,s))
        return TRUE;
    }
  }
  return FALSE;
}

#define for_KombinatorSymbole(/*SymbolT*/ Index)                                                                    \
  SO_forKonstSymbole(Index)                                                                                         \
    if (IstKombinatorSymbol(Index))


static unsigned int KernNr = 1;

static void Erfolg1Testen(TermT l, TermT r)
{
  UTermT s, t;
  if (Z_UngleichungEntmantelt(&s,&t, l,r)) {
    if (U2_NeuUnifiziert(s,t)) {
      TermT K = U2_MGU(s);
      SymbolT C;
      /*printf("KERN "); IO_DruckeTerm(K);*/
      for_KombinatorSymbole(C)
        if (!TO_SymbolEnthalten(C,K))
          {/*printf(" --- unvollstaendig\n");*/ return;}
      /*printf(" --- vollstaendig\n"); */
      if (!--KernNr) {
        KO_KonsultationErforderlich = TRUE; 
        KO_Kern = K; 
      }
    }
  }
}


/* ------- 2. Stufe 2 -------------------------------------------------------------------------------------------- */

static BOOLEAN IstLinksVariable(TermT l, TermT r)
{                                    /* keine Seitenrelaxierung erforderlich, da Bezug auf intern kreiertes Faktum */
  UTermT s, t;
  return Z_UngleichungEntmantelt(&s,&t, l,r) && TO_TermIstVar(s);
}

#define /*BOOLEAN*/ IstFixedPositiv_(/*TermT*/ s, /*TermT*/ t)                                                      \
  ( SO_SymbGleich(TO_TopSymbol(s),SO_EQ) &&                                                                         \
    SO_SymbGleich(TO_TopSymbol(t),SO_TRUE) &&                                                                       \
    SO_SymbIstFkt(TO_TopSymbol(TO_ErsterTeilterm(s)))                                                               \
  )

#define /*BOOLEAN*/ IstFixedPositiv(/*RegelOderGleichungsT*/ Faktum)                                                \
  ( IstFixedPositiv_(TP_LinkeSeite(Faktum),TP_RechteSeite(Faktum)) ||                                               \
    IstFixedPositiv_(TP_RechteSeite(Faktum),TP_LinkeSeite(Faktum)) )

static void Erfolg2Testen(RegelOderGleichungsT Vater, RegelOderGleichungsT Mutter)
{
  BOOLEAN Resultat1 = RE_IstUngleichung(Vater) && IstFixedPositiv(Mutter),
          Resultat2 = !Resultat1 && RE_IstUngleichung(Mutter) && IstFixedPositiv(Vater),
          Resultat = Resultat1 || Resultat2;
  if (Resultat) {
    RegelOderGleichungsT FixedPos = Resultat1 ? Mutter : Vater;
    TermT l = TP_LinkeSeite(FixedPos), r = TP_LinkeSeite(FixedPos);
    UTermT t = SO_SymbGleich(TO_TopSymbol(l),SO_EQ) ? l : r;
    while (TO_TermIstNichtVar(t = TO_Schwanz(t)));
    KO_Fixpunktkombinator = MGU(t);
    /*IO_DruckeFlex("SFP %t\n",KO_Fixpunktkombinator);*/
    KO_KonsultationErforderlich = TRUE; 
  }
}


/* ------- 3. Stufe 3 -------------------------------------------------------------------------------------------- */

static BOOLEAN IstUnifizierbaresZiel(TermT l, TermT r)
{
  UTermT s, t;
  return Z_UngleichungEntmantelt(&s,&t, l,r) && U2_NeuUnifiziert(s,t);
}


static BOOLEAN IstTrueGleichFalse(TermT l, TermT r)
{
  if (PM_Existenzziele()) {
    SymbolT f = TO_TopSymbol(l),
            g = TO_TopSymbol(r);
    return (f==SO_TRUE && g==SO_FALSE) || (f==SO_FALSE && g==SO_TRUE);
  }
  else
    return FALSE;
}

#define Erfolg4 IstTrueGleichFalse


/* ------- 4. Filter --------------------------------------------------------------------------------------------- */

static BOOLEAN KernUeberlappung(RegelOderGleichungsT Vater, UTermT VaterStelle, RegelOderGleichungsT Mutter)
{
  BOOLEAN Resultat = 

    KO_Stufe==1 ? ( IstVerdrehterKombinator(Mutter) ) && 
                  ( IstRechtsInZiel(VaterStelle,Vater) ) && 
                  ( KPTermeProduzieren(), Erfolg1Testen(KPLinks,KPRechts), TRUE ) 
    :
    KO_Stufe==2 ? ( IstVerdrehterKombinator(Mutter) && 
                    IstRechtsInZiel(VaterStelle,Vater) && 
                    (KPTermeProduzieren(), IstLinksVariable(KPLinks,KPRechts)) ) ||
                  ( Erfolg2Testen(Vater,Mutter), FALSE )
    :
    KO_Stufe==3 ? ( FALSE )
    :
    KO_Stufe==4 ? ( KPTermeProduzieren(), IstUnifizierbaresZiel(KPLinks,KPRechts) ) || 
                  ( Erfolg4(KPLinks,KPRechts) )
    :               
                  ( TRUE );

  if (!Resultat)
    HK_fair = FALSE;

  return Resultat;
}


#if 0
  {
    KPTermeProduzieren();
    IO_DruckeFlex("Mutter: %t = %t %s\n", TP_LinkeSeite(Mutter),TP_RechteSeite(Mutter), 
      IstVerdrehterKombinator(Mutter) ? "verdrehter Kombinator!" : "");
    IO_DruckeFlex("Vater:  %t = %t\n", TP_LinkeSeite(Vater),TP_RechteSeite(Vater));
    IO_DruckeFlex("VaterStelle: %t %s\n",VaterStelle,
      IstRechtsInZiel(VaterStelle,Vater) ? "rechts in Ziel!" : "");
    IO_DruckeFlex("Faktum: %t = %t --- %sunifizierbar, %sErfolg\n", KPLinks, KPRechts,
      IstUnifizierbar(KPLinks,KPRechts) ? "" : "nicht ", IstErfolg(KPLinks,KPRechts) ? "" : "kein ");
    IO_DruckeFlex("Relevantes Faktum? %s\n", Resultat ? "TRUE" : "FALSE");
  }
#endif



/*=================================================================================================================*/
/*= XIV. Filter fuer Rueckwaertsgang                                                                              =*/
/*=================================================================================================================*/

static BOOLEAN RichtigeZielseiteLinks;

static void RichtigeZielseiteSuchen(void)
{
  if (PM_Ziele) {
    ListenT Ziel = car(PM_Ziele);
    UTermT t = car(Ziel);
    while (SO_SymbGleich(TO_TopSymbol(t),TO_ApplyOperator))
      t = TO_Schwanz(t);
    RichtigeZielseiteLinks = SO_IstSkolemSymbol(TO_TopSymbol(t));
  }
  else
    RichtigeZielseiteLinks = FALSE;
}

static BOOLEAN IstInRichtigerZielseite(UTermT ti, RegelOderGleichungsT Faktum)
{
  UTermT s, t;
  return RE_UngleichungEntmantelt(&s,&t, Faktum) && TO_PhysischTeilterm(ti,RichtigeZielseiteLinks ? s : t);
}


static BOOLEAN RueckwartigeUeberlappung(RegelOderGleichungsT Vater, UTermT VaterStelle, RegelOderGleichungsT Mutter)
{
  BOOLEAN Resultat = ( IstVerdrehterKombinator(Mutter) && IstInRichtigerZielseite(VaterStelle,Vater) ) || 
                     ( KPTermeProduzieren(), IstUnifizierbaresZiel(KPLinks,KPRechts) ) || 
                     ( IstTrueGleichFalse(KPLinks,KPRechts) );
  if (!Resultat)
    HK_fair = FALSE;
  return Resultat;
}



/*=================================================================================================================*/
/*= XV. KP-Actions                                                                                                =*/
/*=================================================================================================================*/

/*   Die Vaterregel oder -gleichung habe die Gestalt l->r. Die Mutterregel oder -gleichung l'->r' gehe durch Va-
     riablenumbenennung aus l->r hervor, so dass die l->r und l'->r' variablendisjunkt sind. l an der Stelle
     p<>lambda und l' an der Stelle lambda seien mit MGU sigma unifizierbar. Dann ergibt sich die Divergenz
                                                  sigma(l)
                                                 /       \
                                          sigma(r)        sigma(l|p<-r')

     und damit das kritische Paar [sigma(r), sigma(l|p<-r')].
     Ist der Vater l->r eine Gleichung, so ist zusaetzlich sigma(l)!<sigma(r) zu ueberpruefen.
     Ist die Mutter l'->r' eine Gleichung, so ist sigma(l)!<sigma(l|p<-r') zu ueberpruefen.
*/

/* ------- 1. Einmalige Benachrichtigungen ----------------------------------------------------------------------- */

static struct GNTermpaarTS Elter1KopieS;
static const RegelOderGleichungsT Elter1Kopie = &Elter1KopieS;

#define /*TermT*/ l                                                                                                 \
  TP_LinkeSeite(Elter1Kopie)

static void Elter1Anmelden(RegelOderGleichungsT Elter1)
{
  TP_LinkeSeite(Elter1Kopie) = TO_TermkopieVarUmbenannt(TP_LinkeSeite(Elter1),BO_NeuesteBaumvariable());
  TP_RechteSeite(Elter1Kopie) = TO_TermkopieVarUmbenannt(TP_RechteSeite(Elter1),BO_NeuesteBaumvariable());
  SUE_NewParent(Elter1,TP_LinkeSeite(Elter1Kopie),TP_RechteSeite(Elter1Kopie));
}


static void Elter1Andersherum(void)
{
  TermT Merker = TP_LinkeSeite(Elter1Kopie);
  TP_LinkeSeite(Elter1Kopie) = TP_RechteSeite(Elter1Kopie);
  TP_RechteSeite(Elter1Kopie) = Merker;
  SUE_RedirectParent();
}


static void Elter1Abmelden(void)
{
  TO_TermeLoeschen(TP_LinkeSeite(Elter1Kopie),TP_RechteSeite(Elter1Kopie));
  SUE_FinishParent();
}



/* ------- 2. KPActions ------------------------------------------------------------------------------------------ */

#define /*BOOLEAN*/ TermGroesserUnif(/*UTermT*/ s, /*UTermT*/ t, /*{"RG","GG"}*/ Wo)                                \
  (ORD_TermGroesserUnif(s,t) ? TRUE : ((void)IncAnzUnifMisserfolgeOrdnung##Wo(), FALSE))

static BOOLEAN KPActionRR(RegelT Vater, UTermT VaterStelle, RegelT Mutter)
{
  Beginnen(aus,aus);
  if (Weggefiltert(Vater,VaterStelle,Mutter))
    Verwerfen()
  else 
    Vollenden(RR);
}


static BOOLEAN KPActionGR(GleichungsT Vater, UTermT VaterStelle, RegelT Mutter)
{
  Beginnen(an,aus);
  if (TermGroesserUnif(KPLinks,Ueberlappung,RG) || Weggefiltert(Vater,VaterStelle,Mutter))
    Verwerfen()
  else
    Vollenden(RG);
}


static BOOLEAN KPActionRG(RegelT Vater, UTermT VaterStelle, GleichungsT Mutter)
{
  Beginnen(aus,an);                                 /* Vergleich innen sowohl bei KBO als auch bei LPO ausreichend */
  if (TermGroesserUnif(KPRechtsInnen,UeberlappungInnen,RG) || Weggefiltert(Vater,VaterStelle,Mutter))
    Verwerfen()
  else
    Vollenden(RG);
}


static BOOLEAN KPActionGG(GleichungsT Vater, UTermT VaterStelle, GleichungsT Mutter)
{
  Beginnen(an,an);
  if (TermGroesserUnif(KPLinks,Ueberlappung,GG) || TermGroesserUnif(KPRechtsInnen,UeberlappungInnen,GG) || Weggefiltert(Vater,VaterStelle,Mutter))
    Verwerfen()
  else
    Vollenden(GG);
}



/* ------- 3. Benachrichtigungen ueber Ueberlappungen ------------------------------------------------------------ */

#define SUE_CPGeneratedWithVater  KPB_GebildetesKPBehandelnMitMutter
#define SUE_CPGeneratedWithMutter KPB_GebildetesKPBehandelnMitVater               /* Bezeichnung in SUE kreuzweise */


#define MacheElterMit(/*{Vater; Mutter}*/ Elternart, /*Bezeichner*/ KPAction,                                       \
  /*{R; G}*/ Gattung1, /*RegelOderGleichungsT*/ Elter2, /*UTermT*/ VaterStelle, /*UnifArtT*/ UnifArt)               \
{                                                                                                                   \
  if (KPAction) {                                                                                                   \
    SUE_CPGeneratedWith ## Elternart(Elter2,KPLinks,KPRechts,VaterStelle);                                          \
    IncAnzUnifErf ## UnifArt();                                                                                     \
  }                                                                                                                 \
}

#define MacheVaterMit(/*{R; G}*/ Gattung1, /*{R; G}*/ Gattung2, Elter2, VaterStelle, z)                             \
  MacheElterMit(Vater,KPAction ## Gattung1 ## Gattung2(Elter1Kopie,VaterStelle,Elter2),Gattung1,Elter2,VaterStelle,z)

#define MacheMutterMit(/*{R; G}*/ Gattung1,/*{R; G}*/ Gattung2, Elter2, VaterStelle, z)                             \
  MacheElterMit(Mutter,KPAction ## Gattung2 ## Gattung1(Elter2,VaterStelle,Elter1Kopie),Gattung1,Elter2,VaterStelle,z)

#define RVaterMitR(x,y,z)  MacheVaterMit (R,R,x,y,z)
#define RVaterMitG(x,y,z)  MacheVaterMit (R,G,x,y,z)
#define GVaterMitR(x,y,z)  MacheVaterMit (G,R,x,y,z)
#define GVaterMitG(x,y,z)  MacheVaterMit (G,G,x,y,z)
#define RMutterMitR(x,y,z) MacheMutterMit(R,R,x,y,z)
#define RMutterMitG(x,y,z) MacheMutterMit(R,G,x,y,z)
#define GMutterMitR(x,y,z) MacheMutterMit(G,R,x,y,z)
#define GMutterMitG(x,y,z) MacheMutterMit(G,G,x,y,z)



/* ------- 4. Sonstiges ------------------------------------------------------------------------------------------ */

#if TP_PETERGRAF
  #define /*Schleife*/ MitAllen(/*Block*/ Action)                                                                   \
  { Action; }

  #define /*Schleife*/ MitAllenAusser(/*GleichungsT*/ Ausschluss, /*Block*/ Action)                                 \
  { if (GleichungsIndex!=Ausschluss && GleichungsIndex!=TP_Gegenrichtung(Ausschluss)) {                             \
      Action;                                                                                                       \
  } }
#else
  #define /*Schleife*/ MitAllen(/*Block*/ Action)                                                                   \
  { GleichungsT GleichungsIndex = BK_Regeln(BaumIndex);                                                             \
    do {                                                                                                            \
      Action;                                                                                                       \
    } while ((GleichungsIndex = GleichungsIndex->Nachf));                                                           \
  }

  #define /*Schleife*/ MitAllenAusser(/*GleichungsT*/ Ausschluss, /*Block*/ Action)                                 \
  { GleichungsT GleichungsIndex = BK_Regeln(BaumIndex);                                                             \
    do                                                                                                              \
      if (GleichungsIndex!=Ausschluss && GleichungsIndex!=TP_Gegenrichtung(Ausschluss)) {                           \
        Action;                                                                                                     \
      }                                                                                                             \
    while ((GleichungsIndex = GleichungsIndex->Nachf));                                                             \
  }
#endif



/*=================================================================================================================*/
/*= XVI. KP-Bildung zu neuer Regel                                                                                =*/
/*=================================================================================================================*/

void U1_KPsBildenZuRegelFailing(RegelT Regel)
{
/* Die neue Regel sei l->r; R sei die Menge der linken Seiten der alten Regeln.
   Ueberlappungen zwischen Termen und Termmengen sind so zu interpretieren,
   dass der jeweilige Term mit allen Mengenelementen zu unifizieren ist.
   Ist die Menge aber leer, so ist nichts zu tun.
   Ueberlappungen von Mengen A =? B sind ueber das Kreuzprodukt zu bilden,
   also alle Aufgaben a =? b zu betrachten mit a E A, b E B.
   Es sei TT(t):={ t/p : p E O(t)} die Menge aller Teilterme von t (=>t E TT(t)
   und eTT(t):=TT(t)\{t}; ferner seien TT und eTT auf Mengen von Termen fortgesetzt.

   Folgende Ueberlappungen sind zu bilden:
   (1) l     =? eTT(l)
   (2) TT(l) =? R
   (3) l     =? eTT(R)

   Sei R':=R U {l}. Die Unifikation wurde hier so realisiert, dass zum Aufrufzeitpunkt
   die neue Regel schon im Baum steht. Dies stellt die volle Reduktionskraft zur Verfuegung
   und erspart das Puffern der kritischen Paare. Der zu zahlende Preis besteht in einer
   Anzahl ueberfluessiger erfolgreicher Unifikationen, ist jedoch niedriger als der Gewinn.

   Daher werden die Ueberlappungen wie folgt gebildet:
   (1) TT(l) =? R
   (2) l     =? eTT(R')
   Dabei muss R aus R' rekonstruiert werden. Dazu erhaelt TermMitDSBaumUnifizieren ein
   Ausschlussobjekt als zusaetzlichen Parameter. */


  /* 1. Regel kopieren, Variablendisjunktheit herstellen */

  UTermT IndexLinks;
  Elter1Anmelden(Regel);


  /* 2. linke Seite und alle Teilterme ausser Variablen mit Regeln auf Top-Level unifizieren */

  IndexLinks = l;
  do
    if (TO_TermIstNichtVar(IndexLinks)) {
      IncAnzUnifVersRegelnToplevel();
      TermMitDSBaumUnifizierenAusschluss(IndexLinks,RE_Regelbaum,Regel,
        RVaterMitR(BK_Regeln(BaumIndex),IndexLinks,RegelnToplevel));
    }
  while ((IndexLinks = TO_Schwanz(IndexLinks)));


  /* 3. linke Seite mit allen echten Teiltermen aus dem Baum ausser Variablen unifizieren */

  IncAnzUnifVersRegelnTeilterme();
  TermMitDSBaumTeiltermenUnifizieren(TP_LinkeSeite(Elter1Kopie),RE_Regelbaum,
    RMutterMitR(RegelIndex,LinkeSeiteIndex,RegelnTeilterme))


  /* 4. Nicht mehr benoetigte Termkopien freigeben */

  Elter1Abmelden();
}


void U1_KPsBildenZuRegelUnfailing(RegelT Regel)
{
/* Bezeichnungen wie vor; insbesondere l->r die neue Regel.
   E sei die Menge aller linken und rechten Seiten aller Gleichungen.

   Folgende Ueberlappungen sind zu bilden:
   (1) l     =? eTT(l)
   (2) TT(l) =? R
   (3) l     =? eTT(R)
   (4) TT(l) =? E
   (5) l     =? eTT(E) (jeweils mit >-Test)

   Sei R':=R U {l}. Die Unifikation wurde hier so realisiert, dass zum Aufrufzeitpunkt
   die neue Regel schon im Baum steht. Dies stellt die volle Reduktionskraft zur Verfuegung
   und erspart das Puffern der kritischen Paare. Der zu zahlende Preis besteht in einer
   Anzahl ueberfluessiger erfolgreicher Unifikationen, ist jedoch niedriger als der Gewinn.

   Daher werden die Ueberlappungen wie folgt gebildet:
   (1) TT(l) =? R
   (2) l     =? eTT(R')
   (3) TT(l) =? E
   (4) l     =? eTT(E) (jeweils mit >-Test)
   Dabei muss R aus R' rekonstruiert werden. Dazu erhaelt TermMitDSBaumUnifizieren ein
   Ausschlussobjekt als zusaetzlichen Parameter. */


  /* 1. Regel kopieren, Variablendisjunktheit herstellen */

  UTermT IndexLinks;
  Elter1Anmelden(Regel);


  /* 2. linke Seite und alle Teilterme ausser Variablen mit Regeln und Gleichungen auf Top-Level unifizieren */

  IndexLinks = l;
  do
    if (TO_TermIstNichtVar(IndexLinks)) {
      IncAnzUnifVersRegelnToplevel();
      TermMitDSBaumUnifizierenAusschluss(IndexLinks,RE_Regelbaum,Regel,
        RVaterMitR(BK_Regeln(BaumIndex),IndexLinks,RegelnToplevel));
      IncAnzUnifVersGleichgnToplevel();
      TermMitDSBaumUnifizieren(IndexLinks,RE_Gleichungsbaum,
        MitAllen(RVaterMitG(GleichungsIndex,IndexLinks,GleichgnToplevel)));
    }
  while ((IndexLinks = TO_Schwanz(IndexLinks)));


  /* 3. linke Seite mit allen echten Teiltermen aus Regel- und Gleichungsbaum ausser Variablen unifizieren */

  IncAnzUnifVersRegelnTeilterme();
  TermMitDSBaumTeiltermenUnifizieren(l,RE_Regelbaum,
    RMutterMitR(RegelIndex,LinkeSeiteIndex,RegelnTeilterme));
  IncAnzUnifVersGleichgnTeilterme();
  TermMitDSBaumTeiltermenUnifizieren(l,RE_Gleichungsbaum,
    RMutterMitG(RegelIndex,LinkeSeiteIndex,GleichgnTeilterme));


  /* 4. Nicht mehr benoetigte Termkopien freigeben */

  Elter1Abmelden();
}



/*=================================================================================================================*/
/*= XVII. KP-Bildung zu neuer Gleichung                                                                           =*/
/*=================================================================================================================*/

void U1_KPsBildenZuGleichung(GleichungsT Gleichung)
/* Die neue Gleichung sei gegeben durch eine ausgezeichnete Richtung l->r sowie die Gegenrichtung r'->l'.
   Alle Bezeichnungen wie vor. Sei E':=E U {l->r} U {r'->l'}.
   Die Unifikation wurde hier realisiert unter der Voraussetzung, dass zum Aufrufzeitpunkt
   die neue Gleichung schon mit Hin- und Herrichtung im Baum steht.

       zu bildende  |Entf||       realisiert in       |       Abfolge im Programm     |   Abfolge im Programm
     Ueberlappungen |Mono||                           |       bei Stereogleichung     |    bei Monogleichung
   -----------------+----++---------------------------+-------------------------------+-----------------------------
1   l     =? eTT(l) |    || l     =? eTT(E') U eTT(R) |  A  TT(l) =? E U R            | TT(l) =? E U R
2   r     =? eTT(r) | ** || r     =? eTT(E') U eTT(R) |  B  l     =? eTT(E') U eTT(R) | l     =? eTT(E') U eTT(R)
3   l     =? eTT(r) |    || l     =? eTT(E') U eTT(R) |  C  l     =? r                |
4   r     =? eTT(l) | ** || r     =? eTT(E') U eTT(R) |  D  TT(r) =? E U R            |
5   l     =? r      | ** || l     =? r                |  E  r     =? eTT(E') U eTT(R) |
                    |    ||                           |                               |
6   TT(l) =? E      |    || TT(l) =? E U R            |                               |
7   TT(r) =? E      | ** || TT(r) =? E U R            |                               |
8   l     =? eTT(E) |    || l     =? eTT(E') U eTT(R) |                               |
9   r     =? eTT(E) | ** || r     =? eTT(E') U eTT(R) |                               |
                    |    ||                           |                               |
10  TT(l) =? R      |    || TT(l) =? E U R            |                               |
11  TT(r) =? R      | ** || TT(r) =? E U R            |                               |
12  l     =? eTT(R) |    || l     =? eTT(E') U eTT(R) |                               |
13  r     =? eTT(R) | ** || r     =? eTT(E') U eTT(R) |                               |

*/

{
  /* 0. Antigleichung herausfischen und Variablendisjunktheit herstellen */

  UTermT Index;
  Elter1Anmelden(Gleichung);


  /* A.  TT(l) =? E U R                      (6, 10) */

  Index = l;
  do
    if (TO_TermIstNichtVar(Index)) {
      IncAnzUnifVersRegelnToplevel();
      TermMitDSBaumUnifizieren(Index,RE_Regelbaum,
        GVaterMitR(BK_Regeln(BaumIndex),Index,RegelnToplevel));
      IncAnzUnifVersGleichgnToplevel();
      TermMitDSBaumUnifizieren(Index,RE_Gleichungsbaum,
        MitAllenAusser(Gleichung,GVaterMitG(GleichungsIndex,Index,GleichgnToplevel)))
    }
  while ((Index = TO_Schwanz(Index)));


  /* B.  l     =? eTT(E') U eTT(R)     (1, 3, 8, 12) */

  IncAnzUnifVersRegelnTeilterme();
  TermMitDSBaumTeiltermenUnifizieren(l,RE_Regelbaum,
    GMutterMitR(RegelIndex,LinkeSeiteIndex,RegelnTeilterme));
  IncAnzUnifVersGleichgnTeilterme();
  TermMitDSBaumTeiltermenUnifizieren(l,RE_Gleichungsbaum,
    GMutterMitG(RegelIndex,LinkeSeiteIndex,GleichgnTeilterme));


  /* F.  l     =? l                                  */

  IncAnzUnifVersGleichgnEigeneTterme();
  Term1MitTerm2Unifizieren(l,TP_LinkeSeite(Gleichung),
    GVaterMitG(Gleichung,l,GleichgnEigeneTterme));


  /* C.  l     =? r                              (5)      Ab hier nur noch fuer Stereogleichung */

  Elter1Andersherum();                                         /* vertauscht in Elter1Kopie linke und rechte Seite */
  if (TP_IstKeineMonogleichung(Gleichung)) {
    IncAnzUnifVersGleichgnEigeneTterme();
    Term1MitTerm2Unifizieren(l,TP_LinkeSeite(Gleichung),
      GVaterMitG(Gleichung,l,GleichgnEigeneTterme));


  /* D.  TT(r) =? E U R                      (7, 11) */

    Index = l;
    do
      if (TO_TermIstNichtVar(Index)) {
        IncAnzUnifVersGleichgnToplevel();
        TermMitDSBaumUnifizieren(Index,RE_Gleichungsbaum,
          MitAllenAusser(Gleichung,GVaterMitG(GleichungsIndex,Index,GleichgnToplevel)));
        IncAnzUnifVersRegelnToplevel();
        TermMitDSBaumUnifizieren(Index,RE_Regelbaum,
          GVaterMitR(BK_Regeln(BaumIndex),Index,RegelnToplevel));
      }
    while ((Index = TO_Schwanz(Index)));


  /* E.  r     =? eTT(E') U eTT(R)     (2, 4, 9, 13) */

    IncAnzUnifVersGleichgnTeilterme();
    TermMitDSBaumTeiltermenUnifizieren(l,RE_Gleichungsbaum,
      GMutterMitG(RegelIndex,LinkeSeiteIndex,GleichgnTeilterme));
    IncAnzUnifVersRegelnTeilterme();
    TermMitDSBaumTeiltermenUnifizieren(l,RE_Regelbaum,
      GMutterMitR(RegelIndex,LinkeSeiteIndex,RegelnTeilterme));


  /* G.  r     =? r                                  */

    IncAnzUnifVersGleichgnEigeneTterme();
    Term1MitTerm2Unifizieren(l,TP_LinkeSeite(TP_Gegenrichtung(Gleichung)),
      GVaterMitG(TP_Gegenrichtung(Gleichung),l,GleichgnEigeneTterme));
  }


  /* 6. Kopie der Antigleichung wieder freigeben */

  Elter1Abmelden();
}



/*=================================================================================================================*/
/*= XVIII. Rekonstruktion von kritischen Paaren                                                                   =*/
/*=================================================================================================================*/

static void SubstErgaenzenOhneOC(SymbolT Variable, UTermT Term)
{                                                            /* fuer sowieso unifizierbare Terme wie SubstErgaenzt */
  
  SymbolT Topsymbol = TO_TopSymbol(Term);                                 /* triviale Bindung x(i)<-x(i) abfangen; */
  Entbinden(Topsymbol,Term);
  if (SO_SymbUngleich(Topsymbol,Variable)) {
    xU[Variable].Substitut = Term;
    PushVariable(Variable);
    AnzImKnotenAngelegterBindungen++;
  }
}


static void Reunifizieren(UTermT Term1, UTermT Term2)
{
  UTermT Term1Ende = TO_TermEnde(Term1),
         Term2Ende = TO_TermEnde(Term2);
  Endestapel12Initialisieren();
  do {
    SymbolT Symbol1 = TO_TopSymbol(Term1),
            Symbol2 = TO_TopSymbol(Term2);
  Start:
    if (SO_SymbUngleich(Symbol1,Symbol2)) {
      if (SO_SymbIstVar(Symbol1)) {
        if (VarGebunden(Symbol1)) {
          AufSubstitut(1);
          goto Start;
        }
        else {
          SubstErgaenzenOhneOC(Symbol1,Term2);
          AufTeiltermende(2);
        }
      }
      else                                                                           /* Symbol2 muss Variable sein */
        if (VarGebunden(Symbol2)) {
          AufSubstitut(2);
          goto Start;
        }
        else {
          SubstErgaenzenOhneOC(Symbol2,Term1);
          AufTeiltermende(1);
        }
    }
  } while (DoppeltWeitergeschaltet12(&Term1,&Term1Ende,&Term2,&Term2Ende));
}


void U1_KPRekonstruieren(RegelOderGleichungsT Vater, unsigned int Stellennummer, RegelOderGleichungsT Mutter,
/* Die Stellennummer wird ab Null gezaehlt. */                                       TermT *KPLinks, TermT *KPRechts)
{
  UTermT Stelle = TO_TermAnStelle(TP_LinkeSeite(Vater),Stellennummer);

  if (Vater==Mutter) {                                          /* dann kann Mutter nicht variablenumbenannt werden */
    if (TP_Gegenrichtung(Vater)) {
      GleichungsT Antigleichung = TP_Gegenrichtung(Vater);
      RE_RegelExternalisieren(Antigleichung);
      Reunifizieren(Stelle,TP_RechteSeite(Antigleichung));
      *KPRechts = MGUNachTE(TP_LinkeSeite(Vater),Stelle,TP_LinkeSeite(Antigleichung));
      *KPLinks = MGU(TP_RechteSeite(Vater));
      RE_RegelReinternalisieren(Antigleichung);
    }
    else {
      TermT MutterLinks = TO_TermkopieVarUmbenannt(TP_LinkeSeite(Mutter),BO_NeuesteBaumvariable()),
            MutterRechts = TO_TermkopieVarUmbenannt(TP_RechteSeite(Mutter),BO_NeuesteBaumvariable());
      Reunifizieren(Stelle,MutterLinks);
      *KPRechts = MGUNachTE(TP_LinkeSeite(Vater),Stelle,MutterRechts);
      *KPLinks = MGU(TP_RechteSeite(Vater));
      TO_TermeLoeschen(MutterLinks,MutterRechts);
    }
  }
  else {
    RE_RegelExternalisieren(Mutter);
    Reunifizieren(Stelle,TP_LinkeSeite(Mutter));
    *KPRechts = MGUNachTE(TP_LinkeSeite(Vater),Stelle,TP_RechteSeite(Mutter));
    *KPLinks = MGU(TP_RechteSeite(Vater));
    RE_RegelReinternalisieren(Mutter);
  }
  KnotenbindungenAufheben();
}

void U1_KPAusTermenRekonstruieren(TermT VaterLinks, TermT VaterRechts, TermT MutterLinks, TermT MutterRechts, 
                                  unsigned int Stellennummer, TermT *KPLinks, TermT *KPRechts)
/* spezielle Fassung von U1_KPRekonstruieren ohne Termpaare fuer Bernd. */
{
  UTermT Stelle = TO_TermAnStelle(VaterLinks,Stellennummer);
  
  if ((VaterLinks == MutterLinks) || (VaterLinks == MutterRechts)){/* CP mit sich selber ? */
    MutterLinks  = TO_TermkopieVarUmbenannt(MutterLinks,  BO_NeuesteBaumvariable());
    MutterRechts = TO_TermkopieVarUmbenannt(MutterRechts, BO_NeuesteBaumvariable());
    Reunifizieren(Stelle, MutterLinks);
    *KPRechts = MGUNachTE(VaterLinks,Stelle,MutterRechts);
    *KPLinks  = MGU(VaterRechts);
    TO_TermeLoeschen(MutterLinks,MutterRechts);
  }
  else{
    signed int Delta = SO_VarNummer(BO_NeuesteBaumvariable());
    TO_VariablenUmbenennen(MutterLinks,Delta);
    TO_VariablenUmbenennen(MutterRechts,Delta);
    Reunifizieren(Stelle, MutterLinks);
    *KPRechts = MGUNachTE(VaterLinks,Stelle,MutterRechts);
    *KPLinks  = MGU(VaterRechts);
    TO_VariablenUmbenennen(MutterLinks,-Delta);
    TO_VariablenUmbenennen(MutterRechts,-Delta);
  }
  KnotenbindungenAufheben();
}



/*=================================================================================================================*/
/*= XIX. Mass fuer relative Unifizierbarkeit                                                                      =*/
/*=================================================================================================================*/

static void Durchsubstituieren(SymbolT VarSymbol, UTermT SubstitutOriginal,
                               TermT *Start, UTermT *Vorg, UTermT *Stelle)
/* In der Termzellenliste *Start bis NULL wird jedes Vorkommen der 
   Variablen VarSymbol ersetzt durch eine Kopie von SubstitutOriginal.
   Falls es Ersetzungen gibt an den Stellen *Start, *Vorg, *Stelle,
   so werden die entsprechend bezeichneten Variablen ebenfalls veraendert.
   Die Termzellenliste darf auch schon leer sein.
   Die Ende-Zeiger der vorhergehenden Zellen ab TermAnfang werden korrekt angepasst. */
{
  UTermT TermIndexVorg = NULL,
         TermIndex = *Start;
  while (TermIndex) {
    if (SO_SymbGleich(TO_TopSymbol(TermIndex),VarSymbol)) {
      UTermT Substitut = TO_Termkopie(SubstitutOriginal),
             SubstitutEnde = TO_TermEnde(Substitut),
             TermIndex2 = *Start;
      while (TO_PhysischUngleich(TermIndex2,TermIndex)) {                    /* vorhergehende Ende-Zeiger anpassen */
        if (TO_PhysischGleich(TO_TermEnde(TermIndex2),TermIndex))
          TO_EndeBesetzen(TermIndex2,SubstitutEnde);
        TermIndex2 = TO_Schwanz(TermIndex2);
      }
      TO_NachfBesetzen(SubstitutEnde,TO_Schwanz(TermIndex));                        /* Nachf-Verzeigerung anpassen */
      if (TermIndexVorg)
        TO_NachfBesetzen(TermIndexVorg,Substitut);
      else
        *Start = Substitut;                                                           /* ggf. *Start aktualisieren */
      if (TO_PhysischGleich(TermIndex,*Vorg))
        *Vorg = SubstitutEnde;
      if (TO_PhysischGleich(TermIndex,*Stelle))
        *Stelle = Substitut;
      TO_TermzelleLoeschen(TermIndex);
      TermIndex = SubstitutEnde;
    }
    TermIndex = TO_Schwanz(TermIndexVorg = TermIndex);
  }
}


static void ZelleSchneiden(TermT *Start, UTermT *Vorg, UTermT *Term)
{
  TermT Nachf = TO_NaechsterTeilterm(*Term);
  if (*Vorg)
    TO_NachfBesetzen(*Vorg,TO_Schwanz(*Term));
  else
    *Start = TO_Schwanz(*Term);
  if (TO_NichtNullstellig(*Term))
    *Vorg = TO_TermEnde(*Term);
  TO_TermzelleLoeschen(*Term);
  *Term = Nachf;
}


static void TermSchneiden(TermT *Start, UTermT Vorg, UTermT *Term)
{
  TermT Nachf = TO_NaechsterTeilterm(*Term);
  if (Vorg)
    TO_NachfBesetzen(Vorg,Nachf);
  else
    *Start = Nachf;
  TO_TermLoeschen(*Term);
  *Term = Nachf;
}


static void VariableBinden(TermT *VarTermStart, UTermT *VarTermVorg, UTermT *VarTerm, 
                           TermT *SubstTermStart, UTermT *SubstTermVorg, UTermT *SubstTerm,
                           long unsigned int d, long unsigned int *Mass)
{
  UTermT TermIndex = TO_Schwanz(*SubstTerm),
         NULLTerm = TO_NULLTerm(*SubstTerm);
  BOOLEAN ok = TRUE;
  while (TO_PhysischUngleich(TermIndex,NULLTerm)) {                                                 /* occur check */
    if (TO_TopSymboleGleich(*VarTerm,TermIndex)) {
      *Mass += 2*d;
      ok = FALSE;
    }
    TermIndex = TO_Schwanz(TermIndex);
  }
  if (ok) {
    SymbolT VarSymbol = TO_TopSymbol(*VarTerm);
    TermT Substitut = *SubstTerm;
    TermSchneiden(VarTermStart,*VarTermVorg,VarTerm);
    Durchsubstituieren(VarSymbol,Substitut,VarTermStart,VarTermVorg,VarTerm);
    Durchsubstituieren(VarSymbol,Substitut,SubstTermStart,SubstTermVorg,SubstTerm);
  }
  else
    TermSchneiden(VarTermStart,*VarTermVorg,VarTerm);
  TermSchneiden(SubstTermStart,*SubstTermVorg,SubstTerm);
}


unsigned long int U1_Unifikationsmass(TermT links, TermT rechts)
{                         /* links und rechts muessen vom Typ TermT sein. Rueckgabe 0 <=> Terme sind unifizierbar. */
  unsigned long int TiefeLinks = TO_Termtiefe(links),
                    TiefeRechts = TO_Termtiefe(rechts),
                    d = max(TiefeLinks,TiefeRechts),
                    Mass = 0;
  TermT linksStart = TO_Termkopie(links),
        rechtsStart = TO_Termkopie(rechts);
  do {
    UTermT linksVorg = NULL,
           rechtsVorg = NULL;
    links = linksStart;
    rechts = rechtsStart;
    do {
      if (TO_TopSymboleGleich(links,rechts)) {
        ZelleSchneiden(&linksStart,&linksVorg,&links);
        ZelleSchneiden(&rechtsStart,&rechtsVorg,&rechts);
      }
      else if (TO_TermIstVar(links))
        VariableBinden(&linksStart,&linksVorg,&links,
                       &rechtsStart,&rechtsVorg,&rechts,
                       d,&Mass);
      else if (TO_TermIstVar(rechts))
        VariableBinden(&rechtsStart,&rechtsVorg,&rechts,
                       &linksStart,&linksVorg,&links,
                       d,&Mass);
      else {
        Mass += d;
        TermSchneiden(&linksStart,linksVorg,&links);
        TermSchneiden(&rechtsStart,rechtsVorg,&rechts);
      }
    } while (links);
   if (!(--d))
     d = 1;
  } while (linksStart);
  return Mass;
}



/*=================================================================================================================*/
/*= XX. Statistik                                                                                                 =*/
/*=================================================================================================================*/

void U1_Statistik(InfoStringT IS)
{
  if (!(IS_IsIn("all", IS) || IS_IsIn("U", IS)))
    return;
  if (IS_IsIn("standard", IS)){
    int tiefe1;
    if(WerteDa(2,AnzUnifVersRegelnToplevel, AnzUnifVersGleichgnToplevel)){
      IO_DruckeKurzUeberschrift("Unifikation", '+');
      tiefe1 = IO_Einruecktiefe(0, Maximum(6, AnzUnifErfRegelnToplevel, AnzUnifErfRegelnTeilterme,
                                           AnzUnifErfRegelnEigeneTterme, AnzUnifErfGleichgnToplevel,
                                           AnzUnifErfGleichgnTeilterme, AnzUnifErfGleichgnEigeneTterme));
      IO_DruckeStatistikZeile("Kapur-Opfer", AnzKPOpferKapur);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Unterdrueckte Nicht-Eins-Stern-Paare", AnzNonEinsSternKPs);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Unterdrueckte Nusfu-Paare", AnzNonNusfuKPs);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Unifikationsversuche mit der Regelmenge Toplevel",AnzUnifVersRegelnToplevel);
      IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfRegelnToplevel, tiefe1);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Unifikationsversuche mit Teiltermen der Regelmenge",AnzUnifVersRegelnTeilterme);
      IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfRegelnTeilterme, tiefe1);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Unifikationsversuche mit eigenen Teiltermen (RM)",AnzUnifVersRegelnEigeneTterme);
      IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfRegelnEigeneTterme, tiefe1);
      if (RE_AnzGleichungenGesamt > 0){
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Unifikationsversuche mit der Gleichungsmenge Toplevel",AnzUnifVersGleichgnToplevel);
        IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfGleichgnToplevel, tiefe1);
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Unifikationsversuche mit Teiltermen der Gleichungsmenge",AnzUnifVersGleichgnTeilterme);
        IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfGleichgnTeilterme, tiefe1);
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Unifikationsversuche mit eigenen Teiltermen (GM)",AnzUnifVersGleichgnEigeneTterme);
        IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",AnzUnifErfGleichgnEigeneTterme, tiefe1);
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Ueberlappungen erzeugt", AnzUnifUeberlappungenGebildet);
        IO_DruckeStatistikZeile("Wegen >-Test fehlgeschlagene KP-Bildungen Regel-Gleichung", AnzUnifMisserfolgeOrdnungRG);
        IO_DruckeStatistikZeile("Wegen >-Test fehlgeschlagene KP-Bildungen Gleichung-Gleichung", AnzUnifMisserfolgeOrdnungGG);
        IO_DruckeStatistikZeile("Dabei ueberfluessigerweise gebildete kritische Terme", AnzUnifUeberflKritTerme);
      }
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeile("Insgesamt erzeugte KPs", 
                              AnzUnifErfRegelnToplevel+AnzUnifErfRegelnTeilterme+
                              AnzUnifErfRegelnEigeneTterme+AnzUnifErfGleichgnToplevel+
                              AnzUnifErfGleichgnTeilterme+AnzUnifErfGleichgnEigeneTterme);
      tiefe1 = IO_Einruecktiefe(0, Maximum(3, AnzKPRR, AnzKPGG,AnzKPRG));
      IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen 2 Regeln", AnzKPRR, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen 2 Gleichungen", AnzKPGG, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen Regel und Gleichung", AnzKPRG, tiefe1);
    }
  }
  else if (IS_IsIn("latex-head", IS)){
    fprintf(IO_Strom, "Unif: Kapur-Vict & Unif R top & succ & Unif R st & succ & Unif own st & succ & ");
    if (RE_AnzGleichungenGesamt > 0)
      fprintf(IO_Strom, "Unif E top & succ & Unif E st & succ & Unif own st & succ & overlaps & failed > RE & failed > EE & superfl terms & ");
    fprintf(IO_Strom, "CPs total & RR & EE & RE & ");
  }
  else if (IS_IsIn("intern", IS)){
    IO_DruckeStatistikZeile("Maximale Ausdehnung der Unifikationsvariablenbank", -xUDimension);
    IO_DruckeStatistikZeile("Maximale Ausdehnung des Unifikationsstacks", -xUDimension);
    IO_DruckeStatistikZeile("Maximale Ausdehnung der drei U1_Endestapel", -xUDimension);
    IO_DruckeStatistikZeile("Maximale Ausdehnung des U1_Kopierstapels", U1_AusdehnungKopierstapel());
    IO_DruckeStatistikZeile("Maximale Ausdehnung des U1_Kopierstapels2", U1_AusdehnungKopierstapel2());
  }
  else if (IS_IsIn("config", IS)){
#if U1_STATISTIK
    fprintf(IO_Strom, "U1: Statistik  AN\n");
#else
    fprintf(IO_Strom, "U1: Statistik AUS\n");
#endif
  }
}



/*=================================================================================================================*/
/*= XXI. Initialisierung                                                                                          =*/
/*=================================================================================================================*/


void U1_InitApriori(void)
{
  if (WM_ErsterAnlauf()) {
    SymbolT VarIndex;
    xU = negarray_alloc(U1_AnzErsterVariablen,U1_EintragsT,U1_AnzErsterVariablen);
    xUDimension = -U1_AnzErsterVariablen;
    SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb,xUDimension)
      xU[VarIndex].Substitut = NULL;
    AnzImKnotenAngelegterBindungen = 0;
    StackZeiger = 0;
    BindungsStack = posarray_alloc(U1_AnzErsterVariablen,StackElementT,1);
    Endestapel1 = posarray_alloc(U1_AnzErsterVariablen,EndeEintragsT,1);
    Endestapel2 = posarray_alloc(U1_AnzErsterVariablen,EndeEintragsT,1);
    Endestapel3 = posarray_alloc(U1_AnzErsterVariablen,EndeEintragsT,1);
    TI_IterationInitialisieren();
    TI2_IterationInitialisieren();
  }
  KPFilterInitialisieren();
}

void U1_ParamInfo(ParamInfoArtT what, char* KapurString)
{
   if (what == ParamEvaluate){
      int KapurStaerke = 0;
      fprintf(IO_Strom, "\n* SUBCONNECTEDNESS CRITERION:   ");
      if (strchr(KapurString, 'r'))
         KapurStaerke += 1;
      if (strchr(KapurString, 'e'))
         KapurStaerke += 2;
      switch (KapurStaerke){
      case 1:
         fprintf(IO_Strom, "R\n");
         break;
      case 2:
         fprintf(IO_Strom, "E\n");
         break;
      case 3:
         fprintf(IO_Strom, "RE\n");
         break;
      case 0:
      default:
         fprintf(IO_Strom, "-\n");
         break;
      }
   }
   else{
      fprintf(IO_Strom, "Subconnectedness Criterion\n");
      fprintf(IO_Strom, "**************************\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "[r][e]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Strength of the subconnectedness-criterion:\n");
      fprintf(IO_Strom, "  'r':  application of rules,\n");
      fprintf(IO_Strom, "  'e':  application of equations,\n");
      fprintf(IO_Strom, "  're': application of rules and equations.  \n");
   }
}


void U1_InitAposteriori(char* KapurString)
{
  unsigned int Kapurstaerke = 0;
  if (strchr(KapurString, 'r'))
    Kapurstaerke += 1;
  if (strchr(KapurString, 'e'))
    Kapurstaerke += 2;
  if (Kapurstaerke) {
    NFInnenReduzibelT *Kapurnormalisierungen[] = { 
                         &NFInnenReduzibelProzR, 
                         &NFInnenReduzibelProzE, 
                         &NFInnenReduzibelProzRE };
    Kapurreduzibel = Kapurnormalisierungen[Kapurstaerke-1];
    KPFilterErgaenzen(Primueberlappung);
  }
  if (PA_EinsSternAn)
    KPFilterErgaenzen(EinsSternUeberlappung);
  if (PA_NusfuAn)
    KPFilterErgaenzen(NusfUeberlappung);
  if (PA_KernAn)
    KPFilterErgaenzen(KernUeberlappung);
  if (PA_BackAn) {
    TO_KombinatorapparatAktivieren();
    RichtigeZielseiteSuchen();
    KPFilterErgaenzen(RueckwartigeUeberlappung);
  }
#if !COMP_POWER
  fprintf(IO_Strom, "\n* SUBCONNECTEDNESS CRITERION:  %s%s%s\n", 
    Kapurstaerke & 1 ? "R" : "", Kapurstaerke & 2 ? "E" : "", !Kapurstaerke ? "-" : "");
#endif
}
