/**********************************************************************
* Filename : MNF.c
*
* Kurzbeschreibung : MultipleNormalFormen-Bildung
*                    (Verwalten von Nachfolgermengen)
*
* Autoren : Andreas Jaeger
*
* 10.07.96: Erstellung
* 22.02.97: Mehrfachaufrufclean. AJ.
*
*
*
* Benoetigt folgende Module:
*  - MatchOperationen: Reduktion der Terme
*  - Ausgaben: Ausgeben von Termen (wird beim Testen benutzt)
*  - NFBasis: Reduzieren
*  - RUndEVerwaltung: Zum Testen, ob Regelmenge leer ist
*  - Hash: Verwalten der Nachfolgermenge
*  - MNF_PQ: Verwalten der noch zu bearbeitenden Elemente
*  - MNF_Print: Ausgeben der MNF-Menge
*  - Hauptkomponenten: Ordnungsvergleich
*  - DSBaumOperationen: Normieren von Termpaaren
*
* Benutzte Globale Variablen aus anderen Modulen:
*   zur Zeit keine
*
**********************************************************************/

/* ToDo:
  - wie soll Subsumptionstest bei mnfanalyse geschehen?
  - CounterT in interne Structs einfuegen?
  */

/* Konzept fuer Mehrfachaufruf:

   InitAposteriori legt Speicher an
   CleanUp gibt Speicher frei und setzt Variablen wieder auf Initialzustand.
 */

#include "antiwarnings.h"
#include "general.h"
#include "compiler-extra.h"
#include "Hash.h"
#include "MNF_intern.h"
#include "MNF.h"
#include "MNF_Print.h"
#include "MNF_PQ.h"
#include "MNF_CPs.h"
#include "NFBildung.h"
#include "DSBaumOperationen.h"
#include "MatchOperationen.h"
#include "NFBasis.h"
#include "Ordnungen.h"
#include "Ausgaben.h"
#include "RUndEVerwaltung.h"
#include "Hauptkomponenten.h"
#include "Signale.h"
#include "SpeicherVerwaltung.h"
#include "Statistiken.h"
#include "Unifikation1.h"
#include "VariablenMengen.h"
#include "pcl.h"
#include <string.h>

#include "Stack.h"
#include "Deque.h"

#define MNF_STATISTIK normal_power_comp(1,0,0)


typedef enum {ARnoUse, ARuseAnti, ARuseUnifikation} AntiRegelUseT;

   
/* Funktionsprototypen */
static void MNF_BearbeiteRE(void *pq, MNF_MengeT *menge, int colour, MNF_ColTermPtrT acctCTerm);
static void MNF_BearbeiteR(void *pq, MNF_MengeT *menge, int colour , MNF_ColTermPtrT acctCTerm);
static void MNF_BearbeiteE(void *pq, MNF_MengeT *menge, int colour , MNF_ColTermPtrT acctCTerm);
static void MNF_BearbeiteDummy (void *pq, MNF_MengeT *menge, int colour, MNF_ColTermPtrT acctCTerm);
static void MNF_BearbeiteEineR (void *stack, MNF_MengeT *menge, int colour,
                               MNF_ColTermPtrT acctCTerm, RegelT Regel);
static unsigned int count_vars (TermT term) MAYBE_UNUSEDFUNC;
static void MNF_ConvertRegelToAntiRegel (MNF_AntiT antiKind,
                                         RegelT Regel,
                                         TermT *antiLinks, TermT *antiRechts,
                                         AntiRegelUseT *useAnti);
static void SetColTerm(/*@out@*/ MNF_ColTermPtrT *colTermPtr, TermT term,
                       int colour,
                       MNF_ColTermPtrT parentPtr,
                       VergleichsT Vergleich,
                       long int RENummer,
                       HashKeyT *Key1,
                       HashKeyT *Key2);
#if COMP_PCL
# define SetPclInfo(pclInfo,redPosition,REpclID,RESeite, parentPtr, RENummer) \
    SetPclInfoI(pclInfo,redPosition,REpclID,RESeite, parentPtr, RENummer)
     
static void SetPclInfoI(MNF_PCLInfoPtrT pclInfo, TermT redPosition,
                       PCL_ObjektNummerT REpclID, PCL_SeitenT RESeite,
                       MNF_ColTermPtrT parentPtr, long int RENummer);
#else
# define SetPclInfo(pclInfo,redPosition,REpclID,RESeite, parentPtr, RENummer) \
    SetPclInfoI(pclInfo,redPosition, parentPtr, RENummer)
static void SetPclInfoI(MNF_PCLInfoPtrT pclInfo, TermT redPosition,
                       MNF_ColTermPtrT parentPtr, long int RENummer);
#endif

static void DeallocData (MNF_ColTermT *colTerm);
static void MNF_PseudoExpand (MNF_MengeT *menge);
static void MNF_SearchInteresting (MNF_MengeT *menge);
static BOOLEAN MNF_PseudoDurchlaufPaarweise (MNF_MengeT *menge, MNF_DurchlaufPaarProcT DurchlaufProc);



/* Variablen fuer Initialisierungen */
static BOOLEAN           MNF_Initialised = FALSE;

#if MNF_STATISTIK
/* Variablen fuer Statistiken, werden automatisch auf 0 initialisiert. */
static CounterT MNF_AnzDirektZus,
   MNF_AnzReduktionen,
   MNF_AnzMatchVersucheE,
   MNF_AnzMatchVersucheR,
   MNF_AnzCalls,
   MNF_AnzTermGleich,
   MNF_AnzTermGleichCol,
   MNF_AnzTermGleichOpp,
   MNF_MaxNachfolger,
   MNF_MaxEQUp,
   MNF_AbbruchMaxSize,
   MNF_SumAnzNachfolger,
   MNF_AnzCallsHAAddData,
   MNF_AnzAntiR,
   MNF_AnzEQUp;
#endif /* MNF_STATISTIK */

#if MNF_STATISTIK
# define MNF_IncAnzReduktionen      (MNF_AnzReduktionen++)
# define MNF_IncAnzCalls            (MNF_AnzCalls++)
# define MNF_IncAnzTermGleich       (MNF_AnzTermGleich++)
# define MNF_IncAnzTermGleichCol    (MNF_AnzTermGleichCol++)
# define MNF_IncAnzTermGleichOpp    (MNF_AnzTermGleichOpp++)
# define MNF_IncAbbruchMaxSize      (MNF_AbbruchMaxSize++)
# define MNF_IncAnzDirektZusammenfuehrbar (MNF_AnzDirektZus++)
# define MNF_IncAnzCallsHAAddData   (MNF_AnzCallsHAAddData++)
# define MNF_IncAnzEQUp             (MNF_AnzEQUp++)
# define MNF_IncAnzAntiR            (MNF_AnzAntiR++)
# define MNF_IncAnzMatchVersucheE   (MNF_AnzMatchVersucheE++)
# define MNF_IncAnzMatchVersucheR   (MNF_AnzMatchVersucheR++)
#else
# define MNF_IncAnzReduktionen
# define MNF_IncAnzCalls
# define MNF_IncAnzTermGleich
# define MNF_IncAnzTermGleichCol
# define MNF_IncAnzTermGleichOpp
# define MNF_IncAbbruchMaxSize
# define MNF_IncAnzDirektZusammenfuehrbar
# define MNF_IncAnzCallsHAAddData
# define MNF_IncAnzEQUp
# define MNF_IncAnzAntiR
# define MNF_IncAnzMatchVersucheE
# define MNF_IncAnzMatchVersucheR
#endif /* MNF_STATISTIK */


/* diverse Variablen */
static SV_SpeicherManagerT MNF_ColTermSpeicher;
static SV_SpeicherManagerT MNF_MengeSpeicher;

/* Variablen fuer Zielverwaltung */
static BOOLEAN             NMuseE;
static BOOLEAN             NMuseR;
static BOOLEAN             NMuseAnalyse;
static unsigned int        NMSwitchIrred;
static unsigned int        NMSwitchNF;
static unsigned int        NMCountNF;
static MNFPQ_ArtT          NMMNFPQ_Art;
static unsigned int        NMmaxSize;
static unsigned int        NMmaxGlSchritte;
static unsigned int        NMmaxAntiRSchritte;
static BOOLEAN             NMAddTryAllElements = TRUE;
static MNF_AntiT           NMantiKind;
static HashSizeT           NMHashSize;

/* wir verwalten einen eigenen Zaehler fuer die Elemente - dieser hat nichts
   mit dem PCL-Zaehlern zu tun, es ist nur der gleiche Typ */
static PCL_ObjektNummerT current_pclid;

/**************************************************************************************/
/* Initialisierung/Aufraeumen/Statistik des Moduls                                    */
/**************************************************************************************/


void MNF_InitAposteriori (void) 
{
   HA_InitAposteriori ();
   DQ_InitAposteriori ();
   
   current_pclid = 1;
   
   MNF_Initialised = TRUE;
   SV_ManagerInitialisieren (&MNF_ColTermSpeicher, MNF_ColTermT, MNF_ColTermBlock, "Coloured Terms (MNF)");
   SV_ManagerInitialisieren (&MNF_MengeSpeicher, MNF_MengeT, MNF_MengeSpeicherBlock, "MNF Set (MNF)");
}


void MNF_Cleanup (void) {
   if (!MNF_Initialised)
      return;
   MNF_CPCleanup ();

}



void MNF_Statistik (InfoStringT IS) {

   
   if (!(IS_IsIn("all", IS) || IS_IsIn("MNF", IS)))
      return;

   if (MNF_Initialised && IS_IsIn("standard", IS)) {
#if MNF_STATISTIK   
      IO_DruckeKurzUeberschrift("MultipleNormalForms",'+');

      IO_DruckeStatistikZeile("Total Termpairs", MNF_AnzCalls);
      IO_DruckeStatistikZeile("Joinable Termpairs", MNF_AnzTermGleichOpp);
      IO_DruckeStatistikZeile("Immediatly joinable Termpairs", MNF_AnzDirektZus);
      IO_DruckeStatistikZeile("Reductions", MNF_AnzReduktionen);
      IO_DruckeStatistikZeile("Match queries with equations", MNF_AnzMatchVersucheE);
      IO_DruckeStatistikZeile("Match queries with rules", MNF_AnzMatchVersucheR);
      IO_DruckeStatistikZeile("Maximal Size of MNF-Set", MNF_MaxNachfolger);
      IO_DruckeStatistikZeile("Stops: MNF-set > maxSize", MNF_AbbruchMaxSize);
      IO_DruckeStatistikZeile("Total steps up with E", MNF_AnzEQUp);
      IO_DruckeStatistikZeile("Maximal steps up with E", MNF_MaxEQUp);
      IO_DruckeStatistikZeile("Total steps up with R", MNF_AnzAntiR);
   }
   else if (MNF_Initialised && IS_IsIn("latex-head", IS)) { /* korrespondiert zu standard */
      fprintf(IO_Strom, "MNF: total CP & MNF joined & MNF imm. joined"
              "& MNF Red. & Match Queries E & Match Queries R"
              "& MNF Max Size & MNF too big & Tot. Up & Max UpE & Tot UpR&"
             );
   }
   else if (MNF_Initialised && IS_IsIn("short", IS)){
      IO_DruckeKurzUeberschrift("MultipleNormalForms",'+');
      IO_DruckeStatistikZeile("MNF Reductions", MNF_AnzReduktionen);
      IO_DruckeStatistikZeile("Total steps up (R+E)", MNF_AnzEQUp+MNF_AnzAntiR);
   }
   else if (MNF_Initialised && IS_IsIn("slatex-head", IS)){
      fprintf(IO_Strom, "MNF: Red &Tot. Up& ");
   }
   else if (MNF_Initialised && IS_IsIn("intern", IS)) {
      IO_DruckeStatistikZeile("Calls of HA_AddData", MNF_AnzCallsHAAddData);
      IO_DruckeStatistikZeile("Count MNF: same Terms", MNF_AnzTermGleich);
      IO_DruckeStatistikZeile("Count MNF: same Terms, same colour", MNF_AnzTermGleichCol);
      IO_DruckeStatistikZeile("Count MNF: same Terms, opp. colour", MNF_AnzTermGleichOpp);
      if (MNF_AnzCalls > 0)
         IO_DruckeStatistikZeile("Average Size of MNF-Set", MNF_SumAnzNachfolger/MNF_AnzCalls);
#endif /* MNF_STATISTIK */
   }
   else if (IS_IsIn("config", IS)) {
#if MNF_STATISTIK      
      fprintf(IO_Strom, "MNF: Statistics ON\n");
#else      
      fprintf(IO_Strom, "MNF: Statistics OFF\n");
#endif /* MNF_STATISTIK */      
   }
   HA_Statistik (IS);
}


/**************************************************************************************/
/* Routinen fuer Kommunikation mit ADT Hash                                           */
/**************************************************************************************/
const unsigned long prim1 = 41698703L;
const unsigned long prim2 = 44278001L;
const unsigned long mult1 = 103L;
const unsigned long mult2 = 97L;

/* MNF_HashCode berechnet einen Fingerprint des Termes term.
   Dieser kann dann als Eingabe fuer die Hashfunktion benutzt werden.
   */
void MNF_HashCode (TermT term, MNF_HashT *hash1E, MNF_HashT *hash2E) {
   MNF_HashT    hash1=0;
   MNF_HashT    hash2=0;
   SymbolT      symbol;

   /* Hashfkt. muss evtl. noch ueberarbeitet werden */
   /* Achtung: Das kleinste Funktionssymbol hat den Wert 0,
      das groesste SO_GroesstesFktSymb.
      Um Eineindeutigkeit der Funktion zu garantieren, muss deswegen
      mit symbol + 1 gerechnet werden
      */
   TO_doSymbole(term, symbol,
                {if (SO_SymbIstFkt (symbol)) {
                   hash1 = (hash1 * mult1 + symbol + 1) % prim1;
                   hash2 = (hash2 * mult2 + symbol + 1) % prim2;
                }
                else{ /* Variablen sind negativ */
                   hash1 = (hash1 * mult1 + SO_GroesstesFktSymb - symbol + 1) % prim1;
                   hash2 = (hash2 * mult2 + SO_GroesstesFktSymb - symbol + 1) % prim2;
                }
                }
                );
   *hash1E = hash1;
   *hash2E = hash2;
}


static BOOLEAN CmpData (MNF_ColTermT *colTerm1, MNF_ColTermT *colTerm2) 
{
   BOOLEAN vgl;
   
   vgl = TO_TermeGleich(colTerm1->term, colTerm2->term);
#if MNF_TEST > 3
   if (!vgl) {
      MNF_HashT    hash1, hash2;
      fprintf(IO_Strom, "\n\a Terme sind unterschiedlich!\a\n");
      IO_DruckeFlex("Term1: %t", colTerm1->term);
      MNF_HashCode (colTerm1->term, &hash1, &hash2);
      fprintf(IO_Strom, "Key1: %lu, Key2: %lu\n", hash1, hash2);
      IO_DruckeFlex("Term1: %t", colTerm2->term);
      MNF_HashCode (colTerm2->term, &hash1, &hash2);
      fprintf(IO_Strom, "Key1: %lu, Key2: %lu\n", hash1, hash2);
      exit(7);
   }
#endif   
   return vgl;
   
   
}


static BOOLEAN AddData (MNF_ColTermT *colTermOld, MNF_ColTermT *colTermNew)
{
   BOOLEAN        retVal = FALSE;

   MNF_IncAnzTermGleich;
   /* Sind Farben gleich? */
   if (MNF_tcolour_eq(*colTermOld, *colTermNew)) {
#if MNF_TEST
      (colTermOld->in)++;
#endif
      if (colTermOld->anzGesSchritte > colTermNew->anzGesSchritte) {
         /* pclID des alten muss erhalten bleiben */
         colTermOld->anzGesSchritte = colTermNew->anzGesSchritte;
         colTermOld->anzGlSchritte = colTermNew->anzGlSchritte;
         colTermOld->anzAntiRSchritte = colTermNew->anzAntiRSchritte;
         colTermOld->pcl_info.parentID = colTermNew->pcl_info.parentID;
#if COMP_PCL
         colTermOld->pcl_info.REpclID = colTermNew->pcl_info.REpclID;
         colTermOld->pcl_info.RESeite = colTermNew->pcl_info.RESeite;
#endif   
         colTermOld->pcl_info.redPosition = colTermNew->pcl_info.redPosition;
         colTermOld->pcl_info.RENummer = colTermNew->pcl_info.RENummer;
         colTermOld->flags.vgl = colTermNew->flags.vgl;
      }
      MNF_IncAnzTermGleichCol;
   }
   else if (MNF_tcolour_opp(*colTermOld, *colTermNew)) {
      MNF_IncAnzTermGleichOpp;
      retVal=TRUE;
   }
   else {
      our_error ("Fuer mehr als zwei Farben fehlt hier noch die Implementierung");
   }
   
   return retVal;
}


static void DeallocData (MNF_ColTermT *colTerm) 
{
     TO_TermLoeschen (colTerm->term);
     SV_Dealloc (MNF_ColTermSpeicher, colTerm);     
}


/**************************************************************************************/
/* Diverse Routinen                                                                   */
/**************************************************************************************/


INLINE static PCL_ObjektNummerT MNF_getID (void) 
{
   return current_pclid;
}

INLINE static void MNF_incID (void) 
{
   current_pclid++;
}

#if COMP_PCL
static void SetPclInfoI(MNF_PCLInfoPtrT pclInfo, TermT redPosition,
                       PCL_ObjektNummerT REpclID, PCL_SeitenT RESeite,
                       MNF_ColTermPtrT parentPtr, long int RENummer)
#else
static void SetPclInfoI(MNF_PCLInfoPtrT pclInfo, TermT redPosition,
                       MNF_ColTermPtrT parentPtr, long int RENummer)
#endif
{
   pclInfo->pclID = MNF_getID ();
   pclInfo->redPosition = redPosition;
#if COMP_PCL
   pclInfo->REpclID = REpclID;
   pclInfo->RESeite = RESeite;
#endif
   pclInfo->RENummer = RENummer;
   if (parentPtr)
      pclInfo->parentID = parentPtr->pcl_info.pclID;
   else {  
      pclInfo->parentID = 0;
   }
}

static MNF_ColTermPtrT CopyColTerm (MNF_ColTermPtrT colTermPtr) 
{
   MNF_ColTermPtrT copyCTermPtr;

   SV_Alloc(copyCTermPtr,MNF_ColTermSpeicher,MNF_ColTermPtrT);
   
   *copyCTermPtr = *colTermPtr;
   copyCTermPtr->term = TO_Termkopie (colTermPtr->term);
   copyCTermPtr->pcl_info.redPosition = NULL; /* Hack, damit nichts passiert */
   return copyCTermPtr;
}

   
static void SetColTerm(/*@out@*/ MNF_ColTermPtrT *colTermPtr, TermT term,
                       int colour,
                       MNF_ColTermPtrT parentPtr,
                       VergleichsT Vergleich,
                       long int RENummer,
                       HashKeyT *Key1,
                       HashKeyT *Key2)
{
   SV_Alloc(*colTermPtr,MNF_ColTermSpeicher,MNF_ColTermPtrT);
   (*colTermPtr)->term=term;
#if MNF_TEST
   (*colTermPtr)->in = 0;
   (*colTermPtr)->out = 0;
   (*colTermPtr)->outUp = 0;
#endif
   (*colTermPtr)->flags.irred = 1;
   (*colTermPtr)->flags.expanded = 0;
   (*colTermPtr)->flags.newElem = 1;
   MNF_setcolour(**colTermPtr, colour);
   MNF_HashCode (term, Key1, Key2);
   if (parentPtr) {
      (*colTermPtr)->anzGesSchritte = parentPtr->anzGesSchritte + 1;
       (*colTermPtr)->flags.vgl = Vergleich;
       if (RENummer < 0) { /* gleichung */
          (*colTermPtr)->anzAntiRSchritte = parentPtr->anzAntiRSchritte;
          switch (Vergleich) {
          case Groesser: /* old > new */
             (*colTermPtr)->anzGlSchritte = parentPtr->anzGlSchritte;
             MNF_setnirred (*parentPtr);
#if MNF_TEST
             parentPtr->out++;
#endif
             break;
          case Kleiner: /* old < new */
             (*colTermPtr)->anzGlSchritte = parentPtr->anzGlSchritte+1;
#if MNF_TEST
             parentPtr->outUp++;
#endif
#if MNF_STATISTIK             
             MNF_MaxEQUp = max (MNF_MaxEQUp, (*colTermPtr)->anzGlSchritte);
#endif /* MNF_STATISTIK */
             break;
          default:
          case Unvergleichbar:
             (*colTermPtr)->anzGlSchritte = parentPtr->anzGlSchritte;
#if MNF_TEST
             parentPtr->outUp++;
#endif
             break;
          }
       } else {
          /* Regel */
          (*colTermPtr)->anzGlSchritte = parentPtr->anzGlSchritte;
          switch (Vergleich) {
          case Groesser: /* old > new */
             (*colTermPtr)->anzAntiRSchritte = parentPtr->anzAntiRSchritte;
             MNF_setnirred (*parentPtr);
#if MNF_TEST
             parentPtr->out++;
#endif
             break;
          case Kleiner: /* old < new */
             (*colTermPtr)->anzAntiRSchritte = parentPtr->anzAntiRSchritte+1;
#if MNF_TEST
             parentPtr->outUp++;
#endif
             break;
          default:
          case Unvergleichbar:
             (*colTermPtr)->anzAntiRSchritte = parentPtr->anzAntiRSchritte;
#if MNF_TEST
             parentPtr->outUp++;
#endif
             break;
          }
       }
   }
   else {
      (*colTermPtr)->anzGesSchritte = 0;
      (*colTermPtr)->flags.vgl = Gleich;
      (*colTermPtr)->anzGlSchritte = 0;
      (*colTermPtr)->anzAntiRSchritte = 0;
   }
}


static void MNF_StatNM (MNF_MengeT *menge MAYBE_UNUSEDPARAM) 
{

#if MNF_STATISTIK
   MNF_SumAnzNachfolger+= menge->size;
   MNF_IncAnzCalls;
   MaximierenLong (&MNF_MaxNachfolger, menge->size);
#endif
}


void MNF_ClearNM (MNF_MengeT *menge)
{
   MNF_StatNM (menge);
   
   HA_Clear(&menge->HashTabelle);

   menge->MNFPQ_Clear (menge->PQRed);
   menge->MNFPQ_Clear (menge->PQGreen);
   if (menge->deallocJoinedCTerm1)
      DeallocData (menge->joinedCTerm1);
   
   if (menge->deallocJoinedCTerm2)
      DeallocData (menge->joinedCTerm2);
   menge->deallocJoinedCTerm1 = FALSE;
   menge->deallocJoinedCTerm2 = FALSE;
   menge->joinable = FALSE;
   menge->size = 0;
} 


/* Testen, ob TermUni mit einem Term der Farbe MitColour unifizierbar ist */
static BOOLEAN MNF_Unifiziert (MNF_MengeT *menge, TermT TermUni, unsigned short MitColour) 
{
   MNF_ColTermPtrT       colTermPtr1 = NULL;
   HashEntryPtrT         TraversePtr1;

   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr1);
   while (!HA_EndTraverseN (&TraversePtr1)) {
      HA_NextTraverseN (&TraversePtr1, (void **)&colTermPtr1);

      if (MNF_colour_opp (MNF_tcolour(*colTermPtr1), MitColour)) {
         if (U1_Unifiziert (colTermPtr1->term, TermUni)) {
            /* TODO */
            fprintf (IO_Strom, "\n\n\a Unifiziert!!!!!!!!!!!!!!!!!!!!!!\n\n");
            menge->joinedCTerm1 = colTermPtr1;
            menge->joinedCTerm2 = NULL;
            menge->deallocJoinedCTerm1 = FALSE;
            menge->deallocJoinedCTerm2 = FALSE;
            menge->joinable = TRUE;
            return TRUE;
         }
      }
   }
   return FALSE;
}



/* liefert TRUE, wenn Elemente zusammengefuehrt werden konnten */
static BOOLEAN
MNF_BearbeiteInsertElem (MNF_ColTermPtrT parentCTerm, long int RENummer,
                         TermT newTerm,
                         void *stack, MNF_MengeT *menge, int colour,
                         VergleichsT Vergleich, TermT redPosition
#if COMP_PCL
                         ,
                         PCL_ObjektNummerT REpclID, PCL_SeitenT RESeite
#endif
                         )
{
   MNF_ColTermT         *oldCData = NULL;
   MNF_ColTermT         *newCTerm;
   BOOLEAN              Joined = FALSE;
   HashKeyT             Key1, Key2;

   SetColTerm (&newCTerm, newTerm, colour, parentCTerm, Vergleich, RENummer, &Key1, &Key2);
   SetPclInfo (&(newCTerm->pcl_info), redPosition,
               REpclID, RESeite, parentCTerm, RENummer);
   HA_AddData(&(menge->HashTabelle), Key1, Key2, newCTerm, (void **)&oldCData);
   MNF_IncAnzCallsHAAddData;
   if (oldCData != NULL) {
      /* Datensatz existiert schon: Also beide Saetze mergen und Speicher freigeben */
      Joined = AddData (oldCData, newCTerm);
      if (Joined) {
         menge->joinedCTerm1 = oldCData;
         menge->joinedCTerm2 = newCTerm;
         menge->deallocJoinedCTerm1 = FALSE;
         menge->deallocJoinedCTerm2 = TRUE;
         menge->joinable = TRUE;
      }
      else
         DeallocData (newCTerm);
   } else {
      MNF_incID();
      menge->size++;
      menge->MNFPQ_Push(stack, newCTerm);
   }
   return Joined;
}

static BOOLEAN
MNF_AddVarianten (MNF_ColTermPtrT acctCTerm, long int RENummer,
                  TermT newTerm,
                  void *stack, MNF_MengeT *menge, int colour,
                  TermT redPosition
#if COMP_PCL
                  ,
                  PCL_ObjektNummerT REpclID, PCL_SeitenT RESeite
#endif
                  )
{
   SymbolT sym;
   BOOLEAN ret;

   /* Term ist kein Grundterm */
   SO_forFktSymbole(sym) {
      if (SO_IstSkolemSymbol(sym)) {
         UTermT teilTerm;
         TermT term = TO_Termkopie (newTerm);
         IO_DruckeFlex ("%t\n", term);
         TO_doStellen (term, teilTerm,
                       {
                          if (SO_SymbIstVar (TO_TopSymbol(teilTerm)))
                             TO_SymbolBesetzen(teilTerm, sym);
                       });
         ret = MNF_BearbeiteInsertElem(acctCTerm, RENummer,
                                       term, 
                                       stack, menge, colour, Groesser, redPosition
#if COMP_PCL
                                       ,REpclID, RESeite
#endif
                                       );
         if (ret)
            return TRUE;
      }
   }
   return FALSE;
}

static INLINE void
MNF_VMMengeVonTerm (TermT term, VMengenT *VMenge) 
{
   SymbolT      symbol;

   VM_NeueMengeSeiLeer (*VMenge);
   TO_doSymbole(term, symbol,
                {if (SO_SymbIstVar (symbol))
                   VM_ElementEinfuegen(VMenge,symbol);
                }
                );
}


/* Testet, ob die Variablen von Term1 eine Teilmenge der Variablen
   von Term2 sind */
static BOOLEAN
MNF_VMTeilmengenP (TermT Term1, TermT Term2) 
{
   VMengenT       Menge1, Menge2;
   BOOLEAN        teilmenge;

   if (count_vars (Term1) != 0) {
      MNF_VMMengeVonTerm (Term1, &Menge1);
      MNF_VMMengeVonTerm (Term2, &Menge2);
      teilmenge = VM_Teilmenge (&Menge1, &Menge2);
      VM_MengeLoeschen (&Menge1);
      VM_MengeLoeschen (&Menge2);
      return teilmenge;
   }
   else
      return TRUE;
}


static BOOLEAN
MNF_BearbeiteInsertElemE (MNF_ColTermPtrT acctCTerm, long int RENummer,
                          TermT newTerm,
                          void *stack, MNF_MengeT *menge, int colour,
                          TermT redPosition
#if COMP_PCL
                          ,
                          PCL_ObjektNummerT REpclID, PCL_SeitenT RESeite
#endif
                          )
{
   VergleichsT    Resultat;

   if (!MNF_VMTeilmengenP (newTerm, acctCTerm->term)) {
      TO_TermLoeschen (newTerm);
      return FALSE;
   }
   
   Resultat = ORD_VergleicheTerme(acctCTerm->term,newTerm);  /* Umschaltung auf Ordnungszeitmessung in ORD_> */

   

   if ((Resultat == Kleiner) && (acctCTerm->anzGlSchritte < menge->maxGlSchritte)) {
      MNF_IncAnzEQUp;
      return (MNF_BearbeiteInsertElem (acctCTerm, RENummer, newTerm,
                                       stack, menge, colour, Resultat, redPosition
#if COMP_PCL
                                       , REpclID, RESeite
#endif
                                       ));
   }
   else if ((Resultat == Groesser)) {
      return (MNF_BearbeiteInsertElem (acctCTerm, RENummer, newTerm,
                                       stack, menge, colour, Resultat, redPosition
#if COMP_PCL

                                       , REpclID, RESeite
#endif
                                       ));
   }
   
   TO_TermLoeschen (newTerm);
   return FALSE;
}


/* Bearbeite holt ein Element von "stack" und bildet alle direkten
   Nachfolger. Die Nachfolger werden in den Stack "stack" und in
   die HashTabelle eingefuegt.
   MNF_Bearbeite endet, wenn alle Nachfolger gebildet oder
   wenn die Elemente zusammengefuehrt werden konnten
   */
static void MNF_BearbeiteDummy  (void *pq MAYBE_UNUSEDPARAM,
                                 MNF_MengeT *menge MAYBE_UNUSEDPARAM,
                                 int colour MAYBE_UNUSEDPARAM,
                                 MNF_ColTermPtrT acctCTerm MAYBE_UNUSEDPARAM)
{
   ;
}


static void MNF_BearbeiteE(void *stack, MNF_MengeT *menge, int colour, MNF_ColTermPtrT acctCTerm)
{
   TermT                Kopie;
   long int             RENummer;
   PCL_ObjektNummerT    REpclID;
   PCL_SeitenT          RESeite;
   
   MNF_setexpanded(*acctCTerm);
   AdHocPfadNeu(acctCTerm->term, KPsGen);
   do {
      if (TO_TermIstNichtVar(aktTeiltermAH)) {
         MO_AllGleichungMatchInit(aktTeiltermAH);
         MNF_IncAnzMatchVersucheE;
         while (MO_AllGleichungMatchNext (&RENummer, &REpclID, &RESeite)) {
            MNF_IncAnzReduktionen;
            Kopie=MO_AllTermErsetztNeu(acctCTerm->term, aktTeiltermAH);
            /* Kopie muss freigegeben werden */
            if (MNF_BearbeiteInsertElemE (acctCTerm, RENummer, Kopie,
                                          stack, menge, colour, aktTeiltermAH
#if COMP_PCL
                                          ,
                                          REpclID, RESeite
#endif
                                          ))
               return;
            MNF_IncAnzMatchVersucheE;
         }
      }
   } while (AdHocPfadWeitersetzen());
}


static void MNF_BearbeiteR (void *stack, MNF_MengeT *menge, int colour, MNF_ColTermPtrT acctCTerm)
{
   TermT                Kopie;
   long int             RENummer;
   PCL_ObjektNummerT    REpclID;
   
   MNF_setexpanded(*acctCTerm);
   AdHocPfadNeu(acctCTerm->term, KPsGen);
   do {
      if (TO_TermIstNichtVar(aktTeiltermAH)) {
         MO_AllRegelMatchInit(aktTeiltermAH);
         MNF_IncAnzMatchVersucheR;
         while (MO_AllRegelMatchNext (&RENummer, &REpclID)) {
            MNF_IncAnzReduktionen;
            Kopie=MO_AllTermErsetztNeu(acctCTerm->term, aktTeiltermAH);
            /* Kopie muss freigegeben werden! */
            if (MNF_BearbeiteInsertElem (acctCTerm, RENummer, Kopie,
                                         stack, menge, colour, Groesser, aktTeiltermAH
#if COMP_PCL
                                         ,REpclID, linkeSeite
#endif
                                         ))
               return;
            MNF_IncAnzMatchVersucheR;
         }
      }
   } while (AdHocPfadWeitersetzen());
}





static void MNF_BearbeiteRE (void *stack, MNF_MengeT *menge, int colour, MNF_ColTermPtrT acctCTerm) 
{
   
   MNF_BearbeiteR (stack, menge, colour, acctCTerm);
   if (!menge->joinable) {
      MNF_BearbeiteE (stack, menge, colour, acctCTerm);
   }
}


/* versucht eine Regel auf den Term anzuwenden */
static void MNF_BearbeiteEineR (void *stack, MNF_MengeT *menge, int colour,
                               MNF_ColTermPtrT acctCTerm, RegelT Regel)
{
   static int zaehler = 0;
   
   TermT                newTerm;
   PCL_ObjektNummerT    REpclID;
   TermT                antiRechts, antiLinks;
   AntiRegelUseT        useAnti = ARnoUse;

   if (acctCTerm->anzAntiRSchritte < menge->maxAntiRSchritte)
      MNF_ConvertRegelToAntiRegel (menge->antiKind,
                                   Regel,
                                   &antiLinks, &antiRechts, &useAnti);

   REpclID = TP_PclIDAs (Regel);

   AdHocPfadNeu(acctCTerm->term, KPsGen);
   do {
      if (TO_TermIstNichtVar(aktTeiltermAH)) {
         MNF_IncAnzMatchVersucheR;
         if (MO_TermpaarPasstMitResultat(Regel->linkeSeite, Regel->rechteSeite,
                                         acctCTerm->term, &newTerm)) {
            MNF_IncAnzReduktionen;
            /* newTerm muss freigegeben werden */
            if (MNF_BearbeiteInsertElem(acctCTerm, Regel->Nummer,
                                        newTerm,
                                        stack, menge, colour, Groesser, aktTeiltermAH
#if COMP_PCL
                                        ,REpclID, linkeSeite
#endif
                                        )) {
               if (useAnti != ARnoUse) {
                  TO_TermLoeschen (antiLinks);
                  TO_TermLoeschen (antiRechts);
               }
               return;
            }
            
         }
         switch (useAnti) {
         case ARuseAnti:
            if ((acctCTerm->anzAntiRSchritte < menge->maxAntiRSchritte)
                && MO_TermpaarPasstMitResultat(antiLinks, antiRechts,
                                               acctCTerm->term, &newTerm)) {
               MNF_IncAnzReduktionen;
               /* newTerm muss freigegeben werden */
               /* Aufruf stimmt (?) noch nicht, da hier gegen Ordnung gearbeitet wird */
               MNF_IncAnzAntiR;
               if (MNF_BearbeiteInsertElem(acctCTerm, Regel->Nummer,
                                           newTerm, 
                                           stack, menge, colour, Kleiner, aktTeiltermAH
#if COMP_PCL
                                           ,REpclID, rechteSeite
#endif
                                           )) {
                  TO_TermLoeschen (antiLinks);
                  TO_TermLoeschen (antiRechts);
                  return;
               }
            }
            break;
         case ARuseUnifikation:
            if (MO_TermpaarPasstMitResultat(antiLinks, antiRechts, acctCTerm->term, &newTerm)) {
               if (zaehler++ < 2 &&
                   MNF_AddVarianten (acctCTerm, Regel->Nummer, newTerm,
                                     stack, menge, colour, aktTeiltermAH
#if COMP_PCL
                                     ,REpclID, rechteSeite
#endif
                                     ))
                  break;
               if (MNF_Unifiziert (menge, newTerm, MNF_tcolour(*acctCTerm))) {
                  TO_TermLoeschen (newTerm);
                  TO_TermLoeschen (antiLinks);
                  TO_TermLoeschen (antiRechts);
                  return;
               }
               TO_TermLoeschen (newTerm);
            }
            break;
         case ARnoUse:
         default:
            break;
         }
      }
   } while (AdHocPfadWeitersetzen());
   if (useAnti != ARnoUse) {
      TO_TermLoeschen (antiLinks);
      TO_TermLoeschen (antiRechts);
   }
}


/* versucht eine Gleichung auf den Term anzuwenden */
static void MNF_BearbeiteEineGl (void *stack, MNF_MengeT *menge, int colour,
                               MNF_ColTermPtrT acctCTerm, GleichungsT Gleichung)
{
   TermT                newTerm;
   PCL_ObjektNummerT    REpclID;

   REpclID = TP_PclIDAs (Gleichung);
   
   AdHocPfadNeu(acctCTerm->term, KPsGen);
   do {
      if (TO_TermIstNichtVar(aktTeiltermAH)) {
         MNF_IncAnzMatchVersucheE;
         if (MO_TermpaarPasstMitResultat(Gleichung->linkeSeite, Gleichung->rechteSeite,
                                         acctCTerm->term, &newTerm)) {
            /* newTerm muss freigegeben werden */
            MNF_IncAnzReduktionen;
            if (MNF_BearbeiteInsertElemE (acctCTerm, Gleichung->Nummer, newTerm,
                                          stack, menge, colour, aktTeiltermAH
#if COMP_PCL
                                          ,REpclID, linkeSeite
#endif
                                          ))
               return;
         }
         /* jetzt links und rechts Vertauschen und das ganze Nochmal */
         /* Gleichung->Gegenrichtung->linkeSeite, rechteSeite */
         else if (TP_IstKeineMonogleichung(Gleichung) &&
                  MO_TermpaarPasstMitResultat(Gleichung->Gegenrichtung->linkeSeite,
                                              Gleichung->Gegenrichtung->rechteSeite,
                                              acctCTerm->term, &newTerm)) {
            MNF_IncAnzReduktionen;
            if (MNF_BearbeiteInsertElemE (acctCTerm, Gleichung->Nummer, newTerm,
                                          stack, menge, colour, aktTeiltermAH
#if COMP_PCL
                                          ,REpclID, rechteSeite
#endif
                                          ))
               return;
         }
      }
   } while (AdHocPfadWeitersetzen());
}


static BOOLEAN MNF_AnalyseNM (MNF_MengeT *menge) {

   if (menge->Expand != ExpandRewrite) {
      if (menge->size > menge->SwitchNF) {
         menge->Expand = ExpandRewrite;
         MNF_SearchInteresting (menge);
         return TRUE;
      }
      else if (menge->size > menge->SwitchIrred) {
         menge->MNF_AddTryAllElements = FALSE;
      }
   }
   return FALSE;
}


/* Expandieren der Nachfolgermenge */
void MNF_ExpandNM (MNF_MengeT *menge) {
   MNF_ColTermPtrT   acctCTerm;
   BOOLEAN           lastRedWasIrred = FALSE;
   BOOLEAN           lastGreenWasIrred = FALSE;

   for(;;) {
      if (!(menge->MNFPQ_Empty)(menge->PQRed)) {
         acctCTerm = (MNF_ColTermPtrT)(menge->MNFPQ_Pop) (menge->PQRed, lastRedWasIrred);
         menge->BearbeiteFunc(menge->PQRed, menge, MNF_red, acctCTerm);
         lastRedWasIrred = MNF_tirred (*acctCTerm);
         if (menge->joinable)
            break;
      }
      if (!(menge->MNFPQ_Empty)(menge->PQGreen)) {
         acctCTerm = (MNF_ColTermPtrT)(menge->MNFPQ_Pop)(menge->PQGreen, lastGreenWasIrred);
         menge->BearbeiteFunc(menge->PQGreen, menge, MNF_green, acctCTerm);
         lastGreenWasIrred = MNF_tirred (*acctCTerm);
         if (menge->joinable)
            break;
      }
      if ((menge->maxSize != 0) && (menge->size > menge->maxSize)) {
         MNF_IncAbbruchMaxSize;
         break;
      }
      if (menge->useAnalyse && MNF_AnalyseNM (menge))
         break;
      if (menge->MNFPQ_Empty(menge->PQRed) && menge->MNFPQ_Empty(menge->PQGreen))
         break;
      if (SI_InterruptflagGesetzt()) ST_IRMenue("MNF generation");
   }
}


/**************************************************************************************/
/* interne Abfrageroutinen                                                            */
/**************************************************************************************/

INLINE static
BOOLEAN MNF_JoinedNM (MNF_MengeT *menge) {
   return (menge->joinable);
}


/**************************************************************************************/
/* exportierte Abfrageroutinen                                                        */
/**************************************************************************************/

BOOLEAN MNF_JoinedNMenge (MNF_MengeT *menge) 
{
   return menge->joinable;
}

   
/* Der zurueckgegebene Term darf von der aufgerufenen Funktion
   nicht freigegeben werden. Evtl. muss eine Termkopie gemacht werden */
TermT MNF_GetJoinedTerm (MNF_MengeT *menge)
{
   return menge->joinedCTerm1->term;
}

/* GetJoinedTermpair liefert das zusammengefuehrte bzw. subsummierte Termpaar
   zurueck */
/* Die zurueckgegebenen Terme duerfen von der aufgerufenen Funktion
   nicht freigegeben werden. Evtl. muss eine Termkopie gemacht werden */
void MNF_GetJoinedTermpair (MNF_MengeT *menge, TermT *left, TermT *right) 
{
   MNF_ColTermPtrT       leftPtr, rightPtr;

   /* joinedCTerm1 existiert immer, joinedCTerm2 nicht unbedingt */
   if (menge->joinedCTerm2 == NULL) {
      if (MNF_tcolour_red (*menge->joinedCTerm1)) {
         rightPtr = menge->joinedCTerm1;
         leftPtr = menge->initialCTermLeft;
      } else {
         leftPtr = menge->joinedCTerm1;
         rightPtr = menge->initialCTermRight;
      }
   } else {
      if (MNF_tcolour_red (*menge->joinedCTerm1)) {
         rightPtr = menge->joinedCTerm1;
         leftPtr = menge->joinedCTerm2;
      } else {
         rightPtr = menge->joinedCTerm2;
         leftPtr = menge->joinedCTerm1;
      }
   }
   
   *left = leftPtr->term;
   *right = rightPtr->term;
}


unsigned int MNF_GetSize (MNF_MengeT *menge) 
{
   return menge->size;
}


/**************************************************************************************/
/* Hilfsroutinen                                                                      */
/**************************************************************************************/

void MNF_InitFirstMenge (MNF_MengeT *menge) 
{
   menge->PQRed = NULL;
   menge->PQGreen = NULL;
   menge->joinable = FALSE;
   menge->useR = TRUE;
   menge->useE = FALSE;
   menge->maxGlSchritte = 0;
   menge->maxAntiRSchritte = 0;
   menge->antiKind = noAnti;
   menge->maxSize = 1000;
   menge->size = 0;
   menge->BearbeiteFunc = MNF_BearbeiteDummy;
   menge->MNFPQ_Art = MNFPQ_stack;
   menge->MNF_UseClassify = FALSE;
   menge->MNF_ClassInit = NULL;
   menge->MNF_ClassTrav = NULL;
   menge->MNF_ClassExit = NULL;
   menge->MNF_AddTryAllElements = TRUE;
   menge->initialHashSize = 4095;
   menge->initialCTermLeft = NULL;
   menge->initialCTermRight = NULL;
}


void MNF_MengeSetREStatus (MNF_MengeT *menge, BOOLEAN RLeer, BOOLEAN ELeer)
{
   if (menge->useR && menge->useE) {
      if (RLeer && ELeer)
         menge->BearbeiteFunc = MNF_BearbeiteDummy;
      else if (ELeer)
         menge->BearbeiteFunc = MNF_BearbeiteR;
      else if (RLeer)
         menge->BearbeiteFunc = MNF_BearbeiteE;
      else 
         menge->BearbeiteFunc = MNF_BearbeiteRE;
   }
   else if (menge->useE) {
      if (ELeer)
         menge->BearbeiteFunc = MNF_BearbeiteDummy;
      else menge->BearbeiteFunc = MNF_BearbeiteE;
   }
   else if (menge->useR) {
      if (RLeer)
         menge->BearbeiteFunc = MNF_BearbeiteDummy;
      else menge->BearbeiteFunc = MNF_BearbeiteR;
   }
   else our_error ("MNF: R oder E muss angegeben werden!");
}


void MNF_InitMenge (MNF_MengeT *menge, HashSizeT HashSize) 
{

   MNF_MengeSetREStatus (menge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());
   
   HA_New (&(menge->HashTabelle),  (HA_CmpFuncT) &CmpData, (HA_DeallocFuncT) &DeallocData,
           HashSize);

   MNFPQ_InitMenge (menge);
   
   menge->PQRed  = menge->MNFPQ_GetNewPQ();
   menge->PQGreen= menge->MNFPQ_GetNewPQ();

}


void MNF_PutFirstElementsInMenge (MNF_MengeT *menge, TermT links, TermT rechts) 
{
   /* Achtung: Speicher freigeben fuer sk,tk, wenn diese direkt zusammenfuehrbar sind */
   TermT        sk = TO_Termkopie(links),
                tk = TO_Termkopie(rechts);
   MNF_ColTermT *colSk, *colTk;
   MNF_ColTermPtrT oldDataPtr = NULL;
   HashKeyT     Key1T, Key2T, Key1S, Key2S;

   /* sk:linker Term, GRUEN gefaerbt */
   SetColTerm (&colSk, sk, MNF_green, NULL, Gleich, 0, &Key1S, &Key2S);
   SetPclInfo (&(colSk->pcl_info), NULL, 0, linkeSeite, NULL, 0);
   MNF_incID();
   MNF_setold (*colSk);
   menge->MNFPQ_Push(menge->PQGreen, colSk);
   HA_AddData(&(menge->HashTabelle), Key1S, Key2S, colSk, (void **) &oldDataPtr);
   MNF_IncAnzCallsHAAddData;
   menge->initialCTermLeft = colSk;
   

   /* tk: rechter Term, ROT gefaerbt */
   SetColTerm (&colTk, tk, MNF_red, NULL, Gleich, 0, &Key1T, &Key2T);
   SetPclInfo (&(colTk->pcl_info), NULL, 0, rechteSeite, NULL, 0);
   MNF_setold (*colTk);
   
   HA_AddData(&(menge->HashTabelle), Key1T, Key2T, colTk, (void **) &oldDataPtr);
   MNF_IncAnzCallsHAAddData;
   if (oldDataPtr == NULL) {
      menge->MNFPQ_Push (menge->PQRed, colTk);
      MNF_incID();
      /* initialCTermRight wird unten gesetzt */
      menge->initialCTermRight = colTk;
      menge->joinable = FALSE;
      menge->size=2;
      menge->joinedCTerm1 = NULL;
      menge->joinedCTerm2 = NULL;
      menge->deallocJoinedCTerm1 = FALSE;
      menge->deallocJoinedCTerm2 = FALSE;
      /* evtl. auf Rewrite umstellen */
      if (menge->useAnalyse)
	(void) MNF_AnalyseNM (menge);
   } else { /* sk, tk sind gleich */
      menge->initialCTermRight = NULL;
      menge->joinable = TRUE;
      menge->joinedCTerm1 = colSk;
      menge->joinedCTerm2 = colTk;
      /* colSk ist im Hash drin, colTk nicht. Deswegen muss colTk separat
         freigegeben werden */
      menge->deallocJoinedCTerm1 = FALSE;
      menge->deallocJoinedCTerm2 = TRUE;
      menge->size=1;
      /* Normalerweise wuerde hier MNF_AddData aufgerufen,
         aber da die Situation klar ist (opposite colour),
         muss hier nur die Statistik korrigiert werden */
      MNF_IncAnzDirektZusammenfuehrbar;
      MNF_IncAnzTermGleichOpp;
      MNF_IncAnzTermGleich;
   }
}


void MNF_ParseInfoStrings (InfoStringT DEF, InfoStringT USR,
                           BOOLEAN *useE, BOOLEAN *useR, MNFPQ_ArtT *MNFPQ_Art,
                           unsigned int *maxSize, unsigned int *maxGlSchritte,
                           BOOLEAN *TryAllElements, 
                           unsigned int *maxAntiRSchritte,
                           MNF_AntiT *antiKind,
                           BOOLEAN *useAnalyse, unsigned int *SwitchIrred,
                           unsigned int *SwitchNF, unsigned int *CountNF,
                           HashSizeT *HashSize)
{
   char          *valueStr, *kopie, *kopieDummy, *token;
   unsigned long tokenValue;
   
   *useE=TRUE;
   *useR=TRUE;
   *MNFPQ_Art = MNFPQ_stack;
   *maxSize = 0;
   *maxGlSchritte = 0;
   *TryAllElements=TRUE;
   *maxAntiRSchritte = 0;
   *antiKind = noAnti;
   *useAnalyse = FALSE;
   *SwitchIrred = 0;
   *SwitchNF=0;
   *CountNF=0;
   *HashSize=1023;

   valueStr = IS_ValueOfOption (DEF, USR, "mnf");
   /* valueStr ist leer, defaults uebernehmen */
   if (!valueStr)
     return;

   kopie = (char *) our_alloc(sizeof(char)*(strlen(valueStr)+1));
   strcpy (kopie, valueStr);
   kopieDummy = kopie;

   token = strtok (kopie, ".");
   while (token != NULL) {
      tokenValue = strtoul (token, NULL, 10);
      if ((strcmp (token, "re") == 0) || (strcmp(token, "er") == 0)) {
         *useR=TRUE;
         *useE=TRUE;
      }
      else if (strcmp(token, "e") == 0) {
         *useR=FALSE;
         *useE=TRUE;
      }
      else if (strcmp(token, "r") == 0) {
         *useR=TRUE;
         *useE=FALSE;
      }
      else if (strcmp(token, "stack") == 0) {
         *MNFPQ_Art = MNFPQ_stack;
      }
      else if (strcmp(token, "deque") == 0) {
         *MNFPQ_Art = MNFPQ_deque;
      }
      else if (strcmp(token, "fifo") == 0) {
         *MNFPQ_Art = MNFPQ_fifo;
      }
      else if (tokenValue > 0) {
         *maxSize = tokenValue;
      }
      token=strtok (NULL, ".");
   }

   our_dealloc (kopieDummy);

   if (IS_FindValuedOption (DEF, USR, 1, "mnfup") == 0) {
      valueStr = IS_ValueOfOption (DEF, USR, "mnfup");
      if (valueStr) {
         tokenValue = strtoul (valueStr, NULL, 10);
         *maxGlSchritte = tokenValue;
      }
   }
   
   if (IS_FindValuedOption (DEF, USR, 1, "mnfantir") == 0) {
      valueStr = IS_ValueOfOption (DEF, USR, "mnfantir");
      if (valueStr) {
         tokenValue = strtoul (valueStr, NULL, 10);
         *maxAntiRSchritte = tokenValue;
         *antiKind = antiWOVar;
      }
   }
   if (IS_FindValuedOption (DEF, USR, 1, "mnfanalysis") == 0) {
      valueStr = IS_ValueOfOption (DEF, USR, "mnfanalysis");
      *useAnalyse = TRUE;
      kopie = (char *) our_alloc(sizeof(char)*(strlen(valueStr)+1));
      strcpy (kopie, valueStr);
      kopieDummy = kopie;
      token = strtok (kopie, ".");
      if (token != NULL) {
         *SwitchIrred = strtoul (token, NULL, 10);
         token = strtok (NULL, ".");
      }
      if (token != NULL) {
         *SwitchNF = strtoul (token, NULL, 10);
         token = strtok (NULL, ".");
      }
      if (token != NULL) {
         *CountNF = strtoul (token, NULL, 10);
      }
      our_dealloc (kopieDummy);
   }
   if (IS_FindOption (DEF,USR, 1, "mnftrynotall") == 0) {
      *TryAllElements = FALSE;
   }

   if (IS_FindOption (DEF,USR, 1, "mnfantiWOVar") == 0) {
      *antiKind = antiWOVar;
   }
   if (IS_FindOption (DEF,USR, 1, "mnfantiBoundVar") == 0) {
      *antiKind = antiBoundVar;
   }
   if (IS_FindOption (DEF,USR, 1, "mnfantiFreeVar") == 0) {
      *antiKind = antiFreeVar;
   }
   if (IS_FindOption (DEF, USR, 1, "stack") == 0) {
      *MNFPQ_Art = MNFPQ_stack;
   } 
   if (IS_FindOption (DEF, USR, 1, "deque") == 0) {
      *MNFPQ_Art = MNFPQ_deque;
   }
   if (IS_FindOption (DEF, USR, 1, "fifo") == 0) {
      *MNFPQ_Art = MNFPQ_fifo;
   }

   if (IS_FindValuedOption (DEF, USR, 1, "hash") == 0) {
      valueStr = IS_ValueOfOption (DEF, USR, "hash");
      if (valueStr) {
         tokenValue = strtoul (valueStr, NULL, 10);
         *HashSize = tokenValue;
      }
   }
}



void MNF_ParseInfoStringsMenge (InfoStringT DEF, InfoStringT USR, MNF_MengeT *menge)
{
   MNF_ParseInfoStrings (DEF, USR, &menge->useE, &menge->useR,
                         &menge->MNFPQ_Art, &menge->maxSize, &menge->maxGlSchritte,
                         &menge->MNF_AddTryAllElements,
                         &menge->maxAntiRSchritte,
                         &menge->antiKind,
                         &menge->useAnalyse, &menge->SwitchIrred,
                         &menge->SwitchNF, &menge->CountNF,
                         &menge->initialHashSize);
}


void MNF_PrintOptions (BOOLEAN useE, BOOLEAN useR, MNFPQ_ArtT MNFPQ_Art,
                       unsigned int maxSize, unsigned int maxGlSchritte,
                       BOOLEAN AddTryAllElements, 
                       unsigned int maxAntiRSchritte,
                       MNF_AntiT antiKind,
                       BOOLEAN useAnalyse, unsigned int SwitchIrred,
                       unsigned int SwitchNF, unsigned int CountNF,
                       HashSizeT initialHashSize)
{
   fprintf (IO_Strom, "    MNF: ");
   if (useE)
      fprintf (IO_Strom, "Equations, ");
   if (useR)
      fprintf (IO_Strom, "Rules, ");
   if (maxSize != 0)
      fprintf (IO_Strom, "MaxSize=%u, ", maxSize);
   if (maxGlSchritte != 0)
      fprintf (IO_Strom, "Max Up=%u, ", maxGlSchritte);
   if (maxAntiRSchritte != 0) {
      fprintf (IO_Strom, "Try Rules Up=%u ", maxAntiRSchritte);
      switch (antiKind) {
      case antiWOVar:
         fprintf (IO_Strom, " (rules without vars)");
         break;
      case antiBoundVar:
         fprintf (IO_Strom, " (rules without new vars on lhs)");
         break;
      case antiFreeVar:
         fprintf (IO_Strom, " (all rules)");
         break;
      default:
         break;
      }
      fprintf (IO_Strom, ", ");
   }
   if (useAnalyse)
      fprintf (IO_Strom, "Analysis: %u:%u:%u, ",
               SwitchIrred, SwitchNF, CountNF);
   if (!AddTryAllElements)
      fprintf (IO_Strom, "Expand only irreduc. Elements");
   switch (MNFPQ_Art) {
   case MNFPQ_stack:
      fprintf (IO_Strom, "PQ: Stack");
      break;
   case MNFPQ_deque:
      fprintf (IO_Strom, "PQ: Deque");
      break;
   case MNFPQ_fifo:
      fprintf (IO_Strom, "PQ: Fifo");
      break;
   }
   fprintf (IO_Strom, ", initial hash size: %d", initialHashSize);
   fprintf (IO_Strom, "\n");
}


void MNF_PrintOptionsMenge (MNF_MengeT *menge)
{
   MNF_PrintOptions (menge->useE, menge->useR, menge->MNFPQ_Art, menge->maxSize,
                     menge->maxGlSchritte, menge->MNF_AddTryAllElements,
                     menge->maxAntiRSchritte,
                     menge->antiKind,
                     menge->useAnalyse,
                     menge->SwitchIrred, menge->SwitchNF,
                     menge->CountNF, menge->initialHashSize);
}


void MNF_ParamInfo (ParamInfoArtT what, InfoStringT DEF, InfoStringT USR) 
{
   BOOLEAN             useE;
   BOOLEAN             useR;
   BOOLEAN             tryAllElements;
   BOOLEAN             useAnalyse;
   unsigned int        SwitchIrred;
   unsigned int        SwitchNF;
   unsigned int        CountNF;
   MNFPQ_ArtT          MNFPQ_Art;
   unsigned int        maxSize;
   unsigned int        maxGlSchritte;
   unsigned int        maxAntiRSchritte;
   MNF_AntiT           antiKind;
   HashSizeT           initialHashSize;
   
   switch (what) {
   case ParamEvaluate:
      MNF_ParseInfoStrings (DEF, USR, &useE, &useR, &MNFPQ_Art,
                            &maxSize, &maxGlSchritte, &tryAllElements,
                            &maxAntiRSchritte,
                            &antiKind,
                            &useAnalyse,
                            &SwitchIrred, &SwitchNF, &CountNF,
                            &initialHashSize);
      MNF_PrintOptions (useE, useR, MNFPQ_Art, maxSize, maxGlSchritte,
                        tryAllElements, maxAntiRSchritte,
                        antiKind,
                        useAnalyse,
                        SwitchIrred, SwitchNF, CountNF,
                        initialHashSize);
      break;
   case ParamExplain:
      fprintf (IO_Strom, "Multiple Normalforms\n");
      fprintf (IO_Strom, "********************\n");
      fprintf (IO_Strom, "To select: mnf=[r|e|re].<n>[:additional paramameter]*\n");
      fprintf (IO_Strom, "  <n>: Maximal size of MNF set\n");
      fprintf (IO_Strom, "  e  : use only equations\n");
      fprintf (IO_Strom, "  r  : use only rules\n");
      fprintf (IO_Strom, "  re : use rules and equations (default)\n");
      fprintf (IO_Strom, " additional parameter:\n");
      fprintf (IO_Strom, "  [stack|deque|fifo] : priority queue (stack is default)\n");
      fprintf (IO_Strom, "  [hash=<n>]     : initial hash size (default 1023)\n");
      fprintf (IO_Strom, "  [mnfup=<n>]    : max. number of steps up with equations (default 0)\n");
      fprintf (IO_Strom, "  [mnfantir=<n>] : max. number of steps up with rules (default 0)\n");
      fprintf (IO_Strom, "  [mnfantiWOVar|mnfantiBoundVar|mnfantiFreeVar] :\n");
      fprintf (IO_Strom, "   class of rules to try steps up with\n");
      fprintf (IO_Strom, "  [mnftrynotall] : expand only new and irreducible terms\n");
      fprintf (IO_Strom, "  [mnfanalysis=<n1>.<n2>.<n3>] : parameter for analysis to limit size of set,\n");
      fprintf (IO_Strom, "   n1: no. of elements for switching to expand only new and irreducible terms\n");
      fprintf (IO_Strom, "   n2: no. of elements for switching to reduce with normal strategy\n");
      fprintf (IO_Strom, "   n3: no. of elements to reduce with normal strategy for each side\n");
      break;
   }
}


BOOLEAN MNF_DurchlaufPaarweise (MNF_MengeT *menge, MNF_DurchlaufPaarProcT DurchlaufProc)
{
   MNF_ColTermPtrT       colTermPtr1 = NULL;
   MNF_ColTermPtrT       colTermPtr2 = NULL;
   HashEntryPtrT         TraversePtr1, TraversePtr2;
   
   if (menge->Expand == ExpandRewrite)
      return MNF_PseudoDurchlaufPaarweise (menge, DurchlaufProc);
   
   /* Wenn mit Deckelfkt. gearbeitet wird, dann sind einige Elemente noch nicht expandiert.
      Im Extrem-Fall kann es passieren, dass noch keine Normalform unter den Elementen ist.
      Dieser Fall wird im folgenden nicht behandelt.
      Die DurchlaufProc wird fuer alle Paare aus irreduziblen NF aufgerufen.
      In der aeusseren Schleife werden die gruenen (linken), in der inneren Schleife die
      roten (rechten) Elemente behandelt.
      */

   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr1);
   while (!HA_EndTraverseN (&TraversePtr1)) {
      HA_NextTraverseN (&TraversePtr1, (void **)&colTermPtr1);

      if ((MNF_tcolour_green (*colTermPtr1)) && MNF_tirred (*colTermPtr1)) {
         HA_InitTraverseN (&menge->HashTabelle, &TraversePtr2);
         while (!HA_EndTraverseN (&TraversePtr2)) {
            HA_NextTraverseN (&TraversePtr2, (void **)&colTermPtr2);
            if ((MNF_tcolour_red (*colTermPtr2)) && MNF_tirred (*colTermPtr2)) {
               if (DurchlaufProc (colTermPtr1->term, colTermPtr2->term)) {
                  menge->joinedCTerm1 = colTermPtr1;
                  menge->joinedCTerm2 = colTermPtr2;
                  menge->deallocJoinedCTerm1 = FALSE;
                  menge->deallocJoinedCTerm2 = FALSE;
                  menge->joinable = TRUE;
                  return TRUE;
               }
            }
         }
      }
   }
   return FALSE;
}


static unsigned int  count_vars (TermT term) 
{
   SymbolT      symbol;
   unsigned int  count=0;

   TO_doSymbole(term, symbol,
                {if (SO_SymbIstVar (symbol))
                      count++;
                }
                );
   return count;
}


static void MNF_ConvertRegelToAntiRegel (MNF_AntiT antiKind,
                                         RegelT Regel MAYBE_UNUSEDPARAM,
                                         TermT *antiLinks MAYBE_UNUSEDPARAM, TermT *antiRechts MAYBE_UNUSEDPARAM,
                                         AntiRegelUseT *useAnti)
{
   BOOLEAN res;

   *useAnti = ARnoUse;
   switch (antiKind) {
   case antiWOVar:
   /* Wenn beide Seiten keine Variablen enthalten,
      da dies eine Regel ist, reicht es zu pruefen, ob die linke
      Seite keine Variablen enthaelt.
      */
      res = (count_vars (Regel->linkeSeite) == 0);
      if (res) {
         *antiLinks  = TO_Termkopie(Regel->rechteSeite);
         *antiRechts = TO_Termkopie(Regel->linkeSeite);
         *useAnti = ARuseAnti;
      }
      break;
   case antiBoundVar:
      /* Wenn alle Variablen auf beiden Seiten auftauchen, dann koennen
         wir's versuchen */
      /* Testen, ob Variablen von linker Seite Teilmenge von rechts sind,
         dann ist unsere Bedingung erfuellt */
      res = MNF_VMTeilmengenP (Regel->linkeSeite, Regel->rechteSeite);
      if (res) {
         *antiLinks  = TO_Termkopie(Regel->rechteSeite);
         *antiRechts = TO_Termkopie(Regel->linkeSeite);
         /*
           Fuer Matching muessen Terme normiert sein:
           */
         BO_TermpaarNormierenAlsKP(*antiLinks, *antiRechts);
         *useAnti = ARuseAnti;
      }
      break;
   case antiFreeVar:
      res = MNF_VMTeilmengenP (Regel->linkeSeite, Regel->rechteSeite);
      *antiLinks  = TO_Termkopie(Regel->rechteSeite);
      *antiRechts = TO_Termkopie(Regel->linkeSeite);

      /* Fuer Matching muessen Terme normiert sein: */
      BO_TermpaarNormierenAlsKP(*antiLinks, *antiRechts);
      if (res) {
         *useAnti = ARuseAnti;
      } else {
         *useAnti = ARuseUnifikation;
      }
      break;
   case noAnti:
   default:
      break;
   }
}


/**************************************************************************************/
/* Zielverwaltung                                                                     */
/**************************************************************************************/

BOOLEAN MNF_InitNMengeAposteriori (InfoStringT DEF, InfoStringT USR) 
{
   MNF_InitAposteriori ();

   if (IS_FindValuedOption (DEF, USR, 1, "mnf") == -1)
      return FALSE;
   MNF_ParseInfoStrings (DEF, USR, &NMuseE, &NMuseR, &NMMNFPQ_Art,
                         &NMmaxSize, &NMmaxGlSchritte, &NMAddTryAllElements,
                         &NMmaxAntiRSchritte,
                         &NMantiKind,
                         &NMuseAnalyse, &NMSwitchIrred,
                         &NMSwitchNF, &NMCountNF,
                         &NMHashSize);
#if !COMP_POWER
   MNF_PrintOptions (NMuseE, NMuseR, NMMNFPQ_Art, NMmaxSize, NMmaxGlSchritte,
                     NMAddTryAllElements, NMmaxAntiRSchritte,
                     NMantiKind,
                     NMuseAnalyse, NMSwitchIrred, NMSwitchNF, NMCountNF,
                     NMHashSize);
#endif

   return TRUE;
}


struct MNF_MengeTS *MNF_InitNMenge (TermT s, TermT t) 
{
   MNF_MengePtrT mengePtr;

   SV_Alloc(mengePtr,MNF_MengeSpeicher,MNF_MengePtrT);
   MNF_InitFirstMenge (mengePtr);
   mengePtr->useE = NMuseE;
   mengePtr->useR = NMuseR;
   mengePtr->useAnalyse = NMuseAnalyse;
   mengePtr->SwitchIrred = NMSwitchIrred;
   mengePtr->SwitchNF = NMSwitchNF;
   mengePtr->CountNF = NMCountNF;
   mengePtr->MNFPQ_Art = NMMNFPQ_Art;
   mengePtr->maxSize  = NMmaxSize;
   mengePtr->maxGlSchritte = NMmaxGlSchritte;
   mengePtr->maxAntiRSchritte = NMmaxAntiRSchritte;
   mengePtr->antiKind = NMantiKind;
   mengePtr->MNF_AddTryAllElements = NMAddTryAllElements;
   mengePtr->Expand = ExpandAll;
   mengePtr->LastPseudoExpandTabPtr = NULL;
   MNF_InitMenge (mengePtr, mengePtr->initialHashSize);
   MNF_PutFirstElementsInMenge (mengePtr, s, t);
   return mengePtr;
}


void MNF_ExpandNMenge (struct MNF_MengeTS *menge) 
{
   MNF_MengeSetREStatus (menge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());

   if (menge->Expand == ExpandRewrite)
      MNF_PseudoExpand (menge);
   else
      MNF_ExpandNM (menge);
}


void MNF_DeleteNMenge (MNF_MengeT *menge) 
{
   MNF_StatNM (menge);

   HA_Delete(&menge->HashTabelle);
   if (menge->LastPseudoExpandTabPtr) {
      HA_Delete (menge->LastPseudoExpandTabPtr);
      our_dealloc (menge->LastPseudoExpandTabPtr);
   }
   menge->MNFPQ_Delete(menge->PQRed);
   menge->MNFPQ_Delete(menge->PQGreen);

   if (menge->deallocJoinedCTerm1)
      DeallocData (menge->joinedCTerm1);

   if (menge->deallocJoinedCTerm2)
      DeallocData (menge->joinedCTerm2);
   SV_Dealloc (MNF_MengeSpeicher, menge);
}


static void
MNF_AddRENM (MNF_MengeT *menge, RegelOderGleichungsT RE,
             void (*MNF_BearbeiteEineREProc) (void *, MNF_MengeT *, int, MNF_ColTermPtrT,  RegelOderGleichungsT))
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr1, TraversePtr2;
   
   /* Durchlauf durch alle Daten. Problem ist, dass in MNF_BearbeiteEinR Elemente
      eingefuegt werden, die wir im Moment nicht testen wollen. 
      Da aber nur alle bereits expandierten Elemente das texpanded-Flag
      gesetzt haben, sind diese leicht heraus zufiltern.
      Falls zwischendurch die Elemente zusammenfuehrbar waren, dann
      hatten wir Glueck und brechen ab */
   if (menge->MNF_AddTryAllElements) {
      HA_InitTraverseN(&menge->HashTabelle, &TraversePtr1);

      while (!HA_EndTraverseN(&TraversePtr1) && !menge->joinable) {
         HA_NextTraverseN(&TraversePtr1, (void **)&colTermPtr);
         /* Es werden immer nur bereits expandierte Elemente untersucht,
            nicht expandierte werden in MNF_ExpandNM untersucht */
         if (MNF_texpanded (*colTermPtr)) {
            if (MNF_tcolour (*colTermPtr) == MNF_red) {
               MNF_BearbeiteEineREProc (menge->PQRed, menge, MNF_red, colTermPtr, RE);
            }
            else if (MNF_tcolour (*colTermPtr) == MNF_green) {
               MNF_BearbeiteEineREProc (menge->PQGreen, menge, MNF_green, colTermPtr, RE);
            }
         }
      }
   }
   else {
      /* nur "interessante Elemente ansehen */
      HA_InitTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2);
      if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2))
         HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
      while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2) && !menge->joinable) {
         /* wir testen nur Normalformen und Ursprungselemente */
         if ((colTermPtr->flags.irred) || (colTermPtr->anzGesSchritte == 0)) {
            /* nicht expandierte Elemente werden hier nicht betrachtet */
            if (MNF_texpanded (*colTermPtr)) {
               if (MNF_tcolour (*colTermPtr) == MNF_red) {
                  MNF_BearbeiteEineREProc (menge->PQRed, menge, MNF_red, colTermPtr, RE);
               }
               else if (MNF_tcolour (*colTermPtr) == MNF_green) {
                  MNF_BearbeiteEineREProc (menge->PQGreen, menge, MNF_green, colTermPtr, RE);
               }
            }
            HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
         }
         /* Elemente, die weder Normalformen noch Ursprungselemente sind, werden aus der Schlange geloescht */
         else {
            HA_DelAndNextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
         }
      }
   }
}



void MNF_AddRegelNM (MNF_MengeT *menge, RegelT Regel)
{
   if (menge->Expand != ExpandRewrite)
      MNF_AddRENM (menge, Regel, MNF_BearbeiteEineR);

}


void MNF_AddGleichungNM (MNF_MengeT *menge, GleichungsT Gleichung) 
{
   if (menge->Expand != ExpandRewrite)
      MNF_AddRENM (menge, Gleichung, MNF_BearbeiteEineGl);
}


/**************************************************************************************/
/* Test-Routinen                                                                      */
/**************************************************************************************/


BOOLEAN MNF_BildungTest (TermT s, TermT t)
{
   BOOLEAN         retVal;
   MNF_MengePtrT   menge;

   SV_Alloc(menge,MNF_MengeSpeicher,MNF_MengePtrT);
   MNF_InitFirstMenge (menge);
   MNF_PutFirstElementsInMenge (menge, s, t);
   
   retVal = MNF_JoinedNM (menge);

   if (!retVal) {
      MNF_ExpandNM (menge);
      retVal = MNF_JoinedNM (menge);
   }
   MNF_DeleteNMenge (menge);

   return retVal;
}

/**************************************************************************************/
/* Work-In-Progress                                                                   */
/**************************************************************************************/

static void MNF_FilterInteresting (MNF_MengeT *menge, MNF_ColTermPtrT *colTermPtr,
                                   HashEntryPtrT *TraversePtr1, HashEntryPtrT *TraversePtr2,
                                   unsigned int  *countColour, unsigned int  *countTotal) 
{

   if ((*colTermPtr)->anzGesSchritte == 0) {
      HA_NextTraverseI (&menge->HashTabelle, TraversePtr1, TraversePtr2, (void **)colTermPtr);
      /* Ursprungselement wird nicht aufaddiert, ist direkt am Anfang in Summe drin */
      (*countTotal)++;
   }
   else if (((*colTermPtr)->flags.irred) && (*countColour < menge->CountNF)) {
      HA_NextTraverseI (&menge->HashTabelle, TraversePtr1, TraversePtr2, (void **)colTermPtr);
      (*countTotal)++;
      (*countColour)++;
   }
   else
      HA_DelAndNextTraverseI (&menge->HashTabelle, TraversePtr1, TraversePtr2, (void **)colTermPtr);
   
}

/* alle nicht interessanten Elemente rausschmeissen um Menge klein zu halten */
static void MNF_SearchInteresting (MNF_MengeT *menge)
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr1, TraversePtr2;
   unsigned int          count = 0;
   /* Die Zaehler countL/R zaehlen die Elemente pro Farbe, die eingefuegt sind. Da die Ursprungselemente immer
      eingefuegt werden, wird der Zaehler direkt auf 1 gesetzt - dadurch ist der Rest des Programmes auch einfacher
      */
   unsigned int          countL = 1;
   unsigned int          countR = 1;
   
   
   /* nur "interessante Elemente ansehen */
   HA_InitTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2);
   if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2))
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
   while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2) && !menge->joinable) {
      /* wir testen nur Normalformen und Ursprungselemente */
      if (MNF_tcolour_red(*colTermPtr)) {
         MNF_FilterInteresting (menge, &colTermPtr, &TraversePtr1, &TraversePtr2, &countR, &count);
      }
      else if (MNF_tcolour_green (*colTermPtr)) {
         MNF_FilterInteresting (menge, &colTermPtr, &TraversePtr1, &TraversePtr2, &countL, &count);
      }
   }
   /*fprintf (IO_Strom, "SearchInteresting: count is %lu.\n", count);*/
}

				  
static BOOLEAN MNF_PseudoDurchlaufPaarweise (MNF_MengeT *menge, MNF_DurchlaufPaarProcT DurchlaufProc)
{
   MNF_ColTermPtrT       colTermPtr1 = NULL;
   MNF_ColTermPtrT       colTermPtr2 = NULL;
   HashEntryPtrT         TraversePtr1a, TraversePtr1b,TraversePtr2a,TraversePtr2b, TraversePtr1, TraversePtr2;

   /* Die folgenden 4 Schleifen sind alle fast gleich - abgesehen von Kleinigkeiten wie CopyColTerm und
      Initialisierung, gehen zum naechsten Element. */
   /* Durchlaufen durch alle Elemente in LastPseudoExpandTabPtr Hash und durch alle
      interessanten Elemente im normalen Hash
    */
   /* nur interessante Elemente ansehen */
   /* Erst interessante x interessante */
   HA_InitTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b);
   if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b))
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b, (void **)&colTermPtr1);
   while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b)) {
      if (MNF_tcolour_green (*colTermPtr1)) {
	 HA_InitTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b);
	 if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b))
	    HA_NextTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b, (void **)&colTermPtr2);
         while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b)) {
            if (MNF_tcolour_red (*colTermPtr2)) {
               if (DurchlaufProc (colTermPtr1->term, colTermPtr2->term)) {
                  menge->joinedCTerm1 = colTermPtr1;
                  menge->joinedCTerm2 = colTermPtr2;
                  menge->deallocJoinedCTerm1 = FALSE;
                  menge->deallocJoinedCTerm2 = FALSE;
                  menge->joinable = TRUE;
                  return TRUE;
               }
            }
	    HA_NextTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b, (void **)&colTermPtr2);
         }
      }
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b, (void **)&colTermPtr1);
   }
   /* beim ersten Aufruf ist LastPseudoExpandTabPtr NULL */
   if (!menge->LastPseudoExpandTabPtr)
     return FALSE;
   /* Jetzt lastPseudoExpandTabPtr x lastPseudoExpandTabPtr */
   HA_InitTraverseN (menge->LastPseudoExpandTabPtr, &TraversePtr1);
   while (!HA_EndTraverseN (&TraversePtr1)) {
      HA_NextTraverseN (&TraversePtr1, (void **)&colTermPtr1);
      
      if (MNF_tcolour_green (*colTermPtr1)) {
	 HA_InitTraverseN (menge->LastPseudoExpandTabPtr, &TraversePtr2);
         while (!HA_EndTraverseN (&TraversePtr2)) {
            HA_NextTraverseN (&TraversePtr2, (void **)&colTermPtr2);
            if (MNF_tcolour_red (*colTermPtr2)) {
               if (DurchlaufProc (colTermPtr1->term, colTermPtr2->term)) {
                  menge->joinedCTerm1 = CopyColTerm (colTermPtr1);
                  menge->joinedCTerm2 = CopyColTerm (colTermPtr2);
                  menge->deallocJoinedCTerm1 = FALSE;
                  menge->deallocJoinedCTerm2 = FALSE;
                  menge->joinable = TRUE;
                  return TRUE;
               }
            }
         }
      }
   }

   /* interessante x lastPseudoExpandTabPtr */
   HA_InitTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b);
   if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b))
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b, (void **)&colTermPtr1);
   while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b)) {
      if (MNF_tcolour_green (*colTermPtr1)) {
	 HA_InitTraverseN (menge->LastPseudoExpandTabPtr, &TraversePtr2);
         while (!HA_EndTraverseN (&TraversePtr2)) {
            HA_NextTraverseN (&TraversePtr2, (void **)&colTermPtr2);
            if (MNF_tcolour_red (*colTermPtr2)) {
               if (DurchlaufProc (colTermPtr1->term, colTermPtr2->term)) {
                  menge->joinedCTerm1 = colTermPtr1;
                  menge->joinedCTerm2 = CopyColTerm (colTermPtr2);
                  menge->deallocJoinedCTerm1 = FALSE;
                  menge->deallocJoinedCTerm2 = FALSE;
                  menge->joinable = TRUE;
                  return TRUE;
               }
            }
         }
      }
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1a, &TraversePtr1b, (void **)&colTermPtr1);
   }
   /* lastPseudoExpandTabPtr x interessante */
   HA_InitTraverseN (menge->LastPseudoExpandTabPtr, &TraversePtr1);
   while (!HA_EndTraverseN (&TraversePtr1)) {
      HA_NextTraverseN (&TraversePtr1, (void **)&colTermPtr1);

      if (MNF_tcolour_green (*colTermPtr1)) {
	 HA_InitTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b);
	 if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b))
	    HA_NextTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b, (void **)&colTermPtr2);
         while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b)) {
            if (MNF_tcolour_red (*colTermPtr2)) {
               if (DurchlaufProc (colTermPtr1->term, colTermPtr2->term)) {
                  menge->joinedCTerm1 = CopyColTerm (colTermPtr1);
                  menge->joinedCTerm2 = colTermPtr2;
                  menge->deallocJoinedCTerm1 = FALSE;
                  menge->deallocJoinedCTerm2 = FALSE;
                  menge->joinable = TRUE;
                  return TRUE;
               }
            }
	    HA_NextTraverseI (&menge->HashTabelle, &TraversePtr2a, &TraversePtr2b, (void **)&colTermPtr2);
         }
      }
   }

   return FALSE;
}


static void MNF_PseudoExpand (MNF_MengeT *menge)
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   MNF_ColTermPtrT       newCTerm;
   HashEntryPtrT         TraversePtr1, TraversePtr2;
   TermT                 termCopy;
   MNF_ColTermPtrT       oldCData;
   BOOLEAN               Joined;
   BOOLEAN               reduced;
   HashKeyT              Key1, Key2;
   

   if (menge->LastPseudoExpandTabPtr) {
      HA_Clear (menge->LastPseudoExpandTabPtr);
   } else {
      /* 1021 ist eine Primzahl */
      menge->LastPseudoExpandTabPtr = our_alloc (sizeof(HashT));
      HA_New (menge->LastPseudoExpandTabPtr,  (HA_CmpFuncT) &CmpData,
	      (HA_DeallocFuncT) &DeallocData, 1021);
   }

   /* nur "interessante Elemente ansehen */
   HA_InitTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2);
   if (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2))
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
   while (!HA_EndTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2)
          && !menge->joinable) {
      termCopy = TO_Termkopie (colTermPtr->term);
      PCL_StartFeyChain();
      reduced = NF_NFBildung(termCopy, Ziele);
      PCL_DeleteChain();
      if (reduced) {
         /* Reduktion war erfolgreich */
         SetColTerm (&newCTerm, termCopy, MNF_tcolour(*colTermPtr), colTermPtr, Groesser, 0, &Key1, &Key2);
         if (MNF_tcolour_red(*colTermPtr))
            SetPclInfo(&(newCTerm->pcl_info), NULL, 0, rechteSeite, colTermPtr, 0);
         else 
            SetPclInfo(&(newCTerm->pcl_info), NULL, 0, linkeSeite, colTermPtr, 0);
         /* Haben wir diesen Wert bereits in HashTabelle?
            dann ist oldCData != NULL */
         HA_CheckData (&(menge->HashTabelle), Key1, Key2, newCTerm, (void **)&oldCData);
         if (oldCData == NULL) {
            /* einfuegen in lokalen Hash */
            HA_AddData (menge->LastPseudoExpandTabPtr, Key1, Key2, newCTerm, (void **)&oldCData);
            if (oldCData != NULL) {
               /* Datensatz existiert schon: Also beide Saetze mergen und Speicher freigeben */
               Joined = AddData (oldCData, newCTerm);
               if (Joined) {
               /* oldCData ist in hash gespeichert - soll aber in menge->joinedCTerm1 rein,
                  und oldCData wird spaeter freigegeben.
                  Deswegen muss eine Kopie gemacht werden */
                  menge->joinedCTerm1 = CopyColTerm (oldCData);
                  menge->joinedCTerm2 = newCTerm;
                  menge->deallocJoinedCTerm1 = TRUE;
                  menge->deallocJoinedCTerm2 = TRUE;
                  menge->joinable = TRUE;
               }
               else {
                  /* Term freigeben */
                  DeallocData (newCTerm);
               }
            } else { /* oldCData == NULL, Element also nicht in menge->HashTabelle */
               ; /* vorerst gibt's dann nichts zu tun */
            }
         }
         else { /* oldCData != NULL */
            MNF_IncAnzTermGleich;
            if (MNF_tcolour_opp(*newCTerm, *oldCData)) {
               /* joined mit Element in menge->hashTabelle */
               MNF_IncAnzTermGleichOpp;
               menge->joinedCTerm1 = oldCData;
               menge->joinedCTerm2 = newCTerm;
               menge->deallocJoinedCTerm1 = FALSE;
               menge->deallocJoinedCTerm2 = TRUE;
               menge->joinable = TRUE;
            }
            else {
               MNF_IncAnzTermGleichCol;
               DeallocData (newCTerm);
            }
         }
      }
      else {
         /* wurde nicht reduziert */
         TO_TermLoeschen (termCopy);
      }
      HA_NextTraverseI (&menge->HashTabelle, &TraversePtr1, &TraversePtr2, (void **)&colTermPtr);
   }
}

#if 0
/* Bilden der Normalformmenge */
/* liefert TRUE, wenn Elemente zusammengefuehrt werden konnten */
static BOOLEAN
MNF_BearbeiteInsertElemIrred (TermT newTerm, MNF_MengeT *menge, int colour)
{
   MNF_ColTermT         *oldCData = NULL;
   MNF_ColTermT         *newCTerm;
   BOOLEAN              Joined = FALSE;
   HashKeyT             Key1, Key2;

   SetColTerm (&newCTerm, newTerm, colour, NULL, Groesser, 0, &Key1, &Key2);
   SetPclInfo (&(newCTerm->pcl_info), NULL, 0, rechteSeite, NULL, 0);
   HA_AddData(&(menge->HashTabelle), Key1, Key2, newCTerm, (void **)&oldCData);
   if (oldCData != NULL) {
      /* Datensatz existiert schon: Also beide Saetze mergen und Speicher freigeben */
      Joined = AddData (oldCData, newCTerm);
      if (Joined) {
         menge->joinedCTerm1 = oldCData;
         menge->joinedCTerm2 = newCTerm;
         menge->deallocJoinedCTerm1 = FALSE;
         menge->deallocJoinedCTerm2 = TRUE;
         menge->joinable = TRUE;
      }
      else
         DeallocData (newCTerm);
   } else {
      MNF_incID();
      menge->size++;
   }
   return Joined;
}


static BOOLEAN MNF_BearbeiteIrred (MNF_MengeT *menge, int colour, TermT acctTerm)
{
   GNTermpaarT          RegelIndex;
   TermT                TeilTerm, TermNew;
   BOOLEAN              hatNachfolger = FALSE;
   
   TO_doStellen(acctTerm, TeilTerm,
                {
                   BK_forRegeln(RegelIndex, RE_Regelbaum,
                                {
                                   if (TO_TermIstNichtVar(TeilTerm)) {
                                      MNF_IncAnzMatchVersucheR;
                                      if (MO_TermpaarPasst(RegelIndex->linkeSeite,
                                                           RegelIndex->rechteSeite,
                                                           TeilTerm)) {
                                         MNF_IncAnzReduktionen;
                                         TermNew = MO_AllTermErsetztNeu(acctTerm, TeilTerm);
                                         hatNachfolger = TRUE;
                                         if (MNF_BearbeiteIrred (menge, colour, TermNew)) 
                                            {
                                               TO_TermLoeschen (TermNew);
                                               return TRUE;
                                            }
                                         TO_TermLoeschen (TermNew);
                                      }
                                   }
                                }
                                );
                }
                );
   if (!hatNachfolger) {
      TermT kopie = TO_Termkopie (acctTerm);
      
      if (MNF_BearbeiteInsertElemIrred (kopie, menge, colour))
         return TRUE;
   }
   return FALSE;
}


BOOLEAN MNF_ExpandNormalForm (MNF_MengeT *menge, TermT links, TermT rechts) {

   MNF_MengeSetREStatus (menge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());

   if (TO_TermeGleich (links, rechts)) {
      MNF_IncAnzDirektZusammenfuehrbar;
      return TRUE;
   }
   
   if (!RE_RegelmengeLeer ())
      {
         MNF_BearbeiteIrred (menge, MNF_green, links);
         return MNF_BearbeiteIrred (menge, MNF_red, rechts);
      }
   return FALSE;
}
#endif
