/**********************************************************************
* Filename : MNF_CPs.c
*
* Kurzbeschreibung : MultipleNormalFormen: Critical Pairs (Select/Generate)
*
* Autoren : Andreas Jaeger
*
* 18.11.96: Erstellung
* 22.02.97: Mehrfachaufrufclean. AJ.
*
*
*
* Benoetigt folgende Module:
*  - MNF
*
* Benutzte Globale Variablen aus anderen Modulen:
*   zur Zeit keine
*
**********************************************************************/

#include "antiwarnings.h"
#include "general.h"
#include "compiler-extra.h"
#include "Hash.h"
#include "MNF_CPs.h"
#include "MNF_intern.h"
#include "MNF.h"
#include "MNF_PQ.h"
#include "NFBasis.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "Subsumption.h"
#include "SpeicherVerwaltung.h"
#include "Hauptkomponenten.h"

/* Variablen fuer Initialisierungen */
static MNF_MengeT        MNF_KPGMenge;
static MNF_MengeT        MNF_KPSMenge;
static BOOLEAN           MNF_KPGInitialised = FALSE;
static BOOLEAN           MNF_KPSInitialised = FALSE;

static unsigned short    MNF_KPGPermGrad;
static BOOLEAN           MNF_KPGSubsumption;
static MNF_DurchlaufPaarProcT MNF_KPGDurchlauf;

/* Speicherverwaltung */
static SV_SpeicherManagerT MNF_NFSetSpeicher;

static void MNF_Classify (MNF_MengeT *menge) 
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr;

   if (!menge->MNF_UseClassify)
      return;

   if (menge->MNF_ClassInit != NULL)
      menge->MNF_ClassInit (menge->size > menge->maxSize);
   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);

      if (MNF_tcolour_red (*colTermPtr)) {
         menge->MNF_ClassTrav (colTermPtr->term, MNF_tirred(*colTermPtr), FALSE);
      } else if (MNF_tcolour (*colTermPtr) == MNF_green) {
         menge->MNF_ClassTrav (colTermPtr->term,  MNF_tirred(*colTermPtr), TRUE);
      }
   }
   if (menge->MNF_ClassExit != NULL)
      menge->MNF_ClassExit ();
   
}


static void MNF_BuildNFSet (MNF_MengeT *menge, MNF_NFSetT **NFLeft, MNF_NFSetT **NFRight) {

   MNF_ColTermPtrT       colTermPtr = NULL;
   int                   AnzIrredR = 0;
   int                   AnzIrredG = 0;
   MNF_NFSetPtrT         leftSetPtr = NULL;
   MNF_NFSetPtrT         rightSetPtr = NULL;
   MNF_NFSetPtrT         tmpPtr;
   HashEntryPtrT         TraversePtr;
   
   /* Wenn mit Deckelfkt. gearbeitet wird, dann sind einige Elemente noch nicht expandiert.
      Im Extrem-Fall kann es passieren, dass noch keine Normalform unter den Elementen ist.
      Dieser Fall wird im folgenden mitbehandelt.
      Es werden entweder alle Normalformen zurueckgegeben, oder, wenn fuer eine Seite keine
      Normalformen gebildet wurden, alle Elemente, die keine Nachfolger haben (dies sind dann
      Elemente, die noch expandiert werden muessen) */

   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      if ((MNF_tcolour_red (*colTermPtr)) && (colTermPtr->flags.irred)) {
         if (MNF_tirred (*colTermPtr))
            AnzIrredR++;
      } else if ((MNF_tcolour_green (*colTermPtr)) && (colTermPtr->flags.irred)) {
         if (MNF_tirred (*colTermPtr))
            AnzIrredG++;
      }
   }
   
   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      if ((MNF_tcolour (*colTermPtr) == MNF_red)) {
         if (((AnzIrredR == 0) && (colTermPtr->flags.irred)) || (MNF_tirred (*colTermPtr))) {
            SV_Alloc(tmpPtr,MNF_NFSetSpeicher,MNF_NFSetPtrT);
            tmpPtr->next = rightSetPtr;
            tmpPtr->Term = TO_Termkopie(colTermPtr->term);
            rightSetPtr=tmpPtr;
         }
      } else if (MNF_tcolour (*colTermPtr) == MNF_green) {
         if (((AnzIrredG == 0) && (colTermPtr->flags.irred)) || (MNF_tirred (*colTermPtr))) {
            SV_Alloc(tmpPtr,MNF_NFSetSpeicher,MNF_NFSetPtrT);
            tmpPtr->next = leftSetPtr;
            tmpPtr->Term = TO_Termkopie(colTermPtr->term);
            leftSetPtr=tmpPtr;
         }
      }
   }
   *NFLeft=leftSetPtr;
   *NFRight=rightSetPtr;
}


static void MNF_DeleteNFSingleSet (MNF_NFSetT **NFSet) {
   MNF_NFSetPtrT    prevPtr;
   MNF_NFSetPtrT    currPtr=*NFSet;

   while (currPtr != NULL) {
      prevPtr = currPtr;
      currPtr = currPtr->next;
      TO_TermLoeschen(prevPtr->Term);
      SV_Dealloc(MNF_NFSetSpeicher, prevPtr);
   }
}


void MNF_REMengenGeaendert (BOOLEAN RLeer, BOOLEAN ELeer) {

   if (MNF_KPGInitialised)
      MNF_MengeSetREStatus (&MNF_KPGMenge, RLeer, ELeer);

   if (MNF_KPSInitialised)
      MNF_MengeSetREStatus (&MNF_KPSMenge, RLeer, ELeer);
   
}


static BOOLEAN
MNF_KPGDurchlaufPermsub (TermT term1, TermT term2) 
{
   if (SS_TermpaarSubsummiertVonPermsub(term1, term2, MNF_KPGPermGrad)) {
      HK_fair = FALSE;
      SUE_recentCPMNFpermsubsumed();
      return TRUE;
   } else {
      return FALSE;
   }
}


static BOOLEAN
MNF_KPGDurchlaufSub (TermT term1, TermT term2) 
{
   if (SS_TermpaarSubsummiertVonGM(term1, term2)) {
      SUE_recentCPMNFsubsumed();
      return TRUE;
   } else {
      return FALSE;
   }
}


static BOOLEAN
MNF_KPGDurchlaufSubPermsub (TermT term1, TermT term2) 
{
   if (SS_TermpaarSubsummiertVonPermsub(term1, term2, MNF_KPGPermGrad)) {
      SUE_recentCPMNFpermsubsumed();
      HK_fair = FALSE;
      return TRUE;
   } else if (SS_TermpaarSubsummiertVonGM (term1, term2)) {
      SUE_recentCPMNFsubsumed();
      return TRUE;
   } else {
      return FALSE;
   }
}


/**************************************************************************************/
/* KP-Generate                                                                        */
/**************************************************************************************/


BOOLEAN MNF_InitKPGAposteriori (InfoStringT DEF, InfoStringT USR,
                                unsigned short PermGrad, BOOLEAN useSubsumption) 
{
   MNF_InitAposteriori ();
   
   if (IS_FindValuedOption (DEF, USR, 1, "mnf") == -1)
      return FALSE;
   if (!MNF_KPGInitialised) {
      MNF_InitFirstMenge (&MNF_KPGMenge);
      MNF_KPGInitialised = TRUE;
   }
   MNF_ParseInfoStringsMenge (DEF, USR, &MNF_KPGMenge);
   MNF_InitMenge (&MNF_KPGMenge, MNF_KPGMenge.initialHashSize);
   MNF_PrintOptionsMenge (&MNF_KPGMenge);

   MNF_KPGPermGrad = PermGrad;
   MNF_KPGSubsumption = useSubsumption;
   if (MNF_KPGPermGrad && MNF_KPGSubsumption)
      MNF_KPGDurchlauf = MNF_KPGDurchlaufSubPermsub;
   else if (MNF_KPGSubsumption)
      MNF_KPGDurchlauf = MNF_KPGDurchlaufSub;
   else if (MNF_KPGPermGrad)
      MNF_KPGDurchlauf = MNF_KPGDurchlaufPermsub;
   else
      MNF_KPGDurchlauf = NULL;
   
   return TRUE;
}



/* Testen auf Zusammenfuehrbarkeit und setzen von Parametern fuer
   Classifikation bei der Generierung von CP */
BOOLEAN MNF_KPGBearbeite (TermT links, TermT rechts, TermT *joinedTerm)
{
   BOOLEAN      retVal;
   
   MNF_MengeSetREStatus (&MNF_KPGMenge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());

   MNF_PutFirstElementsInMenge (&MNF_KPGMenge, links, rechts);
   retVal = MNF_JoinedNMenge (&MNF_KPGMenge);

   if (!retVal) {
      MNF_ExpandNM (&MNF_KPGMenge);
      retVal = MNF_JoinedNMenge (&MNF_KPGMenge);
   }
   if (!retVal && MNF_KPGPermGrad > 0) {
      
   }
   if (retVal) {
      *joinedTerm = TO_Termkopie(MNF_KPGMenge.joinedCTerm1->term);
   } else {
      MNF_Classify (&MNF_KPGMenge);
   }
   MNF_ClearNM (&MNF_KPGMenge);
   return retVal;
}

#define MNF_HACK_CG 0
/* dieser Hack bringt nichts. AJ */
#if MNF_HACK_CG
/* Testen auf Zusammenfuehrbarkeit und Auswaehlen des besten Paares
   fuer CG */
/* HACK */
#include <limits.h>
#include <NFBildung.h>
#include <Classification.h>
BOOLEAN MNF_CGBearbeite (TermT links, TermT rechts, TermT *joinedTerm, TermT *neuL, TermT *neuR)
{
   BOOLEAN      retVal;
   unsigned int AnzIrredR = 0, AnzIrredG=0, anzTotal=0;
   HashEntryPtrT         TraversePtr;

   
   MNF_MengeSetREStatus (&MNF_KPGMenge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());

   MNF_PutFirstElementsInMenge (&MNF_KPGMenge, links, rechts);
   retVal = MNF_JoinedNMenge (&MNF_KPGMenge);

   if (!retVal) {
      MNF_ExpandNM (&MNF_KPGMenge);
      retVal = MNF_JoinedNMenge (&MNF_KPGMenge);
   }
   if (retVal) {
      *joinedTerm = TO_Termkopie(MNF_KPGMenge.joinedCTerm1->term);
   } else {
      MNF_ColTermPtrT       colTermPtr1 = NULL;
      MNF_ColTermPtrT       colTermPtr2 = NULL;
      TermT                 bestL = NULL, bestR = NULL;
      HashEntryPtrT         TraversePtr1, TraversePtr2;
      WeightEntryT wt;
      long int bestWeight;
      

      HA_InitTraverseN (&MNF_KPGMenge.HashTabelle, &TraversePtr);
      while (!HA_EndTraverseN (&TraversePtr)) {
         HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr1);
         anzTotal++;
         if ((MNF_tcolour_red (*colTermPtr1)) && (colTermPtr1->flags.irred)) {
            if (MNF_tirred (*colTermPtr1))
               AnzIrredR++;
         } else if ((MNF_tcolour_green (*colTermPtr1)) && (colTermPtr1->flags.irred)) {
         if (MNF_tirred (*colTermPtr1))
            AnzIrredG++;
         }
      }
      if ((AnzIrredR != 1) || (AnzIrredG != 1)) {
         bestWeight = INT_MAX;
         HA_InitTraverseN (&MNF_KPGMenge.HashTabelle, &TraversePtr1);
         while (!HA_EndTraverseN (&TraversePtr1)) {
            HA_NextTraverseN (&TraversePtr1, (void **)&colTermPtr1);
            
            if ((MNF_tcolour (*colTermPtr1) == MNF_red)) {
               
               if ((AnzIrredR == 0) && (colTermPtr1->flags.irred)) {
                  NFZuRegelnProz (colTermPtr1->term, KPsGen);
               }
               if (((AnzIrredR == 0) && (colTermPtr1->flags.irred)) || (MNF_tirred (*colTermPtr1))) {
                  HA_InitTraverseN (&MNF_KPGMenge.HashTabelle, &TraversePtr2);
                  while (!HA_EndTraverseN (&TraversePtr2)) {
                     HA_NextTraverseN (&TraversePtr2, (void **)&colTermPtr2);
                     if (MNF_tcolour_green (*colTermPtr2)) {
                        if ((AnzIrredG == 0) && (colTermPtr2->flags.irred)) {
                           NFZuRegelnProz (colTermPtr2->term, KPsGen);
                        }
                        if (((AnzIrredG == 0) && (colTermPtr2->flags.irred)) || (MNF_tirred (*colTermPtr2))) {
                           C_CGCLASSIFY(&wt, colTermPtr1->term, colTermPtr2->term, SUE_ActualParent, SUE_recCPOtherParent, NULL);
                           if (wt.w1 < bestWeight) {
                              bestWeight = wt.w1;
                              bestL = colTermPtr1->term;
                              bestR = colTermPtr2->term;
                           }
                        }
                     }
                  }
               }
            }
         }
         if (bestL != NULL) {
            *neuL = TO_Termkopie (bestL);
            *neuR = TO_Termkopie (bestR);
            /* fprintf (stderr, "Neues bestes Paar gefunden!\n");*/
            
         }
      }
      
      fprintf (stderr, "%5d %5d %5d\n", AnzIrredG, AnzIrredR, anzTotal);
   }
   
   MNF_ClearNM (&MNF_KPGMenge);
   return retVal;
}
#endif

/**************************************************************************************/
/* KP-Classify                                                                        */
/**************************************************************************************/
void MNF_InitClassifyAposteriori (MNF_ClassifyInitT CInit,
                                  MNF_ClassifyTraverseT CTrav,
                                  MNF_ClassifyExitT CExit)
{
   MNF_InitAposteriori ();
   if (!MNF_KPGInitialised)
      MNF_InitFirstMenge (&MNF_KPGMenge);
   MNF_KPGMenge.MNF_ClassInit = CInit;
   MNF_KPGMenge.MNF_ClassTrav = CTrav;
   MNF_KPGMenge.MNF_ClassExit = CExit;
   MNF_KPGMenge.MNF_UseClassify = (CTrav != NULL);
}


/**************************************************************************************/
/* KP-Select                                                                          */
/**************************************************************************************/

BOOLEAN MNF_InitKPSAposteriori (InfoStringT DEF, InfoStringT USR) 
{
   MNF_InitAposteriori ();

   if (IS_FindValuedOption (DEF, USR, 1, "mnf") == -1)
      return FALSE;
   SV_ManagerInitialisieren (&MNF_NFSetSpeicher, MNF_NFSetT, MNF_NFSetBlock, "NF Set (MNF_CPs)");
   if (!MNF_KPSInitialised) {
      MNF_InitFirstMenge (&MNF_KPSMenge);
      MNF_KPSInitialised = TRUE;
   }
   MNF_ParseInfoStringsMenge (DEF, USR, &MNF_KPSMenge);
   MNF_InitMenge (&MNF_KPSMenge, MNF_KPSMenge.initialHashSize);
   MNF_PrintOptionsMenge (&MNF_KPSMenge);
   return TRUE;
}


BOOLEAN MNF_KPSBearbeite (TermT links, TermT rechts,  TermT *joinedTerm,
                          MNF_NFSetT **NFLeft, MNF_NFSetT **NFRight) 
{
   BOOLEAN      retVal;

   MNF_MengeSetREStatus (&MNF_KPSMenge, RE_RegelmengeLeer(), RE_GleichungsmengeLeer());

   MNF_PutFirstElementsInMenge (&MNF_KPSMenge, links, rechts);
   retVal = MNF_JoinedNMenge (&MNF_KPSMenge);

   if (!retVal) {
      MNF_ExpandNM (&MNF_KPSMenge);
      retVal = MNF_JoinedNMenge (&MNF_KPSMenge);
   }

   if (retVal) {
      *NFLeft  = NULL;
      *NFRight = NULL;
      *joinedTerm = TO_Termkopie(MNF_KPSMenge.joinedCTerm1->term);
   }
   else
      MNF_BuildNFSet (&MNF_KPSMenge, NFLeft, NFRight);

   MNF_ClearNM (&MNF_KPSMenge);
   return retVal;
   
}


void MNF_DeleteNFSet (MNF_NFSetT **NFLeft, MNF_NFSetT **NFRight) {

   MNF_DeleteNFSingleSet (NFLeft);
   MNF_DeleteNFSingleSet (NFRight);
}


void MNF_CPCleanup (void) {
   if (MNF_KPGInitialised) {
      HA_Delete (&MNF_KPGMenge.HashTabelle);
      MNF_KPGMenge.MNFPQ_Delete (MNF_KPGMenge.PQRed);
      MNF_KPGMenge.MNFPQ_Delete (MNF_KPGMenge.PQGreen);
      MNF_KPGInitialised = FALSE;
   }
   if (MNF_KPSInitialised) {
      HA_Delete (&MNF_KPSMenge.HashTabelle);
      MNF_KPSMenge.MNFPQ_Delete (MNF_KPSMenge.PQRed);
      MNF_KPSMenge.MNFPQ_Delete (MNF_KPSMenge.PQGreen);
      MNF_KPSInitialised = FALSE;
   }
}
