/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFBildung.c                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (normales Verfahren)                                       */
/* -                                                                                                            */
/* - 23.03.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"

#include "compiler-extra.h"
#include "general.h"		   /* wg. Typ BOOLEAN */
#include "TermOperationen.h"	   /* wg. Typ TermT */
#include "FehlerBehandlung.h"
#include "RUndEVerwaltung.h"
#include "NFInnermost.h"
#include "NFOutermost.h"
#include "NFTopDown.h"
#include "NFBildung.h"
#include "NFBasis.h"              /* wg. Exekutor und Statistik*/

#include <string.h>

/* ------------------------------------------------------------------------*/
/* ---------- globale Variablen (siehe NFBildung.h) -----------------------*/
/* ------------------------------------------------------------------------*/

NFProzedurT           NFBildungProz, NFZuRegelnProz, NFZuGleichungenProz, NFBildungStratProz, NFZuRegelnStratProz, NFZuGleichungenStratProz;
TermpaarIrgendwoT     GleichungAlsTermpaarAnwendbarProz, RegelAlsTermpaarAnwendbarProz;
NFInnenReduzibelT     NFInnenReduzibelProzE, NFInnenReduzibelProzRE, NFInnenReduzibelProzR;
NFRedStellenT         NFAnzRedStellenProzR, NFAnzRedStellenProzE, NFAnzRedStellenProzRE;
BOOLEAN               RMleer, GMleer;
#if NF_TEST
IrrTestProcT          IrrNFBildungProz, IrrNFZuRegelnProz, IrrNFZuGleichungenProz;
BOOLEAN               NF_private_booly;
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Dummy ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFInnenReduzibelDummyProz(UTermT Term MAYBE_UNUSEDPARAM)
{
  return FALSE;
}

BOOLEAN NFDummyProz(TermT Term MAYBE_UNUSEDPARAM, ObjektT objekt MAYBE_UNUSEDPARAM)
{
   return FALSE;
}

int NFRedStellenDummyProz(UTermT Term MAYBE_UNUSEDPARAM)
{
  return 0;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Irreduzibilitaets-Test ----------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
#if NF_TEST
#include "MatchOperationen.h"

long int NFEEges = 0;
long int NFRRges = 0;
long int NFRERges = 0, NFREEges = 0;

static const char *blubb = "|/-\\|/-\\";
static int modulo = 8;
static int zaehler = 0;

#define Blubb() printf("%c\r", blubb[zaehler = (zaehler+1)%modulo])

void NF_IrrTestR(TermT Term)
{
   UTermT TIndex = Term;
   int count = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex) && MO_RegelGefunden(TIndex)){
         IO_DruckeFlex("%t mit R reduzierbar an Stelle %t.\n", Term, TIndex);
         count++;
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFRRges += count;
   if (count>0)fprintf(IO_Strom, "R: Rges = %ld\n", NFRRges);
}

void NF_IrrTestE(TermT Term)
{
   UTermT TIndex = Term;
   int count = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex) && MO_GleichungGefunden(TIndex)){
         IO_DruckeFlex("%t mit E reduzierbar an Stelle %t.\n", Term, TIndex);
         count++;
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFEEges += count;
   if (count>0)fprintf(IO_Strom, "E: Eges = %ld\n", NFEEges);
}

void NF_IrrTestRE(TermT Term)
{
   UTermT TIndex = Term;
   int Rcount = 0, Ecount = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex)){
         if (MO_RegelGefunden(TIndex)){
            IO_DruckeFlex("%t mit R reduzierbar an Stelle %t.\n", Term, TIndex);
            Rcount++;
         }
         if (MO_GleichungGefunden(TIndex)){
            IO_DruckeFlex("%t mit E reduzierbar an Stelle %t.\n", Term, TIndex);
            Ecount++;
         }
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFRERges += Rcount; NFREEges += Ecount;
   if (Rcount+Ecount>0){ 
      fprintf(IO_Strom, "RE: Rges = %ld, Eges = %ld\n", NFRERges, NFREEges);
   }
}

void NF_IrrTestDummy(TermT T MAYBE_UNUSEDPARAM)
{
   ;
}

BOOLEAN NF_IrrtestFehler(void)
{
   return (BOOLEAN) (NFEEges+NFRRges+NFRERges+NFREEges>0);
}

#endif /* NF_TEST */
/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Initialisierung -----------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void NF_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
/* what == ParamEvaluate: Auswerten und Ausgeben, aber nix einstellen.
   what == ParamExplain: Syntax ausgeben.
*/
{
   if (what == ParamEvaluate){
      int Strategie = IS_FindOption(DEF, USR, 2, "outermost", "innermost", "topdown");
      BOOLEAN Wiederholen = (IS_FindOption(DEF, USR, 2, "repeat", "continue") == 0);
      BOOLEAN Aufsteigen  = (IS_FindOption(DEF, USR, 2, "up", "down") == 0);
      fprintf(IO_Strom, "\n* NORMALIZATION:    ");
      switch (Strategie)
      {
      case 1: fprintf(IO_Strom, "innermost"); break;
      case 2: fprintf(IO_Strom, "topdown"); break;
      case 0:
      default: fprintf(IO_Strom, "outermost"); break;
      }      
      fprintf(IO_Strom, ", %s, %s\n", Wiederholen?"repeat":"continue", Aufsteigen?"up":"down");
   }
   else{
      IO_DruckeKurzUeberschrift("Normalization", '-');
      fprintf(IO_Strom, "1. [outermost|innermost|topdown]\n");
      fprintf(IO_Strom, "2. [up|down]\n");
      fprintf(IO_Strom, "3. [repeat|continue]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. Strategy\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "outermost   = leftmost-outermost\n");
      fprintf(IO_Strom, "innermost   = leftmost-innermost\n");
      fprintf(IO_Strom, "topdown     = breadth-first search, level after level\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "2. Backtracking\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "up   = climb up from the current position\n");
      fprintf(IO_Strom, "down = jump to the top-most position and go down\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "3. Behaviour after reduction\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "repeat   = try current position again\n");
      fprintf(IO_Strom, "continue = backtrack immediately\n");
      fprintf(IO_Strom, "\n");
   }      
}   

void NF_InitAposteriori(InfoStringT DEF, InfoStringT USR)

/* Initialisieren der fuer die eingestellte NF-Strategie und -Variante benoetigten Module. 
Strategie: 0 = leftmost-outermost
           1 = leftmost-innermost
           2 = top-down
*/
{
   int Strategie = IS_FindOption(DEF, USR, 3, "outermost", "innermost", "topdown");
   BOOLEAN Wiederholen = (IS_FindOption(DEF, USR, 2, "repeat", "continue") == 0);
   BOOLEAN Aufsteigen  = (IS_FindOption(DEF, USR, 2, "up", "down") == 0);

   RMleer = GMleer = TRUE;

#if !COMP_POWER
   fprintf(IO_Strom, "\n* NORMALIZATION:    ");
#endif

   switch (Strategie){

   case 0:                      /* Leftmost-Outermost */
#if !COMP_POWER
      fprintf(IO_Strom, "Leftmost-Outermost, ");
#endif
      NFO_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFO_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFO_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFO_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFO_NormalformAbDirekt;
        }
      }
      break;

   case 1:                      /* Leftmost-Innermost */
      fprintf(IO_Strom, "Leftmost-Innermost, ");
      NFI_Init();
      if(Wiederholen){
        NFZuRegelnStratProz= NFI_NormalformZuRegelnNochmal;
        NFZuGleichungenStratProz = NFI_NormalformZuGleichungenNochmal;
        NFBildungStratProz = NFI_NormalformBildenNochmal;
      }
      else{
        NFZuRegelnStratProz= NFI_NormalformZuRegelnDirekt;
        NFZuGleichungenStratProz = NFI_NormalformZuGleichungenDirekt;
        NFBildungStratProz = NFI_NormalformBildenDirekt;
      }
      break;

   case 2:                      /* Top-Down */
      fprintf(IO_Strom, "Top-Down, ");
      NFT_Init();
      if (Aufsteigen){
        if (Wiederholen){
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAufNochmal; /* Top-Down, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFT_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAufDirekt;  /* Top-Down, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFT_NormalformAufDirekt;
        }
      }
      else{
        if (Wiederholen){
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAbNochmal; /* Top-Down, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFT_NormalformAbNochmal;
        }
        else{
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAbDirekt;   /* Top-Down, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFT_NormalformAbDirekt;
        }
      }
      break;
    default:
      our_error("Nicht implementierte NF-Variante!\n");
      break;
   }

   NFBildungProz = NFZuRegelnProz = NFZuGleichungenProz = NFDummyProz;
   NFInnenReduzibelProzR = NFInnenReduzibelProzE = NFInnenReduzibelProzRE = NFInnenReduzibelDummyProz;
   NFAnzRedStellenProzR = NFAnzRedStellenProzE = NFAnzRedStellenProzRE = NFRedStellenDummyProz;
#if NF_TEST   
   IrrNFBildungProz = IrrNFZuRegelnProz = IrrNFZuGleichungenProz = NF_IrrTestDummy;
#endif

#if !COMP_POWER
   fprintf(IO_Strom, "%s, %s\n", Wiederholen?"repeat":"continue", Aufsteigen?"up":"down");
#endif

}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Statistik -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void NF_Statistik(InfoStringT IS)
{

   if (!(IS_IsIn("all", IS) || IS_IsIn("NF", IS)))
      return;
   else {
#if NFB_STATISTIK
      unsigned long int 
         AnzRedVersMitRegelnGes = 0, AnzRedVersMitGleichgnGes = 0, 
         AnzRedErfMitRegelnGes = 0, AnzRedErfMitGleichgnGes = 0,
         maxAnz = 0;
      int Zaehler;
      for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
         AnzRedVersMitRegelnGes += AnzRedVersMitRegeln[Zaehler];
         AnzRedVersMitGleichgnGes += AnzRedVersMitGleichgn[Zaehler];
         AnzRedErfMitRegelnGes += AnzRedErfRegeln[Zaehler];
         AnzRedErfMitGleichgnGes += AnzRedErfGleichgn[Zaehler];
         if (AnzRedVersMitRegeln[Zaehler] > maxAnz) maxAnz = AnzRedVersMitRegeln[Zaehler];
         if (AnzRedVersMitGleichgn[Zaehler] > maxAnz) maxAnz = AnzRedVersMitGleichgn[Zaehler];
      }
      if (IS_IsIn("standard", IS)){
         int tiefe1;
         const char *RedObjekteIO[AnzObjektTypen] = {"ues after generation", "ues during IRP", 
                                                  "ues after selection", "rhs of rules",
                                                  "lhs of rules", "equations", "goals"};
         
         if (WerteDa(2, AnzRedVersMitRegelnGes, AnzRedVersMitGleichgnGes)){
            IO_DruckeKurzUeberschrift("rewriting",'+');
   
            tiefe1 = IO_Einruecktiefe(0, Maximum(3, maxAnz, AnzRedMisserfolgeICR, AnzRedMisserfolgeICG));
   
            IO_DruckeStatistikZeile("match queries R",AnzRedVersMitRegelnGes);
            for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
               char str[100];
               strcpy(str, "of them concerning ");
               strcat(str, RedObjekteIO[Zaehler]);
               IO_DruckeStatistikZeileEingerueckt(str, AnzRedVersMitRegeln[Zaehler], tiefe1);
            }
            IO_DruckeStatistikZeileEingerueckt("of them failed due to indir. clashes",AnzRedMisserfolgeICR, 
                                               tiefe1);         
            IO_DruckeStatistikLeerZeile();
            IO_DruckeStatistikZeile("Reductions with R",AnzRedErfMitRegelnGes);
            for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
               char str[100];
               strcpy(str, "of them concerning ");
               strcat(str, RedObjekteIO[Zaehler]);
               IO_DruckeStatistikZeileEingerueckt(str, AnzRedErfRegeln[Zaehler], tiefe1);
            }
   
            IO_DruckeStatistikLeerZeile();
            IO_DruckeStatistikZeile("match queries E",AnzRedVersMitGleichgnGes);
            for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
               char str[100];
               strcpy(str, "of them concerning ");
               strcat(str, RedObjekteIO[Zaehler]);
               IO_DruckeStatistikZeileEingerueckt(str, AnzRedVersMitGleichgn[Zaehler], tiefe1);
            }
            IO_DruckeStatistikZeileEingerueckt("of them failed due to indir. clashes",AnzRedMisserfolgeICG, tiefe1);
            IO_DruckeStatistikLeerZeile();
            IO_DruckeStatistikZeile("Reductions with E",AnzRedErfMitGleichgnGes);
            for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
               char str[100];
               strcpy(str, "of them concerning ");
               strcat(str, RedObjekteIO[Zaehler]);
               IO_DruckeStatistikZeileEingerueckt(str, AnzRedErfGleichgn[Zaehler], tiefe1);
            }
            IO_DruckeStatistikLeerZeile();
            IO_DruckeStatistikZeile("total match queries (CoM)", AnzRedVersMitRegelnGes+AnzRedVersMitGleichgnGes);
            IO_DruckeStatistikZeile("total reductions (Red)", AnzRedErfMitRegelnGes+AnzRedErfMitGleichgnGes);
         }
      }
      else if (IS_IsIn("short", IS)){
         IO_DruckeKurzUeberschrift("rewriting", '+');
         IO_DruckeStatistikZeile("Match Queries", AnzRedVersMitRegelnGes + AnzRedVersMitGleichgnGes);
         IO_DruckeStatistikZeile("Reductions", AnzRedErfMitRegelnGes+ AnzRedErfMitGleichgnGes);
      }
      else if (IS_IsIn("latex-head", IS))
         fprintf(IO_Strom, "Red R: CoM & Red & IC--Fail & Red E: CoM & Red & IC--Fail & Rewrite--Steps: total & CPs & R--rhs & R--lhs & E & Goals & ");
      else if (IS_IsIn("slatex-head", IS))
         fprintf(IO_Strom, "Match Queries &Reductions & ");
      else if (IS_IsIn("intern", IS))
         IO_DruckeStatistikZeile("max. expansion of rewriting path",NFB_AusdehnungPfad());
      else if (IS_IsIn("config", IS)){
         fprintf(IO_Strom, "NF: statistics  ON, ");
         fprintf(IO_Strom, "Irred-test %s, ", NF_TEST ? " ON" : "OFF");
      }
#else
      if (IS_IsIn("standard", IS))
        IO_DruckeStatistikZeile("rewriting-statistics OFF", 0);
      else if (IS_IsIn("intern", IS))
        IO_DruckeStatistikZeile("max. expansion of rewriting path",NFB_AusdehnungPfad());
      else if (IS_IsIn("config", IS)){
        fprintf(IO_Strom, "NF: statistics OFF,");
        fprintf(IO_Strom, "Irred-test %s, ", NF_TEST ? " ON" : "OFF");
      }
#endif
   }
}
   
void NF_AnzReduktionenGesamt(unsigned long int *AnzRedGes)
{
  *AnzRedGes = 0;
#if NFB_STATISTIK
  { int Zaehler;
    for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++)
      *AnzRedGes += AnzRedErfRegeln[Zaehler] + AnzRedErfGleichgn[Zaehler]; }
#endif
}
