/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFOutermost.c                                                                                              */
/* - *************                                                                                              */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (normales Verfahren)                                       */
/* -                                                                                                            */
/* - 23.03.94 Arnim. Angelegt.                                                                                  */
/* - 10.05.94 Arnim. Algorithmen umgestrickt (auf 'direkt' umgestellt) und NF zu R und E eingebaut.             */
/* - 16.05.94 Arnim. Kommentare eingefuegt/redigiert und Kleinigkeiten geaendert.                               */
/* - 23,06.94 Arnim. Sonderbehandlungen fuer Reduktionen auf Wurzelebene VOR Hauptschleife eingebaut und        */
/* -                 IN Hauptschleife herausgeworfen. Begruendungen und Ergebnisse von Tests eingefuegt.        */
/* -                                                  (siehe Kommentar zu NFZuRegeln)                           */
/* - 27.10.95 Arnim. Aus der Outermost-Fassung mit ausfuehrlicher IC-Verwaltung hergestellt.                    */
/* -                 Backtracking zu ICs: es werden nur die IC-Stellen markiert, keine Faedelung vorgenommen.   */
/* -------------------------------------------------------------------------------------------------------------*/



#include "general.h"
#include "TermOperationen.h"
#include "FehlerBehandlung.h"
#include "RUndEVerwaltung.h"
#include "Ausgaben.h"
#include "NFBasis.h"
#include "pcl.h"
#include "NFOutermost.h"

typedef struct ICStruct{
   BOOLEAN IC;
} ICZellenT;

static ICZellenT       *IC = NULL, *IC1 = NULL, *IC2 = NULL;

static void NFO_Ausdehnungsproz(StellenT neueGroesse)
{
   IC1 = our_realloc(IC1, neueGroesse*sizeof(ICZellenT), "Ausdehnung des IC1-Pfades");
   IC2 = our_realloc(IC2, neueGroesse*sizeof(ICZellenT), "Ausdehnung des IC2-Pfades");
   IC = IC1;
}

static BOOLEAN NFO_SchonInitialisiert = FALSE;

void NFO_Init(void)
{
   NFB_Init();
   NFB_AusdehnungsprozedurAnmelden(NFO_Ausdehnungsproz);
   if (!NFO_SchonInitialisiert){
     IC1 = our_alloc(NF_PfadArrayGroesse*sizeof(ICZellenT));
     IC2 = our_alloc(NF_PfadArrayGroesse*sizeof(ICZellenT));
     IC  = IC1;
   }
   NFO_SchonInitialisiert = TRUE;
}

#define AufIC2Schalten() (IC=IC2)
#define VonIC2Zurueckschalten() (IC=IC1)

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Lokale Makros ------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static StellenT        letzteRed,         /* Verweis auf die letzte reduzierte Stelle   */
                       hoechsteRed;       /* Verweis auf die hoechste reduzierte Stelle */

static unsigned long int MaxTiefeRegeln, MaxTiefeGleichungen; 
/* wegen Tiefe("f(a)") == 2 wird hier in PfadNeu MaxTiefe??? auf */
/* RE_MaxTiefe??? ()-1 gesetzt. Sonst Aufstieg eins zu weit!       */

#define PfadInit(/*TermT*/ Term, /*ObjektT*/ objekt)\
{\
    NFB_PfadInit(Term, objekt);\
    letzteRed =0;\
    hoechsteRed = 0;\
    MaxTiefeRegeln = RE_maxTiefeRegel()-1;\
    MaxTiefeGleichungen = RE_maxTiefeGleichung()-1;\
    IC[0].IC=FALSE;\
}

#define hoechsteRedAnpassen(Kandidat) (Kandidat<hoechsteRed?(hoechsteRed = Kandidat):0)

#define GipfelZuR(/*StellenT*/ stelle)            ((stelle > MaxTiefeRegeln)? (stelle - MaxTiefeRegeln):0)
#define GipfelZuE(/*StellenT*/ stelle)            ((stelle > MaxTiefeGleichungen)? (stelle - MaxTiefeGleichungen):0)


/* Unter Gipfel versteht man hier diejenige Stelle im Pfad, bis zu der sich eine Reduktion an der aktuellen Stelle maximal
   auswirken kann (bzgl. Regeln). Hierzu wird die Tiefe der tiefsten Regel nach oben gestiegen, da oberhalb dieser Stelle
   keine Regel mehr anwendbar geworden sein kann, die vorher nicht anwendbar war.*/

   
/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- SetzePfadAufNaechstenInVorOrdnung ----------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Nachdem die aktuelle Stelle im Pfad nicht mehr reduzierbar ist, wird der naechste Teilterm in Vorordnung gesucht und    */
/* zum aktuellen Term gemacht. Dabei wird auch das noetige Backtracking durchgefuehrt                                      */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN WeiterInVO(void)
{
  while (EsGibtKeineZuBehNachf())     /* Aufstieg bis Wurzel oder Knoten mit noch nicht abgearbeiteten Teiltermen */
    if (Wurzel())
      return FALSE;
    else
      PfadEinsAuf();
 
  NFB_PfadEinsWeiterSetzen();                                                        /* sonst: Weiter im Durchlauf */
  IC[aktuelleStelle].IC = FALSE;
  return TRUE;
}



static BOOLEAN WeiterInVOTeilterm(void)
/* Setzt die aktuelle Stelle im Pfad auf ihren Nachfolger in Vorordnung. Dabei wird aber der an der Stelle hoechsteRed
   beginnende Teilterm nicht verlassen.
   Rueckgabe ist TRUE, wenn eine solche Stelle gefunden wurde. In diesem Fall ist aktStelle dann die gefundene Stelle.
   Ist aktStelle bereits die letzte in Vorordnung zu behandelnde Stelle im Teilterm, so wird FALSE zurueckgegeben. aktStelle
   ist dann die Wurzel des in hoechsteRed beginnenden Teiltermes.
   Damit ist am Ende aktuelleStelle==hoechsteRed.
*/
{
   while ((aktuelleStelle != hoechsteRed) && EsGibtKeineZuBehNachf()) /* Aufstieg bis Wurzel des Teilterms oder Knoten 
								   mit noch nicht abgearbeitetem Teilt. */
      PfadEinsAuf();
   if (EsGibtKeineZuBehNachf())                    /* Keinen Knoten mit noch nicht abgearbeiteten Teiltermen gefunden */
      return FALSE;                                     /* => Pfad ist am Ende, dh. Teilterm ist fertig durchlaufen, und*/
                                                        /*    aktStelle ist hoechsteRed, also Wurzel des Teilterms */
   else{			                        /*    (dessen AnzNochZuBehNachf gleich 0 ist) */
      NFB_PfadEinsWeiterSetzen();                   /* sonst: Weiter im Durchlauf */
      IC[aktuelleStelle].IC = FALSE;
      return TRUE;
   }
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Gleichungen ----------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_GleichungAngewendet())
         return TRUE;
      else                                                                      /* ICUpdateBeiAufstiegZuGipfel(); */
         IC[aktuelleStelle].IC = NFB_IC();
   }
   return FALSE;
}

static BOOLEAN GICsAufsteigenBisReduziertOderStop(StellenT start)
/*
- alle ICs excl. start aufsteigen und probieren
- FALSE: aktuelleStelle = 0
- TRUE : aktuelleStelle = letzter reduzierter IC = letzteRed
*/
{
   aktuelleStelle = start;
   while(aktuelleStelle>0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && NFB_GleichungAngewendet())
         return TRUE;
   }
   return FALSE;
}
            
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GICsAbsteigenBisReduziertOderStop(StellenT stop)
/*
- alle ICs von Wurzel ausgehend absteigend probieren, excl. stop
- FALSE: aktuelleStelle = letzterIC oberhalb von stop bzw. MaxStelle, falls keine ICs ex.
- TRUE : aktuelleStelle = reduzierter IC = letzteRed
*/
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && NFB_GleichungAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN GAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop' oder start, falls start=stop
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_GleichungAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
      PfadEinsAb();
   }
   return FALSE;
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* ++++ Kommentare sind analog zu denjenigen in den entsprechenden Prozeduren NormalformZuRegeln***  ++++                  */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFO_NormalformZuGleichungenAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFO_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   PfadInit(Term, objekt);
   while (NFB_GleichungAngewendet());
   IC[aktuelleStelle].IC = NFB_IC();
   weiter =  WeiterInVO();

   while(weiter)
      if (NFB_GleichungAngewendet()){
	 letzteRed = aktuelleStelle;
         while(GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuE(letzteRed) */
            letzteRed = aktuelleStelle;
         while(GICsAufsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
            letzteRed = aktuelleStelle;
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            while(GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuE(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
	 weiter =  WeiterInVO();
      }
   return PfadWurdeReduziert();
}


BOOLEAN NFO_NormalformZuGleichungenAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   PfadInit(Term, objekt);

   while (NFB_GleichungAngewendet());			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */
   IC[aktuelleStelle].IC = NFB_IC();

   while(WeiterInVO()){
      if (NFB_GleichungAngewendet()){			    /* Reduktionserfolg => Stelle merken */
	 while(NFB_GleichungAngewendet());			    /*           Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (GICsAufsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while (GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed))){
               while(NFB_GleichungAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }

   return PfadWurdeReduziert();
}

BOOLEAN NFO_NormalformZuGleichungenAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.*/

{
   PfadInit(Term, objekt);
         
   do{
      if (NFB_GleichungAngewendet()){ 
	 while(NFB_GleichungAngewendet());			    /* Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(aktuelleStelle))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (GAlleAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed), letzteRed)){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
               while(NFB_GleichungAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }while(WeiterInVO());
   return PfadWurdeReduziert();
   
}

BOOLEAN NFO_NormalformZuGleichungenAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   PfadInit(Term, objekt);

   do{
      if (NFB_GleichungAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed)))
            letzteRed = aktuelleStelle;
         while (GAlleAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return PfadWurdeReduziert();
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Regeln ---------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_RegelAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }
   return FALSE;
}

static BOOLEAN RICsAufsteigenBisReduziertOderStop(StellenT start)
/*
- alle ICs excl. start aufsteigen und probieren
- FALSE: aktuelleStelle = hoechsterIC
- TRUE : aktuelleStelle = letzter reduzierter IC = letzteRed
*/
{
   aktuelleStelle = start;
   while(aktuelleStelle > 0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && NFB_RegelAngewendet())
         return TRUE;
   }
   return FALSE;
}
            
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RICsAbsteigenBisReduziertOderStop(StellenT stop)
/*
- alle ICs von Wurzel ausgehend absteigend probieren, excl. stop
- FALSE: aktuelleStelle = letzterIC oberhalb von stop bzw. MaxStelle, falls keine ICs ex.
- TRUE : aktuelleStelle = reduzierter IC = letzteRed
- ICs bleiben ICs, auch nach Aenedrungen weiter unten im Term
*/
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && NFB_RegelAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN RAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop'
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_RegelAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
      PfadEinsAb();
   }
   return FALSE;
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln TEILTERM --------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren arbeiten analog den normalen Normalform-Prozeduren zum Regelsystem weiter unten.               */
/* Sie kommen zur Anwendung, wenn ein Term, der vorher in Normalform bezueglich R war, an einer Stelle (hoechsteRed) ver-  */
/* aendert wurde. Dieser Term wird dann wieder in Normalform bzgl. R gebracht, wobei der Termdurchlauf auf den veraenderten*/
/* Bereich des Termes beschraenkt bleibt (ggf. wird mit Backtracking ein groesserer Bereich abgedeckt).                    */
/* Dabei muss das Backtracking bis zur Wurzel des Termes laufen, da indirekte Clashes zu R nicht vermerkt wurden.          */
/* Alle vier Prozeduren sind im Prinzip wie folgt strukturiert:                                                            */
/* - Aufstieg im Term, um moeglicherweise reduzierbare Vorgaenger zu finden (variantenabhaengig)           (1. Schleife)   */
/* - 'normaler' Normalform-Algorithmus, der aber nur den Teilterm bearbeitet und ggf. hoechsteRed anpasst. (2. Schleife)   */
/*                                                                                                                         */
/* Vorbedingung : aktStelle = hoechsteRed (Dies ist die veraenderte Stelle)                                                */
/* Aenderungen waehrend des Ablaufes: falls beim Aufsteigen ueber hoechsteRed hinaus eine Reduktion stattfand, wird deren  */
/*                                    Stelle als hoechsteRed gespeichert.                                                  */
/* Nachbedingung: der uebergebene Term ist in Normalform zu R, die hoechste veraenderte Stelle (also entweder diejenige,   */
/*                deren Reduktion den Prozeduraufruf ausgeloest hat, oder eine waehrend des Proz.ablaufs reduzierte, im    */
/*                Term hoehergelegene Stelle) ist in hoechsteRed gespeichert; aktStelle=hoechsteRed.                       */
/*                                                                                                                         */
/* Verifikation der Nachbedingung:                                                                                         */
/*  letztes ausgefuehrtes Statement, bevor die hintere do-while-Schleife verlassen wird, ist SetzePfadAufNachfInVOTeilterm,*/
/*  und zwar mit dem Resultat FALSE. => Gemaess Nachbedingung dieser Prozedur ist aktStelle = Gipfel akt. Teilterm =       */
/*  hoechsteRed.                                                                                                           */
/*                                                                                                                         */
/* Nur die erste der vier Prozeduren ist ausfuehrlich kommentiert, da die uebrigen sich nur im Detail von dieser unter-    */
/* scheiden. Bei diesen sind dann die sich unterscheidenden Stellen kommentiert.                                           */
/*                                                                                                                         */
/* Aufstieg im Term (1.Schleife):                                                                                          */
/*                                                                                                                         */
/* - Abstieg:                                                                                                              */
/*    * Von Wurzel absteigen bis hoechsteRed                                                                               */
/*    * Dabei jede Stelle auf Reduzierbarkeit pruefen                                                                      */
/*    * Nach Reduktion: +hoechsteRed anpassen (die reduzierte Stelle kann nicht unterhalb von hoechsteRed liegen)          */
/*                      +Erneut ganz oben beginnen                                                                         */
/*    * Nach Erfolglosen Reduktionsversuchen das IC-Flag anpassen                                                          */
/*                                                                                                                         */
/* - Aufstieg:                                                                                                             */
/*    * Aufstieg von hoechsteRed bis zur Wurzel                                                                            */
/*    * Dabei jede Stelle auf Reduzierbarkeit pruefen                                                                      */
/*    * Nach erfolglosen Versuchen: IC-Flag anpassen                                                                       */
/*    * Nach Reduktion: hoechsteRed auf aktuelleStelle setzen, weiter aufsteigen                                           */
/*                                                                                                                         */
/* Unterschiede zwischen Nochmal und Direkt:                                                                               */
/*    * In der Nochmal-Fassung wird nach Erfolgen zunaechst fertigreduziert, dann das IC-Flag besetzt,                     */
/*      in der Direkt-Fassung wird die betroffenen Stelle erst spaeter wieder angefasst, das IC-Flag erhaelt keinen Wert.  */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

static void NormalformZuRegelnAufDirektTeilterm(void)
{
   BOOLEAN nochmal;

   AufIC2Schalten();

   while (RAlleAufsteigenBisReduziertOderStop(aktuelleStelle, 0))
      hoechsteRed = aktuelleStelle;

   aktuelleStelle = hoechsteRed;

   do{
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
            letzteRed = aktuelleStelle;
         while(RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            letzteRed = aktuelleStelle;
            while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         hoechsteRedAnpassen(letzteRed);
         aktuelleStelle = letzteRed;
      }
      else{
         nochmal = WeiterInVOTeilterm();
         IC[aktuelleStelle].IC=NFB_IC();
      }
   }while(nochmal);
   VonIC2Zurueckschalten();
}


static void NormalformZuRegelnAufNochmalTeilterm(void)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   AufIC2Schalten();

   while(NFB_RegelAngewendet());
   IC[aktuelleStelle].IC=NFB_IC();

   while(RAlleAufsteigenBisReduziertOderStop(aktuelleStelle, 0)){
      while(NFB_RegelAngewendet());
      IC[aktuelleStelle].IC = NFB_IC();
      hoechsteRed = aktuelleStelle;
   }

   aktuelleStelle = hoechsteRed;

   while(WeiterInVOTeilterm()){
      if (NFB_RegelAngewendet()){
	 while(NFB_RegelAngewendet());
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;

         while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;
         hoechsteRedAnpassen(letzteRed);
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }
   VonIC2Zurueckschalten();
}


static void NormalformZuRegelnAbDirektTeilterm(void)
/* nach Reduktion direkt Abstieg von Wurzel (wegen indirekter Clashes keine Begrenzung moeglich). */
{
   BOOLEAN nochmal;

   AufIC2Schalten();
   
   while(RAlleAbsteigenBisReduziertOderStop(0, aktuelleStelle))
      hoechsteRed = aktuelleStelle;
   
   aktuelleStelle = hoechsteRed;

   do{
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
            letzteRed = aktuelleStelle;
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;
         hoechsteRedAnpassen(letzteRed);
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVOTeilterm();
      }
   }while(nochmal);
   VonIC2Zurueckschalten();
}

static void NormalformZuRegelnAbNochmalTeilterm(void)
/* nach Reduktion aktuelle Stelle nochmal versuchen, Abstieg.*/
{
   AufIC2Schalten();

   while(NFB_RegelAngewendet());
   IC[aktuelleStelle].IC = NFB_IC();

   while (RAlleAbsteigenBisReduziertOderStop(0, aktuelleStelle)){
      while(NFB_RegelAngewendet());
      IC[aktuelleStelle].IC = NFB_IC();
      hoechsteRed = aktuelleStelle;
   }

   aktuelleStelle = hoechsteRed;

   while(WeiterInVOTeilterm()){
      if (NFB_RegelAngewendet()){ 
	 while(NFB_RegelAngewendet());			    /* Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(aktuelleStelle))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
         hoechsteRedAnpassen(letzteRed);
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }
   VonIC2Zurueckschalten();
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* Allen Prozeduren gemeinsam ist, da\3 der Term zunaechst auf Toplevel reduziert wird, bis dort keine Reduktionen mehr    */
/* moeglich sind. Anschlie\3end wird der Durchlauf in Vorordnung durchgefuehrt. Die vorgezogene NF-Bildung auf Toplevel hat*/
/* den Vorteil, da\3 hierbei keinerlei Backtracking-Overhead erzeugt werden mu\3, wie das innerhalb der 'normalen' Reduk-  */
/* tionsschleife der Fall ist. Zur Begruendung dient auch folgende Tabelle (Resultate aus ca. 20 Laeufen je Variante+Bsp.)

                 vorher while             innen auch                weder noch
                                                                                                
-------------------------------------------------------------------------------------------------------------------------  
                                                                                               
ackermannViel
                 121.68   0.11913        123.268  0.0995833          121.509   0.105833                                     
                 121.595  0.106842       122.725  0.103              121.314   0.097                                                   
                 118.940 0.080           120.230 0.070               118.580 0.060                                                     
                                                                                                
                                                                                                
                                                                                                
braid4
                20.2867  0.40375         20.4683   0.374167          20.2325   0.39375                                                     
                20.2535  0.383           20.4265   0.371             20.209    0.3795                                               
                19.760 0.530             19.940 0.420                19.830 0.420                                                     
                                                                                                
                                                                                                
                                                                                                
z22             46.6633   0.310417       47.2667  0.224167          46.96     0.242083                                                        
                46.46     0.3105         46.976   0.2255            46.6255   0.242
                46.070 0.250             46.540 0.180               46.120 0.260                                                    
                                                                                                
                                                                                                
-------------------------------------------------------------------------------------------------------------------------  
vorherwhile: Zusaetzliche while-Schleife zur Reduktion der Wurzel vor der eigentlichen Reduktionsschleife
innen auch : Auch in der eigentlichen Reduktionsschleife wird nach jeder Reduktion geprueft, ob man sich an der Wurzel
             befindet, und falls ja, diese zuerst in einer weiteren Schleife kleinreduziert.
weder noch : 'normales' Verfahren, dh. ohne zusaetzliche while-Schleifen.
-------------------------------------------------------------------------------------------------------------------------  

1.Zeile  : jeweils Durchschnitte komplett
2. Zeile : ohne die 2 groessten und die zwei kleinsten Werte
3. Zeile : bester erreichter Wert

ERGEBNIS : 

- Durchschnittswerte : 2xwedernoch, 1xvorherwhile(z22)
- DurchschnittswerteII: 2xwedernoch, 1xvorherwhile(z22)
- Spitzenwerte : 2xvorherwhile, 1xwedernoch(braid4)


Weitere Ueberlegung:
--------------------
Innerhalb der Reduktionsschleifen lassen sich mit der Abfrage, ob man sich an der Wurzel befindet, im Erfolgsfall Ops einsparen.
Dies saehe folgendermassen aus:

if(NichtWurzel()){                        <--- hier wird mit Abfrage abgebrochen
   letzteRed = aktuelleStelle;
   aktStelleSei(neuerGipfel());
   while(aktIstNichtLetzteRed()){         <--- hier ohne
      if(NFB_RegelAngewendet()){
	  letzteRed = aktuelleStelle;
	  while(NFB_RegelAngewendet());
	  aktStelleSei(neuerGipfel());
      }
      else{
	  PfadEinsAb();
      }
   }
}

Daraus ergibt sich folgendes:

- mit Abfrage:

Stelle = Wurzel: 1 Op (Abfrage)                           = 1 Op
Stelle # Wurzel: 1 Op (Abfrage) + 2 Zuweisungen + Abfrage = 4 Ops

- ohne Abfrage:

immer 2 Zuweisungen + 1 Abfrage = 3 Ops

Soll die Abfrage Vorteile bringen, ergibt sich (sei w = P(Stelle = Wurzel)):

    w * (1 Abfrage) + (1-w)*(2 Abfrage + 2 Zuweisungen) < 2 Zuweisungen + 1 Abfrage
<=> w*A + 2A + 2Z - w*2A - w*2Z                         < 2Z + 1A
<=> A - w*A - 2w*Z                                      < 0
<=> A                                                   < w*(A + 2Z)

Unter der Annahme, dass eine Zuweisung doppelten Aufwand zu einer Abfrage erfordert, ergibt sich:
w > 20 %.

==> Es muesste in mehr als 20 % der Faelle die behandelte Stelle tatsaechlich die Wurzel sein, was m.E. utopisch ist.
---------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFO_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFO_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   PfadInit(Term, objekt);
   while (NFB_RegelAngewendet());
   IC[aktuelleStelle].IC = NFB_IC();
   weiter =  WeiterInVO();

   while(weiter)
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuR(letzteRed) */
            letzteRed = aktuelleStelle;
         while(RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            letzteRed = aktuelleStelle;
            while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuR(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
	 weiter =  WeiterInVO();
      }
   return PfadWurdeReduziert();
}

BOOLEAN NFO_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   PfadInit(Term, objekt);

   while (NFB_RegelAngewendet());			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */
   IC[aktuelleStelle].IC = NFB_IC();

   while(WeiterInVO()){
      if (NFB_RegelAngewendet()){			    /* Reduktionserfolg => Stelle merken */
	 while(NFB_RegelAngewendet());			    /*           Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }

   return PfadWurdeReduziert();
}

BOOLEAN NFO_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.
   Das Flag 'Aufstieg' bedeutet in dieser Prozedur eigentlich Abstieg!*/
{
   PfadInit(Term, objekt);
         
   do{
      if (NFB_RegelAngewendet()){ 
	 while(NFB_RegelAngewendet());			    /* Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(aktuelleStelle))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }while(WeiterInVO());
   return PfadWurdeReduziert();
   
}

BOOLEAN NFO_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   PfadInit(Term, objekt);

   do{
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
            letzteRed = aktuelleStelle;
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return PfadWurdeReduziert();
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regel- und Gleichungssystems auf Normalform. Dazu wird zunaechst eine  */
/* 'abstrakte' Variable Pfad initialisiert, die dann gemaess einer -aenderbaren- Strategie das Durchlaufen des Termes      */
/* organisiert, wobei das weitere Vorgehen in Abhaengigkeit vom Erfolg der letzten versuchten Reduktion sein kann.         */
/* Dies wird solange fortgefuehrt, bis es keine reduzierbaren Stellen mehr im Anfrageterm gibt.                            */
/* Es wird erst dann versucht, Gleichungen anzuwenden, wenn es keine anwendbaren Regeln mehr gibt.                         */
/* In diesem Fall wird versucht, eine anwendbare Gleichung zu finden. Gelingt dies, so wird die Gleichung angewendet, und  */
/* anschliessend wird an dieser Stelle mit Regeln fortgefahren. Dies wiederholt sich, bis es weder anwendbare Regeln noch  */
/* Gleichungen gibt.                                                                                                       */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFO_NormalformAufNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN reduziert, nochmal;
   
   reduziert = NFO_NormalformZuRegelnAufNochmal(Term, objekt);
   
   PfadInit(Term, objekt);
   
   do{
      if (NFB_GleichungAngewendet()){
         reduziert = TRUE;
         do{
            hoechsteRed = aktuelleStelle;
            NormalformZuRegelnAufNochmalTeilterm();
            aktStelleVonVorne();
         }while(GAlleAufsteigenBisReduziertOderStop(hoechsteRed, GipfelZuE(hoechsteRed)) || 
                GICsAufsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed)));
         aktuelleStelle = hoechsteRed;
         nochmal = TRUE;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return reduziert;                                                              /* klar */
}

BOOLEAN NFO_NormalformAufDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN reduziert, nochmal;

   reduziert = NFO_NormalformZuRegelnAufDirekt(Term, objekt);
   PfadInit(Term, objekt);

   do{
      if (NFB_GleichungAngewendet()){
         reduziert = TRUE;
         do{
            hoechsteRed = aktuelleStelle;
            NormalformZuRegelnAufDirektTeilterm();
            aktStelleVonVorne();
         }while(GAlleAufsteigenBisReduziertOderStop(hoechsteRed, GipfelZuE(hoechsteRed)) || 
                GICsAufsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed)));
         aktuelleStelle = hoechsteRed;
         nochmal = TRUE;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return reduziert;                                                              /* klar */
}

BOOLEAN NFO_NormalformAbDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN reduziert, nochmal;

   reduziert = NFO_NormalformZuRegelnAbDirekt(Term, objekt);
   PfadInit(Term, objekt);
   do{
      if (NFB_GleichungAngewendet()){
         reduziert = TRUE;
         do{
            hoechsteRed = aktuelleStelle;
            NormalformZuRegelnAbDirektTeilterm();
            aktStelleVonVorne();
         }while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed)) ||
                GAlleAbsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed), hoechsteRed));
         aktuelleStelle = hoechsteRed;
         nochmal = TRUE;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return reduziert;                                                              /* klar */
}


BOOLEAN NFO_NormalformAbNochmal (TermT Term, ObjektT objekt)
{
   BOOLEAN reduziert, nochmal;

   reduziert = NFO_NormalformZuRegelnAbNochmal(Term, objekt);
   PfadInit(Term, objekt);
   do{
      if (NFB_GleichungAngewendet()){
         reduziert = TRUE;
         do{
            hoechsteRed = aktuelleStelle;
            NormalformZuRegelnAbNochmalTeilterm();
            aktStelleVonVorne();
         }while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed)) ||
                GAlleAbsteigenBisReduziertOderStop(GipfelZuE(hoechsteRed), hoechsteRed));
         aktuelleStelle = hoechsteRed;
         nochmal = TRUE;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   return reduziert;                                                              /* klar */
}

