/**********************************************************************
* Filename : Ordnungen.c
*
* Kurzbeschreibung : Ordnungen
*
* Autoren : Andreas Jaeger, Arnim Buch
*
* 14.12.96: Erstellung
* 25.02.97: Vergleichs- und Groessertest-Funktionen aus Hauptkomponenten
*           hierher verlagert. Arnim.
*
*
* Benoetigt folgende Module:
*
* Benutzte Globale Variablen aus anderen Modulen:
*
**********************************************************************/

#include "Ausgaben.h"
#include "compiler-extra.h"
#include <ctype.h>
#include "general.h"
#include "parser.h"
#include "FehlerBehandlung.h"
#include "InfoStrings.h"
#include "KBO.h"
#include "LPO.h"
#include "LPOVortests.h"
#include "Ordnungen.h"
#include "PhilMarlow.h"
#include "Praezedenzgenerator.h"
#include <string.h>
#include "Statistiken.h"
#include "SymbolOperationen.h"



/* -------------------------------------------------------------------------------*/
/* --------------- Term-Vergleich (ORD-1) ----------------------------------------*/
/* -------------------------------------------------------------------------------*/

typedef VergleichsT (*ORDVergleichsT)  (UTermT, UTermT);
typedef BOOLEAN     (*ORDTestT)  (UTermT, UTermT);

static ORDVergleichsT ORD_Vergleichsfkt;
static ORDTestT       ORD_UnifTestfkt, ORD_RedTestfkt;



/* -------------------------------------------------------------------------------*/
/* --------------- Spezialordnungen ----------------------------------------------*/
/* -------------------------------------------------------------------------------*/

#define /*BOOLEAN*/ Vergleich2Test(s,t, eq, gt) \
  (eq(s,t) ? Gleich : gt(s,t) ? Groesser : gt(t,s) ? Kleiner : Unvergleichbar)
  /* zu selten / speziell fuer Plazierung in general.h */

static BOOLEAN LeererTest(UTermT s MAYBE_UNUSEDPARAM, UTermT t MAYBE_UNUSEDPARAM)
{ return FALSE; }

static VergleichsT LeererVergleich(UTermT s, UTermT t)
{ return TO_TermeGleich(s,t) ? Gleich : Unvergleichbar; }

/* Fassung fuer allquantifizierte Ziele.. */
static BOOLEAN Col3Test(UTermT s, UTermT t)
{ 
  if (TO_IstKombinatorTupel(s,t))
    return TRUE;
  else {
    SymbolT c = TO_TopSymbol(s);
    if (SO_SymbIstKonst(c))
      if ( (  !PM_Existenzziele()
	    || (SO_SymbUngleich(c,SO_TRUE) && SO_SymbUngleich(c,SO_FALSE)))
	  && TO_TermIstGrund(t))
	return TRUE;
    return FALSE;
  }
}

static VergleichsT Col3Vergleich(UTermT s, UTermT t)
{ return Vergleich2Test(s,t, TO_TermeGleich, Col3Test); }


static BOOLEAN Col4Test(UTermT s_, UTermT t_)
{ 
  if (PM_Existenzziele()) {
    SymbolT s = TO_TopSymbol(s_),
            t = TO_TopSymbol(t_);
    return 
      (s==SO_EQ && (t==SO_TRUE || t==SO_FALSE)) ||
      (s==SO_TRUE && t==SO_FALSE);
  }
  else
    return FALSE;
}

static VergleichsT Col4Vergleich(UTermT s, UTermT t)
{ return Vergleich2Test(s,t, TO_TermeGleich, Col4Test); }



/* -------------------------------------------------------------------------------*/
/* --------------- Ordnungsserver ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

#if 0
#define ORD_SERVER_SOCK
#elif 0
#define ORD_SERVER_PTY
#endif

#if defined(ORD_SERVER_SOCK) || defined(ORD_SERVER_PTY)
static int sockfd;
#define SERVER "/tmp/ordsocket"
#endif

#ifdef ORD_SERVER_SOCK
#include <stddef.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>

static int init_socket(){
  int sock;
  struct sockaddr_un name;
  int size;

  sock = socket (PF_UNIX, SOCK_STREAM, 0);     /* Create the socket.  */
  if (sock < 0) {perror ("socket"); exit (EXIT_FAILURE);}

  name.sun_family = AF_UNIX;
  strcpy (name.sun_path, SERVER);
  size = strlen (name.sun_path) + sizeof (name.sun_family);

  connect(sock, &name, size);
  return sock;
}

static int ordwrite (char *buf) {
  return write (sockfd, buf, strlen(buf));
}

static int ordread (char *buf, int len){
  int r = read (sockfd, buf, len);
  buf[r] = '\0';
  return r;
}

#endif

#ifdef ORD_SERVER_PTY

static int ptyfd;
static char *pty_argv[]={
"/home/loechner/haskell/bpg/ordserver",
/* "-kbo6", "+RTS", "-H1M", */
NULL};

#include <sys/types.h>
#include <termios.h>
#ifndef TIOCGWINSZ
#include <sys/ioctl.h>   /* 44BSD requires this too */
#endif
#include <stdio.h>               /* for convenience */
#include <stdlib.h>              /* for convenience */
#include <string.h>              /* for convenience */
#include <unistd.h>              /* for convenience */

#if 0
/* SUN ptyopen.c */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
/*
#include "ourhdr.h"
*/
static int 
ptym_open(char *pts_name)
{
  int             ptyfd;
  char    *ptr1, *ptr2;

  strcpy(pts_name, "/dev/ptyXY");
  /* array index: 0123456789 (for references in following code) */
  for (ptr1 = "pqrstuvwxyzPQRST"; *ptr1 != 0; ptr1++) {
    pts_name[8] = *ptr1;
    for (ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++) {
      pts_name[9] = *ptr2;

      /* try to open master */
      if ( (ptyfd = open(pts_name, O_RDWR)) < 0) {
	if (errno == ENOENT)                           /* different from EIO */
	  return(-1);                                  /* out of pty devices */
	else
	  continue;                                   /* try next pty device */
      }

      pts_name[5] = 't';                            /* change "pty" to "tty" */
      return(ptyfd);                            /* got it, return fd of master */
    }
  }
  return(-1);                                          /* out of pty devices */
}

static int
ptys_open(int ptyfd, char *pts_name)
{
  struct group    *grptr;
  int                             gid, fds;

  if ( (grptr = getgrnam("tty")) != NULL)
    gid = grptr->gr_gid;
  else
    gid = -1;               /* group tty is not in the group file */

  /* following two functions don't work unless we're root */
  chown(pts_name, getuid(), gid);
  chmod(pts_name, S_IRUSR | S_IWUSR | S_IWGRP);

  if ( (fds = open(pts_name, O_RDWR)) < 0) {
    close(ptyfd);
    return(-1);
  }
  return(fds);
}

#else

/* SVR4 ptyopen */
#include        <sys/types.h>
#include        <sys/stat.h>
#include        <errno.h>
#include        <fcntl.h>
#include        <stropts.h>
/*
#include        "ourhdr.h"
*/
extern char     *ptsname(int);  /* prototype not in any system header */

static int
ptym_open(char *pts_name)
{
  char    *ptr;
  int             ptyfd;

  strcpy(pts_name, "/dev/ptmx");  /* in case open fails */
  if ( (ptyfd = open(pts_name, O_RDWR)) < 0)
    return(-1);

  if (grantpt(ptyfd) < 0) {         /* grant access to slave */
    close(ptyfd);
    return(-2);
  }
  if (unlockpt(ptyfd) < 0) {        /* clear slave's lock flag */
    close(ptyfd);
    return(-3);
  }
  if ( (ptr = ptsname(ptyfd)) == NULL) {    /* get slave's name */
    close(ptyfd);
    return(-4);
  }

  strcpy(pts_name, ptr);  /* return name of slave */
  return(ptyfd);                    /* return fd of master */
}
static int
ptys_open(int ptyfd, char *pts_name)
{
  int             fds;

  /* following should allocate controlling terminal */
  if ( (fds = open(pts_name, O_RDWR)) < 0) {
    close(ptyfd);
    return(-5);
  }
  if (ioctl(fds, I_PUSH, "ptem") < 0) {
    close(ptyfd);
    close(fds);
    return(-6);
  }
  if (ioctl(fds, I_PUSH, "ldterm") < 0) {
    close(ptyfd);
    close(fds);
    return(-7);
  }
  if (ioctl(fds, I_PUSH, "ttcompat") < 0) {
    close(ptyfd);
    close(fds);
    return(-8);
  }

  return(fds);
}

#endif

/* ptyfork */
#include        <sys/types.h>
#include        <termios.h>
#ifndef TIOCGWINSZ
#include        <sys/ioctl.h>   /* 44BSD requires this too */
#endif

static pid_t
pty_fork(int *ptrfdm, char *slave_name,
	 const struct termios *slave_termios,
	 const struct winsize *slave_winsize)
{
  int             ptyfd, fds;
  pid_t   pid;
  char    pts_name[20];

  if ( (ptyfd = ptym_open(pts_name)) < 0)
    fprintf(stderr, "can't open master pty: %s", pts_name);

  if (slave_name != NULL)
    strcpy(slave_name, pts_name);   /* return name of slave */

  if ( (pid = fork()) < 0)
    return(-1);

  else if (pid == 0) {            /* child */
    if (setsid() < 0)
      fprintf(stderr, "setsid error\n");

    /* SVR4 acquires controlling terminal on open() */
    if ( (fds = ptys_open(ptyfd, pts_name)) < 0)
      fprintf(stderr, "can't open slave pty\n");
    close(ptyfd);             /* all done with master in child */

#if     defined(TIOCSCTTY) && !defined(CIBAUD)
    /* 44BSD way to acquire controlling terminal */
    /* !CIBAUD to avoid doing this under SunOS */
    if (ioctl(fds, TIOCSCTTY, (char *) 0) < 0)
      fprintf(stderr, "TIOCSCTTY error\n");
#endif
    /* set slave's termios and window size */
    if (slave_termios != NULL) {
      if (tcsetattr(fds, TCSANOW, slave_termios) < 0)
	fprintf(stderr, "tcsetattr error on slave pty\n");
    }
    if (slave_winsize != NULL) {
      if (ioctl(fds, TIOCSWINSZ, slave_winsize) < 0)
	fprintf(stderr, "TIOCSWINSZ error on slave pty\n");
    }
    /* slave becomes stdin/stdout/stderr of child */
    if (dup2(fds, STDIN_FILENO) != STDIN_FILENO)
      fprintf(stderr, "dup2 error to stdin\n");
    if (dup2(fds, STDOUT_FILENO) != STDOUT_FILENO)
      fprintf(stderr, "dup2 error to stdout\n");
    if (dup2(fds, STDERR_FILENO) != STDERR_FILENO)
      fprintf(stderr, "dup2 error to stderr\n");
    if (fds > STDERR_FILENO)
      close(fds);
    return(0);              /* child returns 0 just like fork() */

  } else {                                        /* parent */
    *ptrfdm = ptyfd;  /* return fd of master */
    return(pid);    /* parent returns pid of child */
  }
}


static int do_pty_fork (char *argv[])
{
  int   ptyfd;
  pid_t pid;
  struct termios  stermios;

  pid = pty_fork(&ptyfd, NULL, NULL, NULL);

  if (pid < 0){
    fprintf(stderr, "fork error");
  }
  else if (pid == 0) {            /* child */

    /* turn off echo (for slave pty) */
    if (tcgetattr(STDIN_FILENO, &stermios) < 0)
      fprintf(stderr, "tcgetattr error\n");
    stermios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
    stermios.c_oflag &= ~(ONLCR);
    /* also turn off NL to CR/NL mapping on output */
    if (tcsetattr(STDIN_FILENO, TCSANOW, &stermios) < 0)
      fprintf(stderr, "tcsetattr error\n");

    if (execvp(argv[0], argv) < 0)
      fprintf(stderr, "can't execute: %s", argv[0]);
  }

  /* parent */
  return ptyfd;
}

int ordwrite (char *buf){
  char *end;
  char lbuf [LBUFS];
  
  lbuf [ 0] = 'C';
  lbuf [LBUFS-2] = '\n';
  lbuf [LBUFS-1] = '\0';
  end = buf;
  while (end[0] != '\0'){
    end++;
  }
  while (end - buf > LBUFS - 3){
    strncpy(&lbuf[1], buf, LBUFS - 3);
    /*    printf("#%s#\n", lbuf); */
    write (ptyfd, lbuf, strlen(lbuf));
    buf += LBUFS -3;
  }
  /*  printf("###%s###\n", buf);*/
  write(ptyfd, buf, strlen(buf));
  return 0;
}

static int ordread (char *buf, int len){
  int r = read (ptyfd, buf, len);
  buf[r] = '\0';
  return r;
}
#endif

#if defined(ORD_SERVER_SOCK) || defined (ORD_SERVER_PTY)

char *ORD_BerndGedruckt(TermT links, char *res, TermT rechts);
/* Druckfunktion zur Unterstuetzung der Server-Anflanschung; wird mit Praeprozessorflag aus .c-Datei leer */

#define LBUFS 200
  
#endif

/* ORD-1 */
/*-------*/

VergleichsT ORD_VergleicheTerme(UTermT term1, UTermT term2)
/* Vergleicht die beiden angegebenen Terme nach der eingestellten Ordnung. */
{
#define INLEN 101

  VergleichsT result;

#if defined(ORD_SERVER_SOCK) || defined (ORD_SERVER_PTY)

  char *buf;
  int r;
  char in[INLEN];
  static char *vglt[] = {" > ", " < ", " = ", " # "};

#if 0
  buf = ORD_BerndGedruckt(term1, vglt[result], term2);

  printf("### trying %s", buf);                fflush(stdout);
  r = write(ptyfd, buf, strlen(buf));
  printf("### res = %d\n", r);                 fflush(stdout);
  printf("### reading...\n");                  fflush(stdout);
  r = read(ptyfd, in, 100);
  in[r] = '\0';
  printf("### got %d chars: %s...\n", r, in);  fflush(stdout);
#endif

  buf = ORD_BerndGedruckt(term1, " = ", term2);
#if 1
  r = strlen (buf);
  if (r > 200)
    printf("###LARGE %d\n", r);
#endif

  r = ordwrite(buf);

tryagain:  
  r = ordread(in, INLEN-1);

  /*  printf("### %s", in);*/

  switch (in[0]){
  case '>': result = Groesser;                                 break;
  case '<': result = Kleiner;                                  break;
  case '~':
  case '=': 
    result = TO_TermeGleich(term1,term2)? Gleich : Unvergleichbar;  
                                                               break;
  case '#': result = Unvergleichbar;                           break;
  case '\a':
  case '\0':
  case '\n':     
    printf("##%s", buf);
    printf("RES[0] = %d\n", in[0]);
    printf("RES ***%s*** NEXT TRY\n",in); fflush(stdout); 
    goto tryagain;
  case 'E': 
  default: 
    printf("##%s", buf);
    printf("RES[0] = %d\n", in[0]);
    printf("RES ***%s***BAILING OUT\n",in); fflush(stdout); 
    do {
      r = ordread(in, INLEN-1);
      printf("RES ***%s***\n",in); fflush(stdout); 
    }
    while (r>0);
    ordwrite("ERROR VON WM\n");
    our_error("Error/Unbekannter Input von Ordnungsserver!");  break;
  }
#else
  /* ganz normaler Ordnungsaufruf ... */
  result = ((*ORD_Vergleichsfkt) (term1, term2));
#endif

  /* IO_DruckeFlex("## %t%s%t\n", term1, vglt[result], term2);*/
  /* IO_DruckeFlex("##%t%s%t  wird  %s.\n", term1, vglt[result], term2); */

  return result;
}

/* ORD-1a*/
/*-------*/
BOOLEAN ORD_TermGroesserRed(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
#if defined(ORD_SERVER_SOCK) || defined (ORD_SERVER_PTY)
  /*our_error("aha!");*/ /* binaerer Test! */
#endif

  /* EVTL SOCKETS ODER ABORT */
  BOOLEAN res = ((*ORD_RedTestfkt) (term1, term2));

  /* IO_DruckeFlex("##%s %t ? %t\n", res ? "T" : "F", term1, term2);*/


  return  res; /*((*ORD_RedTestfkt) (term1, term2));*/
  /* Ende der Zeitmessung erfolgt in MatchOPs:GleichungGefunden */
}

BOOLEAN ORD_TermGroesserUnif(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
#if defined(ORD_SERVER_SOCK) || defined (ORD_SERVER_PTY)
  /*our_error("soso!");*/ /* binaerer Test! */
#endif
  /* EVTL SOCKETS ODER ABORT */
  if ((*ORD_UnifTestfkt) (term1, term2)) {
    return TRUE;
  }
  else {
    return FALSE;
  }
}

void ORD_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    ordering ord = PA_Ordering;
    char *newprec;
    char *newperm;  
    fprintf(IO_Strom, "\n* ORDERING: "); 
    if (IS_FindOption (DEF, USR, 1, "kbo") == 0)  
      ord  = ordering_kbo; 
    else if (IS_FindOption (DEF, USR, 1, "lpo") == 0)                  
      ord  = ordering_lpo; 
    else if (IS_FindOption (DEF, USR, 1, "empty") == 0)                  
      ord  = ordering_empty; 
    else if (IS_FindOption (DEF, USR, 1, "col3") == 0)                  
      ord  = ordering_col3; 
    else if (IS_FindOption (DEF, USR, 1, "col4") == 0)                  
      ord  = ordering_col4; 
    switch (ord ) { 
    case ordering_kbo:   
      fprintf(IO_Strom, "KBO"); 
      break; 
    case ordering_lpo:   
      fprintf(IO_Strom, "LPO"); 
      break; 
    case ordering_empty:   
      fprintf(IO_Strom, "empty"); 
      break; 
    case ordering_col3:   
      fprintf(IO_Strom, "col3"); 
      break; 
    case ordering_col4:   
      fprintf(IO_Strom, "col4"); 
      break; 
    default:  
      ST_ErrorStatistik("Unknown Ordering",FALSE);
      break; 
    } 
    newprec =IS_ValueOfOption (DEF,USR,"prec"); 
    newperm=IS_ValueOfOption (DEF,USR,"perm"); 
    if (newprec) { 
      fprintf(IO_Strom, ", (new) precedence %s",newprec );
      if (!(SO_PraezedenzKorrekt(newprec)) )
        fprintf(IO_Strom, " invalid, leaving unchanged");
    } 
    else if (newperm) { 
      char *punkt; 
      int permno; 
      int permdegree= SO_AnzahlFunktionen ; 
      punkt=strchr(newperm,'.'); 
      if (punkt){ 
        *punkt='\0'; 
        permdegree = atoi(punkt+1); 
      } 
      permno = atoi(newperm); 
      if ((newprec =PG_Permutator(permno, permdegree)))
        fprintf(IO_Strom, ", permutation %u is %s with degree %d", permno, newprec , permdegree);  
      else
        fprintf(IO_Strom, ", permutation %u with degree %d invalid or precedence not total", permdegree, permno);  
    } 
    fprintf(IO_Strom, "\n            current precedence is "); 
    if (SO_PraezedenzTotal ) { 
      unsigned int i; 
      for(i=0; i< SO_AnzahlFunktionen ; i++) { 
        IO_DruckeSymbol(SO_TotaleKette[i]); 
        if (i<SO_AnzahlFunktionen-1) 
          fprintf(IO_Strom, " > "); 
      } 
    } 
    else 
      fprintf(IO_Strom, "not total"); 
    fprintf(IO_Strom, "\n"); 
  }
  else
    fprintf(IO_Strom, "-ord [kbo|lpo|empty]:prec=a.b.c..d.e.a  redefine ordering and precedence\n"
            "KBO only with weights=1; precedence above would be a>b>c;d>e>a\n");
}
  


/*=================================================================================================================*/
/*= N. Provisorium fuer Ordnungsserver                                                                            =*/
/*=================================================================================================================*/

#define ORD_PRINT 0
#if ORD_PRINT
#include "DSBaumOperationen.h"

#define ErstePufferlaenge 5

static char *Puffer,
            *p;
static unsigned int Pufferlaenge;
static unsigned int *Operatornummer;


static void OperatorenArrangieren()
{
  unsigned int i,
               Index = 0;
  SymbolT Symbol;
  SO_forFktSymbole(Symbol)
    Operatornummer[Symbol] = (unsigned)SO_ErsteFVar;
  for (i = SO_AnzahlFunktionen; i; i--) {
    SymbolT Symbol = SO_TotaleKette[i-1];
    if (Operatornummer[Symbol]==(unsigned)SO_ErsteFVar)
      Operatornummer[Symbol] = Index++;
  }
}


static void DruckInitialisieren(void)
{
  if (!SO_PraezedenzTotal)
    our_error("ORD-Druck ohne totale Praezedenz");
  Puffer = string_alloc(Pufferlaenge = ErstePufferlaenge);
  Operatornummer = array_alloc(SO_AnzahlFunktionen,unsigned int);
  OperatorenArrangieren();
}

static void GgfPufferErweitern(TermT links, TermT rechts)
{
  unsigned int Funktionsindex = SO_AnzahlFunktionen ? SO_AnzahlFunktionen-1 : 0,
               Variablenindex = BO_NeuesteBaumvariable()==SO_VorErstemVarSymb ? 0 : SO_VarNummer(2*BO_NeuesteBaumvariable()),
               Funktionslaenge = LaengeNatZahl(Funktionsindex)+1+2+1,                                      /* F(), */
               Variablenlaenge = LaengeNatZahl(Variablenindex)+1+1,                                          /* x, */
               Symbollaenge = max(Funktionslaenge,Variablenlaenge),
               Linkslaenge = TO_Termlaenge(links)*Symbollaenge,
               Rechtslaenge = TO_Termlaenge(rechts)*Symbollaenge,
               Pufferschranke = Linkslaenge+3+Rechtslaenge+100;                                            /*  =  \0 */
  if (Pufferschranke>Pufferlaenge) {
    Pufferlaenge = minKXgrY(ErstePufferlaenge,Pufferschranke);
    array_realloc(Puffer,Pufferlaenge,char);
  }
}

#define /*void*/ pp(/*char **, char **, ...*/ Argumente)                                                            \
  p += our_sprintf(Argumente)


static void SymbolDrucken(SymbolT Symbol)
{
  if (SO_SymbIstVar(Symbol))
    pp((p,"x%d",SO_VarNummer(Symbol)));
  else
    pp((p,"f%d",Operatornummer[Symbol]));
}

static void TermDrucken(UTermT Term)
{
  SymbolDrucken(TO_TopSymbol(Term));
  if (TO_NichtNullstellig(Term)) {
    UTermT NULLTerm = TO_NULLTerm(Term);
    pp((p,"("));
    TermDrucken(Term = TO_ErsterTeilterm(Term));
    while (TO_PhysischUngleich(Term = TO_NaechsterTeilterm(Term),NULLTerm)) {
      pp((p,","));
      TermDrucken(Term);
    }
    pp((p,")"));
  }
}

static char *ORD_BerndGedruckt(TermT links, char *res, TermT rechts)
{
  GgfPufferErweitern(links,rechts);
  p = Puffer;
  TermDrucken(links);
  pp((p,res));
  TermDrucken(rechts);
  pp((p,"\n"));
  return Puffer;
}

/*Einsatz: IO_DruckeFlex("%t = %t  wird  %s.\n", term1, term2, ORD_BerndGedruckt(term1,res,term2));*/

#else
static void DruckInitialisieren(void) {;}
#endif



/*=================================================================================================================*/
/*= N+1. Weiter im Text                                                                                           =*/
/*=================================================================================================================*/

static void GgfGewichteAendern(InfoStringT DEF, InfoStringT USR);

BOOLEAN ORD_InitAposteriori (InfoStringT DEF, InfoStringT USR) 
{
  ordering oldOrd = PA_Ordering;
  char *newprec;
  char *newperm;  
  DruckInitialisieren();

  fprintf(IO_Strom, "\n* ORDERING: ");
  if (IS_FindOption (DEF, USR, 1, "kbo") == 0)  
    PA_Ordering  = ordering_kbo;
  else if (IS_FindOption (DEF, USR, 1, "lpo") == 0)
    PA_Ordering  = ordering_lpo;
  else if (IS_FindOption (DEF, USR, 1, "empty") == 0)
    PA_Ordering  = ordering_empty;
  else if (IS_FindOption (DEF, USR, 1, "col3") == 0)
    PA_Ordering  = ordering_col3;
  else if (IS_FindOption (DEF, USR, 1, "col4") == 0)
    PA_Ordering  = ordering_col4;
  switch (PA_Ordering) { 
  case ordering_kbo:
    ORD_Vergleichsfkt = KB_KBOVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
    if (oldOrd != PA_Ordering)
      fprintf(IO_Strom, "changed to KBO");
    break;
  case ordering_lpo:   
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to LPO");
    ORD_Vergleichsfkt = LP_LPO;
    ORD_RedTestfkt =  ORD_UnifTestfkt = LP_LPOGroesser;
    LV_InitAposteriori();
    break;
  case ordering_empty:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to empty");
    ORD_Vergleichsfkt = LeererVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = LeererTest;
    break;
  case ordering_col3:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to col3");
    ORD_Vergleichsfkt = Col3Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col3Test;
    TO_KombinatorapparatAktivieren();
    break;
  case ordering_col4:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to col4");
    ORD_Vergleichsfkt = Col4Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col4Test;
    TO_KombinatorapparatAktivieren();
    break;
  default:
    ST_ErrorStatistik("Unknown Ordering",FALSE);
    break;
  } 

  GgfGewichteAendern(DEF,USR);
  if (PA_Ordering==ordering_kbo)
    KB_InitAposteriori();    

  newprec =IS_ValueOfOption (DEF,USR,"prec");
  newperm=IS_ValueOfOption (DEF,USR,"perm");
  if (newprec) {
    fprintf(IO_Strom, ", (new) precedence %s", newprec );
    if (!SO_PraezedenzErneuert(newprec))  
      our_error("Invalid Precedence");
  } 
  else if (newperm) { 
    char *punkt;
    int permno;
    int permdegree= SO_AnzahlFunktionen ;
    punkt=strchr(newperm,'.');
    if (punkt) { 
      *punkt='\0';
      permdegree = atoi(punkt+1);
    } 
    permno = atoi(newperm);
    if ((newprec =PG_Permutator(permno, permdegree))) { 
      fprintf(IO_Strom, ", permutation %u is %s with degree %d", permno, newprec , permdegree);
      if(!SO_PraezedenzErneuert(newprec)) 
        ST_ErrorStatistik("Permutator generated invalid precedence\n",FALSE);
    } 
    else { 
      fprintf(IO_Strom, ", permutation %u with degree %d invalid or precedence not total", permdegree, permno);
      ST_ErrorStatistik("Invalid permutation", FALSE);
    } 
  }  
  fprintf(IO_Strom, "\n            current precedence is ");
  if (SO_PraezedenzTotal) { 
    unsigned int i;
    for(i=0; i< SO_AnzahlFunktionen ; i++) { 
      IO_DruckeSymbol(SO_TotaleKette[i]);
      if (i<SO_AnzahlFunktionen-1) 
        fprintf(IO_Strom, " > ");
    } 
  } 
  else 
    fprintf(IO_Strom, "not total"); 
  fprintf(IO_Strom, "\n"); 


#ifdef ORD_SERVER_SOCK
  sockfd = init_socket(); 
#endif
#ifdef ORD_SERVER_PTY
  ptyfd = do_pty_fork(pty_argv);
#endif

  return TRUE;
}

void ORD_OrdnungUmstellen(ordering Ordnung)
{
  switch (Ordnung){
  case ordering_lpo:
    ORD_Vergleichsfkt = LP_LPO;
    ORD_RedTestfkt = ORD_UnifTestfkt = LP_LPOGroesser;
    LV_InitAposteriori();
    break;
  case ordering_kbo:
    ORD_Vergleichsfkt = KB_KBOVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
    KB_InitAposteriori();   /* Phi auf Gueltigkeit testen.*/
    break;
  case ordering_empty:
    ORD_Vergleichsfkt = LeererVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = LeererTest;
    break;
  case ordering_col3:
    ORD_Vergleichsfkt = Col3Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col3Test;
    break;
  case ordering_col4:
    ORD_Vergleichsfkt = Col4Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col4Test;
    break;
  default:
    ST_ErrorStatistik("Invalid Ordering in ORD_OrdnungUmstellen", FALSE);
    break;
  }
  PA_Ordering = Ordnung;
}


static void GgfGewichteAendern(InfoStringT DEF, InfoStringT USR)
{
  BOOLEAN DefaultAngegeben = IS_ValueOfOption(DEF, USR, "DEF")!=NULL,
          Geaendert = DefaultAngegeben;
  long int Default = DefaultAngegeben ? atoi(IS_ValueOfOption(DEF, USR, "DEF")) : 1,
           *Gewichte;
  SymbolT i;
  Gewichte = array_alloc(SO_AnzahlFunktionen,long int);
  SO_forFktSymbole(i) {
    char *wert = IS_ValueOfOption(DEF, USR, IO_FktSymbName(i));
    if (wert && (isdigit((int)*wert) || *wert=='-')) {
      Gewichte[i] = atoi(wert);
      Geaendert = TRUE;
    }
    else
      Gewichte[i] = -1;
  }
  if (Geaendert)
    SO_PhiUmstellen(Gewichte,DefaultAngegeben,Default);
  our_dealloc(Gewichte);
}
