/* -*-Mode: text;-*-
   lexer fuer Waldmeister
   Geschrieben von Andreas Jaeger
*/
%{
#include "parse-util.h"
#include "wm-parse.tab.h"
#include <stdio.h>
#include <string.h>
int yywrap (void)
{
  return 1;
}
int lineno = 1;
int column = 0;
void yyerror (char *s);
%}

LETTER	[A-Za-z]
DIGIT	[0-9]
%%

\%([^\n])*\n	{ lineno++; /* Comment until end of line */
		  column = 0;
		  return ENDOFLINE;
		}
{DIGIT}|{LETTER}({LETTER}|{DIGIT})* {
                  yylval.ident = strdup (yytext);
		  column += yyleng;
		  return SYMB;
                }
"+"|"*"|"/"|"-"	{ yylval.ident = strdup (yytext);
		  column += yyleng;
		  return OP;
                }
"("|")"|","	{ 
		  column += yyleng;
		  return *yytext;
                }
"="		{
		  column += yyleng;
                  return EQ;
                }

[ \t]+		{ /* ignore space/tabs */
		  column += yyleng;
		}
\n		{ lineno++;
		  column = 0;
       		  return ENDOFLINE;
                }
.		{ fprintf (stderr, "Error in lexer: %c ignored\n", *yytext);}


%%
void
yyerror (char *s)
{
  fprintf (stderr, "Error: %s in line %d, column %d: at `%s'\n", 
	   s, lineno, column, yytext);
}
