#include "Achtzigstel.h"
#include "antiwarnings.h"
#include "compiler-extra.h"
#include <stdio.h>
#include <string.h>
#include "Ausgaben.h"
#include "NewClassification.h"
#include "general.h"
#include "HeapOfHeaps.h"
#include "InfoStrings.h"
#include "ParseInterleave.h"
#include "PQStatistik.h"
#include "RUndEVerwaltung.h"
#include "SUEBasis.h"
#include "SUEM.h"
#include "SpeicherVerwaltung.h"
#include "StrahlMemorium.h"
#include "Termpaare.h"
#include "TermpairRepresentation.h"

#if defined SV_BASIS_MMAP
/* Dieses Modul funktioniert nicht mit malloc.  */

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe HeapOfHeaps.h) -----------------------*/
/* ------------------------------------------------------------------------*/
/* keine mehr! */

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Data-Structures --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

typedef struct SHTS SubHeapZellenT;

typedef struct InfoTS{        /* Info-Block mit Statistik */
   long int ElternNummer;
   unsigned long AnzNochNichtBehandelterKPs;
   RegelOderGleichungsT re;
   BOOLEAN LebtNoch;
   SubHeapZellenT *subheap;
   unsigned long NaechstesFreies;
#if PQ_STATISTIK
  unsigned long AnzInitialerKPs;
#endif
  BOOLEAN InTopHeap;
} InfoT;

typedef struct{
   InfoT *info;
   WeightT w1;                 /* Die Gewichte des Top-Elements, um eine Indirektion zu sparen */
   WeightT w2;
} TopEntryT;


struct SHTS{                  /* Eintrag im SubHeap: 2 long als Gewicht */
   InfoT *OtherParent;
   TermpairT tp;
   WeightT w1, w2;
};



/* --------------------------------------------------------------------------------------------------------*/
/* ------- Globale Variablen ------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static unsigned long  TopHeapAktuelleGroesse;
static unsigned long  DynHeapAktuelleGroesse;

static unsigned long TopHeapNaechstesFreies;

static TopEntryT *TopHeap;

static InfoT *DynInfo;

static SV_SpeicherManagerT InfoManager;

static DomaenenT Domaene;
static unsigned long int DomainGran;

/* fuer Interleaving der Queues */
static unsigned int modulo;
static unsigned int threshold;

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics 1 -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#if PQ_STATISTIK
static unsigned long int AnzRegelnMitInitial0KPs = 0, AnzGleichungenMitInitial0KPs = 0,
                         AnzRegelnMitInitialenKPs = 0, AnzGleichungenMitInitialenKPs = 0,
                         SummeInitialerKPsZuRegeln = 0, SummeInitialerKPsZuGleichungen = 0;
# define IncAnzI0KPR() (AnzRegelnMitInitial0KPs++)
# define IncAnzI0KPG() (AnzGleichungenMitInitial0KPs++)
# define IncSIKPR(anz) (AnzRegelnMitInitialenKPs++, SummeInitialerKPsZuRegeln += anz)
# define IncSIKPG(anz) (AnzGleichungenMitInitialenKPs++, SummeInitialerKPsZuGleichungen += anz)
#else
# define IncAnzI0KPR() 
# define IncAnzI0KPG() 
# define IncSIKPR(anz)
# define IncSIKPG(anz)
#endif

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Heap-Behandlung fuer allgemeine Heaps ----------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* Hier finden sich Prozeduren, die zur Verwaltung allgemeiner Heaps benutzt werden koennen.
   Dabei ist der Top-Heap moeglichst so zu gestalten, dass er sich wie die Sub-Heaps verwalten laesst (insb.
   mit denselben Prozeduren). Diese erhalten dieselbe Aufgliederung wie beim regular heap, dh. mit
   den Varianten e1...e4 und random. Ferner wird unterschieden, ob die Statistik ein- oder ausgeschaltet ist
   und natuerlich, welche Groesse die Weight-Entries haben.
   
   Die Info-Bloecke enthalten im wesentlichen einen Zeiger auf den anhaengigen Subheap und Informationen zu
   dessen Verwaltung. Ferner die ueblichen Informationen bzgl. Status (lebend/tot) und noch anhaengenden KPs.
   
   Ein KP-Eintrag enthaelt:
   - einen Zeiger auf das Termpaar
   - einen Gewichtseintrag
   - einen Zeiger auf den anderen Elternteil

   Jedes Element des Top-Heaps erhaelt das Gewicht des anhaengenden Subheaps fest eingeschrieben.
   Dies bedingt, dass jede INSERT- bzw. SELECT-Ausfuehrung auf einem Subheap die Uebertragung des
   Gewichts des aktuellen Top-Elementes nach sich zieht.
 
   KPs zu Gleichungen werden in einem einzigen Array abgespeichert, welches aber aus zwei Subheaps besteht.
   Dies garantiert die noetige Trennung nach den beiden Richtungen als Eltern.

   Dynamischer Extra-Subheap fuer unsterbliche KPs ist vorzusehen.

   Extra-Funktion, die das Ende der KP-Bildung zu einem Elternteil angibt.

   Operationen auf Top-Heap:
   
   - Einfuegen eines Elementes
   - Entfernen eines Elementes (?)
     InfoBloecke zu toten RE fruehestens dann loeschen, wenn zugehoeriges TopHeap-Element an Top-Position angekommen ist.
     Loeschen aber erst dann, wenn es keine spaeteren KPs dazu mehr gibt.
     ==> Loeschen? == Aus TopHeapEntfernt && KeineKPsMehr 
   - Neu-Einordnen eines Elementes (LasseEinsinken nach oben sowie unten fortlaufend)
   - Durchlauf

   ==> LasseEinsinken fuer TopHeap mit der ueblichen Funktionalitaet, bis auf addref-Vergleiche, die auf
       die den zu vergleichenden Elementen anhaengenden Subheaps durchgreifen. BoundCheck ausfuehren.

   Operationen auf SubHeaps und dem dyn.Heap:
   
   - Einfuegen
   - Top entfernen
   - Neusortieren nach Durchlauf
   
   KPZuElternGeloescht darf die zugehoerigen InfoBloecke noch nicht entfernen. Diese werden erst geloescht, wenn sie
   im TopHeap auf der Top-Position angekommen sind.

   Leere SubHeaps werden NICHT angelegt. Es gibt eine zusaetzliche Funktion, die anzeigt, wenn die KP-Bildung
   zur aktuellen RE abgeschlossen ist. Diese friert den zugehoerigen Strahl ein und sorgt dafuer, dass ggf.
   eine NULL als Subheap eingetragen wird. In diesem Falle wird der Info-Block natuerlich nicht in den TopHeap
   eingetragen.

*/

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Info-Bloecken ------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define GGfStrahlLoeschen(INFO)                                                                    \
/* Voraussetzung: info->NaechstesFreies == 0 */                                                                                 \
{                                                                                                                               \
   if (RE_IstRegel(INFO->re)){                                                                                                  \
      if (INFO->subheap){                                                                                                       \
         SM_StrahlLoeschen(INFO->subheap);                                                                                      \
         INFO->subheap = NULL;                                                                                                  \
      }                                                                                                                         \
   }                                                                                                                            \
   else                                                                                                                         \
      if (((InfoT *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> NaechstesFreies == 0){                      \
         if (TP_RichtungAusgezeichnet(INFO->re) && (INFO->subheap))                                                             \
            SM_StrahlLoeschen(INFO->subheap);                                                                                   \
         else if (TP_RichtungUnausgezeichnet(INFO->re) &&                                                                       \
                  (((InfoT *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap))                          \
            SM_StrahlLoeschen(((InfoT *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap);               \
         INFO->subheap = NULL;                                                                                                  \
         ((InfoT *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap = NULL;                              \
      }                                                                                                                         \
}
         

static INLINE void InfoBlockInitialisieren(InfoT *info)
{
  info -> AnzNochNichtBehandelterKPs = 0;
#if PQ_STATISTIK
  info -> AnzInitialerKPs = 0;
#endif
  info -> NaechstesFreies = 0;
  info -> InTopHeap = FALSE;
  info -> LebtNoch = TRUE;
}


#define GGfInfoBlockLoeschen(INFO)                                                                                              \
{                                                                                                                               \
   if ( (INFO->AnzNochNichtBehandelterKPs==0) && (!INFO->InTopHeap) && (!INFO->LebtNoch) && (!INFO->subheap))                   \
      SV_Dealloc(InfoManager, INFO);                                                                                            \
}

  

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Subheap-Elementen --------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define KPVonElternGeloescht(actPar, oPar)                                                                                      \
{                                                                                                                               \
   actPar->AnzNochNichtBehandelterKPs--;                                                                                        \
   oPar->AnzNochNichtBehandelterKPs--;                                                                                          \
   GGfInfoBlockLoeschen(oPar);                                                                                                  \
}


#define TermeAuspacken_terms(tp, lhsPtr, rhsPtr, actualP, otherP, pclIDPtr) TPR_TERMPAIRtoPARALLEL_INTERMED(tp, lhsPtr, rhsPtr, actualP, otherP, pclIDPtr)

#define TermeWiederEinpacken_terms(tp, lhs, rhs, actualP, otherP, changed, pclID) TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualP, otherP, changed, pclID)

#define TermeZerstoeren_terms(tp, lhs, rhs, actualP, otherP) TPR_DESTROYPARALLEL(tp, lhs, rhs, actualP, otherP)


#define GewichteUebertragen(ziel, quelle) {ziel.w1=quelle.w1; ziel.w2=quelle.w2;}

#define GewichteHochziehen(ziel, quelle) {ziel.w1 = quelle.w1; ziel.w2 = quelle.w2;}

#define LebtNoch(elt) (elt.OtherParent->LebtNoch)



#define C_Groesser(e1, e2) ((e1.w1 > e2.w1) || ((e1.w1 == e2.w1) && (e1.w2 > e2.w2)))


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Subheaps -----------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static BOOLEAN SHLasseEinsinken (InfoT *infoPtr, unsigned long position)
{
   BOOLEAN changed = FALSE,isHeap = FALSE;
   unsigned long j=position*2+1;
   SubHeapZellenT *ElementArray = infoPtr->subheap;
   SubHeapZellenT merker= ElementArray[position];

   while ((!isHeap) && (j<infoPtr->NaechstesFreies)){
      if ((j+1<infoPtr->NaechstesFreies) &&                                      /* Position j+1 noch gueltig? */
          (C_Groesser (ElementArray [j], ElementArray[j+1])))
         j++;                                                  /* => j hat jetzt den Wert des leichteren Nachfolgers */
      if (C_Groesser (merker, ElementArray[j])){
         ElementArray[position]=ElementArray[j];
         changed = TRUE;
         position = j;
         j *=2;
         j++;
      }
      else
         isHeap = TRUE;
   }
   if (changed)
      ElementArray[position] = merker;
   return changed;
}

static INLINE void  SHTopElementLoeschenUndReorganisieren(InfoT *info)
{
   if (--(info->NaechstesFreies)>0){
      info->subheap[0] = info->subheap[info->NaechstesFreies];
      SHLasseEinsinken (info, 0);
   }
}

static INLINE void SHKomplettReorganisieren(InfoT *info)
{
   if (info->NaechstesFreies > 0){
      unsigned long position = (info->NaechstesFreies)/2;
      while (position-->0)
         SHLasseEinsinken (info, position);
   }
}

/* Heap neu aufbauen, nachdem ein Element aus der Mitte geloescht wurde */
static void SHReorganiseHole (InfoT *info, unsigned long hole) 
{
  unsigned long pos = hole;

  /* Loch an Topposition schieben (also alles auf dem Pfad nach unten) */
  while (pos > 0) {
    info->subheap [pos] = info->subheap [(pos-1)/2];
    pos = (pos-1) / 2;
  }
  if (info->NaechstesFreies > 0) {
    info->NaechstesFreies--;
    info->subheap [0] = info->subheap [info->NaechstesFreies];
    SHLasseEinsinken (info, 0);
  }
}

static INLINE void SHVonTotenAufTopBefreien(InfoT *info, BOOLEAN *changed, BOOLEAN *removed)
{ /* changed und removed enthalten am Ende des Makros gueltige Werte, werde aber zu Beginn nicht initialisiert. */
   SubHeapZellenT *sh = info->subheap;

   while ((info->NaechstesFreies>0) && (!LebtNoch(sh[0]))){
      *changed = TRUE;
      TPR_DESTROYTERMPAIR(sh[0].tp, info->re, sh[0].OtherParent->re);
      IncAnzKPSpaeterWaisenmord ();
      IncAnzKPEntfernt ();
      KPVonElternGeloescht(info, sh[0].OtherParent);
      SHTopElementLoeschenUndReorganisieren(info);
   }
   if (info->NaechstesFreies==0){
      *removed = TRUE;
      *changed = FALSE;
      GGfStrahlLoeschen(info);
   }
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf dem Top-Heap -------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define THC_Groesser(e1, e2) ((e1.w1 > e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 > e2.w2)))


static BOOLEAN THLasseEinsinken(unsigned long int position)
/* Einsinken, bis ein nicht-kleineres Gewicht angetroffen wird */
{
   BOOLEAN changed = FALSE, isHeap = FALSE;
   unsigned long j=position*2+1;
   TopEntryT merker= TopHeap[position];

   while ((!isHeap) && (j<TopHeapNaechstesFreies)){
      if ((j+1<TopHeapNaechstesFreies) &&                                      /* Position j+1 noch gueltig? */
          (THC_Groesser (TopHeap [j], TopHeap [j+1])))
         j++;                                                 /* => j hat jetzt den Wert des leichteren Nachfolgers */
      if (THC_Groesser (merker, TopHeap[j])){
         TopHeap [position]=TopHeap [j];
         changed = TRUE;
         position = j;
         j *=2;
         j++;
      }
      else
         isHeap = TRUE;
   }
   if (changed)
      TopHeap [position] = merker;
   return changed;
}



static void THNeuenSubheapEinhaengenUndReorganisieren(InfoT *info)
/* info ist der dem neuen subheap zugehoerige Info-Block */
{
   unsigned long int position = TopHeapNaechstesFreies;

   if (TopHeapNaechstesFreies==TopHeapAktuelleGroesse){
      TopHeapAktuelleGroesse += 1000;
      TopHeap = our_realloc(TopHeap,
                              TopHeapAktuelleGroesse*sizeof(TopEntryT),
                              "HoH: Ausdehnung Top-Heap");
   }
   TopHeap [TopHeapNaechstesFreies].info = info;
   GewichteHochziehen (TopHeap [TopHeapNaechstesFreies], (info->subheap)[0]);
   TopHeapNaechstesFreies++;

   do
      position = (position-1)/2;
   while (THLasseEinsinken (position) && (position>0));
}

static INLINE void THTopElementLoeschenUndReorganisieren(void)
{
   if (--TopHeapNaechstesFreies>0){
      TopHeap [0] = TopHeap [TopHeapNaechstesFreies];

      THLasseEinsinken (0);
   }
}

static INLINE void THVonTotenAufTopBefreien(void)
{
   while( (TopHeapNaechstesFreies>0) && (!TopHeap [0].info->LebtNoch)){
      TopHeap [0].info->InTopHeap = FALSE;
      GGfInfoBlockLoeschen(TopHeap [0].info);
      THTopElementLoeschenUndReorganisieren();
   }
}

static INLINE void THKomplettReorganisieren(void)
{
   unsigned long i = TopHeapNaechstesFreies/2;

   while(i > 0)
      THLasseEinsinken (--i);
}

/* Heap neu aufbauen, nachdem ein Element aus der Mitte geloescht wurde */
static void THReorganiseHole (unsigned long hole) 
{
  unsigned long pos = hole;

  /* Loch an Topposition schieben (also alles auf dem Pfad nach unten) */
  while (pos > 0) {
    TopHeap [pos] = TopHeap [(pos-1)/2];
    pos = (pos-1) / 2;
  }
  if (TopHeapNaechstesFreies > 0) {
    --TopHeapNaechstesFreies;
    TopHeap [0] = TopHeap [TopHeapNaechstesFreies];
    THLasseEinsinken (0);
  }
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Testing ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define HOH_TEST 0
/* Alternativen: 0 fuer aus, 1 fuer an */
#if HOH_TEST

#if PQ_STATISTIK == 0
#error "fuer HOH_TEST muss PQ_STATISTIK gesetzt sein"
#endif

static void HOHAusgeben(void);

static unsigned long int testNr = 0;
static BOOLEAN ok = TRUE;

#define HeapFehler(str, nr1, nr2) (ok = FALSE, printf("%ld: %s (%ld, %ld)\n", i, str, nr1, nr2))


#define TestKleiner(e1, e2) ((e1.w1 < e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 < e2.w2)))

#define GewichteIdentisch(e1, e2) ((e1.w1 == e2.w1) && (e1.w2 == e2.w2))

static void SHTest(InfoT *info)
/* Vor.: Subheap existiert */
{
   unsigned long int i;
   SubHeapZellenT *sh = info->subheap;

   for(i=0; i < info->NaechstesFreies; i++){
      if ((2*i+1< info->NaechstesFreies) &&
         TestKleiner  (sh[2*i+1], sh [i]))
         HeapFehler("SH-Sortierung falsch (Vater, Sohn)",sh[i].w1, sh[2*i+1].w1);

      if ((2*i+2< info->NaechstesFreies) &&
           TestKleiner (sh [2*i+2], sh [i]))
         HeapFehler("SH-Sortierung falsch (Vater, Sohn)", sh[i].w1, sh[2*i+2].w1);
   }
}





static void HeapTest(char *str)
{
   unsigned long i;
   printf("%s HEAPOfHEAPS-TEST %ld %s\n", ok?"++":"--", ++testNr, str);
   
   /*   if (testNr == 50)
      printf("bla\n");
   */
   for (i=0; i<TopHeapNaechstesFreies; i++){
      TopEntryT elt = TopHeap[i];

      if (elt.info->LebtNoch){
         if (elt.info->re){
            if (TP_SUEInfo(elt.info->re) != (void *) elt.info)
               HeapFehler("Info-Block !=  InfoBlock->RE->InfoBlock (Nr)", elt.info->ElternNummer, 0L);
            if (RE_IstRegel(elt.info->re)){
               if (elt.info->ElternNummer > RE_AnzRegelnGesamt())
                  HeapFehler("Regel mit falscher Nummer(reNr, maxNr)", elt.info->ElternNummer, RE_AnzRegelnGesamt());

               if (elt.info->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))
                  HeapFehler("Regel mit zuvielen noch unbeh. KPs (nnbhKPs, max)",elt.info->AnzNochNichtBehandelterKPs,
                              2*(AnzKPEingefuegt-AnzKPGeloescht));
            }
            else{
               if (elt.info->ElternNummer < -RE_AnzGleichungenGesamt())
                  HeapFehler("Gleichung mit falscher Nummer (Nr, maxNr)",elt.info->ElternNummer,  -RE_AnzGleichungenGesamt());

               if (elt.info->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))
                  HeapFehler("Gleichung mit zuvielen noch unbeh. KPs (nnbhKPs, max)",elt.info->AnzNochNichtBehandelterKPs,
                              2*(AnzKPEingefuegt-AnzKPGeloescht));

            }
         }
         else
            HeapFehler("Keine RE an lebendem Info-Block (Nr)", elt.info->ElternNummer, 0L);
         if (elt.info->subheap){
            if (!GewichteIdentisch (elt, elt.info->subheap[0]))
               HeapFehler("Gewichte nicht richtig hochgezogen (RE-Nr)", elt.info->ElternNummer, 0L);
            SHTest(elt.info);
         }
      }
      else{
         if (elt.info->re)
            HeapFehler("RE an totem Info-Block (Nr)", elt.info->ElternNummer, 0L);
         if (elt.info->subheap)
            HeapFehler("Nicht genullter Strahl an totem Info-Block (Nr)", elt.info->ElternNummer, 0L);
      }
      if ((2*i+1< TopHeapNaechstesFreies) &&
         TestKleiner (TopHeap [2*i+1], TopHeap [i]))
         HeapFehler("TH-Sortierung falsch (Vater,Sohn)",i,  2*i+1);

      if ((2*i+2< TopHeapNaechstesFreies) &&
           TestKleiner (TopHeap [2*i+2], TopHeap [i]))
         HeapFehler("TH-Sortierung falsch (Vater,Sohn)", i, 2*i+2);
   }
   SHTest(DynInfo);
   if (!ok) HOHAusgeben();
}



#define ElementGewichtAusgeben(ELT) printf(" w1 = %ld, w2 = %ld\n", (ELT).w1, (ELT).w2)


static void SHAusgeben(InfoT *info)
/* Vor.: Subheap existiert */
{
   unsigned long int i;
   SubHeapZellenT *sh = info->subheap;

   for(i=0; i < info->NaechstesFreies; i++){
      printf("   OtherPar = %ld, %s, ", sh[i].OtherParent->ElternNummer, (sh[i].OtherParent->LebtNoch)?"*":"+");
      ElementGewichtAusgeben (sh[i]);
   }
}



         
void HOHAusgeben(void)
{
   unsigned long i;
   for (i=0; i<TopHeapNaechstesFreies; i++){
      TopEntryT elt = TopHeap [i];
      if (elt.info->LebtNoch){
         if (RE_IstRegel(elt.info->re))
            printf("TH[%3ld]: Regel Nr. %ld, %s, %lu init., %lu unbeh. KPs. NF = %lu.", i,
                   elt.info->ElternNummer, (elt.info->LebtNoch)?"*":"+",
                   elt.info->AnzInitialerKPs,
                   elt.info->AnzNochNichtBehandelterKPs,
                   elt.info->NaechstesFreies);
         else
            printf("TH[%3ld]: Gleichung Nr. %ld (%s), %s, %lu init., %lu unbeh. KPs. NF = %lu.",  i,
                   elt.info->ElternNummer,
                   TP_RichtungAusgezeichnet(elt.info->re)?"ausgez":"unausgez",
                   (elt.info->LebtNoch)?"*":"+",
                   elt.info->AnzInitialerKPs,
                   elt.info->AnzNochNichtBehandelterKPs,
                   elt.info->NaechstesFreies);
         ElementGewichtAusgeben (TopHeap [i]);
         SHAusgeben(elt.info);
      }
      else{
         printf("TH[%3ld]: RE Nr. %ld, +, %lu init., %lu unbeh. KPs. NF = %lu.", i,elt.info->ElternNummer,
                elt.info->AnzInitialerKPs, elt.info->AnzNochNichtBehandelterKPs,
                   elt.info->NaechstesFreies);
         ElementGewichtAusgeben (TopHeap [i]);
      }
   }
   printf("TH[---]: RE Nr. 0, *, %lu init., %lu unbeh. KPs. NF = %lu. \n",
          DynInfo -> AnzInitialerKPs,
          DynInfo -> AnzNochNichtBehandelterKPs,
          DynInfo -> NaechstesFreies);
   SHAusgeben(DynInfo);
}

#else
# define HeapTest(str)
#endif /* HOH_TEST */
      
/* --------------------------------------------------------------------------------------------------------*/
/* ------- Insert -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void HOH_INSERT (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
{
   SubHeapZellenT *sh;
   InfoT *info = (InfoT *) TP_SUEInfo(aP);

   HeapTest("INSERT Anfang");
   /* ggf. dynHeap ausdehnen */
   sh = info->subheap;

   sh[info->NaechstesFreies].tp = tp;
   sh[info->NaechstesFreies].OtherParent = TP_SUEInfo(oP);
   GewichteUebertragen (sh[info->NaechstesFreies], wt);

   info->NaechstesFreies++;
#if PQ_STATISTIK
   info->AnzInitialerKPs++;
#endif
   IncAnzKPEingefuegt ();
   info->AnzNochNichtBehandelterKPs++;
   (((InfoT *) TP_SUEInfo(oP))->AnzNochNichtBehandelterKPs)++;
   HeapTest("INSERT Ende");
}

void HOH_INSERTImmortal (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP MAYBE_UNUSEDPARAM, RegelOderGleichungsT oP)
{
   SubHeapZellenT *sh;
   InfoT *info = DynInfo;
   unsigned int i = info->NaechstesFreies;

   HeapTest("INSERTImmortal Anfang");
   /* ggf. dynHeap ausdehnen */
   if (info->NaechstesFreies == DynHeapAktuelleGroesse){
      DynHeapAktuelleGroesse *= 2;
      info->subheap = our_realloc(info->subheap,
                                  DynHeapAktuelleGroesse*sizeof(SubHeapZellenT),
                                  "HoH: Ausdehnung dyn. Heap");
   }

   sh = info->subheap;

   sh[info->NaechstesFreies].tp = tp;
   sh[info->NaechstesFreies].OtherParent = TP_SUEInfo(oP);
   GewichteUebertragen (sh[info->NaechstesFreies], wt);

   info->NaechstesFreies++;

#if PQ_STATISTIK
   info->AnzInitialerKPs++;
#endif
   IncAnzKPEingefuegt ();

   do
      i = (i-1)/2;
   while(SHLasseEinsinken (info, i) && (i>0));

   info->AnzNochNichtBehandelterKPs += 2;

   HeapTest("INSERTImmortal Ende");
}


BOOLEAN HOH_POSTPONE (TermpairT tp MAYBE_UNUSEDPARAM, WeightEntryT wt MAYBE_UNUSEDPARAM, RegelOderGleichungsT aP MAYBE_UNUSEDPARAM, RegelOderGleichungsT oP MAYBE_UNUSEDPARAM)
{
   return FALSE;
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Select -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static void GetFifoSubheap (SubHeapZellenT *sh, int NaechstesFreies,
                            int *nrMin, int *indexMin)
{
   int i;
   *nrMin = INT_MAX;
   *indexMin = -1;
   
   for (i = 0; i < NaechstesFreies; i++) {
      if (LebtNoch(sh[i]) && sh[i].w2 < *nrMin) {
         *nrMin = sh[i].w2;
         *indexMin = i;
      }
   }
}


static BOOLEAN
HOH_SELECT_Fifo (TermpairT *tpPtr,
                 RegelOderGleichungsT *actParPtr,
                 RegelOderGleichungsT *oParPtr)
{
   SubHeapZellenT *sh;
   unsigned int index;
   int indexMinE, indexMinR, nrMinE, nrMinR;
   int nrMinDyn, indexMinDyn, indexMinRE, nrMinSH, indexMinSH;
   BOOLEAN found;
   
   HeapTest("SELECT-fifo Anfang");

   /* minimales Element aus DynInfo auswaehlen */
   if (DynInfo->NaechstesFreies > 0) {
      GetFifoSubheap (DynInfo->subheap, DynInfo->NaechstesFreies,
                      &nrMinDyn, &indexMinDyn);
   }

   found = FALSE;
   do {
      indexMinE = -1;
      indexMinR = -1;
      nrMinE = INT_MIN;
      nrMinR = INT_MAX;

      /* Suchen der Subheaps mit kleinster Regel/Gleichungsnummer */
      for (index = 0; index < TopHeapNaechstesFreies; index++) {
         if (TopHeap[index].info->LebtNoch) {
            if (TopHeap[index].info->ElternNummer < 0) {
               /* Gleichung */
               if (TopHeap[index].info->ElternNummer > nrMinE) {
                  nrMinE = TopHeap[index].info->ElternNummer;
                  indexMinE = index;
               }
            } else {
               /* Regel */
               if (TopHeap[index].info->ElternNummer < nrMinR) {
                  nrMinR = TopHeap[index].info->ElternNummer;
                  indexMinR = index;
               }
            }
         }
      }
      
      /* Wenn sowohl Regeln als auch Gleichungen existieren, dann muss
         derjenige Subheap ausgewaehlt werden, welcher aelter ist */
      if (indexMinR == -1 && indexMinE == -1) {
         /* Sonderfall! -> Es existieren noch keine REs */
         indexMinRE = -1;
         if (DynInfo->NaechstesFreies == 0)
            /* Es existieren ueberhaupt keine CPs :-( */
            return FALSE;
         found = TRUE;
      } else if (indexMinR == -1) {
         indexMinRE = indexMinE;
      } else if (indexMinE == -1) {
         indexMinRE = indexMinR;
      } else {
         if (TopHeap[indexMinR].w2 < TopHeap[indexMinE].w2) {
            indexMinRE = indexMinR;
         } else {
            indexMinRE = indexMinE;
         }
      }
   
      if (indexMinRE != -1) {
         GetFifoSubheap (TopHeap[indexMinRE].info->subheap,
                         TopHeap [indexMinRE].info->NaechstesFreies,
                         &nrMinSH, &indexMinSH);
         if (indexMinSH == -1) {
            /* Subheap enthaelt nur Waisen/ist Leer */
            BOOLEAN changed, removed;
            
            SHVonTotenAufTopBefreien (TopHeap[indexMinRE].info, &changed, &removed);
            if (removed) {
               TopHeap [indexMinRE].info->InTopHeap = FALSE;
               GGfInfoBlockLoeschen (TopHeap [indexMinRE].info);
	       THReorganiseHole (indexMinRE);
            }
         } else {
            found = TRUE;
         }
      }
   } while (!found);
   
   if (indexMinRE == -1 || (DynInfo->NaechstesFreies > 0 && nrMinDyn < nrMinSH)) {
      /* minimales Gewicht ist in DynInfo */
      sh = DynInfo->subheap;
      *tpPtr = sh[indexMinDyn].tp;
      *oParPtr = *actParPtr = &SUE_Demigod;
      DynInfo->AnzNochNichtBehandelterKPs -=2;

      SHReorganiseHole (DynInfo, indexMinDyn);

   } else {
      
      /* minimales Gewicht ist in TopHeap[indexMinRE] */
      sh = TopHeap[indexMinRE].info->subheap;
      *tpPtr = sh[indexMinSH].tp;
      *actParPtr = (TopHeap [indexMinRE].info)->re;
      *oParPtr = sh[indexMinSH].OtherParent->re;

      KPVonElternGeloescht(TopHeap [indexMinRE].info,
                           TopHeap [indexMinRE].info->subheap[indexMinSH].OtherParent);
      /* Subheap reorganisieren */
      SHReorganiseHole (TopHeap[indexMinRE].info, indexMinSH);
      if (TopHeap [indexMinRE].info->NaechstesFreies == 0) {
         GGfStrahlLoeschen(TopHeap [indexMinRE].info);
      } else {
	if (TopHeap[indexMinRE].w1 != TopHeap[indexMinRE].info->subheap[0].w1
	    && TopHeap[indexMinRE].w2 != TopHeap[indexMinRE].info->subheap[0].w2) {
	  /* Die Gewichte haben sich geaendert. Der TopHeap muss reorganisiert werden */
	  TopEntryT merker;
	  unsigned long pos;
      
	  GewichteHochziehen (TopHeap [indexMinRE], (TopHeap[indexMinRE].info->subheap)[0]);
	  merker = TopHeap [indexMinRE];
	  pos = indexMinRE;
	  while (pos > 0) {
	    TopHeap [pos] = TopHeap [(pos-1)/2];
	    pos = (pos-1) / 2;
	  }
	  TopHeap [0] = merker;
	  THLasseEinsinken (0);
	}
      }
   }
   IncAnzKPEntfernt();

   HeapTest("SELECT-fifo Ende");

   SUE_LastSelectedWasFifo = TRUE;
   return TRUE;
}   


BOOLEAN HOH_SELECT (TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr)
/* Ggf. Subheap fuer zuletzt eingefuegte KPs anlegen und diese dort hineinkopieren. (Von hinten!)
   Dabei jedes Element Einsinken lassen.
   Leichtesten Heap aus Top-Heap auswaehlen, dort Top-Element entnehmen. GGf. Top-Heap reorganisieren.

   while (progress && THNaechstesFreies>0)
     TH von toten auf Top befreien
     Top-SH von toten auf top befreien
     if(Top-SH removed)
        THTopElementLoeschenUndReorganisieren
     else if changed
        THLasseEinsinken(0)
     else
        progress = FALSE
 */
{
   BOOLEAN removed, changed, progress = TRUE;
   SubHeapZellenT top;

   if (SUE_AnzAktivierterRE % modulo < threshold) {
      return HOH_SELECT_Fifo (tpPtr, actParPtr, oParPtr);
   }
  
   HeapTest("SELECT Anfang");

                                /* zuerst den Top-Heap von toten Sub-Heaps auf seinem Gipfel befreien */
                                /* und den obersten Subheap von toten KPs auf seinem Gipfel befreien. */
                                /* Dabei ggf. den Top-Heap reorganisieren und das Procedere wiederholen. */
   while(progress && (TopHeapNaechstesFreies > 0)){

      THVonTotenAufTopBefreien();
      if (TopHeapNaechstesFreies == 0)
         break;
      removed=FALSE, changed=FALSE;
      SHVonTotenAufTopBefreien(TopHeap [0].info, &changed, &removed);
      if (removed){
         TopHeap [0].info -> InTopHeap = FALSE;
         GGfInfoBlockLoeschen(TopHeap [0].info);
         THTopElementLoeschenUndReorganisieren();
      }else if (changed){
         GewichteHochziehen (TopHeap [0],
            (TopHeap [0].info->subheap)[0]);
         THLasseEinsinken (0);
      }else
         progress = FALSE;
   }

   if ((TopHeapNaechstesFreies > 0) &&
       ((DynInfo->NaechstesFreies == 0) ||
        (C_Groesser(DynInfo->subheap[0],
                    (TopHeap [0].info->subheap)[0])))){ /* TopHeap auslesen */

      top = TopHeap [0].info->subheap[0];
      *tpPtr = top.tp;
      *actParPtr = (TopHeap [0].info)->re;
      *oParPtr = top.OtherParent->re;

      KPVonElternGeloescht(TopHeap [0].info,
                           TopHeap [0].info->subheap[0].OtherParent);

      SHTopElementLoeschenUndReorganisieren(TopHeap [0].info);

      if (TopHeap [0].info->NaechstesFreies == 0){
         GGfStrahlLoeschen(TopHeap [0].info);
      }
      else{
         GewichteHochziehen (TopHeap [0],
                               (TopHeap [0].info->subheap)[0]);
         THLasseEinsinken(0);
      }
   }
   else if (DynInfo->NaechstesFreies > 0){
      top = DynInfo->subheap[0];
      DynInfo->AnzNochNichtBehandelterKPs -=2;
      *tpPtr = top.tp;
      *oParPtr = *actParPtr = &SUE_Demigod;
      SHTopElementLoeschenUndReorganisieren(DynInfo);
   }
   else
      return FALSE;

   IncAnzKPEntfernt();

   HeapTest("SELECT Ende");
   SUE_LastSelectedWasFifo = FALSE;

   return TRUE;
}   





/* --------------------------------------------------------------------------------------------------------*/
/* ------- Parents ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void HOH_NEWPARENT (RegelOderGleichungsT re)
/* Info-Block anlegen, zugehoerigen Sub-Heap mit NULL initialisieren.
   Bei Gleichungen zweiten Info-Block anlegen, aber ohne Sub-Heap!*/
{

   if (IsDemigod(re)){
      InfoT *info = DynInfo;
      TP_SUEInfo(re) = (void *) DynInfo;
      InfoBlockInitialisieren(info);
      info -> ElternNummer = TP_ElternNr(re);
      info -> re = re;
   }
   else{
      InfoT *info;

      SV_Alloc(TP_SUEInfo(re),InfoManager,void *);
      info = (InfoT *) TP_SUEInfo(re);

      InfoBlockInitialisieren(info);
      info -> ElternNummer = TP_ElternNr(re);
      info -> re = re;
      info -> subheap = SM_NeuerStrahl(Domaene);

      if (!RE_IstRegel(re)){        /* also Gleichung */
         SV_Alloc(TP_SUEInfo(TP_Gegenrichtung(re)),InfoManager,void *);
         info = (InfoT *) TP_SUEInfo(TP_Gegenrichtung(re));
         InfoBlockInitialisieren(info);
         info -> ElternNummer = TP_ElternNr(re);
         info -> re = TP_Gegenrichtung(re);
         info -> subheap = NULL;
      }
   }
}




static INLINE void SubheapLeeren(InfoT *actualInfo)
/* Loescht alle Eintraege im actualInfo anhaengenden Subheap von Index 0 bis NaechstesFreies-1. */
{
   unsigned long i;
   SubHeapZellenT *sh = actualInfo->subheap;

   for (i=0; i < actualInfo->NaechstesFreies; i++){
     IncAnzKPFrueherWaisenmord();
     IncAnzKPEntfernt();
     TPR_DESTROYTERMPAIR(sh[i].tp, actualInfo->re, sh[i].OtherParent->re);
     KPVonElternGeloescht(actualInfo, sh[i].OtherParent);
   }
}   
   

void HOH_KILLPARENT (RegelOderGleichungsT re)
/* Gesamten Sub-Heap durchlaufen und jedes Element loeschen, Sub-Heap freigeben. */
{
   InfoT *info = (InfoT *) TP_SUEInfo(re);

   HeapTest("KILLPARENT Anfang");
   info -> LebtNoch = FALSE;
#if IO_AUSGABEN
   if ((IO_level4) || (IO_level3))
     fprintf(IO_Strom, "\n>>orphan murder<<\n\n");
#endif
   SubheapLeeren(info);
   if (RE_IstRegel(re)){
      if (IsDemigod(re))
         our_dealloc(info->subheap);
      else if (info->subheap)
         SM_StrahlLoeschen(info->subheap);
      info -> subheap=NULL;
      info -> NaechstesFreies = 0;
      info -> re = NULL;
      GGfInfoBlockLoeschen(info);
   }
   else{                        /* Gleichung */
      InfoT *infoGegenrichtung = (InfoT *) TP_SUEInfo(TP_Gegenrichtung(re));

      infoGegenrichtung  -> LebtNoch = FALSE;
      SubheapLeeren(infoGegenrichtung);
      if (info->subheap)
         SM_StrahlLoeschen(info->subheap);
      info->subheap=NULL;
      info->NaechstesFreies = 0;
      infoGegenrichtung->subheap=NULL;
      infoGegenrichtung->NaechstesFreies = 0;
      info -> re = NULL;
      infoGegenrichtung-> re = NULL;
      GGfInfoBlockLoeschen(info);
      GGfInfoBlockLoeschen(infoGegenrichtung);
   }
   if ((IO_level4) || (IO_level3))
     fprintf(IO_Strom, ">>orphan murder finished<<\n\n");

   HeapTest("KILLPARENT Ende");
}



void HOH_REDIRECTPARENT (void)
/* Neuen Sub-Heap zur Gegenrichtung anlegen und fuellen. Es wird ein gemeinsames Array verwendet, das
   dann auch gemeinsam freigegeben werden kann.
   Die Gleichungsrichtung in SUE_ActualParent ist noch gueltig, wird aber sofort im Anschluss umgedreht.
   Jede Gleichung wird 'umgedreht', bevor FinishParent aufgerufen wird!
 */
{
   InfoT *info = (InfoT *) TP_SUEInfo(SUE_ActualParent);
   InfoT *infoGegenrichtung = (InfoT *) TP_SUEInfo(TP_Gegenrichtung(SUE_ActualParent));

   infoGegenrichtung->subheap = &(info->subheap[info->NaechstesFreies]);
}




void HOH_FINISHPARENT (void)
/* Zeigt an, wenn die KP-Bildung zum zuletzt angemeldeten Elternteil abgeschlossen ist.
   Wurden keine KPs zu diesem Elternteil eingefuegt, so muss der anhaengende Strahl sofort
   freigegeben werden. Info-Bloecke koennen hier noch nicht geloescht werden, weil die
   Elternregel auf jeden Fall noch lebt.
*/
{
   HeapTest("FINISHPARENT Anfang");

   if (RE_IstRegel(SUE_ActualParent)){
      InfoT *info = (InfoT *) TP_SUEInfo(SUE_ActualParent);
      if (info->NaechstesFreies>0){
         IncSIKPR (info->NaechstesFreies);
         SM_StrahlGefrieren(info->subheap, &(info->subheap[info->NaechstesFreies]));
         info->InTopHeap = TRUE;
         SHKomplettReorganisieren (info);
         THNeuenSubheapEinhaengenUndReorganisieren(info);
      }
      else{
         IncAnzI0KPR ();
         SM_NichtGefrorenenStrahlLoeschen(info->subheap);
         info->subheap = NULL;
      }
   }
   else{  /* SUE_ActualParent enthaelt nach REDIRECT die Gegenrichtung der aktuellen Gleichung */
      InfoT *infoGegenrichtung = (InfoT *) TP_SUEInfo(SUE_ActualParent),
      *info = (InfoT *) TP_SUEInfo(TP_Gegenrichtung(SUE_ActualParent));

      if ((info->NaechstesFreies>0) || (infoGegenrichtung->NaechstesFreies>0)){

         IncSIKPG (info->NaechstesFreies + infoGegenrichtung->NaechstesFreies);
         SM_StrahlGefrieren(info->subheap, &(infoGegenrichtung->subheap[infoGegenrichtung->NaechstesFreies]));

         if  (infoGegenrichtung->NaechstesFreies>0){
            infoGegenrichtung->InTopHeap = TRUE;
            SHKomplettReorganisieren(infoGegenrichtung);
            THNeuenSubheapEinhaengenUndReorganisieren(infoGegenrichtung);
         }
         if (info->NaechstesFreies>0){
            info->InTopHeap = TRUE;
            SHKomplettReorganisieren(info);
            THNeuenSubheapEinhaengenUndReorganisieren(info);
         }
      }
      else{
         IncAnzI0KPG ();
         SM_NichtGefrorenenStrahlLoeschen(info->subheap);
         info->subheap = NULL;
         infoGegenrichtung->subheap = NULL;
      }
   }
   HeapTest("FINISHPARENT Ende");
}



/* --------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough ------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

unsigned long int HOH_AnzKPsToGo(void)
{
#if PQ_STATISTIK
  return (AnzKPEingefuegt - AnzKPGeloescht);
#else
  return 0;
#endif
}


static INLINE void SHWALKTHROUGH (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms,
                                  int *result, InfoT *info)
/* AP auf jedes Element des Subheaps anwenden. Subheap am Ende ggf. reorganisieren,
   ggf. auch Strahl loeschen.
   RESULT ist:
   0 - noop
   1 - delete
   2 - reweight
*/
{
   unsigned long int i = 0;
   SubHeapZellenT *sh = info->subheap;
   int CPresult;
   BOOLEAN SHchanged = FALSE;

   if (info->LebtNoch){
      while(i<info->NaechstesFreies){
         IncAnzKPIRBetrachtet ();
         if (LebtNoch(sh[i])){
            TermT lhs=NULL, rhs=NULL;
            PCL_ObjektNummerT pclID;
            /* Daten an Globale Variablen zuweisen */
            if (NeedsTerms)
              TermeAuspacken_terms (sh[i].tp, &lhs, &rhs, info->re, sh[i].OtherParent->re, &pclID);
            if ((CPresult = (* AP) (lhs, rhs, info->re, sh[i].OtherParent->re,
                                    &pclID)) == 1) {                                               /* loeschen */
              if (NeedsTerms)
                 TermeZerstoeren_terms (sh[i].tp, lhs, rhs, info->re,sh[i].OtherParent->re );
              KPVonElternGeloescht(info, sh[i].OtherParent);
              sh[i] = sh[--(info->NaechstesFreies)];
              SHchanged = TRUE;
              IncAnzKPEntfernt ();
              IncAnzKPIRGeloescht ();
            }
            else if (CPresult == 2){                                                             /* neugewichten */
              WeightEntryT wt;

              IncAnzKPIRVeraendert ();
              SHchanged = TRUE;
              GewichteUebertragen (wt, sh[i]);
              C_ReClassify(&wt, lhs, rhs, info->re,sh[i].OtherParent->re);
              if (NeedsTerms)
                sh[i].tp = TermeWiederEinpacken_terms (sh[i].tp, lhs, rhs, info->re, sh[i].OtherParent->re, TRUE, pclID);
              GewichteUebertragen (sh[i], wt);
              i++;
            }
            else{
              if (NeedsTerms)
                sh[i].tp = TermeWiederEinpacken_terms (sh[i].tp, lhs, rhs, info->re, sh[i].OtherParent->re, FALSE, pclID);
              i++;
            }
         }
         else{                                                                             /* KP ist bereits tot */
            TPR_DESTROYTERMPAIR(sh[i].tp, info->re, sh[i].OtherParent->re);

            KPVonElternGeloescht(info, sh[i].OtherParent);

            AZ_DruckeMeldungItem();
            IncAnzKPIRWaisenmord ();
            IncAnzKPEntfernt ();

            sh[i] = sh[--(info->NaechstesFreies)];
            SHchanged = TRUE;
         }
      }
      if ((info->NaechstesFreies == 0) && (!IsDemigod(info->re))){
         GGfStrahlLoeschen(info);   /* Der Info-Block kann noch nicht gel. werden, da er noch im TH steht */
         *result = 1;            /* loeschen */
      }
      else if (SHchanged){
         SHKomplettReorganisieren(info);
         *result = 2;            /* neubewerten */
      }
      else
         *result = 0;            /* nix tun */
   }
   else
      *result = 1;               /* schon tot: loeschen */
}
         

static INLINE void THWALKTHROUGHBODY (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
/* Doppelte Schleife ueber Top-Level-Heap und alle Sub-Heaps, dabei vorzeitigen Waisenmord ausfuehren. */
{
   unsigned long int thindex = 0;
   int result;
   BOOLEAN changed = FALSE;

   while(thindex<TopHeapNaechstesFreies){
      if (TopHeap [thindex].info->LebtNoch){                   /* tot oder nach Durchlauf leer */
         SHWALKTHROUGH(AP, NeedsTerms, &result, TopHeap[thindex].info);
      }
      else
         result = 1;            /* tot => loeschen */

      if (result == 1){
         TopHeap[thindex].info->InTopHeap = FALSE;
         GGfInfoBlockLoeschen(TopHeap[thindex].info);
         TopHeap [thindex] = TopHeap [--TopHeapNaechstesFreies];
         changed = TRUE;
      }
      else if (result == 2){
         GewichteHochziehen (TopHeap [thindex],
                               TopHeap [thindex].info->subheap[0]);
         changed = TRUE;
         thindex++;
      }
      else
         thindex++;
   }
   if (changed){
      THKomplettReorganisieren();
   }
   /* SH-WT auf DynHeap!, anschliessend diesen ggf. Reorganisieren. */
   SHWALKTHROUGH(AP, NeedsTerms, &result, DynInfo);
}


void HOH_WALKTHROUGH (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
{
   if (NeedsTerms)
     THWALKTHROUGHBODY(AP, TRUE);
   else
     THWALKTHROUGHBODY(AP, FALSE);
}


static INLINE void SH_WWBODY(InfoT *info, PQ_WatchingWalkthroughProcedureT AP)
/* Vor.: Subheap existiert */
{
   unsigned long int j;

   for (j = 0; j < info->NaechstesFreies; j++) {
      if (info->subheap[j].OtherParent->LebtNoch) {
        WeightEntryT wt;
        GewichteUebertragen (wt, info->subheap[j]);
        (* AP)(wt, info->subheap[j].tp, info->re,
                   info->subheap[j].OtherParent->re);
      }
   }
}

void HOH_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP)
/* s.o., aber ohne Waisenmord. */
{
  unsigned long i;
  
  for (i=0; i<TopHeapNaechstesFreies; i++){
    if (TopHeap [i].info->LebtNoch){
      SH_WWBODY(TopHeap [i].info, AP);
    }
  }
  if (DynInfo -> NaechstesFreies>0)
    SH_WWBODY(DynInfo, AP);
}

static INLINE void
SH_TVBODY (InfoT *info, void (* OperationAufTermen) (TermT, TermT))
/* Vor.: Subheap existiert */
{
  unsigned long int j;

  for (j = 0; j < info->NaechstesFreies; j++)
    TPR_OperationWaehrendTermverfolgung(OperationAufTermen, info->subheap[j].tp,
                                        info->re, info->subheap[j].OtherParent->re);
}

void HOH_TermVerfolgung (void (* OperationAufTermen) (TermT, TermT))
/* s.o., aber ohne Waisenmord. */
{
  unsigned long i;
  for (i=0; i < TopHeapNaechstesFreies; i++){
      if (TopHeap [i].info->LebtNoch){
         SH_TVBODY(TopHeap[i].info, OperationAufTermen);
      }
   }
   if (DynInfo -> NaechstesFreies > 0)
      SH_TVBODY(DynInfo, OperationAufTermen);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics 2 -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static unsigned long int Lebende;

static void LebendsKPsZaehlen(WeightEntryT wt MAYBE_UNUSEDPARAM, TermpairT tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualP MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherP MAYBE_UNUSEDPARAM)
{
   Lebende++;
}

void HOH_STATISTICS(InfoStringT is)
{
   if (!(IS_IsIn("all", is) || (IS_IsIn("PQ", is))))
      return;                   /* Modul nicht angesprochen */
   if (IS_IsIn("standard", is)){
      IO_DruckeKurzUeberschrift("Priority-Queue: Heap of heaps", '+');
#if PQ_STATISTIK
      Lebende = 0;
      HOH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
      PQStatistik(is, Lebende);
      fprintf(IO_Strom, "\n");
      IO_DruckeStatistikZeile("Regeln ohne initiale KPs", AnzRegelnMitInitial0KPs);
      IO_DruckeStatistikZeile("Regeln mit initialen KPs", AnzRegelnMitInitialenKPs);
      if (AnzRegelnMitInitialenKPs>0)
        IO_DruckeStatistikZeile("Durchschnittliche initiale KPs zu Regeln", SummeInitialerKPsZuRegeln/AnzRegelnMitInitialenKPs);
      IO_DruckeStatistikZeile("Gleichungen ohne initiale KPs", AnzGleichungenMitInitial0KPs);
      IO_DruckeStatistikZeile("Gleichungen mit initialen KPs", AnzGleichungenMitInitialenKPs);
      if (AnzGleichungenMitInitialenKPs>0)
        IO_DruckeStatistikZeile("Durchschnittliche initiale KPs zu Gleichungen", SummeInitialerKPsZuGleichungen/AnzGleichungenMitInitialenKPs);
#endif
   }
   else if(IS_IsIn("latex-head", is)){
#if PQ_STATISTIK
     PQStatistik(is, 0);
     fprintf(IO_Strom, "R: mit init & ohne init & ");
     if (AnzRegelnMitInitialenKPs>0)
       fprintf(IO_Strom, "avg init & ");
     fprintf(IO_Strom, "E: mit init & ohne init & ");
     if (AnzGleichungenMitInitialenKPs>0)
       fprintf(IO_Strom, "avg init & ");
#endif
   }
   else if(IS_IsIn("intern", is))
      IO_DruckeStatistikZeile("domain granularity", DomainGran);
   else if(IS_IsIn("config", is)){
      fprintf(IO_Strom, "HoH: Statistik %s, ", PQ_STATISTIK?" On":"Off");
#if HOH_TEST
      fprintf(IO_Strom, "Heap-Test-Modus  On\n");
#else
      fprintf(IO_Strom, "Heap-Test-Modus Off\n");
#endif
   }      
}

void HOH_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  Lebende = 0;
  HOH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
  
  *AktGueltigeKPZahl = Lebende;

#if PQ_STATISTIK
  *AktKPZahl = AnzKPEingefuegt - AnzKPGeloescht;
#else
  *AktKPZahl = Lebende;
#endif

}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Clean up ---------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void HOH_Abraeumen(void)
/* Alles loeschen*/
{
  unsigned long int i;
  for (i=0; i<TopHeapNaechstesFreies; i++){
    if (TopHeap [i].info->subheap){
      unsigned long j;
      SubHeapZellenT *sh = TopHeap [i].info->subheap;
      
      for (j=0; j < TopHeap[i].info->NaechstesFreies; j++){
        TPR_DESTROYTERMPAIR(sh[j].tp, TopHeap [i].info->re, sh[j].OtherParent->re);
        KPVonElternGeloescht(TopHeap [i].info, sh[j].OtherParent);
      }
      TopHeap [i].info->NaechstesFreies = 0;
      GGfStrahlLoeschen(TopHeap [i].info);
    }
    TopHeap [i].info->InTopHeap = FALSE;
    GGfInfoBlockLoeschen(TopHeap [i].info);
  }
  SM_StrahldomaeneLoeschen(Domaene);
}

void HOH_SoftAbraeumen(void)
     /* Nur die Nicht-SpeicherManager-Objekte freigeben, wiederaufsetzen-kompatibel. */
{
  our_dealloc(DynInfo->subheap);
  SM_StrahldomaeneLoeschen(Domaene);
  our_dealloc(TopHeap);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Initialize -------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/



static BOOLEAN HOH_SchonInitialisiert = FALSE;

BOOLEAN HOH_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* Info-BlockManager init.,  Funktionspointer in .h-File besetzen, Top-Level-Heap allokieren */
{
  int fifo, heuristic;
   
   if (!HOH_SchonInitialisiert){
     if (IS_ValueOfOption(DEF, USR, "domgran"))
       DomainGran = atoi(IS_ValueOfOption(DEF, USR, "domgran"));
     else
       DomainGran = 10;
     
     /* dyn. Heap initialisieren */
     DynHeapAktuelleGroesse = 100;

     TopHeapAktuelleGroesse = PQ_AnzahlInfoBloeckeJeBlock;
   }

   Domaene = SM_NeueStrahldomaene(DM_HOH, DomainGran, "Domain for heap of heaps");

   /* Top-Heap initialisieren */
   TopHeapNaechstesFreies = 0;
   
   
#if !COMP_POWER
   fprintf(IO_Strom, "Heap of Heaps. Statistics = %s. domgran = %ld.\n",
           PQ_STATISTIK ?"on":"off", DomainGran);
#endif

   if (!HOH_SchonInitialisiert){
/* HACKALERT */  
   if (PI_ParseInterleave (USR, &fifo, &heuristic)
       || PI_ParseInterleave (DEF, &fifo, &heuristic)) {
     modulo = fifo + heuristic;
     threshold = fifo;
#if !COMP_POWER
     fprintf (IO_Strom, "  Interleaving of CP Queues\n");
     fprintf (IO_Strom, "    Fifo: %d, Heuristic: %d\n", fifo, heuristic);
#endif
   } else {
     modulo = 1;
     threshold = 0;
   }
   }
   
   TopHeap = our_alloc(TopHeapAktuelleGroesse * sizeof(TopEntryT));
   SV_ManagerInitialisieren(&InfoManager, InfoT, PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (HoH)");
   SV_Alloc(DynInfo, InfoManager, void *);
   DynInfo -> subheap = our_alloc(DynHeapAktuelleGroesse * sizeof(SubHeapZellenT));
   
   HOH_SchonInitialisiert = TRUE;

   return TRUE;
}

#endif /* SV_BASIS_MMAP */
