/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - PriorityQueue.c                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Schnittstelle fuer die Priority Queue.                                                                     */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "compiler-extra.h"
#include "general.h"
#include "Ausgaben.h"
#include "InfoStrings.h"
#include "FehlerBehandlung.h"

#include "PriorityQueue.h"
#include "KPHeap.h"
#include "HeapOfHeaps.h"
#include "SpeicherVerwaltung.h"

#include <stdio.h>

/* Falls nicht SV_BASIS_MMAP wird immer der KPHeap benutzt */


/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe PriorityQueue.h) ---------------------*/
/* ------------------------------------------------------------------------*/

PQEinfuegerT PQ_InsertProc,
             PQ_InsertImmortalProc;
BOOLEAN (* PQ_PostponeProc)         (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);

void    (* PQ_WalkthroughProc) (PQ_WalkthroughProcedureT, BOOLEAN);
void    (* PQ_WatchingWalkthroughProc) (PQ_WatchingWalkthroughProcedureT);
void    (* PQ_TermVerfolgungsProc) (void (*) (TermT, TermT));
unsigned long int (* PQ_AnzKPsToGoProc) (void);

void    (* PQ_StatisticsProc)  (InfoStringT);
void   (* PQ_ActSizeProc) (unsigned long int *, unsigned long int *);

void (* PQ_CleanupProc) (void);
void (* PQ_SoftCleanupProc) (void);

BOOLEAN (* PQ_SelectInitialProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *);
BOOLEAN (* PQ_SelectProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *);
void    (* PQ_NewParentProc)   (RegelOderGleichungsT);
void    (* PQ_KillParentProc)  (RegelOderGleichungsT);
void    (* PQ_RedirectParentProc) (void);
void    (* PQ_FinishParentProc) (void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


static void VoidDummy(void)
{
   ;
}

static BOOLEAN SelectInitialDummy(TermpairT *tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT *v MAYBE_UNUSEDPARAM, RegelOderGleichungsT *m MAYBE_UNUSEDPARAM)
{
   our_error("Falsche Einstellung in einem Nicht-VV-Modus\n");
   return FALSE;
}

void PQ_ParamInfo(ParamInfoArtT what MAYBE_UNUSEDPARAM, InfoStringT DEF MAYBE_UNUSEDPARAM, InfoStringT USR MAYBE_UNUSEDPARAM)
{
   fprintf(IO_Strom, "\n* PRIORITY-QUEUE\n\n  ");
   fprintf(IO_Strom, "PQ_ParamInfo not yet implemented\n\n");
   
/*   switch (IS_FindOption(DEF, USR, 4, "heap", "hoh", "HeapOfHeaps", "list")){
   case 1:
   case 2:
      HOH_ParamInfo(what, DEF, USR);
      break;
   case 0:
   default:
      KPH_ParamInfo(what, DEF, USR);
      break;
   }*/
   
}

static InfoStringT PQ_DEF, PQ_USR;
static int PQ_HeapNr;

void PQ_REINITIALIZE (void)
{
#if defined SV_BASIS_MMAP
   switch (PQ_HeapNr){
   case 1:
   case 2:
      HOH_InitAposteriori (PQ_DEF, PQ_USR);
      break;
   case 0:
   default:
      /* Initialize PQ-Module with IS, NumIntsInWeightEntry, AdditionalComparisonProcedure */
      KPH_InitAposteriori(PQ_DEF, PQ_USR);
      break;
   }
#else
      KPH_InitAposteriori(PQ_DEF, PQ_USR);
#endif
}


BOOLEAN PQ_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
#if !COMP_POWER
   fprintf(IO_Strom, "\n* PRIORITY-QUEUE\n\n  ");
#endif
   PQ_HeapNr = IS_FindOption(DEF, USR, 4, "heap", "hoh", "HeapOfHeaps", "list");
   PQ_DEF = DEF;
   PQ_USR = USR;
#if defined SV_BASIS_MMAP
   switch (PQ_HeapNr){
   case 1:
   case 2:
      HOH_InitAposteriori(DEF, USR);
      PQ_InsertProc         = HOH_INSERT;
      PQ_InsertImmortalProc = HOH_INSERTImmortal;
      PQ_PostponeProc       = HOH_POSTPONE;
      PQ_SelectProc         = HOH_SELECT;
      PQ_NewParentProc      = HOH_NEWPARENT;
      PQ_KillParentProc     = HOH_KILLPARENT;
      PQ_RedirectParentProc = HOH_REDIRECTPARENT;
      PQ_WalkthroughProc    = HOH_WALKTHROUGH;
      PQ_WatchingWalkthroughProc = HOH_WATCHINGWALKTHROUGH;
      PQ_StatisticsProc     = HOH_STATISTICS;
      PQ_CleanupProc        = HOH_Abraeumen;
      PQ_SoftCleanupProc    = HOH_SoftAbraeumen;
      PQ_FinishParentProc   = HOH_FINISHPARENT;
      PQ_TermVerfolgungsProc = HOH_TermVerfolgung;
      PQ_AnzKPsToGoProc     = HOH_AnzKPsToGo;
      PQ_SelectInitialProc  = SelectInitialDummy;
      PQ_ActSizeProc        = HOH_ActSize;
      return TRUE;
      break;
   case 0:
   default:
#endif
      /* Initialize PQ-Module with IS, NumIntsInWeightEntry, AdditionalComparisonProcedure */
      KPH_InitAposteriori(DEF, USR);
      PQ_InsertProc         = KPH_INSERT;
      PQ_InsertImmortalProc = KPH_INSERTImmortal;
      PQ_PostponeProc       = KPH_POSTPONE;
      PQ_SelectProc         = KPH_SELECT;
      PQ_NewParentProc      = KPH_NEWPARENT;
      PQ_KillParentProc     = KPH_KILLPARENT;
      PQ_RedirectParentProc = KPH_REDIRECTPARENT;
      PQ_WalkthroughProc    = KPH_WALKTHROUGH;
      PQ_WatchingWalkthroughProc = KPH_WATCHINGWALKTHROUGH;
      PQ_StatisticsProc     = KPH_STATISTICS;
      PQ_CleanupProc        = KPH_Abraeumen;
      PQ_SoftCleanupProc    = KPH_SoftAbraeumen;
      PQ_FinishParentProc   = VoidDummy;
      PQ_TermVerfolgungsProc = KPH_TermVerfolgung;
      PQ_AnzKPsToGoProc     = KPH_AnzKPsToGo;
      PQ_SelectInitialProc  = KPH_SELECTINITIAL;
      PQ_ActSizeProc        = KPH_ActSize;
      return TRUE;
#if defined SV_BASIS_MMAP
      break;
   }
#endif
}
         
