
#include "antiwarnings.h"
#include <string.h>
#include <stdlib.h>

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "FileMenue.h"
#include "Hauptkomponenten.h"
#include "Interreduktion.h"
#include "KBO.h"
#include "Konsulat.h"
#include "KPBildung.h"
#include "KPMengenInterreduktion.h"
#include "LPOVortests.h"
#include "NFBildung.h"
#include "Ordnungen.h"
#include "Parameter.h"
#include "PhilMarlow.h"
#include "Puffer.h"
#include "SUEM.h"
#include "Statistiken.h"
#include "Unifikation1.h"
#include "Zielverwaltung.h"
#include "WaldmeisterII.h"
#include "pcl.h"


/* ------------------------------------------------------------------------*/
/* -------------------- globale Variablen (siehe Parameter.h) -------------*/
/* ------------------------------------------------------------------------*/

ordering PA_Ordering;
mode PA_ExtModus;
BOOLEAN PA_InteraktionAn,
        PA_FailingCompletion,
        PA_EinsSternAn,
        PA_NusfuAn,
        PA_KernAn,
        PA_GolemAn,
        PA_BackAn,
        PA_WWWAn,
        PA_Automodus;
AufrufOptionT AufrufOptionen[OptNrLAST];


/* ------------------------------------------------------------------------*/
/* -------------------- Initialisierung -----------------------------------*/
/* ------------------------------------------------------------------------*/

const char *GruppenNamen[8] = { "Arbeitsmodi", "KPStrategien", "Normalform-Strategien", "Ziel-Behandlung", 
                            "Interreduktion", "Technische Details", "Zusatz-Optionen", "Debug-Optionen"};

static char *Zeichenkette, *ProtokolldateiName = "stdout";
static BOOLEAN XPRTModus, XPRTModusSichtbar, Vervollstaendigung, VervollstaendigungZugaenglich;

BOOLEAN VerwendungVonGleichungen;                                         /* Ist das Auftreten von Gleichungen ueberhaupt moeglich? */


#define XPRTModusWechseln() (XPRTModus = !XPRTModus)

#define OptionZugaenglich(option) (AufrufOptionen[option].aktiviert && (!AufrufOptionen[option].IsXPRT || XPRTModus) &&\
                                   (!AufrufOptionen[option].NurVervollstaendigung || Vervollstaendigung))

#define HoechsterSichtbarerIndex(option) (XPRTModus?AufrufOptionen[option].ParameterTyp.zweistellig.AnzahlVerschiedenerWerte:\
                                          AufrufOptionen[option].ParameterTyp.zweistellig.ErsterXPRTWert)


BOOLEAN PA_XPRTModus(void)
{
  return XPRTModus;
}

BOOLEAN PA_XPRTModusSichtbar(void)
{
  return XPRTModusSichtbar;
}


void PA_InitApriori(void)
/* Aufruf-Optionen besetzen.*/
{
  int i;

  if (WM_ErsterAnlauf()) {
    PA_InteraktionAn = FALSE;
#if COMP_COMP
    PA_Automodus = TRUE;
#else
    PA_Automodus = FALSE;
#endif
    XPRTModus = TRUE;
    XPRTModusSichtbar = TRUE;
    VervollstaendigungZugaenglich = TRUE;
    Vervollstaendigung = VervollstaendigungZugaenglich;
    /* Fuer Sichtbarkeit der Modi z.B. im Hilfe-Menue */

    Zeichenkette = our_alloc(INPUT_STRLEN * sizeof(char));
  }
  else {
    Zeichenkette[0] = '\0';
  }

/*------ Arbeitsmodi-------------------------------------*/

   AufrufOptionen[OptNrFailing].Kuerzel           = "-f";
   AufrufOptionen[OptNrFailing].Name              = "'f'ailing completion";
   AufrufOptionen[OptNrFailing].Bedeutung         = "Abbruch beim Auftreten unrichtbarer kritischer Paare";
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.BedeutungT        = "Failing Completion";
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.BedeutungF        = "Unfailing Completion";
   AufrufOptionen[OptNrFailing].Gruppe            = Arbeitsmodi;
   AufrufOptionen[OptNrFailing].ParameterArt      = 1;
   AufrufOptionen[OptNrFailing].aktiviert = TRUE;
   AufrufOptionen[OptNrFailing].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrFailing].IsXPRT = FALSE;
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.defaultWert = FALSE;

/*------ SUE-Statistics ---------------------------------*/

   AufrufOptionen[OptNrSUE].Kuerzel           = "-sue";
   AufrufOptionen[OptNrSUE].Name              = "'SUE'";
   AufrufOptionen[OptNrSUE].Bedeutung         = "SUEM. Syntax:\nstatistics=[on|off]";
   AufrufOptionen[OptNrSUE].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrSUE].ParameterArt      = 3;
   AufrufOptionen[OptNrSUE].aktiviert = TRUE;
   AufrufOptionen[OptNrSUE].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrSUE].IsXPRT = FALSE;
   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString = 
      normal_power("statistics=on:paraint=3","statistics=off:paraint=3");
      
/*------ Classification ---------------------------------*/

   AufrufOptionen[OptNrClassification].Kuerzel           = "-clas";
   AufrufOptionen[OptNrClassification].Name              = "'Clas'sification";
   AufrufOptionen[OptNrClassification].Bedeutung         = "Classification of unselected equations. Syntax:\n"
      "heuristic=[addweight|cp-in-goal|glam|goal-in-cp|gtweight|\n"
      "           mixweight|maxweight|occnest|unif]           --- primary heuristic\n"
      "DEF=n1:VAR=n2:f=n3:g=n4...:                            --- symbol weights\n"
      "statistics=[on|off]:                                   --- statistics\n" 
      "killer[R|E|RE]=[regular|half|ultimate|double|last]:    --- killer criterion\n"
      "echild=[regular|half|ultimate|double|last]:            --- children of equations\n"
      "initial=[regular|half|ultimate|double|last]:           --- initial equations\n"
      "killedRE=[regular|half|ultimate|double|last]:          --- killed rules/equations\n"
      "cut1=n:                                                --- throw away CPs with weight w1 > n\n"
      "statint=n                                              --- intervall for time distribution of weights\n";
   AufrufOptionen[OptNrClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString = "def=1:var=1:"
      "initial=ultimate:statistics=off:statint=1:w1=prodsum.gt.length.plus.length:w2=fifo:database=ultimate";
      
   AufrufOptionen[OptNrReClassification].Kuerzel           = "-reclas";
   AufrufOptionen[OptNrReClassification].Name              = "'Re-Clas'sification";
   AufrufOptionen[OptNrReClassification].Bedeutung         = "Reclassification of unselected equations during SUE-. Syntax:\n"
     "killer[R|E|RE]=[regular|half|ultimate|double|penalty]:\n"
     "echild=[regular|half|ultimate|double|penalty]:\n"
     "perm[3|4|5|6]=[regular|half|ultimate|double|penalty]\n"; 
   AufrufOptionen[OptNrReClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrReClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrReClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrReClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrReClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString = "";

   AufrufOptionen[OptNrCGClassification].Kuerzel           = "-cgclas";
   AufrufOptionen[OptNrCGClassification].Name              = "'CG-Clas'sification";
   AufrufOptionen[OptNrCGClassification].Bedeutung         = "classification of critical goals. Syntax like clas\n"; 
   AufrufOptionen[OptNrCGClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrCGClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrCGClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrCGClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrCGClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString = 
      "g1=prod.plus.length.plus.unif:g2=fifo:def=2";

   AufrufOptionen[OptNrCGInterreduktion].Kuerzel           = "-cg";
   AufrufOptionen[OptNrCGInterreduktion].Name              = "'CG'-set normalization";
   AufrufOptionen[OptNrCGInterreduktion].Bedeutung         = "interreduction of critical goals. Syntax similar to ki\n"; 
   AufrufOptionen[OptNrCGInterreduktion].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrCGInterreduktion].ParameterArt      = 3;
   AufrufOptionen[OptNrCGInterreduktion].aktiviert = TRUE;
   AufrufOptionen[OptNrCGInterreduktion].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrCGInterreduktion].IsXPRT = TRUE;
   AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.DefaultInfoString = "";
      
/*------ Priority Queue ---------------------------------*/

   AufrufOptionen[OptNrPriorityQueue].Kuerzel           = "-pq";
   AufrufOptionen[OptNrPriorityQueue].Name              = "'P'riority-'Q'ueue";
   AufrufOptionen[OptNrPriorityQueue].Bedeutung         = "'P'riority-'Q'ueue. Syntax:\n"
      "[regular|pairing|hoh]:"
      "[e1|e2|e3|e4|random|unambiguous]:   --- unambiguous nicht bei 'hoh' \n"
      "[left|bin|right]:                   --- nur bei pairing\n"
      "statistics=[on|off]:\n"
      "domain=Granularity";
   AufrufOptionen[OptNrPriorityQueue].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrPriorityQueue].ParameterArt      = 3;
   AufrufOptionen[OptNrPriorityQueue].aktiviert = TRUE;
   AufrufOptionen[OptNrPriorityQueue].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrPriorityQueue].IsXPRT = FALSE;
   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString = 
      normal_power("heap:e1:domain=30:statistics=on","heap:e1:domain=30:statistics=off");

/*------ Termpair-Representation ------------------------*/

   AufrufOptionen[OptNrTermpairRepresentation].Kuerzel           = "-tpr";
   AufrufOptionen[OptNrTermpairRepresentation].Name              = "'T'erm'p'air-'r'epresentation";
   AufrufOptionen[OptNrTermpairRepresentation].Bedeutung         = "Syntax:\n"
      "[regular|packed|implicit]:statistics=[on|off]";
   AufrufOptionen[OptNrTermpairRepresentation].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrTermpairRepresentation].ParameterArt      = 3;
   AufrufOptionen[OptNrTermpairRepresentation].aktiviert = TRUE;
   AufrufOptionen[OptNrTermpairRepresentation].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrTermpairRepresentation].IsXPRT = FALSE;
   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString = 
      normal_power("packed::statistics=on:managers=50","packed::statistics=off:managers=50");

/*------ CP-Criteria ------------------------------------*/

   AufrufOptionen[OptNrKapur].Kuerzel           = "-subcon";
   AufrufOptionen[OptNrKapur].Name              = "'Subcon'nectedness-criterion";
   AufrufOptionen[OptNrKapur].Bedeutung         = "Vermeidung unnoetiger KPs. Argument = Staerke des Tests\n"
      "['r'|'e']" ;
   AufrufOptionen[OptNrKapur].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKapur].ParameterArt      = 0;
   AufrufOptionen[OptNrKapur].aktiviert = TRUE;
   AufrufOptionen[OptNrKapur].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKapur].IsXPRT = FALSE;
   AufrufOptionen[OptNrKapur].ParameterTyp.frei.wurdeAngegeben = FALSE;
   AufrufOptionen[OptNrKapur].ParameterTyp.frei.defaultWert = "";

/*------ Eins-Stern -------------------------------------*/

   AufrufOptionen[OptNrEinsStern].Kuerzel           = "-einsstern";
   AufrufOptionen[OptNrEinsStern].Name              = "'Eins-Stern'-Kriterium";
   AufrufOptionen[OptNrEinsStern].Bedeutung         = "nur kritische Paare an 1* betrachten";
   AufrufOptionen[OptNrEinsStern].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrEinsStern].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrEinsStern].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrEinsStern].ParameterArt      = 1;
   AufrufOptionen[OptNrEinsStern].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrEinsStern].aktiviert = TRUE;
   AufrufOptionen[OptNrEinsStern].IsXPRT = FALSE;
   AufrufOptionen[OptNrEinsStern].ParameterTyp.einstellig.defaultWert = FALSE;


/*------ Nusfu ------------------------------------------*/

   AufrufOptionen[OptNrNusfu].Kuerzel           = "-nusfu";
   AufrufOptionen[OptNrNusfu].Name              = "Nusfu-Kriterium";
   AufrufOptionen[OptNrNusfu].Bedeutung         = "nicht unterhalb von Skolemfunktionen ueberlappen";
   AufrufOptionen[OptNrNusfu].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrNusfu].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrNusfu].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrNusfu].ParameterArt      = 1;
   AufrufOptionen[OptNrNusfu].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrNusfu].aktiviert = TRUE;
   AufrufOptionen[OptNrNusfu].IsXPRT = FALSE;
   AufrufOptionen[OptNrNusfu].ParameterTyp.einstellig.defaultWert = TRUE;

/*------ Kern -------------------------------------------*/

   AufrufOptionen[OptNrKern].Kuerzel           = "-kern";
   AufrufOptionen[OptNrKern].Name              = "'Kern'-Strategie";
   AufrufOptionen[OptNrKern].Bedeutung         = "auf den Kopf gestellte Vervollstaendigung";
   AufrufOptionen[OptNrKern].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrKern].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrKern].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKern].ParameterArt      = 1;
   AufrufOptionen[OptNrKern].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKern].aktiviert = TRUE;
   AufrufOptionen[OptNrKern].IsXPRT = FALSE;
   AufrufOptionen[OptNrKern].ParameterTyp.einstellig.defaultWert = FALSE;

/*------ Golem ------------------------------------------*/

   AufrufOptionen[OptNrGolem].Kuerzel           = "-golem";
   AufrufOptionen[OptNrGolem].Name              = "'Golem";
   AufrufOptionen[OptNrGolem].Bedeutung         = "";
   AufrufOptionen[OptNrGolem].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrGolem].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrGolem].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrGolem].ParameterArt      = 1;
   AufrufOptionen[OptNrGolem].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrGolem].aktiviert = TRUE;
   AufrufOptionen[OptNrGolem].IsXPRT = FALSE;
   AufrufOptionen[OptNrGolem].ParameterTyp.einstellig.defaultWert = TRUE;


/*------ Back ------------------------------------------*/

   AufrufOptionen[OptNrBack].Kuerzel           = "-back";
   AufrufOptionen[OptNrBack].Name              = "'back'ward reasoning";
   AufrufOptionen[OptNrBack].Bedeutung         = "Rueckwaertsargumentieren bei kritischen Zielen";
   AufrufOptionen[OptNrBack].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrBack].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrBack].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrBack].ParameterArt      = 1;
   AufrufOptionen[OptNrBack].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrBack].aktiviert = TRUE;
   AufrufOptionen[OptNrBack].IsXPRT = FALSE;
   AufrufOptionen[OptNrBack].ParameterTyp.einstellig.defaultWert = FALSE;

/*------ Auto-Modus -------------------------------------*/

   /* wird extra behandelt ! */
   AufrufOptionen[OptNrAutomodus].Kuerzel           = "-auto";
   AufrufOptionen[OptNrAutomodus].Name              = "Automodus";
   AufrufOptionen[OptNrAutomodus].Bedeutung         = "automagic";
   AufrufOptionen[OptNrAutomodus].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrAutomodus].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrAutomodus].Gruppe            = ZusatzOptionen;
   AufrufOptionen[OptNrAutomodus].ParameterArt      = 1;
   AufrufOptionen[OptNrAutomodus].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrAutomodus].aktiviert = TRUE;
   AufrufOptionen[OptNrAutomodus].IsXPRT = FALSE;
#if COMP_COMP
   /* wird eigentlich in PA_InitApriori gesetzt */
   AufrufOptionen[OptNrAutomodus].ParameterTyp.einstellig.defaultWert = TRUE;
#else
   AufrufOptionen[OptNrAutomodus].ParameterTyp.einstellig.defaultWert = FALSE;
#endif
/*------ Treatment of CPs -------------------------------*/

   AufrufOptionen[OptNrKPInterreduktion].Kuerzel           = "-ki";
   AufrufOptionen[OptNrKPInterreduktion].Name              = "'K'P-Mengen-'I'nterreduktions";
   AufrufOptionen[OptNrKPInterreduktion].Bedeutung         = "Durchreduzieren von SUE. Syntax:\n"
      "period=<periode>[res]:[res]=<stichpunkt>.<stichpunkt>:[res]=<stichpunkt>.<stichpunkt>...[:statistis=[on|off]]";
   AufrufOptionen[OptNrKPInterreduktion].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPInterreduktion].ParameterArt      = 3;
   AufrufOptionen[OptNrKPInterreduktion].aktiviert = TRUE;
   AufrufOptionen[OptNrKPInterreduktion].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPInterreduktion].IsXPRT = FALSE;
   AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString = 
      normal_power("statistics=on","statistics=off");

   AufrufOptionen[OptNrKPgen].Kuerzel           = "-kg";
   AufrufOptionen[OptNrKPgen].Name              = "'K'P-Behandlung 'g'eneration";
   AufrufOptionen[OptNrKPgen].Bedeutung         = "Behandlung der KPs after generation.Syntax\n"
      "r:e:s:permsub=[n]";
   AufrufOptionen[OptNrKPgen].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPgen].ParameterArt      = 3;
   AufrufOptionen[OptNrKPgen].aktiviert = TRUE;
   AufrufOptionen[OptNrKPgen].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPgen].IsXPRT = TRUE;
   AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString = "r";

   AufrufOptionen[OptNrKPsel].Kuerzel           = "-ks";
   AufrufOptionen[OptNrKPsel].Name              = "'K'P-Behandlung 's'election";
   AufrufOptionen[OptNrKPsel].Bedeutung         = "Behandlung der KPs after selection.Syntax\n"
      "r:e:s:permsub=[n]";
   AufrufOptionen[OptNrKPsel].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPsel].ParameterArt      = 3;
   AufrufOptionen[OptNrKPsel].aktiviert = TRUE;
   AufrufOptionen[OptNrKPsel].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPsel].IsXPRT = TRUE;
   AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString = "r:e:s:permsubE=4";

/*------ NormalformStrategien ---------------------------*/

   AufrufOptionen[OptNrNFBildung].Kuerzel                   = "-nf";
   AufrufOptionen[OptNrNFBildung].Name              = "'N'ormal'f'ormbildung";
   AufrufOptionen[OptNrNFBildung].Bedeutung = "Normalization-Strategy and Backtracking-Behaviour.";
   AufrufOptionen[OptNrNFBildung].Gruppe = NormalformStrategien;
   AufrufOptionen[OptNrNFBildung].ParameterArt = 3;
   AufrufOptionen[OptNrNFBildung].aktiviert = TRUE;
   AufrufOptionen[OptNrNFBildung].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrNFBildung].IsXPRT = FALSE;
   AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString = 
      "outermost:up:repeat";

/*------ Zielbehandlung ---------------------------------*/

   AufrufOptionen[OptNrZielbehandlung].Kuerzel           = "-zb";
   AufrufOptionen[OptNrZielbehandlung].Name              = "'Z'iel-'B'ehandlung";
   AufrufOptionen[OptNrZielbehandlung].Bedeutung         = "Proof-Modus: Behandlung der Ziele";
   AufrufOptionen[OptNrZielbehandlung].Gruppe            = Zielbehandlung ;
   AufrufOptionen[OptNrZielbehandlung].ParameterArt      = 3;
   AufrufOptionen[OptNrZielbehandlung].aktiviert = TRUE;
   AufrufOptionen[OptNrZielbehandlung].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrZielbehandlung].IsXPRT = TRUE;
   AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString = 
      "period=1:continue:rnf";
   /*"period=1:continue:mnf=re.0:s:mnfanalysis=1000.2000.10:mnfantiWOVar:mnfantir=5:mnfup=5"*/;

 /*------ Interreduktion ---------------------------------*/

   AufrufOptionen[OptNrRMrechtsIR].Kuerzel           = "-irr";
   AufrufOptionen[OptNrRMrechtsIR].Name              = "'I'R 'r'ight hand sides of 'r'ules";
   AufrufOptionen[OptNrRMrechtsIR].Bedeutung         = "Interval between subsequent interreductions of rhs of rules";
   AufrufOptionen[OptNrRMrechtsIR].Gruppe            = Interreduktion;
   AufrufOptionen[OptNrRMrechtsIR].ParameterArt      = 3;
   AufrufOptionen[OptNrRMrechtsIR].aktiviert = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].IsXPRT = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString="period=1:strength=re";

/*------ Ordnungen ---------------------------------*/

   AufrufOptionen[OptNrOrdnungen].Kuerzel           = "-ord";
   AufrufOptionen[OptNrOrdnungen].Name              = "Orderings";
   AufrufOptionen[OptNrOrdnungen].Bedeutung         = "Override ordering from specification ";
   AufrufOptionen[OptNrOrdnungen].Gruppe            = ZusatzOptionen;
   AufrufOptionen[OptNrOrdnungen].ParameterArt      = 3;
   AufrufOptionen[OptNrOrdnungen].aktiviert = TRUE;
   AufrufOptionen[OptNrOrdnungen].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrOrdnungen].IsXPRT = TRUE;
   AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString = "";
   
/*------ Symbolnormierung---------------------------*/

   AufrufOptionen[OptNrSymbnorm].Kuerzel           = "-symbnorm";
   AufrufOptionen[OptNrSymbnorm].Name              = "Symbol setting";
   AufrufOptionen[OptNrSymbnorm].Bedeutung         = "normalize arrangement of symbols\n"
      "This parameter activates the unique reordering of\n"
      "the symbols from the specification file";
   AufrufOptionen[OptNrSymbnorm].ParameterTyp.einstellig.BedeutungT        = "On";
   AufrufOptionen[OptNrSymbnorm].ParameterTyp.einstellig.BedeutungF        = "Off";
   AufrufOptionen[OptNrSymbnorm].Gruppe            = ZusatzOptionen;
   AufrufOptionen[OptNrSymbnorm].ParameterArt      = 1;
   AufrufOptionen[OptNrSymbnorm].aktiviert = TRUE;
   AufrufOptionen[OptNrSymbnorm].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrSymbnorm].IsXPRT = TRUE;
   AufrufOptionen[OptNrSymbnorm].ParameterTyp.einstellig.defaultWert = TRUE;
   
/*------ Datenbank ---------------------------------*/

   AufrufOptionen[OptNrDatabase].Kuerzel           = "-db";
   AufrufOptionen[OptNrDatabase].Name              = "Data base";
   AufrufOptionen[OptNrDatabase].Bedeutung         = "Activate data base for additional reasoning rules";
   AufrufOptionen[OptNrDatabase].Gruppe            = ZusatzOptionen;
   AufrufOptionen[OptNrDatabase].ParameterArt      = 0;
   AufrufOptionen[OptNrDatabase].aktiviert = TRUE;
   AufrufOptionen[OptNrDatabase].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrDatabase].IsXPRT = TRUE;
   AufrufOptionen[OptNrDatabase].ParameterTyp.frei.eingestellterWert="";
   AufrufOptionen[OptNrDatabase].ParameterTyp.frei.defaultWert="";
   
/*------ TechnischeDetails ------------------------------*/

   AufrufOptionen[OptNrAusgabeLevel].Kuerzel                   = "-a";
   AufrufOptionen[OptNrAusgabeLevel].Name              = "'A'usgabe-Level";
   AufrufOptionen[OptNrAusgabeLevel].Bedeutung = "Einstellen des Ausgabe-Levels (Wert ist proportional zur Geschwaetzigkeit)";
   AufrufOptionen[OptNrAusgabeLevel].Gruppe    = TechnischeDetails;
   AufrufOptionen[OptNrAusgabeLevel].ParameterArt = 2;
   AufrufOptionen[OptNrAusgabeLevel].aktiviert = TRUE;
   AufrufOptionen[OptNrAusgabeLevel].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrAusgabeLevel].IsXPRT = FALSE;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.defaultWert = 2;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.AnzahlVerschiedenerWerte = 5;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.ErsterXPRTWert = 5;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[0] = "0";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[1] = "1";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[2] = "2";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[3] = "3";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[4] = "4";

   AufrufOptionen[OptNrPCLProt].Kuerzel           = "-pcl";
   AufrufOptionen[OptNrPCLProt].Name              = "'PCL' protocol";
   AufrufOptionen[OptNrPCLProt].Bedeutung         = "Generation of a proof protocol.\n";
   AufrufOptionen[OptNrPCLProt].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrPCLProt].ParameterArt      = 3;
#if COMP_PCL
   AufrufOptionen[OptNrPCLProt].aktiviert = TRUE;
#else
   AufrufOptionen[OptNrPCLProt].aktiviert = FALSE;
#endif
   AufrufOptionen[OptNrPCLProt].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrPCLProt].IsXPRT = FALSE;
   AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString = normal_power_comp("select=off","iPCL:select=off","iPCL:select=off");
   
   AufrufOptionen[OptNrZeitlimit].Kuerzel           = "-tl";
   AufrufOptionen[OptNrZeitlimit].Name              = "set 't'ime-'l'imit";
   AufrufOptionen[OptNrZeitlimit].Bedeutung         = "Waldmeister terminates after <sec> process time";
   AufrufOptionen[OptNrZeitlimit].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrZeitlimit].ParameterArt      = 0;
   AufrufOptionen[OptNrZeitlimit].aktiviert = TRUE;
   AufrufOptionen[OptNrZeitlimit].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrZeitlimit].IsXPRT = TRUE;
   AufrufOptionen[OptNrZeitlimit].ParameterTyp.frei.defaultWert = "0";

   
   AufrufOptionen[OptNrRealZeitlimit].Kuerzel           = "-rtl";
   AufrufOptionen[OptNrRealZeitlimit].Name              = "set 'r'eal-'t'ime-'l'imit";
   AufrufOptionen[OptNrRealZeitlimit].Bedeutung         = "Waldmeister terminates after <sec> real time";
   AufrufOptionen[OptNrRealZeitlimit].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrRealZeitlimit].ParameterArt      = 0;
   AufrufOptionen[OptNrRealZeitlimit].aktiviert = TRUE;
   AufrufOptionen[OptNrRealZeitlimit].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrRealZeitlimit].IsXPRT = TRUE;
   AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.defaultWert = "0";
   AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.wurdeAngegeben = FALSE;

   AufrufOptionen[OptNrMemlimit].Kuerzel           = "-ml";
   AufrufOptionen[OptNrMemlimit].Name              = "set 'm'emory-'l'imit";
   AufrufOptionen[OptNrMemlimit].Bedeutung         = "Waldmeister should use only <mem> megabytes of memory";
   AufrufOptionen[OptNrMemlimit].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrMemlimit].ParameterArt      = 0;
   AufrufOptionen[OptNrMemlimit].aktiviert = TRUE;
   AufrufOptionen[OptNrMemlimit].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrMemlimit].IsXPRT = TRUE;
   AufrufOptionen[OptNrMemlimit].ParameterTyp.frei.defaultWert = "0";
   
   AufrufOptionen[OptNrErgebnis].Kuerzel           = "-res";
   AufrufOptionen[OptNrErgebnis].Name              = "'Res'ult if interaction is off";
   AufrufOptionen[OptNrErgebnis].Bedeutung         = "Syntax:\n h    - Head\n"
      "r    - Regelmenge\n"
      "e    - Gleichungsmenge\n"
      "s    - Statistik (komplett)\n"
      "k    - Konfiguration\n"
      "i    - Intern\n"
      "p    - Prozess\n"
      "t    - Times\n"
      "pa   - Aufruf-Parameter\n"
      "sh   - Short Statistik\n"
      "tex  - LaTeX-Statistik\n"
      "stex - short LaTeX-Statistik\n";
   AufrufOptionen[OptNrErgebnis].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrErgebnis].ParameterArt      = 3;
   AufrufOptionen[OptNrErgebnis].aktiviert = TRUE;
   AufrufOptionen[OptNrErgebnis].IsXPRT = TRUE;
   AufrufOptionen[OptNrErgebnis].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.DefaultInfoString=normal_power("-:stex","");

   AufrufOptionen[OptNrWWW].Kuerzel           = "-www";
   AufrufOptionen[OptNrWWW].Name              = "WWW-Niceness";
   AufrufOptionen[OptNrWWW].Bedeutung         = "Produce some output especially formatted for WWW output";
   AufrufOptionen[OptNrWWW].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrWWW].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrWWW].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrWWW].ParameterArt      = 1;
   AufrufOptionen[OptNrWWW].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrWWW].aktiviert = TRUE;
   AufrufOptionen[OptNrWWW].IsXPRT = TRUE;
   AufrufOptionen[OptNrWWW].ParameterTyp.einstellig.defaultWert = FALSE;

   AufrufOptionen[OptNrDummy].Kuerzel           = "-dummy";
   AufrufOptionen[OptNrDummy].Name              = "dummy";
   AufrufOptionen[OptNrDummy].Bedeutung         = "Not used by Waldmeister";
   AufrufOptionen[OptNrDummy].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrDummy].ParameterArt      = 0;
   AufrufOptionen[OptNrDummy].aktiviert = TRUE;
   AufrufOptionen[OptNrDummy].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDummy].IsXPRT = TRUE;
   AufrufOptionen[OptNrDummy].ParameterTyp.frei.defaultWert = "0";

   for (i=0; i<OptNrLAST; i++)
    switch (AufrufOptionen[i].ParameterArt){
    case 0:                     /* frei */
      AufrufOptionen[i].ParameterTyp.frei.eingestellterWert =
        AufrufOptionen[i].ParameterTyp.frei.defaultWert;
      AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben = FALSE;
      break;
    case 1:                     /* einstellig */
      AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert =
        AufrufOptionen[i].ParameterTyp.einstellig.defaultWert;
      break;
    case 2:                     /* zweistellig */
      AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert =
        AufrufOptionen[i].ParameterTyp.zweistellig.defaultWert;
      break;
    case 3:                     /* InfoString */
      AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString = "";
      break;
    default: break;
    }
}


void PA_InitAposteriori(void)
{
   int i;
   
   if ((PA_ExtModus == proof) || (PA_ExtModus == completion)){ /* bei Proof oder Completion */
     SGM_InitAposteriori(); /* Subgleichungen */
   }
   else{                                                       /* sonst */
      /* In den verschiedenen Modi ungueltige Parameter ueberschreiben  */

      AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString = "-:r"; /* Nur Regeln */
      AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString = "-"; /* Kein PCL */
      AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert = "-";      /* Kein Kapur */
      AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString = "-:w1=fifo:intial=ultimate";
      AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString = "";
      AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString = "";
   }

   for (i=0; i<OptNrLAST; i++)
      if (AufrufOptionen[i].ParameterArt == 3){ /* InfoString */
         if (IS_IsIn("-", AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString))
            AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString = "";
      }

   IO_DruckeKopf();             /* Programm-Kopf ausgeben */
#if !COMP_POWER
   IO_DruckeBeispiel();
   IO_DruckeKurzUeberschrift("INITIALIZATION", '+');
   IO_ParamInfo(ParamEvaluate, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);   /* nur ausgeben! */
   if (Vervollstaendigung)
     fprintf(IO_Strom, "\n* MODE:  %s completion\n", 
             AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert?
             "failing":"unfailing");
#endif



   ORD_InitAposteriori (AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
			AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
   
   BK_InitAposteriori();

   /* An dieser Stelle muessten noch Ueberpruefungen der Gleichungen und Ziele stattfinden:
      - verwendete Argumentsorten;
      - Verwendung von Funktionen nur mit den richtigen Stelligkeiten.*/
   /* Skolemisierung der Ziele: Passiert aus dem Hauptprogramm heraus.*/
   /* clear_spec(); Erst einfuegen, wenn NACH Ablauf dieser Funktion nicht mehr auf 
		    die eingelesenen Daten zugegriffen werden muss.
		    ! Kann hier noch nicht aufgerufen werden, 
		    da die Ziele erst spaeter eingebaut werden. */

   PU_InitAposteriori();

   NF_InitAposteriori(AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString);

   U1_InitAposteriori(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);

   KPB_InitAposteriori(AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);

   ST_InitAposteriori(AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.UserInfoString);

   HK_InitAposteriori(AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);

   IR_InitAposteriori(AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString);
   SUE_InitAposteriori(AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString,
                       AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.UserInfoString);

   KPIR_InitAposteriori(AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                        AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);


   if_pcl(PCL_InitAposteriori(AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
			      AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);)
   Z_InitAposteriori(AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                     AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
   KO_InitAposteriori();
}


/* ------------------------------------------------------------------------------------*/
/* -------------------- Ausgabe der Einstellungen im Statistik-Kopf -------------------*/
/* ------------------------------------------------------------------------------------*/

static void SchreibeStringOhneHochkommataNachZeichenkette(const char *string)
/* s.Name: nach Ende dieser Prozedur steht string bis auf Hochkommata in Zeichenkette.*/
{
   const char *p;
   char       *ziel;
  for (p = string, ziel = Zeichenkette; *p; p++){ /* Hochkommas aus String eliminieren */
    if (*p!='\''){
      *ziel = *p;
      ziel++;
    }
  }
  *ziel = '\0';
}

void PA_DruckeModi(void)
/* Gibt die eingestellten Modi auf den IO_Strom aus.
   Verwendung bei Ausgabe der Statistik.*/
{
  if (Vervollstaendigung){
    fprintf(IO_Strom, "\n* MODE:  %s completion\n", 
            AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert?
            "failing":"unfailing");
  }
  
  IO_ParamInfo(ParamEvaluate, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);

  ORD_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
			AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
  
  NF_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString,
               AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString);

  U1_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);

  KPB_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);
  HK_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);
  if (Vervollstaendigung){
    IR_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString,
                 AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString);
  }

  SUE_ParamInfo(ParamEvaluate, 99, 
                AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.UserInfoString);
  
  KPIR_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                 AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);

  if (PA_ExtModus==proof){
     Z_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                 AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
  }
  if_pcl(PCL_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
		       AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);)

  fprintf(IO_Strom, "\n\n");
}  

static void DruckeOptionDefault(FILE *fp, AufrufOptionenTyp i)
/* Gibt eine einzelne Option mit Kuerzel und Namen auf Strom fp aus.*/
{
   if (OptionZugaenglich(i)){
      switch (AufrufOptionen[i].ParameterArt){
      case 0:
         if (AufrufOptionen[i].ParameterTyp.frei.eingestellterWert!=NULL)
            fprintf(fp, "%s %s ",AufrufOptionen[i].Kuerzel, AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
         break;
      case 1:
         fprintf(fp, "%s %s ",AufrufOptionen[i].Kuerzel,(AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert)?" ON":"OFF");
         break;
      case 2:
         fprintf(fp, "%s %30s ",AufrufOptionen[i].Kuerzel,
                 AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert]);
      case 3:
         fprintf(fp, "%s %s%s%s ",
                 AufrufOptionen[i].Kuerzel,
                 AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString,
                 AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?":":"",
                 AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
         break;
      default:
         break;
      }
   }
}

void PA_DefaultsAusgeben(char *Filename)
{
   FILE *fp;
   int i;
   /*   ParameterGruppenT gruppe;*/
   unsigned int gruppe;
   
   if (strcmp (Filename, "stdout") ==0)
      fp = stdout;
   else
      fp = fopen(Filename, "a");
   
   for (gruppe = (int)Arbeitsmodi; gruppe < (int)TechnischeDetails; gruppe++)
      for (i=0; i<OptNrLAST; i++)
         if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i))
            DruckeOptionDefault(fp,i);
   
   fprintf(fp, "\n");
   if (strcmp (Filename, "stdout") != 0)
      fclose(fp);
}

/* ------------------------------------------------------------------------*/
/* -------------------- Sonstige Ausgaben bzgl. Optionen ------------------*/
/* ------------------------------------------------------------------------*/

static void DruckeOption(AufrufOptionenTyp i)
/* Gibt eine einzelne Option ausfuehrlich am Bildschirm aus.*/
{
   int j, sueModule=0;
   
   SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[i].Name);

   switch(i){
   case OptNrNFBildung: 
      NF_ParamInfo(ParamExplain, AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString);
      break;
   case OptNrRMrechtsIR:
      IR_ParamInfo(ParamExplain, AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString);
      break;
   case OptNrKapur:
      U1_ParamInfo(ParamExplain, AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);
      break;
   case OptNrZielbehandlung:
      Z_ParamInfo(ParamExplain, AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPgen:
      KPB_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                    AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPsel:
      HK_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                    AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPInterreduktion:
      KPIR_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                           AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrAusgabeLevel:
     IO_ParamInfo(ParamExplain, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
                          WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
                  ProtokolldateiName);
       break;
   case OptNrOrdnungen:
     ORD_ParamInfo(ParamExplain, AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
     break;
   case OptNrTermpairRepresentation: sueModule++;
   case OptNrPriorityQueue:sueModule++;
   case OptNrCGClassification:sueModule++;
   case OptNrReClassification:sueModule++;
   case OptNrClassification:sueModule++;
   case OptNrSUE:
     SUE_ParamInfo(ParamExplain, sueModule, 
                   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
		   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString,
		   AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
		   AufrufOptionen[OptNrCGInterreduktion].ParameterTyp.InfoString.UserInfoString);
     break;
   case OptNrPCLProt:
#if COMP_PCL
     PCL_ParamInfo(ParamExplain, AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
		   AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);
#else
     fprintf(IO_Strom, "PCL not available in this Waldmeister configuration! \n");
#endif
     break;
   default: IO_DruckeKurzUeberschrift(Zeichenkette, '-');
      break;
   }

   fprintf(IO_Strom, "%s >>> %s.\n", AufrufOptionen[i].Kuerzel, AufrufOptionen[i].Name);
   switch (AufrufOptionen[i].ParameterArt){
   case 0:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an argument. ");
     if (AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben)
       fprintf(IO_Strom, "Argument given is %s.", AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
     else
       fprintf(IO_Strom, "Argument set to %s per default.", 
              AufrufOptionen[i].ParameterTyp.frei.eingestellterWert?
              AufrufOptionen[i].ParameterTyp.frei.eingestellterWert:"(nothing)");
     break;
   case 1:
     fprintf(IO_Strom, "This Option is a switch: ");
     fprintf(IO_Strom, "%s.\n Position: %s\n Alternative: %s\n", 
            AufrufOptionen[i].Bedeutung,
            AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert?
            AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF,
            AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert?
            AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT);
     break;
   case 2:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an argument from ");
     for (j=0; j<HoechsterSichtbarerIndex(i); j++){
        fprintf(IO_Strom, " %s",AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[j]);
        if (j==AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert)
           fprintf(IO_Strom, "*");
        if (j+1<HoechsterSichtbarerIndex(i))
           fprintf(IO_Strom, ", ");
     }
     fprintf(IO_Strom, ".\n");
     break;
  case 3:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an Info-String (opt1:opt2:opt3=val:...)\n");
     fprintf(IO_Strom, "Default-Info-String is %s,\n", AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
     fprintf(IO_Strom, "Your input was %s.\n", 
             AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?"nothing":AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString);
     break;
  default:
     break;
  }
   fprintf(IO_Strom, "Option is %s and has influence %s.\n\n\n",
           AufrufOptionen[i].IsXPRT?"available only in expert-mode":"always available",
           AufrufOptionen[i].NurVervollstaendigung?"only on the modes proof and completion":"on all modes");
}


static void DruckeOptionKurz(AufrufOptionenTyp i)
/* Gibt eine einzelne Option mit Kuerzel und Namen am Bildschirm aus.*/
{
   int j;
   
   fprintf(IO_Strom, "%6s  >>>  %s",AufrufOptionen[i].Kuerzel+1, AufrufOptionen[i].Name);
   switch (AufrufOptionen[i].ParameterArt){
   case 0:
      if (AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben)
         fprintf(IO_Strom, ". Value is %s.\n", AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
      else
         fprintf(IO_Strom, ". Default is %s.\n", 
                 AufrufOptionen[i].ParameterTyp.frei.eingestellterWert?
                 AufrufOptionen[i].ParameterTyp.frei.eingestellterWert:"(nichts)");
      break;
   case 1:
      fprintf(IO_Strom, ". Schalter-Stellung: %s.\n",(AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert)?
              AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF);
      break;
   case 2:
      fprintf(IO_Strom, " (Values: ");
      for (j=0; j<HoechsterSichtbarerIndex(i); j++){
         fprintf(IO_Strom, "%s",AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[j]);
         if (j==AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert)
            fprintf(IO_Strom, "*");
         if (j+1<HoechsterSichtbarerIndex(i))
            fprintf(IO_Strom, ", ");
      }
      fprintf(IO_Strom, ").\n");
      break;
   case 3:
      fprintf(IO_Strom, ". Setting: %s%s%s.\n", 
              AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString,
              AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?":":"",
              AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
      break;
   default:
      break;
   }
}


/* ------------------------------------------------------------------------*/
/* -------------------- Option per Kuerzel waehlen ------------------------*/
/* ------------------------------------------------------------------------*/

static int OptionMitKuerzelWaehlen(BOOLEAN NurErklaeren)
/* Eingabe einer Option, die durch das Kuerzel fn angegeben wurde. Beruecksichtigt
   werden nur aktive Optionen, die bzgl. XPRT-Modus zugaenglich sind.
   Bei Eingabe einer Pseudo-Option, deren Behandlung bereits hier erledigt wurde, wird -1 zurueckgegeben.
   Bei Eingabe von c oder q wird abgebrochen, in diesem Falle wird -2 zurueckgeliefert.
*/
{
   int option = 0;
   BOOLEAN Gefunden;

   do{
      fprintf(IO_Strom, "\nChoose option, continue with 'c'\n>>> ");
      fgets(Zeichenkette, INPUT_STRLEN, stdin);
      chop (Zeichenkette);
      fprintf(IO_Strom, "\n");
      if (strcmp("c", Zeichenkette)==0){
         Gefunden = TRUE;
         option = -2;
      }
      else if (strcmp("def", Zeichenkette)==0){
         Gefunden = TRUE;
         option = -1;
         if (NurErklaeren)
            fprintf(IO_Strom, "Current parameter setting as comand line.\n");
         else
            PA_DefaultsAusgeben("stdout");
         IO_WarteAufReturnMitText("RETURN");
      }
      else if (XPRTModusSichtbar && (strcmp("x", Zeichenkette)==0)){
         Gefunden = TRUE;
         option = -1;
         if (NurErklaeren){
            fprintf(IO_Strom, "Turn expert-mode on resp. off.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
            XPRTModusWechseln();
            fprintf(IO_Strom, "EXPERT-MODE  : %s\n", XPRTModus?"On":"Off");
         }
      }
      else if (strcmp("exp", Zeichenkette)==0){
         Gefunden = FALSE;
         if (NurErklaeren)
            fprintf(IO_Strom, "Unknown option.\n");
         else{
            do{
               fprintf(IO_Strom, "\nEnter option to be explained.\n>>> ");
               fgets(Zeichenkette, INPUT_STRLEN, stdin);
	       chop (Zeichenkette);
               fprintf(IO_Strom, "\n");
               option=0; 
               do
                  if ((strcmp((AufrufOptionen[option].Kuerzel+1), Zeichenkette) == 0) && OptionZugaenglich(option))
                     Gefunden = TRUE;
               while((!Gefunden)&&(++option<OptNrLAST));
               if (Gefunden) DruckeOption(option);
               else fprintf(IO_Strom, "Unknown option.\n");
            }while(!Gefunden);
            option = -1;
            Gefunden = TRUE;
            IO_WarteAufReturnMitText("RETURN");
         }
      }
      else if (strcmp("view", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "View currently selected specification.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else
            IO_DateiAusgabe(IO_Spezifikationsname(), TRUE);

         option = -1;
         Gefunden = TRUE;
      }
      else if (strcmp("setup", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "Show current setup.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
           IO_DruckeUeberschrift("SETUP");
           PA_DruckeModi();
           IO_WarteAufReturnMitText("RETURN");
         }
         option = -1;
         Gefunden = TRUE;
      }
      else if (strcmp("ia", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "Turn on resp. off interactivity (File-, Options-, Interrupt- and result-menus).\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
            PA_InteraktionAn = !PA_InteraktionAn;
#if !COMP_POWER
            fprintf(IO_Strom, "OPTION    -ia: %s\n", PA_InteraktionAn?"On":"Off");
#endif
         }
         Gefunden = TRUE;
         option = -1;
      }
      else if (strcmp("optfile", Zeichenkette)==0){
         if (NurErklaeren){
           fprintf(IO_Strom, "Parse more options from file.\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
           fprintf(IO_Strom, "Sorry, no longer available at this stage of computation.\n");
         }
         Gefunden = TRUE;
         option = -1;
      }
        
      else if (strcmp("quit", Zeichenkette)==0){
         if (NurErklaeren){
           fprintf(IO_Strom, "Quits the run. No further actions take place.\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
           exit(0);
         }
         Gefunden = TRUE;
         option = -1;
      }
        
      else{
         option=0;
         Gefunden = FALSE;
         do
            if ((strcmp((AufrufOptionen[option].Kuerzel+1), Zeichenkette) == 0) && OptionZugaenglich(option))
               Gefunden = TRUE;
         while((!Gefunden)&&(++option<OptNrLAST));
      }
      if (!Gefunden)
         fprintf(IO_Strom, "Unknown option.\n");
   }while(!Gefunden);
   return option;
}


/* ------------------------------------------------------------------------*/
/* -------------------- Einstellen von Optionen ---------------------------*/
/* ------------------------------------------------------------------------*/

static BOOLEAN NeuerWertVonToggle(BOOLEAN AlterWert, char *Argument, BOOLEAN *ArgumentVerbraucht,
                                  BOOLEAN EndeGelaende)
/* Liefert den neuen Wert des Toggles in Abhaengigkeit von Argument. */
{
   *ArgumentVerbraucht = FALSE;

   if (EndeGelaende) return !AlterWert;

   if (Argument!=NULL){
      if ((strcasecmp(Argument,"an")==0)||
          (strcasecmp(Argument,"on")==0)) {
         AlterWert = TRUE;
         *ArgumentVerbraucht = TRUE;
      }
      else if ((strcasecmp(Argument,"aus")==0)||
               (strcasecmp(Argument,"off")==0)){
         AlterWert = FALSE;
         *ArgumentVerbraucht = TRUE;
      }
      else
         AlterWert = !AlterWert;
   }
   else 
      AlterWert = !AlterWert;      
   return AlterWert;
}

static BOOLEAN OptionEingestellt(AufrufOptionenTyp OptNr, BOOLEAN ArgumentAlsIntErlaubt, BOOLEAN AnAus, 
                                 char *Argument, BOOLEAN *ArgumentVerbraucht, BOOLEAN EndeGelaende)
{
   BOOLEAN Gefunden = FALSE;

   *ArgumentVerbraucht = TRUE;
#if !COMP_POWER
   fprintf(IO_Strom, "  OPTION %8s: ", AufrufOptionen[OptNr].Kuerzel);
#endif

   switch (AufrufOptionen[OptNr].ParameterArt){
   case 0:
      if (Argument){
         if (AufrufOptionen[OptNr].ParameterTyp.frei.wurdeAngegeben)
            our_dealloc (AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert);
         AufrufOptionen[OptNr].ParameterTyp.frei.wurdeAngegeben = TRUE;
         AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert = (char *) our_alloc(1+strlen(Argument));
         strcpy(AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert, Argument);
#if !COMP_POWER
         fprintf(IO_Strom, "%s ", AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert);
#endif
         Gefunden = TRUE;
      }
      else{
         fprintf(IO_Strom, "*** Missing argument ");
         *ArgumentVerbraucht = FALSE;
      }
#if !COMP_POWER
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
#endif
      break;
   case 1:
      AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert = 
         NeuerWertVonToggle(AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert, 
                            AnAus?Argument:NULL, ArgumentVerbraucht, EndeGelaende);
      Gefunden = TRUE;
#if !COMP_POWER
      fprintf(IO_Strom, "%s ",(AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert)?
             AufrufOptionen[OptNr].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[OptNr].ParameterTyp.einstellig.BedeutungF);
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
#endif
      break;
   case 2:
   {
      int wert , i;
      if (ArgumentAlsIntErlaubt && ((wert= atoi(Argument))>0) &&
          (wert<=HoechsterSichtbarerIndex(OptNr))){
         AufrufOptionen[OptNr].ParameterTyp.zweistellig.eingestellterWert = wert-1;
         Gefunden = TRUE;
      }
      else
         if (Argument)
            for (i=0; i<HoechsterSichtbarerIndex(OptNr); i++)
               if (strstr(AufrufOptionen[OptNr].ParameterTyp.zweistellig.WerteImKlartext[i], Argument)){
                  AufrufOptionen[OptNr].ParameterTyp.zweistellig.eingestellterWert = i;
                  Gefunden = TRUE;
                  break;
               }
      if (Gefunden) {
#if !COMP_POWER
         fprintf(IO_Strom, "%s ", AufrufOptionen[OptNr].
                ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[OptNr].
                                                         ParameterTyp.zweistellig.eingestellterWert]);
#else
         ;
#endif
      }
      else if (Argument){
         *ArgumentVerbraucht = FALSE;
         fprintf(IO_Strom, "%s *** unknown value ", Argument);
      }
      else{
         *ArgumentVerbraucht = FALSE;
         fprintf(IO_Strom, "*** missing argument ");
      }

      if (OptNr == OptNrAusgabeLevel)
        IO_AusgabeLevelBesetzen(atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
           WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]));

#if !COMP_POWER
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
#endif      
      break;
   }
   case 3:
      if (Argument){
         if (strlen (AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString)) {
            AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString =
               (char *) our_realloc(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString,
                                    3+strlen(Argument)+
                                    strlen(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString),
                                    "OptionEingestellt: Str erweitern");
            strcat(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString, ":");            
            strcat(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString, Argument);
         }
         else {
            AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString = (char *) our_alloc(1+strlen(Argument));
            strcpy(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString, Argument);
         }
#if !COMP_POWER
         fprintf(IO_Strom, "%s ", AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString);
#endif
         Gefunden = TRUE;
      }
      else{
         fprintf(IO_Strom, "*** missing argument ");
         *ArgumentVerbraucht = FALSE;
      }
#if !COMP_POWER
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
#endif
      break;
      
   default:
      *ArgumentVerbraucht = FALSE;
      our_error("\nError in SetOption.\n\n");
      break;
   }
   return Gefunden;
}
   

/* ------------------------------------------------------------------------*/
/* -------------------- Konsistenztest fuer Einstellung -------------------*/
/* ------------------------------------------------------------------------*/

static BOOLEAN ParameterKonsistent(void)
/* Testet die innere Konsistenz bzw. Gueltigkeit der Aufruf-Parameter.
   Ferner werden hier voneinander abhaengige Parameter gesetzt.
*/
{
   if (Vervollstaendigung){

/*      PA_KapurR = (BOOLEAN) strstr(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert, "r");
      PA_KapurG = (BOOLEAN) strstr(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert, "g");
      PA_Kapur = PA_KapurR | PA_KapurG;*/
   
      VerwendungVonGleichungen = !PA_FailingCompletion;
   }
   else {                       /* Einfache Modi: keine KP-Behandlung bei ihrer Entstehung, keine Interreduktion */

/*      PA_RMrechtsB  = PA_RMrechtsBR = PA_RMrechtsBG = FALSE;*/
      
      VerwendungVonGleichungen = FALSE;
   }

   return TRUE;
}

/* ------------------------------------------------------------------------*/
/* -------------------- Parameter-Einstellung -----------------------------*/
/* ------------------------------------------------------------------------*/

static void PA_ParameterAuswaehlen(void)
/* Manuelles Einstellen der Parameter (menuegesteuert!!!!)*/
{
   int i, eingabe;
   unsigned int gruppe;
   BOOLEAN Booly;

   do{
      IO_DruckeUeberschrift("Waldmeister Interactive Parameter Menue");
      for (gruppe = (int)Arbeitsmodi; gruppe <= (int)ZusatzOptionen; gruppe++){
         BOOLEAN ZugaenglicheOptionGefunden = FALSE;
         for (i=0; i<OptNrLAST; i++)
            if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i)){
               ZugaenglicheOptionGefunden = TRUE;
               break;
            }
         if (ZugaenglicheOptionGefunden){
            IO_DruckeKurzUeberschrift(GruppenNamen[gruppe], '=');
            for (i=0; i<OptNrLAST; i++)
               if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i))
                  DruckeOptionKurz(i);
         }
      }
      IO_DruckeKurzUeberschrift("Sonstiges", '=');
      fprintf(IO_Strom, "    def >>> current parameter-setting as command-line\n");
      if (XPRTModusSichtbar)
         fprintf(IO_Strom, "      x >>> turn expert-mode %s\n", XPRTModus?"off":"on");
      fprintf(IO_Strom, "     ia >>> turn interactivity %s\n", PA_InteraktionAn?"off":"on");
      fprintf(IO_Strom, "    exp >>> explain option\n");
      fprintf(IO_Strom, "   view >>> view selected spezification\n");
      fprintf(IO_Strom, "  setup >>> Show current parameter-setup\n");
      fprintf(IO_Strom, "   quit >>> Quit Waldmeister.\n");
      eingabe = OptionMitKuerzelWaehlen(FALSE);
      if (eingabe >= 0){
         if (!AufrufOptionen[eingabe].NurVervollstaendigung || Vervollstaendigung){
            switch (AufrufOptionen[eingabe].ParameterArt){
            case 0:
            case 3:
               SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[eingabe].Name);
               fprintf(IO_Strom, "\nEnter argument for option '%s': ", Zeichenkette);
               fgets(Zeichenkette, INPUT_STRLEN, stdin);
	       chop (Zeichenkette);
            case 1:
               OptionEingestellt(eingabe, FALSE, FALSE, Zeichenkette, &Booly, FALSE);
               break;
            case 2:
               do{
                  SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[eingabe].Name);
                  fprintf(IO_Strom, "\n%s (currently %s): \n",
                         Zeichenkette, 
                         AufrufOptionen[eingabe].ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[eingabe].ParameterTyp.zweistellig.eingestellterWert]);
                  for (i=1; i<=HoechsterSichtbarerIndex(eingabe); i++)
                     fprintf(IO_Strom, "%2d. %s\n", i, AufrufOptionen[eingabe].ParameterTyp.zweistellig.WerteImKlartext[i-1]);
                  fprintf(IO_Strom, ">>> ");
		  fgets(Zeichenkette, INPUT_STRLEN, stdin);
		  chop (Zeichenkette);
                  fprintf(IO_Strom, "\n");
               }while(!OptionEingestellt(eingabe, TRUE, FALSE, Zeichenkette, &Booly,FALSE));
               break;
            default:
               our_error("\nFauler Fehler in Optionen eingeben.\n\n");
               break;
            }
         }
         else{
            fprintf(IO_Strom, "Option will be ignored in this mode.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
      }
      if (!ParameterKonsistent()){
         IO_WarteAufReturnMitText("Please correct.");
         eingabe = 1;           /* Damit nicht abgebrochen wird */
      }
   }while  (eingabe >= -1);
}   

BOOLEAN PA_Vorbehandeln(int argc, char *argv[])
/* - Einstellung der Pseudo-Parameter, soweit moeglich
   - Einlesen von Optionen aus Datei
   Rueckgabe ist FALSE, wenn das Vorbehandeln bereits einen Abbruchgrund lieferte.
   Die vernutzten Argumente werden auf NULL gesetzt.
*/
{
   BOOLEAN Fehler = FALSE, protdateiAngegeben = FALSE;
   BOOLEAN MenueVerlangt = FALSE;
   int pos=1;

   /* argv[0] == Programmname,  deshalb ist argc >= 2 (sonst wird diese Prozedur nicht aufgerufen aus main.c) */

   while (pos<argc){
      if (strcmp(argv[pos], "-def") ==0){
         if (pos+1<argc)
            PA_DefaultsAusgeben(argv[pos+1]);
         else
            PA_DefaultsAusgeben("stdout");
         exit(0);
      }
      else if ((strcmp(argv[pos], "-ia") == 0)){                        /* Interaktion verlangt? */
         BOOLEAN Verbraucht;
         PA_InteraktionAn = NeuerWertVonToggle(PA_InteraktionAn, argv[pos+1], &Verbraucht, (pos+1>=argc));
         argv[pos] = NULL;
         if (Verbraucht){
            pos++;
            argv[pos] = NULL;
         }
         pos++;
      }
      else if ((strcmp(argv[pos], "-menue") == 0) ||                            /* Menue verlangt? */
          (strcmp(argv[pos], "menue") == 0) ||
          (strcmp(argv[pos], "-m") == 0)) {
         MenueVerlangt = TRUE;
         pos++;
      }
      else if ((strcmp(argv[pos], "-x") == 0) && XPRTModusSichtbar){
         BOOLEAN ArgVerbraucht;
         XPRTModus = NeuerWertVonToggle(XPRTModus, (pos<argc-1)?argv[pos+1]:NULL, &ArgVerbraucht, (pos+1>=argc));
         argv[pos]=NULL;
         if (ArgVerbraucht){
            pos++;
            argv[pos]=NULL;
         }
         pos++;
      }
      else if ((strcmp(argv[pos], "-prot") == 0)){
         argv[pos]=NULL;
         if (pos+1<argc){
            ProtokolldateiName = argv[pos+1];
            pos++;
            argv[pos]=NULL;
            protdateiAngegeben = TRUE;
         }
         pos++;
      }
      /* Die folgenden Optionen muessen bereits hier geparst werden, damit -auto funktioniert */
      else if (strcmp(argv[pos], "-auto") == 0) {
         BOOLEAN Verbraucht;
         PA_Automodus = NeuerWertVonToggle(PA_Automodus, argv[pos+1], &Verbraucht, (pos+1>=argc));
         argv[pos] = NULL;
         if (Verbraucht){
            pos++;
            argv[pos] = NULL;
         }
         pos++;
      }
      else if (strcmp (argv[pos], "-symbnorm") == 0) {
         BOOLEAN Verbraucht;
         OptionEingestellt (OptNrSymbnorm, FALSE, TRUE, argv[pos+1], &Verbraucht, (pos+1 >= argc));
         argv[pos] = NULL;
         if (Verbraucht){
            pos++;
            argv[pos] = NULL;
         }
         pos++;
      }
      else                      /* sonst: Weiterlaufen */
         pos++;
   }
   
   if (MenueVerlangt)
      PA_InteraktionAn = TRUE;

   if (protdateiAngegeben) PA_InteraktionAn = FALSE;

   if (!isatty(STDIN_FILENO))/* 960825 AJ: Interaktion macht nur Sinn, wenn Benutzer tippen kann */
     PA_InteraktionAn = FALSE;
   
   /* !protdateiAngegeben ist wichtig -> siehe Implmentierung IO_AusgabeStromAus/IO_InitAposteriori
      Das IO_Strom Konzept ist broken. AJ.
    */
   if (!PA_InteraktionAn && !protdateiAngegeben)
     IO_AusgabeStromAus ();

   if (PA_InteraktionAn) {
     if (XPRTModus)
       fprintf(IO_Strom, "  EXPERT-MODE  : On\n");
     if (protdateiAngegeben)
       fprintf(IO_Strom, "  PROTOCOL     : %s\n", ProtokolldateiName);
   }

   return Fehler;
}


BOOLEAN PA_InitAposterioriEarly (void)
{
   SO_InitAposteriori(AufrufOptionen[OptNrSymbnorm].ParameterTyp.einstellig.eingestellterWert);
   TO_InitAposteriori();
   IO_InitAposterioriEarly ();
   return TRUE;
}


BOOLEAN PA_AufrufParameterEinstellen (int argc, char *argv[], mode Modus)
/* Einstellen der beim Aufruf mitgegebenen Parameter. argc und argv sind die Parameter von main,
   es wird also davon ausgegangen, dass argv[0] den Programmnamen enthaelt, und das argv[argc] der
   Name der zu bearbeitenden Spezifikation ist. Die Angaben zwischen argv[1] und argv[argc-1] werden
   ausgewertet. Dabei werden die in Parameter.c spezifizierten Aufruf-Optionen herangezogen.
   Inkorrekte Parameter sowie der Filename werden ueberlesen.
   Auf Anforderung von Hilfe sowie Menue wird reagiert.
   Rueckgabe ist, ob - ggf. mit Menue-Hilfe - eine korrekte Einstellung vorgenommen werden konnte.
   Vor.: Die Spezifikation befindet sich bereits im Speicher.
*/
{
   BOOLEAN Gefunden = FALSE, PseudoOption, ArgumentVerbraucht, MenueVerlangt = FALSE;
   int pos=1, option=0, aktuelleOption = 0, AnzahlFehler = 0;
   
   /* argv[0] == Programmname,  deshalb ist argc >= 2 (sonst wird diese Prozedur nicht aufgerufen aus main.c) */

   PA_ExtModus = Modus;

   Vervollstaendigung = (PA_ExtModus == proof) || (PA_ExtModus == completion);

   if (Vervollstaendigung && !VervollstaendigungZugaenglich){
      fprintf(IO_Strom, "Completion not available!\n\n");
      return FALSE;
   }

   PA_Ordering = ord_name(get_ordering());


#if !COMP_POWER
   fprintf(IO_Strom, "\nparsing options\n");
#endif

   while (pos<argc){

      /* Pseudo-Optionen finden */
      /* ---------------------- */

      PseudoOption = Gefunden = TRUE;
      if (argv[pos] == NULL){
         pos++;
         continue;
      }
      if ((strcmp(argv[pos], "-menue") == 0) ||                            /* Menue verlangt? */
          (strcmp(argv[pos], "menue") == 0) ||
          (strcmp(argv[pos], "-m") == 0)){
         fprintf(IO_Strom, "  MENUE\n");
         MenueVerlangt = TRUE;
      }
      else{                                                                /* sonst: Option suchen */

         /* Richtige Option finden */
         /* ---------------------- */

         PseudoOption = Gefunden = FALSE;
         for (Gefunden = FALSE, option=0; (!Gefunden)&&(option<OptNrLAST); option++)
            if ((strcmp(AufrufOptionen[option].Kuerzel, argv[pos]) == 0) && (OptionZugaenglich(option))){
               Gefunden = TRUE;
               aktuelleOption = option;
            }
      }

      pos++;                                       /* pos steht jetzt auf dem naechsten Element der Kommando-Zeile */

      /* unbekannte Option */
      /* ----------------- */

      if (!Gefunden){
         fprintf(IO_Strom, "Unknown option or invalid filename: %s\n", argv[pos-1]);
         AnzahlFehler++;
      }
      else if (!PseudoOption) {

         /* Richtige Option einstellen */
         /* -------------------------- */
         
         if (!OptionEingestellt(aktuelleOption, FALSE, TRUE, argv[pos], &ArgumentVerbraucht, (pos+1>=argc))) 
            AnzahlFehler++;  /* Bei Fehler wird kein Argument vernutzt! */
         else if (ArgumentVerbraucht) 
            pos++;
      }
   }
   
#if !COMP_POWER
   fprintf(IO_Strom, "... done\n");
#endif


   PA_FailingCompletion = AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert;
   PA_EinsSternAn = AufrufOptionen[OptNrEinsStern].ParameterTyp.einstellig.eingestellterWert;
   PA_NusfuAn = AufrufOptionen[OptNrNusfu].ParameterTyp.einstellig.eingestellterWert;
   PA_KernAn = AufrufOptionen[OptNrKern].ParameterTyp.einstellig.eingestellterWert;
   PA_GolemAn = AufrufOptionen[OptNrGolem].ParameterTyp.einstellig.eingestellterWert;
   PA_BackAn = AufrufOptionen[OptNrBack].ParameterTyp.einstellig.eingestellterWert;
   PA_WWWAn = AufrufOptionen[OptNrWWW].ParameterTyp.einstellig.eingestellterWert;

   IO_InitAposteriori(atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);

   if ((!ParameterKonsistent()) || (AnzahlFehler>0) || MenueVerlangt) {
      if (PA_InteraktionAn){
        IO_DruckeKopf();             /* Programm-Kopf ausgeben */
        IO_DruckeBeispiel();
        PA_ParameterAuswaehlen();
        return TRUE;
      }
      else
         return FALSE;
   }
   else
      return TRUE;
   
}


