/* -------------------------------------------------------*/
/* ------------- Includes --------------------------------*/
/* -------------------------------------------------------*/

#include "antiwarnings.h"
#if defined (SUNOS5)
#include <fcntl.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#endif
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/utsname.h>

#include "compiler-extra.h"
#include "Ausgaben.h"
#include "CritGoals.h"
#include "DSBaumAusgaben.h"
#include "DSBaumKnoten.h"
#include "DSBaumOperationen.h"                                                   /* wegen Ausdehnung Variablenbank */
#include "FehlerBehandlung.h"
#include "general.h"
#include "Interreduktion.h"
#include "KPMengenInterreduktion.h"
#include "MatchOperationen.h"                                                    /* wegen Ausdehnung Variablenbank */
#include "MNF.h"
#include "NFBildung.h"                                                      /* wegen Ausdehnung des Reduktionspfades */
#include "Parameter.h"
#include "parser.h"
#include "pcl.h"
#include "RUndEVerwaltung.h"
#include "SelectClassify.h"
#include "Signale.h"
#include "Statistiken.h"
#include "StrahlMemorium.h"
#include "SUEM.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "TermIteration.h"
#include "TermVerfolgung.h"
#include "Unifikation1.h"                                                 /* wegen Ausdehnung Variablenbank, Stack */
#include "Unifikation2.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"
#include "version.h"

extern char **environ;


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Interne Funktionen ++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct SPTS{
   char strength[6];
   unsigned long int sp;
} StichpunktTyp;

static StichpunktTyp *SPArray;

static int SPArrayPos=0, SPArrayLen=0;

static void StichpunktMerken(unsigned long int sp, char *strength)
/* Stichpunkte der wahlfreien KP-Mengen-Interreduktion merken fuer spaetere Ausgabe. */
{
   if (SPArrayPos == SPArrayLen-1){
      SPArray = our_realloc(SPArray, (SPArrayLen+10) * sizeof(StichpunktTyp), "IO:Stichpunkt-Array");
      SPArrayLen +=10;
   }
   strncat(SPArray[SPArrayPos].strength, strength, 5);
   SPArray[SPArrayPos++].sp = sp;
   
}


/*
  Schreiben der gesamten Regel/Gleichungsmenge als Spezifikation auf Platte
  */

static void SpezifikationSchreiben (void)
{
   char            AusgabeDatei[INPUT_STRLEN];
   mode            Mode = get_mode();
   SymbolT         symbol;
   

   fprintf(IO_Strom, "\nName of file to write specification to: ");
   fgets(AusgabeDatei, INPUT_STRLEN, stdin);
   chop(AusgabeDatei);

   IO_AusgabeStromUmlenken(AusgabeDatei);

   fprintf (IO_Strom, "NAME\t\t%s\n\n", get_spec_name());
   fprintf (IO_Strom, "MODE\t\t");
   switch (Mode) {
   case completion:
      fprintf(IO_Strom,"COMPLETION");
      break;
   case reduction:
      fprintf(IO_Strom,"REDUCTION");
      break;
   case confluence:
      fprintf(IO_Strom,"CONFLUENCE");
      break;
   case proof:
      fprintf(IO_Strom,"PROOF");
      break;
   case termination:
      fprintf(IO_Strom,"TERMINATION");
      break;
   case convergence:
      fprintf(IO_Strom,"CONVERGENCE");
      break;
   default:
      our_error("unexpected label in SpezifikationSchreiben");
      break;
   }
   fprintf (IO_Strom, "\n\n");
   
   fprintf (IO_Strom, "SORTS\t");
   printSorts (IO_Strom);
   fprintf (IO_Strom, "\nSIGNATURE\n");
   printSig (IO_Strom);
   fprintf (IO_Strom, "ORDERING\t");
   switch (PA_Ordering) {
   case ordering_kbo: fprintf (IO_Strom, "KBO\n");break;
   case ordering_lpo: fprintf (IO_Strom, "LPO\n");break;
   default:
      our_error("unexpected label in SpezifikationSchreiben");
      break;
   }
   
   printOrdering (IO_Strom);

   fprintf (IO_Strom, "VARIABLES\n");
   fprintf (IO_Strom, "\t");
   IO_DruckeSymbol (-1);
   for (symbol = -2; symbol >= BO_NeuesteBaumvariable(); symbol--) {
      fprintf (IO_Strom, ", ");
      IO_DruckeSymbol (symbol);
   }
   fprintf (IO_Strom, " : ");
   /* wir geben allen Variablen die gleiche Sorte, nicht schoen,
      aber anscheinend einzige Moeglichkeit, da Variablen keine
      Sorte haben - hier geht Information verloren */
   printSort (IO_Strom, 1);
   fprintf (IO_Strom, "\n\n");
   

   fprintf (IO_Strom, "EQUATIONS\n");
   if (!RE_RegelmengeLeer())
      RE_AusgabeRegelmenge (0);
   if (!RE_GleichungsmengeLeer())
      RE_AusgabeGleichungsmenge(0);
   
   fprintf (IO_Strom, "CONCLUSION\n");
   Z_AusgabeZielmenge (Z_AusgabeSpez);
   IO_AusgabeStromZuruecksetzen();
}

/* -------------------------------------------------------*/
/* -------------- Initialisierung ------------------------*/
/* -------------------------------------------------------*/

static InfoStringT AutoErgDEF, AutoErgUSR;

static BOOLEAN ST_SchonInitialisiert = FALSE;

void ST_InitAposteriori(InfoStringT KeDEF, InfoStringT KeUSR)
{
  if (!ST_SchonInitialisiert){
     SPArrayLen = 10;              /* Stichpunkt-Array initialisieren. */
     SPArray = our_alloc(SPArrayLen*sizeof(StichpunktTyp));
  }
  SPArrayPos = 0;
  AutoErgDEF = KeDEF; 
  AutoErgUSR = KeUSR;
  ST_SchonInitialisiert = TRUE;
}

/* -------------------------------------------------------*/
/* -------------- Statistiken der einzelnen Module -------*/
/* -------------------------------------------------------*/

static void ST_Statistik(const InfoStringT is)
{
   SUE_Statistics(is);
   IR_Statistik(is);
   KPIR_Statistik(is);
   DM_Statistik(is);
   RE_Statistik(is);
   U1_Statistik(is);
   U2_Statistik(is);
   NF_Statistik(is);
   MO_Statistik(is);
   BO_Statistik(is);
   SO_Statistik(is);
   Z_Statistik(is);
   BK_Statistik(is);
   MNF_Statistik(is);
   SC_Statistik(is);
}

/* -------------------------------------------------------*/
/* -------------- Prozess-Statistik ----------------------*/
/* -------------------------------------------------------*/

static void ST_PlatzverbrauchAusgeben(BOOLEAN Kopf)
     /* Ausgabe des Platzverbrauchs im Latex-Tabellenformat */
{
#if defined (SUNOS4) || defined (LINUX)
  if (Kopf) fprintf(IO_Strom, "max.res.mem & ");
  else{
    struct rusage rusageBuf;
    
    if (getrusage(RUSAGE_SELF, &rusageBuf) == -1) fprintf(IO_Strom, " <error> &");
    else IO_DruckeStatistikZeile("bla",rusageBuf.ru_maxrss * getpagesize());
  }
#elif defined (SUNOS5)
  if (Kopf) fprintf(IO_Strom, "process size & res.mem &");
  else{
    int             fd;
    char            proc[40];
    prusage_t       prusage;
    prpsinfo_t      prpsinfo;
    
    sprintf(proc,"/proc/%d", (int)getpid());
    if (((fd = open(proc,O_RDONLY)) == -1) ||
        (ioctl(fd, PIOCUSAGE, &prusage) == -1) ||
        (ioctl(fd, PIOCPSINFO, &prpsinfo) == -1))
      fprintf(IO_Strom, " <error> & <error> &");
    else
      fprintf(IO_Strom, " %ld & %ld &", prpsinfo.pr_size*sysconf(_SC_PAGESIZE),
              prpsinfo.pr_rssize *sysconf(_SC_PAGESIZE));
  }
#else
#error "ST_PlatzverbrauchAusgeben muss angepasst werden!"      
#endif
}

void ST_ProzessStatistik(void)
/* Ausgabe von Prozess-Daten (aus getrusage)  */
{

#if defined (SUNOS4) || defined (LINUX)
   struct rusage rusageBuf;
   
   if (getrusage(RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("IO_ProzessStatistik: getrusage fehlgeschlagen");
   
   IO_DruckeUeberschrift("Prozess-Statistik");
   IO_DruckeStatistikZeile("maximum resident set size/pages", (unsigned long int)rusageBuf.ru_maxrss);
   IO_DruckeStatistikZeile("system pagesize (may differ from HW-pagesize)",(unsigned long int) getpagesize());
   IO_DruckeStatistikZeile("==> maximum resident amount of memory (bytes)", rusageBuf.ru_maxrss * getpagesize());
   IO_DruckeStatistikZeile("memory in use by the process while running (integral value: pages * clock ticks)", (unsigned long int) rusageBuf.ru_idrss);
   IO_DruckeStatistikZeile("page faults serviced without any physical I/O activity", (unsigned long int) rusageBuf.ru_minflt);
   IO_DruckeStatistikZeile("page faults serviced with physical I/O activity", (unsigned long int) rusageBuf.ru_majflt);
   IO_DruckeStatistikZeile("times process was swapped out of main memory", (unsigned long int) rusageBuf.ru_nswap);
   IO_DruckeStatistikZeile("file system inputs (read)", (unsigned long int) rusageBuf.ru_inblock);
   IO_DruckeStatistikZeile("file system outputs (write)", (unsigned long int) rusageBuf.ru_oublock);
   IO_DruckeStatistikZeile("messages sent over sockets", (unsigned long int) rusageBuf.ru_msgsnd);
   IO_DruckeStatistikZeile("messages received from sockets", (unsigned long int) rusageBuf.ru_msgrcv);
   IO_DruckeStatistikZeile("signals delivered", (unsigned long int) rusageBuf.ru_nsignals);
   IO_DruckeStatistikZeile("context switches due to voluntarily giving up the processor before time slice completed", (unsigned long int) rusageBuf.ru_nvcsw);
   IO_DruckeStatistikZeile("context switches due to a higher priority process or exceeded time slice", (unsigned long int) rusageBuf.ru_nivcsw);
#elif defined (SUNOS5)
   int             fd;
   char            proc[40];
   prusage_t       prusage;
   prpsinfo_t      prpsinfo;
   
   sprintf(proc,"/proc/%d", (int)getpid());
   if ((fd = open(proc,O_RDONLY)) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open /proc fehlgeschlagen");
      }
   if (ioctl(fd, PIOCUSAGE, &prusage) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open prusage-ioctl fehlgeschlagen");
   }
   if (ioctl(fd, PIOCPSINFO, &prpsinfo) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open prpsinfo-ioctl fehlgeschlagen");
   }
   IO_DruckeUeberschrift("Prozess-Statistik");
   IO_DruckeStatistikZeile("system pagesize (may differ from HW-pagesize)",(unsigned long int) sysconf(_SC_PAGESIZE));
   IO_DruckeStatistikZeile("size of process /pages", (unsigned long int)prpsinfo.pr_size);
   IO_DruckeStatistikZeile("==> process size (bytes)", prpsinfo.pr_size*sysconf(_SC_PAGESIZE) );
   IO_DruckeStatistikZeile("resident set size/pages", (unsigned long int)prpsinfo.pr_rssize);
   IO_DruckeStatistikZeile("==> resident amount of memory (bytes)", prpsinfo.pr_rssize *sysconf(_SC_PAGESIZE) );
   IO_DruckeStatistikZeile("page faults serviced without any physical I/O activity", (unsigned long int) prusage.pr_minf);
   IO_DruckeStatistikZeile("page faults serviced with physical I/O activity", (unsigned long int) prusage.pr_majf);
   IO_DruckeStatistikZeile("times process was swapped out of main memory", (unsigned long int) prusage.pr_nswap);
   IO_DruckeStatistikZeile("file system inputs (read)", (unsigned long int) prusage.pr_inblk);
   IO_DruckeStatistikZeile("file system outputs (write)", (unsigned long int) prusage.pr_oublk);
   IO_DruckeStatistikZeile("messages sent over sockets", (unsigned long int) prusage.pr_msnd);
   IO_DruckeStatistikZeile("messages received from sockets", (unsigned long int) prusage.pr_mrcv);
   IO_DruckeStatistikZeile("signals delivered", (unsigned long int) prusage.pr_sigs);
   IO_DruckeStatistikZeile("context switches due to voluntarily giving up the processor before time slice completed", (unsigned long int) prusage.pr_vctx);
   IO_DruckeStatistikZeile("context switches due to a higher priority process or exceeded time slice", (unsigned long int) prusage.pr_ictx);
#else
#error "ST_ProzessStatistik muss angepasst werden!"      
#endif
}

typedef enum {Finished, Timeout, Error, Interrupt, NoMemory, GoalProved} ReasonT;


static void STEXZeileAusgeben(ReasonT Situation)
{
  IO_StatistikModusAufLatexSetzen();
  fprintf(IO_Strom, " problem & ");
  ST_Statistik("all:slatex-head");
  ST_PlatzverbrauchAusgeben(TRUE);
  fprintf(IO_Strom, " time & Result \\\\\\hline\n ");
  IO_DruckeBeispiel();
  ST_Statistik("short:all");
  ST_PlatzverbrauchAusgeben(FALSE);
  ZM_Statistik("ZM:latex");
  switch (Situation){
  case Finished:
    if (Z_ZielGesamtStatus == Z_AlleWiederlegt || Z_ZielGesamtStatus == Z_EinigeWiederlegt)
      fprintf(IO_Strom, " & R ");
    else
      fprintf(IO_Strom, " & S ");
    break;
  case Timeout:
    fprintf(IO_Strom, " & T ");
    break;
  case Error:
    fprintf(IO_Strom, " & E ");
    break;
  case NoMemory:
    fprintf(IO_Strom, " & M ");
    break;
  case Interrupt:
    fprintf(IO_Strom, " & I ");
    break;
  case GoalProved:
    fprintf(IO_Strom, " & G ");
    break;
  default:
    break;
  }
  fprintf(IO_Strom, " \\\\\\hline\n");
  IO_StatistikModusAufNormalSetzen();
}

/* -------------------------------------------------------*/
/* -------------- Menue ----------------------------------*/
/* -------------------------------------------------------*/


static void Menue(ReasonT Situation, const char* Ueberschrift, const char *Bemerkung, BOOLEAN Resultat)
/* Ausgabe von Statistiken und Ergebnissen, auch als Interrupt-Menue. */
{
   BOOLEAN MenueWiederholen = TRUE;
   char    s[INPUT_STRLEN];
   char    AusgabeDatei[INPUT_STRLEN];
   char    eing[INPUT_STRLEN];
   BOOLEAN AusgabeDateiAngegeben = FALSE;
   BOOLEAN DruckeMenue = TRUE;
   
   if (Situation ==Interrupt){ 
      ZM_StoppGesamtZeit();
      SI_InterruptflagLoeschen();  /* Zuruecksetzen des Flags */
   }
   do{
     if (DruckeMenue) {
       
      IO_DruckeUeberschrift(Ueberschrift);
      IO_DruckeBeispiel();
      if (Bemerkung) fprintf(IO_Strom, "\n***> %s", Bemerkung);
      fprintf(IO_Strom, "\n\n");

      fprintf(IO_Strom, "* How to proceed\n");
      fprintf(IO_Strom, "       q - quit\n");
      if (Situation != Error && Situation != NoMemory)
        fprintf(IO_Strom, "       c - continue\n"); /* nach Fehler gibt's kein continue */
      if (Situation == Interrupt) fprintf(IO_Strom, "      ol - output-level (0,...,4)\n");

      fprintf(IO_Strom, "\n* Statistics\n");
      fprintf(IO_Strom, "       s - statistics on run\n");
      fprintf(IO_Strom, "      sh - short statistics on run\n");
      fprintf(IO_Strom, "       p - statistics on process\n");
      if (PA_XPRTModusSichtbar()){
        fprintf(IO_Strom, "       i - statistics on innards\n");
        fprintf(IO_Strom, "     tex - statistics in LaTeX-format\n");
        fprintf(IO_Strom, "    stex - short statistics in LaTeX-format\n");
      }

      fprintf(IO_Strom, "\n* Input-Data & Environment\n");
      fprintf(IO_Strom, "     par - parameterization\n");
      fprintf(IO_Strom, "   setup - Waldmeister set-up\n");
      fprintf(IO_Strom, "    view - specification\n");
      fprintf(IO_Strom, "    conf - configuration\n");


      if ((Situation == Finished) || (Situation == Interrupt)){ /* Bei Timeout und Error keine Konsistenz-Garantie */
        fprintf(IO_Strom, "\n* State of Computation\n");
        fprintf(IO_Strom, "       r - current set of rules\n");
        fprintf(IO_Strom, "       e - current set of equations\n");
        fprintf(IO_Strom, "       g - current set of goals\n");
        fprintf(IO_Strom, "     sue - current set of ues --- beware of data overflow\n");
        if (PA_XPRTModusSichtbar()){
          fprintf(IO_Strom, "      wd - weight-distribution in current set of ues\n");
          fprintf(IO_Strom, "      td - time-distribution of weights\n");
        }
        if ( (Situation == Finished) &&((PA_ExtModus == convergence) || (PA_ExtModus == termination) || (PA_ExtModus == confluence)))
          fprintf(IO_Strom, "      or - other results (Sonstige Ergebnisse)\n");
        fprintf(IO_Strom, "    spec - save current system as new specification\n");
      }

      fprintf(IO_Strom, "\n* Redirect outputs from this menue\n");
      fprintf(IO_Strom, "     sfo - following outputs into file (start file output)\n");
      fprintf(IO_Strom, "     ffo - finish file output (finish file output)\n");


      if (PA_XPRTModusSichtbar()){
        if ((Situation ==Interrupt) || (SPArrayPos>0))
          fprintf(IO_Strom, "\n* Action\n");
        if (Situation ==Interrupt)
          fprintf(IO_Strom, "     irp - intermediate reprocessing of SUE --- order coffee first\n");
        if (Situation ==Interrupt)
          fprintf(IO_Strom, "   cgirp - intermediate reprocessing of critical goals \n");
        if (SPArrayPos>0)
          fprintf(IO_Strom, "      ip - IRP-points\n");
        
        fprintf(IO_Strom, "\n* Check consistency\n");
        if (Situation != Timeout){ /* Bei Timeout keine Konsistenz-Garantie */
          fprintf(IO_Strom, "      ic - check WDTs for consistency\n");
#if TV_TEST
          fprintf(IO_Strom, "      tc - check termcells for consistency\n");
#endif
          fprintf(IO_Strom, "      ms - memory statistics\n");
        }
        else
          fprintf(IO_Strom, "         - No consistency-checking because of Timeout\n");
#if COMP_TW        
        fprintf(IO_Strom, "     ref - output referee results\n");
#endif
      }
     }
     if (!DruckeMenue)
       fprintf(IO_Strom, "\n('h' for topic-list) ");
     DruckeMenue = FALSE;
     fprintf(IO_Strom, ">>> ");
     fgets(s, INPUT_STRLEN, stdin);
     chop (s);
      
      fprintf(IO_Strom, "\n");

      if (strcmp(s, "q") == 0)
        exit(0);
      else if (strcmp(s, "h") == 0){
	DruckeMenue = TRUE;
      }
      else if (strcmp(s, "s") == 0){
         char str[100];
         if (PA_InteraktionAn()){
            eing[0]='\0';
            printf("specify (SUEM:C:BO:PQ:IR:KPIR:MO:MNF:NF:RE:SC:DM:U:U2:Z)\n>>> ");
            fgets(eing, INPUT_STRLEN, stdin);
	    chop (eing);
            strcpy (str, "standard:");
            if (strlen(eing)<1)
               strcat(str, "all");
            else
               strcat(str, eing);
         }
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("statistics");
         ST_Statistik(str);
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "sh") == 0){
         char str[100];
         if (PA_InteraktionAn()){
            eing[0]='\0';
            printf("specify (SUEM:C:BO:PQ:IR:KPIR:MO:MNF:NF:RE:SC:DM:U:U2:Z)\n>>> ");
            fgets(eing, INPUT_STRLEN, stdin);
	    chop(eing);
            strcpy (str, "short:");
            if (strlen(eing)<1)
               strcat(str, "all");
            else
               strcat(str, eing);
         }
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("short statistics");
         ST_Statistik(str);
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if ((strcmp(s, "or") == 0) && (Situation == Finished) && 
               ((PA_ExtModus == convergence) || (PA_ExtModus == termination) || 
                (PA_ExtModus == confluence))){
        if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
        RE_SonstigeErgebnisse(Resultat);
        if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
        else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "sfo") == 0){
         fprintf(IO_Strom, "\nName of file: ");
         fgets(AusgabeDatei, INPUT_STRLEN, stdin);
	 chop(AusgabeDatei);
         AusgabeDateiAngegeben = TRUE;
      }
      else if (strcmp(s, "ffo") == 0){
         AusgabeDateiAngegeben = FALSE;
      }
      else if (strcmp(s, "par") == 0){
         PA_DefaultsAusgeben(AusgabeDateiAngegeben?AusgabeDatei:"stdout");
         IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "p") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         ST_ProzessStatistik();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else  IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "setup") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("Waldmeister set-up");
         PA_DruckeModi();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "view") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("specification");
         IO_DateiAusgabe(IO_Spezifikationsname(), TRUE);
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
      }
      else if (strcmp(s, "r") == 0){
         if (RE_RegelmengeLeer()){
           fprintf(IO_Strom, "\n\nset of rules is empty.\n\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
#if TP_PETERGRAF
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeRegelmenge(1);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
#else
            int wert = 1;
#if BA_AUSGABEN
            do{
               fprintf(IO_Strom, "output R:\n");
               fprintf(IO_Strom, " 1 - textual\n");
               fprintf(IO_Strom, " 2 - text-tree\n");
               fprintf(IO_Strom, " 3 - graphics-tree (emacs)\n");
               if (PA_XPRTModusSichtbar()){
                  fprintf(IO_Strom, " 4 - textual tree-dump\n");
                  fprintf(IO_Strom, " 5 - graphical tree-dump\n");
               }
               fprintf(IO_Strom, "\n>>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi(eing);
            }while((wert<1)||(wert>(PA_XPRTModusSichtbar()?5:3)));
#endif
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeRegelmenge(wert);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
#endif
         }
      }
      else if (strcmp(s, "e") == 0){
         if (RE_GleichungsmengeLeer()){
           fprintf(IO_Strom, "\n\nset of equations is empty\n\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
#if TP_PETERGRAF
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeGleichungsmenge(1);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
#else
            int wert = 1;
#if BA_AUSGABEN
            do{
               fprintf(IO_Strom, "output E:\n");
               fprintf(IO_Strom, " 1 - textual\n");
               fprintf(IO_Strom, " 2 - text-tree\n");
               fprintf(IO_Strom, " 3 - graphics-tree (emacs)\n");
               if (PA_XPRTModusSichtbar()){
                  fprintf(IO_Strom, " 4 - textual tree-dump\n");
                  fprintf(IO_Strom, " 5 - graphical tree-dump\n");
               }
               fprintf(IO_Strom, "\n>>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi(eing);
            }while((wert<1)||(wert>(PA_XPRTModusSichtbar()?5:3)));
#endif
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeGleichungsmenge(wert);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
#endif
         }
      }
      else if (strcmp(s, "g") == 0){
	    int wert;
            do{
               fprintf(IO_Strom, "output goals:\n");
               fprintf(IO_Strom, "1 - short\n");
               fprintf(IO_Strom, "2 - long\n\n");
               fprintf(IO_Strom, ">>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi (eing);
            } while((wert<1) &&(wert>2));
	    if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	    switch (wert) {
	    case 1:
               Z_AusgabeZielmenge(Z_AusgabeInitial);
               break;
	    case 2:
	    default:
               Z_AusgabeZielmenge(Z_AusgabeAlles);
	    }
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "ol") == 0){
         int wert;
         do{
            fprintf(IO_Strom, "current output level: %d\n", IO_AusgabeLevel());
            fprintf(IO_Strom, "new level >>> ");
            fgets(eing, INPUT_STRLEN, stdin);
            wert = atoi(eing);
            fprintf(IO_Strom, "\n");
         }while((wert<0) || (wert > 4));
         IO_AusgabeLevelBesetzen(wert);
      }
      else if (strcmp(s, "c") == 0)
        MenueWiederholen = FALSE;
      else if (strcmp(s, "sue") == 0){
         char str[INPUT_STRLEN];
         printf("specify (lowerBound=[n]:upperBound=[n]:noweights:noterms) >>> ");
         fgets(eing, INPUT_STRLEN, stdin);
	 chop(eing);
         strcpy(str, "SUEM:output:");
         strcat(str, eing);
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         SUE_Statistics(str);
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (PA_XPRTModusSichtbar()){
        if (strcmp(s, "ms") == 0){
          int wert;
          do{
            fprintf(IO_Strom, "memory statistics\n");
            fprintf(IO_Strom, "1 - discrepancies only\n");
            fprintf(IO_Strom, "2 - complete\n\n");
            fprintf(IO_Strom, ">>> ");
            fgets(eing, INPUT_STRLEN, stdin);
            wert = atoi(eing);
          }while((wert<1) &&(wert>2));
          if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
          SV_Statistik((Situation == Error || Situation == NoMemory), (wert==1), "complete");
          if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
          else IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "i") == 0){
           if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
           ST_InterneStatistik();
           if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
           else IO_WarteAufReturnMitText("RETURN");
        }
        else if (TV_TEST && strcmp(s, "tc") == 0){
          if ((Situation == Timeout) || (Situation == Error) || Situation == NoMemory)   /* Fehler-Situation */
             fprintf(IO_Strom, "\nerror was encountered ==> test not possible.\n");
          else{
             if (Situation == Interrupt){
                TV_TermzellenVerfolger(3);
             }
             else
                TV_TermzellenVerfolger(3);
          }
          IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "ic") == 0){
           if ((Situation == Timeout) || (Situation == Error) || Situation == NoMemory)   /* Fehler-Situation */
              fprintf(IO_Strom, "\nerror was encountered ==> test not possible.\n");
           else
              RE_RUndETest();
           IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "irp") == 0){
           printf("specify >>> ");
           fgets(eing, INPUT_STRLEN, stdin);
	   chop(eing);
           if ((SPArrayPos>0)&&(SPArray[SPArrayPos-1].sp == RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt()))
              fprintf(IO_Strom, "SUE already reprocessed!\n");
           else{
              ZM_StartGesamtZeit(); /*Zeitmessung soll konsistent bleiben (sonst: T(NFBildung) > T(Gesamt) */
              KPIR_Wahlfrei(eing);
              ZM_StoppGesamtZeit();
              fprintf(IO_Strom, "\nMoment for reproduction of run is %ld.\n", RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt());
              StichpunktMerken(RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt(), eing);
           }
           IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "cgirp") == 0) {
	  printf("specify >>> ");
	  fgets(eing, INPUT_STRLEN, stdin);
	  chop(eing);
	  ZM_StartGesamtZeit(); /*Zeitmessung soll konsistent bleiben (sonst: T(NFBildung) > T(Gesamt) */
	  CG_CGMengeInterreduzieren(TRUE, eing);
	  ZM_StoppGesamtZeit();
	  fprintf(IO_Strom, "\nMoment for reproduction of run is %ld.\n", RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt());
	  IO_WarteAufReturnMitText("RETURN");
        }
        else if ((SPArrayPos>0) && (strcmp(s, "ip") == 0)){
           int i;
           if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
           fprintf(IO_Strom, "IRP-Stichpunkte are: ");
           for (i=0; i<SPArrayPos; i++){
              fprintf(IO_Strom, "%ld %s", SPArray[i].sp, SPArray[i].strength);
              if (i<SPArrayPos-1)
                 fprintf(IO_Strom, ":");
           }
           fprintf(IO_Strom, "\n\n");
           if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
           else IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "conf") == 0){
          if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
          ST_DruckeKonfiguration();
          if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
          else IO_WarteAufReturnMitText("RETURN");
        }
        else if ((strcmp(s, "tex") == 0)){
           char str[INPUT_STRLEN];
	   
           eing[0]='\0';
           
           strcpy (str, "latex-head:");
           
           if (PA_InteraktionAn()){
              printf("specify >>> ");
              fgets(eing, INPUT_STRLEN, stdin);
	      chop(eing);
           }

           if (strlen(eing)<1)
              strcpy(eing, "all");

           strcat(str, eing);
           
           if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);

           ST_Statistik(str);

           ST_PlatzverbrauchAusgeben(TRUE);

           fprintf(IO_Strom, " time \\\\\\hline\n");
           
           strcpy(str, "standard:");
           
           strcat(str, eing);
           
           IO_StatistikModusAufLatexSetzen();
           
           ST_Statistik(str);
           
           ST_PlatzverbrauchAusgeben(FALSE);

           ZM_Statistik("ZM:latex");
           
           fprintf(IO_Strom, " \\\\\\hline\n");
           
           IO_StatistikModusAufNormalSetzen();
           
           if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();

        }
        else if ((strcmp(s, "stex") == 0)){
          if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
          STEXZeileAusgeben(Situation);
          if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
        }
        else if (strcmp(s, "wd") == 0){
           char str[INPUT_STRLEN];
           printf("specify (lowerBound=[n]:upperBound=[n]:sec)>>> ");
           fgets(eing, INPUT_STRLEN, stdin);
           strcpy(str, "SUEM:weight:");
           strcat(str, eing);
	   chop(eing);
           if (AusgabeDateiAngegeben) {IO_AusgabeStromUmlenken(AusgabeDatei);strcat(eing, ":datei");}
           SUE_Statistics(str);
           if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
           else IO_WarteAufReturnMitText("RETURN");
        }
        else if (strcmp(s, "td") == 0){
           char eing[INPUT_STRLEN];
           printf("specify >>> ");
           fgets(eing, INPUT_STRLEN, stdin);
	   chop(eing);
           strcat(eing, ":SUEM:C:time");
           if (AusgabeDateiAngegeben) {IO_AusgabeStromUmlenken(AusgabeDatei);strcat(eing, ":datei");}
           SUE_Statistics(eing);
           if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
           else IO_WarteAufReturnMitText("RETURN");
        }
	else if (strcmp (s, "spec") == 0) {
	   SpezifikationSchreiben();
	}
     }
   }while (MenueWiederholen);

   if (Situation == Interrupt) ZM_StartGesamtZeit(); /* Interrupt-Modus => Zeit wieder anwerfen */

}

/* -------------------------------------------------------*/
/* -------------- Statistik ------------------------------*/
/* -------------------------------------------------------*/

static void AutoStatistik(ReasonT Situation, const char* Ueberschrift, const char *Bemerkung, BOOLEAN Resultat);


/* -------------------------------------------------------*/
/* -------------- Statistik ------------------------------*/
/* -------------------------------------------------------*/


void ST_SuccessStatistik(const char *Bemerkung, BOOLEAN Resultat)
/* Bei eingeschalteter Interaktion: Interaktives Ergebnis-Menue.
   Sonst Ausgabe diverser Daten gemaess Aufruf.
*/
{
   if (PA_InteraktionAn())
      Menue(Finished, "RESULTS-STATISTICS", Bemerkung, Resultat);
   
   else
      AutoStatistik(Finished, "SUCCESS-STATISTICS", Bemerkung, Resultat);
}

#if COMP_POWER
void ST_PowerStatistik(const char *Bemerkung)
{
  unsigned long int Sekunden, 
                    Zehntelsekunden;

#if defined (SUNOS5)
  int fd;
  char proc[40];
  prusage_t prusage;
  prpsinfo_t prpsinfo;
  unsigned long int Prozessgroesse;
  sprintf(proc,"/proc/%d", (int)getpid());
  if ((fd = open(proc,O_RDONLY))==-1)
    our_assertion_failed ("IO_ProzessStatistik: open /proc fehlgeschlagen");
  if (ioctl(fd, PIOCUSAGE, &prusage)==-1)
    our_assertion_failed ("IO_ProzessStatistik: open prusage-ioctl fehlgeschlagen");
  if (ioctl(fd, PIOCPSINFO, &prpsinfo)==-1)
    our_assertion_failed ("IO_ProzessStatistik: open prpsinfo-ioctl fehlgeschlagen");
  Prozessgroesse = prpsinfo.pr_size*sysconf(_SC_PAGESIZE);
  Prozessgroesse = 1000*((Prozessgroesse+500)/1000);
#endif

  fprintf (IO_Strom, "\nWaldmeister states: %s.\n", Bemerkung);

  fprintf(IO_Strom,"\nStatistics:\n-----------\n");
  fprintf(IO_Strom,"Rules         %11ld\n", RE_AnzRegelnGesamt);
  fprintf(IO_Strom,"Equations     %11ld\n", RE_AnzGleichungenGesamt());
  fprintf(IO_Strom,"Critical pairs%11ld\n", AnzKPGesamt);
#if defined (SUNOS5)
  fprintf(IO_Strom,"Process size  %11ld\n", Prozessgroesse);
#endif
  Zehntelsekunden = (ZM_GesamtZeit_ut.tv_usec+ZM_GesamtZeit_st.tv_usec+50000)/100000;
  Sekunden = ZM_GesamtZeit_ut.tv_sec+ZM_GesamtZeit_st.tv_sec+Zehntelsekunden/10;
  Zehntelsekunden %= 10;
  fprintf(IO_Strom,"Search time [s]%9ld.%1ld\n",Sekunden,Zehntelsekunden);
}
#endif

void ST_ErrorStatistik(const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn()){
      IO_RingBell();
      Menue(Error, "ERROR-MENUE", Bemerkung, Resultat);
   }
   else{
      AutoStatistik(Error, "ERROR-STATISTICS", Bemerkung, Resultat);
   }
}

void ST_NoMemory (const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn()) {
      IO_RingBell();
      Menue(NoMemory, "ERROR-MENUE", Bemerkung, Resultat);
   }
   else {
      AutoStatistik(NoMemory, "ERROR-STATISTICS", Bemerkung, Resultat);
   }
}

void ST_TimeoutStatistik(const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn())
      Menue(Timeout, "TIMEOUT-MENUE", Bemerkung, Resultat);
   
   else
      AutoStatistik(Timeout, "TIMEOUT-STATISTICS", Bemerkung, Resultat);
}

static void
ZwischenStatistik (ReasonT Situation, const char *Ueberschrift,
		   const char *Bemerkung, BOOLEAN Resultat)
{
   ZM_StoppGesamtZeit();
   SI_Usr1InterruptflagLoeschen ();
   SI_InterruptflagLoeschen ();
   AutoStatistik(Situation, Ueberschrift, Bemerkung, Resultat);
   fflush (NULL);  /* Alle Output-puffer leeren */
   ZM_StartGesamtZeit();   
}

void ST_SignalUsrStatistik(const char *Bemerkung)
{
  ZwischenStatistik (Interrupt, "SIGUSR-STATISTICS", Bemerkung, FALSE);
}

void ST_GoalProvedStatistik(const char *Bemerkung)
{
  ZwischenStatistik (GoalProved, "GOAL-STATISTICS", Bemerkung, FALSE);
}



/* -------------------------------------------------------*/
/* -------------- Interne Statistik ----------------------*/
/* -------------------------------------------------------*/

void ST_InterneStatistik(void)
{
   IO_DruckeUeberschrift("Interne Statistik");

   ST_Statistik("all:intern");

   ZM_Statistik("ZM:intern");
   
   IO_DruckeStatistikZeile("Maximale Ausdehnung des TO_Kopierstapels", TO_AusdehnungKopierstapel());
   
   IO_DruckeStatistikZeile("Maximale Ausdehnung des IO_Kopierstapels", IO_AusdehnungAusgabestapel());
}

/* -------------------------------------------------------*/
/* -------------- Konfiguration --------------------------*/
/* -------------------------------------------------------*/

void ST_DruckeKonfiguration(void)
/* Ausgabe der durch die defines in diesem Modul bestimmten Konfiguration. */
{
  struct utsname  utsbuf;

  IO_DruckeUeberschrift("Konfiguration");

  fprintf(IO_Strom, "%s", configuration_banner);
  
  IO_DruckeZeile('+');
  uname(&utsbuf);
  fprintf(IO_Strom, "Run executed on %s (%s, %s %s)\n", 
	  utsbuf.nodename, utsbuf.machine, 
	  utsbuf.sysname, utsbuf.release);
  IO_DruckeZeile('+');
  
#if IO_AUSGABEN
  fprintf(IO_Strom, "Ausgaben waehrend des Laufes AN, ");
#else
  fprintf(IO_Strom, "Ausgaben waehrend des Laufes AUS, ");
#endif
#if COMP_PCL
  fprintf(IO_Strom, "PCL-Modus enabled.\n");
#else
  fprintf(IO_Strom, "PCL-Modus not existent.\n");
#endif
#if COMP_OPTIMIERUNG
  fprintf(IO_Strom, "Optimized Compilation.\n");
#endif
#if !COMP_OPTIMIERUNG
  fprintf(IO_Strom, "Non-Optimized Compilation.\n");
#endif
  ST_Statistik("all:config");
  ZM_Statistik("ZM:config");
  SV_Statistik(FALSE,FALSE,"config");
}

/* -------------------------------------------------------*/
/* -------------- Ergebnis-Menue -------------------------*/
/* -------------------------------------------------------*/

void ST_IRMenue(const char *where)
/* Ausgabe von Statistiken und Ergebnissen, auch als Interrupt-Menue. */
{
   char str[200] = "Run interrupted by user in ";
   
   if (SI_Usr1InterruptflagGesetzt ()) {
      ST_SignalUsrStatistik ("Received signal SIGUSR1");
   }
   else {
      strcat (str, where);
      Menue(Interrupt, "INTERRUPT-MENUE", str, FALSE);
   }
}

static void AutoStatistik(ReasonT Situation, const char* Ueberschrift MAYBE_UNUSEDPARAM, 
  const char *Bemerkung MAYBE_UNUSEDPARAM, BOOLEAN Resultat)
{
   /* InfoString ergibt:
      h    - Head
      r    - Regelmenge
      e    - Gleichungsmenge
      s    - Statistik (komplett)
      k    - Konfiguration
      i    - Intern
      p    - Prozess
      pa   - Aufruf-Parameter
      sh   - Short Statistik
      tex  - LaTeX-Statistik
      stex - short LaTeX-Statistik
      w    - weight distribution
      */

   char *filename=NULL;

   if ((filename = IS_ValueOfOption(AutoErgDEF, AutoErgUSR, "file")))
      IO_AusgabeStromUmlenken(filename);
   
   fprintf(IO_Strom, "\n");
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "h") == 0)
      IO_DruckeBeispiel();
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "r") == 0){
      RE_AusgabeRegelmenge(1);
      RE_SonstigeErgebnisse(Resultat);
   }
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "e") == 0)
      RE_AusgabeGleichungsmenge(1);
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "s") == 0){
      IO_DruckeUeberschrift("Statistik");
      ST_Statistik("all:standard");
   }
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "k") == 0)
      ST_DruckeKonfiguration();
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "i") == 0)
      ST_InterneStatistik();
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "p") == 0)
      ST_ProzessStatistik();
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "pa") == 0)
      PA_DruckeModi();
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "sh") == 0){
      IO_DruckeUeberschrift("Kurz - Statistik");
      ST_Statistik("short:all");
   }
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "tex") == 0){
      fprintf(IO_Strom, " problem & ");
      ST_Statistik("latex-head:all");
      fprintf(IO_Strom, " time \\\\\\hline\n");
      IO_StatistikModusAufLatexSetzen();
      IO_DruckeBeispiel();
      ST_Statistik("standard:all");
      ZM_Statistik("ZM:latex");
      switch (Situation){
      case Finished: 
         if (Z_ZielGesamtStatus == Z_AlleWiederlegt || Z_ZielGesamtStatus == Z_EinigeWiederlegt)
           fprintf(IO_Strom, " R ");
         else
           fprintf(IO_Strom, " S ");
         break;
      case Timeout: 
         fprintf(IO_Strom, " T ");
         break;
      case Error: 
         fprintf(IO_Strom, " E ");
         break;
      case NoMemory: 
         fprintf(IO_Strom, " M ");
         break;
      default: 
         break;
      }
      fprintf(IO_Strom, " \\\\\\hline\n");
      IO_StatistikModusAufNormalSetzen();
   }
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "stex") == 0)
     STEXZeileAusgeben(Situation);
   if (IS_FindOption(AutoErgDEF, AutoErgUSR, 1, "w") == 0) {
      SUE_Statistics("SUEM:weight:");
   }
   
#if COMP_POWER
   switch (Situation){
   case Timeout: 
     fprintf(IO_Strom, "Waldmeister run out of time.\n");
     break;
   case NoMemory:
     fprintf(IO_Strom, "Waldmeister run out of memory.\n");
     break;
   case Error:
     fprintf(IO_Strom, "Error: %s\n", Bemerkung);
     break;
   case Finished: 
   default: 
     break;
   }
#else
   switch (Situation){
   case Finished: 
     fprintf(IO_Strom, "\n\nWaldmeister states: %s.\n", Bemerkung);
     break;
   case Timeout: 
     fprintf(IO_Strom, "\n\n TIMEOUT: %s\n", Bemerkung);
     break;
   case Error:
   case NoMemory:
     fprintf(IO_Strom, "\n\n ERROR: %s \n", Bemerkung);
     break;
   default: 
     break;
   }
#endif
   if (filename)
      IO_AusgabeStromZuruecksetzen();
}


