/***********************************************************************************************************************
 * Autor:      Juergen Sellentin, 13.02.94
 * Angepasst : Arnim, 04.03.94
 * Stackverwaltung zugunsten eines Arrays 'rausgeworfen. 11.04.94, Arnim.
 ***********************************************************************************************************************
 * 
 * Bemerkungen zur Zeitmessung:
 * ----------------------------
 *
 * Die Zeitmessungen werden mit der Systemfunktion "getrusage" realisiert. Diese protokolliert die seit dem Beginn des
 * Prozesses verbrauchte Zeit (als user- und system- Anteil). 
 * Die Genauigkeit dieser Messung haengt sehr stark von der update-Frequenz von getrusage ab !!!
 * Bei sehr kurzen Prozeduraufrufen werden die getrusage-Daten evtl. NICHT veraendert, so dass die Laufzeit 0
 * als Ergebnis herauskommt. Andererseits kann eine zu hohe Laufzeit gemessen werden, wenn das Aktualisieren der Daten
 * (von getrusage) gerade in einer sehr kurzen Prozedur durchgefuehrt wird.
 * Im statistischen Mittel gleichen sich diese Fehler hoffentlich aus!
 *
 ***********************************************************************************************************************/

#include "antiwarnings.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "compiler-extra.h"
#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "InfoStrings.h"
#include "ZeitMessung.h"
#include "SpeicherVerwaltung.h"
#include "WaldmeisterII.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe ZeitMessung.h) -----------------------*/
/* ------------------------------------------------------------------------*/

struct timeval       ZM_GesamtZeit_ut;               /* insgesamt verbrauchte Zeit exklusive Parsen und Plattklopfen. */
struct timeval       ZM_GesamtZeit_st;



/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++ Zeitmessung +++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */


#define AnzMikrosekProSek 1000000


#define /*void*/ ZeitdifferenzAddieren(/*struct timeval*/ Summe, /*struct timeval*/ Start, /*struct timeval*/ Ende) \
  (Summe).tv_sec += (Ende).tv_sec - (Start).tv_sec;              \
   if (((Summe).tv_usec += (Ende).tv_usec - (Start).tv_usec)<0)  \
   {  (Summe).tv_usec += AnzMikrosekProSek;                      \
      (Summe).tv_sec--;                                          \
   }                                                             \
   else if ((Summe).tv_usec >= AnzMikrosekProSek)                \
   {  (Summe).tv_usec -= AnzMikrosekProSek;                      \
      (Summe).tv_sec++;                                          \
   }

#define /*void*/ ZeitdifferenzBestimmen(/*struct timeval*/ StartZeit, /*struct timeval*/ EndZeit, /*unsigned long*/ d_sec, /*unsigned long*/ d_usec)\
{\
    d_sec = EndZeit.tv_sec - StartZeit.tv_sec;\
    if (StartZeit.tv_usec > EndZeit.tv_usec){\
        d_usec = AnzMikrosekProSek + EndZeit.tv_usec - StartZeit.tv_usec;\
        d_sec--;\
   }\
   else\
      d_usec = EndZeit.tv_usec - StartZeit.tv_usec;\
}


static struct timeval GesamtzeitStart_user_time;
static struct timeval GesamtzeitStart_system_time;
static struct timeval GesamtzeitStart_wallclock_time;
static struct timeval GesamtzeitWallClock;

void ZM_StartGesamtZeit(void)
{
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("Start time: getrusage didn't work");
   GesamtzeitStart_user_time = rusageBuf.ru_utime, GesamtzeitStart_system_time = rusageBuf.ru_stime;
   /* DruckeZeitenAufStack();*/
   if (gettimeofday(&GesamtzeitStart_wallclock_time, NULL) == -1)
     our_assertion_failed ("Start time: gettimeofday didn't work");
}

void ZM_StoppGesamtZeit(void)
{  
   struct timeval GesamtzeitEnde_wallclock_time;
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("Stop time: getrusage didn't work");
   if (gettimeofday(&GesamtzeitEnde_wallclock_time, NULL) == -1)
      our_assertion_failed ("Stop time: gettimeofday didn't work");

   ZeitdifferenzAddieren(ZM_GesamtZeit_ut,GesamtzeitStart_user_time,rusageBuf.ru_utime);   /* user-time update */
   ZeitdifferenzAddieren(ZM_GesamtZeit_st,GesamtzeitStart_system_time,rusageBuf.ru_stime);   /* system-time update */
   ZeitdifferenzAddieren(GesamtzeitWallClock, GesamtzeitStart_wallclock_time, GesamtzeitEnde_wallclock_time);
}


static void DruckeZeitSolo(FILE *fp, struct timeval time) /* Neufassung. BL. */
{ 
  long seconds, milliseconds;

  milliseconds = (time.tv_usec + 500) / 1000;
  seconds      = time.tv_sec + milliseconds / 1000;
  milliseconds = milliseconds % 1000;

  fprintf(fp, "%ld.%03ld", seconds, milliseconds);
}

void ZM_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("ZM", IS) || IS_IsIn("all", IS)))
     return;

   if (IS_IsIn("standard", IS)){
      IO_DruckeUeberschrift("TIMES (user- and system-time in seconds)");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_ut);
      fprintf(IO_Strom, "s ut ");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_st);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, GesamtzeitWallClock);
      fprintf(IO_Strom, " wct)");
      fprintf(IO_Strom, "   total time (without initialisation)\n");
   }
   else if (IS_IsIn("latex", IS)){
      fprintf(IO_Strom, " ");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_ut);
      fprintf(IO_Strom, "s ut ");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_st);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, GesamtzeitWallClock);
      fprintf(IO_Strom, " wct)");
   }
}
