/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - BasicTPR.c                                                                                            */
/* - ************************                                                                              */
/* -                                                                                                       */
/* - Repraesentation von kritischen Paaren als Flachterm-Paar.                                             */
/* -                                                                                                       */
/* - 15.05.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "BasicTPR.h"
#include "InfoStrings.h"
#include "SUEBasis.h"
#include "TermOperationen.h"
#include "pcl.h"


/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void BTP_NewFlatsGenerated(void)
{
   PCL_StartChain();
}

/* --------------------------------------------------------------------------------------------------------*/
/* -------- Termpaare in Standard-Representation ----------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
#if COMP_PCL

typedef struct{
   TermT tp;
   PCL_ObjektNummerT pclID;
} PCLTermpaarTS;

typedef PCLTermpaarTS *PCLTermpaarT;

SV_SpeicherManagerT PCLTermpaarManager;

TermpairT BTP_Flat2TP(TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM,
                      UTermT actualPosition MAYBE_UNUSEDPARAM, PCL_ObjektNummerT pclID)
{
   PCLTermpaarT paar;
   SV_Alloc(paar,PCLTermpaarManager,PCLTermpaarT);
   PCL_IDZuweisung(&(paar->pclID), pclID);
   PCL_SaveChain();
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
   paar->tp = lhs;
   return (TermpairT) paar;
}

void BTP_Tp2Par(TermpairT tp, TermT *lhs, TermT *rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclIDPtr)
{
   *lhs = ((PCLTermpaarT) tp)->tp;
   *rhs = TO_Schwanz(TO_TermEnde(*lhs));
   TO_NachfBesetzen(TO_TermEnde(*lhs), NULL);
   PCL_IDZuweisung(pclIDPtr, ((PCLTermpaarT) tp)->pclID);
   PCL_StartChain();
}

void BTP_Par2Ft(TermpairT tp, TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                    RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   PCL_SaveChain();
   SV_Dealloc(PCLTermpaarManager, tp);
}

TermpairT BTP_Par2TP(TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                     RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, BOOLEAN HasChanged MAYBE_UNUSEDPARAM,
                     PCL_ObjektNummerT pclID)
{
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
   PCL_IDZuweisung(&(((PCLTermpaarT)tp)->pclID), pclID);
   ((PCLTermpaarT)tp)->tp = lhs;
   PCL_SaveChain();
   return (TermpairT) tp;
}

void BTP_DestroyTP(TermpairT tp, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TO_TermzellenlisteLoeschen(((PCLTermpaarT) tp)->tp, 
                              TO_TermEnde(TO_Schwanz(TO_TermEnde(((PCLTermpaarT) tp)->tp))));
   SV_Dealloc(PCLTermpaarManager, tp);
}

void BTP_DestroyPar(TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                    RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TO_TermeLoeschen(lhs, rhs);
   PCL_DeleteChain();
   SV_Dealloc(PCLTermpaarManager, tp);
}

void BTP_PrintTermpair(TermpairT tp, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, const char *trenner)
{
   TermT lhs=((PCLTermpaarT) tp)->tp, rhs=TO_Schwanz(TO_TermEnde(lhs));

   IO_DruckeFlex("%t %s %t", lhs, trenner, rhs);
}

void BTP_TermOperationBeiVerfolgung(void (* Operation) (TermT, TermT), TermpairT tp,
                                    RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TermT lhs, rhs;
   lhs = ((PCLTermpaarT) tp)->tp;
   rhs = TO_Schwanz(TO_TermEnde(lhs));
   TO_NachfBesetzen(TO_TermEnde(lhs), NULL);
   (* Operation) (lhs, rhs);
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
}

BOOLEAN BTP_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                   TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
{
   if (TP_ElternNr(father1) > TP_ElternNr(father2)) return TRUE;
   else if (TP_ElternNr(father1) < TP_ElternNr(father2)) return FALSE;
   else                         /* father1 == father2 */
      if (TP_ElternNr(mother1) > TP_ElternNr(mother2)) return TRUE;
      else if (TP_ElternNr(mother1) < TP_ElternNr(mother2)) return FALSE;
      else{
         TermT lhs1, rhs1, lhs2, rhs2;
         BOOLEAN ret;

         lhs1 = ((PCLTermpaarT) tp1)->tp;
         rhs1 = TO_Schwanz(TO_TermEnde(lhs1));
         TO_NachfBesetzen(TO_TermEnde(lhs1), NULL);
         lhs2 = ((PCLTermpaarT) tp2)->tp;
         rhs2 = TO_Schwanz(TO_TermEnde(lhs2));
         TO_NachfBesetzen(TO_TermEnde(lhs2), NULL);

         ret = TO_TPTestSymbLex(lhs1, rhs1, lhs2, rhs2);

         TO_NachfBesetzen(TO_TermEnde(lhs1), rhs1);
         TO_NachfBesetzen(TO_TermEnde(lhs2), rhs2);

         return ret;
      }
}

#else

/* In diesem Zweig keinerlei PCL-Operationen!!! */
TermpairT BTP_Flat2TP(TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM,
                      UTermT actualPosition MAYBE_UNUSEDPARAM, PCL_ObjektNummerT pclID MAYBE_UNUSEDPARAM)
{
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
   return (TermpairT) lhs;
}

void BTP_Tp2Par(TermpairT tp, TermT *lhs, TermT *rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclIDPtr MAYBE_UNUSEDPARAM)
{
   *lhs = (TermT) tp;
   *rhs = TO_Schwanz(TO_TermEnde((TermT) tp));
   TO_NachfBesetzen(TO_TermEnde((TermT) tp), NULL);
}

void BTP_Par2Ft(TermpairT tp MAYBE_UNUSEDPARAM, TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                    RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   ;
}

TermpairT BTP_Par2TP(TermpairT tp MAYBE_UNUSEDPARAM, TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                     RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, BOOLEAN HasChanged MAYBE_UNUSEDPARAM,
                     PCL_ObjektNummerT pclID MAYBE_UNUSEDPARAM)
{
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
   return (TermpairT) lhs;
}

void BTP_DestroyTP(TermpairT tp, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TO_TermzellenlisteLoeschen((TermT) tp, TO_TermEnde(TO_Schwanz(TO_TermEnde((TermT) tp))));
}

void BTP_DestroyPar(TermpairT tp MAYBE_UNUSEDPARAM, TermT lhs, TermT rhs, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, 
                    RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TO_TermeLoeschen(lhs, rhs);
}

void BTP_PrintTermpair(TermpairT tp, RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM, const char *trenner)
{
   TermT lhs=(TermT)tp, rhs=TO_Schwanz(TO_TermEnde(lhs));

   IO_DruckeFlex("%t %s %t", lhs, trenner, rhs);
}

void BTP_TermOperationBeiVerfolgung(void (* Operation) (TermT, TermT), TermpairT tp,
                                    RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM)
{
   TermT lhs, rhs;
   lhs = (TermT) tp;
   rhs = TO_Schwanz(TO_TermEnde((TermT) tp));
   TO_NachfBesetzen(TO_TermEnde(lhs), NULL);
   (* Operation) (lhs, rhs);
   TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
}


static BOOLEAN BTP_UnambiguousGreaterThanBODY(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                   TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
{
   if (TP_ElternNr(father1) > TP_ElternNr(father2)) return TRUE;
   else if (TP_ElternNr(father1) < TP_ElternNr(father2)) return FALSE;
   else                         /* father1 == father2 */
      if (TP_ElternNr(mother1) > TP_ElternNr(mother2)) return TRUE;
      else if (TP_ElternNr(mother1) < TP_ElternNr(mother2)) return FALSE;
      else{
         TermT lhs1, rhs1, lhs2, rhs2;
         BOOLEAN ret;

         lhs1 = (TermT) tp1;
         rhs1 = TO_Schwanz(TO_TermEnde((TermT) tp1));
         TO_NachfBesetzen(TO_TermEnde(lhs1), NULL);
         lhs2 = (TermT) tp2;
         rhs2 = TO_Schwanz(TO_TermEnde((TermT) tp2));
         TO_NachfBesetzen(TO_TermEnde(lhs2), NULL);

         ret = TO_TPTestSymbLex(lhs1, rhs1, lhs2, rhs2);

         TO_NachfBesetzen(TO_TermEnde(lhs1), rhs1);
         TO_NachfBesetzen(TO_TermEnde(lhs2), rhs2);

         return ret;
      }
}

BOOLEAN BTP_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                   TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
{
   printf("%4ld, %4ld ", TP_ElternNr(father1), TP_ElternNr(mother1));
   IO_DruckeFlex("%t=%t  ?>?  ", (TermT) tp1, TO_Schwanz(TO_TermEnde((TermT) tp1)));
   
   printf("%4ld, %4ld ", TP_ElternNr(father2), TP_ElternNr(mother2));
   IO_DruckeFlex("%t=%t   ", (TermT) tp1, TO_Schwanz(TO_TermEnde((TermT) tp2)));
   
   if (BTP_UnambiguousGreaterThanBODY(tp1, father1, mother1, tp2, father2, mother2)){
      printf("TRUE\n");return TRUE;
   }
   else{
      printf("FALSE\n");return FALSE;
   }
}

#endif


/* --------------------------------------------------------------------------------------------------------*/
/* -------- Initialisierung -------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN BTP_InitAposteriori(InfoStringT DEF MAYBE_UNUSEDPARAM, InfoStringT USR MAYBE_UNUSEDPARAM)
{
#if COMP_PCL
  SV_ManagerInitialisieren(&PCLTermpaarManager, PCLTermpaarTS, 10000, "PCL-Manager");
#endif
  fprintf(IO_Strom, "\n");
  return TRUE;
}
