#include "Ausgaben.h"
#include "compiler-extra.h"
#include "BasicTPR.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"
#include "SUEBasis.h"
#include "TermOperationen.h"
#include "Unifikation1.h"
#include "pcl.h"
#include "Implicit.h"

/* --------------------------------------------------------------------------------------------------------*/
/* - Implicit.h                                                                                            */
/* - **********                                                                                            */
/* -                                                                                                       */
/* - Implizite Darstellung von Paaren von flatterms durch die Stelle ihrer Ueberlappung.                   */
/* -                                                                                                       */
/* - 01.04.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------
 Doku:

 Paare von Flachtermen werden durch die Stelle der Ueberlappung repraesentiert, durch die sie entstanden sind.
 Zum Restaurieren sind folglich die Eltern sowie die Information, in welchen der Elternteile hineinueberlappt
 wurde, noetig. Ferner sind Gleichungen aus der Spezifikation sowie Interreduktions-Opfer zu beruecksichtigen.

 Zu diesem Zwecke werden die Termpaare wie folgt repraesentiert:
 
 - IR-Opfer, Gleichungen aus der Spezifikation: als Flachterm-Paare (1 Pointer)
 - Unifikations-Produkte: als Nummer der Stelle, an der in die aeussere Regel/Gleichung hineinueberlappt wurde.
 
 Bemerkungen:

 - IR-Opfer und Spez-Gleichungen sind leicht daran erkennbar, dass sie den Halbgott als ELternteile haben.
 - Die Information, in welchen der beiden Elternteile letztendlich hineinueberlappt wurde, wird als Vorzeichen
   der Stelle gespeichert:

   tp<0: In OtherParent hineinueberlappt  (CPWithFather)
   tp>0: In ActualParent hineinueberlappt (CPWithMother)


 Voraussetzungen:
 
 - Terme werden nicht laenger als 2^30-1 Symbole.

   --------------------------------------------------------------------------------------------------------*/



/* --------------------------------------------------------------------------------------------------------*/
/* ------- Initialize--------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN IMP_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   /* Hier ist nix zu tun */
   return BTP_InitAposteriori(DEF, USR);
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_NewFlatsGenerated(void)
{
   PCL_StartFeyChain();
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


TermpairT IMP_Flat2TP (TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, 
                       RegelOderGleichungsT OtherParent, UTermT actualPosition, PCL_ObjektNummerT pclID)
{
   TermpairT tp;
   
   if (IsDemigod(OtherParent)){
      /* Keine Ueberlappung verfuegbar ==> Terme aneinanderhaengen */
      tp = BTP_Flat2TP(lhs, rhs, ActualParent, OtherParent, actualPosition, pclID);
   }
   else{
      PCL_DeleteChain();
      /* Ueberlappungsstelle berechnen */
      if (SUE_recCPGenerationHistory == CPWithFather){
         /* actualPosition ist Teil von OtherParent ==> negatives Vorzeichen */
         tp = (TermpairT) (-TO_StelleInTerm(actualPosition, TP_LinkeSeite(OtherParent)));
      }
      else{
         /* actualPosition ist Teil von ActualParent ==> positives Vorzeichen */
         tp =  (TermpairT) TO_StelleInTerm(actualPosition, SUE_ReplicationOfLHS);
      }
      /* Flatterms wegwerfen */
      TO_TermeLoeschen(lhs, rhs);
   }
   return tp;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_TP2ParIntermed (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT ActualParent, 
                         RegelOderGleichungsT OtherParent, PCL_ObjektNummerT *pclIDPtr)
{
   if (IsDemigod(OtherParent)){
      /* tp ist aneinandergehaengtes Paar von Flatterms */
      BTP_Tp2Par(tp, lhsPtr, rhsPtr, ActualParent, OtherParent, pclIDPtr);
   }
   else{
      long int Stelle = ptr2ul(tp);
      PCL_StartFeyChain();
      if (Stelle < 0){
         /* Stelle in OtherParent */
         U1_KPRekonstruieren(OtherParent, -Stelle, ActualParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, OtherParent, -Stelle, ActualParent);
      }
      else{
         /* Stelle in ActualParent */
         U1_KPRekonstruieren(ActualParent, Stelle, OtherParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, ActualParent, Stelle, OtherParent);
      }
   }
}


void IMP_TP2ParFinal (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT ActualParent, 
                      RegelOderGleichungsT OtherParent, PCL_ObjektNummerT *pclIDPtr)
{
   if (IsDemigod(OtherParent)){
      /* tp ist aneinandergehaengtes Paar von Flatterms */
      BTP_Tp2Par(tp, lhsPtr, rhsPtr, ActualParent, OtherParent, pclIDPtr);
   }
   else{
      long int Stelle = ptr2ul(tp);
      PCL_StartChain();
      if (Stelle < 0){
         /* Stelle in OtherParent */
         U1_KPRekonstruieren(OtherParent, -Stelle, ActualParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, OtherParent, -Stelle, ActualParent);
      }
      else{
         /* Stelle in ActualParent */
         U1_KPRekonstruieren(ActualParent, Stelle, OtherParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, ActualParent, Stelle, OtherParent);
      }
   }
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_Par2Flat (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Destroys the termpair-representation irrecoverable */
{
  if (IsDemigod(OtherParent)) 
    BTP_Par2Ft(tp, lhs, rhs, ActualParent, OtherParent);
if_pcl(
   else 
     PCL_SaveChain();                                                                    /* hier ist nix zu tun!!! */
   );
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

TermpairT IMP_Par2TP (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, 
                      RegelOderGleichungsT OtherParent, BOOLEAN changed,
                      PCL_ObjektNummerT pclID)
{
   if (IsDemigod(OtherParent)){
      /* Flatterms wieder aneinanderhaengen */
      return BTP_Par2TP(tp, lhs, rhs, ActualParent, OtherParent, changed, pclID);
   }
   else{
      /* Altes tp behalten, Flatterms killen */
      PCL_DeleteChain();
      TO_TermeLoeschen(lhs, rhs);
      return tp;
   }
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void IMP_DestroyTP  (TermpairT tp, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Termpaar zerstoeren */
{
   if (IsDemigod(OtherParent)){
      /* aneinandergehaengte Flatterms zerstoeren */
      BTP_DestroyTP(tp, ActualParent, OtherParent);
   }
   else{
      /* nix tun */
   }
}

void IMP_DestroyPar (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Termpaar und Flachterme zerstoeren */
{
   if (IsDemigod(OtherParent)){
      BTP_DestroyPar(tp, lhs, rhs, ActualParent, OtherParent);
   }
   else{
      PCL_DeleteChain();
      TO_TermeLoeschen(lhs, rhs);
   }
}




/* --------------------------------------------------------------------------------------------------------*/
/* ------- TermVerfolgung ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_TermOperationBeiVerfolgung(void (* Operation) (TermT, TermT), TermpairT tp, RegelOderGleichungsT ActualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT OtherParent)
{
   if (IsDemigod(OtherParent)){
      /* aneinandergehaengte Flatterms */
      TermT lhs, rhs;
      lhs = (TermT) tp;
      rhs = TO_Schwanz(TO_TermEnde((TermT) tp));
      TO_Schwanz(TO_TermEnde(lhs)) = NULL;
      (* Operation) (lhs, rhs);
      TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
   }
/*   else
      nix tun */
}


/* ------------------------------------------------------------------------------------------------------*/
/* ------- Unambiguous Comparison -----------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

BOOLEAN IMP_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                   TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
{
   if (TP_ElternNr(father1) > TP_ElternNr(father2)) return TRUE;
   else if (TP_ElternNr(father1) < TP_ElternNr(father2)) return FALSE;
   else                         /* father1 == father2 */
      if (TP_ElternNr(mother1) > TP_ElternNr(mother2)) return TRUE;
      else if (TP_ElternNr(mother1) < TP_ElternNr(mother2)) return FALSE;
      else{                     /* beide Elternpaare identisch */
         if (IsDemigod(mother1))
            return BTP_UnambiguousGreaterThan(tp1, father1, mother1, tp2, father2, mother2);
         else
            return (ptr2ul(tp1) > ptr2ul(tp2));
      }
}
         

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics -------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_Statistics(InfoStringT IS MAYBE_UNUSEDPARAM)
{
   ;
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Output -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void IMP_PrintTermpair (TermpairT tp, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent, const char *trenner)
/* Hier stellt sich die Frage, ob Terme berechnet werden oder nicht!!! */
{
   TermT lhs, rhs;
   PCL_ObjektNummerT bla;

   IMP_TP2ParIntermed(tp, &lhs, &rhs, ActualParent, OtherParent, &bla);
   IO_DruckeFlex("%t %s %t", lhs, trenner, rhs);
   IMP_Par2TP (tp, lhs, rhs, ActualParent, OtherParent, FALSE, bla);
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* existiert nicht */
