/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - TermpairRepresentation.c                                                                              */
/* - ************************                                                                              */
/* -                                                                                                       */
/* - Repraesentation von kritischen Paaren im Speicher.                                                    */
/* -                                                                                                       */
/* - 14.03.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

#include "compiler-extra.h"
#include "Ausgaben.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"
#include "TermpairRepresentation.h"

#include "BasicTPR.h"
#include "Implicit.h"
#include "Packen.h"

/* ------------------------------------------------------------------------*/
/* ---------- globale Variablen (siehe TermpairRepresentation.h) ----------*/
/* ------------------------------------------------------------------------*/

void (* TPR_NewFlatsGeneratedProc) (void);
TermpairT (* TPR_Flat2TPProc) (TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT, PCL_ObjektNummerT);
void (* TPR_TP2ParIntermedProc) (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT, PCL_ObjektNummerT *);
void (* TPR_TP2ParFinalProc) (TermpairT, TermT *, TermT *, RegelOderGleichungsT, RegelOderGleichungsT, PCL_ObjektNummerT *);
void (* TPR_Par2FlatProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);
TermpairT (* TPR_Par2TPProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, BOOLEAN, PCL_ObjektNummerT);
void      (* TPR_DestroyTPProc)  (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);
void      (* TPR_DestroyParProc) (TermpairT, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);
void (* TPR_TermOperationBeiVerfolgungProc) (void (*) (TermT, TermT), TermpairT, 
                                             RegelOderGleichungsT, RegelOderGleichungsT);
void (* TPR_PrintTermpairProc) (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT, const char *);
BOOLEAN (* TPR_UnambiguousGreaterThanProc) (TermpairT, RegelOderGleichungsT, RegelOderGleichungsT, 
                                            TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics -------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static void (* StatisticsProc)(InfoStringT);


void TPR_Statistics(InfoStringT IS)
{
   (* StatisticsProc)(IS);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Dummies ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


static void StatisticsDummy(InfoStringT IS MAYBE_UNUSEDPARAM){;}

static void DummyTermOperationBeiVerfolgung(void (* Operation) (TermT, TermT) MAYBE_UNUSEDPARAM, TermpairT tp MAYBE_UNUSEDPARAM,
                                            RegelOderGleichungsT father MAYBE_UNUSEDPARAM, RegelOderGleichungsT mother MAYBE_UNUSEDPARAM){;}

/* --------------------------------------------------------------------------------------------------------*/
/* -------- Initialisierung -------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN TPR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
#if !COMP_POWER
   fprintf(IO_Strom, "\n* TERMPAIR-REPRESENTATION\n\n  ");
#endif
   switch (IS_FindOption(DEF, USR, 3, "regular", "packed", "implicit")){
   case 1:                      /* packed */
#if !COMP_POWER
      fprintf(IO_Strom, "Packed. ");
#endif
      PK_InitAposteriori(DEF, USR);
      TPR_Flat2TPProc = PK_Flat2TP;
      TPR_TP2ParIntermedProc = PK_TP2Par;
      TPR_TP2ParFinalProc = PK_TP2Par;
      TPR_Par2FlatProc = PK_Par2Flat;
      TPR_Par2TPProc = PK_Par2TP;
      TPR_DestroyTPProc = PK_DestroyTP;
      TPR_DestroyParProc = PK_DestroyPar;
      TPR_PrintTermpairProc = PK_PrintTermpair;
      TPR_TermOperationBeiVerfolgungProc = DummyTermOperationBeiVerfolgung;
      TPR_UnambiguousGreaterThanProc = PK_UnambiguousGreaterThan;
      TPR_NewFlatsGeneratedProc = PK_NewFlatsGenerated;
      StatisticsProc = PK_Statistics;
      return TRUE;
      break;
   case 2:                      /* implicit */
#if !COMP_POWER
      fprintf(IO_Strom, "Implicit. ");
#endif
      IMP_InitAposteriori(DEF, USR);
      TPR_Flat2TPProc = IMP_Flat2TP;
      TPR_TP2ParIntermedProc = IMP_TP2ParIntermed;
      TPR_TP2ParFinalProc = IMP_TP2ParFinal;
      TPR_Par2FlatProc = IMP_Par2Flat;
      TPR_Par2TPProc = IMP_Par2TP;
      TPR_DestroyTPProc = IMP_DestroyTP;
      TPR_DestroyParProc = IMP_DestroyPar;
      TPR_PrintTermpairProc = IMP_PrintTermpair;
      TPR_TermOperationBeiVerfolgungProc = IMP_TermOperationBeiVerfolgung;
      TPR_UnambiguousGreaterThanProc = IMP_UnambiguousGreaterThan;
      TPR_NewFlatsGeneratedProc = IMP_NewFlatsGenerated;
      StatisticsProc = StatisticsDummy;
      return TRUE;
      break;
   case 0:
   default:                     /* Standard */
#if !COMP_POWER
      fprintf(IO_Strom, "Standard. ");
#endif
      BTP_InitAposteriori(DEF, USR);
      TPR_Flat2TPProc = BTP_Flat2TP;
      TPR_TP2ParFinalProc = BTP_Tp2Par;
      TPR_TP2ParIntermedProc = BTP_Tp2Par;
      TPR_Par2FlatProc = BTP_Par2Ft;
      TPR_Par2TPProc = BTP_Par2TP;
      TPR_DestroyTPProc = BTP_DestroyTP;
      TPR_DestroyParProc = BTP_DestroyPar;
      TPR_PrintTermpairProc = BTP_PrintTermpair;
      TPR_TermOperationBeiVerfolgungProc = BTP_TermOperationBeiVerfolgung;
      TPR_UnambiguousGreaterThanProc = BTP_UnambiguousGreaterThan;
      TPR_NewFlatsGeneratedProc = BTP_NewFlatsGenerated;
      StatisticsProc = StatisticsDummy;
      return TRUE;
      break;
   }
}

void TPR_ParamInfo(ParamInfoArtT what MAYBE_UNUSEDPARAM, InfoStringT DEF MAYBE_UNUSEDPARAM, InfoStringT USR MAYBE_UNUSEDPARAM)
{
   fprintf(IO_Strom, "\n* Termpair-Representation\n\n  ");
   fprintf(IO_Strom, "TPR_ParamInfo not yet implemented\n\n");
}
