#include "antiwarnings.h"
#include <string.h>

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "general.h"
#include "InfoStrings.h"
#include "KPMengenInterreduktion.h"
#include "NFBildung.h"
#include "Stichpunkte.h"
#include "Subsumption.h"
#include "SUEM.h"
#include "TermOperationen.h"


static int SPcompare(const void *i, const void *j)
     /* Vergleichsfunktion fuer Quicksort. */
{
   return(((StichpunktT *) i)->sp - ((StichpunktT *) j)->sp);
}

static int AnzEintraege(char* str)
     /* Zaehlt die Anzahl der in str vorkommenden Stichpunkte. */
{
   char *kopie, *token, *kopieDummy;
   int anzahl=0;
   
   if (!str) return 0;          /* String ist leer -> strtok schmiert ab */

   kopie = (char *) our_alloc(sizeof(char)*(strlen(str)+1));
   strcpy(kopie, str);
   kopieDummy = kopie;

   while((token = strtok(kopie, "."))){
      kopie = NULL;             /* gemaess Spezifikation von strtok */
      if (atol(token)>0) anzahl++;
   }
   our_dealloc(kopieDummy);
   return anzahl;
}

static void StichpunkteEintragen(char *str, int strength, StichpunktT *StichpunktListe,
                                 int *AktuellerStichpunkt)
     /* Traegt die zur Staerke strength gehoerenden Stichpunkte aus str in die SP-Liste ein. */
{
   char *kopie, *token, *kopieDummy;
   
   if (!str) return;          /* String ist leer -> strtok schmiert ab */

   kopie = (char *) our_alloc(sizeof(char)*(strlen(str)+1));
   strcpy(kopie, str);
   kopieDummy = kopie;

   while((token = strtok(kopie, "."))){
      kopie = NULL;             /* gemaess Spezifikation von strtok */
      if (atol(token)>0){
         StichpunktListe[*AktuellerStichpunkt].sp = atol(token);
         StichpunktListe[(*AktuellerStichpunkt)++].strength = strength;
      }
   }
   our_dealloc(kopieDummy);
}

#if 0
static void ListeAusgeben (StichpunktT *Liste, int Anzahl)
{
  int i;
  for(i=0; i< Anzahl; i++){
    printf("%ld, %d\n", Liste[i].sp, Liste[i].strength);
  }
  printf("\n");
}
#else   
#define ListeAusgeben(quark,joghurt)
#endif

void SP_StichpunktlisteGenerieren(InfoStringT DEF, InfoStringT USR,
                                  unsigned long int *Period, int *PeriodStrength,
                                  StichpunktT **StichpunktListe,
                                  unsigned long int *AnzahlStichpunkte,
                                  int AnzahlStaerken, const char *strengthArray[])
{
  unsigned int i,j;
  int staerke;
  char *str;
  int AktuellerStichpunkt=0, NeueStichpunktZahl=0;
  char *periodStr = IS_ValueOfOption(DEF, USR, "period");

  if (periodStr){
    *Period = atol(periodStr);
    for (staerke=0; staerke<AnzahlStaerken; staerke++){
      if (strstr(periodStr, strengthArray[staerke])){
        *PeriodStrength = staerke;
      }
    }
  }

  *AnzahlStichpunkte = 0;
  for (staerke=0; staerke<AnzahlStaerken; staerke++) {
    str = IS_ValueOfOption(DEF, USR, strengthArray[staerke]);
    if (str) {
      *AnzahlStichpunkte += AnzEintraege(str);
    }
  }

  if (*AnzahlStichpunkte > 0){
    *StichpunktListe = our_alloc(*AnzahlStichpunkte*sizeof(StichpunktT));
    for (staerke=0; staerke<AnzahlStaerken; staerke++){
      str = IS_ValueOfOption(DEF, USR, strengthArray[staerke]);
      StichpunkteEintragen(str,staerke,*StichpunktListe,&AktuellerStichpunkt);
    }

    ListeAusgeben(*StichpunktListe, *AnzahlStichpunkte);

    qsort(*StichpunktListe,*AnzahlStichpunkte,sizeof(StichpunktT),SPcompare);

    ListeAusgeben(*StichpunktListe,*AnzahlStichpunkte);

    /* Doppelte Stichpunkte eliminieren */
    /* (Liste ist bereits sortiert => doppelte folgen aufeinander) */

    NeueStichpunktZahl = *AnzahlStichpunkte;

    for(i = 0; i<*AnzahlStichpunkte; i++)
      for (j=i+1; j<*AnzahlStichpunkte; j++)
        if (((*StichpunktListe)[i].sp!=0)&&((*StichpunktListe)[i].sp == (*StichpunktListe)[j].sp)){
          (*StichpunktListe)[i].strength |= (*StichpunktListe)[j].strength;
          (*StichpunktListe)[j].sp = 0;
          NeueStichpunktZahl--;
        }
    j=0;
    for (i = 0; i<*AnzahlStichpunkte; i++){
      if ((*StichpunktListe)[i].sp!=0){
	if (i > j) {
	  (*StichpunktListe)[j].sp=(*StichpunktListe)[i].sp;
	  (*StichpunktListe)[j].strength=(*StichpunktListe)[i].strength;
	}
        j++;
      }
    }
    *AnzahlStichpunkte = NeueStichpunktZahl;
    ListeAusgeben(*StichpunktListe, *AnzahlStichpunkte);
  }
}

BOOLEAN SP_IstStichpunkt(StichpunktT *StichpunktListe, unsigned long int AnzahlStichpunkte,
                         unsigned long int punkt, unsigned long int Period, 
                         int PeriodStrength, int *strength)
{
  unsigned long int i;

  for (i=0; i<AnzahlStichpunkte; i++)
    if (StichpunktListe[i].sp==punkt){
      *strength = StichpunktListe[i].strength;
      return TRUE;
    }
  if ((Period>0) && (punkt % Period == 0)){
    *strength = PeriodStrength;
    return TRUE;
  }
  return FALSE;
}

void
SP_Ausgabe (unsigned long int Period, int PeriodStrength,
	    StichpunktT *StichpunktListe,
	    unsigned long int AnzahlStichpunkte,
	    const char *strengthArray[],
	    BOOLEAN StatistikAn)
{
  unsigned int i;

  fprintf (IO_Strom, "  Period = %lu %s, %lu poke points, statistics %s\n", \
	   Period, strengthArray[PeriodStrength], AnzahlStichpunkte, StatistikAn?"on":"off");

  if (AnzahlStichpunkte > 0)
    {
      fprintf(IO_Strom, "                ");
      for (i = 0; i < AnzahlStichpunkte; i++)
	{
	  fprintf(IO_Strom, "%ld %s, ", StichpunktListe[i].sp, strengthArray[StichpunktListe[i].strength]);
	}
      fprintf(IO_Strom, "\n");
    }
}

