/*=================================================================================================================
  =                                                                                                               =
  =  SymbolOperationen.c                                                                                          =
  =                                                                                                               =
  =  In diesem Modul sind die Operationen auf Symbolen, d.h. auf Variablen-,                                      =
  =  Konstanten- und Funktions-Symbolen, realisiert.                                                              =
  =                                                                                                               =
  =  24.02.94 Angelegt. Arnim.                                                                                    =
  =  25.02.94 1 Fehler beseitigt, IntegerVergleich. Thomas.                                                       =
  =  25.02.94 Weitergemacht. Arnim.                                                                               =
  =  04.03.94 Kommentiert, etc.                                                                                   =
  =                                                                                                               =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "general.h"
#include "InfoStrings.h"
#include "parser.h"
#include "PhilMarlow.h"
#include <stdlib.h>
#include <string.h>
#include "SpezNormierung.h"
#include "SymbolOperationen.h"


/*=================================================================================================================*/
/*= II. Globale Variablen                                                                                         =*/
/*=================================================================================================================*/

SymbolT             SO_GroesstesSpezSymb,
                    SO_GroesstesFktSymb,
                    SO_LetzteFVar,
                    SO_EQ,
                    SO_TRUE,
                    SO_FALSE,
                    SO_const1,
                    SO_const2;

unsigned int        SO_AnzahlFunktionen,
                    SO_MaximaleStelligkeit;

struct SO_SymbolInfoS *SO_SymbolInfo;
static unsigned int GroesseSymbolInfo;
#define /*unsigned int*/ AnzVariablen                                                                               \
  (GroesseSymbolInfo-(SO_LetzteFVar+1))
#define /*void*/ SymbolInfoSei(/*SymbolT*/ i, /*unsigned int*/ Stelligkeit_,                                        \
  /*BOOLEAN*/ IstEquation_, /*weight*/ Gewicht_)                                                                    \
  SO_SymbolInfo[i].Stelligkeit = Stelligkeit_,                                                                      \
  SO_SymbolInfo[i].IstEquation = IstEquation_,                                                                      \
  SO_SymbolInfo[i].IstAC = FALSE,                                                                                   \
  SO_SymbolInfo[i].DistribuiertUeber = SO_ErstesVarSymb,                                                            \
  SO_SymbolInfo[i].Gewicht = Gewicht_

BOOLEAN             SO_PraezedenzTotal;
VergleichsT         *SO_Praezedenz;
SymbolT             *SO_TotaleKette;
static weight       Variablengewicht;



/*=================================================================================================================*/
/*= IV. Symbolindizierte Felder                                                                                   =*/
/*=================================================================================================================*/

#define /*VergleichsT*/ Vorordnung(/*VergleichsT*/ Praezedenz, /*SymbolT*/ f, /*SymbolT*/ g)                        \
  Praezedenz[f*SO_AnzahlFunktionen+g]

#define /*VergleichsT*/ Praezedenz(/*SymbolT*/ f, /*SymbolT*/ g)                                                    \
  Vorordnung(Praezedenz,f,g)

static VergleichsT *Praezedenzfeld(void)
{
  SymbolT i,j;
  VergleichsT *Praezedenz = array_alloc(SO_AnzahlFunktionen*SO_AnzahlFunktionen,VergleichsT);
  SO_forFktSymbole(i)
    SO_forFktSymbole(j)
      Praezedenz(i,j) = Unvergleichbar;
  return Praezedenz;
}

static void ArraysMitFunktionsInformationenAnlegen(void)
{   /* legt die globalen Variablen SO_Stelligkeit, SO_Praezedenz etc., mit der angegebenen Anzahl von Plaetzen an. */
  GroesseSymbolInfo = SO_AnzahlFunktionen+SO_AnzErsterVariablen;
  SO_SymbolInfo = negarray_alloc(GroesseSymbolInfo, struct SO_SymbolInfoS, SO_AnzErsterVariablen);
  SO_Praezedenz = Praezedenzfeld();
  SO_TotaleKette = array_alloc(SO_AnzahlFunktionen,SymbolT);
}


void SO_VariablenfeldAnpassen(SymbolT NeuesteBaumvariable)                   /* Anpassen des Feldes SO_Stelligkeit */
{                                                                            /* und des Feldes SO_FunktionsGewicht */
  if (SO_VariableIstNeuer(NeuesteBaumvariable,SO_NummerVar(AnzVariablen))) {
    unsigned int AnzVariablenNeu = minKXgrY(SO_AnzErsterVariablen,-NeuesteBaumvariable),
                 GroesseSymbolInfoNeu = AnzVariablenNeu+(SO_LetzteFVar+1);
    struct SO_SymbolInfoS *SymbolInfoNeu = negarray_alloc(GroesseSymbolInfoNeu, struct SO_SymbolInfoS, AnzVariablenNeu);
    SymbolT i;
    SO_forSymboleAbwaerts(i,SO_LetzteFVar,SO_NummerVar(AnzVariablen))
      SymbolInfoNeu[i] = SO_SymbolInfo[i];
    negarray_dealloc(SO_SymbolInfo,AnzVariablen);
    SO_SymbolInfo = SymbolInfoNeu;
    SO_forSymboleAbwaerts(i,SO_NaechstesVarSymb(SO_NummerVar(AnzVariablen)),SO_NummerVar(AnzVariablenNeu))
      SymbolInfoSei(i, 0,FALSE,Variablengewicht);
    GroesseSymbolInfo = GroesseSymbolInfoNeu;
  }
}

SymbolT SO_VarDelta(SymbolT Variable, unsigned int Faktor)
{
  return -minKXgrY(Faktor,-Variable);
}


static void InTermErmitteln(term t)
{
  if (!variablep(t)) {
    func_symb f = term_func(t);
    SymbolT g = SN_WMvonDISCOUNT(f);
    short int i;
    SO_SymbolInfo[g].IstEquation = TRUE;
    for (i=sig_arity(term_func(t))-1; i>=0; i--)
      InTermErmitteln(term_arg(t,i));
  }
}


static void EquationSymboleErmitteln(void)
{
  eq_list e;
  for (e=get_equations(); e; e=eq_rest(e)) {
    InTermErmitteln(eq_first_lhs(e));
    InTermErmitteln(eq_first_rhs(e));
  }
} /* Gehoert bei Revision "on the fly" erledigt. */



/*=================================================================================================================*/
/*= V. Von der Praezedenz zur Totalordnung                                                                        =*/
/*=================================================================================================================*/

static void NeueFunktionenEinbauen(VergleichsT *Praezedenz)         /* ggf. Eintraege f>p und EQ>TRUE>FALSE setzen */
{
  /*if (PA_KernAn){*/  /* XXX 29.06.98 */
    SymbolT f,c;
    SO_forSymboleAufwaerts(c,SO_const1,SO_const2)
      SO_forSymboleAufwaerts(f,SO_KleinstesFktSymb,c-1)
        Praezedenz(f,c) = Groesser;
  /*}*/
  if (PM_Existenzziele()) {
    SymbolT f,p;
    SO_forSpezSymbole(f)                                                 /* Jeder Operator f aus der Spezifikation */
      SO_forParaSymbole(p)                            /* ist groesser als jeder Existenzquantifizierungsoperator p */
      Praezedenz(f,p) = Groesser;
    Praezedenz(SO_EQ,SO_TRUE) = 
      Praezedenz(SO_TRUE,SO_FALSE) = Groesser;
  }
}

static void TransitivenAbschlussBilden(VergleichsT *Praezedenz)
{ 
  BOOLEAN geaendert;
  do {
    SymbolT f,g,h;
    geaendert = FALSE;
    SO_forFktSymbole(f)                                                                /* Zeilenweise durchlaufen. */
      SO_forFktSymbole(g)
        if (Praezedenz(f,g)==Groesser)                                                                     /* f>g */
          SO_forFktSymbole(h)
            if (Praezedenz(g,h)==Groesser && Praezedenz(f,h)==Unvergleichbar) {                           /* g>h */
              Praezedenz(f,h) = Groesser;                                                               /* => f>h */
              geaendert = TRUE;
            }
  } while (geaendert);
}


VergleichsT *qcmp_Praezedenz;

static int qcmp(const void *x, const void *y)
{
  SymbolT f = *(SymbolT *)x,
          g = *(SymbolT *)y;
  switch (Vorordnung(qcmp_Praezedenz, f,g)) {
  case Groesser:
    return -1;
  case Kleiner:
    return +1;
  default:
    return 0;
  }
}

static BOOLEAN Komplettiert(VergleichsT *Praezedenz, BOOLEAN *PraezedenzTotal, SymbolT *TotaleKette)
/* Fuegt die '=' auf der Diagonalen und die zu den '>' korrespondierenden '<' ein. Dabei wird auf Zyklen getestet
   und festgestellt, ob die Praezedenz total ist.
   Wurde ein Zyklus entdeckt, so wird TRUE zurueckgeliefert.*/
{
  SymbolT i,j;
  *PraezedenzTotal = TRUE;
  SO_forFktSymbole(i)                                                /* Praezedenz-Matrix zeilenweise durchlaufen. */
    SO_forFktSymbole(j) {
      if (i==j) {                                                                                    /* Diagonale. */
        if (Praezedenz(i,j)==Unvergleichbar)                                         /* Noch nicht besetzt -> '=' */
          Praezedenz(i,j) = Gleich;
        else                                                                            /* Schon besetzt -> Fehler */
          return FALSE;
      }
      else if (Praezedenz(i,j)==Groesser) {                                                                /* i>j */
        if (Praezedenz(j,i)==Unvergleichbar || Praezedenz(j,i)==Kleiner)                               /* => j<i */
          Praezedenz(j,i) = Kleiner;
        else
          return FALSE;                                                                /* falls auch j>i => Fehler */
      }
      if (j<i && Praezedenz(i,j) == Unvergleichbar)     /* i und j bleiben nach dem Komplettieren unvergleichbar. */
        *PraezedenzTotal = FALSE;                                                /* => Praezedenz ist nicht total. */
    }
  if (*PraezedenzTotal) {
    SO_forFktSymbole(i)
      TotaleKette[i] = i;
    qcmp_Praezedenz = Praezedenz;
    qsort(TotaleKette, SO_AnzahlFunktionen, sizeof(SymbolT), qcmp);
  }
  return TRUE;                                                                         /* Kein Fehler aufgetreten. */
}



/*=================================================================================================================*/
/*= VI. Symbolinformationen und initiale Praezedenz holen                                                         =*/
/*=================================================================================================================*/

static BOOLEAN NichtGrundterm(term tt)                               /* Zum Erkennen existenzquantifizierter Ziele */
{
  if variablep(tt)
    return TRUE;
  else {
    int runny;
    for (runny=0; runny<sig_arity(term_func(tt)); runny++)
      if (NichtGrundterm(term_arg(tt,runny)))
        return TRUE;
    return FALSE;
  }
}


static void ExistenzzieleSuchen(void)
{
  eq_list ziele = get_conclusions();
  PM_Existenzziele = FALSE;
  while (ziele) {
    if (NichtGrundterm(eq_first_lhs(ziele)) || NichtGrundterm(eq_first_rhs(ziele))) {
      PM_Existenzziele = TRUE;
      return;
    }
    ziele = eq_rest(ziele);
  }
}


static void PraezedenzEinlesen(void)
/* Liest die in der Spezifikation angegebene Praezedenz ein und besetzt das Ordnungs-Array damit. */
{
  ord_chain ketten = ord_chain_list(get_ordering());                           /* Alle OrdnungsKetten durchlaufen. */
  while (ketten) {
    ord_chain aktKette = ord_first(ketten);                                  /* Aktuelle OrdnungsKette durchlaufen */
    while (aktKette) {
      ord_chain rest = ord_chain_rest(aktKette);                   /* Rest der aktuellen Ordnungskette durchlaufen */
      SymbolT Symbol1 = SN_WMvonDISCOUNT(ord_chain_first(aktKette)); 
      while (rest) {                         /* und den Vergleich mit dem ersten Element der akt. Kette eintragen. */
        SymbolT SymbolN = SN_WMvonDISCOUNT(ord_chain_first(rest)); 
        SO_Praezedenz(Symbol1,SymbolN) = Groesser;
        rest = ord_chain_rest(rest);
      }
      aktKette = ord_chain_rest(aktKette);
    }
    ketten = ord_rest(ketten);
  }
}


void SO_InitAposteriori(BOOLEAN Normiere)
{
  func_symb_list fktsliste = get_sig(),
                 p;
  SymbolT i;
  weight Konstantengewicht = UINT_MAX;
  BOOLEAN IstKBO = ord_name(get_ordering())==ordering_kbo;

  SN_SpezifikationAuswerten(Normiere);
  SO_AnzahlFunktionen = (/*PA_KernAn*/ TRUE ? 2 : 0); /* XXX 29.06.98 */
  for (p=fktsliste; p; p = sig_rest(p))                                                /* Funktionssymbole zaehlen */
    SO_AnzahlFunktionen++;
  ExistenzzieleSuchen();
  
  SO_GroesstesSpezSymb = SO_AnzahlFunktionen - 1;
  SO_AnzahlFunktionen += PM_Existenzziele ? 3 : 0;
  SO_LetzteFVar = SO_GroesstesFktSymb = SO_AnzahlFunktionen - 1;
  ArraysMitFunktionsInformationenAnlegen();
  SO_MaximaleStelligkeit = 0;
  for (p=fktsliste; p; p = sig_rest(p)) {
    func_symb f = sig_first(p);
    SymbolT F = SN_WMvonDISCOUNT(f);
    struct SO_SymbolInfoS *I = &(SO_SymbolInfo[F]);
    SymbolInfoSei(F, sig_arity(f),FALSE,IstKBO ? ord_func_weight(f) : 1);
    Maximieren(&SO_MaximaleStelligkeit, I->Stelligkeit);
    if (!I->Stelligkeit)                                   /* erste Konstante oder Konstante mit kleinerem Gewicht */
      Minimieren(&Konstantengewicht,I->Gewicht);          
  }
  EquationSymboleErmitteln();

  Variablengewicht = Konstantengewicht!=UINT_MAX ? Konstantengewicht : 1;
  SO_forSymboleAbwaerts(i,SO_ErstesVarSymb,SO_NummerVar(SO_AnzErsterVariablen))
    SymbolInfoSei(i, 0,FALSE,Variablengewicht);

  /*if (PA_KernAn) {*/    /* XXX 29.06.98 */    /* wenn andernorts, dann Vorsicht mit Feld DISCOUNTvonWM und Funktion SN_WMvonDISCOUNT */
    SO_const1 = SO_GroesstesSpezSymb-1;
    SO_const2 = SO_GroesstesSpezSymb;
    SO_forSymboleAufwaerts(i,SO_const1,SO_const2)
      SymbolInfoSei(i, 0,TRUE,Variablengewicht);
  /*}*/

  if (PM_Existenzziele()) {
    SO_EQ    = SO_GroesstesSpezSymb+1;
    SO_TRUE  = SO_GroesstesSpezSymb+2;
    SO_FALSE = SO_GroesstesSpezSymb+3;
    SO_forSymboleAufwaerts(i,SO_EQ,SO_FALSE)
      SymbolInfoSei(i, 0,FALSE,Variablengewicht);
    SO_SymbolInfo[SO_EQ].Stelligkeit = 2;
  }

  PraezedenzEinlesen();                                                             /* Praezedenz-Matrix besetzen. */
  NeueFunktionenEinbauen(SO_Praezedenz);               /* Die Spalten und Zeilen mit den neuen Konstanten besetzen */
  TransitivenAbschlussBilden(SO_Praezedenz);
  if (!Komplettiert(SO_Praezedenz, &SO_PraezedenzTotal, SO_TotaleKette))         /* die entsprechenden '<' und '=' */
    our_error("precedence is cyclic!");                          /* einbauen und auf Zykel sowie Totalitaet testen */
}

static unsigned int AnzahlFunktionsvariablen(void)
{
  unsigned int Anzahl = 0;
  func_symb_list sig;
  for (sig = get_sig(); sig; sig = sig_rest(sig))
    Anzahl += sig_is_fvar(sig_first(sig)) ? 1 : 0;
  return Anzahl;
}
  
void SO_FunktionsvariablenEintragen(void)
{
  func_symb_list sig = get_sig();
  SymbolT i;
  SO_LetzteFVar = SO_ErsteFVar+AnzahlFunktionsvariablen()-1;
  GroesseSymbolInfo = SO_LetzteFVar+1+SO_AnzErsterVariablen;
  negarray_realloc(SO_SymbolInfo, GroesseSymbolInfo, struct SO_SymbolInfoS, SO_AnzErsterVariablen);
  SO_forFktVarSymbole(i) {
    func_symb f;
    while (!sig_is_fvar(f = sig_first(sig)))
      sig = sig_rest(sig);
    SymbolInfoSei(i, sig_arity(f),FALSE,Variablengewicht);
    sig = sig_rest(sig);
  }
}


void SO_PhiUmstellen(long int *Gewichte, BOOLEAN DefaultAngegeben, weight Default)
{
  weight Konstantengewicht = UINT_MAX;
  SymbolT i;
  SO_forFktSymbole(i) {
    struct SO_SymbolInfoS *I = &(SO_SymbolInfo[i]);
    I->Gewicht = Gewichte[i]!=-1 ? Gewichte[i] : DefaultAngegeben ? Default : I->Gewicht;
    if (!I->Stelligkeit)                                   /* erste Konstante oder Konstante mit kleinerem Gewicht */
      Minimieren(&Konstantengewicht,I->Gewicht);          
  }
  Variablengewicht = Konstantengewicht!=UINT_MAX ? Konstantengewicht : 1;
  SO_forSymboleAbwaerts(i,SO_ErstesVarSymb,SO_NummerVar(SO_AnzErsterVariablen))
    SymbolInfoSei(i, 0,FALSE,Variablengewicht);
}



/*=================================================================================================================*/
/*= VII. Praezedenz aus Kommandozeile lesen und auswerten                                                         =*/
/*=================================================================================================================*/

static BOOLEAN PraezedenzAnalyse(char *ps, BOOLEAN PraezedenzSetzen)
{
  VergleichsT *neuePraezedenz;
  BOOLEAN neuePraezedenzTotal;
  SymbolT *neueTotaleKette;
  
  /* 1. Neues Praezedenz-Array anlegen + initialisieren (eq,false,true beachten) */
  neuePraezedenz = Praezedenzfeld();
  /* Der Einbau der Existenzquantifizierungsfunktionen erfolgt nur in der initialen Praezedenz.
     Gibt man in der Kommandozeile eine Praezedenz an oder laesst eine solche durch Permutation oder 
     Ordnungsgenerator erzeugen, so muessen sie explizit angegeben werden. */

  /* 2. Neue Praezedenz dort eintragen (Fehler -> freigeben, return FALSE) */
  do {
    SymbolT GroesseresSymbol = SO_ErstesVarSymb;
    do {                                                                      /* Start einer neuen Praezedenzkette */
      char *NaechsterPunkt = strchr(ps,'.');
      SymbolT Symbol;
      if (NaechsterPunkt)
        *NaechsterPunkt = '\0';
      SO_forFktSymbole(Symbol)
        if (!strcmp(ps,IO_FktSymbName(Symbol)))
          break;
      if (Symbol>SO_GroesstesFktSymb){
        our_dealloc(neuePraezedenz);
        return FALSE;
      }
      if (SO_SymbUngleich(GroesseresSymbol,SO_ErstesVarSymb))
        Vorordnung(neuePraezedenz,GroesseresSymbol,Symbol) = Groesser;
      GroesseresSymbol = Symbol;
      if (NaechsterPunkt) {
        *NaechsterPunkt = '.';
        ps = NaechsterPunkt+1;
      }
      else
        ps += strlen(ps);
    } while (*ps && *ps!='.');
    ps++;
  } while (*(ps-1));

  /* 3. Trans. Abschl. bilden */
  TransitivenAbschlussBilden(neuePraezedenz);

  /* 4. Komplettieren (Fehler -> freigeben, return FALSE) */
  neueTotaleKette = array_alloc(SO_AnzahlFunktionen,SymbolT);
  if (!Komplettiert(neuePraezedenz, &neuePraezedenzTotal, neueTotaleKette)){
    our_dealloc(neuePraezedenz);
    our_dealloc(neueTotaleKette);
    return FALSE;
  }

  /* 5. Ggf. neue Praezedenz setzen */
  fprintf(IO_Strom, "\n");
  if (PraezedenzSetzen) {
    our_dealloc(SO_Praezedenz);
    SO_PraezedenzTotal = neuePraezedenzTotal;
    SO_Praezedenz = neuePraezedenz;
    our_dealloc(SO_TotaleKette);
    SO_TotaleKette = neueTotaleKette;
  }
  else {
    our_dealloc(neuePraezedenz);
    our_dealloc(neueTotaleKette);
  }
  return TRUE;
}  


BOOLEAN SO_PraezedenzKorrekt(char *ps)
{
  return PraezedenzAnalyse(ps,FALSE);
}

BOOLEAN SO_PraezedenzErneuert(char *ps)
{
  return PraezedenzAnalyse(ps,TRUE);
}

    
/*=================================================================================================================*/
/*= VIII. Cleanup und Statistik                                                                                   =*/
/*=================================================================================================================*/

void SO_CleanUp(void)
{
  negarray_dealloc(SO_SymbolInfo,AnzVariablen);
  our_dealloc(SO_Praezedenz);
}


void SO_Statistik(InfoStringT IS)
{
  if (!(IS_IsIn("all", IS) || IS_IsIn("SO", IS)))
    return;
  if (IS_IsIn("intern", IS))
    IO_DruckeStatistikZeile("Max. Expansion of SO_SymbolInfo", GroesseSymbolInfo);
}
